/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */


#include <iostream.h>
#include <LiDIA/nmbrthry_functions.h>
#include <LiDIA/bigint_matrix.h>
#include <LiDIA/quadratic_order.h>
#include "arith.h"
#include "fields.h"
#include "matrix.h"
#include "sparse.h"
#include "vector.h"
#include "charpoly.h"
#include "my_ntl.h"


int next_prime(int p) 
{ 
    return to_int(next_prime(bigint(p))) ; 
}

bigint classno(const bigint& D)
{
    quadratic_order QO(D);
    return QO.class_number();
}


bigint nfdisc(const Vector<bigrational>& f)
{
    polynomial<bigint> g;
    bigint d=to_polybigint(g, f);
    if(d!=1) {
	cerr << "f="; f.polprint(); cerr<<endl;
	cerr << "denom = " << d << "!= 1"<<endl;
	die("nfdisc: non-1 denom not yet implimented.");
    }
    number_field K(g);
    order O(K);
    return disc(O);
}



bigint nth_prime(int n) 
{
#ifdef DEBUG
    if(n<=0) {cerr<<"nth_prime called with n="<<n<<endl; die();}
#endif
    bigint p=2;
    while((n--)>1) p=next_prime(p);
    return p;
}

void power(int& a, int b, int n) 
{
    bigint aa,bb(b),nn(n);
    power(aa,bb,nn);
#ifdef DEBUG
    if(!is_int(aa)) {
	cerr << "power:: overflow" << b << "^" << n << " is not an int."<<endl;
	exit(-1);
    }
#endif
    aa.intify(a);
}

bigint sqrtmod (bigint a, bigint p) 
{
    galois_field field(p, 1);
    gf_p_base base(field);
    gf_p_element alp(base);
    Fp_polynomial f ;
    f.set_modulus(p);
    f.assign_one ();
    f[0] = a;
    alp.set_polynomial (f) ;
    gf_p_element b;
    sqrt(b,alp);
    return (b.polynomial_rep())[0];
}

				// returns some b s.t. b^2=a (mod p^n)
bigint
sqrtmod (bigint a, bigint p, int n) 
{
    bigint pn; power(pn,p,n); 
    a %= pn; if (a<0) a += pn;
    
#ifdef DEBUG
    if (!is_prime(p)) {
	cerr << "sqrtmod error: Non-prime p = " << p <<endl;
	exit(-1);
    }
    if (n<=0) {
	cerr << "sqrtmod error: negative power n = "<<n<<endl;
	exit(-1);
    }
    if (p==2) {
	cerr << "sqrtmod error: only implimented for odd primes p." <<endl;
	exit(-1);
    }    
    if (jacobi(a,p)==-1) {
	cerr << "sqrtmod error: a = " << a << " is not a square "
	     << "modulo p = " << p << endl ;
	exit(-1);
    }
#endif
    if (a==0) return 0; if (a==1) return 1;
    
    // Step 1) Find an answer modulo p
    bigint b=sqrtmod(a,p);
    
    // Step 2) Solve modulo p^n by using the following computation:
    //    b^2=a(mod p^k) 
    //    (b+cp^k)^2=b^2+2bcp^k+p^(2k)=b^2+2bcp^k=a(mod p^(k+1))
    //    <==> c= (a-b^2)/(2bp^k),
    //           this last division makes sense because p^k divides a-b^2
    //           and p!=2, and p doesn't divide b.
    //     In any case, 
    //           b+cp^k = b + (a-b^2)/(2b) is a square root of a mod p^(k+1).
    for(int k=1; k<n; k++) {
	bigint x; power(x, p, k+1);	// x=p^(k+1)
	b = b + (a-b*b)*invmod(2*b,x);
    }
    b %= pn; if (b<0) b+=pn;
    
#ifdef DEBUG
    if ((b*b-a)%pn != 0) {
	cerr << "sqrtmod error: ERROR IN PROGRAMMING."<<endl;
	exit(-1);
    }
#endif
    return b;
}

int lcm(int a, int b)
{
    bigint A(a), B(b);
    bigint C=lcm(A,B);
    if(!is_int(C)) {
	cerr << "lcm: overflow."<<endl;
	exit(-1);
    }
    int ans; 
    C.intify(ans); 
    return ans;
}

bool is_approx_zero(const bigcomplex& a)
{
    return(abs(a)<1e-10);	// perhaps a bad choice?
}




				// prop. 8.2 of hungerford's algebra.
				// pari is at least twice as fast, so
				// this is probably not an optimal algorithm
				// but for most applications it should be fine.
polynomial<bigint> 
polcyclo(const int& n) 
{
    polynomial<bigint> f;
    if (n==1) {
	f.set_degree(1);
	f[1]=1; f[0]=-1;		// f=x-1
	return f;
    } else if (n==2) {
	f.set_degree(1);
	f[1]=1; f[0]=1;
	return f;
    } else if (is_prime(n,100)) {
	f.set_degree(n-1);
	for(int i=0; i<=n-1; i++)
	    f[i]=1;
	return f;
    } else {
	f.assign_one();
	f.set_degree(n);
	f[n]=1; f[0]=-1;
	sort_vector<bigint> d=divisors(n);
	polynomial<bigint> g;
	g.assign_one();
	for (int i=0; i<d.get_size()-1; i++) {// iterate over proper divisors.
	    int div; d[i].intify(div);
	    g *= polcyclo(div);
	}
	return f/g;
    }
}

// adapted by was from JEC arith.h :
// finds x, y, such that ax+by = gcd(a,b)
int 
Bezout(int aa, int bb, int& xx, int& yy)
{
    int a = aa, b = bb, x = 0, oldx = 1, y = 1, oldy = 0, q;
    while( b!=0 ) {
	q = a/b; 
	int c    = a    - q*b; a    = b; b = c;
	int newx = oldx - q*x; oldx = x; x = newx;
	int newy = oldy - q*y; oldy = y; y = newy;
    }
    if (a<0) { xx = -oldx; yy = -oldy; return(-a); }
    else     { xx = oldx;  yy = oldy;  return( a); }
}

bigint binom(const bigint& n, const bigint& m) 
{
    if (n==0 || m==0) return 1;
    bigint b=n;
    for(int i=1;i<m; i++)
	b*=(n-i);
    return b/factorial(m);
}

bigint factorial(const bigint& n) 
{
    bigint a=1;
    for(bigint i=2; i<=n; i++) a*=i;
    return a;
}


int 
invmod(int a, int p)
{				// Not the most efficient, but avoids
				// overflow.  This is not a time critical
				// part of the client code. 
    bigint ans=invmod(bigint(a),bigint(p));
    int i_ans;
    ans.intify(i_ans);
    return i_ans;
}


bigint invmod(const bigint& a, const bigint& modulus)
{
    bigint ainv, dummy;
    if (xgcd(ainv, dummy, a, modulus) != 1) {
	cerr << "invmod called with " << a 
	     << " and " << modulus << " -- not coprime!\n" << endl;
	exit (0) ;
    }
    ainv%=modulus;
    if(ainv<0)ainv+=modulus;
    return ainv;
}

int 
phi (int N) 
{
    bigint ans=eulerphi(N);
    int a; ans.intify(a); return a;
    
    /*
      int a = 1 ;
      
      while (N > 1) {
      for (int p = 2; p <= N ; p++)    // find smallest prime divisor
      if (N%p == 0) {
      // multiply a by (p^n - p^(n-1)), n=ord_p(N).
      while (N%p == 0) {
      a *= p ;
      N /= p ;
      }
      a -= a/p ;
      break ;   // out of for loop.
      }
      }
      
      return a ;
      
      // it takes about 10 seconds to do all phi's from 10000 to 20000 on a P200.
    */
}

				// this might be a stupid algorithm.
int znorder (int a, int N)
{
    if (gcd(N,a)!=1) return 0 ;
    if ((a-1)%N==0) return 1;
    sort_vector<bigint> divs = divisors(phi(N));
    for (int i=0; i<divs.size(); i++) {
	int u;
	divs[i].intify(u);
	if (powmod(a,u,N)==1)
	    return u;
    }
}

				// perhaps slow, but avoids overflow.
int 
powmod (int a, int r, int N) 
{
    bigint ans;
    power_mod(ans, a, r, N) ;
    int i_a; ans.intify(i_a);
    return i_a;
}

int ordp (int p, int n, int& prime_to_p) 
{
#ifdef DEBUG
    if (n==0) {
	cerr << "ordp("<<p<<","<<n<<") : answer is infinite."<<endl;
	exit(-1);
    }
    if (!is_prime(p)) {
	cerr << "ordp, p="<<p<<" is not prime."<<endl;
	exit(-1);
    }
#endif
    
    int ord = 0;
    while (n % p == 0) {
	ord ++ ;
	n /= p ;
    }
    prime_to_p = n ;
    return ord ;
}

int ordp(int p, int n)
{
    int dummy; 
    return ordp(p,n,dummy); 
}

int num_prime_divs(const bigint& N) 
{
    rational_factorization f(N);
    f.factor();
    return f.no_of_comp();
}

int num_divs(const bigint& N)
{
    if(N<0) return -num_divs(N);
    if(N==0) { die("num_divs: 0 has infinitely many divisors!");}
    if(N==1) return 1;
    sort_vector<bigint> divs = divisors(N);
    return divs.size();
}

void Divisors(Vector<bigint>& divs, const bigint& N)
{
    sort_vector<bigint> d = divisors(N);
    divs.init(d.get_size());
    for (int i=0; i<d.get_size(); i++) 
	divs[i+1]=d[i];
    for (int i=1; i<divs.size(); i++) // make sure order is decreasing.
	if(divs[i]>=divs[i+1]) {
	    cerr << "divs = " << divs << endl;
	    die("Divisors -- order is not increasing!");
	}
}


bigint smallest_prime_divisor(const bigint& N)
{
    if(N<1) die("smallest_prime_divisor: N<1");
    if(N==1) return 1;
    rational_factorization f(N);
    f.factor();
    return f.base(0);
}

void Factor(Vector<bigint>& base, Vector<bigint>& exp, const bigint& N)
{
    rational_factorization fac(N);
    fac.factor();
    base.init(fac.no_of_comp());
    exp.init(fac.no_of_comp());
    for(int i=0;i<fac.no_of_comp();i++){
	base[i+1]=fac.base(i);
	exp[i+1]=fac.exponent(i);
    }
}


bigint eulerphi (const bigint& N)
{
    if (N==1) return 1;
    rational_factorization fac(N);
    fac.factor();
    bigint ans=1;
    for(int i=0;i<fac.no_of_comp();i++){
	bigint pn;
	power(pn,fac.base(i),fac.exponent(i)); 
	ans *= pn-pn/fac.base(i);
    }
    return ans;
}

void abcd (bigint& a, bigint& b, bigint& c, bigint& d,
	   const bigint& uu, const bigint& vv, const bigint& N) 
{
    c=uu; d=vv;
    if(c==0) c+=N; if(d==0) d+=N;
    bigint m=c, g ;
    while ((g=gcd(m,d)) != 1) { m/=g; }
    while ((g=gcd(N,m)) != 1) { m/=g; }
    d += N*m ;
    xgcd(b,a,-c,d);
    if(a*d-b*c != 1) die("abcd: ERROR!!!");
}


/* LIFTING ALGORITHM:  Let (a,b;c,d) in SL_2(Z/LZ).  Lift c,d to coprime
   u,v such that A=(x,y;u,v) in SL_2(Z).  Multiply (a,b;c,d) by A^{-1} so
   we *assume* (a,b;c,d)=(a,b;0,1).  Since det=1, a=1, so a lift is (1,b;0,1).
   Now multiply by A to obtain the original lift.  */
                         
void LiftSL2modN(bigint& aa, bigint& bb, bigint& cc, bigint& dd,
		 const bigint& a, const bigint& b, const bigint& c, const bigint& d,
		 const bigint& L)
{
    if(((a*d-b*c)-1) % L != 0)
	die("LiftSL2modN: det != 1.");
    bigint x,y;
    abcd(x,y,cc,dd,  c,d,L);
    if (((a*dd-b*cc)-1)%L != 0) die("LiftSL2modN: algorithm error.");
    aa=x+cc*(-a*y+b*x);
    bb=y+dd*(-a*y+b*x);
    if ((aa*dd-bb*cc != 1) ||
	(aa-a)%L !=0 || (bb-b)%L != 0 || (cc-c)%L != 0 || (dd-d)%L != 0)
	die("LiftSL2modN: major algorithm error!");
}



Vector<bigint> CRTdata(const int& N)
{
    rational_factorization fac(N);
    fac.factor();
    // CRT data
    Vector<bigint> CRT;
    CRT.init(fac.no_of_comp());
    for(int i=0;i<fac.no_of_comp();i++){
	bigint pn;
	power(pn,fac.base(i),fac.exponent(i)); 
	CRT[i+1]=invmod(bigint(N)/pn, pn)*(bigint(N)/pn) % N; 
    }
    return CRT;
}


void contfrac(Vector<bigint>& v,
	      const bigint& aa, const bigint& bb)
{
    bigint a=aa, b=bb, q, r;
    int i=0; 
    v.init(10);
    do {
	i++;
	div_rem(q, r, a, b);
	if(v.size()<i) v.resize(i+i/2);
	v[i]=q;
	a=b; b=r;
    } while(r!=0);
    v.resize(i);
}

bigrational convergent(const Vector<bigint>& v, const int n)
{
    int i=n;
    bigrational x=v[i];
    i--;
    while(i>=1) {
	x = v[i] + 1/x;
	i--;
    }
    return x;
}

bool issquarefree(const bigint& N)
{
    if(N<=1) return false;
    rational_factorization fac(N);
    fac.factor();
    for(int i=0;i<fac.no_of_comp();i++)
	if(fac.exponent(i)>1) return false;
    return true;
}

bool issquare(bigint& root, const bigint& N)
{
    if(N<0) return false;
    if(N==0 || N==1) {
	root=N;
	return true;
    }
    sqrt(root,N);
    return (root*root==N);
}

bigint int_sqrt(const bigint& N)
{
    if(N<0) 
	die("int_sqrt: N is negative");
    if(N==0 || N==1) return N;
    bigint a;
    sqrt(a,N);
    return a;
}

///////////////////////////////////////////////////////

// Let V and W be Z-submodules of a fixed Q^n, and assume that 
// W is contained in V.  This function returns the index 
//                     [V:W] = #(V/W).
// THE COLUMNS of V and W are assumed linearly independent!
// Representation: "V" and "W" are given as the Z span of the columns 
// of V and W.  The columns are ASSUMED linearly independent. 
// Algorithm: 
//   (1) Write the elements of W in terms of V,
//   (2) Since the index [V:W] is invariant under change of variables we may 
//       think of V as being the standard basis (1,0,0,...),
//       More formally, an isomorphism tau:V-->V' induces an isomorphism
//       on the quotient groups   V/W --> V'/tau(W). 
//   (3) Now we see that, with the identification of part (2), and having
//       written W in terms of V, that the index is the volume of W.
// What this all boils down to is 
//             [V:W] = abs(det(V^{-1}*W)).
// Note, this is a priori a good number to compute because if A is an invertible 
// matrix then
//      abs(det((AV)^{-1}*(AW))) = abs(det(V^{-1}A^{-1}AW))=abs(det(V^{-1}*W)).
// Also, this agrees with the "module index" as defined in Cassels-Frohlich,
// as W*V^{-1} is a linear transformation which sends V to W. 

bigrational ModuleIndex(const Matrix<bigrational>& V,
			const Matrix<bigrational>& W)
{
    if(V.ncols()!=W.ncols()) return 0;
    Matrix<bigrational> A;
    V.augment(A, W);
    Vector<int> elimcol, elimrow;
    A.SimpleEchelon(elimcol, elimrow); // assumes proceeds from left to write.

//    cout << "A = " << A << endl;
//    for(int i=1;i<=elimcol.size();i++)
//	cout << elimcol[i] << " " ;
//    cout<<endl;

    for(int i=1;i<=V.ncols();i++)
	if(elimcol[i]==0) 
	    die("ZModIndex -- V has lin. dep. columns!");
    for(int i=V.ncols()+1; i<=elimcol.size(); i++)
	if(elimcol[i]!=0) 
	    die("ZModIndex -- W is not even contained in the Q span of V!");
    Matrix<bigrational> B(V.ncols(), V.ncols());
    for(int r=1;r<=B.nrows();r++)
	for(int c=1;c<=B.ncols();c++)
	    B(r,c)=A(elimcol[r],c+V.ncols());
    return abs(Determinant(B));
}


void Vector_denominator(bigint& d,
			const Vector<bigrational>& v)
{
   // find lcm of denominators in M.
    d = 1;
    for (int i=1 ; i<=v.size() ; i++)
	    if (v[i] != 0)
		d = lcm (d, denominator(v[i]) ) ;
    d = abs(d);	
}

void Matrix_denominator(bigint& d,
			const Matrix<bigrational>& M)
{
   // find lcm of denominators in M.
    d = 1;
    for (int r=1 ; r<=M.nrows() ; r++)
	for (int c=1 ; c<=M.ncols() ; c++) 
	    if (M(r,c) != 0)
		d = lcm (d, denominator(M(r,c)) ) ;
    d = abs(d);			// necessary???
}

void Matrix_bigintify(bigint_matrix& A, bigint& d,
		      const Matrix<bigrational>& M)
{
  Matrix_denominator(d,M);
    A.resize(M.nrows(),M.ncols());
    for (int r=0 ; r<M.nrows(); r++)
	for (int c=0; c<M.ncols(); c++) {
	    A(r,c) = numerator(M(r+1,c+1) * d);
	}
}

void Matrix_bigintify(bigint_matrix& A,
		      const Matrix<bigint>& M)
{
    A.resize(M.nrows(),M.ncols());
    for (int r=0 ; r<M.nrows(); r++)
	for (int c=0; c<M.ncols(); c++) 
	    A(r,c) = M(r+1,c+1);
}


void _ZBasis(Matrix<bigrational>& B,
	    const Matrix<bigrational>& Gen)
{
    if(Gen.iszero()) {
	B.init(Gen.nrows(),0);
	return;
    }
    bigint_matrix A;
    bigint d;
    Matrix_bigintify(A,d,Gen);
    bigint_matrix Im;
    Im.image(A);
    B.init(Im.get_no_of_rows(),Im.get_no_of_columns());
    for(int r=1; r<=B.nrows(); r++)
	for(int c=1; c<=B.ncols(); c++)
	    B(r,c)=bigrational(Im(r-1,c-1))/bigrational(d);
    mesg(".");
}

// Break into 1/16'ths -- reduce each, then breaks into 16ths, 8ths, 4th, 1/2s, etc.
void ZBasis(Matrix<bigrational>& B,
	    const Matrix<bigrational>& G)
{
    mesg("[Lattice ZBasis P=");
    int n=G.nrows();
    Matrix<bigrational> Gen=G;
    int P=16, c;
    while(P>=2 && (c=Gen.ncols())>P) {
	mesg(P);
	Matrix<bigrational> L[P];
	int bd[P+1];
	bd[0]=1; bd[P]=c+1;
	for(int i=1;i<P;i++) 
	    bd[i]=i*c/P+1;
	int Lsize=0;
	for(int i=0;i<P;i++) {
	    Matrix<bigrational> C(n,bd[i+1]-bd[i]);
	    for(int j=1;j<=bd[i+1]-bd[i];j++)
		C.setcol(j,Gen.col(bd[i]+j-1));
	    _ZBasis(L[i],C);
	    Lsize+=L[i].ncols();
	}
	Gen.init(n,Lsize);
	int pos=1;
	for(int i=0;i<P;i++) 
	    for(int j=1;j<=L[i].ncols();j++)
		Gen.setcol(pos++,L[i].col(j));
	P/=2;
    }
    mesg("final");
    _ZBasis(B,Gen);    
    mesg("]\n");
}



bigrational LatticeVol (const Matrix<bigrational>& L) 
{
    mesg("[LatticeVol...");
//    cout << "L = " << L << endl;
    Matrix<bigrational> B;
    ZBasis(B,L);
    if(B.ncols()!=B.nrows()) 
	die("LatticeVol: input is NOT a lattice, i.e., doesn't span the space (or overflow error!");
    mesg("]\n");
    return abs(Determinant(B));
}

bigrational AlgNumNorm(const Vector<bigrational>& a,
		       const Vector<bigrational>& g)
{
    if(g.poldeg()<2) return a[1];
    bigint l=1;
    for (int i=1; i<=a.size(); i++)
	if (a[i] != 0)
	    l=lcm(l,denominator(a[i]));
    Vector<bigrational> al=bigrational(l)*a;
    bigint ln; power(ln,l,g.poldeg());
    polynomial<bigint> A,T;
    to_polybigint(A,al);
    to_polybigint(T,g);
    // resultants can be used to quickly compute norms.
    return bigrational(resultant(T,A))/bigrational(ln);	
    
				// see Cohen, section 4.3.
}

bigint PolDisc(const Vector<bigrational>& g)
{
    polynomial<bigint> f;
    bigint d=to_polybigint(f,g);
    if(d!=1) {
	cout << "g = " << g << endl;
	die("PolDisc: doesn't lie in Z[x]!");
    }
    bigint ans=discriminant(f);     
    return ans;
}

bigrational FieldDisc(const Vector<bigrational>& f)
{
    mesg("[FieldDisc (warning, uses number_field)");
    Vector<bigrational> g=f;
    // normalize g.
    bigint d=1;
    for(int i=1;i<=g.size();i++)
	d=lcm(d,denominator(g[i]));
    g*=d;
    // make integral
    // a0+a1*x+...+an*x^n |--> an^(n-1)*(a0+...+an*x^n) |--> an^(n-1)a0+an^(n-2)*a1*x+...+x^n
    int n=g.poldeg();
    bigint a=numerator(g.polgetcoeff(n));
    if(denominator(g.polgetcoeff(n)) != 1) die("ERROR!");
    bigint b=a;
    g.polcoeff(n)=1;
    for(int i=n-2;i>=0;i--) {
	g.polcoeff(i)*=b;
	b*=a;
    }
    polynomial<bigint> h;
    to_polybigint(h,g);
    number_field F(h);
    order O(F);
    return(disc(O));
}

bigint ZAlgDisc(const Vector<bigrational>& f,
		const Matrix<bigrational>& E)
{
    Matrix<bigrational> tE;
    E.transpose(tE);
    bigrational Vol=LatticeVol(tE);
    bigrational ans=Vol*Vol*PolDisc(f);
    if(denominator(ans)!=1) {
       cerr << "E = " << E <<endl;
       cerr << "Vol = " << Vol << endl;
       cerr << "PolDisc(f) = " << PolDisc(f)<<endl;
       cerr << "ans = " << ans << endl;
       die("ZAlgDisc -- error computing discriminant.");
    }
    return numerator(ans);
}

void SpanningColumns_modp (Vector<int>& cols,
			   const Matrix<bigrational>& T)
{
    mesg("[SpanningColumns");
    // [Step 1. Compute small prime not dividing 
    //          the denominators of any element of T.]
    int p=next_prime(random()%4096);
    for(int i=1;i<=T.nrows();i++)
	for(int j=1;j<=T.ncols();j++)
	    while(denominator(T(i,j)) % p == 0) 
		p=next_prime(p);
    
    int oldp=intmod::get_pr();
    if(oldp<0||!is_prime(oldp)) oldp=2;
    intmod::set_pr(p);
    Matrix<intmod> Tmod;
    to_intmod(Tmod,T);
    Vector<int> elimcol, elimrow;
    Tmod.SimpleEchelon(elimcol, elimrow);

    int j=0;
    for(int i=1;i<=Tmod.ncols();i++)
	if(elimcol[i]!=0) j++;
    cols.init(j);
    j=1;
    for(int i=1;i<=Tmod.ncols();i++)
	if(elimcol[i]!=0) cols[j++]=i;
    mesg("\n]");

    intmod::set_pr(oldp);
}


#include "display.h"

bigrational ZModDisc(Vector<bigrational>& v, bool& badrank,
		     const Matrix<bigrational>* T, int n) 
{
    mesg("[ZModDisc");
/* Algorithm: 1. Choose a random vector v in Q^m so that probably
                 dim(sp(T1*v,...,Tn*v)) = m and the map R|-->Q^m 
                 defined by A|->A*v is then injective.
              2. Working mod p find Ti1,...,Tim so that 
                 Ti1*v,...,Tim*v span Q^m.  If this fails, go to step 1.
     [steps 3 and 4 may be done in parallel.]
              3. Compute the matrix B=(tr(Tij*Tik)) (use symmetry to half work).
              4. Use Hermite normal form to compute a Z-basis
                 w1,...,wm for R viewed as contained in Q^m.
              5. Write each wi in terms of the Tij*v:
                  [Ti1*v ... Tim*v | w1 ... wm]
                  [ identity matrix | wi in terms of Tij*v ]
                  [ I | A ]
                 Then (<wi,wj>) = A^t*B*A so output
                  disc(R) = det(A^t*B*A) = det(A)^2*det(B). 
               Trick: Let W=[w1...wm] and Ti=[Ti1*v...Tim*v].  
                Then A=Ti^{-1}*W, so 
                  disc(R) = (det(W)/det(Ti))^2*det(B).
                
*/
    badrank=false;
    if(n<=0) {mesg("invalid n=",n); die("ZModDisc: n must be >0");}
    int m=T[0].nrows();
    
    // [Step 1.  Choose random vector.]
    Vector<int> cols;
    Matrix<bigrational> Tv(m,n);
    int failsafe=0;
    do {
	v.init(m);
	v.random(100);

	// [Step 2.  Find spanning set.]
	for(int i=1;i<=n;i++)
	    Tv.setcol(i,T[i-1]*v);
	SpanningColumns_modp(cols, Tv);
	if(cols.size()<m) {
	    mesg("ZModDisc: Warning, random vector probably doesn't define an injection.");
	    failsafe++;
	    if(failsafe>50) {
		mesg("\n** The Ti almost certainly span a space of too SMALL a dimension! ");
		badrank=true;
		return 0;
	    }
	}
    } while (cols.size()<m);
    if(cols.size()>m) 
	die("ZModDisc: The Ti span a space of too large dimension.");

    // [Step 3. Compute B = tr(Tij*Tik).]
    mesg("[compute B]");
    Matrix<bigrational> B(m,m);
    for(int i=1;i<=m;i++) 
	for(int j=i;j<=m;j++) 
	    B(i,j)=B(j,i)=TraceOfProduct(T[cols[i]-1],T[cols[j]-1]);

    // [Step 4. Compute Z-basis.]
    mesg("[compute Lattice Zbasis]");
    Matrix<bigrational> W;
    ZBasis(W, Tv);

    // [Step 5. Compute disc.]
    Matrix<bigrational> Ti(m,m);
    for(int i=1; i<=m; i++) 
	Ti.setcol(i,Tv.col(cols[i]));
    
//    cout << "Tv = " << Tv << endl;
//    cout << "B = " << B <<endl;
//    cout << "W = " << W << endl;
    mesg("[compute dets]");
    bigrational dW=W.determinant(),
	        dTi=Ti.determinant(), dB=B.determinant();
//    cout << "dW = " << dW << endl
//	 << "dTi= " << dTi << endl
//	 << "dB = " << dB << endl;
    bigrational a=dW*dW*dB/(dTi*dTi);

/*
    mesg("(");
    DisplayFactored(numerator(a));
    if(denominator(a)!=1) {
	cout<<"/";
	DisplayFactored(denominator(a));
    }
*/
    mesg("finished ZModDisc]\n");
    return dW*dW*dB/(dTi*dTi);
}


#include "NTL/ZZX.h"

void CharPolyMod(Vector<bigrational>& chi,
		 const Vector<bigrational>& f,
		 const Vector<bigrational>& g)
{
    if(g.polgetcoeff(g.poldeg()) != 1)
	die("CharPolyMod: g is not monic!");
    ZZX  A, B;
    bigint a, b;
    to_ZZX(A,a,f);   // A = a*f
    to_ZZX(B,b,g);   // B = b*g
    // g is assumed to have integer coefficient
    if(b!=1)
	die("CharPolyMod: the coefficients of g are not integers!");
    ZZX C;
    CharPolyMod(C, A, B, 1);	// 1 = deterministic
    to_bigrationalvec(chi, C);
    if(chi.poldeg()!=g.poldeg())
	die("CharPolyMod: charpoly should have the same degree as g!");
    // renormalize, i.e., note that f = A/a, so the correct charpoly is
    // X^n + a*a_{n-1}*X^{n-1} + ... + a^(n-1)*a_1*X + a^n. 
    bigint c=a;
    for(int i=chi.poldeg(); i>=1; i--) {
	chi[i]/=c;
	c *= a;
    }
}

bigrational TracePolyMod(const Vector<bigrational>& f,
			 const Vector<bigrational>& g)
{
    Vector<bigrational> chi;
    CharPolyMod(chi, f, g);
    return -chi.polcoeff(chi.poldeg()-1);
}


// TorsionReduce:  A is assumed to have no denominators!!

/*
void TorsionReduce(Matrix<bigrational>& A,
		   const bigint& p)
{
  if(A.nrows()!=A.ncols())
    die("TorsionReduce: A is not square.");

  bigmod_matrix B(A.nrows(), A.ncols(),p);
  for(int r=0;r<A.nrows();r++)
    for(int c=0;c<A.ncols();c++)
      B(r,c)=A(r+1,c+1).numerator();

  mesg("Removing the prime p = ");
  cout<<p<<endl;
  bigmod_matrix K;
  cout << "B = " << B << endl;

  cout << "start computing kernel." << endl;
  K.kernel(B);
  cout << "done computing kernel." << endl;
 
  cout << "Finished, kernel = " << K << endl;
  if(K.get_no_of_columns()<1)
    die("TorsionReduce: Error -- rank is full mod p!");
  int i=K.get_no_of_rows();
  while(i>=1&&K(i-1,0)==0) i--;
  if(i==0) 
    die("TorsionReduce: Error finding kernel mod p!!");

  Vector<bigrational> w(K.get_no_of_rows());
  bigint a=invmod(K(i-1,0),p);
  for(int j=1;j<=K.get_no_of_rows();j++) {
    bigint wj;
    remainder (wj, K(j-1,0)*a,p);
    w[j]=wj;
  }
  
  cout << "w = " << w << endl;
  Vector<bigrational> v=A*w;
  v/=bigrational(p);
  A.setcol(i,v);

  cout << "A = " << A << endl;

}
*/

int RankModP(const Matrix<bigrational>& A,
	      const bigint& p)
{
  if(A.nrows()!=A.ncols())
    die("TorsionReduce: A is not square.");

  ZZ_p::init(to_ZZ(p));  
  mat_ZZ_p B; 
  B.SetDims(A.ncols(), A.nrows());

  for(int r=1;r<=A.nrows();r++)
      for(int c=1;c<=A.ncols();c++) {
	  if(A(r,c).denominator()!=1)
	     die("RankModP: error, nontrivial denominator.");
	  conv(B(c,r),to_ZZ(A(r,c).numerator()));
      }

  mat_ZZ_p K;
  kernel(K,B);
  return K.NumRows();
}


void TorsionReduce(Matrix<bigrational>& A,
		   const bigint& p)
{
  if(A.nrows()!=A.ncols())
    die("TorsionReduce: A is not square.");

  ZZ_p::init(to_ZZ(p));  
  mat_ZZ_p B; 
  B.SetDims(A.ncols(), A.nrows());

  for(int r=1;r<=A.nrows();r++)
    for(int c=1;c<=A.ncols();c++)
      conv(B(c,r),to_ZZ(A(r,c).numerator()));

  mesg("TorsionReduce p = ");
  cout<<p<<"  "<<flush;
  mat_ZZ_p K;
  //  cout << "B = " << B << endl;

  //  cout << "start computing kernel." << endl;
  kernel(K,B);
  cout << "rank="<<K.NumRows()<<endl;
  //  cout << "Finished, kernel = " << K << endl;
  int i=K.NumCols();
  while(i>=1&&K(1,i)==0) i--;
  if(i==0) 
    die("TorsionReduce: Error finding kernel mod p!!");
  for(int j=1;j<=K.NumCols();j++)
    K(1,j)/=K(1,i);
  //  cout << "scaled K = " << K  << endl;

  Vector<bigrational> w(K.NumCols());
  for(int j=1;j<=K.NumCols();j++) 
    w[j] = to_bigint(rep(K(1,j)));
  
  //  cout << "w = " << w << endl;
  Vector<bigrational> v=A*w;
  v/=bigrational(p);
  A.setcol(i,v);
  //  cout << "A = " << A << endl;
}


void TorsionFreeQuotientMatrix (Matrix<bigrational>& B,
				const Matrix<bigrational>& M)
{

  Matrix<bigrational> K, Mt;
  M.transpose(Mt);
  Kernel(K, Mt);
  
  bigint denom;
  Matrix_denominator(denom,K);
  if(denom!=1) {
    K*=denom;
  }

  M.augment(B,K);

  //cout << "First B = " << B << endl;
  bigrational det=abs(B.determinant());
  cout << "TorsFreeQuotient: det = " << det << endl;
  bigint d = det.numerator();

  if(det.denominator()!=1)
    die("TorsionFreeQuotientMatrix: denom(B)!=1 !!\n");
  
  if (d==0)
    die("TorsionFreeQuotientMatrix: rank(B)==0!!\n");
  if (d==1)
    return;

  rational_factorization fac(d);
  fac.factor();
  for (int i=0; i<fac.no_of_comp(); i++) 
    for (int j=0; j<fac.exponent(i); j++) 
      TorsionReduce(B,fac.base(i));

/*  cout << "good B = " << B << endl;

  Matrix<bigrational> A;
  K.init(B.nrows(),B.ncols()-M.ncols());
  for(int r=1; r<=K.nrows(); r++)
     for(int c=1; c<=K.ncols(); c++)
       K(r,c)=B(r,c+B.ncols()-M.nrows());

  ZBasis(A,K);
  for(int r=1; r<=K.nrows(); r++)
     for(int c=1; c<=K.ncols(); c++)
       B(r,c+B.ncols()-M.ncols())=A(r,c);

  cout << "reduced B = " << B << endl;
*/
}

// THIS IS A SOLID WORKING VERSION.
void IntegerKernel(Matrix<bigrational>& K,
		   const Matrix<bigrational>& A)
{
    mesg("Integer Kernel of ",A.nrows());
    mesg("x",A.ncols()); mesgln(" matrix.");
    if(A.iszero()) {    // program around *stupid* design of LiDIA:
	K.idmat(A.ncols());
	return;
    }
    
    bigint_matrix B, C;
    bigint d;
    Matrix_bigintify(B,d,A);	// B = d*A.
    C.kernel(B);			// C = ker(B)
    to_bigrationalmat(K,C);	// 
    mesgln("Integer kernel computed.");
}


/*
// EXPERIMENTAL VERSION.
void IntegerKernel(Matrix<bigrational>& K,
		   const Matrix<bigrational>& A)
{
    mesg("Integer Kernel of ",A.nrows());
    mesg("x",A.ncols()); mesgln(" matrix.");
    if(A.iszero()) {    // program around *stupid* design of LiDIA:
	K.idmat(A.ncols());
	return;
    }
    
    Matrix<bigrational> L;
    ntl_cheapLLL(L, A);

    bigint_matrix B, C;
    bigint d;
    Matrix_bigintify(B,d,L);	// B = d*A, but kernel(dA)=ker(A), so who cares

    C.kernel(B);			// C = ker(B)
    to_bigrationalmat(K,C);	// 
    mesgln("Integer kernel computed.");
}
*/


void IntegerKernel(Matrix<bigrational>& K,
		   const Matrix<bigrational>& A,
		   const Vector<bigrational>& f)
{
    Matrix<bigrational> B;
    A.polsubst(B,f);
    IntegerKernel(K,B);
}

void Saturate(Matrix<bigrational>& S,
	      const Matrix<bigrational>& A)
{
    Matrix<bigrational> At,B;
    A.transpose(At);
    Kernel(B,At);
    B.transpose();
    IntegerKernel(S,B);
}


void ElementaryDivisors(Vector<bigint>& d,
			const Matrix<bigrational>& A)
{
    bigint_matrix B;
    bigint denom;
    if(A.nrows()!=A.ncols()) {
	Matrix<bigrational> C;
	int n=max(A.nrows(),A.ncols());
	C.init(n,n);
	for(int r=1;r<=A.nrows();r++)
	    for(int c=1;c<=A.ncols();c++)
		C(r,c)=A(r,c);
	Matrix_bigintify(B,denom,C);
    } else 
	Matrix_bigintify(B,denom,A);

    if(denom!=1) 
	die("ElementaryDivisors: input matrix not integral.");

    B.snf_simple();
//    cout << "A = " << A << endl;
//    cout << "B = " << B << endl;
    d.init(A.nrows());
    for(int i=1;i<=d.size();i++)
	d[i]=B(i-1,i-1);
//    cout << "d = " << d << endl;
}


const char* letter(int i)
{
    static bool in_str=false;
    static char* str=NULL;

    if(in_str) delete str;
    if(i<=0) { str=new char[1]; str[0]='\0'; return str; }
    if(i<=26) {
	str=new char[2];
	str[0]=char(i+64);
	str[1]='\0';
	return str;
    }
    if(i<=26*26) {
	i--;
	str=new char[3];
	str[0]=char(i/26+64);
	str[1]=char(i%26+65);
	str[2]='\0';
	return str;
    }
    
    die("letter, i is too large.");
}

int parse_letter_to_num(const char* str)
{
    // TODO
    return int(str[0]-'A')+1;	// actually will need to parse two letter strings as well!!
}
