/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/* 
   FILE: arith.h
   DESC: arithmetic functions  
*/

#ifndef ARITH_H
#define ARITH_H

#include <LiDIA/lidia.h>
#include <LiDIA/bigint.h>
#include <LiDIA/polynomial.h>

#include "vector.h"

// returns the least integer number strictly greater than a that satisfies 
// the primality test in function is_prime(a). For any input a <=1 
// this function returns 2. 
int next_prime(int p);

// true iff N has no square factor.
// false if N<=1.
bool issquarefree(const bigint& N);

// true iff N is a perfect square.
bool issquare(bigint& root, const bigint& N);

// returns integer part of square root
bigint int_sqrt(const bigint& N);

// returns the class number of the order of discriminant D.
bigint classno(const bigint& D);

// returns the discriminant of the number field defined by f(x)
bigint nfdisc(const Vector<bigrational>& f);

bigint nth_prime(int n);

void power(int& a, int b, int n); // a=b^n.

bigint sqrtmod (bigint a, bigint p) ;

// returns some square root of a modulo a prime power p^n, when
// one exists.  Otherwise returns 0.
bigint sqrtmod (bigint a, bigint p, int n) ;

Vector<bigint> CRTdata(const int& N);

bigint binom(const bigint& n, const bigint& m) ;
bigint factorial(const bigint& n) ;

void Factor(Vector<bigint>& base, Vector<bigint>& exp,
	    const bigint& N);

// returns a sorted vector of all positive divisors of N.
// the order is decreasing, e.g., N=12 then divs=[12, 6, 4, 3, 2, 1].
void Divisors(Vector<bigint>& divs, const bigint& N);

int lcm(int a, int b);

bool is_approx_zero(const bigcomplex& a);

				// computes the nth cyclotomic polynomial.
polynomial<bigint> polcyclo(const int& n) ;


//int gcd(int a, int b) ;


bigint invmod(const bigint& a, const bigint& modulus) ;

int Bezout(int aa, int bb, int& xx, int& yy);

// The following two functions are "guaranteed"
// not to give valid output as long as it will fit 
// in an int.  The internal computations are done 
// using higher precision data types to insure valid output. 
int invmod(int a, int modulus) ;
int powmod (int a, int r, int N) ;   // a^r % N
int znorder (int a, int N); // order of a modulo N. 

inline int intmax(int a, int b) {   return (a>b)?a:b; }
inline int intmin(int a, int b) {   return (a<b)?a:b; }
int phi (int N) ;
bigint eulerphi (const bigint& N);
int ordp (int p, int n, int& prime_to_p) ;

int num_prime_divs(const bigint& N) ;

int num_divs(const bigint& N);

bigint smallest_prime_divisor(const bigint& N);

// returns nearest integer to a/b.  if a/b = d.5 then rounds up
// if d>=0 and rounds down otherwise. 
inline int nearest_integer (int a, int b) {
  return (2*a)/b - a/b ;
  // uses that * / * is the integer part of the quotient. 
}

inline int floor (int a, int b) {
  return a/b - ((a*b<0 && a%b) ? 1 : 0) ;
}

inline int ceil (int a, int b) {
  if (a%b==0) return a/b;
  return floor(a,b)+1;
}

// rounds a/b to nearest whole number, rounding up to break ties. 
inline int round (int a, int b) { 
  return floor(2*a,b) - floor(a,b) ;
}

inline unsigned int factorial (unsigned int n) {
  return (n>1) ? n*factorial(n-1) : 1 ;
}

inline int sign(int a) { return a>0 ? 1 : -1 ; }

void contfrac(Vector<bigint>& v,
	      const bigint& a, const bigint& b);

// Compute (a,b;c,d) in SL_2(Z) so that (c,d)=(u,v) (mod N).
// Assumes gcd(u,v,N)=1.
void abcd (bigint& a, bigint& b, bigint& c, bigint& d,
	   const bigint& u, const bigint& v, const bigint& N) ;


/* LIFTING ALGORITHM:  Let (a,b;c,d) in SL_2(Z/LZ).  Lift c,d to coprime
   u,v such that A=(x,y;u,v) in SL_2(Z).  Multiply (a,b;c,d) by A^{-1} so
   we *assume* (a,b;c,d)=(a,b;0,1).  Since det=1, a=1, so a lift is (1,b;0,1).
   Now multiply by A to obtain the original lift.  */
                         
void LiftSL2modN(bigint& aa, bigint& bb, bigint& cc, bigint& dd,
		 const bigint& a, const bigint& b, const bigint& c, const bigint& d,
		 const bigint& L);

// Returns the nth continued fraction convergent, assuming v was computed
// using the contfrac function above. 
// Here n>=1.  If the continued fraction expansion is
//        v = [a1 a2 a3 ... ak]
// then convergent(2) is
//            a_(k-1) + 1 / ak
// and convergent(k) is a1 + 1/(a2+1/(...) ... ).
bigrational convergent(const Vector<bigint>& v, const int n);

//////////////////////////////////////////////////////////////////
// some functions for working with lattices and
// algebraic numbers. 

bigrational ModuleIndex(const Matrix<bigrational>& V,
			const Matrix<bigrational>& W);

// compute Z basis for the Z-span of 
// the *columns* of matrix Gen.
void ZBasis(Matrix<bigrational>& B,
	    const Matrix<bigrational>& Gen);

// volume of lattice spanned by the columns of L.
bigrational LatticeVol (const Matrix<bigrational>& L);

bigrational AlgNumNorm(const Vector<bigrational>& a,
		       const Vector<bigrational>& g);

void Matrix_bigintify(bigint_matrix& A, bigint& d,
		      const Matrix<bigrational>& M);
void Matrix_bigintify(bigint_matrix& A,
		      const Matrix<bigint>& M);

// assumes entries of g are bigints!!!
bigint PolDisc(const Vector<bigrational>& g) ;

// Compute the discriminant of the field Q(x)/(f(x)).  
// There is no assumption that f in Z[x]. 
bigrational FieldDisc(const Vector<bigrational>& f);


// The rows of E represents elements of the field Q[x]/(f(x)).
// They generate a subring R of this field.
// It is assumed they generate this subring as a *Z-module*!
// This function returns the discriminant of this Z-module. 
bigint ZAlgDisc(const Vector<bigrational>& f,
		const Matrix<bigrational>& E);

// Attempt to very quickly compute a set of columns which span 
// the column space of T.  The algorithm involves working modulo
// p, and the algorithm finds too few columns if p accidently 
// divides the discriminant of T.
void SpanningColumns_modp (Vector<int>& cols,
			  const Matrix<bigrational>& T);


/* Input: T = array of n square mxm matrices which generate
              a Z-module of rank m, which is also a Z-algebra
              R under matrix multiplication. 
   Output: disc(R) = det(tr(Ai*Aj)) where A1,...,Am generate
              R as a Z-module.  
           If the Ti generate a module of rank < m then 0 is returned 
           and the flag badrank is set to true!
           v is a vector which defines an injection into Q^m.
   Algorithm: 1. Choose a vector v in Q^m so that 
                 dim(sp(T1*v,...,Tn*v)) = m.  The map R|-->Q^m 
                 defined by A|->A*v is then injective.
              2. Working mod p find Ti1,...,Tim so that 
                 Ti1*v,...,Tim*v span Q^m.
     [steps 3 and 4 may be done in parallel.]
              3. Compute the matrix B=(tr(Tij*Tik)) (use symmetry to half work).
              4. Use Hermite normal form to compute a Z-basis
                 w1,...,wm for R viewed as contained in Q^m.
              5. Write each wi in terms of the Tij*v:
                  [Ti1*v ... Tim*v | w1 ... wm]
                  [ identity matrix | wi in terms of Tij*v ]
                  [ I | A ]
              6. Then (<wi,wj>) = A^t*B*A so output
                  disc(R) = det(A^t*B*A) = det(A)^2*det(B). 
 */

bigrational ZModDisc (Vector<bigrational>& v, bool& badrank,
		      const Matrix<bigrational>* T, int n);


// characteristic polynomial of f modulo g.
// g is assumed MONIC with INTEGER COEFFICIENTS.
// f is arbitary.
void CharPolyMod(Vector<bigrational>& chi,
		 const Vector<bigrational>& f,
		 const Vector<bigrational>& g);

// returns the trace of Mod(f,g).  Same assumptions on g as in CharPolyMod().
bigrational TracePolyMod(const Vector<bigrational>& f,
			 const Vector<bigrational>& g);

// Computes the rank of (A / F_p).
int RankModP(const Matrix<bigrational>& A,
	      const bigint& p);

// Input: an m x n matrix M with linearly indepedent columns
// Output: a matrix A whose last n-m columns form a basis 
//         for the _torsion free quotient_ F of V = Z^m / M.
// If V is already free then the first m columns of A will equal M.
// If not, then the first m columns of A will generate a module
// which contains A with finite index and so that the quotient
// is torsion free.

// Algorithm:  Construct the augmented matrix 
//              B = [M | K]
// where K is a basis for the Q-vector space ker(M^t)
// having entries in Z, and M^t is the transpose of M.
//
// Let d = det(B).  If d=1 or -1 set A=B and terminate.
// If p is a prime which divides d:
//    Compute a vector w in ker (B mod p), and lift
//    w to characteristic zero so that the gcd of the entries is 1.
//    Let z=B*w.  Note that z=0 (mod p).
//    Let i be the last nonzero position of z.
//    Replace the ith columnn of B by z/p.
//    This is an elementary column operation, over Z,
//     which doesn't change the rank of B. However,
//     it does divide the determinant by p, so now
//      det(B) = d/p.
// Repeat until det(B) = 1 or -1.

void TorsionReduce(Matrix<bigrational>& A,
		   const bigint& p);

void TorsionFreeQuotientMatrix (Matrix<bigrational>& A,
				const Matrix<bigrational>& M);


void Matrix_denominator(bigint& d,
			const Matrix<bigrational>& M);
void Vector_denominator(bigint& d,
			const Vector<bigrational>& v);

void IntegerKernel(Matrix<bigrational>& K,
		   const Matrix<bigrational>& A);

void IntegerKernel(Matrix<bigrational>& K,
		   const Matrix<bigrational>& A,
		   const Vector<bigrational>& f);

// Compute the saturation of the Z-module spanned by the columns of A. 
// The columns of A are not assumed linearly independent, but it is assumed
// that the dim of column space = number of rows.     The algorithm is:
//       1) B <--- the rational kernel of A transpose, then 
//       2) S <--- the integer kernel of B transpose. 
void Saturate(Matrix<bigrational>& S,
	      const Matrix<bigrational>& A);

// Let A be an nxn matrix with integer entries and of full rank.
// Then G = "Z^n/A" is a finite abelian group isomorphic to
//    Z/d[1]Z + ... + Z/d[n]Z and
//    d[i] | d[i+1] for i=1,...,n-1.
// In fact, if A isn't nxn, then the torsion part of the quotient is computed.
void ElementaryDivisors(Vector<bigint>& d,
			const Matrix<bigrational>& A);


// 1="A", 2="B", etc. the pointer is subject to change.
const char* letter(int i);
int parse_letter_to_num(const char* str);




#endif 







