/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

// array.cc
#ifndef ARRAY_CC
#define ARRAY_CC

#include "array.h"

template <class T>
Array<T>::Array( const T* ptr, int s ) : items( new T [ _size = s ] )
{
  T *p1 = items; const T *p2 = ptr;
  for( int s = 0; s < size(); s++ ) *p1++ = *p2++;
}

template <class T>
Array<T>::Array( const Array<T>& L ) : items( new T [ _size = L.size() ] )
{
  T *p1 = items, *p2 = L.items;
  for( int s = 0; s < size(); s++ ) *p1++ = *p2++;
}

template <class T>
Array<T>& Array<T>::operator=(const Array<T>& L)
{
  if( this == &L ) return *this;
  if( size() != L.size() ) {
    delete [] items;
    items = new T [_size = L.size()];
  }
  T *t1 = items, *t2 = L.items;
  for(int i = 0; i < size(); i++ ) *t1++ = *t2++;
  return *this;
}

template <class T>
void Array<T>::display(ostream& s, const int _n) const
{
  int n=_n ;
  if( n == -1 ) n = size();
  if( n > size() ) 
    error("Argument out of bounds in function display");
  s << "[";
  for(long j=0;j<n;j++) {
      s << items[j];
      if (j<n-1) s << ",";
  }
  s << "]";
}

template <class T>
void Array<T>::resize(int m)
{
  T* oldArray = items;
  int low = ( m <? size() );
  items = new T [ _size = m ];
  
  T* oa = oldArray, *na = items;
  for( int i = 0; i < low; i++ ) *na++ = *oa++;
  
  delete [] oldArray;
}

template <class T> void Array<T>::error(const char* msg) const
{
  cerr << "Error in class Array. " << msg << endl;
  // int i = 1; while(i);
  exit(-1);
}

template <class T> bool Array<T>::inList(const T& X) const
{
  T* ptr = items;
  for( int i = 0; i < size(); i++ )
    if( *ptr++ == X ) return true;
  return false;
}

template <class T> bool Array<T>::inList(const T& X, int ub, int lb ) const
{
  if( (lb < 0) || (lb > ub) || (ub > size()-1) ) 
    error("Bad bounds in Array inList function");
  
  T* ptr = items +lb;
  for( int i = lb; i <= ub; i++ )
    if( *ptr++ == X ) return true;
  return false;
}

template <class T> int Array<T>::position(const T& X) const
{
  T* ptr = items;
  for( int i = 0; i < size(); i++ )
    if( *ptr++ == X ) return i;
  return -1;
}

template <class T> int Array<T>::position( const T& X, int ub, int lb ) const
{
  if( (lb < 0) || (lb > ub) || (ub > size()-1) ) 
    error("Bad bounds in Array position function");
  
  T* ptr = items +lb;
  for( int i = lb; i <= ub; i++ )
    if( *ptr++ == X ) return i;
  return -1;
  
}

template <class type>
static void swap ( type *ia, int i, int j )
{ // swap two elements of an array
  type tmp = ia[ i ];
  ia[ i ] = ia[ j ];
  ia[ j ] = tmp;
}

template <class T> void Array<T>::qsort( int high, int low )
  // sorts array 
  //  using quick sort algorithm
{
#ifndef UNORDERED
  if ( low < high ) {     // stopping condition for recursion
    int lo = low;
    int hi = high + 1;
    T elem = items[ low ];
    
    for (;;) {
      items[0] < items[1] ;
      while ( (lo < high) && (items[ ++lo ] < elem) ); //Luiz in 03/12/94
      while ( elem < items[ --hi ] ) ;
      
      if ( lo < hi ) {
	swap( items, lo, hi );
      }
      else break;
    }
    
    swap( items, low, hi );
    qsort( hi-1,low);
    qsort( high,hi + 1);
  } 
#endif
}

#endif // ARRAY_CC
