/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

//  array.h

#ifndef ARRAY_H
#define ARRAY_H

#include <iostream.h>
#include <LiDIA/lidia.h>

template <class T>
class ArrayLooper ;

template < class T >
class Array 
{
    friend class ArrayLooper<T>;
 protected:
    T *items;
    int _size;
public:
    Array( int m = 0 );
    Array( const Array<T>& );
    Array( const T* ptr, int size );
    virtual ~Array( );
    
    Array& operator = (const Array<T>& );
    T& operator [] ( int n );          // check bounds of array 0 <= n < size
    const T& operator [] ( int n ) const;
    T& operator () ( int n ) { return operator[](n-1); }
    const T& operator () ( int n ) const { return operator[](n-1); }
    void set(int,const T&);            // does not check bounds
    const int length() const;          // size of array
    const int size() const { return _size; }
    
    // resizing array

    void grow ();
    void resize( int );                // resizes and copies contents
    void reset( int );                 // resizes forgetting the contents
    void init(int i) { reset(i); }
    
// other functions and operators 

    bool inList(const T& X) const;
    bool inList( const T& X, int ub, int lb = 0) const;  // see note below
    int position(const T& X) const;
    int position(const T& X, int ub, int lb = 0) const;
    void qsort();
    virtual void qsort(int ub, int lb = 0); // sorts array between lb and ub
				// using the quicksort algorithm 
    // output functions

    void display(ostream& s = cout, const int n = -1) const; // display 
			 // first n elements, with n = -1 displays whole array
    friend ostream& operator<< (ostream&, const Array<T>&);
    
    void    error(const char* msg) const;     // error handling
};

/* Function bool Array::inList(X,ub,lb) searches the element X in the array
 * between the limits lb (lower bound) and ub (upper bound). Returns true
 * if found and false otherwise. Default is to search all the array.
 *
 * Function int Array::position(X,ub,lb) does the same but returns
 * position of X in array(starts from 0). Returns -1 if X not in array.
 */

template <class T>
class ArrayLooper 
{
public:
    ArrayLooper( const Array<T>& L ) { ArrayRef = &L; index = 0; }
    int OK() { return index < ArrayRef->size(); }
    T next() { return ArrayRef->items[index++]; }
    void reset() { index = 0; }
private:
    const Array<T>* ArrayRef;
    long index;
};

// inlined array member and friend functions

template <class T> 
inline Array<T>::Array(int m) : items ( new T [_size = m] ) {}    

template <class T> inline Array<T>::~Array()
{
    delete [] items;
}

template <class T> inline void Array<T>::grow()
{
    resize(size()+size()/2+1);
}

template <class T> inline T& Array<T>::operator [] (int n)
{
    if( n < 0 || n >= size() ) error( "argument for operator[] out of bounds");
    return items[n];
}

template <class T> inline const T& Array<T>::operator [] (int n) const
{
    if( n < 0 || n >= size() ) error( "argument for operator[] out of bounds");
    return items[n];
}

template <class T> inline void Array<T>::reset( int n )
{
    delete [] items;
    items = new T [_size = n];
    for(int i=0; i<n; i++)
	items[i]=T(0);
}

template <class T> inline const int Array<T>::length() const
{
    return _size; 
}

template <class T> inline void Array<T>::set(int n, const T& X) 
{
//    if( n < 0 || n >= size ) error( "argument for operator[] out of bounds");
    items[n] = X;
}

template <class T> inline 
ostream& operator<< (ostream& s, const Array<T>& A)
{
    A.display(s);
    return s;
}

template <class T> inline void Array<T>::qsort()
{
    qsort(size()-1,0);
}
#endif
