/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */
//   character.cc


#include <sys/stat.h>       
#include <sys/types.h>       
#include <fcntl.h>       
#include <unistd.h>
#include <stdio.h>
#include <LiDIA/bigrational.h>
#include "character.h"
#include "arith.h"
#include "fields.h"
#include "display.h"
#include "system.h"


void
Character::init_all ()
{
  // nothing to do.
}

void 
Character::init_trivial ()
{
  CHI = new rootof1[N] ;
  for (int i = 0 ; i < N ; i++)
    CHI [i] = gcd(i,N) == 1 ? 1 : 0 ;
}

// compute b=(1,1,1,1,1,a,1,1,...,1,1,1) in prod_p Z/p^n = Z/N
int lift(int a, int N, Vector<bigint>& CRT, int pos)
{
  bigint A=0;
  for (int i=1; i<=CRT.size(); i++) 
    if (i!=pos)
      A=(A+CRT[i])%N;
    else
      A=(A+(a*CRT[i]))%N;
  int ans; A.intify(ans); 
  if (ans<0) ans+=N;
  return ans;
}

// degs
void 
Character::init_arbitrary ()
{
  // 0. initialize things: * CHI[i]=0 for all i s.t. gcd(i,N)!=1, but set
  //    equal to 0 for all for efficiency.
  //            * factor N. 
  // 1. compute smallest even n such that character maps to <zeta_n>.
  // 2. compute generators for each odd prime factor storing in an array.
  // 3. iterate over (Z/NZ)^* computing value of character.

  // Step 0.
  fac = rational_factorization(N);
  fac.factor();
  int npd = fac.no_of_comp();

  CHI=new rootof1[N];
  for(int i=0;i<N;i++) CHI[i]=0;
  
  // Step 1. 
  bigint bign=1;
  for (int i=0; i<degrees.length(); i++)
    bign = lcm(bign,degrees[i]);
  bool is_odd=false;		// only used for printing degree of chi.
  if (bign%2 != 0) { is_odd = true ;  bign *= 2 ; } 
				// n *must* be even,
				// so we can represent -1.
  int n; bign.intify(n);
  rootof1::set_prim (n) ;
  rootof1 zeta_n ;    
  deg = is_odd ? n/2:n ;
  mesg("Chi is a character of degree ",deg);mesg(".\n");
  
  // CRT data
  Vector<bigint> CRT = CRTdata(N);

  // Step 2.
  int numfactors = npd + ((N%8==0)?1:0);
  Array<bigint> gens(numfactors), ords(numfactors);
  int j=0;
  if (N%8==0) {
    gens[0]=lift(-1,N,CRT,1); gens[1]=lift(5,N,CRT,1); j=2;
    ords[0]=2; 
    power(ords[1],2,fac.exponent(0)-2);
    if ((degrees[0]!=1 && degrees[0]!=2) || ords[1]%degrees[1]!=0) {
      cerr << "Character::init -- Invalid character at 2."<<endl;
      valid = false;
      return;
    }
  } else if (N%4==0) {
    gens[0]=lift(-1,N,CRT,1); j=1;
    ords[0]=2;
    if (degrees[0]!=1 && degrees[0]!=2) {
      cerr << "Character::init -- Invalid character at 2."<<endl;
      valid = false;
      return;
    }
  } else if (N%2==0) {
    gens[0]=lift(1,N,CRT,1); j=1;
    ords[0]=1;
  }

  for(int i=0; i<npd; i++) {    
    bigint bi_pn, bi_pnone;
    if(fac.base(i)==2) continue; // we've done this case above. 
    power(bi_pn,fac.base(i),fac.exponent(i));
    power(bi_pnone,fac.base(i),fac.exponent(i)-1);
    ords[j]=bi_pn-bi_pnone;
    int pn; bi_pn.intify(pn);
    if (ords[j]%degrees[j] != 0) {
      cerr << "Character::init -- Invalid character at "
	   << fac.base(i) << "." << endl ;
      valid = false;
      return;
    }
    
    for (int g=2; g<pn; g++) 
      if(bigint(znorder(g,pn))==ords[j]) {
	gens[j++]=lift(g,N,CRT,i+1); break; 
      }
  }
  if (j!=numfactors) {
    cerr << "Error making character.  Please enter the right "
	 << "number of degrees."<<endl; 
    valid=false;
    return;
  }
  
  // Step 3. 
  Array<int> cur(numfactors);
  for(int i=0; i<numfactors; i++)
    cur[i]=0;
  for (int i=0; i<phi(N); i++) {
    bigint c=1, temp;
    rootof1 zeta(1);
    for (int j=0; j<numfactors; j++) {
      power(temp, gens[j], cur[j]);
      c = (c*temp) % N;
      zeta *= zeta_n ^ (n*cur[j]/degrees[j]);
      
    }
    // compute next cur.
    int m=0;
    cur[0]++;
    while(cur[m]>=ords[m]) { // carry.
      cur[m]=0; m++; if (m<cur.length()) cur[m]++; else break; 
    }    
#ifdef DEBUG
    if (c < 0 || c >= N) {
      cerr << "Error computing image.  (Programming Error)" << endl;
      exit(-1);
    }
#endif
    int cc; c.intify(cc);
    CHI[cc] = zeta;
  }
}

void
Character::init_chip () 
{
  // CRT data
  Vector<bigint> CRT=CRTdata(N);

  CHIP = new (rootof1*)[fac.no_of_comp()];
  for (int i=0; i<fac.no_of_comp(); i++) {
    bigint bi_pn; power(bi_pn,fac.base(i),fac.exponent(i));
    int pn; bi_pn.intify(pn);	// pn=p^n||N.
    CHIP[i] = new rootof1[pn];
    for(int a=0; a<pn; a++) 
      // lift() = (1,1,1,1,1,a,1,1,...,1,1,1) in prod_p Z/p^n = Z/N
      CHIP[i][a]=of(lift(a,N,CRT,i+1));
  }
}

void 
Character::name (char* buf, int len) const
{
  ostrstream ost(buf,len);
  ost<<_type;
  if (_type > 0) 
    for (int i = 1 ; i < degrees.length() ; i++)
      ost<<"_"<<degrees[i];
  ost<<'\0';
}

void 
Character::ppname (char* buf) const
{
  if (_type >= 0) {
    ostrstream ost(buf,512);
    ost<<"Gamma_0("<<N;
    if(!is_prime(N)) {
      ost<<"=";
      DisplayFactored(N,ost);
    }
    ost<<")"<<'\0';

    if(_type>0) {
      sprintf (buf+strlen(buf), ", eps=[") ;
      for (int i = 0 ; i < degrees.length() ; i++) {
	if(i==0&&N%2==0&&N%4!=0) continue; // annoying special case
	sprintf (buf+strlen(buf), "%d", degrees[i]) ;
	if (i+1<degrees.length()) 
	  sprintf(buf+strlen(buf), " ") ;
	else
	  sprintf(buf+strlen(buf), "]") ;
      }
    }
  } else {
    sprintf (buf, "Gamma_1(%d)",N) ;
  }
}


bool
Character::load() 
{
  if (!do_save) return false ;

  char buf[600];
  ostrstream ost(buf,600);
  ost<<name()<<"_Chi" << '\0';
  ifstream in(data_read(buf,N));  
  if (!in) { 
     mesg("The character ");
     mesg(buf);
     mesg("\nhasn't yet been created, so we are now creating it.\n");
     return false;
  }

  mesg("Loading ");
  mesgln(buf);

  int prim ;
  in >> prim ;
  rootof1::set_prim(prim) ;

  in >> deg;
  mesg("The character has degree ",deg);
  mesgln(".");

  int k ;
  for (int a=0;a<N ;a++) {
    in >> k ;
    CHI[a].set_k(k) ;
  }
  fac = rational_factorization(N);
  fac.factor();
  return true;
}

void 
Character::save() const 
{
  char buf[600];
  ostrstream ost(buf,600);
  ost<<name()<<"_Chi" << '\0';
  ofstream os(data_write(buf,N));  
  if(os) {
      mesg("[saving level ",N); mesg(" character "); mesg(buf); mesg("]\n");
  } else {
      mesg("[couldn't save level ",N); mesg(" character "); mesg(buf); mesg("]\n");
  }

  os << rootof1::get_prim() << endl ;
  os << deg << endl;
  for (int a=0;a<N ;a++)
    os << CHI[a].get_k() << endl ; 
}
  

Character::Character (bool _save) 
  : do_save (_save), CHI(NULL), CHIP(NULL)
{  
  do {
    N=get_int("level N = ",1);
  } while (N<1);
  if (N>1 && !is_prime(N)) {
    cout << "("<<N<<" = "; DisplayFactored(N); 
    cout << ")" << endl;
  }

  cout << "-1=Gamma_1, 0=Gamma_0, -2=help."<<endl;
  do {
    _type=get_int("character chi = ",-2);
    if(_type==-2) {
      cout << "For nontrivial characters, enter a space separated list of"<<endl
	   << "the degrees of the images of the factors corresponding to the"<<endl
	   << "primes which divide the level.  Not all characters can be entered"<<endl
	   << "in this way, but most of interest can.  There is some subtlety at"<<endl
	   << "2 because (Z/2^n)^* need not be cyclic.  If 2|N but 4 does not divide"<<endl
	   << "N, ignore 2.  Otherwise, give either 1 or 2 degrees for 2, if 4 or"<<endl
	   << "8 divides N, respectively."<<endl;
    }      
  } while(_type==-2);
  
  do {
    if (_type > 0) {
      int npd=num_prime_divs(N);
      if (N%8==0) npd++;
      degrees.reset(npd);
      int i=1;
      if (npd>0) {
	if (N%2==0 && N%4!=0) {
	  i=2;
	  degrees[0] = 1;
	  if (N>2) 
	    degrees[1]=_type;
	}
	else
	  degrees[0]=_type;
      }
      //      cout << "Enter " ; 
      //      if (degrees.length()>1) cout << degrees.length() << " degrees: ";
      //      else cout << "the degree of chi: ";
      for (; i < degrees.length() ; i++) 
	degrees[i]=get_int(NULL,1);
    }
    init();
    if (!is_valid()) 
      _type=get_int("Please re-enter chi.",-2);
  } while (!is_valid());
}

Character::Character (int _N, int t, bool _save,
		      Array<int>* p_degrees, bool)
  : N(_N), _type (t), do_save(_save), CHI(NULL), CHIP(NULL)
{
  if(t>0) {
    if (p_degrees==NULL) {
      cerr << "Invalid input to character constructor.  Please " << endl 
	   << "specify the degrees." << endl ;
      exit(-1);
    }
    degrees=*p_degrees;
    _type = degrees[0];		// so that the character naming works uniformly,
				// the Character::name function in fact assumes
				// that _type is the the first degree.
  } else
    degrees.reset(0);
  init();
}

Character::Character(int M, const Character& chi) 
  : CHI(NULL), CHIP(NULL)
{
  N = M;
  do_save = chi.do_save;
  _type = chi._type;
#ifdef DEBUG
  if (M==0 || chi.N % M != 0) {
    cerr << "Character constructor: M must divide N."<<endl;
    exit(-1);
  }
#endif
  if (chi._type<=0) {
    init();
  } else {
    rational_factorization f = rational_factorization(chi.N);
    f.factor();
    int npd = f.no_of_comp();
    if (chi.N%8==0) npd++;
    fac=rational_factorization(N); // maybe this line should go elsewhere...
    fac.factor();
    int npdM = fac.no_of_comp();
    if (M%8==0) npdM++;
    degrees.reset(npdM);
    int posdeg=0, poschi=0;
    Array<int> cdeg = chi.degrees;
    for(int i=0;i<f.no_of_comp();i++) {
      if(M%f.base(i)==0) {
	  degrees[posdeg++] = cdeg[poschi++];
      if(f.base(i)==2 && f.exponent(i)>2) 
	  degrees[posdeg++] = cdeg[poschi++];
    } else
      poschi++;
    }
    init();
  }
}

void 
Character::init()
{
  if(N==1&&_type>=0) {
    _type=-1;
  }
  delete CHI;
  deg = 1;
  valid=true;

  name(_name,512); // record character name. 

  rootof1::set_prim (2) ;    
  CHI = new rootof1[N] ; 
  if (!load()) {
    switch(_type) {
    case -1 : // all characters
      init_all () ;
      break ;
    case 0 :  // trivial character
      init_trivial () ;
      break ;
    default : // arbitrary Dirichlet character. 

	mesg("Initializing an arbitrary Dirichlet character.\n");

      if (_type <= 0) {
	cerr << "Invalid data to Character constructor.  (type = " 
	     << _type << endl; 
	exit (-1) ;
      }
      init_arbitrary () ;
      if (valid) {
	mesg("Chi is ");
	if (of(-1)==-1) 
	    mesg("odd.\n");
	else
	    mesg("even.\n");
      }
    }
    save();
  }
}

Character::Character (const Character& chi) 
{
  N = chi.N ;
  _type = chi._type ;
  degrees = chi.degrees ;
  fac=chi.fac ;
  CHI = new rootof1[N] ;
  for (int i = 0 ; i < N ; i++) 
      CHI[i] = chi.CHI[i] ;
  if (chi.CHIP != NULL) {
      CHIP = new (rootof1*)[fac.no_of_comp()];
    for (int i=0; i<fac.no_of_comp(); i++) 
	CHIP[i] = chi.CHIP[i];
  } else
      CHIP=NULL;
  for(int i=0; i<=strlen(chi._name); i++)
      _name[i]=chi._name[i];
}

Character& 
Character::operator=(const Character& chi)
{ 
    N = chi.N ;
    _type = chi._type ;
    degrees = chi.degrees ;
    if (CHIP != NULL) 
        for(int i=0;i<fac.no_of_comp(); i++)
            delete[] CHIP[i];
    CHIP = NULL;
    fac=chi.fac ;
    delete[] CHI;
    CHI = new rootof1[N] ;
    for (int i = 0 ; i < N ; i++) 
    CHI[i] = chi.CHI[i] ;
    
    if (chi.CHIP != NULL) {
	CHIP = new (rootof1*)[fac.no_of_comp()];
	for (int i=0; i<fac.no_of_comp(); i++) 
	CHIP[i] = chi.CHIP[i];
    }
    for(int i=0; i<=strlen(chi._name); i++)
       _name[i]=chi._name[i];

    return *this;
}


Character::~Character ()
{
  delete CHI ;
  if (CHIP != NULL) {
    for(int i=0; i<fac.no_of_comp(); i++)
      delete[] CHIP[i];
    delete[] CHIP;
  }
}

rootof1 
Character::of (int a) const
{
  if (type() == -1) 
    return rootof1(a);

  a %= N ;
  if (a < 0) 
    a += N ;
  
  return CHI [a] ;
}

rootof1
Character::of (const bigint& a) const
{
  int n; a.intify(n); return of(n);
}





rootof1 
Character::ofp (const int& p, int a) 
{
#ifdef DEBUG
  if (N%p != 0) {
    cerr << "Character::ofp -- Illegal p="<<p
	 << " doesn't divide N="<<N << endl;
    exit(-1);
  }
#endif
  if (type() == -1) 
    return rootof1(a);
  if (type() == 0)
    return (a%p)?1:0;
  if (!CHIP)
    init_chip();

  int i=0;
  while (i<fac.no_of_comp() && fac.base(i)!=p) i++;
  if (i==fac.no_of_comp()) {
    cerr << "Character::ofp -- programming error occured."<<endl;
    exit(-1);
  }
  bigint b_pn;
  power(b_pn,fac.base(i),fac.exponent(i)); 
  int pn; b_pn.intify(pn);
  a %= pn ;
  if (a < 0) 
    a += pn ;
  return CHIP[i][a];

}

rootof1
Character::ofp (const int& p, const bigint& a)
{
  int n; a.intify(n); return ofp(p,n);
}


int
Character::degree (const int& p) const
{
  if (type() <= 0) return 1;
  // determine which prime divisor p is.
  int i=0;
  for (; i<fac.no_of_comp(); i++)
    if(fac.base(i)==p) break;
#ifdef DEBUG
  if(i>=fac.no_of_comp()) {
    cerr << "Character: degree -- p="<<p
	 << " doesn't divide the modulus N="<<N
	 <<" of the character."<<endl;
    exit(0);
  }
#endif
  if(p!=2) {
    if(N%8==0) i++;
    return degrees[i];
  } else
    if (N%8==0) {
      return lcm(degrees[0],degrees[1]);
    }
    else return degrees[0];
}


////////////////////////////////////////////////////
// rootof1
////////
int rootof1::n=2;
bool rootof1::init_charmod=false;
bool rootof1::init_intmod=false;
bool rootof1::init_algnum=false;
bool rootof1::init_ffield=false;
bool rootof1::init_bigcomplex=false;

bigcomplex
rootof1::to_bigcomplex() const
{
    static bigfloat PI;
    switch(k) {
    case -1: return bigcomplex(0);
    case 0: return bigcomplex(1);
    }

    if(!init_bigcomplex) {
	constant_Pi(PI);
	init_bigcomplex=true;
    }

    // k is the power of the primitive root.
    // exp(k*2*pi*I/n) 
    return exp(k*2*PI*bigcomplex(0,1)/bigcomplex(n,0));
}

charmod 
rootof1::to_charmod () const 
{
    static Vector<charmod> roots_charmod;

    switch(k) {
    case -1: return 0;
    case 0: return 1;
    }

    if(!init_charmod) {
	mesgln("\nInitializing rootof1_to_charmod.");
	int p = charmod::get_pr();
	if ((p-1) % n != 0) {
	    cerr << "ERROR: p must split completely in Z[zeta_"<<n<<"]"<<endl
		 << "in order to have a reduction map <zeta_n> --> (Z/pZ)^*." << endl 
		 << "Note, p splits completely iff p = 1 (mod N)."<<endl ;
	    exit(-1) ;
	}
	
	roots_charmod.init(n-1);
	int a = powmod (charmod::get_prim_root(), (p-1)/n, p) ;
	for(int i=1;i<n;i++)
	    roots_charmod[i] = powmod(a, i, p); 
	init_charmod=true;
    }
    
    return roots_charmod[k];
}

charmod* 
rootof1::charmod_table () 
{
  charmod* t = new charmod[n+1] ;
  t[0] = 0 ;
  rootof1 prim, z(1) ;
  for (int i = 0 ; i < n ; i++) {
    t [z.table_idx()] = z.to_charmod () ;
    z *= prim ;
  }
  return t ;
}

intmod 
rootof1::to_intmod () const
{
    static Vector<intmod> roots_intmod;

    switch(k) {
    case -1: return 0;
    case 0: return 1;
    }
    
    if(!init_intmod) {
	mesgln("\nInitializing rootof1_to_intmod.");
	int p = intmod::get_pr();
	if ((p-1) % n != 0) {
	    cerr << "ERROR: p must split completely in Z[zeta_"<<n<<"]"<<endl
		 << "in order to have a reduction map <zeta_n> --> (Z/pZ)^*." << endl 
		 << "Note, p splits completely iff p = 1 (mod N)."<<endl ;
	    exit(-1) ;
	}
	
	roots_intmod.init(n-1);
	int a = powmod (intmod::get_prim_root(), (p-1)/n, p) ;
	for(int i=1;i<n;i++)
	    roots_intmod[i] = powmod(a, i, p); 
	init_intmod=true;
    }
    
    return roots_intmod[k];
}

bigrational 
rootof1::to_bigrational () const
{
    switch(k) {
    case -1: return 0;
    case 0: return 1;
    }
    if(k==n/2) return -1;
  
  cerr << "\n\nzeta_"<<n<<"^"<<k
       <<" is not a rational number" <<endl;
  die("to_bigrational error.");
}

ffield_element
rootof1::to_ffield_element () const
{
    static Vector<ffield_element> roots_ffield;

    switch(k) {
    case -1: return 0;
    case 0: return 1;
    }
    if(k==n/2) return ffield_element(-1);

  if(!init_ffield) {
      mesg("\nInitializing rootof1_to_ffield_element...");
      if ((ffield_element::cardinality()-1) % n != 0) {
	  cerr << "\n\nThe current finite field does not contain a zeta_"
	       << n <<"^"<<k<<"."<<endl;
	  die("rootof1_to_ffield_element error.");
      }

      roots_ffield.init(n-1);

      ffield_element nthroot;
      power (nthroot, ffield_element::primitive_elem(), 
	     (ffield_element::cardinality()-1)/n);
      roots_ffield[1]=nthroot;
      for(int i=2;i<n;i++)
	  roots_ffield[i] =nthroot*roots_ffield[i-1];

      init_ffield=true;
      mesgln("Done.");
  }

  return roots_ffield[k];
}


algnum
rootof1::to_algnum () const
{
  static Vector<algnum> roots_algnum;

    switch(k) {
    case -1: return algnum(0);
    case 0 : return algnum(1);
    }

  if(!init_algnum) {
      mesgln("\nInitializing rootof1_to_algnum.");
      roots_algnum.init(n-1);
      roots_algnum[1] = algnum::Zeta(1);
      for(int i=2;i<n;i++)
	  roots_algnum[i] = roots_algnum[i-1]*roots_algnum[1];
      init_algnum=true;
  }

  return roots_algnum[k];
}


rootof1::rootof1() 
{
  k = 1 ;
}

rootof1::rootof1(const int& i) 
{
  if (i==0)
    k = -1 ;
  else if (i == 1)
    k = 0 ;
  else if (i == -1)
    k = n/2 ; 
  else {
    cerr << "Only the integers -1,0, and 1 can be cast to rootof1" << endl
	 << "not " << i << " as this program has tried to do." << endl;
    exit(-1) ;
  }
}

void 
rootof1::set_prim (int _n) 
{
  if (_n % 2 != 0) {
    cerr << "set_prim(" << _n << ") invalid, argument must be even." << endl;
    exit (-1) ;
  }
  n = _n ; 

  init_charmod = init_intmod 
      = init_algnum = init_ffield = false;
}

rootof1& 
rootof1::operator*=(const rootof1& r) 
{
  if (r.k == -1) 
    k = -1 ;
  else
    k = (k + r.k) % n ;
}

rootof1& 
rootof1::operator/=(const rootof1& r) 
{
  if (r.k == -1) {
    cerr << "Error: One can't divide a root of unity by 0." << endl;
    exit(-1) ;
  }
  k = (k - r.k) % n ;
  if (k < 0) k += n ;
}

bool 
rootof1::operator==(const rootof1& r) const
{
  return k==r.k ;
}

bool 
rootof1::operator!=(const rootof1& r) const
{
  return k!=r.k ;
}

bool 
iszero (const rootof1& r) 
{
  return r.k == -1 ;
}

rootof1 
operator- (const rootof1& r)
{
  return r * rootof1(-1) ;
}

rootof1 
operator+ (const rootof1& r) return ans (r) 
{
}

rootof1 
operator* (const rootof1& r, const rootof1& s) return ans (r) 
{
  ans *= s ;
}

rootof1 
operator/ (const rootof1& r, const rootof1& s) return ans (r) 
{
  ans /= s ;
}

rootof1 
operator^ (const rootof1& r, int n) return ans (1) 
{
  if (r.k == -1) {
    if (n != 0) 
      return rootof1(0) ;
    cerr << "rootof1:  0^0 is not defined." << endl;
    exit(-1) ;
  }
  n %= rootof1::n ;
  if (n < 0) n += rootof1::n;
  rootof1 t = r ;
  while (n >= 1) {
    if (n % 2) 
      ans *= t ;
    t *= t ;
    n /= 2 ;
  }
}

ostream& 
operator<< (ostream& s, const rootof1& z)
{
  if (z.k == -1) 
    s << "0" ;
  else if (z.k == 0)
    s << "1" ;
  else if (z.k == z.n/2)
    s << "-1" ;
  else
    if (z.k != 1)
      s << "(zeta_" << z.n << ")^" << z.k;
    else
      s << "zeta_" << z.n ;
  return s;
}
