/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */
//   character.h

#ifndef CHARACTER_H
#define CHARACTER_H

#include <LiDIA/bigrational.h>

#include "fields.h"
#include "charmod.h"
#include "intmod.h"
#include "vector.h"
#include "arith.h"

class rootof1 ;

class Character 
{
  char _name[512];
  int N ;
  int _type ;
  bool do_save;
  Array<int> degrees ;
  int deg;
  bool valid ;			// false iff something went wrong making chi.
  int num_gens ;
  rootof1* CHI ;
  rootof1** CHIP;
  rational_factorization fac;
  void init() ;
  void init_all () ;
  void init_trivial () ;
  void init_arbitrary () ;
  void init_chip () ;
  void save() const ;
  bool load();
  
public:
  Character (const Character& chi) ;    

				// If type=0 then the character is trivial.
				// If type=-1 then the character is interpreted
				// by other objects as "all characters", i.e.,
				// is used for computing the full space of
				// modular forms on Gamma_1(N).
				// If type=1 then the characters is nontrivial:
  // The vector p_degrees gives the degrees of the images of the
  // factors of (Z/NZ)^*=prod (Z/p^nZ)^*, where if N is divisible by
  // 8 then there are two factors corresponding to p=2, generated by
  // -1 and 5.   THE FOLLOWING REFERS TO SOMETHING NOT YET IMPLIMENTED!
  // Though this is a very convenient way to specify characters, 
  // not every character, up to automorphism, can be represented in this
  // way.  If the power_not_degrees flag is set to true then the integers
  // in the array pointed to by p_degrees give the power of zeta that a 
  // generator of each factor of (Z/NZ)^* maps to. If n is the least
  // common multiple of the orders of the factors in the product,
  // multiplied by 2 if this number is odd, then zeta is a prim. n-th root
  // of unity. 
  Character (int N, int type=0, bool save=false,
	     Array<int>* p_degrees=NULL, 
	     bool powers_not_degrees=false) ; 
  // construct this from chi by restricting to Z/MZ.  
  Character (int M, const Character& chi) ;
  
  Character (bool save=false) ;
  ~Character () ;
  Character& operator=(const Character&) ;
  int modulus () const { return N ; } 
  const rational_factorization& factors() const { return fac ; } 
				// When chi is ALLCHARS then this just
				// returns rootof1(a).  Note that a *must*
				// be either 1 or -1 in that case. 
  rootof1 of (int a) const ;
  rootof1 of (const bigint& a) const ;

  //     As sets,  Z/p^n \--> Z/NZ = prod_{q|N} (Z/q^n)
  //                a |----------> (a at p and 1 at all other factors).
  //     ofp(a) = of(image of a).
  rootof1 ofp (const int& p, int a) ;
  rootof1 ofp (const int& p, const bigint& a) ;
  int type () const { return _type ; }
  void name (char* buf, int len=512) const ;  
  const char* name() const { return _name; }
  void ppname (char* buf) const ;  // sizeof(buf) = 512 characters. pretty print

				// order of image of chi.
  int degree () const { return deg ; }
  int degree (const int& p) const ;
  bool is_valid() const { return valid ; }
} ;


// M encodes the character 
// (NOTE: not *all* characters can be
//        encoded in this way.)
// Write N = prod p_i^{n_i}, so 
//    (Z/NZ)^* = prod (Z/p_i^{n_i})^*
//    * the primes dividing M are exactly the primes at which
//      the character is quadratic. 
//    * if M == 0 then the character is "all characters",
//      i.e., do the computations on the full Gamma_1. 
///////////////////////////////////////////////////////





 
///////////////////////////////////////////////////////
// roots of unity class
// For efficiency, can not be used with multiple primitive
// roots in the same program!  I.e., Call set_prim only when
// no rootof1 objects are currently instantiated. 
// This design was chosen for efficiency, because it takes
// half as much space.  It's usually OK because most of the time
// all computations will take place over a fixed base field and
// its residue class fields. 

class rootof1 {
  static int n ;
  static bool rootof1::init_bigcomplex;
  static bool rootof1::init_charmod;
  static bool rootof1::init_intmod;
  static bool rootof1::init_algnum;
  static bool rootof1::init_ffield;

  int k ;

public: 
  rootof1() ;    // primitive n-th root. 
  rootof1 (const int& i) ;     // initialize with +/- 1 or 0 (only!)

  
  static void set_prim (int n) ;
  static int get_prim () { return rootof1::n ; }

  int get_k () const { return k ; }
  int table_idx () const { return k + 1 ; }
  void set_k (int _k) { k = _k ; }

  bigcomplex to_bigcomplex() const ;
  operator bigcomplex() const { return to_bigcomplex(); }

  charmod to_charmod () const ;
  operator charmod() const { return to_charmod () ; }
				// user must delete.
				// indexed by table_idx().
  static charmod* charmod_table () ; 


  intmod to_intmod () const ;
  operator intmod() const { return to_intmod () ; }

  bigrational to_bigrational () const ;
  operator bigrational() const { return to_bigrational () ; }

  //  bigint to_bigint () const ;
  //  operator bigint() const { return to_bigint() ; }

  ffield_element to_ffield_element () const ;
  operator ffield_element() const { return to_ffield_element () ; }

  algnum to_algnum() const ;
  operator algnum() const { return to_algnum () ; }

  rootof1& operator=(const rootof1& r) { k = r.k ; }
  rootof1& operator*=(const rootof1& r) ;
  rootof1& operator/=(const rootof1& r) ;

  bool operator==(const rootof1& r) const ;
  bool operator!=(const rootof1& r) const ;
  
  friend rootof1 operator- (const rootof1& r); // unary
  friend rootof1 operator+ (const rootof1& r); // unary 

  friend rootof1 operator* (const rootof1& r, const rootof1& s) ;
  friend rootof1 operator/ (const rootof1& r, const rootof1& s) ;
  friend rootof1 operator^ (const rootof1& r, int n) ;  // r^n
  friend bool iszero(const rootof1& r) ;
  friend ostream& operator<< (ostream&s, const rootof1& z);
} ;


#endif


