/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

// charmod.h

// implements class of small integers module some a static odd prime pr 
// integers being represented by an unsigned character. 
// The  maximum possible prime pr is 251.

#ifndef CHARMOD_H
#define CHARMOD_H

#include <iostream.h>
//#include <builtin.h>
#include "arith.h"
#include <LiDIA/bigint.h>

typedef unsigned char uchar;

class charmod {
    uchar val;
    static int pr;
    static int halfpr;
    static int prim_root ; 

public:
    static void set_pr(int p);
    static int get_pr() { return pr; }
    static int get_prim_root () { return prim_root ; }

    charmod() { val = 0; }
    charmod(int n);
    charmod(long n);
    charmod(const charmod& n) { val = n.val; }
    charmod(const uchar uc) { val = uc; } // quick constructor
    charmod(const bigint& N) ;
    charmod& operator = (const charmod& n) { val = n.val; return *this; }

    charmod& operator += (const charmod& n); 
    charmod& operator -= (const charmod& n); 
    charmod& operator *= (const charmod& n); 
    charmod& operator /= (const charmod& n);

    friend charmod  operator - (const charmod& n); // unary
    friend charmod  operator + (const charmod& n); // unary 
 
    friend charmod  operator + (const charmod&, const charmod& );
    friend charmod  operator - (const charmod&, const charmod& );
    friend charmod  operator * (const charmod&, const charmod& );
    friend charmod  operator / (const charmod&, const charmod& );

    friend void power (charmod& ator, charmod a, int r) ;

    bool operator == (const charmod& n) const;
    bool operator != (const charmod& n) const;

    friend bool iszero(const charmod& n);
    friend charmod inv(const charmod&);
    friend int lift(const charmod&);
    friend uchar ulift(const charmod&);
    friend int normlift(const charmod&);  //lifts to int i, -pr/2 < i < pr/2

    friend ostream& operator<< (ostream&s, const charmod&);
    friend istream& operator>> (istream&s, charmod&);

    void    error(const char* msg) const;     // error handling


};

inline charmod::charmod(int n)
{
  int j = n%pr; 
  if( j < 0 ) j += pr; 
  val = j;
}
inline charmod::charmod(long n)
{
  int j = n%pr; 
  if( j < 0 ) j += pr; 
  val = j;
}    

inline charmod::charmod(const bigint& N) 
{
  bigint M = N % pr ;
  int n; M.intify(n); *this=n; 
} 
inline charmod& charmod::operator += ( const charmod& n )
{
  int t = val; 
  t += n.val;
  t %= pr;
  val = t;
  return *this;
}

inline charmod& charmod::operator -= ( const charmod& n )
{
  int t = val; 
  t -= n.val;
  t %= pr;
  if( t < 0 ) t += pr;
  val = t;
  return *this; 
}

inline charmod& charmod::operator *= ( const charmod& n )
{
  int t = val;
  t *= n.val;
  t %= pr;
  val = t;
  return *this;
}

inline charmod& charmod::operator /= ( const charmod& n )
{
  int t = val;
  t *= inv(n).val; 
  t %= pr;
  val = t;
  return *this;
}

inline bool charmod::operator == (const charmod& n) const
{
  return (val == n.val) ;
}

inline bool charmod::operator != (const charmod& n) const
{
  return ( val != n.val);
}
 
inline charmod  operator - (const charmod& n)
{
  int t = n.val;
  return charmod(-t);
}

inline bool iszero(const charmod& n)
{
  return (n.val == 0);
}

inline int lift(const charmod& n)       // lifts to non-negative integer 
{
  return n.val;
}
inline uchar ulift(const charmod& n)      // lifts to unsigned character 
{
  return n.val;
}
inline int normlift(const charmod& n)   // lifts to integer n, -p/2 < n <= p/2
{
  int i = n.val;
  if( i > charmod::halfpr ) return( i - charmod::pr );
  else return i;
}

// using named returns below

inline charmod operator + (const charmod& n1, const charmod& n2) return ans(n1)
{
  ans += n2;
}

inline charmod operator - (const charmod& n1, const charmod& n2) return ans(n1)
{
  ans -= n2;
}

inline charmod operator * (const charmod& n1, const charmod& n2) return ans(n1)
{
  ans *= n2;
}

inline charmod operator / (const charmod& n1, const charmod& n2) return ans(n1)
{
  ans /= n2;
}

inline charmod  operator + (const charmod& n) return ans(n) {}

// these are false operators:

inline bool operator <(class charmod c, class charmod) 
{ 
    c.error("operator < not implemented for type charmod");
    return true;
}
inline bool operator >(class charmod c, class charmod)
{ 
    c.error("operator < not implemented for type charmod");
    return true;
}

inline void power (charmod& ator, charmod a, int r) 
{
  if (r == 0) { ator=1; return; }
  if (r < 0) {
    a=1/a;
    r=-r;
  }

  ator = 1 ;
  charmod t = a ;
  while (r >= 1) {
    if (r % 2) 
      ator *= t ;
    t *= t ;
    r /= 2 ;
  }
}

#endif




