/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */
//   charpoly.cc 

#include <LiDIA/bigint.h>
#include "charpoly.h"
#include "charmod.h"
#include "intmod.h"

#include "NTL/mat_ZZ.h"
#include "NTL/mat_poly_ZZ.h"

#include "NTL/mat_ZZ_p.h"
#include "NTL/mat_poly_ZZ_p.h"

#include "NTL/ZZX.h"
#include "NTL/ZZ_pX.h"

#include "fields.h"

ZZ to_ZZ (const bigint& n) ;  // defined in fields.cc



void Matrix_ZZify(mat_ZZ& A, bigint& l, const Matrix<bigrational>& M)
{
  // find lcm of denominators in M.
  l = 1;
  for (int r = 1 ; r <= M.nrows() ; r++)
    for (int c = 1 ; c <= M.ncols() ; c++) 
      if (M(r,c) != 0)
	l = lcm (l, denominator(M(r,c)) ) ;

  // Multiply M through by l to make M integral.
  // Then convert M to a ZZ matrix, and compute the charpoly 
  // using the wonderful NTL algorithm. 
  A.SetDims (M.nrows(), M.ncols()) ;
  for (int r = 1 ; r <= M.nrows() ; r++)
    for (int c = 1 ; c <= M.ncols() ; c++) {
      bigrational s = M(r,c) * bigrational(l) ;
      A(r,c)=to_ZZ(numerator(s));
    }
}

bigrational Determinant (const Matrix<bigrational>& M)
{
    if(M.nrows()!=M.ncols())
	die("Determinant of nonsquare matrix!");
    bigint l;
    mat_ZZ A ;
    Matrix_ZZify (A, l, M);
    bigint d=to_bigint(determinant(A, 1));
    bigint lton;
    power(lton, l, M.nrows());
    return bigrational(d)/bigrational(lton);
}


void CharPoly(Vector<bigrational>& f,  const Matrix<bigrational>& M) 
{
  if (M.nrows() != M.ncols()) 
    die("CharPoly can only be called with a square matrix.");
    
  mesg("[charpoly ", M.nrows()); mesg("]");

  bigint l;
  mat_ZZ A ;
  Matrix_ZZify (A, l, M);

  ZZX g;
  CharPoly (g, A) ;

  // Next renormalize (while copying to f) to account for the denominators.
  // Use that the char poly of M is g(lx)/l^deg(g). 

  bigrational a=l;
  f.init(deg(g)+1);
  f[deg(g)+1]=1;
  for (int i=deg(g)-1 ; i >= 0 ; i--) {
    f[i+1]=bigrational(to_bigint(coeff(g,i)));
    f[i+1] /= a;
    a *= l;
  }

  mesg("\n");
}

void CharPoly (const Matrix<bigrational>& M, ZZX& g) 
{
   Vector<bigrational> f;
   CharPoly(f, M);
   g.kill();
   g.SetMaxLength(f.size());
   for(int i=1;i<=f.size();i++)
       if(denominator(f[i])==1)
	   SetCoeff(g,i-1,to_ZZ(numerator(f[i])));
       else
	   die("CharPoly(Matrix<bigrational>, ZZX) nonintegral char poly.");
   
}

void CharPoly (const Matrix<charmod>& M, ZZ_pX& pol) 
{
    mesgln("Charpoly(const Matrix<charmod>&, ZZ_pX&)");
    if (charmod::get_pr() != ZZ_p::modulus()) {
	cerr << "Before calling CharPoly, you must first set the " << endl
	     << "modulus of ZZ_p to be the same as that of charmod." << endl;
	exit (0) ;
    }
    mat_ZZ_p A ;
    A.SetDims (M.nrows(), M.ncols()) ;
    for (int r = 1 ; r <= M.nrows() ; r++)
	for (int c = 1 ; c <= M.ncols() ; c++) 
	    A(r,c) = lift(M(r,c)) ;
    
    if(M.nrows()*M.ncols()>1500) 
	mesg("[NTL charpoly]");
    CharPoly (pol, A) ;

  if(M.nrows()*M.ncols()>1500) mesg("\n");
  mesgln("Charpoly done.");
}


void CharPoly (const Matrix<intmod>& M, ZZ_pX& pol) 
{
  if (intmod::get_pr() != ZZ_p::modulus()) {
    cerr << "Before calling CharPoly, you must first set the " << endl
	 << "modulus of ZZ_p to be the same as that of intmod." << endl;
    exit (0) ;
  }
  mat_ZZ_p A ;
  A.SetDims (M.nrows(), M.ncols()) ;
  for (int r = 1 ; r <= M.nrows() ; r++)
    for (int c = 1 ; c <= M.ncols() ; c++) 
      A(r,c) = lift(M(r,c)) ;

  if(M.nrows()*M.ncols()>1500) 
    mesg("[NTL charpoly]");

  CharPoly (pol, A) ; 

  if(M.nrows()*M.ncols()>1500) mesg("\n");
}

bool DiscIsZero (const ZZ_pX& f) 
{
  ZZ_pX fprime, gcd ;
  diff (fprime, f) ;
  GCD (gcd, f, fprime) ;
  return deg(gcd) > 0 ;
}

bool DiscIsZero (const ZZX& f) 
{
  ZZX fprime, gcd ;
  diff (fprime, f) ;
  GCD (gcd, f, fprime) ;
  return deg(gcd) > 0 ;
}






