/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */
//   cusps.cc

#include "arith.h"
#include "cusps.h"


// put in lowest terms
void reduce (bigint& p, bigint& q) 
{
    bigint d = abs(gcd(p,q)) ;
#ifdef DEBUG
    if (p%d!=0 || q%d!=0) { cerr << "error in reduce's call to gcd."<<endl; exit(-1); }
#endif
    p /= d ;  q /= d ;
}

// return true if we don't learn that the cusp is killed by
// the chi relations.  If it is is killed, return false.
bool Cusps::_list_add (bigint p, bigint q, bigint u) 
{
    reduce (p,q) ;
    
    bigint p2, q2 ;
    for (int i=0 ; i < _list_p.length() ; i++) { 
	p2 = _list_p[i] ;
	q2 = _list_q[i] ;
	if (((q - q2) % N == 0 && (p-p2)%gcd(q,N)==0)
	    ||((q - q2) % N == 0 && (-p-p2)%gcd(q,N)==0))
	    return chi.of (_list_u[i]) == chi.of (u) ;
    }
    
    // add to list
    int i = _list_p.length () ;
    _list_p.resize (i + 1) ;
    _list_q.resize (i + 1) ;
    _list_u.resize (i + 1) ;
    _list_p[i] = p ;
    _list_q[i] = q ;
    _list_u[i] = u ;
    return true ;
}


// returns chi(u) such that chi(u)[p1/q1] = [gamma_u(a/b)] 
// where p1/q1 was the first one encountered in this class.
rootof1 Cusps::_chi_u (bigint p, bigint q) 
{
    if (chi.type() == -1) return 1 ;    
    if (chi.type () == 0) return 1 ;
    
    reduce (p,q) ;
    
    // find p/q in the gamma1 list.
    for (int j=0 ; j < g1_p.length() ; j++) { 
	bigint p2 = g1_p[j] ;
	bigint q2 = g1_q[j] ;
	if (((q - q2) % N == 0 && (p-p2)%gcd(q,N)==0)
	    ||((q - q2) % N == 0 && (-p-p2)%gcd(q,N)==0))
	    return chi.of(g1_u[j]) ;
	
    }
    cerr << "Error in Cusps:_chi_u ("<<p<<","<<q<<")!" << endl
	 << "g1 length = " << g1_p.length () << endl;
    exit(0) ;
}


///////////////////////////////////////////////////////////////////
// returns true iff the cusp [p/q] is zero modulo the relations:
//       (I) [p/q] = M[p/q] for all M in Gamma_1(N)*(-1,0;0,1).  
//     which is q2 = +/- q1 (mod N) and p2 = +- p1 (mod gcd(q1,N)). 
//
//      (II) [gamma_u(p/q)] = chi(u) [p/q]. 
// 
// Notes: * this test is only necessary for nontrivial characters.
//        * it's quite conceivable that there is a more clever
//          and efficient algorithm which could be inserted here.

bool Cusps::is_cusp_killed (bigint p, bigint q) 
{
    if (chi.type() == -1) return false ;    // full Gamma_1(N).
    
    _list_p.reset(0) ;
    _list_q.reset(0) ;
    _list_u.reset(0) ;
    for (int j = 0 ; j < phiN ; j++) {
	bigint u = gam[j].d ;
	if ( ! _list_add (gam[j].a * p + gam[j].b * q, 
			  gam[j].c * p + gam[j].d * q, 
			  u))
	    return true ;
    }
    
    // now save our newly constructed list to the gamma1 list
    // for future use by _chi_u. 
    int i = g1_p.length () ;
    g1_p.resize (i+_list_p.length()) ;
    g1_q.resize (i+_list_q.length()) ;
    g1_u.resize (i+_list_u.length()) ;
    for (int j = 0 ; j < _list_p.length () ; j++) {
	g1_p[i+j] = _list_p[j] ;
	g1_q[i+j] = _list_q[j] ;
	g1_u[i+j] = _list_u[j] ;
    }
    
    return false ;
}


void Cusps::free (bigint p, bigint q, int& idx, rootof1& coeff) 
{
    coeff = 1 ;
    reduce (p, q) ;
    if (chi.type() == -1 && p*q < 0) { p = -abs(p); q = abs(q) ; }    
    
    for (int j=0 ; j < num_cusps ; j++) {
	// If we are working in the +1 eigenspace for the complex
	// conjugation (-1,0;0,1), i.e., plus=true, then
	// [-p/q] and [p/q] are identified.
	if ( equiv_signed (p, q, g0_p[j], g0_q[j]) ||
	     (plus && equiv_signed (-p, q, g0_p[j], g0_q[j]))) {
	    if (g0_is_zero[j]) {
		coeff = 0 ;
		idx = 1 ;
	    } else {
		coeff = _chi_u (p,q) ;
		idx = j+1 ;
	    }
	    return ;
	}
    }
    g0_p.resize (num_cusps + 1) ;
    g0_q.resize (num_cusps + 1) ;
    g0_is_zero.resize (num_cusps + 1) ;
    g0_p[num_cusps] = p ;
    g0_q[num_cusps] = q ;
    g0_is_zero[num_cusps] = is_cusp_killed (p, q) ;

    if(!g0_is_zero[num_cusps]) 
	if(mesg_on())
	    cerr <<"["<<p<<"/"<<q<<"]"<<flush;
    // is_cusp_killed adds to the g1 list. 
    if (g0_is_zero[num_cusps]) 
	coeff = 0 ;
    
    idx = num_cusps + 1 ;
    num_cusps++ ;
}







////////////////////////////////////////////////////////

bool Cusps::equiv_signed (bigint p1, bigint q1, bigint p2, bigint q2) 
{
    if (chi.type() == -1) {  // full Gamma_1(N).
	// lemma 3.2 of cremona's paper on gamma_1(N)
	if ( ((q2-q1)%N == 0) && ( ((p2-p1)%gcd(q1,N)) == 0) )
	    return true ;
	if ( ((q2+q1)%N == 0) && ( ((p2+p1)%gcd(q1,N)) == 0) )
	    return true ;
	return false ;
    }
    
    if (chi.type() != -1) {   // any other character
	// proposition 2.2.3 of cremona's book
	bigint s1=0, s2=0 ;
	if (q1 != 1 && p1 != 0) s1 = invmod (p1, q1) ;
	if (q2 != 1 && p2 != 0) s2 = invmod (p2, q2) ;
	if ((s1*q2 - s2*q1) % gcd(q1*q2,N) == 0)
	    return true ;
	else
	    return false ;
    }
    
    cerr << "cuspequiv: Uknown chi! Bailing!" << endl ;
    exit (0) ;
}

bool Cusps::equiv (bigint p1, bigint q1, bigint p2, bigint q2) 
{
    return equiv_signed (p1, q1, p2, q2) ||
	equiv_signed (-p1, q1, p2, q2) ;
}


Cusps::Cusps (const Character& i_chi, int i_k, bool _plus) 
    : chi(i_chi), k(i_k), plus(_plus)
{
    N = chi.modulus() ;
    num_cusps = 0 ;
    g0_p.reset (0) ;
    g0_q.reset (0) ;
    g0_is_zero.reset (0) ;
    g1_p.reset (0) ;
    g1_q.reset (0) ;
    g1_u.reset (0) ;
    _init_gamma () ;
}

Cusps::~Cusps() 
{
    if (gam!=NULL)
	delete[] gam;
}


// construct, for each u in (Z/NZ)^*, a matrix (a,b;c,d) in G_0(N)
// congruent to (u^{-1},0;0,u) mod N.
void Cusps::_init_gamma ()
{
    phiN = phi(N) ;
    gam = new matrix[phiN] ;
    int i = 0 ;
    for (int u = 1 ; u < N ; u++) {
	if (gcd (u, N) == 1) {
	    bigint uinv = invmod (bigint(u), bigint(N)*bigint(N)) ;
	    // so u * uinv + a*N^2 = 1 for some a, and in fact 
	    bigint a = (1 - bigint(u) * uinv) / (bigint(N)*bigint(N)) ;
				// since a/b = floor(a/b) for int's. 
	    if (i>=phiN) {
		cerr << "Error initializing gam[]" << endl;
		exit(-1) ;
	    }
	    gam[i].a = uinv ;
	    gam[i].b = - bigint(N) * a ;
	    gam[i].c = (a == 0) ? 0 : bigint(N) ;
	    gam[i].d = bigint(u) ;
	    i++ ;
	}
    }
}







