/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/* 
 FILE: fields.cc
 DESC: Implimentation of fields classes and conversion between 
       arithmetic types. 
*/

#include "fields.h"
#include "character.h"
#include "NTL/ZZX.h"
#include "my_ntl.h"

FIELD_TYPE field_type (bigrational& F) { return FT_BIGRAT ; }
FIELD_TYPE field_type (ffield_element& F) { return FT_FFIELD ; }
FIELD_TYPE field_type (charmod& F) { return FT_CHARMOD ; }
FIELD_TYPE field_type (intmod& F) { return FT_INTMOD ; }
FIELD_TYPE field_type (algnum& F) { return FT_ALGNUM ; }
// needed so can instantiate Matrix<int>
FIELD_TYPE field_type (int& F) { return FT_INT ; } 

const char* field_name (bigrational& F) { return "bigrational"; }

const char* field_name (algnum& F) 
{
  static char buf[1024]; 
  const polynomial<bigint>& f=algnum::irred_polynomial();
  ostrstream ost(buf,1024);
  ost<<"algnum_";
  for(int i=f.degree()-1;i>=0;i--) {
    ost<<f[i];
    if(i>0) ost<<'_';
  }
  ost<<'\0';			// make it null terminated. 
  return buf;
}

const char* field_name (ffield_element& F) 
{
  static char buf[1024]; 
  const Fp_polynomial& f
    = ffield_element::irred_polynomial();
  ostrstream ost(buf,1024);
  ost<<"ffield"<<ffield_element::characteristic()<<"_";
  for(int i=f.degree()-1;i>=0;i--) {
    ost<<f[i];
    if(i>0) ost<<'_';
  }
  ost<<'\0';			// make it null terminated. 
  return buf;
}

const char* field_name (charmod& F) 
{
  static char buf[1024];
  sprintf(buf, "charmod%d", charmod::get_pr());
  return buf;
}

const char* field_name (intmod& F) 
{
  static char buf[1024];
  sprintf(buf, "intmod%d", intmod::get_pr());
  return buf;
}


//////////////////// ffield_element implimentation ////////////////

void
ffield_element::irred_polynomial(Vector<bigint>& f)
{
    mesgln("start irred_poly");
    const Fp_polynomial& g=def_base->irred_polynomial();
    f.init(g.degree()+1);
    for(int i=1; i<=f.size(); i++)
	f[i] = g[i-1];
    mesgln("finished irred_poly");
}

const gf_p_base* ffield_element::def_base = 0;
const ffield_element* ffield_element::prim_elem = 0;
bigint ffield_element::card = 0;

void  ffield_element::set_default_base (const gf_p_base* base) 
{
  delete ffield_element::def_base ;
  delete ffield_element::prim_elem ;

  ffield_element::def_base = base ;
  ffield_element::prim_elem =
    new ffield_element(base->gen_primitive_elem());
				// The last remaining pointer to base 
                                // and prim elem are deleted only when 
                                // the heap is destroyed.  This is probably
                                // perfectly fine. 
  card = def_base->field().number_of_elements();
}

const ffield_element& ffield_element::primitive_elem() 
{
  if (!ffield_element::def_base) {
    cerr << "Initialize ffield_element default first!!!" << endl ;
    exit (-1) ;
  }
  return *ffield_element::prim_elem ; 
}

const gf_p_base& ffield_element::get_def_base () {
  if (!ffield_element::def_base) {
    cerr << "Initialize ffield_element default first!!!" << endl ;
    exit (-1) ;
  }
  return *ffield_element::def_base ; 
}



bool operator == (const ffield_element & a,
				  const bigint & i)
{
  return a == ffield_element(i) ;
}

ostream & operator << (ostream & os, const ffield_element & a)
{
  const Fp_polynomial& f = a.polynomial_rep() ;

  if (f.degree()<0) {
    os << "0";
    return os ;
  }
  int d = f.degree();
  os<<"(";
  for(int i=d; i>=0; i--) {
    if(f[i] != 0) {
      if (f[i] != 1 || i==0)  {
	os << f[i] ;
	if (i>0) os << "*";
      }
      if (i>1)
	os << "x^" << i ;
      if (i==1)
	os << "x" ;
    }
    if(i>0 && f[i-1] != 0)
      os << "+";
  }
  os<<")";
  return os;
}

// b = a^r
void power (ffield_element& b, const ffield_element& a, int r)
{
  if (r == 0) { b=1; return ; }
  if (r < 0) {
    power (b, a, -r);
    b = 1/b;
    return;
  }

  b = 1 ;
  ffield_element t = a ;
  while (r >= 1) {
    if (r % 2) 
      b *= t ;
    t *= t ;
    r /= 2 ;
  }
}

void power (algnum& b, const algnum& a, int r)
{
  if (r == 0) { b=1; return ; }
  if (r < 0) {
    power (b, a, -r);
    b = 1/b;
    return;
  }

  b = 1 ;
  algnum t = a ;
  while (r >= 1) {
    if (r % 2) 
      b *= t ;
    t *= t ;
    r /= 2 ;
  }
}


void 
ffield_element::save_default_base (ofstream& of)
{
  of << def_base->irred_polynomial() << endl;
}

void 
ffield_element::load_default_base (ifstream& in)
{
  Fp_polynomial pol; 
  in >> pol ;
  InitFieldFinite(pol);
}

ffield_element::ffield_element (ifstream& in)
{
  Fp_polynomial pol;
  pol.set_modulus(characteristic());
  pol.set_max_degree(degree());
  for(int i=0; i<degree(); i++) {
    bigint a; in >> a;
    pol.set_coefficient(a,i);
  }
  set_base(*def_base);
  set_polynomial(pol);
}

ffield_element::ffield_element(const Fp_polynomial& pol)
{
  set_base(*def_base);
  set_polynomial(pol);
}

void 
ffield_element::save(ofstream& of) const 
{
  for(int i=0; i<degree(); i++) 
    of << polynomial_rep()[i] << " ";
  of << endl;
}

int
ffield_element::degree() 
{
  return def_base->irred_polynomial().degree();
}

bigint
ffield_element::characteristic() 
{
  return def_base->irred_polynomial().modulus();
}


///////////////////// algnum implimentation///////////////////

const algnum* algnum::zeta =0; 

algnum algnum::Zeta(const int& n) 
{
  algnum a;
  power(a,*algnum::zeta,n);
  return a;
}

void 
algnum::save(ofstream& out) const
{
  out << alg_number(*this) <<endl;
}

void 
algnum::load(ifstream& in)
{
  alg_number a;
  in >> a;
  *this = a;
}


ostream & operator << (ostream & os, const algnum & a)
{
  if (a == algnum(0)) {
    os << "0" ;
    return os ;
  }

  const math_vector<bigint>& f = coeff_vector(a) ;

  os << "(";
  for(int i=f.size()-1; i>=0; i--) {
    if(f[i] != 0) {
      if (f[i] == -1 && i>0) 
	os << "-" ;
      if (abs(f[i]) != 1 || i==0)  {
	os << f[i] ;
	if (i>0) os << "*";
      }
      if (i>1)
	os << "a^" << i;
      if (i==1)
	os << "a" ;
    }
    if(i>0 && f[i-1] > 0)
      os << "+";
  }
  os << ")";
  if (denominator(a)!=1) os<<"/"<<denominator(a);
  return os;
}

bool
algnum::is_bigint() const
{
  if (denominator() != 1) return false;
  for(int i=1; i<coeff_vector().size(); i++)
    if (coeff_vector()[i]!=0) return false;
  return true;
}

void 
algnum::bigintify(bigint& n) const
{
#ifdef DEBUG
  if (!is_bigint()) {
    cerr << "algnum::Can't convert " << (*this) << " to a bigint."<<endl;
    exit(-1);
  }
#endif
  n=coeff_vector()[0];
}

polynomial<bigrational>
algnum::lift() const
{
    const math_vector<bigint>& v=coeff_vector();
    bigrational w[v.size()];
    for(int i=0;i<v.size();i++)
	w[i]=bigrational(v[i])/bigrational(denominator());
    return polynomial<bigrational>(w,v.size()-1);
}

void
algnum::set_rootirredpoly()
{
    bigint a[degree()];
    for(int i=0;i<degree();i++) a[i]=0;
    a[1]=1;
    *this = algnum(alg_number(a,1));
}

int 
algnum::degree () 
{
  number_field F;		// initialized to default field.
  return ::degree(F);
}

const polynomial<bigint>& 
algnum::irred_polynomial() 
{ 
  number_field K;
  return which_polynomial(K); 
}


algnum::algnum (const rootof1& mu) 
{
  *this = mu.to_algnum();
}

void UserInitField(FIELD_TYPE FT) 
{
  int p, d;
  if (FT==FT_FFIELD) {
    p=get_int("Finite field characteristic = ",2);
    d=get_int("Finite field degree = ",1);
    // ffield_element takes care of deleting this heap based galois_field.
    galois_field* field = new galois_field (p, d);
    ffield_element::set_default_base (new gf_p_base(*field)) ;
    cout << "F=Fp[x]/(f) where f = " <<
      ffield_element::get_def_base() << " initialized." << endl ;
  } else if (FT==FT_CHARMOD) {
    p =get_int("charmod Characteristic = ", 3, 255);
    charmod::set_pr(p) ;
    cout << "finite field F_"<<p<<" initialized." << endl ;
  } else if (FT==FT_INTMOD) {
    p=get_int("intmod Characteristic = ", 3, 32767) ;
    intmod::set_pr(p) ;
    cout << "finite field F_"<<p<<" initialized." << endl ;
  } else if (FT==FT_BIGRAT) {
    // nothing to do.
  } else if (FT==FT_ALGNUM) {
    polynomial<bigint> f ;        // f = x^2+1
    cout << "Defining K=Q[x]/(f).  "
	 << "What is f? (when done hit return twice): " << endl ;
    cin >> f ;
    number_field K(f) ;	// K = Q[x]/(f) is now the current base field.
  
    cout << endl << "Let n (or 2n if n is odd) be the degree of the character."
	 << endl << "Enter an n-th root of unity, as a vector (e.g., [0 1]) "
	 << endl << "on the basis 1,x,...,x^{deg f-1}.:  " ;
    base_vector<bigint> r;
    do {
      cin >> r;
    } while (r.size() != f.degree()) ;
    alg_number* zeta = new alg_number(r, 1, K) ;
    algnum::Zeta((algnum*)zeta) ;	// Record this as the zeta for use by rootof1. 
  }
}

void InitFieldAlgnumCyclo(const int& n)
{
  mesg("Initializing Q(zeta_",n); 
  mesg(") = ");

  polynomial<bigint> f=polcyclo(lcm(2,n));
  cout << "Q[x]/("<<f<<")"<<endl;
  number_field K(f) ;		// K = Q[x]/(f) is now the current base field.
				// it is the lcm(2,n)-th cyclotomic field.
  
  base_vector<bigint> r(f.degree(),FIXED); 
  if (n>2) {
    r[f.degree()-1]=0;
    r[1]=1; 
  } else if (n==1) 
    r[0]=1;
  else if (n==2)
    r[0]=-1;
  alg_number* zeta = new alg_number(r, 1, which_base(K)) ;
  algnum::Zeta((algnum*)zeta) ;	// Record this as the zeta for use by rootof1. 
}

void InitFieldCharmod(int p)
{
    if (3<=p && p<=253) {
	charmod::set_pr(p);
	ZZ_p::init(ZZ(INIT_VAL,p));
	mesg("Initializing (charmod) F_",p);mesg("\n");
    } else {
	die("InitFieldCharmod: p must be between 3 and 253.");
    }

    ntl_init_to_charmod();
}

void InitFieldIntmod(int p)
{
    if (3<=p && p<=45083) {
	intmod::set_pr(p);
	ZZ_p::init(ZZ(INIT_VAL,p));
	mesg("Initializing (intmod) F_",p);mesg("\n");
    } else {
	die("InitFieldIntmod: p must be between 3 and 45083.");
    }

    ntl_init_to_intmod();
}

void InitFieldFinite(const Fp_polynomial & pol)
{
    static Fp_polynomial last;
    if(pol==last) return;
    last=pol;
  galois_field* field = new galois_field (pol.modulus(), pol.degree());
  ffield_element::set_default_base (new gf_p_base(*field, pol)) ;
  if(mesg_on()) 
      cerr <<"Initializing F=F_{"<<pol.modulus()<<"}[x]/f(x), " <<endl
	   << "f(x)= "<<ffield_element::get_def_base();
}

// Same as the built in LiDIA function "build_irred", except
// this is deterministic in the most stupid way, and hence
// correspondingly slower. 
// Use the slowest, but safest algorithm, because this probably
// isn't a time critical part of this system.  If it becomes one,
// rewrite.  
void determ_build_irred(Fp_polynomial& f, int p, int d)
{
  f.set_modulus(p);
  f.set_max_degree(d);
  for(int j=1;j<d;j++) f[j]=0;
  f[0]=bigint(1);
  f[d]=bigint(1);
  int i=0;
  while(!det_irred_test(f)) {
    f[i]=f[i]+bigint(1);
    while(f[i]%p==0) {
      f[i++]=0;
      f[i]=f[i]+bigint(1);
      if(f[i]<bigint(p)) { i=0; break; }
      if(i>d) 
	  die("determ_build_irred: logical contradiction, bug in LiDIA.");
    }
  }
}


void InitFieldFinite(int p, int d)
{
  Fp_polynomial pol;
  determ_build_irred(pol, p, d); 
  InitFieldFinite(pol);
}

void InitFieldFiniteCyclo(int n, int p)
{
#ifdef DEBUG
  if (n <= 0) { 
    cerr << "InitFieldFiniteCyclo: error, negative n" << endl;
    exit(-1);
  }
  if (!is_prime(p)) {
    cerr << "InitFieldFiniteCyclo: error, p not prime" << endl ;
    exit(-1);
  }
#else
  if (n<=0) n==1;
#endif


  mesg("Initializing ",n); 
  mesg("-th cyclotomic field/F_",p);
  mesg(".\n");
  if (n%p==0) 
    cout << p << " is ramified in Q(zeta_" << n << ")" << endl 
	 << "so in characteristic " << p << " there is no primitive "
	 <<n<<"-th root of unity"<<endl;

  while (n%p==0) n/= p;
  int d = znorder(p,n);		// n|p^d-1
  // construct field
  InitFieldFinite(p, d);
}


// HERE!

// 4096 digits should be enough.  
// I'm using this fixed length for efficiency purposes
// as allocating a new buffer for every conversion would be slow.
static const int buflen=4096;
static char buf[buflen];	

ZZ to_ZZ (const bigint& n) {
// allocat
  bigint_to_string (n, buf);
  return ZZ(INIT_VAL, buf);
}

bigint to_bigint (const ZZ& n) {
  if(n==0) return 0;
  ostrstream ost(buf,buflen);
  ost<<n<<'\0';
  bigint ans;
  string_to_bigint(buf, ans);
  return ans;
}

void to_bigrationalvec (Vector<bigrational>& g, const ZZX& f) 
{
    if (deg(f)<0) { g.init(0); return; }
    g.init(deg(f)+1);
    for(int i=0; i<=deg(f); i++)
	g[i+1]=bigrational(to_bigint(coeff(f,i)));
}

void to_bigrationalvec (Vector<bigrational>& g, const polynomial<bigrational>& f) 
{
  if (f.degree()<0) { g.init(0); return; }
  g.init(f.degree()+1);
  for(int i=0; i<=f.degree(); i++)
    g[i+1]=f[i];
}

void to_bigrationalvec (Vector<bigrational>& g, const Vector<bigint>& f)
{
    g.init(f.size());
    for(int i=1; i<=f.size();i++)
	g[i]=f[i];
}

void to_bigintvec (Vector<bigint>& g, const ZZX& f) 
{
  if (deg(f)<0) { g.init(0); return; }
  g.init(deg(f)+1);
  for(int i=0; i<=deg(f); i++)
    g[i+1]=to_bigint(coeff(f,i));
}

void to_polybigint (polynomial<bigint>& g, const ZZX& f) 
{
  if (deg(f)<0) { polynomial<bigint> zero; g=zero; return; }

  base_vector<bigint> c(deg(f)+1,FIXED);
  for(int i=0; i<=deg(f); i++)
    c[i]=to_bigint(coeff(f,i));
  g = polynomial<bigint>(c) ;
}

bigint to_polybigint (polynomial<bigint>& g, const Vector<bigrational>& ff) 
{
    bigint d=denominator(ff.polgetcoeff(0));
    for(int i=1; i<=ff.poldeg(); i++)
	d=lcm(d,denominator(ff.polgetcoeff(i)));
    
    Vector<bigrational> f = bigrational(d)*ff; 

    if (f.poldeg()<0) { polynomial<bigint> zero; g=zero; return 0; }
    base_vector<bigint> c(f.poldeg()+1,FIXED);
    for(int i=1; i<=f.poldeg()+1; i++) {
	if(denominator(f[i])!=1) {
	    cerr<<"to_polybigint: error -- non an integer: f = " << f << endl;
	    die();
	}
	c[i-1]=numerator(f[i]);
    }
    g = polynomial<bigint>(c) ;
    return d;
}


void to_polybigrational(polynomial<bigrational>& g, const Vector<bigrational>& f) 
{
  if (f.poldeg()<0) { polynomial<bigrational> zero; g=zero; return; }

  base_vector<bigrational> c(f.poldeg()+1,FIXED);
  for(int i=1; i<=f.poldeg()+1; i++) 
    c[i-1]=f[i];
  g = polynomial<bigrational>(c) ;
}

void to_ZZX (ZZX& g, bigint& denom, const Vector<bigrational>& f)
{
    g.SetMaxLength(f.size());
    denom=1;
    for(int i=1;i<=f.size();i++) 
	denom=lcm(denom,denominator(f[i]));
    for(int i=1;i<=f.size();i++) 
	SetCoeff(g,i-1,to_ZZ(numerator(f[i]*denom) ));
}


charmod to_charmod (const algnum& alpha, charmod a)
{
    const math_vector<bigint>& v=alpha.coeff_vector();
    bigint d=alpha.denominator();
    if(d%charmod::get_pr()==0) 
	die("to_charmod: conversion from algnum to charmod not possible, divide by zero!");
    charmod b(0);
    for(int i=v.size()-1;i>=0;i--) {
	b+=v[i];
	if(i>0) b*=a;
    }
    return b/d;
}

bool can_reduce_charmod(const algnum& alpha)
{
    return (alpha.denominator() % charmod::get_pr()) != 0;
}
bool can_reduce_charmod(const charmod&) { return true; }
bool can_reduce_charmod(const intmod&) { die("can_reduce_charmod(intmod) not implimented."); }
bool can_reduce_charmod(const bigrational& x) {
    return denominator(x)%charmod::get_pr() != 0;
}

bool can_reduce_charmod(const ffield_element&) { return false; }


template <class F>
void to_charmod(Matrix<charmod>& B,
		const Matrix<F>& A, charmod z)
{
    B.init(A.nrows(),A.ncols());
    for(int r=1;r<=A.nrows();r++)
	for(int c=1;c<=A.ncols();c++)
	    B(r,c)=to_charmod(A(r,c),z);
}

charmod to_charmod (const charmod& alpha, charmod z=0){ return alpha;}
charmod to_charmod (const intmod& alpha, charmod z=0) {
    die("to_charmod(intmod) not implimented");
}
charmod to_charmod (const bigrational& alpha, charmod z=0) {
    return charmod(numerator(alpha))/charmod(denominator(alpha));
}
charmod to_charmod (const ffield_element& alpha, charmod z=0) {
    die("to_charmod(ffield_element) not implimented");
}


template <class F>
bool can_reduce_charmod(const Matrix<F>& A)
{
    for(int r=1;r<=A.nrows();r++)
	for(int c=1;c<=A.ncols();c++)
	    if(!can_reduce_charmod(A(r,c))) return false;
    return true;
}


template <class F>
void to_charmod(Array<charmod>& w,
		const Array<F>& v, charmod z)
{
    w.reset(v.length());
    for(int i=0;i<v.length();i++)
	w[i]=to_charmod(v[i],z);
}

template <class F>
bool can_reduce_charmod(const Array<F>& v)
{
    for(int i=0;i<v.length();i++)
	if(!can_reduce_charmod(v[i]))
	    return false;
    return true;
}


bool can_reduce_intmod(const bigrational& x)
{
    return denominator(x)%intmod::get_pr() != 0;
}
intmod to_intmod (const bigrational& alpha, intmod z=0)
{
    return intmod(alpha);
}
void to_intmod(Matrix<intmod>& B,
		const Matrix<bigrational>& A)
{
    B.init(A.nrows(),A.ncols());
    for(int r=1;r<=A.nrows();r++)
	for(int c=1;c<=A.ncols();c++)
	    B(r,c)=to_intmod(A(r,c));
}

void to_bigrationalmat (Matrix<bigrational>& B, const bigint_matrix& A) 
{
    B.init(A.get_no_of_rows(), A.get_no_of_columns());
    for(int r=1;r<=B.nrows();r++)
	for(int c=1;c<=B.ncols();c++)
	    B(r,c)=A(r-1,c-1);
}

void to_algnum (Vector<algnum>& g, const Vector<bigrational>& f)
{
    g.init(f.size());
    for(int i=1;i<=f.size();i++)
	g[i]=algnum(f[i]);
}


void to_bigcomplex(Matrix<bigcomplex>& B,
		  const Matrix<bigrational>& C)
{
    B.init(C.nrows(),C.ncols());
    for(int r=1; r<=C.nrows(); r++)
	for(int c=1; c<=C.ncols(); c++)
	    B(r,c)=bigcomplex(C(r,c));
}

void to_bigcomplex(Vector<bigcomplex>& B,
		   const Vector<bigrational>& C)
{
    B.init(C.size());
    for(int i=1; i<=C.size(); i++)
	B[i]=bigcomplex(C[i]);
}

// calling function must delete string pointed to by returned pointer.
char* strip_nonumeric(const char* str)
{
    if(str==NULL) return NULL;
    int cnt=0;
    for(int i=0;i<strlen(str);i++)
        if(('0' <= str[i] && str[i] <= '9') || str[i]=='-')
            cnt++;
    char* s=new char[cnt+1];
    cnt=0;
    for(int i=0;i<strlen(str);i++)
        if(('0' <= str[i] && str[i] <= '9') || str[i]=='-')
            s[cnt++]=str[i];
    s[cnt]='\0';
    return s;                   
}

bigrational parse_bigrational(const char* str)
{
    if(str==NULL) return 0;
    const char* s1=str; 
    const char* s2=str;
    int i=0;
    int n=strlen(str);
    while (i<n && s2[0]!='/') s2++;
    bigrational ans;
    char *s1n, *s2n;
    if(i==n) {
        s1n=strip_nonumeric(s1);
        s2n=new char[2];
	s2n[0]='1';
	s2n[1]='\0';
    } else {
	s2++;
	s1n=strip_nonumeric(s1);
	s2n=strip_nonumeric(s2);
    }

    if(strlen(s1n)==0 || strlen(s2n)==0) {
	delete s1n;
	delete s2n;
	return bigrational(0);
    }
    string_to_bigrational(s1n, s2n, ans);
    delete s1n;
    delete s2n;
    return ans;
}

bigint parse_bigint(const char* str)
{
    if(str==NULL) return 0;
    char* s=strip_nonumeric(str);
    if(strlen(s)==0) {
	delete s;
	return bigint(0);
    }
    bigint ans;
    string_to_bigint(s, ans);
    delete s;
    return ans;
}




// TODO: make these parse in a nicer way.
Vector<bigrational> parse_bigrational_poly(const char* str)
{
    istrstream ist(str);
    Vector<bigrational> v;
    ist >> v;
    return v;
}

Vector<bigint> parse_bigint_poly(const char* str)
{
    istrstream ist(str);
    Vector<bigint> v;
    ist >> v;
    return v;
}

int parse_int(const char* str)
{
    if(str==0) return 0;
    char* s=strip_nonumeric(str);
    int r=atoi(s);
    delete s;
    return r;
}


#ifdef INST_BIGRATIONAL
template bool can_reduce_charmod(const Matrix<bigrational>& A);
template void to_charmod(Matrix<charmod>&,
		    const Matrix<bigrational>&, charmod);
template bool can_reduce_charmod(const Array<bigrational>& A);
template void to_charmod(Array<charmod>&,
		    const Array<bigrational>&, charmod);
#endif
#ifdef INST_FFIELD
template bool can_reduce_charmod(const Matrix<ffield_element>& A);
template void to_charmod(Matrix<charmod>&,
		    const Matrix<ffield_element>&, charmod);
template bool can_reduce_charmod(const Array<ffield_element>& A);
template void to_charmod(Array<charmod>&,
		    const Array<ffield_element>&, charmod);
#endif
#ifdef INST_CHARMOD
template bool can_reduce_charmod(const Matrix<charmod>& A);
template void to_charmod(Matrix<charmod>&,
		    const Matrix<charmod>&, charmod);
template bool can_reduce_charmod(const Array<charmod>& A);
template void to_charmod(Array<charmod>&,
		    const Array<charmod>&, charmod);
#endif
#ifdef INST_INTMOD
template bool can_reduce_charmod(const Matrix<intmod>& A);
template void to_charmod(Matrix<charmod>&,
		    const Matrix<intmod>&, charmod);
template bool can_reduce_charmod(const Array<intmod>& A);
template void to_charmod(Array<charmod>&,
		    const Array<intmod>&, charmod);
#endif
#ifdef INST_ALGNUM
template bool can_reduce_charmod(const Matrix<algnum>& A);
template void to_charmod(Matrix<charmod>&,
		    const Matrix<algnum>&, charmod);
template bool can_reduce_charmod(const Array<algnum>& A);
template void to_charmod(Array<charmod>&,
		    const Array<algnum>&, charmod);
#endif

