/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/* 
 FILE: fields.h 
 DESC: Implimentation of fields classes and conversion between 
       arithmetic types. 
 NOTES:
  * A cyclotomic field Q(zeta_n) is DEFINED by giving n.  
    thus the algorithm which generates an irreducible
    polynomial whose root generates zeta_n has to be
    nondeterministic.  Of course I choose the n-th cyclotomic
    polynomial. 
  * A finite field F_{p^r} is DEFINED by p and r, 
    in the sense that the irreducible polynomial defining
    the finite field MUST be the same each time the field
    is initialized.  I've used a SLOWER nondeterministic
    algorithm in order to accomplish this. 
*/

#ifndef FIELDS_H
#define FIELDS_H

#include <LiDIA/bigrational.h>
#include <LiDIA/gf_p_element.h>
#include <LiDIA/alg_number.h>
#include "charmod.h"
#include "intmod.h"
#include "matrix.h"

const char* field_name (bigrational&) ;
class ffield_element;		// forward decl.
const char* field_name (ffield_element&) ;
const char* field_name (charmod&) ;
const char* field_name (intmod&) ;
class algnum;
const char* field_name (algnum&) ;

////////////////////
// ffield_element
//
// This class extends LiDIA's gf_p_element class to
// allow for a default base field so that objects are
// automatically initialized to 0 and casting from
// other types (like integers) is well-defined. 
//
// The function "set_default_base" must be called before
// this class is used.  The argument to set_default_base
// is assumed to be a pointer to a gf_p_base object which
// has been created on the heap (using operator new, for example).
// It is a bad idea to pass a pointer to a stack based object!
// When set_default_base is called, a fixed generator for 
// the multiplicative group is computed which can be obtained 
// by calling primitive_elem().

class ffield_element : public gf_p_element {
  static const gf_p_base* def_base ;
  static const ffield_element* prim_elem ;   
  static bigint card;

public:
  static void set_default_base (const gf_p_base* base) ;
  static const ffield_element& primitive_elem() ;
  static const gf_p_base& get_def_base () ;
  static void save_default_base (ofstream& of);
  static void load_default_base (ifstream& in);
  static int degree() ;
  static bigint cardinality() { return card; }
  static bigint characteristic();
  static const Fp_polynomial&  irred_polynomial() 
    { return def_base->irred_polynomial() ; }
  static void irred_polynomial(Vector<bigint>& f);

  ffield_element () ;
  ffield_element (const ffield_element& a) ;
  ffield_element (const int& n) ;
  ffield_element (const bigint& n) ;
  ffield_element (const gf_p_element& a) ;
  ffield_element (const Fp_polynomial& pol) ;
  ffield_element (const char* str) ;
  ffield_element (ifstream& in);
  Fp_polynomial  rep() { return polynomial_rep(); }

  void load(ifstream& in) { *this = ffield_element(in) ; }
  void save(ofstream& of) const ;

  ffield_element& operator = (const ffield_element& a) ;

  ffield_element operator-() ;

  ffield_element& operator += (const ffield_element& x) ;
  ffield_element& operator -= (const ffield_element& x) ;
  ffield_element& operator *= (const ffield_element& x) ;
  ffield_element& operator /= (const ffield_element& x) ;
  friend bool operator == (const ffield_element & a, const bigint & i) ;
  friend ostream & operator <<  (ostream & os, const ffield_element & a) ;
} ;

				// b=a^r
void power (ffield_element& b, const ffield_element& a, int r);


// definitions of inlined functions.

inline ffield_element::ffield_element (const gf_p_element& a) {
  set_base(a.base());
  set_polynomial(a.polynomial_rep());
}

inline ffield_element::ffield_element (const ffield_element& a) 
{
  set_base(a.base());
  set_polynomial(a.polynomial_rep());
}

inline ffield_element::ffield_element () 
{
  set_base(*def_base);
  assign_zero();
}


inline ffield_element::ffield_element (const int& n) 
{
  set_base(get_def_base()) ;	
  Fp_polynomial f ;
  f.set_modulus(def_base->field().characteristic());
  f.assign_one ();

#ifdef DEBUG
  if (f[0] != 1) { cerr << "there is a bug in ffield_element::ffield_element (const int& n)"<<endl; exit(-1); }
#endif

  f[0] = n;
  set_polynomial (f) ;
}

inline ffield_element::ffield_element (const bigint& n) 
{
  set_base(get_def_base()) ;	
  Fp_polynomial f ;
  f.set_modulus(def_base->field().characteristic());
  f.assign_one ();
  f[0] = n;
  set_polynomial (f) ;
}

inline ffield_element::ffield_element (const char* str) 
{
  bigint n; 
  string_to_bigint(str, n);
  *this = ffield_element(n);
}

inline ffield_element&
ffield_element::operator=(const ffield_element& a) 
{
  set_base(a.base());
  set_polynomial(a.polynomial_rep());
  return *this;
}

inline ffield_element ffield_element::operator-() 
{ 
  return ffield_element(-1) * (*this) ; 
}

inline ffield_element& ffield_element::operator += (const ffield_element& x) 
{
  *this = *this+x ;
  return *this ;
}

inline ffield_element& ffield_element::operator -= (const ffield_element& x) 
{
  *this = *this - x ;
  return *this ;
}

inline ffield_element& ffield_element::operator *= (const ffield_element& x) 
{
  *this = *this * x ;
  return *this ;
}

inline ffield_element& ffield_element::operator /= (const ffield_element& x) 
{
  *this = *this / x ;
  return *this ;
}

///////////// end ffields_element //////////////////////////////////////



///////////// algnum  -- algebraic number field element ///////////////
// NOTE: if Zeta is for the convenience of rootof1 which must
// cast from roots of unity to algnum.  Before working with a 
// character of degree n, set zeta to be a primitive nth root
// of unity. At some point, for speed reasons, Zeta(int) may
// be implimented using a table

class algnum : public alg_number {
  static const algnum* zeta; 
public:
  static algnum Zeta(const int& n);	// returns zeta^n
  static void Zeta (const algnum* z);	// sets zeta
  static int degree () ;	// returns the degree of the default field.
  static const polynomial<bigint>& irred_polynomial() ;

  algnum () ;
  algnum (const int& n) ;
  algnum (const rootof1& mu) ;
  algnum (const bigint& n) ;
  algnum (const bigrational& n) ;
  algnum (const algnum& a) ;
  algnum (const alg_number& a) ;

  void set_rootirredpoly();

  void save(ofstream& out) const;
  void load(ifstream& in);

  algnum& operator = (const algnum& a) ;
  algnum operator-() ;
  //  algnum& operator* (const algnum& x);
  algnum& operator += (const algnum& x) ;
  algnum& operator -= (const algnum& x) ;
  algnum& operator *= (const algnum& x) ;
  algnum& operator /= (const algnum& x) ;
  friend bool operator == (const algnum & a, const bigint & i) ;
  friend ostream & operator << (ostream & os, const algnum & a) ;
  bool is_bigint() const ;
  void bigintify(bigint& n) const ;
  polynomial<bigrational> lift() const;
} ;

void power (algnum& b, const algnum& a, int r);




// definitions of inlined functions.

inline void algnum::Zeta (const algnum* z)
{
  algnum::zeta = z ;
}


inline algnum::algnum (const alg_number& a) {
  this->assign(a) ;
  /*  den=a.denominator();
  coeff = a.coeff_vector() ;
  O = a.which_base() ;*/
}

inline algnum::algnum (const algnum& a) 
{
  this->assign(a) ;
}

inline algnum::algnum () 
{
  assign_zero();
}

inline algnum::algnum (const int& n) 
{
  assign(bigint(n));
}

inline algnum::algnum (const bigint& n) 
{
  assign(n);
}

// this should be optimized
inline algnum::algnum (const bigrational& x) 
{
  alg_number num; num.assign(::numerator(x));
  alg_number denom; denom.assign(::denominator(x));
  *this = num/denom;
}

inline algnum& algnum::operator = (const algnum& a) 
{
  this->assign(a) ;
  return *this;
}

inline algnum algnum::operator-() 
{ 
  return algnum(-1) * (*this) ; 
}

/*
inline algnum& algnum::operator* (const algnum& x)
{
  return *this = algnum (alg_number(*this) * alg_number(x));
}
*/

inline algnum& algnum::operator += (const algnum& x) 
{
  *this = *this+x ;
  return *this ;
}

inline algnum& algnum::operator -= (const algnum& x) 
{
  *this = *this - x ;
  return *this ;
}

inline algnum& algnum::operator *= (const algnum& x) 
{
  *this = *this * x ;
  return *this ;
}

inline algnum& algnum::operator /= (const algnum& x) 
{
  *this = *this / x ;
  return *this ;
}

//////////////////////// CONVERSIONS ///////////////////////////////
template <class K1, class K2> 
void ChangeField (Matrix<K1>& A,
		  const Matrix<K2>& B) 
{
  A.init(B.nrows(), B.ncols());
  for(int r=1;r<=B.nrows();r++)
    for(int c=1;c<=B.ncols();c++)
      A(r,c)=B(r,c);
}


//////////////////////// FIELD INITIALIZATION FUNCTIONS ////////////

enum FIELD_TYPE {FT_FFIELD, FT_CHARMOD, FT_INTMOD, FT_BIGRAT,
		 FT_ALGNUM, FT_INT};
				// prompt the user for parameters
				// for this field
void UserInitField(FIELD_TYPE F) ;
				// initialize algnum to be the 
				// n-th cyclotomic extension.
void InitFieldAlgnumCyclo(const int& n);
void InitFieldFiniteCyclo(int n, int p);
void InitFieldFinite(const Fp_polynomial& pol);
void InitFieldFinite(int p, int d);
void InitFieldCharmod(int p);
void InitFieldIntmod(int p);

FIELD_TYPE field_type (bigrational& F) ;
FIELD_TYPE field_type (ffield_element& F) ;
FIELD_TYPE field_type (charmod& F) ;
FIELD_TYPE field_type (intmod& F) ;
FIELD_TYPE field_type (algnum& F) ;
FIELD_TYPE field_type (int& F) ;


inline int to_int(const bigint& n) {
#ifdef DEBUG
  if(!is_int(n)) {
    cout << "overflow casting " << n ;
    die(" from bigint to int.");
  }
#endif
    int a; n.intify(a); return a;
}

class ZZ; 
class ZZX;
ZZ to_ZZ (const bigint& n) ;
bigint to_bigint (const ZZ& n) ;
void to_polybigint (polynomial<bigint>& g, const ZZX& f) ;
// returns the d such that d*f = g.
bigint to_polybigint (polynomial<bigint>& g, const Vector<bigrational>& f) ;
void to_polybigrational(polynomial<bigrational>& g, const Vector<bigrational>& f) ;
void to_bigrationalvec (Vector<bigrational>& g, const polynomial<bigrational>& f) ;
void to_bigrationalvec (Vector<bigrational>& g, const ZZX& f) ;
void to_bigrationalvec (Vector<bigrational>& g, const Vector<bigint>& f);
void to_bigintvec (Vector<bigint>& g, const ZZX& f) ;
void to_ZZX (ZZX& g, bigint& denom, const Vector<bigrational>& f);
void to_bigrationalmat (Matrix<bigrational>& B, const bigint_matrix& A) ;

void to_algnum (Vector<algnum>& g, const Vector<bigrational>& f);

// view v = v[0] + v[1]*x + ... + v[n]*x^n
bigfloat to_bigfloat(const Vector<bigrational>& v, const bigfloat& x);

// algnum --> charmod reduction
// alpha = sum a_i x^i with a_i in Q.  
// to_charmod reduces the a_i modulo charmod::get_pr() (assuming this
// is possible!), then replaces x^i by a^i to obtain
// an element of F_p.  For this to make sense the user should
// assure that a is a root of the irreducible polynomial
// defining the current algebraic number field.
/*
charmod to_charmod (const algnum& alpha, charmod z);
bool can_reduce_charmod(const algnum& alpha);
void to_charmod(Matrix<charmod>& B,
		const Matrix<algnum>& A, charmod z);
bool can_reduce_charmod(const Matrix<algnum>& A);
void to_charmod(Vector<charmod>& w,
		const Vector<algnum>& v, charmod z);
bool can_reduce_charmod(const Vector<algnum>& v);
void to_charmod(Array<charmod>& w,
		const Array<algnum>& v, charmod z);
bool can_reduce_charmod(const Array<algnum>& v);
*/

bool can_reduce_charmod(const algnum& alpha);
bool can_reduce_charmod(const charmod&);
bool can_reduce_charmod(const intmod&);
bool can_reduce_charmod(const bigrational&);
bool can_reduce_charmod(const ffield_element&);
charmod to_charmod (const algnum& alpha, charmod z=0);
charmod to_charmod (const charmod& alpha, charmod z=0);
charmod to_charmod (const intmod& alpha, charmod z=0);
charmod to_charmod (const bigrational& alpha, charmod z=0);
charmod to_charmod (const ffield_element& alpha, charmod z=0);

template <class F>
void to_charmod(Matrix<charmod>& B,
		const Matrix<F>& A, charmod z=0);
template <class F>
bool can_reduce_charmod(const Matrix<F>& A);

/*
template <class F>
void to_charmod(Vector<charmod>& w,
		const Vector<F>& v, charmod z);
template <class F>
bool can_reduce_charmod(const Vector<F>& v);
*/

template <class F>
void to_charmod(Array<charmod>& w,
		const Array<F>& v, charmod z);
template <class F>
bool can_reduce_charmod(const Array<F>& v);


bool can_reduce_intmod(const bigrational&);
intmod to_intmod (const bigrational& alpha, intmod z=0);
void to_intmod(Matrix<intmod>& B,
		const Matrix<bigrational>& A);


void to_bigcomplex(Vector<bigcomplex>& B,
		  const Vector<bigrational>& C);
void to_bigcomplex(Matrix<bigcomplex>& B,
		  const Matrix<bigrational>& C);


// IO. 
char* strip_nonumeric(const char* str);

bigrational parse_bigrational(const char* str);
bigint parse_bigint(const char* str);

bool parse_bigrational_poly(Vector<bigrational>& v, const char* str);
bool parse_bigint_poly(Vector<bigint>& v, const char* str);

int parse_int(const char* str);


#endif // FIELDS_H



