/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/* 
   FILE: h1nchi.cc
   DESC: Modular symbols of weight k>=2 of any field.
*/

#include <fstream.h>  
#include "NTL/ZZ.h"
#include "system.h"
#include "fields.h"
#include "h1nchi.h"
#include "sparse.h"
#include "sparsematrix.h"
#include "cusps.h"
#include "arith.h"
#include "modformulas.h"
#include <LiDIA/nmbrthry_functions.h>

const int HEILBRONN_CACHE_MAXP=500; // number of Heilbronn matrices to precompute.

template <class F>
H1NChi<F>::HCache H1NChi<F>::Heil;

template <class F>
H1NChi<F>::H1NChi (const Character& _chi, int _k,
		   bool _save, bool _plus)
    : chi(_chi) , msymbols (_chi,_k,_save), k(_k),
      save (_save), plus(_plus)
{
    N = chi.modulus () ;
    init_mesg();
    init_fname() ;
    init_BINOM();
    init_Mk() ;
    init_Heilbronn();
    init_delta = init_sk = init_sknew = init_eis = false ;
    init_Wq();
    mesg("Initialization of Manin symbols data complete.\n");
}

template<class F>
void H1NChi<F>::init_mesg() const
{
    mesg("***********************************************************\n");
    mesg("Initializing M_",k);
    mesg("("); 
    char s[1024]; chi.ppname(s);
    mesg(s);
    mesg(")");
    if(plus) mesg("^+");
    mesg("\n");
    F x;
    mesg("Base field = ");mesg(field_name(x));mesgln(".");
    mesg("***********************************************************\n");
}

template <class F>
H1NChi<F>::~H1NChi ()
{
    term_Wq();			// call before deleting fname, as it _uses_ fname!
    if (save) delete fname ;
}

template <class F>
void H1NChi<F>::error (const char* msg, etype e) const 
{
    cerr << "H1NChi Error: " << msg << endl ;
    if (e == TERMINAL)
	exit(-1) ;
}

template <class F>
void H1NChi<F>::assert(bool assertion, const char* func_name,
		       const char* mesg) const 
{
#ifdef DEBUG
    if (!assertion) {
	cerr << "H1NChi assertion failed in function " << func_name << endl
	     << "'" << mesg << "'" << endl ;
	exit(-1);
    }
#endif
}

template <class F>
int H1NChi<F>::num_Heilbronn (const int& p) const
{
    Array<int> H ;
    Heilbronn (p, H) ;
    return H.length();
}


template <class F>
void H1NChi<F>::init_fname ()
{
    if (!save) return ;
    fname = new char[1024] ;
    char b[512] ;  chi.name(b) ;  
    F x ;
    if(plus)
	sprintf (fname, "%s_H1NChi_%d_%s_", b, k, field_name(x)) ;
    else
	sprintf (fname, "%s_H1NChiALL_%d_%s_", b, k, field_name(x)) ;
}


template <class F>
void H1NChi<F>::init_BINOM () 
{
    mesg ("Making binomial table.");
    /*
      F dummy;
      char s[512]; sprintf(s,"BINOM_%d_%s",k,field_name(dummy));
      ifstream in(data_read(s,N));
      if (save && in) 
      BINOM.Load(in);
      else {
    */
    BINOM.init(k-2+1,k-2+1);	// I know I could shrink this
				// by using properties of binomial
				// coefficients.
    for(int i=0; i<=k-2; i++) {
	for(int j=0; j<=k-2; j++) 
	    BINOM(i+1,j+1)=F(binom(bigint(i),bigint(j)));
    }
    mesg(" Done.\n");
    /*
      if (save) {
      ofstream out(data_write(s,N));
      if (out) BINOM.Save(out);
      }
      }
    */
}


template <class F>
void H1NChi<F>::act_on_MSym_q (Array<F>& coeff, const int* m,
			       const int& q) const
{
    if(k==2) {
	coeff[0]=F(1);
	return;
    }

    int a=m[0], b=m[1], c=m[2], d=m[3]; 
    
    Array<F> aXbY(k-2+1), cXdY(k-2+1);
    if (a==0) {
	power (aXbY[0], F(b), q);
    } else if (b==0) {
	power (aXbY[q], F(a), q);
    } else {
	for (int i=0; i<=q; i++) {
	    F ai, bqi;
	    power(ai,F(a),i); power(bqi,F(b),q-i);
	    //      aXbY[i]=ai*bqi*F(binom(bigint(q),bigint(i)));
	    aXbY[i]=ai*bqi*BINOM(q+1,i+1);  // the "+1" is just because
	    // BINOM is a matrix so
	    // indexing starts at 1.
	}
    }
    
    if (c==0) {
	power(cXdY[0],F(d),k-2-q);
    } else if (d==0) {
	power(cXdY[k-2-q],F(c),k-2-q);
    } else {
	for (int i=0; i<=k-2-q; i++) {
	    F ci, dk2qi;
	    power(ci,F(c),i); power(dk2qi,F(d),k-2-q-i);
	    //      cXdY[i]=ci*dk2qi*F(binom(bigint(k-2-q),bigint(i)));
	    cXdY[i]=ci*dk2qi*BINOM(k-2-q+1,i+1);
	}
    }
    
    for(int qq=0; qq<=k-2; qq++) 
    				// all ways to sum two naturals and obtain i.
	for (int j=max(0,qq + q-k-2); j<=min(qq, q); j++) {
	    coeff[qq] += aXbY[j]*cXdY[qq-j];
	}
    
}


template <class F>
void H1NChi<F>::act_on_MSym(Array<int>& idx, Array<F>& coeff,
			    const int& i, const int* m,
			    ACTION_TYPE action=HECKE,
			    int param=0,
			    bigint spec_u=0, bigint spec_v=0, int spec_q=0) const 
{
    int q;
    bigint u,v ;
    MSym x;
    if (action == ETA_LOWER) 
	x = MSym(spec_u,spec_v,spec_q);
    else
	x = msymbols.x(i);
    
    u=x.u(); v=x.v(); q=x.q();
    
    int a=m[0], b=m[1], c=m[2], d=m[3]; 
    
    bigint aa, bb ;
    bigint ga, gb, gc, gd ;
    
    switch (action) {
    case HECKE:
	aa = bigint(a)*u+bigint(c)*v;
	bb = bigint(b)*u+bigint(d)*v;
	
	if (gcd(N,gcd(aa,bb))!=1) {	// occurs when computing U_p
	    idx.reset(1); idx[0]=1;	
	    coeff.reset(1); coeff[0]=0; return;
	}
	break;
	
    case ATKIN_LEHNER:
	// for readability I would make the following assignments, but
	// C++ doesn't allow them in a case statement. 
	// M=param;
	// pi_M=spec_u, pi_MoverN=spec_v;
	if (((bigint(a)*u+bigint(c)*v) % param != 0) ||
	    ((bigint(b)*u+bigint(d)*v) % param != 0)) {
	    idx.reset(1); idx[0]=1;	
	    coeff.reset(1); coeff[0]=0; return;
	}
	x.abcd(ga, gb, gc, gd, N) ;       
#ifdef DEBUG
	assert((gc-u)%N==0&&(gd-v)%N==0,"act_on_MSym","error!");
	assert(ga*gd-gb*gc==1,"act_on_MSym","abcd error!");
#endif
	/*  
	    bigint gMa=ga*a+gb*c, gMb=ga*b+gb*d,
	    gMc=gc*a+gd*c, gMd=gc*b+gd*d;
	*/
	
	aa = (ga*a+gb*c)*spec_u + (gc*a+gd*c)*spec_v ;
	bb = (ga*b+gb*d)*spec_u + (gc*b+gd*d)*spec_v;
	
	break ;
	
    case ETA_LOWER:
	aa = bigint(a)*u+bigint(c)*v;
	bb = bigint(b)*u+bigint(d)*v;
	if (((aa % param) != 0) ||
	    ((bb % param) != 0)) {
	    idx.reset(1); idx[0]=1;	
	    coeff.reset(1); coeff[0]=0; return;
	}
	aa /= param;
	bb /= param; 
	break ;
    }
    
    aa=aa%N; if(aa<0)aa+=N; bb=bb%N; if(bb<0)bb+=N;
    int lambda, index;
    msymbols.MSym_to_list(index, lambda, MSym(aa,bb,q));
    F zeta(chi.of(lambda));
    
    idx.reset(k-2+1); coeff.reset(k-2+1);
    int base_idx = index - q;	// the index of MSym(aa,bb,0).

    act_on_MSym_q (coeff, m, q);// compute polynomial part of the action.
    
    for(int qq=0; qq<=k-2; qq++) {
	idx[qq] = base_idx+qq;
	coeff[qq] *= zeta;
    }
    
}

template <class F>
void H1NChi<F>::init_Mk_SI_table ()
{
  if(init_mk_si_table) return;

  init_mk_si_table=true;
    //                   S              I             S*I
    int m[3][4] = { {0, -1, 1, 0}, {-1, 0, 0, 1}, {0,-1,-1,0} } ;
    
    mesg("Step 1: mod out by S and I relations.");
    
    int max_idx = msymbols.length() ;  
    SI_rels.reset(max_idx+1);     // encodes M-symbols in relations
    SI_coeff.reset(max_idx+1);    // encodes coefficients in relations
    SI_gens.reset(max_idx+1);     // gives generators which, after modding
				// out by the S and I relations, satisfy
				// no further relations.  This array
				// will be resized later as it doesn't
				// have to be this big.
    for (int i = 1 ; i <= max_idx ; i++) {
	SI_rels[i] = 0 ; SI_coeff[i] = F(0) ; SI_gens[i] = 0 ; 
    }
    
    int cur_gen = 1 ;
    for (int i = 1; i <= max_idx; i++) 
	if (SI_rels[i] == 0) {
	    SI_gens[cur_gen] = i ;
	    SI_rels[i] = cur_gen ;
	    SI_coeff[i] = F(1) ;
	    int id[4]; 
	    id[0] = i ;
	    
	    bool is_trivial = false ;
	    for (int j=0; j<(plus?3:1); j++) {
		Array<int> n;  Array<F> c;
		act_on_MSym(n, c, i, m[j]);
		if (n.length() != 1) {
		    int num_nonzero=0, nonzero_one=0;
		    for (int l=0; l<n.length(); l++)
			if (c[l] != F(0)) {
			    num_nonzero++;
			    nonzero_one = l ;
			}
#ifdef DEBUG
		    if (num_nonzero>1 || num_nonzero==0)
			error ("init_SI_table can't work -- SI don't preserve lines.");
#endif
		    if (nonzero_one>0) {
			n[0]=n[nonzero_one];
			c[0]=c[nonzero_one];
		    }
		}
		if (j!=1) c[0] = -c[0];	// x-xI=x+xS=x+xSI=0, so need a minus
				// sign in all but the I case. 
		
		id[j+1] = n[0]; 
		
		if (SI_rels[n[0]] != 0 &&    // the relations kill this basis symbol.
		    SI_coeff[n[0]]*c[0] != F(1)) {
		    is_trivial=true;	
		}
		
		if (!is_trivial) {
		    SI_coeff[n[0]] = F(1) / c[0] ;
		}
		SI_rels[n[0]] = cur_gen ;
	    }
	    
	    if (is_trivial) 
		for (int j = 0 ; j < (plus?4:2) ; j++) {
		    SI_rels[id[j]] = 1 ;
		    SI_coeff[id[j]] = F(0) ;
		}
	    else 
		cur_gen++ ;
	}
    SI_gens.resize(cur_gen) ;
    mesg(" Done.\n");
}



template <class F>
void H1NChi<F>::init_Mk_add (Vector<F>& row,
			     const int& n, const int* matrix)
{
    Array<int> idx ; 
    Array<F> coeff ; 
    act_on_MSym (idx, coeff,
		 SI_gen(n), matrix) ;
    for(int i=0;i<idx.length();i++) {
	row[SI_rel(idx[i])] += coeff[i]*SI_c(idx[i]);
    }
}

template <class F>    
void H1NChi<F>::init_Mk_saverow (Vector<int>& r,
				 Vector<int>& c,
				 Vector<F>& val,
				 int& i,
				 int& row_num,
				 const Vector<F>& row_val) const
{
    int ii=i;
    for(int j=1; j<=row_val.size(); j++)
	if (row_val[j]!=F(0)) {
	    if (r.size()<i) {
		int newsize=r.size()+r.size()/5+1;
		r.resize(newsize); c.resize(newsize); val.resize(newsize);
	    }
	    r[i]=row_num;
	    c[i]=j;
	    val[i++]=row_val[j];
	}
    if (ii!=i)
	row_num++ ;
}

template <class F>
void H1NChi<F>::init_Mk_BuildRelMat (Sparse<F>& RelMat)
{
    mesg("Building Manin symbols relations matrix.\n");
    
    int T[]={0,-1,1,-1}, T2[]={-1,1,-1,0};
    int S[]={0, -1, 1, 0}, I[]={-1, 0, 0, 1}, IS[]={0,1,1,0};
    int ST[]={-1,1,0,-1}, ST2[]={1,0,-1,1};
    int IT[]={0,1,1,-1}, IT2[]={1,-1,-1,0}; 
    int IST[]={1,-1,0,-1},IST2[]={-1,0,-1,1};
    int IDEN[]={1,0,0,1};
    //  int m[3][4] = { {0, -1, 1, 0}, {-1, 0, 0, 1}, {0,-1,-1,0} } ;
    // Then m-space has basis v_1, ..., v_{max_idx}.
    int n, row ;
    

    init_Mk_SI_table () ;	

    int max_idx = SI_numgens() ;  
    start_timing ("init_Mk, # cols of relations matrix = ", max_idx);
    
    Vector<int> r(1), c(1) ;
    Vector<F> val(1) ;
    int rownum = 1 ;
    int i=1 ;
    for (n=1; n<=max_idx; n++) {
	if (n%100==0) { mesg(n); mesg(" "); }
	
	Vector<F> row(max_idx) ;
	init_Mk_add (row, n, IDEN);
	init_Mk_add (row, n, T);
	init_Mk_add (row, n, T2);
	init_Mk_saverow (r, c, val, i, rownum, row);
	
	init_Mk_add (row, n, S);
	init_Mk_add (row, n, ST);
	init_Mk_add (row, n, ST2);
	init_Mk_saverow (r, c, val, i, rownum, row);
	
	if(plus) {
	    init_Mk_add (row, n, I);
	    init_Mk_add (row, n, IT);
	    init_Mk_add (row, n, IT2);
	    init_Mk_saverow (r, c, val, i, rownum, row);
	    
	    init_Mk_add (row, n, IS);
	    init_Mk_add (row, n, IST);
	    init_Mk_add (row, n, IST2);
	    init_Mk_saverow (r, c, val, i, rownum, row);
	}
    }		
    if (i>1) {
	r.resize(i-1); c.resize(i-1); val.resize(i-1);
    } else {
	r[1]=1; c[1]=1; val[1]=F(0);
    }
    
//    cout << "The sparse relation matrix is as follows:"<<endl;
//    cout << "{Rel=[ \\\\ [row,column,value]"<<endl;
//    for(int i=1; i<=r.size();i++) 
//	cout << "[" << r[i]<<","<<c[i]<<","<<val[i]<<"] ";
//    cout << endl << "]}" << endl;
    RelMat.BuildFromSparseRep (r, c, val, rownum-1, max_idx) ;
}

    
template <class F>
void H1NChi<F>::init_Mk () 
{
    mesg("init_Mk: Computing M_k=(Manin Symbols)/(S, T, etc relations).\n");

    init_mk_si_table=false;
    init_Mk_SI_table () ;		// should be auto-save/loaded

    if (load_Mk()) return ;
    
    if (SI_numgens()==0) {
        mesg("The space has dimension zero - functions besides dim\n");
        mesg("may not work correctly.\n");
	dim_m = 0 ;
	return ;
    }

    // Mk   := list of free generators of M_k
    // Lift := expression of each symbol/<S,I> in terms
    //         of the above basis for Mk.
    
    Sparse<F> Sp;
    init_Mk_BuildRelMat (Sp);

/*
      Matrix<F> M;
      Sp.as_Matrix(M);
      cout << "Relation Matrix: M= " << M << endl ;
*/

    start_timing("Finished creating, now using sparse reduction");    
    Sp.kernel (Lift_data, Mk, Lift, true) ;
    dim_m = Mk.size();
    //  cerr << "Lift = " << Lift << endl
    //       << "Lift_data = " << Lift_data << endl 
    //       << "Mk = " << Mk << endl ;
    

    mesg("Done, now saving.\n");

    save_Mk();
    stop_timing();
}

template <class F>
Vector<F> H1NChi<F>::msymidx_to_free (const int& i) const 
{
#ifdef DEBUG
    assert (1<=i && i<=Lift_data.size(),
	    "msymidx_to_free", "Something is a matter, so terribly a matter!");
#endif
    if (Lift_data[i]<0) {
	// construct a row with a one in the -Lift_data[i] th position.
	Vector<F> v; v.init(Lift.ncols());
	v[-Lift_data[i]] = F(1);
	return v ;

    } else
	return Lift.row(Lift_data[i]);
}

template <class F>
void
H1NChi<F>::ImageOfManin(Matrix<F>& M) 
{
    M.init(dim_Mk(),Lift_data.size());
    for(int c=1;c<=Lift_data.size();c++)
	M.setcol(c,msymidx_to_free(c));
}



/* Here's what \delta "does": The Manin symbol (c,d) corresponds
   to the path (a,b;c,d){0,inf} = geodesic path from [b/d] to [a/b].  
   \delta sends (c,d) to the difference of the two endpoints 
   of that npath.  If the difference is zero then the path
   is a closed loop on the quotient curve X(gamma), otherwise 
   it is a path.

   The matrix computed below defines the map 
       M_k(N) --->> C' subset C(N)
   where C(N) is the vector space on the cusps
   and C' is the IMAGE of the delta map.  
*/

template <class F>
void H1NChi<F>::init_Delta ()
{
    if(init_delta) return ;	// already initialized
    mesg ("Computing the boundary map (Modular symbols) --> C[cusps].\n");
    mesg ("The cusps ");
    if(plus) mesg (" (modulo complex conjugation!) ");
    mesg ("are: ");
    Cusps cusps(chi, k, plus) ;
    
    Delta.init(10,dim_m) ; 
    int max_row = 0 ;
    for (int i=1; i<=dim_m; i++) {
	bigint a,b,c,d;
	msymbols.abcd (a, b, c, d, SI_gen(Mk[i])) ;
				// change this when cusp class is re-written!
	//    if (!(is_int(aa) && is_int(bb) && is_int(cc) && is_int(dd))) 
	//      error("Overflow in computing Ker(Delta).");
	//    int a,b,c,d; 
	//    aa.intify(a); bb.intify(b); cc.intify(c); dd.intify(d);
	int idx; rootof1 coeff ;
	MSym x = msymbols.x(SI_gen(Mk[i]));
	if(x.q()==k-2) {
	    cusps.free(a,c,idx,coeff); 
	    if(Delta.nrows()<idx) Delta.resize(idx+10,dim_m);
	    Delta(idx,i)+=F(coeff) ;
	    if(max_row<idx) max_row=idx;
	} 
	if (x.q()==0) {
	    cusps.free(b,d,idx,coeff); 
	    if(Delta.nrows()<idx) Delta.resize(idx+10,dim_m);
	    Delta(idx,i)-=F(coeff) ;
	    if(max_row<idx) max_row=idx;
	}
    }
    Delta.resize(max_row,dim_m);
    mesg (" Done.\n");
    init_delta=true;
}

template <class F>
int H1NChi<F>::dim_Sk ()
{
    if (!init_sk) init_Sk() ;
    return Sk.ncols();
}

template <class F>
void H1NChi<F>::init_Sk ()
{
    if(init_sk) return;
    init_sk=true;
    if (load_Sk()) return ;
    start_timing ("Computing ker(Delta)");
    init_Delta();
    
    Kernel(Sk,Delta);
    stop_timing() ;
    save_Sk();
}

template<class F>
void H1NChi<F>::init_Eis ()
{
    if(init_eis) return;
    init_eis=true;
    Eis.idmat(dim_Mk());
    start_timing ("Initializing Eisenstein space (complement of Sk)\n");
    // Compute successive Tp, until the Eisenstein space is split off.
    // p <- 2
    int p=2;
    // Compute Tp
    mesg("Eis space has dimension ",dim_Mk()-dim_Sk());

    do {
	mesg("\nUsing T_",p);
	Matrix<F> T;
	Tp_Mk(p,T);
	
	// f = charpoly (T_p | M_k)
	Vector<F> f; T.charpoly(f);
	
	Matrix<F> TonS;
	Tn_Sk(p,TonS);
	// g = charpoly (T_p | S_k)
	Vector<F> g; TonS.charpoly(g);
	
	// h is divisible by charpoly (T_p | Eis)
	Vector<F> h = poldiv(f,g);
	
	// Restrict T_p to Eis and replace Eis by ker h(T_p).
	Matrix<F> A;
	T.RestrictedTo(A, Eis);
	Kernel(Eis, A, h);
	mesg(", dim=",Eis.ncols());
	p=next_prime(p);
    } while (Eis.ncols() > dim_Mk() - dim_Sk()) ;
    stop_timing() ;
}

template<class F>
void H1NChi<F>::Tn_SparseVector (Vector<F>& v, // v = T_n * x
				 const Vector<F>& x,
				 const Vector<int>& nonzero,
				 int n) const
{
    if(nonzero.size()<1) {
	int n=1;
	for(int i=1;i<=x.size();i++) 
	    if(x[i]!=F(0))
		nonzero[n++]=i;
    }
    v.init(dim_Mk());
    Array<int> H;
    Tp_Heilbronn(n,H);
    for(int i=1;i<=nonzero.size();i++) {
	Vector<F> w;
	Tp_StandardSymbol (w, H, nonzero[i]);
	v += x[nonzero[i]]*w;
    }
}


template <class F>
void H1NChi<F>::Tp_StandardSymbol (Vector<F>& v, Array<int>& H, int i) const
{  
    if (i<1 || i>dim_m) {
	cerr << "H1NChi::Tp_StandardSymbol, invalid i="<<i<<", 1<=i<="<<dim_m<<endl;
	die();
    }

    v.init(dim_Mk());
    Array<int> idx; 
    Array<F> c; 

    for (int j=0; j<H.length()/4; j++) {
	act_on_MSym(idx,c,SI_gen(Mk[i]),&H[4*j]);
	for(int t=0;t<idx.length();t++) 
	    if(c[t]!=F(0)) 
		v += c[t]*index_to_free(idx[t]);
    }
}

template <class F>
void H1NChi<F>::Tp_StandardSymbol (Vector<F>& v, int i, int p) const
{  
    Array<int> H;
    Tp_Heilbronn(p,H);
    Tp_StandardSymbol(v,H,i);
}


template<class F>
void H1NChi<F>::standard_to_free(Vector<F>& v,
				 const Vector<F>& w) const
{
    if(w.size()!=msymbols.length())
	die("standard_to_free -- invalid w.");
    v.init(dim_Mk());
    for(int i=1; i<=w.size();i++)
	if(w[i]!=F(0))
	    v+=w[i]*index_to_free(i);
}


template <class F>
void H1NChi<F>::Tp_Mk (int p, Matrix<F>& M) const
{  
    
    if(p<=0) 
	die("H1NChi::Tp_Mk -- Can't compute Hecke operator T_n for n negative.");

    if(p==1) {
	M.idmat(dim_Mk());
	return;
    }

    max_Tp_Got = max(p,max_Tp_Got); // This is for accounting purposes by clients only.
                                    // doesn't affect algorithm.

    if (load_Tn_Mk(p, M)) return ;
    start_timing ("Computing Tp on Mk, p = ", p) ;
    
    Array<int> H ;
    Tp_Heilbronn(p, H);
    M.init (dim_m, dim_m) ;
    Vector<F> v(dim_m) ;
    
    mesg("[");
    for (int i = 1 ; i <= dim_m; i++) {
	Tp_StandardSymbol (v, H, i) ;
	
	if(i%25==0) { mesg(i*100/dim_m); mesg("% "); }
	M.setcol (i, v) ;
    }
    
    mesg("100%] ");
    stop_timing();
    mesgln("\nSaving T",p); 
    save_Tn_Mk (p, M) ;
}

template <class F>
void H1NChi<F>::CyclicHeckeModule(Matrix<F>& Tx,
				  const Vector<F>& x)
{
    // Step 1: find nonzero entries
    Vector<int> nz(x.numnonzero()); // nonzero positions of x.
    int n=1;
    for(int i=1;i<=x.size();i++) 
	if(x[i]!=F(0))
	    nz[n++]=i;
    
    // Step 2: Compute sturm bound on generators
    ModularFormulas formulas;
    int b= k*formulas.idxG0(N)/12;	  // Sturm's bound.
    if(chi.type()!=0) {
	mesgln("Warning -- Sturm bound used in ");
	mesgln("CyclicHeckeModule might be wrong.");
    }
    mesg("Cyclic module (1-",b);
    mesg("):  ");


    // Step 3. compute hecke span.
    Tx.init(dim_Mk(),b);
    Tx.setcol(1,x);		// T_1*x
    Vector<F> Tnx;

    for(int n=2;n<=b;n++) {
	mesg(" ",n);
	Tn_SparseVector(Tnx,x, nz, n);
	Tx.setcol(n,Tnx);
    }
    mesgln(" done.");
}

template<class F>
void H1NChi<F>::CyclicHeckeImage(Matrix<F>& PTx,
				 const Matrix<F>& P,
				 const Vector<F>& x,
				 int b=0)
{
    mesgln("CyclicHeckeImage");
    if((plus && P.nrows()==1) || (!plus && P.nrows()==2)) {
	mesgln("Shortcut: By dimension count, Hecke should act as scalars on P.");
	PTx.init(P.nrows(),1);
	PTx.setcol(1,P*x);
	return;
    }

    /* Let Mn be the matrix representing the right
       action of the Hecke operator Tn on P.
       We must compute gens for an F-module who (Z-span) equals
             span {Mn*c : 1<=n<=b}
       where b is a bound on the number of Tn necessary
       to generate the Hecke algebra as a Z-module.

       We do this by:
       1) Computing Mp for p prime using Hecke action on easy elements of M_k(Q).
       2) Iteratively computing prod Mp^i*c where n=prod p^i varies from 1 to b.
    */

    Vector<F> c=P*x;

    if(b==0) {
	// Compute Sturm's bound b.
	ModularFormulas formulas;
	b = k*formulas.idxG0(N)/12;	  
	if(chi.type()!=0) {
	    mesgln("Warning -- Sturm bound used in ");
	    mesgln("CyclicHeckeModule might be wrong.");
	}
	mesgln("Sturm = ",b);
    }
    PTx.init(P.nrows(),b);

    // Prepare to compute Hecke action.
    Matrix<F> V; P.transpose(V);
    Vector<int> e; Matrix<F> D;
    Dual_E (e, D, V);

    // Compute Hecke operators.
    Matrix<F>* T=new Matrix<F>[b+1];
    for(int p=2; p<=b; p++)
	if(is_prime(p)) {
	    mesg(" ",p);
	    Dual_Tp_On(T[p],p,V,e,D);
	}

    mesgln("Computing generators.");

    // Fill in generators.
    PTx.setcol(1,c);
    for(int n=2; n<=b; n++) {
	if (is_prime(n))
	    PTx.setcol(n,T[n]*c);
	else {
	    // find smallest p|n and set the n-th column to
	    // Tp*(n/p-th column). This works since the n/p-th
	    // column has already been computed. 
	    int p=to_int(smallest_prime_divisor(n));
	    PTx.setcol(n,T[p]*PTx.col(n/p));
	}
    }
    delete[] T;
    
}


template <class F>
void H1NChi<F>::Tpr_Mk (int p, int r, Matrix<F>& M) const
{
    switch(r) {
    case 0: M.idmat(dim_m); break;
    case 1: Tp_Mk(p, M);  break;
    default:
	if (chi.type()==-1) {
	    // T_{p^r}, r>1 not programmed for Gamma_1(N) forms, but 
	    // can't simply use Heilbronn matrices!
	    bigint pr;
	    power(pr,p,r);
	    Tp_Mk(to_int(pr),M);
	    return;
	}
	if (N%p != 0) {		// if p does not divide the level N
	    Matrix<F> A, B, C ;
	    Tpr_Mk (p, r-1, A) ;
	    Tpr_Mk (p, 1, B) ;
	    Tpr_Mk (p, r-2, C) ;
	    bigint q;			// q=p^{k-1}
	    power(q,p,k-1);
	    M = A*B - F(chi.of(p))*F(q)*C;
	} else {			// p divides N
	    Matrix<F> A ;
	    Tp_Mk (p, A) ;
	    M = A^r ; 
	}
    }
}

template <class F>
void H1NChi<F>::Tn_Mk (int n, Matrix<F>& M) const
{
    if (n<0) {
	error("Invalid Tn arguments. n = ", NONTERMINAL);
	return ;
    }

    if (load_Tn_Mk (n, M)) return ;
    
    M.idmat(dim_m) ;
    if (n==1) return ;
    
    bool first_time=true;
    while (n > 1) {
	int p = 2;
	while (n % p) p++ ;
	int prime_to_p_part ;
	int r = ordp (p, n, prime_to_p_part) ;
	if (first_time && prime_to_p_part == 1) { // avoid the nasty stupid M*=M_pr. 
	    Tpr_Mk(p, r, M) ;		// with M=idmatrix.
	} else {
	    Matrix<F> M_pr ;
	    Tpr_Mk(p, r, M_pr) ;
	    M *= M_pr ;
	}
	n = prime_to_p_part ;
	first_time=false;
    }
}

template <class F> 
void H1NChi<F>::Tn_Mk_Eig (int n, const F& lambda, Matrix<F>& ES) const
{
    start_timing("Tn_Mk_Eig n = ", n) ;
    Matrix<F> M ;			// M will equal T-lambda*I
    Tn_Mk (n, M); 
    for (int i=1; i<=M.ncols(); i++) M(i,i)-=lambda;
    stop_timing();
    start_timing("Tn loaded, now computing kernel");
    Kernel(ES, M);
    stop_timing();
}

template <class F> 
void H1NChi<F>::Tn_Mk_Eig (int n, const Vector<F>& f, Matrix<F>& ES) const
{
    start_timing("Tn_Mk_Eig n = ", n) ;
    Matrix<F> fT, T ;			// fT will equal f(Tn)
    Tn_Mk (n, T); 
    T.polsubst(fT, f);
    T.init(0);
    stop_timing();
    start_timing("Tn loaded and f(Tn) computed, now computing kernel.");
    Kernel(ES, fT);
    stop_timing();
}


template <class F>
void H1NChi<F>::ZmodNZ_Mk(Matrix<F>& U, int u) const
{
    mesgln("ZmodNZ, u=",u);

    U.init(dim_Mk(),dim_Mk());
    if (gcd(u,N)!=1) 
	return;

    Matrix<bigint> G;
    MSym x(0,u);
    x.abcd(G,N);

    Vector<F> v;
    for(int i=1; i<=dim_Mk(); i++) {
	if(i%25==0) { mesg(i*100/dim_Mk()); mesg("% "); }
	ActStandardModularSymbol(v, G, i);
	U.setcol(i, v);
    }
    
}

template <class F>
void H1NChi<F>::ZmodNZ_On (Matrix<F>& U, const Matrix<F>& V, int u) const
{
    if (dim_Mk() != V.nrows ()) {
	error("ZmodNZ_On: V doesn't sit inside M_k(Q).");
	return ;
    }
    Matrix<F> A;
    ZmodNZ_Mk(A,u);
    A.RestrictedTo(U, V) ;	// U = A|V.
}

template<class F>
void
H1NChi<F>::init_Wq()
{
    Vector<bigint> base,exp;
    Factor(base,exp,N);
    WqList.init(base.size());
    if(save) {
	char name[strlen(Name())+8];
	sprintf(name,"%sWq",Name());
	ifstream in(data_read(name,N)) ;
	if (in) WqList.Load(in);
    }
}

template<class F>
void
H1NChi<F>::term_Wq()
{
    if(!save) return;

    char name[strlen(Name())+8];
    sprintf(name,"%sWq",Name());
    ofstream out(data_write(name,N)) ;
    if (out) WqList.Save(out);
}

template<class F>
const Matrix<F>& 
H1NChi<F>::get_Wq(int i)
{
    mesgln("get_Wq");
    if(i>WqList.size())
	die("H1NChi<F>::get_Wq -- i doesn't give a divisor of N.");
    if(i>WqList.len() || WqList.cM(i-1).nrows()==0) { // must compute it
	Vector<bigint> base, exp;
	Factor(base,exp,N);
	Wq_Mk(WqList.M(i-1), to_int(base[i]));
    }
    mesgln("Done (get_Wq).");
    return WqList.cM(i-1);
}

template<class F>
void H1NChi<F>::Wq_Mk (Matrix<F>& Wq, int p) 
{
    mesgln("W_",p);
    if(!is_prime(p)) {
	mesgln("Wq_Mk: p must be prime, but p = ",p);
	die();
    }
    Wq.init(dim_Mk(),dim_Mk());
    if(N%p != 0)		// Wq = identity matrix.
	return;
    int q=p;
    while((N/q)%p==0) q*=p;


    bigint x,y;
    xgcd(x,y,q,-N/q);		// x*q-y*N/q = 1
    if(x*q-y*N/q != 1) 
	die("Wq_Mk: error in xgcd?");
    Matrix<bigint> W(2,2);

    W(1,1)=q*x; W(1,2)=y;
    W(2,1)=N;   W(2,2)=q;

    Vector<F> v;
    for(int i=1; i<=dim_Mk(); i++) {
	if(i%25==0) { mesg(i*100/dim_Mk()); mesg("% "); }
	ActStandardModularSymbol(v, W, i);
	Wq.setcol(i, v);
    }

    if(F(q)!=F(0)) {
	bigint scale;
	power(scale,q,k/2-1);
	Wq /= F(scale);
    } else {
	mesg("**WARNING: Atkin-Lehner W",p);
        mesgln(" in characteristic ",p);
	mesgln("is not understood by the implimenter of this program.");
    }
}

template <class F>
void H1NChi<F>::ActStandardModularSymbol(Vector<F>& v,
					 const Matrix<bigint>& M,
					 int i) const
{
    // i-th standard symbol = [X^qY^(k-2-q),A] := A.(X^qY^(k-2-q){0,oo}).
    Matrix<bigint> A;
    int ith=IthStandardSymbol(i);
    msymbols.x(ith).abcd(A,N);
//    cout << "ActStandardModularSymbol: i = " << i << ", A = "
//	 << A(1,1) << ","<< A(1,2) << ";"<<A(2,1)<<","<<A(2,2) << endl
//	 << "q = " << msymbols.x(ith).q() << endl;
    Matrix<bigint> B=M*A;	
    ActWindingModularSymbol(v, B, msymbols.x(ith).q());
}

// Let M be in M_2(Z).  This function
// compute M.(X^qY^(k-2-q){0,oo}) in terms of the free basis.
template <class F>
void H1NChi<F>::ActWindingModularSymbol(Vector<F>& v,
				 const Matrix<bigint>& M,
				 int q) const
{
    Array<F> coeff;
    M2ZAction(coeff, M, q);
    v.init(dim_Mk());
    Vector<F> w;
    for(int i=0; i<=k-2;i++) {
	modular_to_free(w, M(1,2), M(2,2), M(1,1), M(2,1), i) ;
	v += coeff[i]*w;
    }			
}


// INPUT: a matrix M=(a,b;c,d) in M_2(Z) and a polynomial P(X,Y) = X^qY(k-2-q).
// OUTPUT: Array of coefficients of 
//            M.P := P(dX-bY, -cX+aY)
//                   coeff[0]*Y^(k-2)+ coeff[1]*X*Y^(k-3) +
//                            ... + coeff[k-2]*X^(k-2)
template <class F>
void H1NChi<F>::M2ZAction (Array<F>& coeff,
			   const Matrix<bigint>& M,
			   const int& q) const
{
    coeff.init(k-1);
    if(k==2) {
	coeff[0]=F(1);
	return;
    }

    // (a,b;c,d) := M^(-1)*det(M)
    F a(M(2,2)), b(-M(1,2)), c(-M(2,1)), d(M(1,1));

    Array<F> aXbY(k-2+1), cXdY(k-2+1);
    if (a==F(0)) {
	power (aXbY[0], b, q);
    } else if (b==F(0)) {
	power (aXbY[q], a, q);
    } else {
	for (int i=0; i<=q; i++) {
	    F ai, bqi;
	    power(ai,a,i); power(bqi,b,q-i);
	    aXbY[i]=ai*bqi*BINOM(q+1,i+1);  // the "+1" is just because
	    // BINOM is a "Matrix" so indexing starts at 1.
	}
    }
    
    if (c==F(0)) {
	power(cXdY[0],d,k-2-q);
    } else if (d==F(0)) {
	power(cXdY[k-2-q],c,k-2-q);
    } else {
	for (int i=0; i<=k-2-q; i++) {
	    F ci, dk2qi;
	    power(ci,c,i); power(dk2qi,d,k-2-q-i);
	    //      cXdY[i]=ci*dk2qi*F(binom(bigint(k-2-q),bigint(i)));
	    cXdY[i]=ci*dk2qi*BINOM(k-2-q+1,i+1);
	}
    }
    
    for(int qq=0; qq<=k-2; qq++) 
    				// all ways to sum two naturals and obtain i.
	for (int j=max(0,qq + q-k-2); j<=min(qq, q); j++) {
	    coeff[qq] += aXbY[j]*cXdY[qq-j];
	}
}


template <class F>
void H1NChi<F>::WN_Mk (int _NN, Matrix<F>& M) const
{  
    int NN=1;
    // force N/NN and NN to be coprime by increasing the powers of the
    // prime divisors of NN. 
    rational_factorization fac(N);
    fac.factor();
    for(int i=0;i<fac.no_of_comp();i++)
	if (_NN % fac.base(i)==0) {
	    bigint pn;
	    power(pn,fac.base(i),fac.exponent(i)); 
	    NN*=to_int(pn);
	}
    
#ifdef DEBUG
    if (chi.type()==-1) {
	cerr << "WARNING: Atkin Lehner involution for Gamma_1(N)" << endl
	     << "not programmed!" << endl ;
	// e.g., N=20, k=2 doesn't work.
	M.init(dim_m,dim_m);
	return ;
    }
#endif
    
#ifdef DEBUG
    if (chi.type()>0 && chi.degree()>2) {
	cerr << "Warning: Atkin Lehner involution isn't defined when "<<endl
	     << "character has degree > 2!"<<endl;
	M.init(dim_m,dim_m);
	return ;
    }
#endif
    
    if (load_WN_Mk(NN,M)) return;
    
    start_timing ("Computing");
    cerr << "W_{"<<NN<<"} on M_{"<<k<<"}  "<<flush;
    
    Array<int> H ;
    HeilbronnC (NN, H) ;		// it is *essential* to use HeilbronnC
				// and not Heilbronn!
    M.init (dim_m, dim_m) ;
    Vector<F> v(dim_m) ;
    
    // init CRT data:
    bigint pi_M=invmod(N/NN,NN)*N/NN, pi_NoverM=invmod(NN,N/NN)*NN;
    
    for (int i = 1 ; i <= dim_m; i++) {
	Wq_StandardSymbol(v, H,
			  i, NN,
			  pi_M, pi_NoverM);
	M.setcol(i, v) ;
	if(i%10==0) { mesg((i*100/dim_m)); mesg("% "); }
    }
    save_WN_Mk(NN,M);
    stop_timing();
}

template<class F>
void H1NChi<F>::Wq_StandardSymbol(Vector<F>& v,
				  const Array<int>& H,
				  int i, int q,
				  bigint& pi_M, bigint& pi_NoverM) const 
{
    v.zero () ;
    Array<int> idx; 
    Array<F> c; 
    for (int j=0; j<H.length()/4; j++) {
	act_on_MSym(idx,c,SI_gen(Mk[i]),&H[4*j],
		    ATKIN_LEHNER,q, pi_M, pi_NoverM) ;
	for(int t=0;t<idx.length();t++) {
	    v += c[t]*index_to_free(idx[t]) ;
	}
    }
}

template<class F>
void H1NChi<F>::Conjugation_Mk(Matrix<F>& C) const
{
    Array<int> J(4) ;
    J[0]=-1; J[1]=0; J[2]=0; J[3]=1;

    C.init (dim_m, dim_m) ;
    Vector<F> v(dim_m) ;
    
    mesgln("Conjugation_Mk");
    for (int i = 1 ; i <= dim_m; i++) {
	Tp_StandardSymbol (v, J, i) ;
	C.setcol (i, v) ;
    }
}

template <class F>
void H1NChi<F>::Tn_On (int n, const Matrix<F>& W, Matrix<F>& B) const
{
    if (dim_m != W.nrows ()) {
	error("TpResTo: asked to compute on space of invalid dimension.");
	return ;
    }
    mesgln("T", n) ;
    Matrix<F> T;
    Tn_Mk(n,T);
    T.RestrictedTo(B, W) ;	// B = T|W
}

template <class F>
void H1NChi<F>::Wq_On (Matrix<F>& Wq, const Matrix<F>& V, int q) 
{
    mesgln("H1NChi::Wq_on");
    if (dim_Mk() != V.nrows ()) {
	error("Wq_On: V doesn't sit inside M_k(Q).");
	return ;
    }
    Vector<bigint> base, exp; Factor(base,exp,N);
    int s=0;
    for(int i=1; i<=base.size();i++)
	if(base[i]==q) {
	    s=i;
	    break;
	}
    get_Wq(s).RestrictedTo(Wq, V) ;	// Wq = W|V.
    mesgln("Done (H1NChi::Wq_On).");
}


// THIS CODE CAN EASILY BE OPTIMIZED BUT HAS NOT BEEN YET!
template <class F>
Matrix<F>* H1NChi<F>::TnList_Mk(int n)
{
    Matrix<F>* T=new Matrix<F>[n];
    for(int i=1;i<=n;i++)
	Tn_Mk(i,T[i-1]);
    return T;
}

template <class F>
Matrix<F>* H1NChi<F>::TnList_On(int n, const Matrix<F>& V)
{
    Matrix<F>* T=new Matrix<F>[n];
    for(int i=1;i<=n;i++)
	Tn_On(i,V,T[i-1]);
    return T;
}

template <class F>
void 
H1NChi<F>::Tn_Eig_On (int n , const Matrix<F>& W,
		      const F& lambda, Matrix<F>& ES) const
{
    if (dim_m != W.nrows ()) 
	error ("Tn_Eig_On: Invalid dimension.");
    
    mesg ("Computing an eigenspace of T_", n);
    Matrix<F> T;
    Tn_On(n, W, T) ;
    for (int i=1; i<=T.nrows(); i++) T(i,i)-=lambda;
    Matrix<F> E;
    Kernel(E,T);
    ES=W*E;
    mesg("Done computing kernel.\n");
}

template <class F>
void 
H1NChi<F>::Tn_Eig_On (int n , const Matrix<F>& W,
		      const Vector<F>& f, Matrix<F>& ES) const
{
    if (dim_m != W.nrows ()) 
	error ("Tn_Eig_On: Invalid dimension.");
    
    mesg ("Computing the kernel of f(T_", n);
    mesg (") for some polynomial f(x).\n");
    Matrix<F> fT, T;
    Tn_On(n, W, T) ;
    T.polsubst(fT, f);
    T.init(0);
    Matrix<F> E;
    Kernel(E,fT);
    ES=W*E;
    mesg("Done computing kernel.\n");
}

template <class F>
void 
H1NChi<F>::Dual_Eig (Matrix<F>& V,
		     int n, const F& lambda) const
{
    start_timing("Dual_Eig n = ", n) ;
    Matrix<F> M ;			// M will equal T-lambda*I
    Tn_Mk (n, M); 
    M.transpose();
    for (int i=1; i<=M.ncols(); i++) M(i,i)-=lambda;
    stop_timing();
    start_timing("Tn loaded and transposed, now computing eigenspace");
    Kernel(V, M);
    stop_timing();

}

template <class F>
void 
H1NChi<F>::Dual_Eig (Matrix<F>& V,
		     int n, const Vector<F>& f) const
{
    start_timing("Dual_Eig f(T), n = ", n) ;
    Matrix<F> M, fM ;			// M will equal f(T_n^transpose)
    Tn_Mk (n, M); 
    M.transpose();
    M.polsubst(fM, f);
    M.init(0);			// free space
    stop_timing();
    start_timing("Tn loaded and transposed, now computing eigenspace.");
    Kernel(V, fM);
    stop_timing();

}

template <class F>
void 
H1NChi<F>::Dual_Eig_On(Matrix<F>& V,
		       int n, const F& lambda, const Matrix<F>& W) const
{
    Vector<F> f(2);
    f[1]=-lambda;
    f[2]=F(1);
    Dual_Eig_On(V,n,f,W);
}

template <class F>
void 
H1NChi<F>::Dual_Eig_On(Matrix<F>& V,
			int n, const Vector<F>& f, const Matrix<F>& W) const
{
    if (dim_m != W.nrows ()) 
	error ("Tndual_Eig_On: Invalid dimension.");
    
    mesg ("Computing the kernel of f(transpose(T_", n);
    mesg (")) for some polynomial f(x).\n");
    Matrix<F> fT, A, T;
    Tn_Mk(n,A);
    A.transpose();
    A.RestrictedTo(T,W);
    A.init(0);		// free space
    T.polsubst(fT, f);
    T.init(0);
    Matrix<F> E;
    Kernel(E,fT);
    V=W*E;
    mesg("Done computing Tndual_Eig_On.\n");
}


				// pr=list of pr.size() primes p_0, ..., p_s.
				// f = array of pr.size() polys f(x) in F[x].
				// computes intersect ker(f(T_{p_i}^t)).
				// Computes first T_{p_0} in the usual way,
				// then uses Dual_Tp_On	to compute each successive T_p.
/* Algorithm for Dual_Eig_Intersect
   1. Compute T = T_{pr[0]}^t using modular symbols.
   2. Find V=ker(f[0](T)) subset Hom(M_k,Q).
   3. For each i=1,...,s:
      a. Compute the "Dual_E" subspace E of M_k corresponding to V.
      b. Use E to quickly compute T' = T_{pr[i]} acting on V.
      c. Let V <-- ker(f[i](T')) 

*/ 

template <class F>
void 
H1NChi<F>::Dual_Eig_Intersect (Matrix<F>& V,
			       const Array<int>& pr,
			       const Vector<F>*& f,
			       int stop_dimension)
{
    if (pr.length()==0) {
      mesg("Dual_Eig_Intersect: using whole space.\n");
      V.idmat(dim_Mk());
      return;
    }
     
    mesg("Computing first eigenspace.\n");
    Matrix<F> Start;
    Dual_Eig(Start, pr[0], f[0]);
    mesg("Dimension = ", Start.ncols()); mesg("\n");
    if(Start.ncols()<1) {
      mesg ("No further eigenspaces will be computed.\n");
      V=Start;
      return;
    }
    
    Dual_Eig_Intersect(V, pr, f, Start, 1, stop_dimension);
}


template <class F>
void 
H1NChi<F>::Dual_Eig_Intersect (Matrix<F>& V,
			       const Array<int>& pr,
			       const Vector<F>*& f,
			       const Matrix<F>& Start,
			       int start_at=0,
			       int stop_dimension=0)
{
    V=Start;
    for(int i=start_at;i<pr.length();i++) {
      mesg ("Computing ", i+1);
      mesg("th eigenspace.\n");

      Matrix<F> T;
      Vector<int> e; Matrix<F> D;
      Dual_E(e,D,V);
      Dual_Tp_On(T,pr[i],V,e,D);
      //      mesg("V = "); cout << V << endl;
      //      mesg("T = "); cout << T << endl;

      T.transpose();  // since Dual_Tp_On gives the *right* action of the Hecke operator.
      Matrix<F> W;
      Kernel (W, T, f[i]);
      V = V*W;
      mesg("Dimension = ",V.ncols());
      if(V.ncols()<=stop_dimension)
	  return;
      mesg("\n");
    }
}

template <class F>
void 
H1NChi<F>::Dual_FindEigenvector (Matrix<F>& v, Vector<F>& K,
				 const Matrix<F>& V)
{
    /* Compute some linear combinations of Hecke operators until characteristic
       polynomial is irreducible.  Use this to find a left eigenvector
       over some field. */

    Vector<int> e;
    Matrix<F> EVinv;
    Dual_E(e, EVinv, V);

    Matrix<F> T, Tbig;
    T.init(V.ncols(), V.ncols());
    int p=2;
    while(1) {
        while(N%p == 0) p=next_prime(p);
	Matrix<F> Tp;
	Dual_Tp_On(Tp, p, V, e, EVinv);
        T += Tp;
        T.charpoly(K);
        if(K.pol_is_irreducible()) break;
        p=next_prime(p);
	if(p>97) {
	    mesgln("WARNING! H1NChi<F>::Dual_FindEigenvector -- V doesn't");
	    mesgln("seem irreducible!  I will keep trying...");
	}
    }

    T.transpose();
    Eigenvector(v, T, K);
    v = V*v;
}



/* Fast computation of T_p:

  Let V be an nxm matrix, viewed as a set of rows in an m dimensional space of 
  row vectors.  Suppose that V has rank n and that V*T is contained in the row
  space of V.  Let E be an mxn matrix so that the nxn matrix V*E is invertible,
  with inverse D.
    Proposition.  V*T = V*T*E*D*V.
    Proof.
      Observe that V*(E*D*V) = (V*E*D)*V = I*V = V.
      Thus right multiplication by E*D*V 
                       v  |----> v*E*D*V 
      induces the _identity map_ on the row space of V.
      Since V*T is contained in the row space of V, we have
                     (V*T)*E*D*V = V*T,
      as claimed.
    
  We obtain an algorithm for computing the action of T on V.
     Algorithm.
        Compute V*T = V*(T*E)*D*V quickly by computing T*E quickly. 
        The matrix representing the action of T with respect to the rows of V is:
                V*T*(any right inverse of V)
                  = V*(T*E)*D*V*(E*D)
                  = V*(T*E)*D*V*E*D.

        A right inverse A of V is a matrix so that V*A =I=nxn identity.

        Observe that, by the proposition, 
              V*T*(any right inverse of V) * V = V*T.
        Thus  V*T*(any right inverse of V) gives the correct linear combinations
        of the rows of V.  [too vague, but true.]
*/

template <class F>
void H1NChi<F>::Dual_E (Vector<int>& e,
			Matrix<F>& D,
			const Matrix<F>& W) const 
{
    mesg("Computing Dual_E... \n");

    // 1. [Compute e] e = set of positions so that elements of the space spanned by
    //    the columns of V are determined by the entries in these spots.
    //    This is accomplished by row reducing, and setting e equal to
    //    the pivot columns.

    // V <-- transpose of W
    Matrix<F> V; W.transpose(V);   
    int n=V.nrows(), m=V.ncols();
    Vector<int> elimcol, elimrow;
    V.SimpleEchelon(elimcol,elimrow);
    //    mesg("Found pivot columns... ");
    e.init(n);
    int i=1;
    for(int c=1;c<=elimcol.size();c++) 
	if(elimcol[c]!=0) 
	    e[i++]=c;
    if(i<=n) {
	cerr << "W = " << W << endl << ", V = " << V << endl;
	die("H1NChi::Dual_E -- Columns of V not linearly independent.");
    }
    
    // Before using V again, set back to the original matrix.
    W.transpose(V);  

    // E is the matrix whose columns correspond to the e[i]:
    Matrix<F> E; E.init(m,n);
    for(int i=1; i<=n; i++)
      E(e[i],i)=1;
    
    // The matrix V*E must be invertible.  Compute its inverse D.
    mesg("Computing (VE)^(-1).\n");
    (V*E).Inverse(D);
    //    cout << "Dual_E:"<<endl;
    //    cout << "V = " << V<<endl;
    //    cout << "E = " << E << endl;
    //    cout << "(VE)^(-1) = " << D <<endl;

    mesg("Done.\n");
}

template <class F>
void H1NChi<F>::Dual_Tp_On(Matrix<F>& T,
			   int p,
			   const Matrix<F>& W,
			   const Vector<int>& e,
			   const Matrix<F>& D) const
{
//    start_timing("Computing Dual_Tp_On, p=",p);
    Matrix<F> V; W.transpose(V);
    int n=V.nrows(), m=V.ncols();

    // E is the matrix whose columns correspond to the e[i]:
    Matrix<F> E; E.init(m,n);
    for(int i=1; i<=n; i++)
      E(e[i],i)=1;
    
    // Compute (T_p)*E quickly.
    Matrix<F> TE(m,n);
    Array<int> H; 

    // The *HARD* work is in the following two steps!
    Tp_Heilbronn(p,H);
    for(int i=1;i<=n;i++) {	
	Vector<F> Tei;
	Tp_StandardSymbol(Tei,H,e[i]);
	TE.setcol(i,Tei);
    }

    // Compute matrix of V.
    //    cout << "TE = " << TE <<endl;
    //    cout << "V = " << V << endl;
    //    cout << " D = " << D << endl;
    //    cout << " E = " << E << endl;     
     
    T = V*TE*D*V*E*D;
//    cout << "V*TE*D*V*E*D = " << T << endl;
}


template <class F>
void H1NChi<F>::WN_On (int M, const Matrix<F>& W, Matrix<F>& WNon) 
{
    start_timing ("WN_on ") ;
    Matrix<F> WN;
    WN_Mk (M, WN) ;
    WN.RestrictedTo(WNon, W) ;  
    stop_timing();
}


template <class F>
int H1NChi<F>::load_Mk () 
{
    if (!save) return 0 ;
    char s[1024] ; sprintf (s, "%sMk", fname) ;
    
    ifstream in(data_read(s,N)) ;
    if (!in) return 0 ;
    
    in >> dim_m ;
    Mk.resize (dim_m) ;
    for (int i = 1 ; i <= dim_m ; i++)
	in >> Mk[i] ;
    
    int Lift_data_length ;
    in >> Lift_data_length ;
    Lift_data.resize (Lift_data_length) ;
    for (int i = 1 ; i <= Lift_data_length ; i++)
	in >> Lift_data[i] ;
    
    Lift.Load (in) ;
    return 1 ; 
}



template <class F>
void H1NChi<F>::save_Mk () const
{
    if (!save) return ;
    char s[1024] ; sprintf (s, "%sMk", fname) ;
    
    ofstream of(data_write(s,N));
    if (!of) {
	error("Error trying to save Mk object.", NONTERMINAL) ;
	return ;
    }
    
    of << dim_m << endl ;
    for (int i = 1 ; i <= dim_m ; i++) 
	of << Mk[i] << " " ;
    of << endl ;
    of << Lift_data.size() << endl ;
    for (int i = 1 ; i <= Lift_data.size() ; i++) 
	of << Lift_data[i] << " " ;
    of << endl ;
    
    Lift.Save (of) ;
}

template <class F>
int H1NChi<F>::load_Sk() 
{
    if (!save) return 0 ;
    char s[1024] ; sprintf (s, "%sSk", fname);
    ifstream in(data_read(s,N)) ;
    if (!in) return 0 ;
    Sk.Load(in);
    return 1 ; 
}

template <class F>
void H1NChi<F>::save_Sk() const
{
    if (!save) return ;
    char s[1024] ; sprintf (s, "%sSk", fname) ;
    ofstream of(data_write(s,N));
    if (!of) {
	error("Error trying to save Sk.", NONTERMINAL) ;
	return ;
    }
    Sk.Save(of);
}

template <class F>
int H1NChi<F>::load_Sknew() 
{
    if (!save) return 0 ;
    char s[1024] ; sprintf (s, "%sSknew", fname);
    ifstream in(data_read(s,N)) ;
    if (!in) return 0 ;
    Sknew.Load(in);
    return 1 ; 
}


template <class F>
void H1NChi<F>::save_Sknew() const
{
    if (!save) return ;
    char s[1024] ; sprintf (s, "%sSknew", fname) ;
    ofstream of(data_write(s,N));
    if (!of) {
	error("Error trying to save Sknew.", NONTERMINAL) ;
	return ;
    }
    Sknew.Save(of);
}


template <class F>
int H1NChi<F>::load_Tn_Mk (int n, Matrix<F>& T) const
{
    if (!save) return 0 ;
    char s[1024]; 
    sprintf (s, "%sTn%d", fname, n) ;
    
    ifstream in(data_read(s,N)) ;
    if (!in) return 0 ;
    
    T.Load (in) ;
    return 1 ; 
}

template <class F>
void H1NChi<F>::save_Tn_Mk (int n, const Matrix<F>& T) const
{  
    if (!save) return ;
    char s[1024] ;  sprintf (s, "%sTn%d", fname, n) ;
    
    ofstream of(data_write(s,N)) ;
    if (!of) {
	error("Couldn't save Tn_Mk.", NONTERMINAL) ;
	return ;
    }
    
    T.Save(of) ;
}

template <class F>
int H1NChi<F>::load_WN_Mk (int M, Matrix<F>& W) const
{
    if (!save) return 0 ;
    char s[1024]; 
    sprintf (s, "%sW%d", fname, M) ;
    
    ifstream in(data_read(s,N)) ;
    if (!in) return 0 ;
    
    W.Load (in) ;
    return 1 ; 
}

template <class F>
void H1NChi<F>::save_WN_Mk (int M, const Matrix<F>& W) const
{  
    if (!save) return ;
    char s[1024] ;  sprintf (s, "%sW%d", fname, M) ;
    
    ofstream of(data_write(s,N)) ;
    if (!of) {
	error("Couldn't save WN_Mk.", NONTERMINAL) ;
	return ;
    }
    
    W.Save(of) ;
}

template<class F>
void
H1NChi<F>::MSymRelations(Matrix<F>& M)
{
     Matrix<F> A(dim_m, msymbols.length());
        for(int i=1; i<=msymbols.length(); i++)
	       A.setcol(i,msymidx_to_free(SI_rel(i)));
	   Kernel(M, A);
}

template <class F>
const Matrix<F>& 
H1NChi<F>::get_Sk() 
{
    if (!init_sk) { init_Sk();  }
    return Sk ;
}

template <class F>
const Matrix<F>& 
H1NChi<F>::get_Eis() 
{
    if (!init_eis) { init_Eis();  }
    return Eis ;
}

template <class F>
const Matrix<F>& 
H1NChi<F>::get_Sknew() 
{
    if (!init_sknew) {   Matrix<F> Dummy; get_Sknew(Dummy, 1); }
    return Sknew ;
}


/* this function is too long */
template <class F>
void 
H1NChi<F>::get_Mknew (Matrix<F>& Mn, int at) const
{
    start_timing("get_Mknew");
    int d = dim_Mk();
    
    if(N==1) {
	Mn.idmat(d);
	return;
    }
    // compute the intersection of the kernels of the trace maps.
    
    // algorithm: for each pair (M,t) with M|N and t|(N/M), compute
    // a matrix repn for the trace map
    //         eta_t : Mk(N)-->Mk(M).
    // then augment that matrix at the bottom of A. 
    // The kernel of A is the intersection of the kernel
    // of the etas. 
    Matrix<F> A; A.init(1,d);
    int cr=2;			// current row. 
    
    rational_factorization fac(N);
    fac.factor();
    for(int i=0;i<fac.no_of_comp();i++) 
	if(chi.degree(to_int(fac.base(i)))==1) {
	    int M=N/to_int(fac.base(i));	// M = N/p
	    mesg("M=",M); mesg("    ");
	    
	    // construct the space of forms at level M.
	    Character chiM(M,chi);  
	    H1NChi<F>* H = new H1NChi<F>(chiM,k,save,plus);
	    if(H->dim_Mk()==0) continue; // skip this space.
	    
	    // for each divisor t | N/M
	    sort_vector<bigint> tlist=divisors(N/M);
	    
	    // add space in A matrix for the eta maps. 
	    int nrows = A.nrows() + H->dim_Mk()*tlist.get_size();
//	    cout << "Allocating matrix of size "<<nrows<<"x" << d<<endl;
	    A.resize(nrows, d);
	    
	    for (int j=0; j<tlist.get_size(); j++) { 
		int t; tlist[j].intify(t);	// t = ith divisor of N/M
		mesg(j*100/tlist.get_size()); mesg("% ");
		
		Array<int> Heil;
//		cout << "About to construct level " << t << " Heilbronn matrices."<<endl;
		HeilbronnC(t,Heil);
//		cout <<"Success."<<endl;
		mesg("  [Eta t=",t); mesg("   ");
		for (int s=0; s<d; s++) {
		    if (s%20==0) {mesg(s*100/d); mesg("% ");}
		    // construct m.
		    Vector<F> mtemp(d); mtemp.zero(); mtemp[s+1]=1;
		    MSymVec<F> m; 
		    VecToMSymVec(m,mtemp);
		    // compute image of m in H
		    Vector<F> v;
		    H->Eta_Lower (v, m, t, Heil);
#ifdef DEBUG
		    if(v.size()>d) error("get_Mknew: ERROR! smaller level, bigger dim?");
#endif
		    // add column to matrix
		    for(int z=cr; z<H->dim_Mk()+cr; z++)
			A(z,s+1)=v[z-cr+1];
		}
		cr += H->dim_Mk();
		mesg("100%]\n");
	    }
	    delete H ;
	}
    mesg("\n");
    Kernel(Mn, A);
}

template<class F>
void 
H1NChi<F>::get_Sknew (Matrix<F>& Sn, int at) 
{
    if(at==1) {
	if(!init_sknew) {
	    if (!load_Sknew()) {
		Matrix<F> Mn;
		get_Mknew(Mn, 1);
		get_Sk_in_V(Sknew, Mn);
	    }
	    init_sknew=true;
	    save_Sknew();
	}
	Sn=Sknew;
    } else {
	Matrix<F> Mn;
	get_Mknew(Mn, at);
	get_Sk_in_V (Sn, Mn);
    }
}

template <class F>
void
H1NChi<F>::get_SknewPlus (Matrix<F>& SnPlus, int at)
{
    Matrix<F> V;
    get_Sknew(V,at);
    ConjugationSubspace(SnPlus, V);
}


template <class F>
void 
H1NChi<F>::get_Mknewdual (Matrix<F>& Mndual, int at) const
{
    start_timing("get_Mknewdual");
    
    if(N==1 || chi.type()==-1) {
	Mndual.idmat(dim_Mk());
	return;
    }
    
    // algorithm: for each pair (M,t) with M|N and t|(N/M), compute
    // a matrix repn for the raising map
    //         Mk(M)-->Mk(N).
    // then augment A on the right.  
    Matrix<F> A; A.init(dim_Mk(),1);
    int cc=2;			// current column
    
    rational_factorization fac(N);
    fac.factor();
    for(int i=0;i<fac.no_of_comp();i++) 
	if(chi.degree(to_int(fac.base(i)))==1) {
	    int M=N/to_int(fac.base(i));	// M = N/p
	    mesg("M=",M);
	    if(M*M<N) { mesg("Skipping  "); continue; }
	    // construct the space of forms at level M.
	    Character chiM(M,chi);  
	    H1NChi<F>* H = new H1NChi<F>(chiM,k,save);
	    if(H->dim_Mk()==0) continue; // skip this space.
	    
	    // for each divisor t | N/M
	    sort_vector<bigint> tlist=divisors(N/M);
	    
	    // add space in A matrix for the eta maps. 
	    int ncols = A.ncols() + H->dim_Mk()*tlist.get_size();
	    A.resize(A.nrows(), ncols);
	    
	    for (int j=0; j<tlist.get_size(); j++) { 
		int t; tlist[j].intify(t);	// t = ith divisor of N/M
		mesg(j*100/tlist.get_size(),"% ");
		
		mesg("[Eta ");
		Array<int> Heil;
		HeilbronnC(t,Heil);
		for (int s=0; s<H->dim_Mk(); s++) {
		    if (s%20==0) mesg(s*100/dim_m,"% ");
		    // construct m.
		    MSymVec<F> m; 
		    Vector<F> col; col.init(H->dim_Mk()); col[s+1]=F(1);
		    H->VecToMSymVec(m,col);
		    
		    // compute image of m at level N.
		    Vector<F> v;
		    Eta_Raise (v, m, t, Heil);
#ifdef DEBUG
		    if(v.size()!=dim_m) 
			error("get_Mknewdual: Eta_Raise -- gives wrong dimension");
#endif
		    A.setcol(cc+s,v);
		}
		cc += H->dim_Mk();
		mesg("100%]");
	    }
	    delete H ;
	}
    mesg("\n");
    A.transpose();
    Kernel(Mndual, A);
    stop_timing();
}

/* 
NOTE: The following turned out to be USELESS because
      the kernel of the dual of a splitting of delta
      is ** NOT ** well-defined. 

   Now Mkndual contains the new subspace of the dual of Mk.  
   The cuspidal subspace of M_k^{\dual} is the kernel of           <-- LIE.
   the dual(=transpose) of essentially a splitting of the map
   
                       delta
              M_k(N) --------> C(N)
   
   where C(N) is the vector space on the cusps.  Let C' in C(N) 
   be the image of delta.  Let D be a matrix which represents
   delta: M_k(N) ------> C'.  To find a splitting s:C'-->M_k 
   we find a matrix S so that DS=I. 

template <class F>
void 
H1NChi<F>::get_Sknewdual (Matrix<F>& Skndual, int at)
{
    Matrix<F> Mkndual;
    get_Mknewdual(Mkndual, at);

    // Compute reduced row-echolen form of Delta and delete all zero rows, copying into D. 
    init_Delta();  // assure that Delta is defined 
    Matrix<F> D1=Delta;
    Vector<int> elimcol, elimrow;
    D1.SimpleEchelon(elimcol,elimrow);
    Matrix<F> D(D1.nrows(),D1.ncols());
    int r=0;
    for(int i=1;i<=D1.nrows();i++)
	if(elimrow[i]!=0)
	    D.setrow(++r,D1.row(i));
    D.resize(r,D1.ncols());
    Matrix<F> S;

    cout << "Delta = " << D << endl;

    D.RightInverse(S);
    S.transpose();
    cout << "S = " << S <<endl;
    // Restrict S to Mkndual and compute the kernel.
    Matrix<F> Sres;
    cout << "Mkndual = " << Mkndual << endl;
    Sres = S * Mkndual;
    cout << "Sres = " << Sres << endl;
    Matrix<F> K;
    Kernel(K, Sres);
    Skndual = Mkndual * K;
    cout << "Skndual = " << Skndual << endl ;
}
*/

template<class F>
void 
H1NChi<F>::get_Sk_in_V (Matrix<F>& W, const Matrix<F>& V) 
{
    init_Delta();
    Matrix<F> DV=Delta*V;
    Matrix<F> T;
    Kernel(T,DV);
    W=V*T;
}


template<class F>
void 
H1NChi<F>::modular_to_free_wt2_helper(Vector<F>& v,
				   const bigint& a, const bigint& b) const
{
//    cout << "{0,"<<a<<"/"<<b<<"} = ";
    MSymVec<F> m; m.init(1,N);
    Vector<bigint> cf; cf.init(0);
    if(b!=0)
	contfrac(cf, a, b);
    v.init(dim_Mk());
    bigint c=0, d=1;
    bigrational part=0;

    int j=-1;
    do {
	m.set(1, MSym(c,d,0), F(1));
	Vector<F> w;
	MSymVecToVec (w, m);
	v += w;
	j++;
	// compute next pair c,d
	if(j+1<=cf.size()) {
	    d=abs(c);
	    c=denominator(convergent(cf,j+1));
	    if((j+1)%2 != 0)
		c*=-1;
	}
    } while (j+1<=cf.size());
//    cout <<v<<endl;
}

template<class F>
void 
H1NChi<F>::modular_to_free_wt2(Vector<F>& v,
			    const bigint& a, const bigint& b,
			    const bigint& c, const bigint& d) const
{
    // {a/b,c/d}={a/b,0}+{0,c/d}=-{0,a/b}+{0,c/d}
    modular_to_free_wt2_helper(v,a,b);
    v*=F(-1);
    Vector<F> w;
    modular_to_free_wt2_helper(w,c,d);
    v+=w;
}


template<class F>
void 
H1NChi<F>::modular_to_free_helper(Vector<F>& v,
				  const bigint& a, const bigint& b,
				  const int& t) const
{
  if(t<0||t>k-2)
    die("modular_to_free_helper: q is out of bounds, crazy madness!!!");
  if(b==0) {
    if(a==0)
      die("modular_to_free_helper: a/b=0/0 not defined!!!!");
    MSym x=MSym(0,1, t);
    MSymToVec(v,x);
    return;
  }
    

  Vector<bigint> cf; 
  contfrac(cf, a, b); 

  v.init(dim_Mk());
  
  bigint p, pp, q, qq;
  p = 0; q = 1; pp =1; qq=0;
  for(int j=1; j<=cf.size()+1; j++) {

//    cout << "Adding the term p="<<p<<",q="<<q<<",pp="<<pp<<",qq="<<qq<< endl;

    bigint det=q*pp-p*qq ;

    if(abs(det)!=1) 
      die("modular_to_free_helper: error in cont. frac. convergent.");

    int mat[4]; 
    mat[0]=to_int(pp); mat[1]=to_int(p);
    mat[2]=to_int(qq); mat[3]=to_int(q);
    if(det<1) {
      mat[1]*=-1;
      mat[3]*=-1;
    }

    Array<F> coeff;
    coeff.reset(k-2+1);
    act_on_MSym_q(coeff,mat,t);

    MSymVec<F> msv; msv.init(k-2+1,N);
    for(int i=0; i<=k-2; i++) 
      msv.set(i+1, MSym(mat[2],mat[3],i), coeff[i]);

    Vector<F> w;
    MSymVecToVec (w, msv);

//    cout << "Now adding " << endl
//	 << msv << "=" <<endl;
//    msv.ModSymPrint(cout,k);
//    cout << endl;
//    cout << " = " << w << endl << endl;
    
    v += w;

    if(j<=cf.size()) {
      bigrational conv = convergent(cf,j);
      p=pp;
      q=qq;
      pp = numerator(conv);
      qq = denominator(conv);
    }
  }
}


template<class F>
void 
H1NChi<F>::modular_to_free(Vector<F>& v,
			   const bigint& a, const bigint& b,
			   const bigint& c, const bigint& d,
			   const int& q) const
{
    // P = X^q*Y^(k-2-q).
    // P*{a/b,c/d}=P*{a/b,0}+P*{0,c/d}=-P*{0,a/b} + P*{0,c/d}
    modular_to_free_helper(v,a,b,q);
    v*=F(-1);
    Vector<F> w;
//    cout<<endl<<endl<<endl;
    modular_to_free_helper(w,c,d,q);
    v+=w;
}

template<class F>
void H1NChi<F>::modular_to_free(Vector<F>& v,
				const bigint& a, const bigint& b,
				const bigint& c, const bigint& d,
				const Array<F>& qar) const
{
    v.init(dim_Mk());
    for(int i=0; i<=k-2; i++) 
	if(qar[i]!=F(0)) {
	    Vector<F> w;
	    modular_to_free(w,a,b,c,d,i);
	    v+= qar[i]*w;
	}
}

template<class F>
void H1NChi<F>::Invariant_Symbol(Vector<F>& v,
				 Vector<bigint>& P,
				 const bigint& a, const bigint& b,
				 const bigint& c, const bigint& d) const 
{
    // Compute:     ((1-ad)X^2-b(d-a)XY+b^2Y^2)^((k-2)/2) * {0, c/d}.
    if(chi.type()!=0) 
	mesgln("WARNING: H1NChi<F>::Invariant_Symbol -- chi should be trivial.");
    if(k%2 != 0)
	die("H1NChi<F>::Invariant_Symbol -- weight must be even.");
    if(c%N !=0)
	die("H1NChi<F>::Invariant_Symbol -- g must be in Gamma_0(N).");
    if(k==2) {
	modular_to_free(v,0,1,b,d,0);
	return;
    }
    Vector<bigint> f;
    f.init(3);
    f[1] = b*b;
    f[2] = -b*(d-a);
    f[3] = 1-a*d;
    f.polpower(P, (k-2)/2);
    Array<F> qar(k-1);
    for(int i=0; i<k-1; i++)
	qar[i] = F(P[i+1]);
    modular_to_free (v, 0, 1, b, d, qar);
}

// This isn't too serious, i.e., it doesn't
// really iterate through all of Gamma_0(N).  It
// just sort of bounces around.
void NextElementGammaZeroN(Matrix<bigint>& B,
			   const Matrix<bigint>& A,
			   int N)
{
    B.init(2,2);
    B(2,1)=A(2,1); B(2,2)=A(2,2);
    if(B(2,1) % N !=0) die("NextElementGammaZeroN -- A is not in Gamma_0(N).");
    if(B(2,2) > B(2,1)*B(2,1)) {
	B(2,1)+=N;
	B(2,2)=0;
    }
    // Go to next one, then...
    B(2,2)++;
    // increase B(2,2) until coprime to B(2,1).
    while(gcd(B(2,2),B(2,1))!=1) B(2,2)++;
    xgcd(B(1,2),B(1,1),-B(2,1),B(2,2));
}

template<class F>
void
H1NChi<F>::Invariant_Symbol_Matrices(MatVecList<bigint>& List,
				     Matrix<F>& PhiGood,
				     const Matrix<F>& Phi) const
{
    mesgln("Invariant_Symbol_Matrices, Phi = ");
    cerr << Phi<<endl;
    if(chi.type()!=0) 
	mesgln("WARNING H1NChi<F>::Invariant_Symbol_Matrices -- chi should be trivial.");
    if(k%2 != 0)
	die("H1NChi<F>::Invariant_Symbol_Matrices -- weight must be even.");
    
    int dim=Phi.nrows();
    /* Algorithm: List elements (a,b;c,d) in Gamma_0(N) for which c=Nc', and 
       c'>0 is small.  For each there is a corresponding vector in M_k(Q), 
       as computed by Invariant_Symbol.  Compute the images under Phi of
       these until the rank of the image = dim.  
    */
    MatVecList<bigint> tList;
    tList.init(2*dim);
    Matrix<F> Im; Im.init(Phi.nrows(),2*dim);

    Vector<int> elimcol, elimrow;
    int i=0;
    do {
	// add a few columns.
	if(i>dim*dim) {
	    mesgln("H1NChi<F>::Invariant_Symbol_Matrices: WARNING!!");
	    mesgln("It is taking way to long to find enough symbols.");
	    mesgln("N= ",N);
	    mesgln("k= ",k);
	}
	// add dim more rows.
	for(int j=1; j<=dim+1; j++) {
	    if(Im.ncols() < i+1) {
		Im.resize(Im.nrows(), Im.ncols()+dim);
		tList.resize(Im.ncols()+dim);
	    }
	    if(i>0) 
		NextElementGammaZeroN(tList.M(i), tList.M(i-1), N);
	    else {
		Matrix<bigint> A(2,2);
		A(1,1)=1; A(1,2)=0;
		A(2,1)=N; A(2,2)=1;
		NextElementGammaZeroN(tList.M(i), A, N);
	    }
	    
	    Vector<F> v;
	    Invariant_Symbol(v, tList.v(i),
			    tList.M(i)(1,1), tList.M(i)(1,2),
			    tList.M(i)(2,1), tList.M(i)(2,2));
	    Im.setcol(i+1,Phi*v);
	    i++;
	    if(i%5==0) mesg(i/5,"  ");
	}
    } while(Im.rank(elimcol,elimrow)<dim);
    
    // pick off pivot columns.
    List.init(dim);
    PhiGood.init(Im.nrows(),dim);
    i=0;
    mesgln("The invariant symbol data is:");
    for(int c=1;c<=Im.ncols();c++)
	if(elimcol[c]!=0) {
	    if(i>=dim)
		die("H1NChi<F>::Invariant_Symbol_Matrices -- a contradiction!");
	    List.M(i)=tList.M(c-1);
	    List.v(i)=tList.v(c-1);
	    Vector<F> v; 
	    Vector<bigint> dummy;
	    Invariant_Symbol(v, dummy,
			    List.M(i)(1,1), List.M(i)(1,2),
			    List.M(i)(2,1), List.M(i)(2,2));
//	    cout << "g = " << List.M(i) << endl;
//	    cout << "v = " << v << endl;
//	    cout << "Phi*v = " << Phi*v << endl;

	    PhiGood.setcol(i+1,Im.col(c));
	    i++;
	}
}


// THESE TIMERS MUST BE REWRITTEN IN A LESS SILLY WAY.
template <class F>
void H1NChi<F>::start_timing(char* msg, int param=-1) const
{
    reset_timer () ;
    mesg(msg);
    if (param != -1) mesg(param) ;
    mesg(" ..");
}

template <class F>
void H1NChi<F>::stop_timing() const
{
    mesg(get_time()); mesg("s\n");
}

template <class F>
void H1NChi<F>::status_message(char* msg, int param=-1) const
{
    
    reset_timer () ;
    mesg("\n"); mesg(msg);
    if (param != -1) mesg(param);
    mesg("\n");
}

/*
				// HORRIBLY INEFFICIENT!
template <class F>
void H1NChi<F>::HeilbronnC_old (int n, Array<int>& H) const
{
    int i=0;
    for (int a=1; a<=n; a++)
	for (int d=1; d<=n; d++)
	    for (int c=0; c<d; c++)
		for (int b=0; b<a; b++)
		    if (a*d-b*c==n) {
			if (H.length () <= 4*i+3) 
			    H.resize ((H.length() + 4) + H.length()/2) ;
			H[4*i]   = a; H[4*i+1] = b ;
			H[4*i+2] = c; H[4*i+3] = d ;
			i++ ;
		    }
    H.resize (4*i) ;
}
*/

template <class F>
void H1NChi<F>::HeilbronnC (int n, Array<int>& H) const
{
    H.reset(24);
    int i=0;
    for (int a=1; a<=n; a++)
	for (int d=1; d<=n; d++) {
	    // ad-bc=n so c=0 and ad=n, or b=(ad-n)/c
	    int bc=a*d-n;
	    if(bc<0) continue;
	    if(bc==0) {
		for(int b=0;b<a;b++) {
		    if (H.length () <= 4*i+3) 
			H.resize ((H.length() + 4) + H.length()/2) ;
		    H[4*i]   = a; H[4*i+1] = b;
		    H[4*i+2] = 0; H[4*i+3] = d ;
		    i++ ;
		}
	    } 
	    for (int c=1; c<d; c++)// should factor!
		if((bc % c==0) && (bc/c < a)) {
		    if (H.length () <= 4*i+3) 
			H.resize ((H.length() + 4) + H.length()/2) ;
		    H[4*i]   = a; H[4*i+1] = bc/c ;
		    H[4*i+2] = c; H[4*i+3] = d ;
		    i++ ;
		}
	}
    H.resize (4*i) ;
}


template <class F>
void H1NChi<F>::Heilbronn (int p, Array<int>& H) const
{
    if(!is_prime(p)) {
	HeilbronnC(p,H);
	return;
    }
    if (p==2) {
	H.reset(16) ;
	H[0] = 1; H[1] = 0; H[2] = 0; H[3] = 2;
	H[4] = 2; H[5] = 0; H[6] = 0; H[7] = 1;
	H[8] = 2; H[9] = 1; H[10] = 0; H[11] = 1;
	H[12] = 1; H[13] = 0; H[14] = 1; H[15] = 2;
	return ;
    }
    
    H.reset(4) ;
    H[0] = 1; H[1] = 0; 
    H[2] = 0; H[3] = p ;
    
    if (p==1) return ;
    
    int i = 1, r = round(-p,2) ;
    for (; r <= p/2; r++) {
	int x1 = p, x2 = -r, y1 = 0, y2 = 1, a = -p, b = r, c,
	    x3, y3, q ;
	if (H.length () <= 4*i+3) 
	    H.resize ((H.length() + 4) + H.length()/2) ;
	H[4*i]   = x1; H[4*i+1] = x2 ;
	H[4*i+2] = y1; H[4*i+3] = y2 ;
	i++ ;
	while ( b ) {
	    q = nearest_integer (a, b) ;
	    c = a - b * q ;
	    a = -b ;
	    b = c ;
	    x3 = q * x2 - x1 ; 
	    x1 = x2 ;
	    x2 = x3 ;
	    y3 = q * y2 - y1 ;
	    y1 = y2 ;
	    y2 = y3 ;
	    if (H.length () <= 4*i+3) 
		H.resize ((H.length() + 4) + H.length()/2) ;
	    H[4*i]   = x1; H[4*i+1] = x2 ;
	    H[4*i+2] = y1; H[4*i+3] = y2 ;
	    i++ ;
	}
    }
    H.resize (4*i) ;
}

template <class F>
void H1NChi<F>::init_Heilbronn() 
{
    if(H1NChi<F>::Heil.H) return;
    if(!save) { H1NChi<F>::Heil.H=NULL; return; } // BUG ALERT: a possible memory leak.
    ifstream in(data_read("Heilbronn",0));
    int a;
    if(in) {
        mesgln("Loading Heilbronn matrix list.");
        in>>a;
        if(a==HEILBRONN_CACHE_MAXP) {
            H1NChi<F>::Heil.H = new Array<int>[HEILBRONN_CACHE_MAXP+1];
            int len;
            for(int p=2; p<=HEILBRONN_CACHE_MAXP; p=next_prime(p)) {
                in >> len ;
                H1NChi<F>::Heil.H[p].reset(len);
                for(int i=0;i<len; i++)
                    in>>H1NChi<F>::Heil.H[p][i];
            }
        }
    }
    if(!in || a!=HEILBRONN_CACHE_MAXP) {
	mesgln("Precomputing Heilbronn matrices.\n");
        H1NChi<F>::Heil.H = new Array<int>[HEILBRONN_CACHE_MAXP+1];
        for(int p=2; p<=HEILBRONN_CACHE_MAXP; p=next_prime(p)) {
	    mesg(p);
            Heilbronn(p,H1NChi<F>::Heil.H[p]);
	    mesg("(",H1NChi<F>::Heil.H[p].length()/4);
	    mesg(")  ");
	}
        ofstream out(data_write("Heilbronn",0));
        out << HEILBRONN_CACHE_MAXP << endl ;
        for(int p=2; p<=HEILBRONN_CACHE_MAXP; p=next_prime(p)) {
            out << H1NChi<F>::Heil.H[p].length() <<endl;
            for(int i=0;i<H1NChi<F>::Heil.H[p].length();i++)
                out << H1NChi<F>::Heil.H[p][i]<< ' ';
            out << endl;
        }
    }
    mesgln("");
}


template <class F>
void H1NChi<F>::Tp_Heilbronn (int n, Array<int>& H) const
{
    if(save && n<=HEILBRONN_CACHE_MAXP && H1NChi<F>::Heil.H && is_prime(n))
	H=H1NChi<F>::Heil.H[n];
    else 
	Heilbronn(n,H);
}

template<class F>
void H1NChi<F>::DegeneracyReps (Array<bigint>& R,
				const bigint& N, const bigint& M, const bigint& t) const
{
    /* PROBLEM: Find coset reps for Gamma_0(N) \ T Gamma_0(M) T^{-1}.
       SOLUTION:  To avoid using non-integers compute this algorithm
       computes reps for T^{-1} Gamma_0(N) T = Gamma_0(N/t,t)
       inside of Gamma_0(M).  The calling routine may conjugate by T
       at the end.  Let L=lcm(N/t,t).
       COSET EQUIVALENCE: Two right cosets represented by (a,b;c,d) & 
       (a',b';c',d') of Gamma_0(N/t,t) in SL_2(Z) are equivalent iff
       (a,b)=(a',b') as points of P^1(t), i.e., ab'=ba'(mod t),
       and (c,d)=(c',d') as points of P^1(N/t).  
    */

    // The index of Gamma_0(A,B) in SL_2(Z) is #P^1(A*B) = index Gamma_0(A*B).
    ModularFormulas f;
    R.reset(4*(f.idxG0(N)/f.idxG0(M)));	
//    mesg(R.length()/4," cosets.\n");
    // Thus this gives an upper bound for the index in Gamma_0(M).

    if(N%t != 0) die("DegeneracyReps, t doesn't divide N.");
    if(N%M != 0) die("DegeneracyReps, M doesn't divide N.");
    bigint Not=N/t;
    if((N/M) %t != 0) die("DegeneracyReps, t doesn't divide N/M.");

    int n=0, c=0;
    int max=R.length()+4;
    while(n<R.length()) {
	// generate random element of SL_2(Z).
	bigint cc=M*((::random() % max) - max/2), dd=(::random()%max)-max/2, aa,bb;
	bigint d=xgcd(bb,aa,-cc,dd);
	if(d==0) continue;
	cc/=d; dd/=d;
	if((cc%M)!=0) continue;
	// Test if it is new. 
	bool isnew=true;
	for(int j=0; j<n; j+=4) 
	    if((R[j+1]*aa-R[j]*bb)%t==0 && (R[j+3]*cc-R[j+2]*dd)%Not==0)
		{isnew=false; break;}
	
	// If so add it to the list.
	if(isnew) {
	    R[n]=aa; R[n+1]=bb;   
	    R[n+2]=cc; R[n+3]=dd;
	    n+=4;
	} else {
	    if((++c)%500==0) max*=2; // double range.
	}
    }
}


template<class F>
Vector<F> H1NChi<F>::Eta_Raise_Single (const MSym& x, 
				       const Array<bigint>& Gam,
				       const Array<int>& Heil,
				       const int& t, const int& M) const 
{
    Vector<F> ans;
    ans.init(dim_m);
    bigint_matrix g(3,3);   // 3x3 because LiDIA indexes matrices starting at 0.
    // ?? AND ALSO TO MAKE the code WAY WAY SLOWER?????????? -- was
    x.abcd(g(1,1),g(1,2),g(2,1),g(2,2),M);   // definite *not* N here.
    
    // sum_{u in Gam}  sum_{H in Heil}  
    //        Action of H on q (= weight of x)
    //        and bottom two entries of u*g*H for P^1(N) part (if they give
    //        an element of P^1(N), i.e., gcd(c,d,N)=1).
    //     Note Heil is the set of Heilbronn matrices
    //     of det t.   It is assumed that t|N.
    bigint_matrix U(3,3), H(3,3), tTinv(3,3);
    for (int i=0; i<Gam.length()/4; i++) {
	U(1,1)=Gam[4*i]; U(1,2)=Gam[4*i+1]; U(2,1)=Gam[4*i+2]; U(2,2)=Gam[4*i+3];
	for (int j=0; j<Heil.length()/4; j++) {
	    H(1,1)=Heil[4*j]; H(1,2)=Heil[4*j+1]; H(2,1)=Heil[4*j+2]; H(2,2)=Heil[4*j+3];
	    int u, v;
	    
	    // compute U * g * H;
	    bigint_matrix UgH = U * g * H ;
	    if ((UgH(1,1)%t)==0 && (UgH(1,2)%t)==0) {
		UgH(2,1).intify(u); UgH(2,2).intify(v);
		
		// for higher weight
		Array<F> coeff(k-2+1); 
		int m[4]; H(1,1).intify(m[0]); H(1,2).intify(m[1]);
		H(2,1).intify(m[2]); H(2,2).intify(m[3]);
		act_on_MSym_q (coeff, m, x.q());
		int index, lambda; 
		msymbols.MSym_to_list(index, lambda, MSym(u,v,0));
		F zeta = F(chi.of(lambda));
		for(int l=0; l<k-2+1;l++) 
		    ans += zeta * (coeff[l] * index_to_free(index+l));
	    }
	}
    }
    return ans ;
}

// Proposition 16 in Merel, SLNM 1585, 1994.
// IT IS IMPORTANT THAT Heil be HeilbronnC and 
// *not* Heilbronn from Cremona's book!
// The proof of prop 16 only works under the assumption
// that the matrices satisfy Loic's criterion. I don't
// think Cremona's do.
template<class F>
void H1NChi<F>::Eta_Raise (Vector<F>& v,	
			   const MSymVec<F>& m,
			   const int& t,
			   const Array<int>& Heil) const 
{
    Array<bigint> Gam;
    DegeneracyReps(Gam,N,m.N(),t);
    v.init(dim_m);
    for (int i=1; i <= m.size(); i++) {
	v += m.c(i)*Eta_Raise_Single(m.x(i), Gam, Heil, t, m.N());
    }
}

template<class F>
Vector<F> H1NChi<F>::Eta_Lower_Single (const MSym& x, 
				       const Array<int>& Heil,
				       const int& t, const int& M) const 
{
    Vector<F> ans;
    ans.init(dim_m);
    
    for (int j=0; j<Heil.length()/4; j++) {
	Array<int> idx; 
	Array<F> c; 
	act_on_MSym(idx,c,0,&Heil[4*j],ETA_LOWER, t, x.u(), x.v(), x.q()) ;
	for(int i=0;i<idx.length();i++) 
	    if(c[i]!=F(0))
		ans += c[i]*index_to_free(idx[i]) ;
    }
    return ans ;
}


// Proposition 15 in Merel, SLNM 1585, 1994.
template<class F>
void H1NChi<F>::Eta_Lower (Vector<F>& v,	
			   const MSymVec<F>& m,
			   const int& t,
			   const Array<int>& Heil) const 
{

    v.init(dim_m);
    for (int i=1; i <= m.size(); i++) 
	if(m.c(i) != F(0))
	    v += m.c(i)*Eta_Lower_Single(m.x(i), Heil, t, m.N());
}


template<class F>
void H1NChi<F>::MSymVecToVec (Vector<F>& v,
			      const MSymVec<F>& m) const 
{
    v.init(dim_m);
    for(int i=1; i<=m.size(); i++) {
	int lambda, index; 
	msymbols.MSym_to_list(index, lambda, m.x(i));
	F zeta(chi.of(lambda));
	v += m.c(i) * (zeta*index_to_free(index));
    }
}

template<class F>
void H1NChi<F>::MSymToVec (Vector<F>& v,
			   const MSym& m) const 
{
    MSymVec<F> msym; msym.init(1,N); msym.set(1,m,1);
    MSymVecToVec(v,msym);
}

template<class F>
void H1NChi<F>::VecToMSymVec (MSymVec<F>& m,
			      const Vector<F>& v) const 
{
    m.init(v.numnonzero(),N);
    int i=1;
    for(int j=1; j<=v.size(); j++)
	if (v[j]!=F(0)) 
	    m.set(i++,msymbols.x(SI_gen(Mk[j])),v[j]);
}



template <class F> 
void 
H1NChi<F>::Dual(Vector<F>& fdual, const Vector<F>& f) const
{
#ifdef DEBUG
    if(f.size()!=dim_m) error("Dual: f is of the wrong size.");
#endif
    
    Matrix<F> V;
    V.idmat(dim_m);
    int nz=0;
    for(int i=1;i<=dim_m;i++)
	if(f[i]!=F(0)) { nz=i; break; }
    if(nz==0) error("Dual: f is the zero vector.");
    
    int p=2;
    while (V.ncols() > 1) {
	Matrix<F> T; 
	Tp_Mk(p,T);
	Vector<F> Tf=T*f;
	F ap=Tf[nz]/f[nz];
#ifdef DEBUG
	for(int i=1;i<=dim_m;i++)
	    if (Tf[i]!=ap*f[i]) {
		cerr << "Dual: f is not an eigenvector for T_"<<p<<endl;
		error("exiting.");
	    }
#endif
	T.transpose();		// T := transpose of T.
	Matrix<F> TonV;
	T.RestrictedTo(TonV, V);
	Matrix<F> I;
	I.idmat(V.ncols());
	TonV -= ap*I;
	Matrix<F> K;
	Kernel(K,TonV);
	V = V*K;
	// check
#ifdef DEBUG   // this is really unnecessary
	Matrix<F> Vcheck;
	Vcheck = T*V;
	for(int i=1;i<=V.nrows();i++)
	    for(int j=1;j<=V.ncols();j++)
		if(Vcheck(i,j)!=ap*V(i,j)) error("Dual: programming mistake!");
#endif
	p=next_prime(p);
    }
    if (V.ncols() != 1)
	error("Dual: couldn't find dual.");
    V.col(1,fdual);
}

template<class F>
bool H1NChi<F>::ComputeHeckeDual(Matrix<F>& VDual, const Matrix<F>& V,
				 int maxn) 
{
    mesg("[ComputeHeckeDual");
    mesg("-->short circuiting to 'useimage'-->\n");
    if(ComputeHeckeDual_useimage(VDual, V, 2, maxn)) {
	if(VDual.ncols()!=V.ncols())
	    die("ComputeHeckeDual: VDual.ncols()!=V.ncols()");
	return true;
    }

    mesg("\nComputeHeckeDual_useimage failed. Using kernels instead.\n");
    if(!plus) {
	mesg("Kernel method not programmed for non-plus spaces.");
	return false;
    }
    get_Mknewdual(VDual);
    for(int p=2; VDual.ncols() > V.ncols(); p=next_prime(p)) {
	mesg("[Compute dual: p=",p); mesg(", dim=",VDual.ncols());
	mesg(", dim V=",V.ncols());mesg("]");
	if(p>maxn) return false;
	Matrix<F> T; Tp_Mk(p,T); 
	Matrix<F> A; T.RestrictedTo(A, V);
	Vector<F> f; A.charpoly(f);
	T.transpose();
	Matrix<F> B;
	T.RestrictedTo(B,VDual);
	Matrix<F> Vd;
	Kernel(Vd,B,f);
	VDual = VDual*Vd;
    }
    return true;
}

template<class F>
bool H1NChi<F>::ComputeHeckeDual_useimage(Matrix<F>& VDual, const Matrix<F>& V,
					  int startn, int stopn) 
{
    mesgln("[ComputeHeckeDual_useimage]");
    /* IMAGE algorithm.

       Proposition 1. Let W = space which contains the sought after VDual,
       let f=charpoly(T|VdDual), let h=charpoly(T|W), and
       write h = f*g for some g.  Then Im(g(T)) is contained in
       the kernel of f(T).  
       Proof. h(T)=0 so f(T)*g(T)(W)=0 so g(T)(W) is contained in ker(f(T)). 

       Proposition 2. Suppose h=f*g, as above, and gcd(f,g)=1. 
       Then Im(g(T))=ker(f(T)).  
       Proof. Find polynomials s and t so that s*f + g*t = 1.  
       Let v in ker(f(T)) subset W.  Then
               v = s(T)*f(T)*v + g(T)*t(T)*v 
                 = g(T)*t(T)*v    is in Im(g(T)).
       
       In generaly VDual is contained in ker(f(T)), but unfortunately 
       ker(f(T)) might be larger, and there is no simple relationship
       between Im(g(T)) and VDual. 

       Our algorithm will be to simply compute f and g as above until gcd(f,g)=1,
       and then compute Im(g(T)).  This should work well when W=Sknew since the
       eigenvalues of newforms should probably be somewhat different than eigenvalues
       of old forms and eigenvalues of eisenstein series. 
    */
    
    Matrix<F> W;
    get_Mknewdual(W);

    if(V.ncols()>W.ncols()) {
	mesg("ComputeHeckeDual: It is not possible that V is contained\n");
	mesg("in the space of newforms.");
	return false;
    }

    if(!plus) {			// if easy, first cut down using *
	Matrix<F> C; Conjugation_Mk(C);
	Matrix<F> A; C.RestrictedTo(A, V);
	Matrix<F> I; I.idmat(A.nrows());

	if(A==I) {
	    C.transpose();
	    Matrix<F> B;
	    C.RestrictedTo(B,W);
	    for(int i=1;i<=B.nrows();i++)
		B(i,i)-=F(1);
	    Matrix<F> K;
	    Kernel(K, B);
	    W=W*K;
	} if(A==F(-1)*I) {
	    C.transpose();
	    Matrix<F> B;
	    C.RestrictedTo(B,W);
	    for(int i=1;i<=B.nrows();i++)
		B(i,i)+=F(1);
	    Matrix<F> K;
	    Kernel(K, B);
	    W=W*K;
	}
    }
    
    for(int n=startn; n<=stopn; n++) {
	if(chi.type()==-1 && !issquarefree(n)) continue;
	mesg("[Trying T_",n);
	Matrix<F> T; Tn_Mk(n,T);
	Matrix<F> A; T.RestrictedTo(A, V);
	Vector<F> f; A.charpoly(f);
	T.transpose();
	Matrix<F> B; T.RestrictedTo(B, W);
	Vector<F> h; B.charpoly(h);
        if(polgcd(f,h).poldeg()==f.poldeg()) {
	    Vector<F> g=poldiv(h,f);
	    Matrix<F> Im;
	    Image(Im, B, g);
	    VDual = W*Im;
	    if(VDual.ncols()==V.ncols()) {
		mesg("success!]");
		return true;
	    }
	}
	mesg("]");
    }
    mesg("failure.]");
    return false;
}

template <class F>
bool
H1NChi<F>::HeckeDual_UsingImage(Matrix<F>& VDual, const Matrix<F>& V, int stop)
{
    mesgln("HeckeDual_UsingImage");
    /* IMAGE algorithm.

       Proposition 1. Let W = space which contains the sought after VDual,
       let f=charpoly(T|VdDual), let h=charpoly(T|W), and
       write h = f*g for some g.  Then Im(g(T)) is contained in
       the kernel of f(T).  
       Proof. h(T)=0 so f(T)*g(T)(W)=0 so g(T)(W) is contained in ker(f(T)). 

       Proposition 2. Suppose h=f*g, as above, and gcd(f,g)=1. 
       Then Im(g(T))=ker(f(T)).  
       Proof. Find polynomials s and t so that s*f + g*t = 1.  
       Let v in ker(f(T)) subset W.  Then
               v = s(T)*f(T)*v + g(T)*t(T)*v 
                 = g(T)*t(T)*v    is in Im(g(T)).
       
       In generaly VDual is contained in ker(f(T)), but unfortunately 
       ker(f(T)) might be larger, and there is no simple relationship
       between Im(g(T)) and VDual. 

       Our algorithm will be to simply compute f and g as above until gcd(f,g)=1,
       and then compute Im(g(T)).  This should work well when W=Sknew since the
       eigenvalues of newforms should probably be somewhat different than eigenvalues
       of old forms and eigenvalues of eisenstein series. 
    */
    
    Matrix<F> W; W.idmat(dim_Mk());
    int n=2;
    Matrix<F> T; T.init(dim_Mk(),dim_Mk());
    while(n<stop) {
	while(gcd(n,N)!=1) n++;
	if(chi.type()==-1 && !issquarefree(n)) continue;
	mesgln("Adding on T_",n);
	Matrix<F> Tn; Tn_Mk(n,Tn);
	T += Tn;
	Matrix<F> A; T.RestrictedTo(A, V);
	Vector<F> f; A.charpoly(f);
	T.transpose();
	Matrix<F> B; T.RestrictedTo(B, W);
	Vector<F> h; B.charpoly(h);
        if(polgcd(f,h).poldeg()==f.poldeg()) {
	    Vector<F> g=poldiv(h,f);
	    Matrix<F> Im;
	    Image(Im, B, g);
	    VDual = W*Im;
	    if(VDual.ncols()==V.ncols()) {
		mesgln("success!");
		return true;
	    }
	}
	n++;
    }
    mesgln("Failure.");
    return false;
}


template <class F>
bool 
H1NChi<F>::Lfone_vanishes(const Vector<F>& f) const
{
#ifdef DEBUG
    if(k>2) error("Lfone_vanishes: not implimented in case k>2");
#endif
    Vector<F> e, fdual;
    Dual(fdual,f);
    WindingElement(e);
    
    return e*fdual == F(0);
}


template <class F> 
void
H1NChi<F>::WindingElement(Vector<F>& e, int i) const
{
    if(i<0||i>k-2)
	die("WindingElement -- invalid i");
    MSymVec<F> m; m.init(1,N);
    m.set(1, MSym(0,1,i), F(-1));
    MSymVecToVec (e, m);
}

template <class F> 
void
H1NChi<F>::LoopSymbol(Vector<F>& v,
		      const bigint& a, const bigint& b,
		      const bigint& c, const bigint& d) const 
{
    if(k==2) {
	modular_to_free_wt2_helper(v,b,d);
	return;
    } else if(k==4) {
	bigint p2=1-bigint(a)*bigint(d),
	    p1=-bigint(b) * (bigint(d)-bigint(a)),
	    p0=bigint(b)*bigint(b);

	Vector<F> w;
	modular_to_free(v,0,1,b,d,0);
	v*=F(p0);
        modular_to_free(w,0,1,b,d,1);
	v+=F(p1)*w;
	modular_to_free(w,0,1,b,d,2);
	v+=F(p2)*w;
	return;
    }

    mesg("LoopSymbol not implimented for k>4 or k odd!!!");
    v.init(dim_m);
    return;
}


// The function below does not provably terminate, as written.
// See comments below. 
template<class F>
bool
H1NChi<F>::DualBasisOfEigenforms_recurse(MatVecList<F>& E, 
					 const Matrix<F>& VDual,
					 int n) 
{
    Matrix<F> Tt;			// must be something in Hecke.
    const int cutoff=50;
    if(n<=cutoff) {		// 50 is a "random" choice
	Tn_Mk(n,Tt); 
    } else if (n<300) {		// 300 is even more random
	cerr << "DualBasisOfEigenforms_recurse: space isn't breaking up,"<<endl
	     << "maybe Q(a_n) =/= Q(a_1,a_2,a_3,...) all n!"<<endl
	     << "try Q(a_p+a_q), instead"<<endl;
	int m=n-cutoff;		// m<- n-cutoff, so  m>=1.
	int p=to_int(nth_prime(m));	// p << max int, or we'd never be
	int q=next_prime(p);	// able to compute anything anyway
	Matrix<F> Tp, Tq;
	Tn_Mk(p,Tp); Tn_Mk(q,Tq);
	Tt = Tp + Tq;	       // this should be good enough for the most part.
    } else {// there should be something else here,
	// perhaps a "desperate approach".
	cerr << "DualBasisOfEigenforms_recurse: space isn't breaking up,"<<endl
	     <<"even after trying 250 Hecke operators and some a_p+a_q's."<<endl;
	die();
    }
    
    
    Tt.transpose();		// and then transpose it.
    Matrix<F> B; Tt.RestrictedTo(B, VDual); // B <- Tt|VDual
    Vector<F> f; B.charpoly(f); // f <- charpoly(B)
    VecFac<F> fac;  f.polfactor(fac); // factor f
    for(int i=1;i<=fac.numfactors();i++) {
	Matrix<F> VDualfac;
	Kernel(VDualfac, B, fac.base(i)); // can use fac.base(i) instead
	// fac.base(i)^fac.exp(i) because
	// of semisimplicity assumption.
	if(fac.exp(i)==1) {		// multiplicity 1 factor
	    Matrix<F> C;
	    B.RestrictedTo(C, VDualfac);
	    Matrix<F> v;
	    Eigenvector(v, C, fac.base(i));
	    int len=E.len();
	    E.M(len)=VDual*(VDualfac*v); // lift up and up to Mk
	    E.v(len)=fac.base(i);
	} else {			// multiplicity > 1 factor
	    VDualfac = VDual*VDualfac;
	    MatVecList<F> Efac(VDualfac.ncols());
	    if(!DualBasisOfEigenforms_recurse(Efac, VDualfac, n+1))
		return false;
	    int len=E.len();
	    for(int j=0; j<Efac.len(); j++) {
		E.M(j+len)=Efac.M(j);
		E.v(j+len)=Efac.v(j);
	    }
	}
    }
    return true;
}

template<class F>
bool
H1NChi<F>::DualBasisOfEigenforms_compute(MatVecList<F>& E,    
                                         const Matrix<F>& V) 
{
    mesgln("[DualBasisOfEigenforms_compute(E,V)]");
    E.init(V.ncols());
    Matrix<F> VDual;
    if(!ComputeHeckeDual(VDual, V, 53)) {
	cerr << "couldn't find hecke dual of V." << endl;
	return false;
    }
    return DualBasisOfEigenforms_recurse(E,VDual,2);
}

template<class F>
void
H1NChi<F>::ConjugationSubspace(Matrix<F>& W,
			       const Matrix<F>& V,
			       bool plusone) const
{
    if(plus) {			// easy case.
	W=V;
	return ;
    }

    Matrix<F> C, B, I;
    Conjugation_Mk(C);
    C.RestrictedTo(B, V);
    I.idmat(B.nrows());
    if(plusone)
	B=B-I;
    else
	B=B+I;
    Matrix<F> K;
    Kernel(K,B);
    W=V*K;
}

template<class F>
bool
H1NChi<F>::DualBasisOfEigenforms(MatVecList<F>& E,
				 const Matrix<F>& V) 
{
    mesgln("[DualBasisOfEigenforms(E,V)]");
    if(!plus)
	mesgln("********** WARNING: Don't run DualBasisOfEigenforms if not on plus space!");
    if (V.ncols() != dim_Sknew()) // can't load.
      return DualBasisOfEigenforms_compute(E,V);
    
    if (save) {
	char s[1024] ; sprintf(s, "%sEigen", fname);
	ifstream in(data_read(s,N)) ;
	if(in) {
	    mesg("** Loading dual eigenbasis from datafile.\n");
	    E.Load(in);
	    return true; 
	}
    } 
    
    // compute
    if(!DualBasisOfEigenforms_compute(E, get_Sknew())) 
      return false;
    
    if(save) {
	char s[1024] ; sprintf (s, "%sEigen", fname) ;
	ofstream of(data_write(s,N));
	if (!of) 
	    error("Error trying to save Eigen.", NONTERMINAL) ;
	else
	    E.Save(of);
    }
    return true;
}


template <class F> 
bool
H1NChi<F>::BasisOfEigenforms(MatVecList<F>& E,
			     const Matrix<F>& V,
			     int maxn, int n) 
{
    die("BasisOfEigenforms not programmed!");
}


template <class F> 
bool
H1NChi<F>::HeckeEigenvalues_ap(MatVecList<F>& aplist,
			       MatVecList<F>& W,
			       const Matrix<F>& V,  
			       int maxp) 
{
    // 1. Find a dual basis of eigenforms using DualBasisOfEigenforms
    
    // 2. Systematically find a very simple modular symbol e such
    //    that <e,f> =/= 0 for all f in the dual basis.  Save the
    //    nonzero values <e,f> once they are found. 
    
    // 3. Without first computing T_p on the whole space, compute
    //    e_p:=T_p(e) as quickly as possible.  This is the MAIN
    //    bottleneck in the computation. 
    
    // 4. Record the quotients a_p=<e_p,f>/<e,f>
    
    // Step 1.
    //  MatVecList<F> W;
    if(!DualBasisOfEigenforms(W, V)) return false;
    cout << "dim V = " << V.ncols()<<endl;
    cout << "num eigenforms found = " << W.len()<<endl;
    
    // Step 2.
    mesg("Finding e.\n");
    
    // Find a vector of length dim_m, as simple as possible, so that it is
    // *not* in the kernel of matrix multiplication on the left
    // by any of the W[i].  The entries of e are either 1 or 0.
    Vector<int> e; e.init(dim_m); 
    // Step 2a. quickly try vectors with exactly one nonzero entry
    bool e_simple=false;		// e_simple means we e is simple.
    int e_simple_pos;
    Vector<F> ef[W.len()];		
    for(int i=1;i<=dim_m;i++) {
	bool all_nonzero=true;
	for(int j=0;j<W.len();j++) {
	    W.M(j).row(i,ef[j]);
	    if(ef[j].iszero()) { all_nonzero=false; break; }
	}
	if (all_nonzero) {
	    e[i]=1; e_simple=true; e_simple_pos=i;
	    break; 
	}
    }
    
    // Step 2b. 2a didn't find anything, so use a systematic (but slower) search.
    if(!e_simple) {
	mesg("No monomial e. Trying randomized strategy: ");
	// (i) Find an e not orthogonal to any W[i]
	//     Using a slow search here makes sense because choosing
	//     a good e is crucial to the speed of the rest of the algorithm. 
	int i=2;
	bool not_done=true;
	while(i<=dim_m && not_done) {
	    // try a few hundred random vectors of length i.
	    for(int j=0;j<300;j++) {
		int a[i];
		for(int n=0;n<i;n++)
		    a[n]=RandomBnd(dim_m)+1;
		bool all_nonzero=true;
		for(int n=0;n<W.len();n++) {
		    Vector<F> row;
		    ef[n].init(W.M(n).ncols());
		    for(int m=0;m<i;m++) {
			W.M(n).row(a[m],row);
			ef[n] += row;
		    }
		    if(ef[n].iszero()) { all_nonzero=false; break; }
		}
		if(all_nonzero) {			   // found it!
		    for(int n=0;n<i;n++)
			e[a[n]]=1;
		    not_done=false;
		    break;
		}
	    }
	    if(not_done) i++;
	}
	
	mesg("Randomized strategy suceeded.");
	// (ii) find all pairings ef
	// ... this is automatic in the above algorithm. 
	
    }
    
    // Compute and save the values 1/<e,f>.
    mesg("Computing the values 1/<e,f>. [");
    Vector<F> oneover_ef[W.len()];
    for(int i=0;i<W.len();i++) {
	reset_timer();
	oneover_ef[i] = ef[i].polinv_polymod(W.v(i));
	mesg((i*100)/W.len(),"% ");
	mesg("(");mesg(get_time());mesg("s)");
    }
    mesg("]\n");

    int numprimes=0;
    for (int p=2; p<=maxp; p=next_prime(p)) 
	if(is_prime(p)) numprimes++;
    
    aplist.init(W.len());
    for(int i=0;i<W.len();i++) {
	aplist.M(i).init(numprimes,W.v(i).poldeg());
	aplist.v(i)=W.v(i);
    }
    int pidx=1;
    mesg("Computing ap:  [");
    for (int p=2; p<=maxp; p=next_prime(p)) {
        mesg(p," ");
	// Step 3. 
	// Compute T_p(e). 
	Vector<F> Tpe;
	Array<int> H;
	Tp_Heilbronn (p, H) ;
	
	if(e_simple) {
	    // Compute T_p ( e_simple_pos -th generating Manin symbol )
	    Tp_StandardSymbol(Tpe,H,e_simple_pos);
	} else {
	    Tpe.init(dim_m);
	    for(int i=1;i<=e.size();i++) 
		if(e[i]) {
		    Vector<F> Tpei;
		    Tp_StandardSymbol(Tpei,H,i);
		    Tpe+=Tpei;
		}
	}
	
	// Step 4. 
	// Compute ap
	for(int i=0; i<W.len(); i++) {
	    Vector<F> Tpef = Tpe * W.M(i);
	    Vector<F> ap=polmul_polymod(Tpef,oneover_ef[i],W.v(i));
	    aplist.M(i).setrow(pidx,ap);
	}
	pidx++;
    }

    mesg("]");
    return true;
}

/************************ WARNING *****************
 THE RECURSIVE ALGORITHM USED IN THIS ROUTINE IS
 AMAZINGLY SLOW (COMPARED TO THE OPTIMAL ALGORITHM)!  
 THIS IS JUST TO GET SOMETHING WORKING QUICKLY AND 
 WILL BE REPLACED.
***************************************************/
template <class F>
void 
H1NChi<F>::HeckeEigenvalues_get_apr(Vector<F>& apr,
				    int p, int r,
				    const Matrix<F>& aplist,
				    const Vector<F>& f,
				    int* pidx) 
{
    if (r==1) {
	aplist.row(pidx[p], apr);
	return ;
    }
    if (r==0) {
	apr.init(f.size()-1);
	apr[1]=1;
	return;
    }
    if (r>1) {
	Vector<F> a, b, c;
	HeckeEigenvalues_get_apr(a,p,r-1,aplist,f,pidx);
	HeckeEigenvalues_get_apr(b,p,1,aplist,f,pidx);
	HeckeEigenvalues_get_apr(c,p,r-2,aplist,f,pidx);
	bigint q;			// q=p^{k-1}
	power(q,p,k-1);
	if(chi.type()==-1) {
	    apr = polmul_polymod(a,b,f) -F(F(p)*F(q))*c; 
	    mesg("WARNING: eigenvalue recurrence for Gamma_1 not yet correctly programmed.\n");
	} else
	    apr = polmul_polymod(a,b,f) -F(F(chi.of(p))*F(q))*c; 
    }
}

template <class F> 
bool
H1NChi<F>::HeckeEigenvalues_ap(MatVecList<F>& aplist,
			       const Matrix<F>& V,  
			       int maxn) 
    
{
    MatVecList<F> W;
    return HeckeEigenvalues_ap(aplist, W, V, maxn);
}


template <class F> 
bool
H1NChi<F>::HeckeEigenvalues_an(MatVecList<F>& anlist,
			       const Matrix<F>& V,  
			       int maxn) 
    
{
    MatVecList<F> W;
    return HeckeEigenvalues_an(anlist, W, V, maxn);
}


template <class F> 
bool
H1NChi<F>::HeckeEigenvalues_an(MatVecList<F>& anlist,
			       MatVecList<F>& W,
			       const Matrix<F>& V,  
			       int maxn) 
{
    // 1. Compute HeckeEigenvalues_ap and then use the recurrences, 
    //    which we know correctly since we're "inside" the appropriate
    //    H1NChi object, to compute the an.  Note that the character
    //    plays a role in computing the recurrence. 
    
    MatVecList<F> aplist;
    if(!HeckeEigenvalues_ap(aplist, W, V, maxn))
	return false;

    mesg("[Recursively computing an...");
    
    // now we know a_p for all p<=n.  
    // 
    anlist.init(aplist.len());
    
    int pidx[maxn+1];
    int i=1;
    for(int p=1; p<=maxn; p++)
	if(is_prime(p)) {
	    pidx[p]=i;
	    i++;
	} else pidx[p]=0;
    
    for(int i=0; i<aplist.len(); i++) {
	anlist.v(i) = aplist.v(i);
	anlist.M(i).init(maxn,anlist.v(i).size()-1);
	mesg("\nf",i);
	mesg(": [");
	for(int n=1; n<=maxn; n++) {
	    mesg(" ",n);
	    if(pidx[n]) {
		Vector<F> aprow;
		aplist.M(i).row(pidx[n],aprow);
		anlist.M(i).setrow(n,aprow);
	    } else {
		Vector<F> an(aplist.v(i).size()-1);
		an[1]=F(1);				   // make an=1 polymod.
		int m=n;
		while(m>1) {
		    int p=2;
		    while (m % p) p++ ;
		    int prime_to_p_part ;
		    int r = ordp (p, m, prime_to_p_part) ;
		    Vector<F> apr;
		    HeckeEigenvalues_get_apr(apr,p,r,aplist.M(i),aplist.v(i),pidx);
		    an = polmul_polymod(apr, an, aplist.v(i));
		    m = prime_to_p_part;
		}
		an.resize(anlist.M(i).ncols());
		anlist.M(i).setrow(n,an);
	    }
	}
	mesg("]\n");
    }
    mesg("]");
    return true;
}

template <class F> 
bool
H1NChi<F>::Basis_q_expansions(Matrix<F>& Q,
			      const Matrix<F>& V,
			      int maxn) 
{
    // 1. Compute the basis given by HeckeEigenvalues_an,
    //    and then simply project.  This will give a basis
    //    with coefficients in F. 
    
    MatVecList<F> anlist;
    if(!HeckeEigenvalues_an(anlist, V, maxn))
	return false; 
    int d=0;
    for(int i=0; i<anlist.len(); i++)
	d+=anlist.v(i).size()-1;
    Q.init(maxn,d);
    d=1;
    for(int i=0; i<anlist.len(); i++) 
	for (int j=1; j<=anlist.M(i).ncols(); j++) {
	    Vector<F> col;
	    anlist.M(i).col(j,col);
	    Q.setcol(d,col);
	    d++;
	}
    return true;
}

template<class F>
void
H1NChi<F>::BSD(F& bsd, 
	       const Vector<F>& f,
	       const Matrix<F>& v,
	       const bigrational& discZf,
	       int i) 
{
    mesg(" ** BSD not supported in H1NChi<F>, need integral structure (use H1NChiZ).");
}

template<class F>
void
H1NChi<F>::AllBSD(Matrix<F>& bsd,
		  Vector<F>& disc,
		  const MatVecList<F>& eigenforms,
		  const MatVecList<F>& anlist) 
{
    mesg(" ** AllBSD not supported in H1NChi<F>, need integral structure (use H1NChiZ).");
}

template<class F>
void
H1NChi<F>::OrderByTrace(Vector<int>& perm,
			const MatVecList<F>& anlist) 
{
    mesg("In Hecke, only the bigrational field is ordered.\n");
    mesg("Unless you use H1NChiZ I'm not sure your in that field, so no reordering is taking place.\n");
}

template<class F>
void
H1NChi<F>::OrderByTrace(MatVecList<F>& anlist,
			MatVecList<F>&) 
{
    mesg("In Hecke, only the bigrational field is ordered.\n");
    mesg("Unless you use H1NChiZ I'm not sure your in that field, so no reordering is taking place.\n");
}

template<class F>
void
H1NChi<F>::OrderByTrace(MatVecList<F>& anlist) 
{
    mesg("In Hecke, only the bigrational field is ordered.\n");
    mesg("Unless you use H1NChiZ I'm not sure your in that field, so no reordering is taking place.\n");
}


template <class F>
bool H1NChi<F>::CanReduce_charmod() const
{
    return can_reduce_charmod(Lift) && can_reduce_charmod(SI_coeff) ;
}

template<class F>
void H1NChi<F>::build(const Character&  b_chi,
		      const MSymbols&   b_msymbols,
		      const int&        b_k,
		      const int&        b_N,
		      const bool&       b_save,
		      const bool&       b_plus,
		      char*             b_fname, // b_fname is deleted by this.
		      const int&        b_dim_m,
		      const Vector<int>&b_Mk,
		      const Matrix<F>&  b_Lift,
		      const Array<F>&   b_SI_coeff,
		      const Array<int>& b_SI_gens,
		      const Array<int>& b_SI_rels)
{
    mesgln("Building h1nchi object...");
    init_delta = init_sk = init_sknew = init_eis = false ;
    chi        = b_chi;
    msymbols   = b_msymbols;
    k          = b_k;
    N          = b_N;
    save       = b_save;
    plus       = b_plus;
    fname      = b_fname;
    dim_m      = b_dim_m;
    Mk         = b_Mk;
    Lift       = b_Lift;
    SI_coeff   = b_SI_coeff;
    SI_gens    = b_SI_gens;
    SI_rels    = b_SI_rels;
}
		      

template<class F>
void H1NChi<F>::Reduce_charmod(H1NChi<charmod>& Hmod, charmod z) const
{
    mesgln("Reduce_charmod: Warning reducing modulo a 'random' lattice.");
    if(!CanReduce_charmod()) 
	die("Reduce_charmod() -- not possible to reduce!");

    int p=charmod::get_pr();

    char* Hmodname=NULL;
    if (save) {
	Hmodname = new char[strlen(fname)+64] ;
	sprintf(Hmodname, "%s_Mod%d_%d", fname, p, lift(z)) ;
    }

    Matrix<charmod> HmodLift;
    to_charmod(HmodLift,Lift,z);
    Array<charmod> HmodSI_coeff;
    to_charmod(HmodSI_coeff,SI_coeff,z);

    Hmod.build(chi,msymbols,k,N,save,plus,Hmodname,
	       dim_m,Mk,
	       HmodLift, HmodSI_coeff, SI_gens, SI_rels);
}




//////////// TEMPLATE INSTANTIATION /////////////////////

template <class F>
ostream& operator<<(ostream& os, const MSymVec<F>& m)
{
    for(int i=1;i<=m.size();i++) {
      if(m.c(i)!=F(1))
	os << m.c(i) ;
      os << m.x(i) ;
	if (i<m.size()) os << "+" ;
    }
    return os ;
}


#ifdef INST_BIGRATIONAL
template class H1NChi<bigrational> ;
template class MSymVec<bigrational> ;
template ostream& operator<< (ostream&, const MSymVec<bigrational>&);
#endif
#ifdef INST_FFIELD
template class H1NChi<ffield_element> ;
template class MSymVec<ffield_element> ;
template ostream& operator<< (ostream&, const MSymVec<ffield_element>&);
#endif
#ifdef INST_CHARMOD
template class H1NChi<charmod> ;
template class MSymVec<charmod> ;
template ostream& operator<< (ostream&, const MSymVec<charmod>&);
#endif
#ifdef INST_INTMOD
template class H1NChi<intmod> ;
template class MSymVec<intmod> ;
template ostream& operator<< (ostream&, const MSymVec<intmod>&);
#endif
#ifdef INST_ALGNUM
template class H1NChi<algnum> ;
template class MSymVec<algnum> ;
template ostream& operator<< (ostream&, const MSymVec<algnum>&);
#endif

template <class F>
void MSymVec<F>::ModSymPrint(ostream& os, int k) const
{
    for(int i=1;i<=size();i++) {
      if(c(i)!=F(1))
	os << c(i) << "*";
      x(i).ModSymPrint(os,N(),k);
      if (i<size()) os << "+" ;
    }
}
