/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/* 
   FILE: h1nchi.cc
   DESC: Modular symbols of weight k>=2 of any field.
*/

#ifndef H1NCHI_H
#define H1NCHI_H

#include "character.h" 
#include "matrix.h"
#include "sparse.h"
#include "array.h"
#include "msymbols.h"
#include "fields.h"		

// F is the base field:
// Tested fields now include
//               * ffield_element (essentially LiDIA's finite field)
//               * bigrational (LiDIA's arb. prec. rational numbers)
//               * algebraic_number (LiDIA's algebraic number field class)
//               * charmod (Luiz's 1 "8 byte" mod p, 3<=p<=253 class.)
//               * intmod (Luiz's int mod p)
// See "fields.h".
// At some point it might be good to add a longlongmod type.

template <class F> class MSymVec;

static int max_Tp_Got;

template <class F>
class H1NChi 
{
 public:
    H1NChi (const Character& chi, int k = 2,	
	    bool save = true, bool plus=true) ;
    virtual ~H1NChi () ; // virtual destructors are good so objects 
    // can best be polymorphic.

    friend class H1NChiZ;

    // use at own risk.
    void build(const Character&  b_chi,
	       const MSymbols&   b_msymbols,
	       const int&        b_k,
	       const int&        b_N,
	       const bool&       b_save,
	       const bool&       b_plus,
	       char*             b_name,	// b_name is deleted by this,
	       const int&        b_dim_m,       // iff save=true.
	       const Vector<int>&  b_Mk,
	       const Matrix<F>&  b_Lift,
	       const Array<F>&   b_SI_coeff,
	       const Array<int>& b_SI_gens,
	       const Array<int>& b_SI_rels);

    const MSymbols& get_MSymbols() { return msymbols; }
    void MSymRelations(Matrix<F>& M);
    
    // returns a basis for Sk.
    const Matrix<F>& get_Sk() ;

    // returns a basis for the Eisenstein part of Mk, that is, the
    // Hecke complement of Sk in Mk.
    const Matrix<F>& get_Eis() ;
    
    // computes a basis for the subspace of cusp forms in V. 
    void get_Sk_in_V (Matrix<F>& W, const Matrix<F>& V) ;
    

    // This computes action of T_p using Heilbronn matrices.
    // In fact, this should work fine even for T_n, n composite.
    void Tp_Mk  (int p, Matrix<F>& M) const ;
    void Tn_Mk  (Matrix<F>& M, int n) const { Tp_Mk(n,M); }

    void Tp_Mk  (Matrix<F>& M, int n) const { Tp_Mk(n,M); }

    // When Tp_Mk(p,M) is called, it sets max_Tp_Got to max(p,max_Tp_Got).
    // Note that there is only one variable max_Tp_Got for all H1 objects.
    int  get_max_Tp_Got () const { return max_Tp_Got; }
    void reset_max_Tp_Got() { max_Tp_Got = 2; }


				// action of T_n on M_k.
    void Tn_Mk   (int n, Matrix<F>& T) const ;
    void Tn_Sk   (int n, Matrix<F>& T) { Tn_On(n, get_Sk(), T); }
				// compute the lambda eigenspace for the
				// action of T_n on the whole M_k.
    void Tn_Mk_Eig (int n, const F& lambda, Matrix<F>& ES) const ;
    void Tn_Mk_Eig (int n, const Vector<F>& f, Matrix<F>& ES) const ;
    
    void Tn_Sk_Eig (int n, const F& lambda, Matrix<F>& ES)
	{ Tn_Eig_On (n, get_Sk(), lambda, ES); }
    void Tn_Sk_Eig (int n, const Vector<F>& f, Matrix<F>& ES)
	{ Tn_Eig_On (n, get_Sk(), f, ES); }
    
				// GammaU = action of u in (Z/NZ)^*
    void U_Mk(int u, Matrix<F>& U) const ;

    // Compute the action of u in Z/NZ on M_k(Q). 
    void ZmodNZ_Mk(Matrix<F>& U, int u) const;
    void ZmodNZ_On(Matrix<F>& U, const Matrix<F>& V, int u) const;

    // Compute the Atkin-Lehner involution using modular symbols,
    // instead of Manin symbols.  (Cremona suggested that this is much faster.)
    void Wq_Mk(Matrix<F>& Wq, int q);
    void Wq_On(Matrix<F>& Wq, const Matrix<F>& V, int q);
    // i=ith prime divisor of N.
    const Matrix<F>& get_Wq(int i);
    // Compute modular symbols action of M in M_2(Z) on the
    // i-th freely generating Modular symbol.
    void ActStandardModularSymbol(Vector<F>& v,
				  const Matrix<bigint>& M,
				  int i) const;
    
    /* INPUT: a matrix M=(a,b;c,d) in M_2(Z) and
       a polynomial P(X,Y) = X^qY(k-2-q).
       OUTPUT: Array of coefficients of 
               M.P := P(dX-bY, -cX+aY)
                   coeff[0]*Y^(k-2)+ coeff[1]*X*Y^(k-3) +
		   ... + coeff[k-2]*X^(k-2)
    */
    void M2ZAction (Array<F>& coeff,
		    const Matrix<bigint>& M,
		    const int& q) const;
    // Compute action of M in M_2(Z) on the specific
    // modular symbol X^qY(k-2-q){0,oo} and return
    // the answer in terms of the free basis for M_k(Q).
    void ActWindingModularSymbol(Vector<F>& v,
				 const Matrix<bigint>& M,
				 int q) const;




    
				// Atkin-Lehner involutions. 
				// I don't think this map must be well-defined
				// for an arbitrary nontrivial character.
				// The map computed is W_q where
				// q is the product p^n over all
				// p dividing both M and N and n=ord_p(N).
				// If M=0 then the involution W_N is computed.
				// If M=p is a prime divisor of N then 
				// W_q is computed where p^n exactly divides N.

				// This is a slow manin symbols implimentation,
                                // use Wq_Mk instead.
    void WN_Mk(int M, Matrix<F>& WN)  const; 
    void WN_Sk(int M, Matrix<F>& WN)  { WN_On(M, get_Sk(), WN); }

    void Wq_StandardSymbol(Vector<F>& WqSymi,
			   const Array<int>& H, int i, int q,
			   bigint&, bigint&) const ;


                                // Returns a matrix representing the
                                // map (c:d) |--> (-c:d) (on modular symbols),
                                // defined by the matrix T=(-1,0;0,1).
                                // This map MUST be trivial if we are
                                // working in the +1 quotient space. 
    void Conjugation_Mk(Matrix<F>& C) const;
    // sets W = +1 or -1 subspace for conjugation action on V.
    void ConjugationSubspace(Matrix<F>& W,
			     const Matrix<F>& V,
			     bool plusone=true) const;

    
				// Matrix of action of T_n on an invariant
				// subspace W, wrt basis of columns of W. 
    void Tn_On   (int n, const Matrix<F>& W, Matrix<F>& T) const ;
				// lambda eigenspace of action of T_n on W. 
    void Tn_Eig_On (int n, const Matrix<F>& W,
		    const F& lambda, Matrix<F>& ES) const ;

				// ES = ker(f(T_n)|W)
    void Tn_Eig_On (int n, const Matrix<F>& W,
		    const Vector<F>& f, Matrix<F>& ES) const ;

    /* Quickly compute a list of Hecke operators T1, ..., Tn
       doing all recurrences in memory.  The returned pointer pT points 
       to heap and must be deleted by the client.  The first entry pT[0]
       holds the matrix repn. for T1=identity, and the last entry pT[n-1]
       holds the matrix repn. for Tn.
    */
    Matrix<F>* TnList_Mk(int n);
    Matrix<F>* TnList_On(int n, const Matrix<F>& V);

///////////////////////// Dual /////////////////////////////////

    void Dual_Eig (Matrix<F>& V,
		   int n, const F& lambda) const;
    void Dual_Eig (Matrix<F>& V,
		   int n, const Vector<F>& f) const;
				// W *must* be an invariant subspace of Mkdual!
				// V = ker((mattranspose(T_n)-lambda)|W)) subset Mkdual.
    void Dual_Eig_On (Matrix<F>& V,
			int n, const F& lambda, const Matrix<F>& W) const;
				// V = ker(f(mattranspose(T_n)|W)) subset Mkdual.
    void Dual_Eig_On (Matrix<F>& V,
			int n, const Vector<F>& f, const Matrix<F>& W) const;

				// pr=list of pr.size() primes p_0, ..., p_s.
				// f = array of pr.size() polys f(x) in F[x].
				// computes intersect ker(f(T_{p_i}^t)).
				// Computes first T_{p_0} in the usual way,
				// then uses Dual_Tp_On	to compute each successive T_p.
                                // V is the subspace of the dual.
    void Dual_Eig_Intersect (Matrix<F>& V,
			     const Array<int>& pr,
			     const Vector<F>*& f,
			     int   stop_dimension=0);
    void Dual_Eig_Intersect (Matrix<F>& V,
			     const Array<int>& pr,
			     const Vector<F>*& f,
			     const Matrix<F>& Start,
			     int   start_at=0,
			     int   stop_dimension=0);
				// INPUT: A subspace V of EDual, found for example using
				// one of the above functions.  It is assumed that V
				// is strongly IRREDUCIBLE for the Hecke action, i.e., 
				// there is some element t in the Hecke algebra whose
				// characteristic polynomial is irreducible.
                                // OUTPUT: A left eigenvector v for t over K.
    void Dual_FindEigenvector (Matrix<F>& v, Vector<F>& K,
			       const Matrix<F>& V);

				// Dual_E: Given an n dim invariant subspace V of Mkdual (with
                                // basis the columns of V), compute a list of positions ei
				// so that the map
				//    v|-->(<v,e1>,<v,e2>,...,<v,en>) 
				// is an isomorphism. Here <v,ei>=v[i]=ith entry of v.
                                // The ei are the pivot columns of the reduced row echelon
                                // form of mattranspose(V). 
    void Dual_E (Vector<int>& e,
		 Matrix<F>& D,
		 const Matrix<F>& V) const ;

    void Dual_Tp_On(Matrix<F>& T,
		    int p,
		    const Matrix<F>& V,
		    const Vector<int>& e,
		    const Matrix<F>& D) const;

    bool HeckeDual_UsingImage(Matrix<F>& VDual,
			      const Matrix<F>& V, 
			      int stop=97);

///////////////////////////////////////////////////////////////    

				// Atkin-Lehner involution W_N
				// slow manin symbols implimentation,
                                // use Wq_Mk instead.
    void WN_On(int M, const Matrix<F>& W, Matrix<F>& WN) ;
    
    int dim_Mk() const { return dim_m ; }
    int dim_Sk(); 
    int dim_Sknew() { return get_Sknew().ncols(); }


    // Weight 2 algorithm:
    // Use continued fractions (Manin's trick) to convert to
    // modular symbols {a/b,c/d} to a sum of Manin symbols, expressed 
    // on the free basis.  The helper function converts {0,c/d}.
    void modular_to_free_wt2_helper(Vector<F>& v,
				 const bigint& c, const bigint& d)const;
    void modular_to_free_wt2(Vector<F>& v,
			  const bigint& a, const bigint& b,
			  const bigint& c, const bigint& d) const;


    // Weight >= 2 algorithm:
    // Conversion from modular symbol to expression on free basis:
    // Input:  modular symbol X^q*Y^(k-2-q) {a/b,c/d}
    // Output: the corresponding representation v on the free basis 
    void modular_to_free(Vector<F>& v,
			 const bigint& a, const bigint& b,
			 const bigint& c, const bigint& d,
			 const int& q) const;
    void modular_to_free_helper(Vector<F>& v,
				const bigint& a, const bigint& b,
				const int& q) const;
    // Input:  modular symbol (qar[0]*Y^(k-2)+...+qar[k-2]*X^(k-2)){a/b,c/d}
    // Output: the corresponding representation of v on the free basis 
    void modular_to_free(Vector<F>& v,
			 const bigint& a, const bigint& b,
			 const bigint& c, const bigint& d,
			 const Array<F>& qar) const;
    // Assume k=even, trivial character.
    // Given a matrix g=(a,b;c,d) in Gamma_0(N), returns vector on free basis
    // corresponding to the modular symbol
    //      ((1-ad)X^2-b(d-a)XY+b^2Y^2)^((k-2)/2) * {0, c/d}.
    // Also, P[1]*Y^(k-2)+... = ((1-ad)X^2-b(d-a)XY+b^2Y^2)^((k-2)/2).
    void Invariant_Symbol(Vector<F>& v, Vector<bigint>& P,
			 const bigint& a, const bigint& b,
			 const bigint& c, const bigint& d) const ;

    void Invariant_Symbol_Matrices(MatVecList<bigint>& List,
				   Matrix<F>& PhiGood,
				   const Matrix<F>& Phi) const;

    // functions related to newforms and oldforms
    void HeilbronnC (int p, Array<int>& H) const ;
    
				// M = level of f. 
				// M|N, t|(N/M), g=Eta_{M,t}(f)=d^{k/2-1}f(q^t)??
				// Heil = Heilbronn matrices of determinant t,
				//        as computed by HeilbronnC!
    void Eta_Raise (Vector<F>& g,	
		    const MSymVec<F>& f,
		    const int& t,              // USE HEILBRONNC!!, not Heilbronn!
		    const Array<int>& Heil) const ;
    
    void Eta_Lower (Vector<F>& g,	       // M = level of f
		    const MSymVec<F>& f, // N|M, t|(M/N)
		    const int& t,              // USE HEILBRONNC!!, not Heilbronn!
		    const Array<int>& Heil) const ;

/*    // Compute Lower_s ( Raise_t (f) ). 
    // If s==t the result should be a scalar multiple of f. 
    void Eta_RaiseThenLower (Vector<F>& g,
			     const MSymVec<F>& f,
			     int t,
			     int s,
			     int NN,
			     const Array<int>& Heil) const;
*/
    
    // NOTE : "at" part is not yet implimented. 
    // all forms new at "at":  at is a product of primes dividing N. 
    // This is the orthogonal complement of the images of the 
    // inclusion maps from lower levels M|N with N/M a prime divisor
    // of at.   
    void get_Mknew (Matrix<F>& Mn, int at=1) const ;
    void get_Sknew (Matrix<F>& Sn, int at=1) ;
    void get_SknewPlus (Matrix<F>& SnPlus, int at=1) ;
    const Matrix<F>& get_Sknew();
    
    // (quickly) computes a Hecke invariant space
    // which contains (and usually equals) Mknewdual.
    void get_Mknewdual (Matrix<F>& Mn, int at=1) const ;
    
    
				// Converts a vector on the free generators
				// to a linear combination of Manin symbols.
				// The level N is saved in g, but the 
				// character is *not*.
    void VecToMSymVec(MSymVec<F>& g, const Vector<F>& f) const ;
    void MSymVecToVec(Vector<F>& v, const MSymVec<F>& m) const ;
    void MSymToVec(Vector<F>& v, const MSym& m) const ;
    
				// Given an eigenvector v for all of the
				// Hecke operators, return a dual v^ of v.
				// The dual is a left eigenvector for the
				// Hecke operators, w.r.t our fixed basis,
				// which has the same system of eigenvalues
				// as v.  Note that v^ is only defined
				// up to a nonzero scalar. 
    void Dual(Vector<F>& fdual, const Vector<F>& f) const;
    
				// V = Hecke equivariant subspace of M_k
				// maxn = positive integer
				// VDual = dual Hecke equivariant space.
				// This means that the transposes of the
				// Hecke operators act on VDual in
				// the same way as the usual Hecke operators
				// act on V.  To find VDual, the program
				// computes matker(charpoly(T_p|V)(T_p^{t}))
				// on the VDual computed so far, 
                                // for p<=paxp or until dim VDual = dim V. 
				// It is certainly possible that no such
				// VDual can be found if p is too small.
				// In that case the VDual returned will
				// have too large of dimension. 
				// The algorithm is recursive, and knows
				// the base case from when startp=0.
    bool ComputeHeckeDual(Matrix<F>& VDual, const Matrix<F>& V,
			  int maxn) ;
				// Use the dual approach, i.e., build up 
                                // V instead of cutting down to get V.
                                // Much faster if codim(V) is small.
    bool ComputeHeckeDual_useimage(Matrix<F>& VDual, const Matrix<F>& V,
			  int startn, int stopn) ;

    
				// Given an eigenform f for all T_p, p
				// prime to N, there is an L-function
				// L(f,s).  This function returns true
				// iff L(f,1)==0.  
				// (see 2.8 and 2.9 of Cremona's book)
    bool Lfone_vanishes(const Vector<F>& f) const;
    
    
				// return -X^i*Y^(k-2-i)*{0,\infty} in 
				// terms of free basis.
    void WindingElement(Vector<F>& e, int i=0) const ;
    bigrational WindingBSD() ;
    
    // Compute the cyclic Hecke Z-module 
    //     Tx = {T*x : T in Hecke algebra } 
    // generated by the modular symbol x in M_k(Gamma,Q).
    // x is represented in terms of the free basis.
    // Algorithm: (1) find the nonzero entries of x, so if
    //                x is sparse the algorithm is MUCH faster.
    //            (2) compute the Sturm bound n so that
    //                T_1,...,T_n generate the Hecke algebra.
    //            (3) compute M = [T_1*x | ... |T_n*x]
    //  -- return Tx=M.
    // In the derived class H1NChiZ the following additional step
    // is taken.
    //            (4) Let Tx = [ Z-basis for column space of M ].
    virtual void CyclicHeckeModule(Matrix<F>& Tx,
				   const Vector<F>& x);

				// LoopSymbol:
				// Assume k is even (only implimented for k<=4.)
				// Given g=(a,b;c,d), return m=P(X,Y)*{0,g(0)}, in
				// terms of the free basis, where
				//  P(X,Y) = ((1-ad)X^2 - b(d-a)XY+b^2Y^2)^((k-2)/2)).
				// Observe that P is fixed by g and so that m
				// is in S_k(Gamma,Z).  
    void LoopSymbol(Vector<F>& v,
		    const bigint& a, const bigint& b,
		    const bigint& c, const bigint& d) const ;

    // use first bound Hecke operators, if bound=0, compute Sturm bound.
    virtual void CyclicHeckeImage(Matrix<F>& PTx,
				  const Matrix<F>& P,
				  const Vector<F>& x,
				  int bound=0);

				// Print the number of Heilbronn matrices
				// of determinent p. 
    int  num_Heilbronn (const int& p) const;
    void transform();
    
    
    //              IN ALL OF THE FOLLOWING, V IS ***ASSUMED***
    //              TO BE A SUBSPACE OF S_k^{new}!
    //   
    // Attempt to compute a basis of eigenforms (really conjugacy
    // classes of eigenforms over F).  This is sometimes *not* possible due
    // to the structure of the Hecke algebra over F.
    // Give up if a decomposition can't be found after trying the
    // Hecke operators T_n for n<=maxn. 
    bool DualBasisOfEigenforms(MatVecList<F>& E,
			       const Matrix<F>& V) ; 
    bool DualBasisOfEigenforms_compute(MatVecList<F>& E,    // helper function
				       const Matrix<F>& V); 
    bool DualBasisOfEigenforms_recurse(MatVecList<F>& E,    // helper function
				       const Matrix<F>& VDual,  
				       int n) ; 
    
    bool BasisOfEigenforms(MatVecList<F>& E,
			   const Matrix<F>& V,  // Hecke stable subspace
			   int maxn, int n=2) ; 
    
    bool HeckeEigenvalues_ap(MatVecList<F>& aplist,
			     MatVecList<F>& W,  // basis of (dual) eigenforms.
			     const Matrix<F>& V,  // Hecke stable subspace
			     int maxp) ;
    
    bool HeckeEigenvalues_ap(MatVecList<F>& aplist,
			     const Matrix<F>& V,  // Hecke stable subspace
			     int maxp) ;
    
    void HeckeEigenvalues_get_apr(Vector<F>& apr,
				  int p, int r,
				  const Matrix<F>& aplist,
				  const Vector<F>& f,
				  int* pidx) ;
    
				// same as HeckeEigenvalues but gives a list
				// of a_n, instead of just the a_p. 
    bool HeckeEigenvalues_an(MatVecList<F>& anlist,
			     const Matrix<F>& V,  // Hecke stable subspace
			     int maxn) ;
    bool HeckeEigenvalues_an(MatVecList<F>& anlist,
			     MatVecList<F>& eigenforms,
			     const Matrix<F>& V,  // Hecke stable subspace
			     int maxn) ;
    
    // Attempt to find a basis of q-expansions with coefficients in F.
    // This is not always possible.  It is assumed that V is a subspace
    // of the space of cusp forms.
    // The n-th column of Q is the q-expansion of the nth basis vector.
    bool Basis_q_expansions(Matrix<F>& Q,
			    const Matrix<F>& V,
			    int maxn) ; // Hecke stable subspace
    
    // Not supported -- needs integral structure. 
    virtual void BSD(F& bsd,
		     const Vector<F>& f,
		     const Matrix<F>& v,
		     const bigrational& discZf,
		     int i=0) ;
    
    virtual void AllBSD(Matrix<F>& bsd,
			Vector<F>& disc,
			const MatVecList<F>& eigenforms,
			const MatVecList<F>& anlist) ;

    virtual void BSDRatio(F& bsd, 
			  int m ,const Vector<F>& f,
			  int s=1) {
	die("BSDRatio: not supported, need integral structure.");
    }

    virtual void IntersectionPairing(Matrix<F>& Delta,
				     const Vector<F>& f,
				     const Matrix<F>& an)
	{ die("IntersectionPairing not defined.");}
    virtual void ModularDegree(F& moddeg,
		       const Vector<F>& f,
		       const Matrix<F>& an) 
	{ die("ModularDegree not defined.");}
    virtual void AllModularDegree(Vector<F>& moddeg,
			  const MatVecList<F>& anlist)
	{ die("AllModularDegree not defined.");}

    virtual void OrderByTrace(Vector<int>& perm,
			      const MatVecList<F>& anlist);
    virtual void OrderByTrace(MatVecList<F>& anlist);
    virtual void OrderByTrace(MatVecList<F>& anlist,
			      MatVecList<F>& eigenbasis);

    virtual bigint TorusPointsFp(Matrix<F>& V, int p) {
	die("Not implimented.");
    }
    virtual bigint TorusTorsionBound(Matrix<F>& V, int maxp) {
	die("Not implimented.");
    }
    
    // Compute the discriminant of the Hecke algebra on M_k
    // as a Z-module.
    virtual F heckedisc() {
	die("nonintegral heckedisc not defined yet."); 
    }
    virtual F heckedisc(const Matrix<F>& V) {
	die("nonintegral heckedisc(V) not defined yet.");
    }
    virtual void heckedisc_Sknew(F& discTnew,
				 F& discTfnewprod,
				 F& Vol, F& Idx) {
	die("nonintegral heckedisc_Sknew not defined yet.");
    }
    virtual const Matrix<F>& get_ZBasis() {
	die("Nonintegral get_ZBasis not defined.");
    }
    virtual const Matrix<F>& get_ZBasis_Mk() {
	die("Nonintegral get_ZBasis_Mk not defined.");
    }
    virtual const Matrix<F>& get_ZBasis_Sk() {
	die("Nonintegral get_ZBasis_Sk not defined.");
    }

    virtual F ModularDegree(int n, const Vector<F>& f, bool fullspace=false) {
	die("Nonintegral ModularDegree not defined.");
    }

    virtual void IntPairing(Matrix<F>& Pairing,
			       int n, const Vector<F>& f, bool fullspace=false) {
	die("Nonintegral IntPairing not defined.");
    }

    // Call CanReduce_charmod first. Just do it!
    bool CanReduce_charmod() const ;
    void Reduce_charmod(H1NChi<charmod>& Hmod,charmod z) const ;

    int get_k() const { return k; }
    const Character& get_chi() const { return chi; }
    bool get_plus() const { return plus; }
    bool get_save() const { return save; }

    void DegeneracyReps (Array<bigint>& R,
			 const bigint& N, const bigint& M, const bigint& t) const;

				// Compute a matrix whose columns Z-span
                                // the image of the Z-module map
                                //   (manin symbol) |----> M_k(Gamma,F). 
                                // I.e., every manin symbol is equivalent
                                // to a column of M, up to a unit (or 0).
    void ImageOfManin(Matrix<F>& M) ;

    void Tpr_Mk (int p, int r, Matrix<F>& Tpr) const  ;
    void Tp_StandardSymbol(Vector<F>& TpSymi, Array<int>& H, int i) const ;
    void Tp_StandardSymbol(Vector<F>& TpSymi, int i, int p) const ;

    const char* Name() const { return fname; }
private:
    
    bool plus ;			// If true, do all computations
                                // in the *quotient* H_1^+ = H_1/"I relations".
                                // For the most part, the + part is the
                                // one which has been implimented.
    bool save ;			// If true, save all computations of
				// Hecke operators to the directory "data"
				// and reload them if they are asked for again.
    int k ;			// k is the weight which must be >= 2.
    int N ;			// the level
    Character chi ;		// An arbitrary mod N character
				//        chi:(Z/NZ)^*-->C^*
    MSymbols msymbols ;		// Reduced list of Manin symbols at level N.
				// Span a space, let's call it V.
				// 
				// We're computing S_k in the following diagram:
				// 0 --> S_k --> M_k --> B_k
				//    cusp forms  \--- some mod. forms.
				//               M_k = V/<relations>
				// 
    int dim_m;		        // dim_m = dim(M_k)
    
				// Basis for a subspace of V which maps
    Vector<int> Mk ;		// isomorphically to M_k(Q) under the natural
				// projection V --> M_k(Q). 
    Matrix<F>   Lift ;		// A section of the map V --> M_k with
    Vector<int> Lift_data;
				// with image lieing in the span of FreeGens.
    
    
    Matrix<F>  Sk;		// A basis for S_k.
    Matrix<F>  Sknew;		// A basis for S_k^{new}.
    Matrix<F>  Delta;		// Delta:M_k-->{cusps}, whose kernel is S_k. 
    Matrix<F>  Eis;		// Mk = Eis \oplus Sk as Hecke modules.  

    class HCache {
    public:
	HCache() { H=NULL; }
	~HCache() { delete[] H; }
	Array<int>* H;
    };
    static HCache Heil;		// Lots of Heilbronn matrices.
    
    char* fname ;			// Common filename prefix for saving objects.
    
				// given an index i, for example as returned
				// by the first component of act_on_MSym,
				// return the linear combination of free
				// generators which is equivalent to the
				// i-th standard msymbol.
				// [This replaces the old code:
				//      Lift.row(SI_rel(idx[t]))) 
				// with
				//      msymidx_to_free(SI_rel(idx[t]));
    
    Vector<F> msymidx_to_free (const int& i) const ;
    void init_msymidx_to_free () ;

    // w = vector on the _standard list_ of Manin symbols.
    // v = corresponding element of M_k(Q), wrt free basis.
    void standard_to_free(Vector<F>& v,
			  const Vector<F>& w) const;

    
    // this is the result of modding out by absolutely all relations.
    Vector<F> index_to_free (const int& index) const {
	return SI_c(index)*msymidx_to_free(SI_rel(index));
    }
    
    Matrix<F> BINOM ;  // used by act_on_MSym_q for efficiency reasons.
    void init_BINOM() ;
    
    enum ACTION_TYPE { HECKE, ATKIN_LEHNER, ETA_LOWER } ;
    void act_on_MSym(Array<int>& n, Array<F>& c,
		     const int& i, const int* mat,
		     ACTION_TYPE action=HECKE,
		     int param=0,
		     bigint spec_u=0, bigint spec_v=0, int spec_q=0) const ;


    // Input:  * a monomial P(X,Y) = X^q*Y^(k-2-q)
    //         * a 2x2 matrix m=[ m[0], m[1]; m[2], m[3] ] = [a,b;c,d]
    // Output: * the coefficients of the monomial
    //           P(m(X,Y)) = P(a*X+b*Y, c*X+d*Y)
    //                     = coeff[0]*X^0*Y^(k-2) + ... + coeff[k-2]*X^(k-2).
    void act_on_MSym_q(Array<F>& coeff, const int* m,
		       const int& q) const ;
    
    
    Array<F>   SI_coeff ;
    Array<int> SI_gens ;
    Array<int> SI_rels ;
    
    void init_Mk_SI_table () ;
    bool init_mk_si_table;
    int  SI_gen (const int& n) const { return SI_gens[n] ; }
    int  SI_rel (const int& i) const { return SI_rels[i] ; }
    F    SI_c   (const int& i) const { return SI_coeff[i] ; }
    int  SI_numgens () const { return SI_gens.length()-1 ; }
    
    void init_Mk () ;
    void init_Mk_BuildRelMat (Sparse<F>& RelMat);
    void init_Mk_add (Vector<F>& row,
		      const int& n, const int* matrix) ;
    void init_Mk_saverow (Vector<int>& r, Vector<int>& c,
			  Vector<F>& val, int& i,
			  int& rownum,
			  const Vector<F>& row_val) const ;
    
    //  void init_Mk_add(Matrix<F>&,const int&, const int&, const int*);  
    // init_Mk_add is a helper function.
    bool init_Mk_rownew (Matrix<F>&, const int& row) ;  // helper function. 
    
    void init_fname () ;
    void init_mesg() const;
    
    void init_Delta () ;   
    bool init_delta;
    void init_Sk () ;  
    bool init_sk ;
    void init_Eis () ;  
    bool init_eis ;
    
    bool init_sknew ;
    
    int  load_Mk() ;
    void save_Mk() const ;
    int  load_Sk() ;
    void save_Sk() const ;
    int  load_Sknew() ;
    void save_Sknew() const ;
    int  load_Tn_Mk (int p, Matrix<F>& M) const ;
    void save_Tn_Mk (int p, const Matrix<F>& M) const ;   
    int  load_WN_Mk (int M, Matrix<F>& W) const ;
    void save_WN_Mk (int M, const Matrix<F>& W) const ;   
				// compute the Heilbronn matrices of det=p.

    // Perhaps this is the format all of the operators should take.
    // init_Wq is called before any call to get_Wq.  It looks in
    // the disk database for WqList.  If found, it loads it.  Otherwise
    // it sets WqList to contains the right number of 0x0 matrices.
    // term_Wq is called when *this is being destroyed.  It saves
    // WqList to the database. 
    MatVecList<F> WqList;
    void          init_Wq();
    void          term_Wq();
    
    void init_Heilbronn() ;
    void Heilbronn (int p, Array<int>& H) const ;
    void Tp_Heilbronn (int p, Array<int>& H) const ;
    
    enum etype {TERMINAL, NONTERMINAL} ;
    void error (const char* msg, etype e=TERMINAL) const ;  // error message. 
    void assert(bool assertion, const char* func_name="?",
		const char* mesg="assertion failed") const;
    

    // return position of i-th free generator in full msymbols list. 
    int IthStandardSymbol(int i) const {
	return SI_gen(Mk[i]);
    }

    void Tn_SparseVector (Vector<F>& v, // v = T_n * x
			  const Vector<F>& x,
			  const Vector<int>& nonzero,
			  int n) const;
    
    void start_timing(char* msg, int param=-1) const;
    void status_message(char* msg, int param=-1) const;
    void stop_timing()const ;
    
    Vector<F> Eta_Raise_Single (const MSym& x, 
				const Array<bigint>& Gam,
				const Array<int>& Heil,
				const int& t, const int& M) const ;
    Vector<F> Eta_Lower_Single (const MSym& x, 
				const Array<int>& Heil,
				const int& t, const int& M) const ;
} ;


// MSymVec: Vector of Manin symbols.

template <class F>
class MSymVec {
 public:
    void init(int i_size, int newN) {
	m_X.reset(i_size+1);
	m_C.init(i_size); N(newN); 
    }
    void N(int newN) { m_N=newN; }
    int N() const { return m_N ; }
    int size() const { return m_C.size() ; }
    void set(const int& i, const MSym& x, const F& c)
	{ verify(i); m_X[i]=x; m_C[i]=c; }
    MSym x(int i) const { verify(i); return m_X[i] ; }
    F c(int i) const { verify(i); return m_C[i] ; }
    friend ostream& operator<<(ostream& os, const MSymVec<F>&);
    void ModSymPrint(ostream& os, int k) const;
 private:
    void verify(int i) const { if(i<1||i>size()) die("MSymVec: i out of bounds."); }
    Array<MSym>  m_X ;
    Vector<F>    m_C ; 
    int          m_N ;
};


#endif 






