/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/* 
   FILE: h1nchiz.cc
   DESC: Modular symbols of weight k>=2 over the field of rationals
         with extensions for integral structure.
*/

#include "h1nchiz.h"
#include <fstream.h>
#include "NTL/ZZ.h"
#include "system.h"
#include "fields.h"
#include "h1nchi.h"
#include "sparse.h"
#include "sparsematrix.h"
#include "cusps.h"
#include "arith.h"
#include "modformulas.h"
#include <LiDIA/nmbrthry_functions.h>

void 
H1NChiZ::CyclicHeckeModule(Matrix<bigrational>& Tx,
			   const Vector<bigrational>& x)
{
    Matrix<bigrational> Gen;
    H1NChi<bigrational>::CyclicHeckeModule(Gen,x);
    ::ZBasis(Tx,Gen);
}

void 
H1NChiZ::CyclicHeckeImage(Matrix<bigrational>& PTx,
			  const Matrix<bigrational>& P,
			  const Vector<bigrational>& x,
			  int bound)
{
    Matrix<bigrational> Gen;
    H1NChi<bigrational>::CyclicHeckeImage(Gen,P,x,bound);
    ::ZBasis(PTx,Gen);
}

void 
H1NChiZ::init_ZBasis_Mk()
{
    mesgln("[init_ZBasis_Mk]");
    // Compute Z-module inside M_k(Q) generated by the standard 
    // Manin symbols
    //         [X^i*Y^(k-2-i), (c,d)]

    Matrix<bigrational> M;
    ImageOfManin(M);
    bigint d;
    Matrix_denominator(d,M);
    if(d!=1) {
	mesg("\ninit_ZBasis_Mk: index is >=",to_int(d));
	mesg(" so we reduce ",M.nrows());
	mesg(" x ",M.ncols());
	mesgln(" matrix.");
	::ZBasis(ZBasis_Mk, M);
    } else 
	ZBasis_Mk.idmat(dim_Mk());
    init_zbasis_mk = true;
}


void
H1NChiZ::init_ZBasis_Sk()
{
  mesgln("[init_ZBasis_Sk]");

  if(load_ZBasis_Sk()) {
      init_zbasis_sk=true;
      return;
  }

  if(!init_zbasis_mk)
    init_ZBasis_Mk();

  // compute ZBasis_Sk = matkerint(delta).

  init_Delta();
  // Now we have at our disposal a matrix Delta
  // which gives the delta map on the Mk space.
  // We also have the ZBasis_Mk of integral symbols.
  // So we compute the map defined by delta on our ZBasis_Mk,
  // by computing the image of each basis element by Delta.
  // Then we compute the integer kernel.
  Matrix<bigrational> D;	// represents Delta wrt ZBasis_Mk.
  //mesgln("Delta is:");
  // cout << "Delta = " << Delta << endl;
  D = Delta* ZBasis_Mk;
  // cout << "D = " << D << endl;
  
  mesgln("[Compute integer ker(M_k(Z) ---> C_k(Z))]");
  Matrix<bigrational> K;
  IntegerKernel(K,D);

  // finally ZBasis_Sk is the linear combination of columns
  // of ZBasis_Mk given by K.
  ZBasis_Sk = ZBasis_Mk * K;

  init_zbasis_sk = true ;
  save_ZBasis_Sk();
}


bool
H1NChiZ::load_Te(Matrix<bigrational>& Te, int s) 
{
    if (!save) return false;
    char buf[1024] ; sprintf (buf, "%sTe%d", fname,s);
    ifstream in(data_read(buf,N)) ;
    if (!in) return false;
    Te.Load(in);
    return true; 
}

void 
H1NChiZ::save_Te(const Matrix<bigrational>& Te, int s) const
{
    if (!save) return ;
    char buf[1024] ; sprintf (buf, "%sTe%d", fname,s);
    ofstream of(data_write(buf,N));
    if (!of) {
	error("Error trying to save Te.", NONTERMINAL) ;
	return ;
    }
    Te.Save(of);
}

bool
H1NChiZ::load_ZBasis_Sk() 
{
    if (!save) return false;
    char s[1024] ; sprintf (s, "%sZBasisSk", fname);
    ifstream in(data_read(s,N)) ;
    if (!in) return false;
    ZBasis_Sk.Load(in);
    return true; 
}

void 
H1NChiZ::save_ZBasis_Sk() const
{
    if (!save) return ;
    char s[1024] ; sprintf (s, "%sZBasisSk", fname) ;
    ofstream of(data_write(s,N));
    if (!of) {
	error("Error trying to save Z-basis.", NONTERMINAL) ;
	return ;
    }
    ZBasis_Sk.Save(of);
}

const Matrix<bigrational>& 
H1NChiZ::get_ZBasis_Mk() 
{
    if(!init_zbasis_mk)
	init_ZBasis_Mk();
    return ZBasis_Mk;
}

const Matrix<bigrational>& 
H1NChiZ::get_ZBasis_Sk() 
{
    if(!init_zbasis_sk)
	init_ZBasis_Sk();
    return ZBasis_Sk;
}

void
H1NChiZ::BSD(bigrational& bsd,
	     const Vector<bigrational>& f,
	     const Matrix<bigrational>& v,
	     const bigrational& discZf,
	     const int i=0) 
{
    // Let e = -{0,\infty}.  Then L(v,1)=<e,v>.
    // If <e,v> = 0, the BSD value is 0 and we are done
    // so assume <e,v> =/= 0.
    // 
    // This function computes
    //         Norm(<e,v>)
    //     ------------------ * [Z[f]:Z[alp]]
    //           Vol(L)
    // where L is the lattice generated by the row vectors
    //       (<c,v_1>,<c,v_2>,...,<c,v_n>)
    // Here v = v_1 + x*v_2 + x^2*v_3 + ... + x^{n-1}*v_n, x a root of f(x),
    // and c varies through enough modular symbols to generate H_1(X_0(N),Z).

    const Matrix<bigrational> Z = get_ZBasis_Sk();
    
    //  cout << "Entering BSD"<<endl;
    //  cout << "f="<<f<<endl;
    mesg("[Nev");
    Vector<bigrational> e;  WindingElement(e,i);

    //  cout << "e="<<e<<endl;
    Vector<bigrational> ev = e*v;
    if(ev.iszero()) {bsd=0; return;}
    bigrational Nev = abs(AlgNumNorm(ev,f));
    mesg("]");

    //  cout << "Nev="<<Nev<<endl;
    int max_sym = msymbols.length() ;  
    Matrix<bigrational> L; 
    L.init(f.poldeg(),Z.ncols());
    for (int i=1; i<=Z.ncols(); i++) 
	L.setcol(i,Z.col(i)*v);

    mesg("[Omega(f)  ");
    bigrational V = LatticeVol(L);
    mesg("]");
    
    //  cout << "Vol(L)="<<V<<endl;
    //  cout << "disc(f) = "<<d<<endl;
    
    bigrational d = abs(PolDisc(f)/discZf);
    if(denominator(d)!=1)
	die("BSD -- error in discriminant, index not an integer!");

/************** IDEA ***************/
/* It might be possible to compute d without computing discZf or
   PolDisc(f) by noting that that index is (probably!) the constant term of  
   the eigenform normalized minimally so it's components are expressed as 
   integer linear combinations of a root of f. */ 
    
    
    bigint D;
    sqrt(D,numerator(d));
    if (D*D != numerator(d)) {
	cout << "discZf="<<discZf<<endl;
	cout << "f="<<f<<endl;
	cout << "e="<<e<<endl;
	cout << "Norm(e mod f) = " << Nev << endl;
	cout << "L = " << L << endl;
	cout << "Vol(L)="<<V<<endl;
	cout << "disc(f) = "<<d<<endl;
	die("BSD -- error in discriminant, not a square!");
    }

    
    bsd=Nev/(V*D);

    // CURRENTLY we are working in the +1 quotient so that power of 2 is wrong
    // in all cases it _seems_ that adjusting by 2^dimension gives a consistent
    // answer -- so we'll code it here.  Note: this is only a guess, really
    // the power of 2 should be IGNORED.

      bigint x;
      power(x,2,f.poldeg());
      bsd/=x;
}

void
H1NChiZ::AllBSD(Matrix<bigrational>& bsd,
		Vector<bigrational>& disc,
		const MatVecList<bigrational>& eigenforms,
		const MatVecList<bigrational>& anlist) 
{
    disc.init(eigenforms.len());
    bsd.init(eigenforms.len(),k-1);
    for(int i=0;i<eigenforms.len();i++) {
	mesg("[Disc(O_f) ");
	if(eigenforms.cv(i)!=anlist.cv(i)) {
	    mesgln("Error in ALLBsd, eigenforms and anlist out of sync.");
	}
//	cout << "anlist.cv(i)="<<anlist.cv(i)<<endl;
//	cout << "anlist.cM(i)="<<anlist.cM(i)<<endl;
	disc[i+1]=ZAlgDisc(anlist.cv(i),anlist.cM(i));
        mesg("disc="); cout<<disc[i+1];
	mesg("]\n");
	for(int s=1;s<=k-1;s++) 
	    BSD(bsd(i+1,s),eigenforms.cv(i),
		eigenforms.cM(i),disc[i+1],
		s-1);
    }
}

/*
ALGORITHM idea (before writing code):
1. Using basis of eigenforms, know the system of Hecke action on an invariant subspace.
2. Compute H_1(X_0(N),Z)   (integral basis, so could be the "* - 1" quotient.)
3. Compute T_2 on H_1(Z) and cut out subspace (matkerint)
4. Repeat 3 until get down to correct dimension.      call V
5. Do the same for the dual (i.e., left eigenvectors)  call it W
6. compute the sublattice of the span of the rows of V generated by
   dotting with W.  This is probably something like the determinant
   of W*V.
7. Output the modular degree.
8. Maybe the structure of ker(delta) would be nice?? To compute that,
   I think I must compute the smith normal form of W*V. 
*/


// use f and an to extract the charpoly of T_2, then T_3, etc. on the eigenform.
// cut down space to get kernel.
void
H1NChiZ::IntersectionPairing(Matrix<bigrational>& Delta,
			     const Vector<bigrational>& f,
			     const Matrix<bigrational>& an) 
{
    Matrix<bigrational> intZ, intZdual;
    intZ.idmat(get_ZBasis().ncols());
    intZdual = intZ;

    int p=2;
//    cout << "Z = " << intZ<<endl;
//    cout << "Zdual = " << intZdual<<endl;
    while(intZ.ncols() > f.poldeg()) {
	if(p>an.nrows()) {
	    mesg("H1NChiZ::IntersectionPairing -- ERROR!  Not able to cut down to small enough subspace.!");
	    Delta.init(1,1);
	    Delta(1,1)=0;	// invalid value -- indicates an error.
	    return;
	}
	// 1. Compute charpoly(T_p) on subspace of S_2(Gamma_0(N)) correspondong to f. 
	Vector<bigrational> g;
	CharPolyMod(g, an.row(p), f);
//	cout << "Got charpoly = " ; g.polprint(); cout<<endl;
	// Self Check: g must have integer coefficients.
	for(int i=1; i<=g.size(); i++)
	    if(denominator(g[i])!=1)
		die("H1NChiZ::ModularDegree -- ERROR, charpoly of an eigenvalue is integral!");
	
	// 2. Compute matrix A of T_p on intZ.
	Matrix<bigrational> Tp;
	Tn_On(p, get_ZBasis(), Tp);
//	cout << "T_"<<p<<" on H^1(Z) = " << Tp << endl;
	Matrix<bigrational> A;
	Tp.RestrictedTo(A, intZ);
//	cout << "On subspace intZ = " << A << endl;

	Matrix<bigrational> gA;
	A.polsubst(gA, g); 
//	cout << "g(A) = " << gA << endl;
	
	// if g(A)==0, skip it.
	if(gA.iszero()) { p = next_prime(p); continue; }

	bigint_matrix lid;
	bigint d;
	Matrix_bigintify(lid, d, gA);
	// [check] all entries of gA must be integers, or there is an error.
	if(d!=1)
	    die("H1NChiZ::ModularDegree -- T_p didn't act integrally on the integral modular symbols!");

	// 3a. compute integral kernel of g(A).
	bigint_matrix kerlid;
	kerlid.kernel(lid);
	Matrix<bigrational> K;
	to_bigrationalmat(K, kerlid);

	// 3b. replace intZ by intZ*ker(chi(A)) subset intZ. 
	intZ = intZ*K;

	// 4a. compute integral kernel of g(A^t)
	Tp.transpose();
	Tp.RestrictedTo(A, intZdual);
	A.polsubst(gA, g); 
	Matrix_bigintify(lid, d, gA);
	// [check] all entries of gA must be integers, or there is an error.
	if(d!=1)
	    die("H1NChiZ::ModularDegree -- T_p didn't act integrally on the integral modular symbols!");

	kerlid.kernel(lid);
	to_bigrationalmat(K, kerlid);
	// 4b. replace Zdual by Zdual*ker(chi(A)) subset Z. 
	intZdual=intZdual*K;


//	cout << "Z = " << intZ<<endl;
//	cout << "Zdual = " << intZdual<<endl;
	p = next_prime(p);
    }
    if(intZ.ncols()!=intZdual.ncols()) 
	die("H1NChiZ::ModularDegree -- Z and Zdual incompatible!");

    // It is now necessary to express each of Z and Zdual in terms of ZBasis, instead of
    // in terms of the standard basis of Manin symbols. 
    
    intZdual.transpose();
    Delta = intZdual*intZ;
    return;
}

bigrational
H1NChiZ::ModularDegree(int n, const Vector<bigrational>& f,
		       bool fullspace)
{
    Matrix<bigrational> pairing;
    IntPairing(pairing, n, f, fullspace);
    return abs(pairing.determinant());
}

void 
H1NChiZ::IntPairing(Matrix<bigrational>& Pairing,
		int n, const Vector<bigrational>& f,
		bool fullspace=false) 
{
    Matrix<bigrational> Tn;
    if(fullspace) 
	Tn_On(n,get_ZBasis_Mk(),Tn);
    else {
	if (k>2) 
	    Tn_On(n,get_ZBasis_Sk(),Tn);
	else 
    	    Tn_On(n,get_ZBasis(),Tn);
    }
    Matrix<bigrational> T;
    Tn.polsubst(T, f);

    Matrix<bigrational> Left, Right;
    IntegerKernel(Right, T);
    T.transpose();
    IntegerKernel(Left, T);
    Left.transpose();
    Pairing=Left*Right;
}


void
H1NChiZ::ModularDegree(bigrational& moddeg,
		       const Vector<bigrational>& f,
		       const Matrix<bigrational>& an) 
{
    Matrix<bigrational> Delta;
    IntersectionPairing(Delta,f,an);
    moddeg=abs(Delta.determinant());
}

void
H1NChiZ::AllModularDegree(Vector<bigrational>& moddeg,
			  const MatVecList<bigrational>& anlist)
{
    moddeg.init(anlist.len());  
    for(int i=0;i<anlist.len();i++) 
	ModularDegree(moddeg[i+1], anlist.cv(i), anlist.cM(i));
    if(plus) {
	mesg("Warning: working in plus space, so modular degrees returned\n");
	mesg("are sqrt(odd_part(deg(delta_f)))*('meaningless' 2-power)!");
    }
}



bigrational 
H1NChiZ::heckedisc() 
{
    ModularFormulas formulas;
    int n = k*formulas.idxG0(N)/12;	  // Sturm's bound.
    Matrix<bigrational>* T=TnList_Mk(n);
    Vector<bigrational> v; bool bad;
    bigrational disc=ZModDisc(v, bad, T, n);
    if(bad) die("heckedisc: bad rank!!!");
    delete[] T;
    return disc;
}

bigrational 
H1NChiZ::heckedisc(const Matrix<bigrational>& V)
{
    ModularFormulas formulas;
    int n = k*formulas.idxG0(N)/12;	  // Sturm's bound.
    Matrix<bigrational>* T=TnList_On(n,V);
    Vector<bigrational> v;
    bool bad;
    bigrational disc=ZModDisc(v, bad, T, n);
    if(bad) die("heckedisc: bad rank!!!");
    delete[] T;
    return disc;
}

bigrational 
H1NChiZ::heckedisc_T1_Tg(bool& dontgenerate)
{
    int g = dim_Sk();
    Matrix<bigrational> T[g];
    for(int n=1; n<=g; n++) 
	Tn_Sk(n,T[n-1]);
//	Tn_On(n,get_Sknew(),T[n-1]);
    Vector<bigrational> v;
    return ZModDisc(v,dontgenerate,T,g);
}


bigrational
H1NChiZ::heckedisc_normalization_Sknew()
{
    MatVecList<bigrational> E;
    DualBasisOfEigenforms(E,get_Sknew());
    bigrational disc=1;
    for(int i=0;i<E.size();i++)
	if(E.v(i).poldeg() > 1) 
	    disc*=FieldDisc(E.v(i));
    return disc;
}

void H1NChiZ::heckedisc_Sknew(bigrational& discTnew,
			      bigrational& discTfnewprod,
			      bigrational& Vol,
			      bigrational& Idx)
{
    /* ALGORITHM:  (Designed, Jan 31, 1999, W. Stein.)
       
       Let f1,...,fn be a basis of newforms for S_2(Gamma_0(N)), up to conjugacy.
       This defines an injection
            Phi: Tnew \-------> C^d
            Tp|----------------> (an(f1),an(sigma(f1)),...,an(fn),...)
       As a Z-module Tnew is generated by the image of Phi, and the multiplication
       is coordinate wise.  So disc(Tnew) is related to Vol(Im(Phi)).
       Write f1=sum hi a^i, for some a in O=Z[a1,a2,a3,...], for example as
       given by the eigenbasis type functions above. 
       Then Vol((an(f1),...,an(sigma(f1))) for all p)
                    = Vol((an(h1),...,an(hd) for all p) * [O:Z[a]]. 
       Proof. Proceed exactly as in the computation of L(A_f,1)/Omega(A_f).
       Choosing h's for all fi at once and making the same sort of change 
       of basis as arises in the proof, to the whole system of f's gives
          |disc(Tnew)| 
                 = (Vol (------ hi --------)
                     *prod [O_fi : Z[a_i]])^2 * prod(disc(O_fi))
       Since Tnew sits inside the product of totally real number fields,
       we know the sign is +1. 
     */

    // [1. Compute basis of eigenforms.]
    MatVecList<bigrational> eigenbasis, anlist;
    ModularFormulas m;
    int max=int(float(k)*float(m.idxG0(N))/float(12))+1;
    if(!HeckeEigenvalues_an(anlist,eigenbasis,get_Sknew(),max))
	die("heckedisc_Sknew: Could not compute basis of eigenforms!");

    // [2. Construct "h matrix".]
    Matrix<bigrational> H(dim_Sknew(),max);
    for(int n=1;n<=max;n++) {
	int c=1;
	for(int j=0;j<eigenbasis.len();j++) {
	    Vector<bigrational> h=anlist.cM(j).row(n);
	    for(int k=1;k<=h.size();k++) 
		H(c++,n)=h[k];
	}
    }

    // [3. Compute basis for Z-span of columns of H.]
    Vol = LatticeVol(H);

    // [4. Compute each [\O:Z[a_i]] and disc(\O).]
    bigrational D=1; 
    Idx=1; // D = \prod_O disc(O), Idx = \prod[O:Z[a]].
    
    for(int i=0;i<eigenbasis.len();i++) {
	bigrational d = ZAlgDisc(anlist.cv(i),anlist.cM(i));
	D*=d;
	bigrational t = PolDisc(eigenbasis.cv(i))/d;
	bigint idx; sqrt(idx,numerator(t));
	if (idx*idx != t)
	    die("heckedisc_Sknew -- error in discriminant, not a square!");
	Idx*=bigrational(idx);
    }
    discTnew = Vol*Vol*Idx*Idx*D;
    discTfnewprod = D;
}

bool LessThan(int i, int j,
	      Matrix<bigrational>& traces,
	      Vector<bigrational>& numcomputed, 
	      const MatVecList<bigrational>& anlist)
{
    int c=1;
    while (c<=anlist.cM(0).nrows()) {
	if (c>numcomputed[i]) {	// in fact c must then equal numcomputed(i)+1.
	    traces(i,c)=TracePolyMod(anlist.cM(i-1).row(c), anlist.cv(i-1));
	    ++(numcomputed[i]);
	}
	if (c>numcomputed[j]) {
	    traces(j,c)=TracePolyMod(anlist.cM(j-1).row(c), anlist.cv(j-1));
	    ++(numcomputed[j]);
	}
	if(traces(i,c)<traces(j,c)) return true;
	if(traces(i,c)>traces(j,c)) return false;
	c++;
    }
    die("H1NChiZ::OrderByTrace -- insufficient precision to separate eigenforms!");
}
	      
void
H1NChiZ::OrderByTrace(MatVecList<bigrational>& anlist)
{
    Vector<int> perm;
    OrderByTrace(perm, anlist);
    anlist.reorder(perm);
}

void
H1NChiZ::OrderByTrace(MatVecList<bigrational>& anlist,
		      MatVecList<bigrational>& eigenbasis)
{
    Vector<int> perm;
    OrderByTrace(perm, anlist);
    anlist.reorder(perm);
    eigenbasis.reorder(perm);
}

void
H1NChiZ::OrderByTrace(Vector<int>& perm,
		      const MatVecList<bigrational>& anlist)
{
    mesg("[OrderByTrace");
    // Because the number of items to sort is VERY small, and
    // we want to finish writing this in a few minutes, I'm going
    // to impliment a simple buble sort.  If this turns out to
    // be slow, change to quicksort later.

    // Newforms are labeled f_1, ... f_n.
    int n=anlist.len();
    // NOTE: anlist.cM(...) starts at 0 and goes to n-1.

    Matrix<bigrational> traces;   // trace(i,j) will give Tr(a_n(f_i)), once computed
    traces.init(n,  anlist.cM(0).nrows());
    Vector<bigrational> numcomputed;   // numcomputed(i) = max k so that Tr(a_n(f_i)) has
                                       // been computed for all n<=k.
    numcomputed.init(n);
    
    perm.init(n);
    for(int i=1; i<=n; i++)	// start with identity permutation.
	perm[i]=i;
    
    // Now we must reorder the array perm so that LessThan(perm[i],perm[i+1]) 
    // is true for all 1<=i<=n.

    for(int i=1; i<=n; i++)
	for(int j=i+1; j<=n; j++) 
	    if( ! LessThan(perm[i],perm[j],traces,numcomputed,anlist) ) {
		int t=perm[i];
		perm[i]=perm[j];
		perm[j]=t;
	    }
    mesg("]\n");

    for(int i=1; i<=n-1; i++)
	if (!LessThan(perm[i],perm[i+1],traces,numcomputed,anlist))
	    die("H1NChiZ::OrderByTrace:  Error in sorting algorithm!");
//    cout << "numcomputed = " << numcomputed << endl;
//    cout << "trace = " << traces << endl;
}


bigint
H1NChiZ::TorusPointsFp(Matrix<bigrational>& V, int p)
{
    if(N%p==0)
	mesg("TorusPointsFp not so sensible when p|N.\n");
    Matrix<bigrational> T;
    Tn_On(p, V, T);
    Vector<bigrational> f;
    T.charpoly(f);
    bigrational a=f.polsubst(p+1);
    if(a.denominator()!=1)
	die("Error in TorusPointsFp.");
    return a.numerator();
}

bigint
H1NChiZ::TorusTorsionBound(Matrix<bigrational>& V, int maxp)
{
    
    bigint a=0;
    for(int p=3; p<=maxp; p=next_prime(p)) 
	if(N%p!=0) {
	    if(a==0)
		a = TorusPointsFp(V,p);
	    else
		a=gcd(a,TorusPointsFp(V,p)); 
	    cout<<a<<"  "; 
	}
    mesg("\n");
    return a;
}

void
H1NChiZ::BSDRatio(bigrational& bsd, 
		  int m, const Vector<bigrational>& f,
		  int s)
{
    mesgln("BSDRatio");
    if(chi.type()!=0) 
	mesgln("WARNING: BSDRatio not guaranteed if Gamma=/=Gamma_0(N).");
    
    //   (1) P <--- left kernel of Tn(f) inside M_k(Gamma) dual,
    //       (the j rows of P are a basis for the kernel.)
    Matrix<bigrational> P;
    Matrix<bigrational> Tm;
    Tn_Mk(Tm,m);
    Tm.transpose();
    Kernel(P, Tm, f);
    P.transpose();

    int n=P.nrows();

    //   (2) Z <--- integral basis for P*S_k(Gamma,Z).
    const Matrix<bigrational>& Zbas=get_ZBasis();
    Matrix<bigrational> Z;
    ::ZBasis(Z,P*Zbas);


/*
    //   (3) Te <-- Z-basis for Z-submodule T*e subset M_k.
    Matrix<bigrational> Te;
    if(!load_Te(Te,s)) {
	Vector<bigrational> e;
	WindingElement(e,s-1);
	CyclicHeckeModule(Te,e);
	save_Te(Te,s);
    }

    //   (4) PTe<-- Z-basis for P*Te. 
    Matrix<bigrational> PTe;
    ::ZBasis(PTe,P*Te);
*/

    // PTe
    Vector<bigrational> e;
    WindingElement(e,s-1);
    Matrix<bigrational> PTe;
    CyclicHeckeImage(PTe, P, e);

    //   (5) Compute complex conjugation C on Z:
    //       (a) choose n independent columns i_1,...,i_n of P.
    Matrix<bigrational> PP=P;
    Vector<int> elimcol, elimrow;
    PP.SimpleEchelon(elimcol, elimrow);
    Vector<int> ind(n);
    int i=1;
    for(int c=1;c<=PP.ncols();c++)
	if(elimcol[c]!=0)
	    ind[i++]=c;
    
    //       (b) B <-- matrix consisting of n linearly independent 
    //           columns i_1,...,i_n
    Matrix<bigrational> B(n,n);
    for(int i=1;i<=n;i++)
	B.setcol(i,P.col(ind[i]));

    //       (c) A <-- * involution on M_k
    Matrix<bigrational> A;
    Conjugation_Mk(A);

    //       (d) A' <-- matrix whose columnes are i_1,...,i_n of A.
    Matrix<bigrational> AA(dim_Mk(),n);
    for(int i=1;i<=n;i++)
	AA.setcol(i,A.col(ind[i]));

    //       (e) C <-- Z^(-1)*P*A'*B^(-1)*Z
    Matrix<bigrational> Zinv,Binv;
    Z.Inverse(Zinv);
    B.Inverse(Binv);
    Matrix<bigrational> C = Zinv*P*AA*Binv*Z;

    //   (6) Zp <-- Z*kerint(C-1)  (plus one eigenspace for conjugation
    //                              on period lattice)
    bigrational val;
    if(s%2==0)
	val=-1;
    else
	val=+1;
    for(int i=1;i<=n;i++)
	C(i,i)-=val;
    Matrix<bigrational> KerC;
    IntegerKernel(KerC, C);
    Matrix<bigrational> Zp=Z*KerC;

    //   (7) cinf <-- 2^(dim(ker(Cmod2-1))-d),    (# connected components)
    //       where d=n/2 if plus=false, or d=n if plus=true.
    int dim=RankModP(C,2);
    int d=n/2;
    if(plus) d=n;
    bigint cinf;
    power(cinf,2,dim-d);
    cout << "c_infinity = " << cinf<<endl;

    //   Finally,
    //    BSDRatio = [Zp : PTe] * cinf.

    bsd = ModuleIndex(Zp, PTe) / cinf;
}

/*
void
H1NChiZ::BSDRatio(bigrational& bsd,
		  int s, Motive& M);
{
    mesgln("BSDRatio(Motive)");

    if(chi.type()!=0) 
	mesgln("WARNING: BSDRatio not guaranteed if Gamma=/=Gamma_0(N).");
    
    //   (1) P <--- left kernel of Tn(f) inside M_k(Gamma) dual,
    //       (the j rows of P are a basis for the kernel.)
    const Matrix<bigrational>& P = M.get_VD();
    int n=P.nrows();

    //   (2) Z <--- integral basis for P*S_k(Gamma,Z).
    const Matrix<bigrational>& Z=M.get_PhiSkZ();

    //   (3) & (4) 
    Vector<bigrational> e;
    WindingElement(e,s-1);
    Matrix<bigrational> PTe;
    M.PhiTx(PTe, e);

    //   (5) Compute complex conjugation C on Z:
    const Matrix<bigrational>& C=M.Conjugation_PhiSkZ();
    
    //   (6) Zp <-- Z*kerint(C-1)  (plus one eigenspace for conjugation
    //                              on period lattice)
    const Matrix<bigrational>& Zp=M.PhiSkZPlus();

    //   Finally,
    //    BSDRatio = [Zp : PTe] 
    bsd = ModuleIndex(Zp, PTe);
}
*/












