/* ------------------------------------------------------------------
   h1nchiz.h
   Integral modular symbols of weight k>=2.
   William Stein
   Last Modified: April, 1999
   ------------------------------------------------------------------ */

#ifndef H1NCHIZ_H
#define H1NCHIZ_H

#include "h1nchi.h"

// Modular symbols, but with additional integral structure. 
class H1NChiZ : public H1NChi<bigrational>  
{
public:
    H1NChiZ(const Character& chi, int k = 2,	
	    bool save = true, bool plus=true)
	: H1NChi<bigrational>(chi,k,save,plus),
      init_zbasis_mk(false), init_zbasis_sk(false)
	{ }
    ~H1NChiZ() {}			// virtual H1NChi has a virtual destructor.

    const Matrix<bigrational>& get_ZBasis() {return get_ZBasis_Sk();}
    const Matrix<bigrational>& get_ZBasis_Mk();
    const Matrix<bigrational>& get_ZBasis_Sk();

    
    // Let v be an eigenform expressed on a power basis in terms of a
    // root of f.  The function BSD returns the rational number
    //             Prod_i L(v_i,1)
    //            ----------------
    //              Omega(A_v)
    // where the product is taken over the Galois conjugates 
    // v_1,...,v_n of v and
    //    L(v_i,1)   = special value at s=1 of L-function attached to v_i.
    //    Omega(A_v) = the volume of the lattice generated by the n tuples
    //                 (<c,v_1>,...,<c,v_n>) as c varies over H_1^+(X_0(N);Z)
    //                 (or the whole H_1(X_0(N);Z), depending on whether the
    //                 member variable "plus" is set to true or not. 
    // 
    // IMPORTANT: v is assumed to be a *left* eigenvector for all
    // of the Hecke operators.  This is not the same as the transpose
    // of a right eigenvector.
    
    void BSD(bigrational& bsd,
	     const Vector<bigrational>& f,
	     const Matrix<bigrational>& v,
	     const bigrational& discZf,
	     int i=0) ;
    
    void AllBSD(Matrix<bigrational>& bsd,
		Vector<bigrational>& disc,
		const MatVecList<bigrational>& eigenforms,
		const MatVecList<bigrational>& anlist) ;


    // f = min poly of generator of Q(alpha)
    // an = list of eigenvalues expressed in terms of alpha.
    void IntersectionPairing(Matrix<bigrational>& Delta,
			     const Vector<bigrational>& f,
			     const Matrix<bigrational>& an) ;
    void ModularDegree(bigrational& moddeg,
		       const Vector<bigrational>& f,
		       const Matrix<bigrational>& an) ;

    void AllModularDegree(Vector<bigrational>& moddeg,
			  const MatVecList<bigrational>& anlist);


// ***************************************
// AWESOME CODE below:

    // compute the modular degree of S_k[f(T_n)], or if
    // fullspace=true, M_k[f(T_n)]. 
    bigrational ModularDegree(int n,
			      const Vector<bigrational>& f,
			      bool fullspace=false);

    // BSDRatio: compute the exact BSD ratio L(s)/Omega, where
    // Omega is real volume wrt integral basis of cusp forms.
    // The power of 2 is only guaranteed to be correct if plus=false.
    // Algorithm:
    //   (1) P <--- left kernel of Tn(f) inside M_k(Gamma) dual,
    //       (the n rows of P are a basis for the kernel.)
    //   (2) Z <--- integral basis for P*S_k(Gamma,Z).
    //   (3) Te <-- Z-basis for Z-submodule T*e subset M_k.
    //   (4) PTe<-- Z-basis for P*Te. 
    //   (5) Compute complex conjugation C on Z:
    //       (a) choose n independent columns i_1,...,i_n of P.
    //       (b) B <-- matrix consisting of n linearly independent 
    //           columns i_1,...,i_n
    //       (c) A <-- * involution on M_k
    //       (d) A' <-- matrix whose columnes are i_1,...,i_n of A.
    //       (e) C <-- Z^(-1)*P*A'*B^(-1)*Z
    //   (6) Zp <-- Z*kerint(C-1)  (plus one eigenspace for conjugation
    //                              on period lattice)
    //   (7) cinf <-- 2^(dim(ker(Cmod2-1))-d),    (# connected components)
    //       where d=n/2 if plus=false, or d=n if plus=true.
    //   Finally,
    //    BSDRatio = [Zp : PTe] / cinf.

    void BSDRatio(bigrational& bsd, 
		  int n, const Vector<bigrational>& f,
		  int s);

    void IntPairing(Matrix<bigrational>& Pairing,
		    int n, const Vector<bigrational>& f,
		    bool fullspace=false);


    // Compute the cyclic Hecke Z-module 
    //     Tx = {T*x : T in Hecke algebra } 
    // generated by the modular symbol x in M_k(Gamma,Q).
    // x is represented in terms of the free basis.
    // Algorithm: (1) find the nonzero entries of x, so if
    //                x is sparse the algorithm is MUCH faster.
    //            (2) compute the Sturm bound n so that
    //                T_1,...,T_n generate the Hecke algebra.
    //            (3) compute M = [T_1*x | ... |T_n*x]
    //            (4) Let Tx = [ Z-basis for column space of M ].
    void CyclicHeckeModule(Matrix<bigrational>& Tx,
			   const Vector<bigrational>& x);
    void CyclicHeckeImage(Matrix<bigrational>& PTx,
			  const Matrix<bigrational>& P,
			  const Vector<bigrational>& x,
			  int bound=0);

    bigrational heckedisc();	// on M_k 
    bigrational heckedisc(const Matrix<bigrational>& V);
    // if Z{T1,...,Tg} has finite index as a Z-submodule of Hecke acting on cusp forms,
    // returns the discriminant of this submodule and sets dontgenerate=false.  
    // Otherwise sets the dontgenerate flag to true.
    bigrational heckedisc_T1_Tg(bool& dontgenerate);
    // returns disc(Q(f1))x...xdisc(Q(fn)), f1,...,fn a basis of normalized newforms.
    bigrational heckedisc_normalization_Sknew();
    void heckedisc_Sknew(bigrational& discTnew,
			 bigrational& discTfnewprod,
			 bigrational& Vol,
                         bigrational& Idx);

    // Compute the order of the difference of two cusps in J_0(N).
    // Theorem. Let x,y be cusps of X_0(N).  Then the order of [x]-[y] in
    // J_0(N) is equal to the order of the modular symbols {x,y} in
    // the quotient group H_1(X_0(N),Q)/H_1(X_0(N),Z). 


    // reorder newforms lexicographically by their traces. 
   void OrderByTrace(Vector<int>& perm,
		     const MatVecList<bigrational>& anlist);
   void OrderByTrace(MatVecList<bigrational>& anlist);
   void OrderByTrace(MatVecList<bigrational>& anlist,
		     MatVecList<bigrational>& eigenbasis);

   bigint TorusPointsFp(Matrix<bigrational>& V, int p);
   bigint TorusTorsionBound(Matrix<bigrational>& V, int maxp);

private:
    void init_ZBasis_Mk() ;
    void init_ZBasis_Sk() ;

    Matrix<bigrational> ZBasis_Mk, ZBasis_Sk;
    bool init_zbasis_mk, init_zbasis_sk;
    void save_ZBasis_Sk() const;
    bool load_ZBasis_Sk();
    void save_Te(const Matrix<bigrational>& Te, int s) const;
    bool load_Te(Matrix<bigrational>& Te, int s);
} ;


#endif  // h1nchiz






