/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

// file intmod.h

// implememts class of integers module some a static prime pr.

#ifndef INTMOD_H
#define INTMOD_H

#include <iostream.h>

#include <LiDIA/bigint.h>
#include <LiDIA/bigrational.h>



class rootof1 ;			// forward declaration.

class intmod {
private:
    int val;
    static int pr;
    static int halfpr;
    static int prim_root ; 
public:
    static void set_pr(int p) ; 
    static int get_pr() { return pr; }
    static int get_prim_root () { return prim_root ; }
    
    intmod() { val = 0; }
    intmod(int n) { val = n%pr; }
    intmod(const intmod& n) { val = n.val; }
    intmod(const bigint& N) { 
      bigint M = N % pr ;
      int n; M.intify(n); *this=n; 
    } 
    intmod(const bigrational& x) ;
    intmod(const rootof1& x) ; 

    void normalize();

    intmod& operator = (const intmod& n) { val = n.val; }
    bool operator < (const intmod& n) 
      { cerr << "< makes no sense for intmods! " << endl ; }    
    bool operator > (const intmod& n) 
      { cerr << "> makes no sense for intmods! " << endl ; }    

    intmod& operator += (const intmod& n); 
    intmod& operator -= (const intmod& n); 
    intmod& operator *= (const intmod& n); 
    intmod& operator /= (const intmod& n);
    
    friend intmod  operator - (const intmod& n);// unary
      
    
    friend intmod  operator + (const intmod&, const intmod& );
    friend intmod  operator - (const intmod&, const intmod& );
    friend intmod  operator * (const intmod&, const intmod& );
    friend intmod  operator / (const intmod&, const intmod& );

    friend void power (intmod& ator, intmod a, int r) ;

    bool operator == (const intmod& n) const;
    bool operator != (const intmod& n) const;
    friend bool iszero(const intmod& n);
    
    friend intmod inv(const intmod&);
    friend int lift(const intmod&);
    friend int normlift(const intmod&);

    friend ostream& operator<< (ostream&s, const intmod&);
    friend istream& operator>> (istream&s, intmod&);

    void    error(const char* msg) const;     // error handling
};

inline intmod& intmod::operator /= ( const intmod& n )
{
    val *= inv(n).val;
    val %= pr;
    return *this;
}

inline intmod& intmod::operator += ( const intmod& n )
{
    val += n.val;
    val %= pr;
    return *this;
}

inline intmod& intmod::operator -= ( const intmod& n )
{
    val -= n.val;
    val %= pr;
    return *this;
}

inline intmod& intmod::operator *= ( const intmod& n )
{
    val *= n.val;
    val %= pr;
    return *this;
}

inline bool intmod::operator == (const intmod& n) const
{
    return ( (val-n.val)%pr == 0 );
}

inline bool intmod::operator != (const intmod& n) const
{
    return ( (val-n.val)%pr != 0 );
}

inline bool iszero(const intmod& n)
{
    return (n.val == 0);
}

inline void intmod::normalize()  // puts val in interval -p/2 < val <= p/2 
{
    if( val > halfpr ) val -= pr;
    else if( val <= -halfpr ) val += pr;
}

inline int lift(const intmod& n)   // lifts to integer 
{
  return n.val;
}

inline int normlift(const intmod& n)   // lifts to integer n, -p/2 < n <= p/2
{
    if( n.val > intmod::halfpr ) return( n.val - intmod::pr );
    else if( n.val <= -intmod::halfpr ) return( n.val + intmod::pr);
    else return n.val;
}

// using named returns below

inline intmod  operator + (const intmod& n1, const intmod& n2) return ans(n1)
{
  ans += n2;
}

inline intmod  operator - (const intmod& n1, const intmod& n2) return ans(n1)
{
  ans -= n2;
}

inline intmod  operator * (const intmod& n1, const intmod& n2) return ans(n1)
{
  ans *= n2;
}

inline intmod  operator / (const intmod& n1, const intmod& n2) return ans(n1)
{
  ans /= n2;
}

inline intmod  operator - (const intmod& n) return ans(n)
{
  ans.val *= -1;
}

inline void power (intmod& ator, intmod a, int r) 
{
  if (r == 0) { ator=1; return; }
  if (r < 0) {
    a=1/a;
    r=-r;
  }
  ator = 1 ;
  intmod t = a ;
  while (r >= 1) {
    if (r % 2) 
      ator *= t ;
    t *= t ;
    r /= 2 ;
  }
}


#endif
