/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

/* 
   FILE: lines.cc
   DESC:
*/


#ifndef LINES_CC
#define LINES_CC

#include "lines.h"

Lines::~Lines()
{
    delete [] entries;
    delete [] Start;
    delete [] Nline;
}

void Lines::error(const char* msg) const
{
    cout << "Fatal error in class Lines.\n " << msg << endl;
    exit(-1);
}

template <class T> void LinesVal<T>::error(const char* msg) const
{
    cout << "Fatal error in class LinesVal.\n " << msg << endl;
    exit(-1);
}

void Lines::clean() 
{
    delete [] entries; 
    entries = 0;
}

template <class T> void 
LinesVal<T>::entry( int row, int index, Elem<T>& elem )
{
    int d = Nline[row];
    if( index < 0 || index >= d ) 
	error("index out of bound in function entry");
    int st = Start[row]+index;
    elem.pos = entries[st];
    elem.val = values[st];
}

template <class T> void 
LinesVal<T>::setrow( int row, int d, long* pos, T* val )
{
    if( row < 1 || row > nro ) error("row out of bounds in setrow");
    if( d < 0 ) error( "d < 0 in function setrow");
    if( d > available() ) error("not enough space for setrow");
    Nline[row] = d;
    Start[row] = next;
    for( int i = 0; i < d; i++ ) {
	entries[next] = pos[i];
	values[next++] = val[i];
    }
}

template <class T> 
void LinesVal<T>::multiply( int row, const T& alpha)
{
    int d = Nline[row];
    T* val = values+Start[row];
    while(d--) *val++ *= alpha;
}

void Lines::setrow( int row, int d, long* pos)
{
    if( row < 1 || row > nro ) error("row out of bounds in setrow");
    if( d < 0 ) error( "d < 0 in function setrow");
    if( d > available() ) error("not enough space for setrow");
    Nline[row] = d;
    Start[row] = next;
    for( int i = 0; i < d; i++ ) entries[next++] = pos[i];
}	

template <class T> 
LinesVal<T>::LinesVal( const LinesVal<T>& lv ) : Lines(lv)
{
    values = new T [lv.size];
    for( int i = 0; i < next; i++ )
	values[i] = lv.values[i];
}

// obs: first entry of arrays Nline and Start are not used, 
// this is why these arrays have nro+1 entries.

Lines::Lines( long nr )
{
    nro = nr;
    Nline = new long [nro+1];
    Start = new long [nro+1];
    for( int r = 0; r <= nro; r++ ) {
	Nline[r] = 0;
	Start[r] = 0;
    }
    entries = 0;
    size = 0;
    slack = 0;
    next = 0;
}

Lines::Lines( const Lines& L )
{
    nro = L.nro;
    size = L.size;
    slack = L.slack;
    next = L.next;
    Nline = new long [nro+1];
    Start = new long [nro+1];
    for( int r = 0; r <= nro; r++ ) {
	Nline[r] = L.Nline[r];
	Start[r] = L.Start[r];
    }
    entries = new long [ size ];
    for( int i = 0; i < size; i++ ) entries[i] = L.entries[i];
}

void Lines::resize(int new_size) 
{
    if( entries ) {
	long* old_entries = entries;
	entries = new long [size = new_size];
	for( int i = 0; i < next; i++ ) entries[i] = old_entries[i];
	delete [] old_entries;
    }
    else entries = new long [size = new_size];
    for( int i = next; i < size; i++ ) entries[i] = 0;
    entries[size-1] = 1;    // signals end of array
}

template <class T> void
LinesVal<T>::resize(int new_size) 
{
    Lines::resize(new_size);
    if( values ) {
	T* old = values;
	values = new T [size=new_size];
	for( int i = 0; i < next; i++ ) values[i] = old[i];
	delete [] old;
    }
    else values = new T [size = new_size];
}

// reset value of slack, but doesn't apply the new value,
// ie. doesn't redo the array entries inserting slack zeros
// after each line
//
void Lines::set_slack( long new_slack ) 
{
    if( new_slack < 0 ) error(" negative value in function set_slack");
    slack = new_slack;
}

// Function reset_nro sets value of nro to a higher value.
//
void Lines::reset_nro( long new_nro ) 
{
    if( nro == new_nro ) return;
    if( nro > new_nro ) 
	error("Reseting nro will cause loss of information.");
    long* old_start = Start;
    long* old_nline = Nline;
    Nline = new long [new_nro+1];
    Start = new long [new_nro+1];
    Nline[0] = Start[0] = 0;
    for( int j = 1; j <= nro; j++ ) {
	Nline[j] = old_nline[j];
	Start[j] = old_start[j];
    }
    for( int j = nro+1; j <= new_nro; j++ ) {
	Nline[j] = 0;
	Start[j] = next;
    }
    nro = new_nro;
    delete [] old_nline;
    delete [] old_start;
}

// Does the permutation of rows given by the 
// vector v, i.e. if v[i] = j then row j will become row i.
//
void Lines::permutation( const Vector<int>& v)
{
    // first a complete sanity check:
    bool good = true;
    if( nro == dim(v) ) {
        int* check = new int [nro+1];
	for( int i = 1; i <= nro; i++ ) check[i] = 0;
	for( int i = 1; i <= nro; i++ ) check[v[i]] = 1;
	for( int i = 1; i <= nro; i++ ) good = good & check[i];
    }
    if( (nro != dim(v)) || (!good) )
	error( "Vector is not permutation of `nro' elements");
    
    // now do the permutation
    
    long * old_nline =  Nline;
    long * old_start =  Start;
    Nline = new long [nro+1];
    Start = new long [nro+1];
    Nline[0] = Start[0] = 0;
    for( int i = 1; i <= nro; i++ ) {
      Nline[i]  =  old_nline[v[i]];
      Start[i] = old_start[v[i]];
    }
    delete [] old_nline;
    delete [] old_start;
}

ostream& operator << ( ostream& s, const line_info& r) { return s;}

void Lines::garbage_collect( const Array<line_info>& RI )
{
    next = 0;
    int nrows = RI.length();
    for( int i = 0; i < nrows; i++ ) {
	int row = RI[i].row;          // row to be reallocated
	int d = Nline[row];
	int first = Start[row];
	if( first > next ) {
	    long* pos1 = entries + next;
	    long* pos2 = entries + first;
	    for( int j = 0; j < d; j++ ) *pos1++ = *pos2++;
	}
	else if( Start[row] < next ) 
	  error("in function garbage_collect");
	Start[row] = next;
	next += d;
	    
	// now tries to put slack zeros after end of row/column
	if( i < nrows-1 ) {
	    int next_row = RI[i+1].row;
	    if( Start[next_row]-next >= slack ) {
		for( int k = 0; k < slack; k++ ) 
		    entries[next++] = 0;
	    }
	}
    }
    
    // now complete to the end of array with zeros:
    long* pos = entries + next;
    for( int i = next; i < size; i++ ) *pos++ = 0;
    pos[-1] = 1;  // marks end of array;
}

void Lines::copy_to_end( int row)
{
    long* pos1 = entries+Start[row];
    long* pos2 = entries+next;
    for( int i = 0; i < Nline[row]; i++ ) {
	*pos2++ = *pos1;
	*pos1++ = 0;
    }
    Start[row] = next;
    next += Nline[row]+slack;
}

// Function take_out(int line, int elem) takes element elem
// out of line, adjusting Nline.
//
void Lines::take_out( int line, int elem )
{
    int d = Nline[line];
    int index = -1;
    long* pos = entries+Start[line];
    for( int i = 0; i < d; i++ ) 
	if( pos[i] == elem ) index = i;
    if( index == -1 ) 
	error( "elem not in line in function take_out");
    pos[index] = pos[--d];
    pos[d] = 0;
    Nline[line]--;
}

void Lines::display( int r ) const
{
    if( r <= 0 || r > nro ) error( "arg out of range in function display");
    int d = Nline[r];
    long st = Start[r];
    cout << "[";
    if( d > 0 ) cout << entries[st++];
    for( int i = 1; i < d; i++ ) cout << "," << entries[st++];
    cout << "]";
}

void Lines::set( int pos, int value)
{
    if( pos < 0 || pos > size-1 ) {
      cerr << "pos = " << pos << ",  size = " << size << endl ;
      error("in function set");
    }
    entries[pos] = value;
    if( pos >= next ) next = pos+1;
}

#include "charmod.h"
template class LinesVal<charmod>;
template class Elem<charmod>;

#include "array.cc"
template class Array<line_info>;

#endif
