/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

//  list.cc 
//  BaseList, List, Ordlist and ListLooper
#ifndef LIST_CC
#define LIST_CC

#include "list.h"

template <class T> void List<T>::error(const char* msg) const
{
    cerr << "Error in class List. " << msg << endl;
    exit(-1);
}

template <class T> List<T>& List<T>::operator=(const List<T>& L)
{ 
    if( this == &L ) return *this;
    num = L.num;
    if( size != L.size ) {
	delete [] items;
	items = new T [size = L.size];
    }
    T *t1 = items, *t2 = L.items;
    for(int i = 0; i < size; i++ ) *t1++ = *t2++;
    return *this;
}

template <class T> bool List<T>::inList(const T& X) const
{
    T* ptr = items;
    for( int i = 0; i < num; i++ )
	if( *ptr++ == X ) return true;
    return false;
}

template <class T> int List<T>::position(const T& X) const
{
    T* ptr = items;
    for( int i = 0; i < num; i++ )
	if( *ptr++ == X ) return i;
    return -1;
}

template <class type>
istream& operator>>( istream& is, List<type> &L )
{   // format for input is: first size of list then all elements
    int max;
    is >> max;
    L.reset(max);
    type t;
    for( int i = 0; i < max; i++ ) { is >> t; L.put(t); }
    return is;
}

template <class type>
void List<type>::remove( const type& X )
{
    type* plist = items;
    int s = 0;
    for( ; s < num; s++ ) if( *plist++ == X ) break;
    if ( s < num ) {   // X is in list
	plist--; type* qlist = plist + 1;
	for( int i = s; i < num-1; i++ ) *plist++ = *qlist++;
	num--;
    }
}

template <class T> bool Ordlist<T>::inList(const T& X) const
{   // uses binary search
    if( num == 0  || items[num-1] < X ) return false;
    int lb = 0;
    find(X,num-1,lb);
    if( items[lb] == X ) return true;
    return false;
}

template <class T> bool Ordlist<T>::inList(const T& X, int ub, int lb ) const
{
    if( (lb < 0) || (lb > ub) || (ub > num-1) ) 
	error("Bad bounds in Ordlist inList function");
    if( items[ub] < X ) return false;
    find(X,ub,lb);
    if( items[lb] == X ) return true;
    return false;
}

template <class T> int Ordlist<T>::position(const T& X) const
// returns -1 if it can't find X
{
    if( num == 0  || items[num-1] < X ) return -1;
    int lb = 0;
    find(X,num-1,lb);
    if( items[lb] == X ) return lb;
    return -1;
}

template <class T> int Ordlist<T>::position( const T& X, int ub, int lb ) const
{
    if( (lb < 0) || (lb > ub) || (ub > num-1) ) 
	error("Bad bounds in Ordlist position function");
    if( items[ub] < X ) return -1;
    find(X,ub,lb);
    if( items[lb] == X ) return lb;
    return -1;
}
 
template <class type>
void Ordlist<type>::put( const type& X )
{
    if( num == size ) grow();
    if( num == 0 || items[num-1] < X ) items[num++] = X; 
    else {
	int lb = 0;
	find(X,num-1,lb);
	if( items[lb] != X ) {
	    type *array = items + num -1;
	    for( long r = num; r > lb; r-- ) array[1] = *array--;
	    array[1] = X;
	    num++;
	}
    }
}

template <class type>
void Ordlist<type>::put( const Ordlist<type>& list )  
{
    if( num == 0 ) *this = list;
    else if( list.num == 0 ) ;   // do nothing!
     else {
	type *newArray = new type [ size = (num + list.num) ];
	type *na = newArray;
	ListLooper<type> L( list );
	int last = 0, lb = 0;
	while( L.OK() && last<num ) {
	    type X = L.next();
	    find(X,num-1,lb);
	    if( items[lb] != X ) {
		while( last < lb ) *na++ = items[last++];
		if( items[last] < X ) *na++ = items[last++];
		*na++ = X;
	    }
	}
	if( last == num ) while( L.OK() ) *na++ = L.next();
	else while( last < num ) *na++ = items[last++];
	delete [] items;
	items = newArray;
	num = na - newArray;
    }
}

template <class type>
void Ordlist<type>::remove( const type& X )
{
    if( num == 0 ) error("Removing from empty list");
    int lb = 0;
    find(X,num-1,lb);
    if( items[lb] != X ) error("element is not in list to remove");
    type *array = items + lb;
    for( long s = lb + 1; s < num; s++, array++ ) *array = array[1];
    num--;
}

template <class T>
void Ordlist<T>::find(const T& X, int ub, int& lb ) const
{
    while( items[lb] < X && lb != ub) {
	int i = (ub + lb)/2;
	items[i] < X ?  (lb = i+1) : (ub = i);
    } 
}

template <class type>
void Ordlist<type>::remove( const Ordlist<type>& L )  // L must be ordered
{
    if( L.num == 0 ) return;   // do nothing
    ListLooper<type> LP(L);
    type X = LP.next(); 
    int lb = 0;
    find(X,num-1,lb);
    if( items[lb] != X ) error("element is not in list to remove");
    int lb2 = lb;
    type *ar = items + lb;
    long last = lb+1;
    while( LP.OK() ) {
	X = LP.next();
	find( X, num-1, lb2 );
	if( items[lb2] != X ) error("element is not in list to remove");
	while( last < lb2 ) *ar++ = items[last++];
	last++;
    }
    while( last < num ) *ar++ = items[last++];
    num = ar - items;
}

#endif
