/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

// list.h
// template classes BaseList, List, Ordlist and ListLooper
 
#ifndef LIST_H
#define LIST_H

#include <iostream.h>
#include "array.h"

template < class T >
class List ;

template <class T>
class ListLooper ;


template <class T>
class BaseList 
{
    // virtual base class. It is the interface for all 
    // implementations of lists
protected:
    int num;              // number of elements in list
public:
    // constructors
    
    BaseList( int m ) : num(m) {}
    BaseList( const BaseList<T>& L ) : num(L.num) {}
    
    // basic access functions
    
    virtual T first() const = 0;            // returns first element
    virtual T last() const = 0;              // return last element
    virtual T& operator[] ( int n ) = 0;     // nth element in list
    
    // inserting and removing from list
    
    virtual void append( const T& X) = 0;    // append at end of list  
    virtual void put( const T& X) = 0;       // puts only if not there already
    virtual void remove( const T& X) = 0;    // removes from list if there
    
    const int count() const { return num; } // number of elements 
    bool empty() const { return num==0; }
    virtual bool inList(const T& X) const = 0; 
    virtual int position(const T&) const = 0;   // position starts at 0
    friend ostream& operator<< (ostream&, const List<T>&);
    friend istream& operator>> (istream&, List<T>&);
};
   
/* class List uses an array to store elements of list */

template < class T >
class List: public BaseList<T>, public Array<T>
{
    friend class ListLooper<T>;
public:
    // constructors and destructor
    
    List( int m = 0 ) : Array<T>(m), BaseList<T>(0) {}
    List( const List<T>& L ) : Array<T>(L), BaseList<T>(L.num) {}
    virtual ~List( ) {}
    List& operator=(const List<T>&);
       
    // basic access functions
    
    T first() const;
    T last() const;
    T& operator[] ( int n );
    const T& operator[] ( int n ) const;
    virtual bool inList(const T& X)   const; 
    virtual int  position(const T& X) const;
    
    // inserting and removing from list
    
    virtual void append( const T& X);  
    virtual void put( const T& X);     
    virtual void remove( const T& X);
    void reset(int);           // resizes forgetting the contents
    void clear();        // makes num = 0 without resizing array
    
    friend ostream& operator<< (ostream&, const List<T>&);
    friend istream& operator>> (istream&, List<T>&);   // first input number
				// of elements then the elements
    void    error(const char* msg) const;
};

template <class T>
class Ordlist : public List<T> 
{
public:
    // constructors and destructor
    
    Ordlist( int m = 0 ) : List<T>(m) {}
    Ordlist( const List<T>& L ); 
    Ordlist( const Ordlist<T>& L) : List<T>(L) {}
    
    ~Ordlist() {}
    // uses binary search for these functions

    bool inList(const T& X) const;
    bool inList( const T& X, int ub, int lb = 0) const; 
    int position(const T& X) const;
    int position(const T& X, int ub, int lb = 0) const;

    // functions to insert and remove
    
    void append( const T& X );
    void put( const T& X);
    void put( const Ordlist<T>& L);   
    void remove( const T& X );
    void remove( const Ordlist<T>& L );

private:
    void find(const T& X, int, int&) const;
};

template <class T>
class ListLooper
{
public:
    ListLooper( const List<T>& L ) { ListRef = &L; index = 0; }
    ListLooper( const Ordlist<T>& L ) { ListRef = &L; index = 0; }
    int ok() { return (index < (ListRef->num) ? 1 : 0 ); }
    int OK() { return (index < (ListRef->num) ? 1 : 0 ); }
    T next() { return ListRef->items[index++]; }
    void reset() { index = 0; }
    long getIndex() const { return index; }
private:
    const List<T>* ListRef;
    long index;
};


// inline member functions  of class List

template <class T> inline void List<T>::append( const T& X )
{
    if( num == size ) grow();
    items[num++] = X;
}

template <class T> inline void List<T>::put( const T& X )
{
    if( !inList(X) ) append(X);
}

template <class T> inline T List<T>::first() const
{
    if( num == 0 ) error("Can't get first element of empty list!");
    return *items;
}

template <class T> inline T List<T>::last() const
{
    if( num == 0 ) error("Can't get last element of empty list!");
    return items[num-1];
}

template <class T> inline void List<T>::clear()
{
  num = 0;
}

template <class T> inline void List<T>::reset( int n )
{
    delete [] items;
    items = new T [size = n];
    num = 0;
}

template <class T> inline ostream& operator<< (ostream& s, const List<T>& L)
{
    L.display(s,L.num);
    return s;
}

template <class T> inline T& List<T>::operator[] ( int n )
{
    if( n < 0 || n >= num ) error("Argument out of bounds in operator[]");
    return items[n];
}

template <class T> inline const T& List<T>::operator[] ( int n ) const
{
    if( n < 0 || n >= num ) error("Argument out of bounds in operator[]");
    return items[n];
}

// inline member functions of Ordlist

template <class T> inline 
Ordlist<T>::Ordlist( const List<T>& L ) : List<T>(L)
{
    qsort(num-1);
}

template <class T> inline void Ordlist<T>::append( const T& X )
{
    if( num != 0 && X <= items[num-1] ) 
	error("Appending to Ordlist would break the order");
    if( num == size ) grow();
    items[num++] = X;
}

#endif



