/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

// matrix.cc
#ifndef MATRIX_CC
#define MATRIX_CC

#include "matrix.h"
#include "charpoly.h"
#include "system.h"

template <class T>
void
Matrix<T>::reorder_rows(const Vector<int>& s) 
{
    if(s.size()!=nrows())
	die("Matrix::reorder_rows -- invalid permutation length.");
    Matrix<T> t(nrows(), ncols());
    for(int i=1; i<=nrows(); i++)
	t.setrow(i, this->row(s[i]));
    *this = t;
}

template <class T>
void
Matrix<T>::reorder_cols(const Vector<int>& s) 
{
    if(s.size()!=ncols())
	die("Matrix::reorder_cols -- invalid permutation length.");
    Matrix<T> t(nrows(), ncols());
    for(int i=1; i<=ncols(); i++)
	t.setcol(i, this->col(s[i]));
    *this = t;
}

template <class T>
void
Matrix<T>::reorder(const Vector<int>& s)
{
    reorder_cols(s);
    reorder_rows(s);
}

template <class T>
bool Matrix<T>::iszero() const 
{
    for(int r=1;r<=nrows();r++)
	for(int c=1;c<=ncols();c++)
	    if((*this)(r,c)!=T(0)) return false;
    return true;
}


template <class T> void Matrix<T>::error(const char* msg) const
{
    cerr << "Matrix error:" << endl << msg << endl;
    exit(1);
}

template <class T>
Matrix<T>::Matrix(const Vector<T>& v)
{
    mat=NULL; nro=nco=0;
    init(1,v.size());
    for(int i=1;i<=v.size();i++) (*this)(1,i)=v[i];
}

template <class T>
Matrix<T>::Matrix(const Matrix& m)
{
    nro=m.nro;
    nco=m.nco;
    long n = nro*nco;
    mat=new T[n];
    T *m1=mat, *m2=m.mat;
    while(n--) *m1++ = *m2++;
}

template <class T>
void Matrix<T>::init(long nr=0, long nc=0) 
    // resizes Matrix to dimension nr x nc and assign zero to all entries.
{                
    
    if(nr<0 || nc<0) { 
	cerr << "Matrix::init("<<nr<<","<<nc<<") obviously invalid params!"<<endl;
	die();
    }
    long n = nr*nc;
    if( nro*nco != n ) {        // delete old space and replace with new 
	delete [] mat;
	mat = new T[n]; 
	
    }
    nro = nr;
    nco = nc;
    T *m1=mat;
    while(n--) *m1++ = T(0);
}

template <class T>
void Matrix<T>::resize(long nr=0, long nc=0) 
    // resizes Matrix to dimension nr x nc, copying contents. 
{                            
    long n = nr*nc;
    T* newmat=new T[n];
    T *m1=newmat;
    while(n--) *m1++ = T(0);
    int mnro=nro<nr?nro:nr; int mnco=nco<nc?nco:nc;
    for(int r=0; r<mnro;r++)
	for(int c=0; c<mnco;c++) 
	    newmat[r*nc+c]=mat[r*nco+c];
    delete[] mat;
    nro = nr;
    nco = nc;
    mat = newmat;
}

// 1 <= i <= nro
//
template <class T> Vector<T> Matrix<T>::row(long i) const
{  
    if( 1 <= i && i <= nro ) {
	Vector<T> ans(nco);
	T *matij = mat+(i-1)*nco, *v = ans.vec;
	for( long j = nco; j; j-- ) *v++ = *matij++;
	return ans; 
    }
    else {
	cerr << "row Matrix error. The matrix has " << nro << "rows" << endl
	     << "but the program has tried to access row " << i << endl;
	die();
    }
}

// 1 <= i <= nro
//
template <class T> void Matrix<T>::row(long i, Vector<T>& V) const
{  
    if( 1 <= i && i <= nro ) {
	if( V.dim != nco ) {
	    delete [] V.vec;
	    V.vec = new T [V.dim = nco];
	}
	T *matij = mat+(i-1)*nco, *v = V.vec;
	for( long j = nco; j; j-- ) *v++ = *matij++;
    }
    else {
	cerr << "row Matrix error. The matrix has " << nro << "rows" << endl
	     << "but the program has tried to access row " << i << endl;
	die();
   }
}

template <class T> Vector<T> Matrix<T>::col(long j) const
{  
    if( j <= 0 || j > nco ) {
	cerr << "col Matrix error. The matrix has " << nco << "columns" << endl
	     << "but the program has tried to access column " << j << endl;
	die();
    }
    
    Vector<T> mj(nro);
    T *matij=mat+(j-1), *v=mj.vec;
    for( long i = nro; i; i-- ) { *v++ = *matij; matij+=nco; }
    return mj;
}

template <class T> void Matrix<T>::col(long j, Vector<T>& V) const
{  
    if( j <= 0 || j > nco ) {
	cerr << "col Matrix error. The matrix has " << nco << "columns" << endl
	     << "but the program has tried to access column " << j << endl;
	die();
    }

    if( V.dim != nro ) {
	delete [] V.vec;
	V.vec = new T [V.dim = nro];
    }
    
    T *matij = mat+(j-1), *v = V.vec;
    for( long i = nro; i; i-- ) { *v++ = *matij; matij+=nco; }
}


template <class T>
Matrix<T> Matrix<T>::slice(long r1,long r2,long c1,long c2) const
{
    if(c1<0) { // abbreviated form with firsts=1
	c2 = r2-1; 
	r2 = r1-1; 
	r1 = c1 = 0;
    }
    else {
	r1--; c1--; r2--; c2--;
    }
    long n = r2-r1+1, c = c2-c1+1; 
    long cc = c;
    Matrix<T> ans(n,cc);
    T* ap=ans.mat, *mp=mat+r1*nco+c1;
    while(n--) {
	c = cc;
	while(c--) *ap++ = *mp++; 
	mp += (nco-cc);
    }
    return ans;
}

template <class T>
Matrix<T>& Matrix<T>::operator = ( const Matrix<T>& m )
{
    if (this==&m) return *this;
    long n = m.nro*m.nco;
    if( nro*nco != n ) {        // delete old space and replace it with new 
	delete [] mat;
	mat = new T[n]; 
    }
    nro = m.nro;
    nco = m.nco;
    T *m1=mat, *m2=m.mat;
    while(n--) *m1++ = *m2++;
    return *this;
}

template <class T>
void Matrix<T>::setrow(long i, const Vector<T>& v)
{
    if ((0<i) && (i<=nro) && (dim(v)==nco))
	{
	    T * rowi = mat + (i-1)*nco;
	    T * vec = v.vec;
	    long c = nco; 
	    while(c--) *rowi++ = *vec++;
	}
    else {
	cerr << "setrow Matrix error. The matrix has " << nro << "rows" << endl
	     << "but the program has tried to access row " << i << endl;
	die();
    }
}

template <class T>
void Matrix<T>::setcol(long j, const Vector<T>& v)
{
    if ((0<j) && (j<=nco) && (dim(v)==nro))
	{
	    T * colj = mat+(j-1);
	    T * vec = v.vec;
	    long n = nro;
	    while(n--) {*colj = *vec++; colj+=nco;}
	}
    else {
	if(dim(v)==nro) {
	    cerr << "setcol Matrix error. The matrix has " << nco << " columns" << endl
		 << "but the program has tried to access column " << j << "." << endl;
	    die();
	} else {
	    cerr << "setcol Matrix error. The matrix has " << nro << " rows " << endl
		 << "but program has tried to set a column of length " << dim(v)<<endl;
	    die();
	}
    }
}

template <class T>
void Matrix<T>::addtocol(long i, const Vector<T>& v) 
{
#ifdef DEBUG
    if (nrows() != v.size() || i<1 || i>ncols()) {
	cerr << "Matrix<T>::addtocol -- invalid input -- size mismatches."<<endl;
	die();
    }
#endif
    for(int j=1; j<=v.size(); j++)
	(*this)(j,i) += v[j];
}

template <class T>
void Matrix<T>::addtorow(long i, const Vector<T>& v) 
{
#ifdef DEBUG
    if (ncols() != v.size() || i<1 || i>nrows()) {
	cerr << "Matrix<T>::addtorow -- invalid input -- size mismatches."<<endl;
	die();
    }
#endif
    for(int j=1; j<=v.size(); j++)
	(*this)(i,j) += v[j];
}


template <class T> void Matrix<T>::swaprows(long r1, long r2)
{
    if ((0<r1)&&(0<r2)&&(r1<=nro)&&(r2<=nro))
	{
	    T *mr1 = mat + (r1-1)*nco;
	    T *mr2 = mat + (r2-1)*nco;
	    long nc=nco;
	    T temp;
	    while(nc--) {temp = *mr1; *mr1++ = *mr2; *mr2++ = temp; }
	}
    else {
	cerr << "Bad row numbers " << r1 << " and " << r2
	     << " in Matrix::swaprows" << endl;
	cerr << "The matrix has " << nro << " rows." << endl;
	die();
    }
}

template <class T> void Matrix<T>::swapcols(long c1, long c2)
{
    if ((0<c1)&&(0<c2)&&(c1<=nco)&&(c2<=nco)) {
	T *mc1 = mat + c1-1;
	T *mc2 = mat + c2-1;
	T tmp;
	for( long i = 0; i < nro; i ++) {
	    tmp = *mc1;
	    *mc1 = *mc2;
	    *mc2 = tmp;
	    mc1 += nco;
	    mc2 += nco;
	}
    }
    else {
	cerr << "Bad column numbers " << c1 << " and " << c2
	     << " in Matrix::swapcols" << endl;
	cerr << "The matrix has " << nco << " columns." << endl;
	die();
    }
}

template <class T>
Matrix<T>& Matrix<T>::operator += ( const Matrix<T>& mat2 )
{
    if ((nro==mat2.nro) && (nco=mat2.nco)) {
	long n=nro*nco; T *m1=mat, *m2=mat2.mat;
	while(n--) (*m1++) += (*m2++);
    }
    else error( "Incompatible matrices in operator +=\n" );
    return *this;
}

template <class T>
Matrix<T>& Matrix<T>::operator -= ( const Matrix<T>& mat2 )
{
    if ((nro==mat2.nro) && (nco=mat2.nco)) {
	long n=nro*nco; T *m1=mat, *m2=mat2.mat;
	while(n--) (*m1++) -= (*m2++);
    }
    else error( "Incompatible matrices in operator -=\n" );
    return *this;
}

template <class T>
Matrix<T>& Matrix<T>::operator *= ( const T& scalar )
{
    T* mij = mat; 
    long n=nco*nro;
    while(n--) (*mij++) *= scalar;
    return *this;
}

template <class T>
Matrix<T>& Matrix<T>::operator /= ( const T& scalar )
{
    T* mij = mat;  
    long n=nco*nro;
    while(n--) (*mij++) /= scalar;
    return *this;
}

//Definitions of non-member, friend operators and functions

template <class T>
Matrix<T> operator*(const Matrix<T>& m1, const Matrix<T>& m2)
{
    long i,j,k, m=m1.nro, n=m1.nco, p=m2.nco;
    Matrix<T> m3(m,p); 
    
    m3.init(m,p);
    /*
      long z=nr*nc;
      T *pm3=m3.mat;
      while (z--) *pm3++ = 0 * (*m1.mat);
    */
    
    
    T *a=m1.mat, *b=m2.mat, *c=m3.mat, *bp, *cp;
    if (n==m2.nro)  // algorithm from Dr Dobb's Journal August 1993
	{
	    while(m--)
		{
		    bp=b; k=n;
		    while(k--)
			{
			    cp=c; j=p;
			    while(j--)
				{
				    *cp++ += *a * *bp++;
				}
			    a++;
			}
		    c += p;
		}
	}
    else m1.error( "Incompatible sizes in Matrix product\n" );
    return m3;
}

template <class T>
T TraceOfProduct(const Matrix<T>& A, const Matrix<T>& B)
{
    if(A.ncols()!=B.nrows() || A.nrows()!=B.ncols()) 
	die("TraceOfProduct: both matrices much be square");
    T t(0);
    for(int i=1;i<=A.nrows();i++)
	t+=A.row(i) * B.col(i);
    return t;
}

template <class T>
bool operator==(const Matrix<T>& m1, const Matrix<T>& m2)
{
    long nr=m1.nro, nc=m1.nco;
    bool equal = ((nr==m2.nro) && (nc==m2.nco)); 
    if(!equal) return false;
    T *m1ij=m1.mat, *m2ij=m2.mat; 
    long n=nr*nc;
    while((n--)&&equal) equal=((*m1ij++)==(*m2ij++));
    return equal;
}

template <class T> ostream& operator<<(ostream& s, const Matrix<T>& m)
{
    T* mij = m.mat;
    s << "[";
    long nr=m.nro;
    while(nr--) {
	long nc = m.nco;
	//    s<<"[";
	while(nc--) {s<<(*mij++); if(nc) s<<",";}
	//    s<<"]"; 
	if(nr) s<<";\n";
    }
    s << "]";
    return s;
}

template <class T> istream& operator>>(istream& s, Matrix<T>& m)
{
    int nr, nc;
    nr = get_int("nrows = ");
    nc = get_int("ncols = ");
    m.init(nr,nc);
    for(int r=1;r<=nr;r++)
	for(int c=1;c<=nc;c++)
	    s>>m(r,c);
    return s;
}

template <class T>
void Matrix<T>::transpose(Matrix<T>& TR) const
{
    TR.init(nco,nro);
    T * Tmat = TR.mat;
    for(long r=1; r<=nco; r++)
	for(long c=1; c<=nro; c++)
	    TR(r,c) = (*this)(c,r);
}  

template <class T> 
void Matrix<T>::transpose()
{
    Matrix<T> temp;
    transpose(temp);
    *this = temp;
}


template <class T> T Matrix<T>::determinant() const
{
    Vector<T> cp(nro+1); 
    charpoly(cp);
    T det = cp[1];
    if( nro%2 == 1 ) det = -det;
    return det;
}

/*
// Determinant Using Gauss-Bareiss (Algorithm 2.2.6 of Cohen).
// After some testing, it turns out that this is *MUCH* slower
// than the NTL code, so I won't use this implimentation. 
template <class T> T Matrix<T>::determinant() const
{
    if (nrows()!=ncols()) die("determinant: non square matrix!");
    int n=nrows();
    Matrix<T> M=*this;

    // 1. [Initialize]
    int k=0;
    T c=T(1), s=T(1);

    // 2. [Increase k]
    while(k+1<n) {
	k++;
	T p=M(k,k);
	
	// 3. [Is p=0?]
	if(p==T(0)) {
	    int i=0;
	    for(int j=1;j<=n;j++)
		if(M(j,k)!=T(0)) {
		    i=j;
		    break;
		}
	    if (i==0) return T(0);
	    for(int j=k;j<=n;j++) {
		T t=M(k,j);
		M(k,j)=M(i,j);
		M(i,j)=t;
	    }
	    s=-s;
	    p=M(k,k);
	}
	
	// 4. [Main step]
	for(int i=k+1;i<=n;i++)
	    for(int j=k+1;j<=n;j++) {
		T t=p*M(i,j)-M(i,k)*M(k,j);
		M(i,j) = t/c;
	    }
	c=p;
    }
    return s*M(n,n);
}
*/

template <class T> T Matrix<T>::trace() const
{ 
    T ans(0);
    for( long i = 0; i < nro; i++) ans += mat[i*nco+i];
    return ans;
}

#include<stdlib.h>
template <class T>
void 
Matrix<T>::random(int max)
{
    for(int r=1;r<=nro;r++)
	for(int c=1;c<=nco;c++)
	    (*this)(r,c)=T(::random()%max);
}

//////////////////// CODE FOR COMPUTING CHARACTERISTIC POLYNOMIALS /////

void charpoly(Vector<bigcomplex>& cpoly, const Matrix<bigcomplex>& M) {
    long m = M.nrows();
    if (m < 1) { cpoly.init(1); cpoly[1]=bigcomplex(0); return; }
    Matrix<bigcomplex> b(M);
    Matrix<bigcomplex> id; id.idmat(m);
    bigcomplex t = M.trace();
    cpoly.init(m+1);
    cpoly[m+1]   = bigcomplex(1);
    cpoly[m] = -t;
    for( long i = 2; i <= m; i++ ) {
	b = M*(b-t*id);   
	t = b.trace()/i;
	cpoly[m-i+1] = -t;
    }
}

#include "my_ntl.h"
void charpoly(Vector<charmod>& cpoly, const Matrix<charmod>& M) {
    ntl_charpoly(cpoly, M);
}

void charpoly(Vector<intmod>& cpoly, const Matrix<intmod>& M) {
    ntl_charpoly(cpoly, M);
}




void charpoly(Vector<algnum>& cpoly, const Matrix<algnum>& M) {
    long m = M.nrows();
    if (m < 1) { cpoly.init(1); cpoly[1]=algnum(0); return; }
    Matrix<algnum> b(M);
    Matrix<algnum> id; id.idmat(m);
    algnum t = M.trace();
    cpoly.init(m+1);
    cpoly[m+1]   = algnum(1);
    cpoly[m] = -t;
    for( long i = 2; i <= m; i++ ) {
	b = M*(b-t*id);   
	t = b.trace()/i;
	cpoly[m-i+1] = -t;
    }
}
void charpoly(Vector<ffield_element>& cpoly, const Matrix<ffield_element>& M) {
    die("Warning: Charpoly ffield_element not implimented."); }
void charpoly(Vector<bigrational>& cpoly, const Matrix<bigrational>& M) 
{
    // Use the function from the charpoly package.  (This uses the NTL library.) 
    CharPoly(cpoly, M);
}
void charpoly(Vector<int>& cpoly, const Matrix<int>& M) {
    die("Charpoly Matrix<int> not implimented.");
}
void charpoly(Vector<bigint>& cpoly, const Matrix<bigint>& M) {
    Matrix<bigrational> N(M.nrows(),M.ncols());
    for(int i=1;i<=M.nrows();i++)
	for(int j=1;j<=M.ncols();j++)
	    N(i,j)=M(i,j);
    Vector<bigrational> v;
    charpoly(v,N);
    cpoly.init(v.size());
    for(int i=1; i<=v.size(); i++)
	cpoly[i]=numerator(v[i]);
}

template <class T> void Matrix<T>::charpoly( Vector<T>& cpoly) const
{ 
    if(nro != nco) 
	die("Matrix::charpoly -- nonsquare matrix.");
    ::charpoly(cpoly, *this);
}

// Function `SimpleEchelon' performs gauss elimination with back
// substitution but no pivoting. Used for finite fields where  
// overflow and stability is not a concern.
// permutation of rows is only implicit
// Last maintained: 27.09.96
//
template <class T> void  
Matrix<T>::SimpleEchelon(Vector<int>& elimcol, Vector<int>& elimrow)
{
    // elimcol[c] = r means col c is pivotal with pivot at row r
    // elimrow[r] = c means row r is pivotal with pivot at col c
    elimcol.init(nco);
    elimrow.init(nro);
    for( int i = 1; i <= nro; i++ ) elimrow[i] = 0;
    
    mesg("[simple echelon");
    
    for( long c = 1; c <= nco; c++) {   // loops through the columns

	if (c%25==0) { mesg(" "); mesg(c*100/nco); mesg("%"); }

	// choosing  first non zero entry as a pivot
	int row = 0;  
	for( int n = c-1, i = 1; i <= nro; i++, n += nco ) 
	    if( mat[n] != T(0) && elimrow[i] == 0) { 
		row = i; 
		break; 
	    }
	if( row == 0 )    // col c is empty 
	    elimcol[c] = 0;
	else {   
	    // normalize pivot row
	    //
	    T piv = mat[(row-1)*nco+c-1]; 
	    if( piv != T(1) ) {
		T invpiv = T(1)/piv;
		T* mij = mat + (row-1)*nco+(c-1); 
		for( int i = c; i <= nco; i++ ) *mij++ *= invpiv;
	    }
	    // clear this column
	    //
	    T *m1 = mat + (row-1)*nco + (c-1);
	    for( int r = 1; r <= nro; r++ ) {
		if( r == row ) continue;
		T *m2 = mat + (r-1)*nco + (c-1);
		T val = -m2[0];  
		if( val != T(0) ) 
		    for( int i = c; i <= nco; i++ ) *m2++ += (val * (*m1++) );
		m1 = mat + (row-1)*nco + (c-1);
	    }
	    elimcol[c] = row;
	    elimrow[row] = c;
	}
    }
    mesg("]\n");
}

/*
bigrational 
SimpleVolume(Matrix<bigrational> A)
{
    // elimcol[c] = r means col c is pivotal with pivot at row r
    // elimrow[r] = c means row r is pivotal with pivot at col c
    long nro=A.nrows(), nco=A.ncols();
    Vector elimcol; elimcol.init(A.nco);
    Vector elimrow; elimrow.init(A.nro);
    bigational V=1;
    
    mesg("[SimpleVolume");

    for(long c=1; c<=nco; c++) {   // loops through the columns
	if (c%25==0) { mesg(" "); mesg(c*100/nco); mesg("%"); }

	// choosing  first non zero entry as a pivot
	int row=0;  
	for(int r=1; r<=nro; r++) 
	    if(A(r,c)!=0 && elimrow[r]==0) {
		row=r; break; 
	    }
	if(row==0)    // col c is empty 
	    elimcol[c]=0;
	else {   
	    // compute least common multiple of other denominators in this row.
	    bigint l=1;
	    for(int r=1;r<=nro;r++) {
		if(r==row) continue;
		l=lcm(denominator(A(r,c)));
	    }

	    // normalize pivot row
	    bigrational piv=A(row,c)/l;
	    V*=piv;
	    if(piv!=1) {
		bigrational invpiv=1/piv;
		for(int i=c; i<=nco; i++) A(row,i) *= invpiv;
	    }
	    
	    // clear column, using only integer multiples of pivot row.
	    for(int r=1;r<=nro;r++) {
		if(r==row) continue;
		bigrational val=-A(r,c)/A(row,c);
		if(denominator(val)!=1) die("denominator(val)!=1");
		for(int i=c;i<=nco;i++)
		    A(r,i)+=val*A(row,c);
	    }

	    elimcol[c] = row;
	    elimrow[row] = c;
	}
    }
    mesg("]\n");
}
*/

inline int find_pivot (const bigrational* mij, const int& nro,
		       const int& nco, const int& r) {
    bigrational big = abs(*mij) ;
    long rmax = r, r2 = r+1; 
    for( mij += nco; r2 < nro; r2++, mij += nco) {
	bigrational mr2c = abs(*mij);
	if( 0 < mr2c && mr2c > big ) { big = mr2c; rmax = r2 ; }
    }
    return (big == 0) ? -1 : rmax;
}


inline int find_pivot (const bigcomplex* mij, const int& nro,
		       const int& nco, const int& r) {
    bigcomplex big = *mij; 
    long rmax = r, r2 = r+1 ;
    for( mij += nco; r2 < nro && big == 0; r2++, mij += nco) {
	big = *mij; rmax = r2;
    }
    return (big == 0) ? -1 : rmax ;
}


inline int find_pivot (const bigint* mij, const int& nro,
		       const int& nco, const int& r) {
    bigint big = abs(*mij) ;
    long rmax = r, r2 = r+1; 
    for( mij += nco; r2 < nro; r2++, mij += nco) {
	bigint mr2c = abs(*mij);
	if( 0 < mr2c && mr2c > big ) { big = mr2c; rmax = r2 ; }
    }
    return (big == 0) ? -1 : rmax;
}

inline int find_pivot (const algnum* mij, const int& nro,
		       const int& nco, const int& r) {
    bigrational big = abs(norm(*mij)) ; // pivot is thing of largest *norm*.
    long rmax = r, r2 = r+1; 
    for( mij += nco; r2 < nro; r2++, mij += nco) {
	bigrational mr2c = abs(norm(*mij));
	if( 0 < mr2c && mr2c > big ) { big = mr2c; rmax = r2 ; }
    }
    return (big == 0) ? -1 : rmax;
}

inline int find_pivot (const ffield_element* mij, const int& nro,
		       const int& nco, const int& r) {
    ffield_element big = *mij; 
    long rmax = r, r2 = r+1 ;
    for( mij += nco; r2 < nro && big == ffield_element(0); 
	 r2++, mij += nco) {
	big = *mij; rmax = r2;
    }
    return (big == ffield_element(0)) ? -1 : rmax ;
}


inline int find_pivot (const charmod* mij, const int& nro,
		       const int& nco, const int& r) {
    charmod big = *mij; 
    long rmax = r, r2 = r+1 ;
    for( mij += nco; r2 < nro && big == 0; r2++, mij += nco) {
	big = *mij; rmax = r2;
    }
    return (big == 0) ? -1 : rmax ;
}

inline int find_pivot (const intmod* mij, const int& nro,
		       const int& nco, const int& r) {
    intmod big = *mij; 
    long rmax = r, r2 = r+1 ;
    for( mij += nco; r2 < nro && big == 0; r2++, mij += nco) {
	big = *mij; rmax = r2;
    }
    return (big == 0) ? -1 : rmax ;
}

inline int find_pivot (const int* mij, const int& nro,
		       const int& nco, const int& r) {
    int big = *mij; 
    long rmax = r, r2 = r+1 ;
    for( mij += nco; r2 < nro && big == 0; r2++, mij += nco) {
	big = *mij; rmax = r2;
    }
    return (big == 0) ? -1 : rmax ;
}

template <class T> void 
Matrix<T>::echelon(Vector<int>& pcols, Vector<int>& npcols,
		   long& rk, long& ny )
{
    mesg("[echelon");
    T *mij = mat;
    pcols.init(nco);
    npcols.init(nco);
    rk=0; ny=0; 
    for (long r = 0, c = 0; c < nco && r < nro; c++) {
	if (c%25==0) { mesg(" "); mesg(c*100/nco); mesg("%"); }
	mij = mat + r*nco + c;
	
	// PIVOTING -- choose the row with the biggest abs
	// value in that col, or if the field is finite, choose
	// the first nonzero entry. 
	long rmax = find_pivot (mij, nro, nco, r) ;
	
	if( rmax == -1 ) npcols[++ny] = c+1;  // only zeros in this column
	
	else  {     
	    pcols[++rk] = c+1;
	    if( rmax > r ) { //swap rows
		T *mi1 = mat + r*nco, *mi2=mat + rmax*nco, temp; 
		long n = nco;
		while(n--) {temp = *mi1; *mi1++ = *mi2; *mi2++ = temp;}
	    }
	    
	    // now divide pivot row by pivot element
	    
	    T piv = mat[r*nco+c]; 
	    T biginv = T(1)/piv;
	    mij = mat + r*nco; 
	    for( long n = nco; n; n-- ) 
		*mij++ *= biginv;
	    mat[r*nco+c] = T(1);      // helps with non-exact types
	    // eliminate from rows below r
	    
	    for( long r3 = r+1; r3 < nro; r3++) {
		T *mi1 = mat + r*nco, *mi2 = mat + r3*nco;
		T dum = *(mi2+c); 
		for( long n = nco; n; n-- ) *mi2++ -= dum * *mi1++;
		mat[r3*nco+c] = T(0);   // helps with non-exact types
	    }
	    r++;
	}
    }
    for (long c = rk+ny; c < nco; c++) npcols[++ny] = c+1 ;
    pcols  =  pcols.slice(rk);
    Vector<int> sl = npcols.slice(ny);
    npcols.init(ny); npcols = sl;
    //npcols =  npcols.slice(ny);    // truncate index vectors
    
    // Back elimination 
    mesg("back elim");
    
    if (ny>0) {
	for( long r = 1; r <= rk; r++ )
	    for( long r2 = r+1; r2 <= rk; r2++ )  
		elim( r2, r, pcols[r2] ); 
    }
    else {
	mij = mat;
	for( long r = 1; r <= rk; r++ )
	    for( long c = 1; c <= nco; c++ ) 
		*mij++ = T((c==pcols[r])?1:0); 
    }
    slice(rk, nco);
    mesg("]\n");
}

template <class T>
void Matrix<T>::elim(long r1, long r2, long col)
{
    // this function "eliminates" column col from row r2 by multiplying
    // row r1 by a suitable scalar and summing to r2. It only works for types
    // which are fields. Integers should be converted to rationals.
    T *mr1 = mat + (r1-1)*nco;
    T *mr2 = mat + (r2-1)*nco;
    T factor = mr2[col-1]/mr1[col-1];
    if( factor == T(0) ) return;
    for( long nc = nco; nc; nc--) {
	*mr2++ -= factor * (*mr1++);
    }
    mat[(r2-1)*nco+col-1] = 0;   // this helps with non-exact types
}
template <class T>
void Matrix<T>::elim2(long r1, long r2, long col)
{
    // As elim, but entries of r1 before col are all zero.
    
    T *mr1 = mat + (r1-1)*nco + (col-1);
    T *mr2 = mat + (r2-1)*nco + (col-1);
    T factor = *mr2 / *mr1;
    if( factor == T(0) ) return;
    for( long nc = col-1; nc < nco; nc++) {
	*mr2++ -= factor * (*mr1++);
    }
}

bool fastsub(Matrix<charmod>& fT, const Vector<charmod>& v,
	     const Matrix<charmod>& th){return false;}
bool fastsub(Matrix<intmod>& fT, const Vector<intmod>& v,
	     const Matrix<intmod>& th){return false;}
bool fastsub(Matrix<algnum>& fT, const Vector<algnum>& v,
	     const Matrix<algnum>& th){return false;}
bool fastsub(Matrix<ffield_element>& fT, const Vector<ffield_element>& v,
	     const Matrix<ffield_element>& th){return false;}
bool fastsub(Matrix<bigrational>& fT, const Vector<bigrational>& v,
	     const Matrix<bigrational>& th)
{
    if(v.poldeg()<=0) return false;
    for(int i=1;i<=v.size();i++)
	if(v[i].denominator()!=1) return false;
    bigint d;
    mat_ZZ A, fA;
    my_to_ntl(A,d,th);
    if(d!=1) return false;
    mesgln("Using fast NTL sub.");
    fA.SetDims(th.nrows(),th.nrows());
    ZZ a;
    my_to_ntl(a,v[v.size()].numerator());
    fA=a*A;
    for (int i=v.size()-1; i>=1; i--) {
	my_to_ntl(a,v[i].numerator());
	if(a!=0)
	    for(int j=1;j<=th.nrows();j++) fA(j,j)+=a;
	if (i>1) fA *= A;
    }
    ntl_to_my(fT,fA);
    mesgln("finished fast sub.");
    return true;
}
bool fastsub(Matrix<int>& fT, const Vector<int>& v,
	     const Matrix<int>& th){return false;}
bool fastsub(Matrix<bigint>& fT, const Vector<bigint>& v,
	     const Matrix<bigint>& th){return false;}
bool fastsub(Matrix<bigcomplex>& fT, const Vector<bigcomplex>& v,
	     const Matrix<bigcomplex>& th){return false;}



template <class T>
void Matrix<T>::polsubst(Matrix<T>& fT, const Vector<T>& v) const
{
    mesg("polsubst: A is ",nro);mesg("x");mesg(nco);
    mesgln(" and deg(f)=",v.poldeg());
    if(fastsub(fT,v,*this)) return;
    if(nro!=nco) error("polsubst: nonsquare matrix.");
    fT.init(nro,nro);
    if(v.poldeg()<0) 		// 0 matrix
	return;
    if(v.poldeg()==0) {         // constant matrix
	for(int j=1;j<=nro;j++) fT(j,j)=v[1];
	return;
    }
    if(v[v.size()]==T(1))
	fT = *this;
    else
	fT = v[v.size()] * (*this);

    for (int i=v.size()-1; i>=1; i--) {
	if(v[i]!=T(0))
	    for(int j=1;j<=nro;j++) fT(j,j)+=v[i];
	if (i>1) fT *= (*this);
    }
    mesg("Done.\n");
}


// functions to write Matrix to file:

template <class T>
void Matrix<T>::WriteToFile( const char* fileName , const char* MatrixName)
{
    ofstream outFile(fileName, ios::out);
    writeMatrix(outFile, MatrixName);
}

template <class T>
void Matrix<T>::AppendToFile( const char* fileName , const char* MatrixName)
{
    ofstream outFile(fileName, ios::app);
    writeMatrix(outFile, MatrixName);
}

template <class T>
void Matrix<T>::writeMatrix(ofstream& outFile, const char* MatrixName)
{
    if( !outFile ) {
	cerr << "cannot open file for output\n";
	exit(-1);
    }
    if( MatrixName ) {
	outFile << MatrixName << " = ";
    }
    outFile << "[ ";
    T* ans = mat;
    for( int i = 1; i <= nro; i++ ) {
	for( int j = 1; j < nco; j++ ) {
	    outFile << *ans++<< ", ";   // matrix elements
	}
	outFile << *ans++ ;
	if( i < nro ) outFile << "; ";
    }
    outFile << " ]\n";
}

// function `count()' counts number of non-zero entries.
// Last mantained: 26.09.96
//
template <class T> int Matrix<T>::count() const
{
    int n=0;
    for(int r=1; r<=nrows(); r++)
	for(int c=1; c<=ncols(); c++)
	    if((*this)(r,c)!=T(0)) n++;
    return n;
}

// This lack of factoring out common functionality sucks
// but I can't see a good way around it without using macros
// or null pointers, and later on when I want to optimize
// each version this will be the only way which makes sense.
// The reason it is necessary now is because saving charmods
// is completely different, as we can store them as a list
// of char's.

const int VERBOSE_SIZE=100;

void Load (ifstream& in, Matrix<charmod>& M) 
{
    for (int i = 1 ; i <= M.nrows(); i++) { 
	if (i%100==0) {
	    mesg(i*100/M.nrows()); mesg("% ");
	}
	for (int j = 1; j <= M.ncols(); j++) {
	    uchar u; in.get(u);
	    M(i,j)=u;
	}
    }
}


void Load (ifstream& in, Matrix<intmod>& M)
{
    for (int i = 1 ; i <= M.nrows(); i++) { 
	if (i%100==0) {
	    mesg(i*100/M.nrows());
	    mesg("% ");
	}
	for (int j = 1; j <= M.ncols(); j++) {
	    in >> M(i,j) ; 
	}
    }
}
void Load (ifstream& in, Matrix<bigrational>& M)
{
    for (int i = 1 ; i <= M.nrows(); i++) { 
	if (i%100==0)
	    mesg(i*100/M.nrows(),"% ");
	for (int j = 1; j <= M.ncols(); j++) {
	    in >> M(i,j) ; 
	}
    }
}

void Load (ifstream& in, Matrix<bigint>& M)
{
    for (int i = 1 ; i <= M.nrows(); i++) { 
	if (i%100==0)
	    mesg(i*100/M.nrows(),"% ");
	for (int j = 1; j <= M.ncols(); j++) {
	    in >> M(i,j) ; 
	}
    }
}

typedef polynomial<bigrational> polbigrat;

void Load (ifstream& in, Matrix<polbigrat>& M)
{
    die("Load polynomial<bigrational> not implimented.");
}

void Load (ifstream& in, Matrix<algnum>& M)
{
    for (int i = 1 ; i <= M.nrows(); i++) {
        if (i%100==0)
            mesg(i*100/M.nrows(),"% ");
        for (int j = 1; j <= M.ncols(); j++) {
            M(i,j).load(in) ;
        }
    }
}

void Load (ifstream& in, Matrix<ffield_element>& M)
{
    for (int i = 1 ; i <= M.nrows(); i++) { 
	if (i%100==0)
	    mesg(i*100/M.nrows(),"% ");
	for (int j = 1; j <= M.ncols(); j++) {
	    M(i,j).load(in) ; 
	}
    }
}

void Load (ifstream& in, Matrix<int>& M)
{
    for (int i = 1 ; i <= M.nrows(); i++) { 
	if (i%100==0)
	    mesg(i*100/M.nrows(),"% ");
	for (int j = 1; j <= M.ncols(); j++) {
	    in >> M(i,j) ; 
	}
    }
}

void Load (ifstream& in, Matrix<bigcomplex>& M)
{
    for (int i = 1 ; i <= M.nrows(); i++) { 
	if (i%100==0)
	    mesg(i*100/M.nrows(),"% ");
	for (int j = 1; j <= M.ncols(); j++) {
	    in >> M(i,j) ; 
	}
    }
}



void Save (ofstream& of, const Matrix<polbigrat>& M) 
{
    for (int i = 1 ; i <= M.nrows(); i++) 
	for (int j = 1; j <= M.ncols(); j++)
	    of << M(i,j) << endl ;
    of << endl ;
}

void Save (ofstream& of, const Matrix<bigint>& M) 
{
    for (int i = 1 ; i <= M.nrows(); i++) 
	for (int j = 1; j <= M.ncols(); j++)
	    of << M(i,j) << endl ;
    of << endl ;
}
void Save (ofstream& of, const Matrix<charmod>& M) 
{
    for (int i = 1 ; i <= M.nrows(); i++) 
	for (int j = 1; j <= M.ncols(); j++)
	    of.put(ulift(M(i,j))) ;
    of << endl ;
}
void Save (ofstream& of, const Matrix<intmod>& M) 
{
    for (int i = 1 ; i <= M.nrows(); i++) 
	for (int j = 1; j <= M.ncols(); j++)
	    of << M(i,j) << endl ;
}
void Save (ofstream& of, const Matrix<algnum>& M) 
{
    for (int i = 1 ; i <= M.nrows(); i++) 
	for (int j = 1; j <= M.ncols(); j++)
	    M(i,j).save(of);
}
void Save (ofstream& of, const Matrix<bigrational>& M) 
{
    for (int i = 1 ; i <= M.nrows(); i++) 
	for (int j = 1; j <= M.ncols(); j++)
	    of << M(i,j) << endl ;
}
void Save (ofstream& of, const Matrix<ffield_element>& M) 
{
    for (int i = 1 ; i <= M.nrows(); i++) 
	for (int j = 1; j <= M.ncols(); j++)
	    M(i,j).save(of);
}
void Save (ofstream& of, const Matrix<int>& M) 
{
    for (int i = 1 ; i <= M.nrows(); i++) 
	for (int j = 1; j <= M.ncols(); j++)
	    of << M(i,j) << endl ;
}
void Save (ofstream& of, const Matrix<bigcomplex>& M) 
{
    for (int i = 1 ; i <= M.nrows(); i++) 
	for (int j = 1; j <= M.ncols(); j++)
	    of << M(i,j);	// the lieing twisted bigcomplex class does *NOT* work
				// *unless* there are no spaces between the complex numbers!!
    of<<endl;
}

template <class T>
void Matrix<T>::Load(ifstream& in) 
{
    int new_nro, new_nco ;
    in >> new_nro >> new_nco ;
    uchar dummy; in.get(dummy);   // get the carriage return.
    if(new_nro*new_nco>VERBOSE_SIZE) {
	mesg("Load "); mesg(new_nro); mesg("x");
	mesg(new_nco); mesg(" matrix. ");
    }
    init (new_nro, new_nco) ;
    ::Load (in, *this);         // :: for global scope

    if(new_nro*new_nco>VERBOSE_SIZE) mesgln("Done.");

}

template <class T>
void Matrix<T>::Save(ofstream& of) const
{
    of << nrows() << ' ' << ncols() << endl ;
    ::Save(of,*this);		      // :: for global scope
}

template <class T>
void Matrix<T>::LoadSparse(ifstream& in) 
{
    int new_nro, new_nco ;
    in >> new_nro >> new_nco ;
    uchar dummy; in.get(dummy);   // get the carriage return.

    if(new_nro*new_nco>VERBOSE_SIZE) {
	mesg("[loading "); 
	mesg(new_nro); mesg(" x ");
        mesg(new_nco); mesg(" (sparse saved) matrix: ");
    }

    init (new_nro, new_nco) ;
    while (1) {
	int x,y; in>>x>>y;
	if(x==-1) break;
	T t; in>>t;
	(*this)(x,y)=t;
    }

    if(new_nro*new_nco>VERBOSE_SIZE) { mesg("]\n"); }

}

template <class T>
void Matrix<T>::SaveSparse(ofstream& of) const
{
    of << nrows() << ' ' << ncols() << endl ;
    for(int i=1;i<=nrows();i++)
	for(int j=1;j<=ncols();j++)
	    if((*this)(i,j)!=T(0)) 
		of << i << " " << j << " " << (*this)(i,j) << endl ;
    of << "-1 -1 0"<<endl;
}

template <class T>
void Matrix<T>::RestrictedTo(Matrix<T>& B, const Matrix<T>& W) const
{
    if (W.nrows() != nrows())
	error("Bad input to RestrictedTo -- wrong number of rows.");
    
    Matrix<T> TW = (*this)*W ;
    Matrix<T> W_TW(W.nrows(), 2*W.ncols());
    for (int c = 1 ; c <= W.ncols(); c++) {
	W_TW.setcol(c, W.col(c)) ;
	W_TW.setcol(c + W.ncols(), TW.col(c)) ;
    }
    TW.init(0,0);			// free up space
    Vector<int> v1, v2;
    W_TW.SimpleEchelon(v1,v2);
    for (int c = W.ncols()+1; c<=2*W.ncols(); c++) 
	if (v1[c]!=0) {
	    cerr<<"** SERIOUS ERROR: subspace is not invariant!!! ** "<<endl ;
	    cerr<<"Restricting A to W, where " << endl << "A = " << (*this)
		<< endl << "W = " << W << endl;
	    die();
	}
    
    B.init (W.ncols (), W.ncols ()) ;
    for (int r = 1 ; r <= W.ncols () ; r++) {
	if (v1[r] == 0) {
	    cerr<<"** SERIOUS ERROR: subspace is not invariant!!! ** "<<endl ;
	    cerr<<"To aid in debugging I will not shut down... Have fun."<<endl;
	}
	for (int c = 1 ; c <= W.ncols (); c++)
	    B(r,c) = W_TW(v1[r], c+W.ncols());
    }
}

/* LeftInverse:
   Let B=this.
   We must find some A so that AB = I.
   Form the augmented matrix [B:I].
   If its reduced row echolon form 
   is [I:A] then AB=I.
*/

template<class T>
void Matrix<T>::LeftInverse(Matrix<T>& A) const
{
    Matrix<T> B=*this;
    Matrix<T> BI, I;
    I.idmat(B.nrows());
    B.augment(BI,I);
    Vector<int> elimcol, elimrow;
    BI.SimpleEchelon(elimcol,elimrow);
    A.init(B.ncols(),B.nrows());
    for(int c=1;c<=B.ncols();c++) {
	if(elimcol[c]==0) die("Matrix<T>::LeftInverse error!");
	for(int i=1;i<=B.nrows();i++)
	    A(c,i)=BI(elimcol[c],i+B.ncols());
    }
}

template<class T>
void Matrix<T>::RightInverse(Matrix<T>& A) const
{
    Matrix B=*this;
    B.transpose();
    B.LeftInverse(A);
    A.transpose();
}

template<class T>
void Matrix<T>::Inverse(Matrix<T>& A) const
{
    if(A.ncols()!=A.nrows())
	die("Matrix<T>::Inverse -- nonsquare matrix.");
    LeftInverse(A);
}

template<class T>
bool Matrix<T>::is_invertible() const
{
    return nrows() == ncols() && rank()==nrows();
}

template<class T>
int Matrix<T>::rank() const
{
    Vector<int> v1, v2;
    return rank(v1,v2);
}

template<class T>
int Matrix<T>::rank(Vector<int>& v1, Vector<int>& v2) const
{
    Matrix<T> M(*this);		// make a copy -- we're const so we
				// can't be allowed to change *this.
    M.SimpleEchelon(v1,v2);
    int r=0;
    for(int i=1; i<=v1.size();i++)
	if(v1[i]!=0) r++;
    return r;
}


template<class T> 
void Matrix<T>::iterate(Matrix<T>& Mv, const Vector<T>& _v, int n) const
{
    const Matrix<T>& A=*this;
    Vector<T> v=_v;
#ifdef DEBUG
    if(v.size()!=A.ncols()) {
	cerr << "Matrix<T>::iterate -- v.size()==A.ncols(), but "
	     <<v.size()<<"!="<<A.ncols()<<endl;
	die();
    }
    if(n<0) {
	cerr << "Matrix<T>::iterate -- negative n = " << n << endl;
	die();
    }
#endif
    
    Mv.init(v.size(),n+1);
    for(int i=1;i<=n+1;i++) {
	Mv.setcol(i,v);
	v = A*v;
    }
}

template<class T>
void
Matrix<T>::augment(Matrix<T>& AB, const Matrix<T>& B) const
{
    AB.init(nrows(),ncols()+B.ncols());
    for(int r=1;r<=nrows();r++) {
	for(int c=1;c<=ncols();c++)
	    AB(r,c)=(*this)(r,c);
	for(int c=1;c<=B.ncols();c++)
	    AB(r,c+ncols())=B(r,c);
    }
}

template<class F>
void
MatVecList<F>::reorder(const Vector<int>& perm)
{
  MatVecList<F> copy=*this;

  for (int i=0; i<len(); i++) {
    M(i)=copy.M(perm[i+1]-1);
    v(i)=copy.v(perm[i+1]-1);
  }
}

template<class F>
void 
MatVecList<F>::Save(ofstream& of) const
{
    of<<thesize<<endl<<thelen<<endl;
    for(int i=0;i<thesize;i++) {
	pM[i].Save(of); of<<endl;
	pv[i].Save(of); of<<endl;
    }
}

template<class F>
void 
MatVecList<F>::Load(ifstream& in) 
{
    in>>thesize;
    init(thesize);
    in>>thelen;
    mesg("Loading MatVecList, size=",thelen);mesg("\n[");
    for(int i=0;i<thesize;i++) {
	if(i<thelen) {
	    if(i>0) mesg(" ");
	    mesg(i+1);
	}
	pM[i].Load(in); 
	pv[i].Load(in); 
    }
    mesgln("]");
}





#endif


