/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

// matrix.h

// To instantiate with a completely new type it is necessary to define
// an appropriate find_pivot() function.  See matrix.cc for 
// some examples.

#ifndef _MATRIX_H
#define _MATRIX_H   

#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include "vector.h"
#include "arith.h"

template<class T> class Subspace;

template<class T>
class Matrix {
    friend class Subspace<T>;
public:
    Matrix();
    Matrix(long nr, long nc);
    Matrix(const Matrix&);
    Matrix(const Vector<T>&);
    ~Matrix();
    void init(long nr=0, long nc=0);  //resizes matrix and make entries=0
    void resize(long nr=0, long nc=0);//resizes matrix and copies contents
    void zero();                            // all entries = 0
    void random(int max);	// Set entries to be random integers
				// in the interval [0,max).
				// Uses the stdlib function random, thus
				// srandom can be used to set the seed.

    bool iszero() const;		   // returns true<==> all entries=0.
    void idmat(long n);              // makes matrix equal identity of dim n
    int count() const;   // counts number of non-zero entries;

    // reorder_rows: row i <-- row s[i] for i=1,...,dim. 
    void reorder_rows(const Vector<int>& s);
    // reorder_cols: column i <-- column s[i] for i=1,...,dim. 
    void reorder_cols(const Vector<int>& s);
    // reorder (reorders both rows and columns, i.e., changes
    // basis by the permutation s.)
    void reorder(const Vector<int>& s);

    Matrix& operator=(const Matrix<T>&);
    bool is_invertible() const ;	// returns true iff rank==nr==nc.
    int rank() const ;
    int rank(Vector<int>& v1, Vector<int>& v2) const ;
    const T& operator () (long row,  long col) const;
    T& operator () (long row, long col);
    Matrix slice(long r1, long r2, long c1 = -1, long c2 = -1) const;
    
				// construct the augmented matrix AB = [*this, B]
    void augment(Matrix<T>& AB, const Matrix<T>& B) const;
    
    inline void set(long, long, const T&);
    void setrow(long i, const Vector<T>& v);  // 1 <= i <= nro
    void setcol(long i, const Vector<T>& v);  // 1 <= i <= nco
    void addtocol(long i, const Vector<T>& v);
    void addtorow(long i, const Vector<T>& v);
    void swaprows(long r1, long r2);
    void swapcols(long c1, long c2);
    inline long nrows() const { return nro; }
    inline long ncols() const { return nco; }
    
    Vector<T> row(long i) const;          // returns row i (as a vector)
    void row(long, Vector<T>& V) const;  // assigns row to Vector V
    Vector<T> col(long j) const;          // returns col j (as a vector)
    void col(long, Vector<T>& V) const;  // assigns col to Vector V
    void addscalar(const T& s);           // adds s to diagonals 
    Matrix& operator +=(const Matrix<T>&);
    Matrix& operator -=(const Matrix<T>&);
    Matrix& operator *=(const Matrix<T>&);
    Matrix& operator *=(const T&);
    Matrix& operator /=(const T&);
    
    const T* getMat() const { return mat; }
    
    void transpose(Matrix<T>& T) const;
    void transpose() ;
    T determinant() const;
    T trace() const;
    void charpoly(Vector<T>& poly) const;
    // substitutes M=*this into the polynomial defined by v, thus
    // sets fM equal to f[1]+f[2]M+f[3]M^2+...+f[n]M^(n-1).
    void polsubst(Matrix<T>& fM, const Vector<T>& f) const;
    // Compute the sequence of iterates 
    //         v, M*v, M^2*v, M^3*v, M^4*v, ..., M^n*v
    // These are stored as the columns of the matrix Mv, with the
    // first column, on the left, being v. 
    void iterate(Matrix<T>& Mv, const Vector<T>& v, int n) const;
    void elim(long r1, long r2, long col);
    void elim2(long r1, long r2, long col);
    void echelon(Vector<int>&, Vector<int>&, long&, long&);
    void SimpleEchelon(Vector<int>&, Vector<int>&);
				// compute action of *this on W.
    void RestrictedTo(Matrix<T>& B, const Matrix<T>& W) const ;
    void LeftInverse(Matrix<T>& A) const ;  // T=this,  AT=I
    void RightInverse(Matrix<T>& A) const ; // TA = I.
    void Inverse(Matrix<T>& A) const;
    
    friend Matrix operator*(const Matrix<T>&, const Matrix<T>&);
    friend Matrix operator*(const T&, const Matrix<T>&);
    friend Vector<T> operator*(const Matrix<T>&, const Vector<T>&);  
    friend Matrix operator^(const Matrix<T>&, const int) ;
    friend bool operator==(const Matrix<T>&, const Matrix<T>&);
    friend bool operator!=(const Matrix<T>&, const Matrix<T>&);
    friend T TraceOfProduct(const Matrix<T>&, const Matrix<T>&);
    friend ostream& operator<< (ostream&, const Matrix<T>&);
    friend istream& operator>> (istream&, Matrix<T>&);
    void error(const char* msg) const;     // error handling
    void Load(ifstream& in) ;
    void Save(ofstream& of) const ;
    void LoadSparse(ifstream& in) ; // at the moment -- only for bigrational
    void SaveSparse(ofstream& of) const ;	// only for bigrational now.
    void WriteToFile( const char* fileName, const char* MatrixName=0);
    void AppendToFile( const char* fileName, const char* MatrixName=0);
private:
    long nro;        // number of rows
    long nco;        // number of columns
    T * mat;         // entries stored in one array, by rows
    void writeMatrix(ofstream& outFile, const char* MatrixName);
};


template<class F>
class MatVecList {
public:
    MatVecList(int _size=1) { allocate(_size); }
    void init(int newsize) { delete[] pM; delete[] pv; allocate(newsize); }
    void resize(int newsize) {
	Matrix<F>* oldpM=pM;
	Vector<F>* oldpv=pv;
	pM=NULL; pv=NULL;
	int oldsize=thesize;
	init(newsize);
	for(int i=0; i<oldsize; i++)
	    { pM[i]=oldpM[i]; pv[i]=oldpv[i]; }
	delete[] oldpM;
	delete[] oldpv;
    }
    MatVecList(const MatVecList<F>& list) {
	allocate(list.thesize);
	for(int i=0; i<size(); i++) { M(i)=list.pM[i]; v(i)=list.pv[i]; }
    }
    MatVecList& operator=(const MatVecList<F>& list) {
	init(list.thesize);
	for(int i=0; i<size(); i++) { M(i)=list.pM[i]; v(i)=list.pv[i]; }
	return *this;
    }
    ~MatVecList() { delete[] pM; delete[] pv; }
    Matrix<F>& M(int i) { validate(i); return pM[i]; }
    Vector<F>& v(int i) { validate(i); return pv[i]; }
    const Matrix<F>& cM(int i) const { constvalidate(i); return pM[i]; }
    const Vector<F>& cv(int i) const { constvalidate(i); return pv[i]; }
    int size() const { return thesize; }
    int len() const { return thelen; }	 // largest array element set. 
    void reorder(const Vector<int>& perm) ;
    void Save(ofstream& of) const;
    void Load(ifstream& in);
private:
    int thesize, thelen;
    Matrix<F>* pM;
    Vector<F>* pv;
    void validate(int i) {
	constvalidate(i);
	if(i>=thelen) thelen=i+1;
    }
    void constvalidate(int i) const {
	if(!(0<=i && i<size())) {
	    cerr<<"MatVecList: assert(0<="<<i<<"<"
		<<size()<<") failed"<<endl; die(); 
	}
    }
    void allocate(int _size) {
	thesize=_size>=1?_size:1;
	thelen=0;
	pM=new Matrix<F>[thesize];
	pv=new Vector<F>[thesize];
    }
};


//////////////////////////////////////////////////////////////

template <class T> inline
Matrix<T>::Matrix( ) { nro=0; nco=0; mat=0; }

template <class T> inline
Matrix<T>::Matrix(long nr, long nc) 
{
    nro=nr;
    nco=nc;
    mat = new T[nr*nc];     
}

template <class T> inline Matrix<T>::~Matrix()
{
    if (mat) delete[] mat;
}

template <class T> inline 
void Matrix<T>::zero()
{
    T *m = mat;
    for( long n = nco*nro, j = 0; j < n; j++ ) *m++ = 0;
}

template <class T> inline 
void Matrix<T>::idmat(long n) 
{
    init(n,n);
    for( long j = 0; j < n; j++ ) mat[j*nco+j] = 1;
}

template <class T> inline 
void Matrix<T>::set(long i, long j, const T& elt)
{
    if( 1<=i && i<=nro && 1<=j && j<=nco ) mat[(i-1)*nco+j-1 ] = elt;
    else error("Argument out of bounds in function set(long,long,T&)\n");
}

template <class T> inline T& Matrix<T>::operator()(long i, long j) 
{
    if(0<i && i<=nro && 0<j && j<=nco)
	return mat[(i-1)*nco+(j-1)];
    else {
	cerr << "Matrix error.  M is a " << nro << "x" << nco << " matrix"
	    << endl << "but the program has attempted to access the " 
	    << i << "," << j << " entry." << endl;
	die();
    }
}

template <class T> inline 
const T& Matrix<T>::operator()(long i, long j) const
{
    if(0<i && i<=nro && 0<j && j<=nco)
	return mat[(i-1)*nco+(j-1)];
    else {
	cerr << "Matrix error.  M is a " << nro << "x" << nco << " matrix"
	     << endl << "but the program has attempted to access the " 
	     << i << "," << j << " entry." << endl;
	die();
   }
}

template <class T> inline void Matrix<T>::addscalar( const T& c )
{
    for(int i=0; i<nro; i++)
	mat[i*nco+i] += c;
}

template <class T> inline long nrows( const Matrix<T>& M )
{
    return M.nrows();
}

template <class T> inline long ncols( const Matrix<T>& M ) 
{ 
    return M.ncols(); 
}

template <class T> inline 
Matrix<T>& Matrix<T>::operator *= ( const Matrix<T>& M )
{
    *this = (*this) * M;
    return *this;
}

template <class T> inline
bool operator != ( const Matrix<T>& M1,  const Matrix<T>& M2)
{
    return !(M1 == M2);
}

template <class T> inline
Matrix<T> operator+(const Matrix<T>& m)
{ 
    return m; 
}

template <class T> inline
Matrix<T> operator-(const Matrix<T>& m)
{ 
    return (-1)*m; 
}

template <class T> inline
Matrix<T> operator+(const Matrix<T>& m1, const Matrix<T>& m2)
{
    Matrix<T> ans(m1); ans+=m2;  return ans;
}

template <class T> inline
Matrix<T> operator-(const Matrix<T>& m1, const Matrix<T>& m2) 
{
    Matrix<T> ans(m1); ans-=m2;  return ans;
}

template <class T> inline
Matrix<T> operator*(const T& scalar, const Matrix<T>& m)
{
    Matrix<T> ans(m); ans*=scalar;  return ans;
}

template <class T> inline
Matrix<T> operator^(const Matrix<T>& m, const int _n)
{
    int n = _n ;
    if (m.nro != m.nco) {
	cerr << "Matrix::operator^(): matrix must be square." << endl ;
	exit (-1) ;
    }
    
    if (n < 0) {
	cerr << "Matrix::operator^(): class inverse not yet implemented." << endl ;
	exit (-1) ;
    }
    
    if (n==1) return m ;
    
    Matrix<T> ans ; 
    ans.idmat (m.nro) ;
    Matrix<T> tmp (m) ;
    
    while (n) {
	if (n%2) 
	    ans *= tmp ;
	n /= 2 ;
	tmp = tmp * tmp ;
    }
    
    return ans;
}


template <class T> inline
Matrix<T> operator/(const Matrix<T>& m, const T& scalar)
{
    Matrix<T> ans(m); ans/=scalar;  return ans;
}

/*************************************************************************
 * Notes:
 *
 * Constructors here DON'T initialize entries of Matrix with zero.
 * You must use function Matrix<T>::zero() to acomplish that.
 *
 * When a function should return a Matrix or a Vector I try to avoid 
 * the wasteful copy in the return statement by passing the Matrix/Vector
 * as an argument to the function as in function 
 * Matrix<T>::transpose( Matrix<T> TP ), which does TP = transpose of *this.
 * Such function will always check that the Matrix/Vector passes are
 * of the right size and will resize them if necessary.
 * Named returns would be a nice alternative, although GCC especific, but
 * at the present version of GCC, named returns don't work for templates.
 *
 * function Matrix<T>::charpoly passes the resulting characteristic 
 * polynomial as a vector, where the coefficients are in ascending 
 * order. 
 *
 **********************************************************************/
#endif

