/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

// methodofgraphs.cc - Mestre-Oesterle graph method

#include "display.h"
#include <LiDIA/factorization.h>
#include "methodofgraphs.h"
#include "system.h"

const int MAXP=7;
 
MethodOfGraphs::MethodOfGraphs(const int& _N, bool save)
  : N(_N), do_save(save)
{
#ifdef DEBUG
  if (!is_prime(N)) error ("MethodOfGraphs(): N must be prime.");
#endif
  basis_size=0;
  if(!loadbasis())
    InitFieldFinite(N,2);
  init_coeff();
  weights.init(0);
}

MethodOfGraphs::~MethodOfGraphs()
{
  delete[] C2;
  delete[] C3;
  delete[] C5;
  delete[] C7;
}

const Vector<ffield_element> &
MethodOfGraphs::ssbasis() 
{
  if(basis.size()==0) {
    Matrix<bigrational> T; 
    Tp(T,2);			// computing Tp defines basis. 
  }
  return basis;
}

int 
MethodOfGraphs::max_p() const
{
  return MAXP;
}

void 
MethodOfGraphs::status()
{
  cout << endl<<endl;
  cout << "\t==============================================" << endl 
       << "\t         Method of Graphs, N = " << N << endl 
       << "\t    [";
  if(matrix) cout<<"X"; else cout<<" ";
  cout<<"] matrix          [";
  if(charpoly) cout<<"X"; else cout<<" ";
  cout<<"] charpoly"<<endl;
  cout << "\t.............................................."<<endl
       << "\t    c = toggle charpoly,    m = toggle matrix."<<endl
       << "\t    2,3,5,7 = compute that T_p"<<endl
       << "\t    s = supersingular basis"<<endl
       << "\t=========================== x = exit =========" << endl 
       << endl;
}

void 
MethodOfGraphs::interface() 
{
  matrix=false; charpoly=true;
  status();
  cout << "Type ? for help."<<endl;
  while(1) {
    char buf[256];
    cout << "(Graphs: N="<<N<<") ";
    get_str(buf,256,"? ");
    int p=atoi(buf);
    if(buf[0]=='?') { 
      status();
      continue;
    }
    if(buf[0]=='s') { 
      cout << ssbasis() << endl;
      cout << "0 = " << ffield_element::get_def_base() ;
      continue; 
    }
    if(buf[0]=='m') {
      if(matrix)matrix=false;else matrix=true; continue;
    }
    if(buf[0]=='c') {
      if(charpoly)charpoly=false;else charpoly=true; continue;
    }
    if(buf[0]=='x' || buf[0]=='q' || (buf[0]=='\\' && buf[1]=='q')) break;
    if(buf[0]=='t') {
	cout << "To compute T_p, simply enter p."<<endl;
	continue;
    }
    if(is_prime(p) && p<=max_p()) {
      Matrix<bigrational> T;
      Tp(T,p);
      cout << endl << endl;
      if(matrix) {
	  cout << "With respect to supersingular basis:"<<endl;
	  cout << "TM" << p << " = " << T <<endl; 
	  cout << "With respect to divisors of degree 0:"<<endl;
	  Matrix<bigrational> X(T.nrows(),T.ncols()-1);
	  for(int c=1;c<=X.ncols();c++) {
	      X(1,c)=1; X(c+1,c)=-1;
	  }
	  Matrix<bigrational> TX;
	  T.RestrictedTo(TX, X);
	  cout << endl << "TX" << p << " = " << TX << endl;
      }
      if(charpoly) {
	cout << "charpoly(T) = ";
	DisplayCharPoly(T); cout <<endl;
      }
    } else
      cout << "I don't know how to compute T_{"<<p<<"}"<<endl;
  }
  cout << "Bye"<<endl;
}

void 
MethodOfGraphs::error(const char* msg) const
{
#ifdef DEBUG
  cerr<<"MethodOfGraphs -- "<<msg<<endl;
  exit(-1);
#endif
}

void
MethodOfGraphs::pisog(Vector<ffield_element>& list,
		      const ffield_element& j,
		      const int& p) const 
{
  list.init(1);
  polynomial<gf_p_element> f;
  isogpoly(f, j, p);
  list.init (f.degree());
  int n=1;

  // make list=roots of f, counted with multiplicity.
  factorization<polynomial<gf_p_element> > fac
    = berlekamp (f);

  for(int i=0; i<fac.no_of_prime_components(); i++) 
    for (int m=0; m<fac.prime_exponent(i); m++)
      list[n++]=-fac.prime_base(i).base()[0];

#ifdef DEBUG
  if (n <= f.degree()) error("pisog: error in programming.");
#endif
}
 
 
void 
MethodOfGraphs::isogpoly(polynomial<gf_p_element>& f,
			 const ffield_element& j,
			 const int& p) const 
{
  // 1. compute the powers j,j^2,...,j^{p+1} of j.
  Vector<ffield_element> v(p+2);
  v[1]=1;
  for(int i=2;i<=p+2;i++) v[i]=j*v[i-1];
  
  // 2. dot with coeff vectors to obtain the coeffs of f.
  f.set_base(ffield_element::get_def_base());
  f.set_degree(p+1);
  f[p+1]=ffield_element(1);
  for(int i=0;i<=p;i++)
    f[i]=v*coeff(p)[i];		// dot product
}

Vector<ffield_element>* 
MethodOfGraphs::coeff(const int& p) const
{
  switch(p) {
  case 2: return C2; 
  case 3: return C3;
  case 5: return C5;
  case 7: return C7;
#ifdef DEBUG
  default:
    error("coeff vector not programmed.");
#endif
  }
}

void
MethodOfGraphs::init_coeff() 
{
  // create C2, C3, C5, C7.
  // the following pari code was useful in producing these tables. 
  //conv(phi)=for(i=0,poldegree(phi)-1,print1("{");for(j=0,poldegree(phi),print1("\"",polcoeff(polcoeff(phi,i),j),"\""); if(j<poldegree(phi),print1(","))); print("},"))
  char* sC2[3][4]={{"-157464000000000","8748000000","-162000","1"},
		   {"8748000000", "40773375", "1488", "0"},
		   {"-162000","1488","-1","0"}};
  char* sC3[4][5]={{"0","1855425871872000000000","452984832000000","36864000","1"},
{"1855425871872000000000","-770845966336000000","8900222976000","-1069956","0"},
{"452984832000000","8900222976000","2587918086","2232","0"},
{"36864000","-1069956","2232","-1","0"}};

  char* sC5[6][7]={{"141359947154721358697753474691071362751004672000","53274330803424425450420160273356509151232000","6692500042627997708487149415015068467200","280244777828439527804321565297868800","1284733132841424456253440","1963211489280","1"},
{"53274330803424425450420160273356509151232000","-264073457076620596259715790247978782949376","36554736583949629295706472332656640000","-192457934618928299655108231168000","128541798906828816384000","-246683410950","0"},
{"6692500042627997708487149415015068467200","36554736583949629295706472332656640000","5110941777552418083110765199360000","26898488858380731577417728000","383083609779811215375","2028551200","0"},
{"280244777828439527804321565297868800","-192457934618928299655108231168000","26898488858380731577417728000","-441206965512914835246100","107878928185336800","-4550940","0"},
{"1284733132841424456253440","128541798906828816384000","383083609779811215375","107878928185336800","1665999364600","3720","0"},
{"1963211489280","-246683410950","2028551200","-4550940","3720","-1","0"}};

char* sC7[8][9]={{"0","0","1464765079488386840337633731737402825128271675392000000000000000000","13483958224762213714698012883865296529472356352000000000000000","41375720005635744770247248526572116368162816000000000000","42320664241971721884753245384947305283584000000000","3643255017844740441130401792000000","104545516658688000","1"},
{"0","1221349308261453750252370983314569119494710493184000000000000000000","-838538082798149465723818021032241603179964268544000000000000000","-129686683986501811181602978946723823397619367936000000000000","553293497305121712634517214392820316998991872000000000","-40689839325168186578698294668599003971584000000","1038063543615451121419229773824000","-34993297342013192","0"},
{"1464765079488386840337633731737402825128271675392000000000000000000","-838538082798149465723818021032241603179964268544000000000000000","-46666007311089950798495647194817495401448341504000000000000","72269669689202948469186346100000679630099972096000000000","308718989330868920558541707287296140145328128000000","11269804827778129625111322263056523132928000","10685207605419433304631062899228","720168419610864","0"},
{"13483958224762213714698012883865296529472356352000000000000000","-129686683986501811181602978946723823397619367936000000000000","72269669689202948469186346100000679630099972096000000000","-5397554444336630396660447092290576395211374592000000","17972351380696034759035751584170427941396480000","-901645312135695263877115693740562092344","16125487429368412743622133040","-4079701128594","0"},
{"41375720005635744770247248526572116368162816000000000000","553293497305121712634517214392820316998991872000000000","308718989330868920558541707287296140145328128000000","17972351380696034759035751584170427941396480000","88037255060655710247136461896264828390470","14066810691825882583305340438456800","4460942463213898353207432","9437674400","0"},
{"42320664241971721884753245384947305283584000000000","-40689839325168186578698294668599003971584000000","11269804827778129625111322263056523132928000","-901645312135695263877115693740562092344","14066810691825882583305340438456800","-18300817137706889881369818348","177089350028475373552","-10246068","0"},
{"3643255017844740441130401792000000","1038063543615451121419229773824000","10685207605419433304631062899228","16125487429368412743622133040","4460942463213898353207432","177089350028475373552","312598931380281","5208","0"},
{"104545516658688000","-34993297342013192","720168419610864","-4079701128594","9437674400","-10246068","5208","-1","0"}};

  C2 = new (Vector<ffield_element>)[3];
  for(int i=0;i<=2;i++) {
    C2[i].init(4);
    for(int j=1;j<=4;j++) C2[i][j]=ffield_element(sC2[i][j-1]);
  }

  C3 = new (Vector<ffield_element>)[4];
  for(int i=0;i<=3;i++) {
    C3[i].init(5);
    for(int j=1;j<=5;j++) C3[i][j]=ffield_element(sC3[i][j-1]);
  }
  C5 = new (Vector<ffield_element>)[6];
  for(int i=0;i<=5;i++) {
    C5[i].init(7);
    for(int j=1;j<=7;j++) C5[i][j]=ffield_element(sC5[i][j-1]);
  }
  C7 = new (Vector<ffield_element>)[8];
  for(int i=0;i<=7;i++) {
    C7[i].init(9);
    for(int j=1;j<=9;j++) C7[i][j]=ffield_element(sC7[i][j-1]);
  }
}

void 
MethodOfGraphs::Tp(Matrix<bigrational>& T, const int& p) 
{
#ifdef DEBUG
  if(!is_prime(p)||p>7) error("Tp: Unsupported p, not prime or > 7.");
#endif
  
  if(load(T,p)) return ;
  // Compute the dimension of M_2. 
  int dim = count_supersingular_j() ;

  // basis will eventually contain all supersingular j's
  bool basis_is_new=false;
  if(basis_size==0) {
    basis.init(dim);
    basis_size = 1 ;
    basis_is_new=true;
    basis[1]=ssj();
  }

  // M will contain a matrix representing the action
  // of the hecke operator T_2 on M_2(Gamma_0(N)). 
  T.init(dim,dim);
  
  // Initialize our basis (so far) with the initial 
  // supersingular j-invariant.  We'll find a basis
  // for all of M_2(N) by hitting what we've got so 
  // far by T_p.  
  // (For connectedness of this graph, see the 
  // discussion on page 223 of Mestre.) 
  
  // For each basis vector, compute the column of the matrix of T_p.

  for(int i=1; i<=dim; i++) {
    //    cout << "T_{"<<p<<"}("<<basis[i]<<") = " << endl ;
      if (i%(dim/15+1)==0) { mesg(i*100/dim); mesg("% "); }
       
    // Compute all curves p-isogeneous to basis[i], with multiplicity.
       
    Vector<ffield_element> v;
    pisog(v, basis[i], p) ; 

    // For each element of v:
    //  1.  If we haven't seen it before add it to our basis,
    //      otherwise find it in our basis.
    //  2.  Add it to the column of the matrix M of T_2 
    //      which we are constructing.
       
    for(int j=1; j<=v.size(); j++) {
      int pos=basis.find(v[j]);
      if(pos>0 && pos<=basis_size)			// found it
	++T(pos,i);
      else {
	if (basis_size+1>dim) {
	    cerr << "v[j] = " << v[j] << endl;
	    cerr << "basis = " << basis << endl;
	    error("Tp: bug in programming, j not found.");
	}
	basis[++basis_size]=v[j];
	++T(basis_size,i);
      }
    }
  }
  if(basis_is_new) {		// sort basis, change rep of T, and save.
      Vector<int> s;
      sortbasis(s);
      T.reorder(s);
      savebasis();
  }
  save (T, p);
}

void
MethodOfGraphs::sortbasis(Vector<int>& s)
{
    mesgln("sortbasis");
    int n=basis.size();
    s.init(n);

    // 1. separate out the Fp-rational j's.
    Array<bigint> rat(basis.size());// note arrays start at 0.
    int j=1;
    for(int i=1;i<=n;i++) {
	ffield_element x=basis[i];

	Fp_polynomial f=x.rep();
	if(f.degree()<1)   // Fp rational j.
	    if(f.degree()<0)
		rat(j++)=0;
	    else {
		rat(j++)=f[0];
	    }
    }
    rat.resize(j-1);
	
    // 2. sort them using the Array quicksort function.
    rat.qsort();

    Vector<ffield_element> new_basis(n);
    for(j=1; j<=rat.size(); j++) 
	new_basis[j]=rat(j);
        
    // 3. list each nonrational pair j, j^p
    bool fin[n+1]; for(int i=1;i<=n;i++) fin[i]=false;
    for(int i=1;i<=n;i++) 
	if(basis[i].rep().degree()>=1 && !fin[i]) {// non Fp rational j.
	    new_basis[j]=basis[i];
	    j++;
	    power(new_basis[j], basis[i],N);
	    for(int k=i+1;k<=n;k++) //i+1 should be the match, so this is quick
		if(basis[k]==new_basis[j]) {
		    fin[k]=true;
		    break;
		}
	    j++;
	}
    
    // 4. compute the permutation vector s.
    //    (This is O(n^2) and probably rather slow.)
    for(int i=1;i<=n;i++) 
	s[i]=basis.find(new_basis[i]);

    if(j!=n+1) 
	die("MethodOfGraphs::sortbasis -- bug in programming.");
    basis=new_basis;
}

ffield_element 
MethodOfGraphs::ssj() const 
{
  if(jacobi(N-1,N) == -1) return 1728;
  if(jacobi(N-2,N) == -1) return 8000;
  if(jacobi(N-3,N) == -1) return 0;
  if(jacobi(N-7,N) == -1) return 16581375;    // 255^3 
  if(jacobi(N-11,N) == -1) return -32768;     // (-32)^3
  if(jacobi(N-19,N) == -1) return 884736;
  if(jacobi(N-43,N)== -1) {
    bigint n; string_to_bigint("-884736000",n);
    return n;
  }
  if(jacobi(N-67,N) == -1) {                 // (-5280)^3
    bigint n; string_to_bigint("-147197952000",n);
    return n;
  }
  if(jacobi(N-163,N) == -1) {
    bigint n; string_to_bigint("-262537412640768000",n);
    return n;
  }
  if(N==15073) return 5797;
  if(N==18313) return 10629;
  if(N==38833) return 3235;
  if(N==69337) return 2840;
  if(N==71593) return 33119;
#ifdef DEBUG
  cerr<<"There is no s.s. j-invariant in characteristic "<<N
      <<" known to this ssj()."<<endl;
  error("Terminating.");
#endif
}

int 
MethodOfGraphs::count_supersingular_j() const
{
  int n=floor(N,12);
  switch (N%12) {
  case 5: case 7: n++; break;
  case 11: n+=2; 
  }
  return n;
}


bool 
MethodOfGraphs::loadbasis() 
{
  if(!do_save) return false;
  ifstream in(data_read("methodofgraphs_basis",N));
  if(!in) return false;

  ffield_element::load_default_base(in);
    
  in>>basis_size;
#ifdef DEBUG
  if(basis_size !=count_supersingular_j())
    error("load: Error in basis datafile");
#endif 
  basis.init(basis_size);
  for(int i=1; i<=basis_size; i++) 
    basis[i] = ffield_element(in);
//  cerr << "supersingular basis:" << endl << basis << endl;
  return true;
}

void
MethodOfGraphs::savebasis() const
{
  if(!do_save) return;
#ifdef DEBUG
  if(basis_size==0) error("savebasis: basis doesn't exist yet.");
#endif
  ofstream out(data_write("methodofgraphs_basis",N));
  if(out.bad()) error("Could not save basis.");

  ffield_element::save_default_base(out);
  out<<endl<<basis_size<<endl;
  for(int i=1; i<=basis_size; i++) {
    basis[i].save(out);
    out<<endl;
  }
}

bool 
MethodOfGraphs::load(Matrix<bigrational>& T, const int& p) 
{
  if(!do_save) return false;
  if (basis_size==0) return false;
  char buf[512];
  sprintf(buf,"methodofgraphs_%d",p);
  ifstream in(data_read(buf,N));
  if(!in) return false;
  mesg("Loading T_",p); mesg("\n");

  // load the matrix for T. 
  T.LoadSparse(in);
  return true;
}
  
void 
MethodOfGraphs::save(const Matrix<bigrational>& T, const int& p) 
{
  if(!do_save) return;
  char buf[512];
  sprintf(buf,"methodofgraphs_%d",p);
  ofstream out(data_write(buf,N));
  if(!out) error("save: couldn't open output file.");
  T.SaveSparse(out);
}

void 
MethodOfGraphs::coker_mx(bigint& image, bigint& mx,
			 const IntVecList<bigrational>& I)
{
    // The following code is written more to be readable than efficient...

    image=mx=0;			// return values, in case of an error.

    // Step 1: define X = degree 0 divisors
    Matrix<bigrational> X; X.init(dim(),dim()-1);
    for(int i=1;i<dim();i++) {
	X(1,i)=1; X(i+1,i)=-1;
    }

    // Step 2: cut out X[I]
    Matrix<bigrational> XI=X;
    if(I.len()>=1) {
	int goaldim=I.cv(0).poldeg();
	int i=0;
	while (XI.ncols()>goaldim) {
	    if(i>=I.len()) return; // error, not enough elements of I.
	    int p=to_int(I.cZ(i));
	    if(p>max_p()) return; // can't compute enough Tp.
	    Matrix<bigrational> T, TXI, K;
	    Tp(T,p);		// T <-- T_p on full space
	    T.RestrictedTo(TXI, XI);   // TXI <-- T|current XI
	    IntegerKernel(K, TXI, I.cv(i));
	    XI *= K;		// XI <-- ker(f(T_p)|XI)
	    i++;
	};
    };

    // Step 3: compute monodromy weights 
    const Vector<int>& w = get_weights();
    // Let W be the diagonal matrix whose entries are the weights.
    Matrix<bigrational> WXI(XI.nrows(),XI.ncols());
    // W*XI is got by multiplying row i of XI by weight w[i].
    for(int i=1; i<=w.size(); i++) {
	Vector<bigrational> v=XI.row(i);
	v *= w[i];
	WXI.setrow(i,v);
    }
    
    // Step 4: compute d=disc(X[I]xX[I]-->Z).
    // This is det(transpose(XI)*WXI).
    Matrix<bigrational> tXI; XI.transpose(tXI);
    Matrix<bigrational> D=tXI*WXI;
    bigint d=abs(D.determinant()).numerator();

    // Step 5: compute c=coker(X-->Hom(X[I],Z))
    X.transpose();
    // The rows of tx*W*XI give generators for image
    // of X in Hom(X[I],Z).
    Matrix<bigrational> tXWXI=X*WXI; 
    tXWXI.transpose();		// now columnes generate image
    bigint c=LatticeVol(tXWXI).numerator();

    // Step 6: image = c, mx = d/c.
    image=c; mx=d/c;
}

const Vector<int>&            
MethodOfGraphs::get_weights()
{
    if(weights.size()==0) {	// initialize weight vector
	weights.init(dim());
	Matrix<bigrational> T;	// T <-- T_2 - 3.
	Tp(T,2);
	for(int i=1;i<=dim();i++) T(i,i)-=3;
	Matrix<bigrational> E;	// E <-- Normalized Eisenstein series
	IntegerKernel(E, T);
	if(E.ncols()!=1)
	    die("MethodOfGraphs::get_weights -- where is E????");
	bigint n=E(1,1).numerator(); // no denom, since an int kernel.
	for(int i=2;i<=dim();i++)
	    n=lcm(n,E(i,1).numerator());
	E/=n;
	for(int i=1;i<=dim();i++)
	    weights[i]=to_int((1/E(i,1)).numerator());
    }
    if(weights[1]<0) weights*=-1;
    return weights;
}

