/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

// methodofgraphs.h  - Mestre-Oesterle graph method

#ifndef METHODOFGRAPHS_H
#define METHODOFGRAPHS_H

#include <LiDIA/bigrational.h>
#include "vector.h"
#include "matrix.h"
#include "fields.h"

class MethodOfGraphs {
public:
    MethodOfGraphs(const int& N, bool save=true);	// N must be prime.
    ~MethodOfGraphs();
    const Vector<ffield_element>& ssbasis();
    const Vector<int>&            get_weights();
    
				// 2 <= p <= 7.
    void  Tp(Matrix<bigrational>& T, const int& p);
    int   dim(){ return count_supersingular_j(); }
    int   max_p()const;		// returns max p such that
				// we know how to compute Tp.

    // coker_mx:
    // compute the invariants #image(Phi_J-->Phi_A) and the
    // character group version of the modular degree: mx=[alp(X):alp(X[I])]
    // where alp : X --> Hom(X[I],Z) is the natural map and A=J/IJ.
    // I should be the annihilator of an eigenform of level N.
    // I is a list of pairs, a prime l and a poly h(X) of degree = dim A
    // such that h(T_l)|A = 0.
    // If it is not possible to cut out a subspace of X of dim = dim A,
    // both image and mx are set to 0.  
    void  coker_mx(bigint& image, bigint& mx,
		   const IntVecList<bigrational>& I);

    
    void  interface();		// interactive user interface.
    
private:
    // sortbasis: 
    // reorder the supersingular basis as follows and store the
    // permutation in s (new basis = [old_s[1], old_s[2], old_s[3], ...])
    // It is assumed that the basis has been defined!
    // The ssbasis is a list of elements of Fp^2 written in
    // terms of a root x of a quadratic polynomial:
    //     a+bx, c+dx, ...
    // The ordering is:
    //    First the elements of Fp, from smallest (=0) to largest (=p-1).
    //    Second the quadratic elements with conjugates together,
    //    but in no particular order.
    // NOTE: the basis member *is* changed by this function, and this
    // function is called the first time a T_p is computed. 
    void  sortbasis(Vector<int>& s);
    void  status();		// used by interface.

    Vector<ffield_element> basis;
    Vector<int> weights;
    int basis_size;
    int N;
    bool matrix, charpoly;
    bool do_save;
    void error(const char* msg) const;
    // helper functions
				// creates list(with multiplicities)
				// of all j's p-isogeneous to j.
    void pisog(Vector<ffield_element>& list,
	       const ffield_element& j, const int& p) const ;
    void isogpoly(polynomial<gf_p_element>& f,
		  const ffield_element& j, const int& p) const ;
				// returns some s.s. j-inv. in
				// characteristic N.  only works
				// for N<=100000.
    ffield_element ssj() const ;
    int count_supersingular_j() const ;
    Vector<ffield_element>* coeff(const int& p) const;
    void init_coeff() ;
    
    Vector<ffield_element> *C2, *C3, *C5, *C7;
    
    bool loadbasis();
    void savebasis() const;
    bool load(Matrix<bigrational>& T, const int& p) ;
    void save(const Matrix<bigrational>& T, const int& p) ;
};




#endif //METHODOFGRAPHS_H
