/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

// modformulas.cc  - Modular formulas

#include <LiDIA/nmbrthry_functions.h>
#include "modformulas.h"
#include "system.h"

////////////////////// HIJIKATA TRACE FORMULA ///////////////
int
ModularFormulas::dimSk (Character& _chi, const int& _k) 
{
  InitFieldAlgnumCyclo (lcm(_chi.degree(),2));
  chi=&_chi; k=_k; N=_chi.modulus();

  ::mesg("Computing dim S_{",k);
  ::mesg("}(",N); ::mesg(",chi)...");

  fac = rational_factorization(N);
  fac.factor();
#ifdef DEBUG
  if (k<=1) { error("dimSk: Weight <=1 not yet programmed."); return 0 ; }
#endif
  if (_chi.of(-1) != (k%2?-1:1)) return 0;
  algnum x = idxG0(N)*algnum(k-1)/algnum(12)-hij_abc()
    + ((k==2&&_chi.degree()==1)?1:0) ;
#ifdef DEBUG
  if (!x.is_bigint()) {
    error("dimSk: PROGRAMMING ERROR, dimension not an integer."); 
  }
#endif 
  bigint a; x.bigintify(a); 
#ifdef DEBUG
  if (!is_int(a)) {
    error("dimSk: OVERFLOW, dimension bigger than machine type int."); 
  }
#endif 
  int b; a.intify(b);
  mesg(" finished.\n");
  return b;
}

algnum 
ModularFormulas::hij_a (int s) 
{
  switch(s) {
  case 0:
    switch(k%4) {
    case 0: return algnum(-1)/algnum(2);
    case 2: return algnum(1)/algnum(2);
    default: return 0;
    }
  case -1:
    switch(k%3) {
    case 1: return 0;
    case 2: return algnum(1)/algnum(2);
    default: return algnum(-1)/algnum(2);
    }
  case 1:
    switch(k%6) {
    case 1: case 4: return 0;
    case 2: case 3: return algnum(1)/algnum(2);
    case 0: case 5: return algnum(-1)/algnum(2);
    }
  case -2:
    return algnum((k%2)?-1:1)/algnum(4*N);
  case 2:
    return algnum(1)/algnum(4*N);
#ifdef DEBUG
  default:
    cerr << "ModularFormulas::hij_a -- PROGRAMMING ERROR. s = "<<s<<endl;
    exit(-1);
#endif
  }
}

algnum 
ModularFormulas::hij_b (int s) 
{
  switch(s) {
  case 2: case -2: return 1;
  case 1: case -1: return algnum(1)/algnum(3);
  case 0: return algnum(1)/algnum(2);
#ifdef DEBUG
  default:
    cerr << "ModularFormulas::hij_b -- PROGRAMMING ERROR. s = "<<s<<endl;
    exit(-1);
#endif
  }
}

algnum 
ModularFormulas::hij_csfp (int s, int f, int p) 
{
  int dummy;
  int rho=ordp(p,f,dummy), nu=ordp(p,N,dummy);
  switch(s) {
  case 0:
    if (p==2) 
      if (nu==1 && rho==0)
	return 1;
      else
	return 0;
    else
      if (p%4==3)
	return 0;
      else
	if (rho>0)
	  return 0;
	else {
	  bigint a=sqrtmod(-1,p,nu);
	  return algnum(chi->ofp(p,a))+algnum(chi->ofp(p,-a));
	}

  case 1: case -1:
    if (p==2) 
      return 0;
    else
      if (rho>0 || jacobi(-3,p)==-1 || (p==3 && nu>1))
	return 0;
      else {
	bigint z=sqrtmod(-3,p,nu);
	bigint pnu; power(pnu,p,nu);
	bigint w=invmod(2,pnu);
	return algnum(chi->ofp(p,s*(1+z)*w)) +
	  algnum(((p!=3)?chi->ofp(p,s*(1-z)*w):0)) ;
      }
	
  case 2: case -2:
    algnum ans(0);
    int x;			// x=p^floor(nu/2).	
    power(x,p,floor(nu,2)); 
#ifdef DEBUG 
    if (x<1) error("Error in csfp: programming error.");
#endif 
    for(int i=1; i<=x; i++) {
      int y; power(y,p,ceil(nu,2)+rho);
      ans+=chi->ofp(p,(2/s)+i*y);
    }
    power(x,p,floor(nu-1,2));
#ifdef DEBUG 
    if (x<1) error("Error in csfp: programming error.");
#endif 
    for(int i=1; i<=x; i++) {
      int y; power(y,p,ceil(nu+1,2)+rho);
      ans+=chi->ofp(p,(2/s)+i*y);
    }      
    return ans;
  }
}

algnum 
ModularFormulas::hij_c (int s, int f) 
{
  algnum ans=1;
  for(int i=0;i<fac.no_of_comp();i++) {
    int p; fac.base(i).intify(p);
    ans *= hij_csfp(s,f,p);
  }
  return ans;
}

algnum 
ModularFormulas::hij_bc (int s, int f) 
{
  return hij_b(s) * hij_c(s,f);
}

algnum 
ModularFormulas::hij_abc () 
{
  algnum x(0);
  for(int f=1; f<=N; f++) x+=hij_bc(-2,f);
  algnum y(0);
  for(int f=1; f<=N; f++) y+=hij_bc(2,f);
  return hij_a(0) * hij_bc(0,1)
      +  hij_a(-1) * hij_bc(-1,1)
      +  hij_a(1) * hij_bc(1,1)
      +  hij_a(-2) * x
      +  hij_a(2) * y ;
}



///////////////////////////////////////////////////////////
int
ModularFormulas::dimSknew (Character& chi, const int& k) 
{
  cout << "Not yet programmed."<<endl;
}

int
ModularFormulas::dimMk (const Character& chi, const int& k) const
{  cout << "Not yet programmed."<<endl;
}

int
ModularFormulas::idxG0 (const bigint& N) const
{
  bigint a = 1;	       
  rational_factorization f(N);
  f.factor();
  for (int i=0; i<f.no_of_comp(); i++) {
    bigint pn;
    power(pn,f.base(i),f.exponent(i));
    a *= pn + pn/f.base(i);
  }
#ifdef DEBUG
  if (!is_int(a)) error("idxG0: overflow.");
#endif
  int ans; a.intify(ans);
  return ans;
}


/* PARI CODE
mu0(n,m,npd)=m=factor(n);npd=matsize(m)[1];prod(x=1,npd,1+1/m[x,1],n);
mu20(n,m,npd)=if(n%4==0,0,m=factor(n);npd=matsize(m)[1];prod(x=1,npd,1+kronecker(-4,m[x,1]),1));
mu30(n,m,npd)=if(n%2==0 || n%9==0,0,m=factor(n);npd=matsize(m)[1];prod(x=1,npd,1+kronecker(-3,m[x,1]),1));
c0(n)=sumdiv(n,x,eulerphi(gcd(x,n/x)));
g0(n)=1+mu0(n)/12-mu20(n)/4-mu30(n)/3-c0(n)/2;
\\
mu1(n)=if(n<=2,mu0(n),eulerphi(n)*mu0(n)/2);
mu21(n)=if(n<4,mu20(n),0);
mu31(n)=if(n<4,mu30(n),0);
c1(n)=if(n<=3,c0(n),if(n==4,3,sumdiv(n,x,eulerphi(x)*eulerphi(n/x))/2));
g1(n)=1+mu1(n)/12-mu21(n)/4-mu31(n)/3-c1(n)/2;
\\
ss0(n,p)=if(isprime(p)==0 || n%p==0,printp("Function format: ss0(N,p) where p is prime and doesn't divide N."),g0(n*p)-2*g0(n)+1);
\\
muXNp(n,p)=mu1(n)*mu0(p);
mu2XNp(n,p)=0;
mu3XNp(n,p)=0;
cXNp(n,p)=2*c1(n);
gXNp(n,p)=if(n<4,g0(N*p),1+muXNp(n,p)/12-mu2XNp(n,p)/4-mu3XNp(n,p)/3-cXNp(n,p)/2);
ss1(n,p)=if(isprime(p)==0 || n%p==0,printp("Function format: ss1(N,p) where p is prime and doesn't divide N."),gXNp(n,p)-2*g1(n)+1);
\\
eisen(p)=if(isprime(p)==0,printp("Function format: eisen(N) where N is prime."),numer((p-1)/12));
\\
S0(n,k=2)=if(n<1,printp("S0(N,k), N must be a positive integer"),if(k<=0 || k%2,0,if(k==2,g0(n),(k-1)*(g0(n)-1)+(k/2-1)*c0(n)+mu20(n)*(k\4)+mu30(n)*(k\3))));
\\
S1(n,k=2)=if(n<1,printp("S1(N,k), N must be a positive integer"),if(k<=0 || (n<3 && k%2),0,if(k==1,printp("S1(N,k), k=1 not programmed for N>2."),if(k==2,g1(n),if(n<3,S0(n,k),(k-1)*(g1(n)-1)+(k/2-1)*c1(n)+(n==4 && k%2)/2+(n==3)*k\3)))));
\\  coefficients of zeta(s)^-2
mumu(n,x,k,p)=if(type(n)!="t_INT" || n<1,printp("mumu(N): N must be a positive integer."),x=factor(n);k=matsize(x)[1];p=1;for(y=1,k,if(x[y,2]>2,p=0,if(x[y,2]==1,p=-2*p,)));p);
\\
nf0(n,k=2)=if(n<1 || type(n)!="t_INT",printp("nf0(N,k): N must be a positive integer."),if(type(k)!="t_INT",printp("nf0(N,k): k must be an integer."),sumdiv(n,x,S0(x,k)*mumu(n/x))));
\\
nf1(n,k=2)=if(n<1 || type(n)!="t_INT",printp("nf1(N,k): N must be a positive integer."),if(type(k)!="t_INT",printp("nf1(N,k): k must be an integer."),sumdiv(n,x,S1(x,k)*mumu(n/x))));
printp("THIS IS A LAZY HACK FROM A 1.39 PROG SO DON'T RELY ON IT!")
printp()
printp("c0(N) -- calculates the number of cusps on X_0(N).");
printp("g0(N) -- calculates the genus of X_0(N).");
printp("ss0(N,p) -- calculates the number of s.s. points at p of X_0(N).");
printp("S0(N,k) -- calculates the dim of the wt. k cusp forms on Gamma_0(N).");
printp("nf0(N,k) -- calculates the dim of the wt. k newforms on Gamma_0(N).");
printp("eisen(N) -- (N PRIME!) calcs. the order of (0)-(\infty) on J_0(N).");
printp("c1(N) -- calculates the number of cusps on X_1(N).");
printp("g1(N) -- calculates the genus of X_1(N).");
printp("ss1(N,p) -- calculates the number of s.s. points at p of X_1(N).");
printp("S1(N,k) -- calculates the dim of the wt. k cusp forms on Gamma_1(N).");
printp("nf1(N,k) -- calculates the dim of the wt. k newforms on Gamma_1(N).");
printp("Please notify me (kaskel@math.berkeley.edu) if any of these functions seem");
printp("to be giving erroneous answers.");
\\
*/


// c0(n)=sumdiv(n,x,eulerphi(gcd(x,n/x)));
int
ModularFormulas::cusps0 (const int& N) const
{  
  if (N==1) return 1;		// divisors doesn't work right when N=1.
  sort_vector<bigint> div = divisors(N);
  bigint ans=0;
  for(int i=0; i<div.get_size(); i++) 
    ans+=eulerphi(gcd(div[i],N/div[i]));
#ifdef DEBUG
  if(!is_int(ans)) error("cusps0: overflow.");
#endif
  return to_int(ans);
}




//mu20(n,m,npd)=if(n%4==0,0,m=factor(n);npd=matsize(m)[1];prod(x=1,npd,1+kronecker(-4,m[x,1]),1));
int 
ModularFormulas::mu20 (const int& N) const
{
  if(N==1) return 1;
  if (N%4==0) return 0;
  bigint a = 1;	       
  rational_factorization f(N);
  f.factor();
  for (int i=0; i<f.no_of_comp(); i++) 
    a *= 1+jacobi(-4,f.base(i));
#ifdef DEBUG
  if (!is_int(a)) error("mu20: overflow.");
#endif
  return to_int(a);
}

//mu30(n,m,npd)=if(n%2==0 || n%9==0,0,m=factor(n);npd=matsize(m)[1];prod(x=1,npd,1+kronecker(-3,m[x,1]),1));
int 
ModularFormulas::mu30 (const int& N) const
{
  if(N==1) return 1;
  if (N%4==0 || N%9==0) return 0;
  bigint a = 1;	       
  rational_factorization f(N);
  f.factor();
  for (int i=0; i<f.no_of_comp(); i++) 
    a *= 1+jacobi(-3,f.base(i));
#ifdef DEBUG
  if (!is_int(a)) error("mu20: overflow.");
#endif
  return to_int(a);
}


//mu21(n)=if(n<4,mu20(n),0);
int 
ModularFormulas::mu21 (const int& N) const
{ 
  if(N<4) return mu20(N); else return 0 ;
}

//mu31(n)=if(n<4,mu30(n),0);
int 
ModularFormulas::mu31 (const int& N) const
{ 
  if(N<4) return mu30(N); else return 0 ;
}

//g0(n)=1+mu0(n)/12-mu20(n)/4-mu30(n)/3-c0(n)/2;
int
ModularFormulas::genus0 (const int& N) const
{ 
  if (N==1) return 0;
  bigrational a=1+idxG0(N)/bigrational(12)-
            mu20(N)/bigrational(4) - 
            mu30(N)/bigrational(3) - 
            cusps0(N)/bigrational(2);
#ifdef DEBUG
  if (denominator(a)!=1)
    error("genus0: not an integer, programming error");
#endif
  return to_int(numerator(a));
}

//mumu(n,x,k,p)=if(type(n)!="t_INT" || n<1,printp("mumu(N): N must be a positive integer."),x=factor(n);k=matsize(x)[1];p=1;for(y=1,k,if(x[y,2]>2,p=0,if(x[y,2]==1,p=-2*p,)));p);
int
ModularFormulas::mumu (const int& N) const
{ 
  if (N==1) return 1;
  int p = 1;	       
  rational_factorization f(N);
  f.factor();
  for (int i=0; i<f.no_of_comp(); i++)  
    if (f.exponent(i)>2) p=0;
    else if (f.exponent(i)==1) p=-2*p;
  return p;
}

//ss0(n,p)=if(isprime(p)==0 || n%p==0,printp("Function format: ss0(N,p) where p is prime and doesn't divide N."),g0(n*p)-2*g0(n)+1);
int
ModularFormulas::ss0 (const int& N, const int& p) const
{
#ifdef DEBUG
  if (!is_prime(p) || N%p==0)
    error("Function format: ss0(N,p) where p is prime and doesn't divide N.");
#endif
  return genus0(N*p)-2*genus0(N)+1;
}
  

//S0(n,k=2)=if(n<1,printp("S0(N,k), N must be a positive integer"),if(k<=0 || k%2,0,if(k==2,g0(n),(k-1)*(g0(n)-1)+(k/2-1)*c0(n)+mu20(n)*(k\4)+mu30(n)*(k\3))));
int
ModularFormulas::dimSk (const int& N, const int& k) const
{
#ifdef DEBUG
  if(N<1) error("dimSk(N,k), N must be a positive integer");
#endif
  if(k<=0||k%2!=0) return 0;
  if(k==2) return genus0(N);
  return (k-1)*(genus0(N)-1)
    +(k/2-1)*cusps0(N)+mu20(N)*(k/4)+mu30(N)*(k/3);
}

int
ModularFormulas::dimMk (const int& N, const int& k) const
{  
  cout << "Not yet programmed."<<endl;
}

//nf0(n,k=2)=if(n<1 || type(n)!="t_INT",printp("nf0(N,k): N must be a positive integer."),if(type(k)!="t_INT",printp("nf0(N,k): k must be an integer."),sumdiv(n,x,S0(x,k)*mumu(n/x))));
int
ModularFormulas::dimSknew (const int& N, const int& k) const
{  
  sort_vector<bigint> div = divisors(N);
  if (N==1) div.set_size(1) ;  	// divisors doesn't work right when N=1.
  bigint ans=0;
  for(int i=0; i<div.get_size(); i++) {
    int a; div[i].intify(a);
    ans+=dimSk(a,k)*mumu(N/a) ;
  }
  return to_int(ans);
}

//eisen(p)=if(isprime(p)==0,printp("Function format: eisen(N) where N is prime."),numer((p-1)/12));
int
ModularFormulas::eisen(const int& N) const
{ 
#ifdef DEBUG
  if (!is_prime(N)) error("Function format: eisen(N) where N is prime.");
#endif
  return to_int(numerator(bigrational(N-1)/bigrational(12)));
}

int
ModularFormulas::idxG1 (const int& N) const
{  
  if(N==1) return 1;
  return to_int(eulerphi(N)*idxG0(N));
}

//c1(n)=if(n<=3,c0(n),if(n==4,3,sumdiv(n,x,eulerphi(x)*eulerphi(n/x))/2));
int
ModularFormulas::cusps1 (const int& N) const
{
  if(N<=3)return cusps0(N);
  if(N==4)return 3;
  sort_vector<bigint> div = divisors(N);
  bigint ans=0;
  for(int i=0; i<div.get_size(); i++) 
    ans+=eulerphi(div[i])*eulerphi(N/div[i]) ;
  return to_int(ans/2);
}

//g1(n)=1+mu1(n)/12-mu21(n)/4-mu31(n)/3-c1(n)/2;
int
ModularFormulas::genus1 (const int& N) const
{ 
  if (N==1) return 0;
  bigint mu1=idxG1(N); if (N>2) mu1/=2;	// because of (-1,0;0,-1) not in Gamma_1(N). 
  bigrational a=1+mu1/bigrational(12)-
            mu21(N)/bigrational(4) - 
            mu31(N)/bigrational(3) - 
            cusps1(N)/bigrational(2);
#ifdef DEBUG
  if (denominator(a)!=1) {
    cout << "Error computing genus1("<<N<<"). Got "<<a<<endl;
    error("genus1: not an integer, programming error");
  }
#endif
  return to_int(numerator(a));
  
}


//muXNp(n,p)=mu1(n)*mu0(p);
//mu2XNp(n,p)=0;
//mu3XNp(n,p)=0;
//cXNp(n,p)=2*c1(n);

int 
ModularFormulas::muXNp (const int& N, const int& p) const
{return (idxG1(N)/2)*idxG0(p);}
int
ModularFormulas::mu2XNp (const int& N, const int& p) const
{return 0;}
int
ModularFormulas::mu3XNp (const int& N, const int& p) const
{return 0;}
int
ModularFormulas::cXNp (const int& N, const int& p) const
{return 2*cusps1(N);}
//gXNp(n,p)=if(n<4,g0(N*p),1+muXNp(n,p)/12-mu2XNp(n,p)/4-mu3XNp(n,p)/3-cXNp(n,p)/2);
int
ModularFormulas::gXNp (const int& N, const int& p) const
{
  if (N<4) return genus0(N*p);
  bigrational a=1+muXNp(N,p)/bigrational(12)
    - mu2XNp(N,p)/bigrational(4)
    - mu3XNp(N,p)/bigrational(3)
    - cXNp(N,p)/bigrational(2);
#ifdef DEBUG
  if (denominator(a)!=1)
    error("genus0: not an integer, programming error");
#endif
  return to_int(numerator(a));
}


//ss1(n,p)=if(isprime(p)==0 || n%p==0,printp("Function format: ss1(N,p) where p is prime and doesn't divide N."),gXNp(n,p)-2*g1(n)+1);
int
ModularFormulas::ss1 (const int& N, const int& p) const
{
#ifdef DEBUG
  if(!is_prime(p) || N%p==0) error("Function format: ss1(N,p) where p is prime and doesn't divide N.");
#endif
  return gXNp(N,p)-2*genus1(N)+1 ;
}  

//S1(n,k=2)=if(n<1,printp("S1(N,k), N must be a positive integer"),if(k<=0 || (n<3 && k%2),0,if(k==1,printp("S1(N,k), k=1 not programmed for N>2."),if(k==2,g1(n),if(n<3,S0(n,k),(k-1)*(g1(n)-1)+(k/2-1)*c1(n)+(n==4 && k%2)/2+(n==3)*k\3)))));
int
ModularFormulas::dimSk1 (const int& N, const int& k) const
{
#ifdef DEBUG
  if(N<1) error("S1(N,k), N must be a positive integer");
#endif
  if(k<=0 || (N<3 && k%2!=0)) return 0;
  if(k==1) error("dimSk1(N,k), k=1 not programmed for N>2."); 
  if(k==2) return genus1(N); 
  if(N<3) return dimSk(N,k);
  bigrational a = (k-1)*(bigrational(genus1(N))-1)
    +(k/bigrational(2)-1)*cusps1(N)
    + ((N==4 && k%2!=0)?(1/bigrational(2)):0)
    + ((N==3)?bigrational(k/3):0) ;
#ifdef DEBUG
  if (denominator(a)!=1) {
    cout << "Error computing N="<<N<<",k="<<k<<endl
	 << "a="<<a<<endl;
    error("dimSk1: not an integer, programming error");
  }
#endif
  return to_int(numerator(a));
}

int
ModularFormulas::dimMk1 (const int& N, const int& k) const
{ 
  cout << "Not yet programmed."<<endl;
}

//nf1(n,k=2)=if(n<1 || type(n)!="t_INT",printp("nf1(N,k): N must be a positive integer."),if(type(k)!="t_INT",printp("nf1(N,k): k must be an integer."),sumdiv(n,x,S1(x,k)*mumu(n/x))));
int
ModularFormulas::dimSk1new (const int& N, const int& k) const
{ 
  sort_vector<bigint> div = divisors(N);
  if (N==1) div.set_size(1) ;  	// divisors doesn't work right when N=1.
  bigint ans=0;
  for(int i=0; i<div.get_size(); i++) {
    int n; div[i].intify(n);
    ans+=dimSk1(n,k)*mumu(N/n) ;
  }
  return to_int(ans);
}

void 
ModularFormulas::help()
{
  cout 
       << "\tcusps0   : number of cusps on X_0(N)" << endl
       << "\tcusps1   : number of cusps on X_1(N)" << endl
       << "\tdimSk    : dimension of S_k(N) using Shimura's formulas." << endl
       << "\tdimSk1   : dimension of S_k(Gamma_1(N)) using Shimura's formulas." << endl
       << "\tdimSk1new: dimension of S_k^{new}(Gamma_1(N)) using Shimura's formulas."
       << endl
       << "\tdimSkchi : dimension of S_k(N,chi) using Hijikata's trace formula." << endl
       << "\tdimSknew : dimension of S_k^{new}(N) using Shimura's formulas." << endl
       << "\teisen    : (N PRIME!) calcs. the order of (0)-(infty) on J_0(N)." << endl
       << "\tidxG0    : index [SL2(Z):Gamma_0(N)]" << endl
       << "\tidxG1    : index [SL2(Z):Gamma_1(N)]" << endl
       << "\tgenus0   : genus of X_0(N)" << endl
       << "\tgenus1   : genus of X_1(N)" << endl
       << "\tss0      : number of s.s. points at p on X_0(N)" << endl
       << "\tss1      : number of s.s. at p " << endl
       << "\tx        : exit"<<endl<<endl;

    //	 << "dimMkchi : Not yet done." << endl
    //	 << "dimMk1   : Not done." << endl
    //	 << "dimMk    : Not yet done." << endl
}
  
void 
ModularFormulas::interactive () 
{
  cout << "Formulas to computer modular type numbers." << endl;
  cout << "Type ? for help."<<endl;
  while (1) {
    char buf[256];
    get_str(buf,256,"(Formulas) ? ");
    if (strcmp(buf,"x")==0 || strcmp(buf,"q")==0) break;
    if (strcmp(buf,"?")==0) {
      help();
    } else if (strcmp(buf,"dimSkchi")==0) {
	Character chi;		// default: user initializes chi.
	cout << dimSk(chi,get_int("k = ",2)) << endl;
    } else if (strcmp(buf,"dimMkchi")==0) {
      cout<<"Not implimented."<<endl;
    } else if (strcmp(buf,"idxG0")==0) {
      cout << idxG0(get_int("N = ",1))<<endl;
    } else if (strcmp(buf,"cusps0")==0) {
      cout << cusps0(get_int("N = ",1))<<endl;
    } else if (strcmp(buf,"genus0")==0) {
      cout << genus0(get_int("N = ",1))<<endl;
    } else if (strcmp(buf,"ss0")==0) {
	int N=get_int("N = ",1);
	int p=get_int("p = ",2);
	cout << ss0(N,p)<<endl;
    } else if (strcmp(buf,"dimSk")==0) {
	int N=get_int("N = "); 
	int k=get_int("k = ");
	cout << dimSk(N,k)<<endl;
    } else if (strcmp(buf,"dimMk")==0) {
	int N=get_int("N = "); 
	int k=get_int("k = ");
	cout << dimMk(N,k)<<endl;
    } else if (strcmp(buf,"dimSknew")==0) {
	int N=get_int("N = "); 
	int k=get_int("k = ");
	cout << dimSknew(N,k)<<endl;
    } else if (strcmp(buf,"eisen")==0) {
      cout << eisen(get_int("N = ",1))<<endl;
    } else if (strcmp(buf,"idxG1")==0) {
      int N=get_int("N = "); cout << idxG1(N)<<endl;
    } else if (strcmp(buf,"cusps1")==0) {
      int N=get_int("N = "); cout << cusps1(N)<<endl;
    } else if (strcmp(buf,"genus1")==0) {
      int N=get_int("N = "); cout << genus1(N)<<endl;
    } else if (strcmp(buf,"ss1")==0) {
      int N=get_int("N = "); 
      int p = get_int("p = "); cout << ss1(N,p)<<endl;
    } else if (strcmp(buf,"dimSk1")==0) {
      int N=get_int("N = "); 
      int k=get_int("k = ");
      cout << dimSk1(N,k)<<endl;
    } else if (strcmp(buf,"dimSk1new")==0) {
      int N=get_int("N = "); 
      int k=get_int("k = ");
      cout << dimSk1new(N,k)<<endl;
    } else if (strcmp(buf,"dimMk1")==0) {
      int N=get_int("N = "); 
      int k=get_int("k = ");
      cout << dimMk1(N,k)<<endl;
    }
  }
}

void 
ModularFormulas::error (const char* msg) const
{
#ifdef DEBUG
  cerr << "ModularFormulas error.  "<<msg <<endl;
  exit(-1);
#endif
}

void ModularFormulas::mesg (const char* s) const
{
  ::mesg("ModularFormulas: ");
  ::mesg(s);
  ::mesg("\n");
}


