/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

// modularcalc.cc
// Impliments templates defined in modularcalc.h.

#include "modularcalc.h"
#include "h1nchiz.h"

template <class F>
void ModularCalculator<F>::status() 
{
    char s[1024]; chi.ppname(s);
//    cout << title <<endl;
    cout << "---------------------------------------------------------------"<<endl;
    cout << "Current space:   " ;
    cout << "M_" << k << "(" << s << ";";
    if (field_type(foo)==FT_FFIELD) {
	cout<<" F_"<<ffield_element::characteristic()<<"[x]/("
	    << ffield_element::irred_polynomial() << ") )";
    } else if (field_type(foo)==FT_ALGNUM) {
	cout<<" Q[a]/("<< algnum::irred_polynomial()<<") )";
    } else if (field_type(foo)==FT_BIGRAT) {
	cout<<" Q)";
    } else if (field_type(foo)==FT_CHARMOD) {
	cout<<" F_"<<charmod::get_pr()<<")";
    } else if (field_type(foo)==FT_INTMOD) {
	cout<<" F_"<<intmod::get_pr()<<")";
    } else {
	cout << " " << field_name(foo) << ")";
    }
    if(plus) cout<<"^+";
    
    cout << ", dim="
	 << H->dim_Mk() ;
    cout << "\nHecke action on: V=" ;
    if (mk) cout << "M_" << k << ", dim=" << H->dim_Mk() << endl ;
    else if (sk) cout << "S_"<< k << ", dim=" << H->dim_Sk() << endl ;
    else cout << Vname << ", dim=" << V.ncols()<<endl ;
    cout << "---------------------------------------------------------------"<<endl;
}


template <class F>
void ModularCalculator<F>::helpnum(int topic) 
{
    cout << endl;
    switch(topic) {
    case 1:  // 1: computing action of OPERATORS
	cout << "actatkin        actconj         acthecke "<<endl
	     << "actfasthecke    actznz" << endl;
	break;
	
    case 2:  // 2: setting current SPACE
	cout << "spacecusp       spaceeisenstein     spaceintegral "<<endl
	     << "spacefull       spaceload           spacemkz"<<endl
	     << "spacenew        spacesave"<<endl;
	break;
	
    case 3: // 3: cutting out subspace
	cout << "subconjminus    subconjplus         subeigenpoly "<<endl;
	cout << "subeigenval    "<<endl;
	break;
	
    case 4:  // 4: computing BASIS
	cout << "basisap           basiscurrent       basisfree"<<endl
	    
	     << "basisheckedual    basisintegral      basisnew"<<endl
	     << "basisrational     basisrationalnew"<<endl;
	break;
	
	
    case 5:  // 5: CONVERSIONS between representations
	cout << "convmanin         convmatrix      convmodular"<<endl
	     << "convredmodp"<<endl;
	break;
	
    case 6:   // 6: INVARIANTS of associated torus.
	cout << "torusbsd            toruscinfinity     torusellc4c6 "<<endl
	     << "torusellj           toruslseries       torusmoddeg  "<<endl
	     << "torusintpairing     torusperiodlattice  torusrealperiod"<<endl
	     << "torustorsion"<<endl;
	break;

    case 7:  // 7: INVARIANTS of Hecke algebra
	cout << "heckedisc           heckediscnew" <<endl;
	break;
	
    case 8:   // 8: OPTIONS
	cout << "opcharpoly       opmatrix      optrace" << endl;
	break;
	
    }
    cout << endl;
}

template <class F>
void ModularCalculator<F>::helpword(char* name) 
{
    cout << endl;
    if (strncmp(name,"actatkin",4)==0 || strcmp(name,"w")==0) 
	cout << "actatkin: compute the Atkin-Lehner involution W_q on V."<<endl;
    else if (strncmp(name,"actconj",4)==0)
	cout << "actconj: compute the complex conjugation involution * on V."<<endl;
    else if (strncmp(name,"actfasthecke",4)==0 || strcmp(name,"f")==0)
	cout << "actfasthecke: quickly compute the action of Hecke operators on a Hecke invariant subspace of the dual of M_k."<<endl;
    else if (strncmp(name,"acthecke",4)==0 || strcmp(name,"t")==0)
	cout << "acthecke: compute the action of the Hecke operators on V."<<endl;
    else if (strncmp(name,"actznz",4)==0)
	cout << "atcznz: compute the action of (Z/NZ)^* on M_k(Q)."<<endl;
    else if(strcmp(name,"heckedisc")==0)
	cout << "heckedisc: compute discriminant of Hecke acting on V."<<endl;
    else if(strncmp(name,"heckediscnew",10)==0)
	cout << "heckediscnew: special algorithm to compute the discriminant of the action of T on new subspace of S_k."<<endl;
    
    else if (strncmp(name,"spacecusp",6)==0)
	cout << "spacecusp: V = the space of cuspidal symbols."<<endl;
    else if (strncmp(name,"spaceeisenstein",6)==0)
	cout << "spaceeisenstein: V = the space of Eisenstein symbols."<<endl;
    else if (strncmp(name,"spaceintegral",6)==0)
	cout << "spaceintegral: V = the space of integral cuspidal symbols."<<endl;
    else if (strncmp(name,"spacemkz",6)==0)
	cout << "spacemkz: V = the space of integral symbols."<<endl;
    else if (strncmp(name,"spacefull",6)==0)
	cout << "spacefull: V = M_k"<<endl;
    else if (strncmp(name,"spaceload",6)==0)
	cout << "spaceload: load V from a file in userdata."<<endl;
    else if (strncmp(name,"spacenew",6)==0)
	cout << "spacenew: V = new cuspidal modular symbols."<<endl;
    else if (strncmp(name,"spacesave",6)==0)
	cout << "spacesave: save V to a file in userdata."<<endl;
    
    else if (strncmp(name,"subconjminus",8)==0)
	cout << "subconjminus: set V = -1 eigenspace for conjugation."<<endl;
    else if (strncmp(name,"subconjplus",8)==0)
	cout << "subconjplus: set V = +1 eigenspace for conjugation."<<endl;
    else if (strncmp(name,"subeigenpoly",9)==0)
	cout << "subeigenpoly: set V = ker(f(T_p)), some polynomial f(x)."<<endl;
    else if (strncmp(name,"subeigenval",9)==0)
	cout << "subeigenval: set V = ker(T_p-a_p), some eigenvalue a_p."<<endl;
    
    else if (strncmp(name,"basisap",6)==0)
	cout << "basisap: compute a list of Hecke eigenvalues a_p for each newform in V."<<endl;
    else if (strncmp(name,"basiscurrent",6)==0)
	cout << "basiscurrent: display a basis for V."<<endl;
    else if (strncmp(name,"basisfree",6)==0)
	cout << "basisfree: display the free basis for M_k"<<endl;
    else if (strncmp(name,"basisheckedual",6)==0)
	cout << "basisheckedual: display a basis for the dual of V inside of Hom(M_k,K)."<<endl;
    else if (strncmp(name,"basisintegral",6)==0)
	cout << "basisintegral: display a basis for the space of integral cuspidal modular symbols."<<endl;
    else if (strncmp(name,"basisnewforms",6)==0)
	cout << "basisnewforms: display the q-expansions of the normalized newforms corresponding to newforms in V^+."<<endl;
    else if (strcmp(name,"basisrational")==0)
	cout << "basisrational: display a basis of q-expansions, with coefficients in K, for the whole of V"<<endl;
    else if (strcmp(name,"basisrationalnew")==0)
	cout << "basisrationalnew: display a basis of q-expansions, with coefficients in K, for the newforms in V."<<endl;
    else if (strncmp(name,"convmanin",7)==0)
	cout << "convmanin: given a manin symbol, gives the corresponding element of M_k"<<endl;
    else if (strncmp(name,"convmatrix",7)==0)
	cout << "convmatrix: given a matrix gives the corresponding element of M_k"<<endl;
    else if (strncmp(name,"convmodular",6)==0)
	cout << "convmodular: given a modular symbol, gives the corresponding element of M_k"<<endl;
    else if (strncmp(name,"convredmodp",4)==0)
	cout << "convredmodp: compute the space M_k\\tensor F_p."<<endl;
    
    else if (strncmp(name,"torusbsd",6)==0)
	cout << "torusbsd: algebraically compute bsd data for each newform in V."<<endl;
    else if (strncmp(name,"toruscinfinity",6)==0)
	cout << "toruscinfinity: compute the number of real components of the complex torus A_V."<<endl;
    else if (strncmp(name,"torusellc4c6",9)==0)
	cout << "torusellc4c6: if the torus A_V has dimension 1, approximate the Weierstrass invariants c4 and c6"<<endl;
    else if (strcmp(name,"torusellj")==0)	    
	cout << "torusellj: if the torus A_V has dimension 1, approximate the j-invariant of the corresponding elliptic curve"<<endl;
    else if (strncmp(name,"toruslseries",6)==0)
	cout << "toruslseries: approximate L^(r)(A_V,1)/r!"<<endl;
    else if (strncmp(name,"torusmoddeg",6)==0)
	cout << "torusmoddeg: compute the modular degree of A_V."<<endl;
    else if (strncmp(name,"torusintpairing",6)==0)
	cout << "torusintpairing: compute the intersection pairing of A_V."<<endl;
    else if (strncmp(name,"torusperiodlattice",6)==0)
	cout << "torusperiodlattice: approximate the period lattice associate to V."<<endl;
    else if (strncmp(name,"torusrealperiod",6)==0)
	cout << "torusrealperiod: approximate the volume of the space of real points A_V(R)."<<endl;
    else if (strncmp(name,"torustorsion",6)==0)
	cout << "torustorsion: compute gcd{2<p<bound, (p,N)=1:charpoly(T_p|V)(p+1)}.  This is an upperbound on A_V(Q)_tor in the weight 2 case."<<endl;
    
    else if (strncmp(name,"opcharpoly",3)==0)
	cout << "opcharpoly: toggle characteristic polynomial display of computed operators."<<endl;
    else if (strncmp(name,"opmatrix",3)==0)
	cout << "opmatrix: toggle matrix display of computed operators."<<endl;
    else if (strncmp(name,"optrace",3)==0)
	cout << "optrace: toggle trace display of computed operators."<<endl;
    
    else cout << " ***   " << name << ": unknown function."<<endl;
    cout << endl;
}

template <class F>
void ModularCalculator<F>::help() 
{
    cout << "  1: computing OPERATORS" << endl;
    cout << "  2: setting current SPACE" <<endl;
    cout << "  3: cutting out SUBSPACES" << endl;
    cout << "  4: computing BASIS" << endl;
    cout << "  5: CONVERSIONS between representations" << endl;
    cout << "  6: arithmetic INVARIANTS of torus A_V" << endl;
    cout << "  7: INVARIANTS of Hecke algebra" << endl;
    cout << "  8: OPTIONS" << endl;
    cout << "Further help (list of relevant functions): ?n (1<=n<=5)"<<endl;
    cout << "Also: ? functionname (short on-line help)"<<endl;
    cout << "      x = exit"<<endl;
}

template <class F>
void ModularCalculator<F>::interface1() 
{
    cout << "Welcome to the modular symbols calculator." << endl;
    cout << "Type ? for help."<< endl << endl;

    while (1) {
	status();
	cout << "M_"<<k<<"(" << N << ") ";
	char buf[1024];
	get_str(buf, 1024, "? ");
	char* c=buf;
	while(c[0]==' ') c=c+1;
	if(c[0]=='?') {
	    c=c+1;
	    while(c[0]==' ') c++;
	    int n=atoi(c);
	    if(1<=n && n<=8)
		helpnum(n);
	    else if(c[0]!='\0')
		helpword(c);
	    else 
		help();
	    continue;
	}

	char* name =c;
	
	if('1'<=name[0] && name[0]<='8') {
	    name[1]='\0';
	    helpnum(atoi(name));
	}
	else if(strcmp(name,"s")==0)
	    status();
	else if (strncmp(name,"actatkin",4)==0) 
	    compute_atkin();
	else if (strncmp(name,"actconj",4)==0)
	    compute_conjugation();
	else if (strncmp(name,"actfasthecke",4)==0 || strcmp(name,"f")==0)
	    compute_fast_Tp();
	else if (strncmp(name,"acthecke",4)==0 || strcmp(name,"t")==0)
	    compute_Tn();
	else if (strncmp(name,"actznz",4)==0)
	    act_ZmodNZ();
	else if (strcmp(name,"heckedisc")==0)
	    compute_discriminant();
	else if(strncmp(name,"heckediscnew",10)==0)
	    compute_discriminant_new();
	
	else if (strncmp(name,"spacecusp",6)==0)
	    init_V('c');
	else if (strncmp(name,"spaceeisenstein",6)==0)
	    init_V('e');
	else if (strncmp(name,"spaceintegral",6)==0)
	    init_V('z');
	else if (strncmp(name,"spacemkz",6)==0)
	    init_V('Z');
	else if (strncmp(name,"spacefull",6)==0)
	    init_V('m');
	else if (strncmp(name, "spaceload", 6)==0)
	    init_V('l');
	else if (strncmp(name,"spacenew", 6)==0)
	    init_V('n');
	else if (strncmp(name, "spacesave", 6)==0)
	    init_V('s');
	else if (strcmp(name, "v")==0)
	    init_V('?');
	
	else if (strncmp(name,"subconjminus",8)==0 || strcmp(name,"subm")==0)
	    compute_conjeigen(false);
	else if (strncmp(name,"subconjplus", 8)==0 || strcmp(name,"subp")==0)
	    compute_conjeigen(true);
	else if (strncmp(name,"subeigenpoly", 9)==0 || strcmp(name,"g")==0)
	    compute_gen_eigen();
	else if (strncmp(name,"subeigenval", 9)==0 || strcmp(name,"e")==0)
	    compute_eigen();
	
	else if (strncmp(name,"basisap", 6)==0)
	    compute_aplist();
	else if (strncmp(name,"basiscurrent", 6)==0)
	    display_basiscurrent();
	else if (strncmp(name,"basisfree", 6)==0)
	    display_basisfree();
	
	else if (strncmp(name,"basisheckedual", 6)==0) 
	    basisheckedual();
	else if (strncmp(name,"basisintegral", 6)==0)
	    compute_integral_basis();
	else if (strncmp(name,"basisnewforms", 6)==0) 
	    compute_anlist();
	else if (strcmp(name,"basisrational")==0)
	    basisrational();
	else if (strcmp(name,"basisrationalnew")==0)
	    basisrationalnew();
	
	else if (strncmp(name,"convmanin", 7)==0)
	    manin_to_free();
	else if (strncmp(name,"convmatrix", 7)==0)
	    matrix_to_free();
	else if (strncmp(name,"convmodular", 7)==0)
	    modular_to_free();
	else if (strncmp(name,"convredmodp", 4)==0)
	    reduce_modp();
	
	else if (strncmp(name,"torusbsd", 6)==0 || strcmp(name,"o")==0)
	    torusbsd();
	else if (strncmp(name,"toruscinfinity", 6)==0)
	    toruscinfinity();
	else if (strncmp(name,"torusellc4c6", 9)==0)
	    torusellc4c6();
	else if (strcmp(name,"torusellj")==0)	    
	    torusellj();
	else if (strncmp(name,"toruslseries",6)==0)
	    toruslseries();
	else if (strncmp(name,"torusmoddeg",6)==0)
	    torusmoddeg();
	else if (strncmp(name,"torusintpairing",6)==0)
	    torusintpairing();
	else if (strncmp(name,"torusperiodlattice",6)==0)
	    torusperiodlattice();
	else if (strncmp(name,"torusrealperiod",6)==0)
	    torusrealperiod();
	else if (strncmp(name,"torustorsion",6)==0)
	    torustorsion();
	
	else if (strncmp(name,"opcharpoly",3)==0 || strcmp(name,"c")==0)
	    opcharpoly();
	else if (strncmp(name,"opmatrix",3)==0 || strcmp(name,"m")==0 )
	    opmatrix();
	else if (strncmp(name,"optrace",3)==0)
	    optrace();
	
	else if(strcmp(c,"x")==0||strcmp(c,"q")==0
		||strcmp(c,"\\q")==0||strcmp(c,"quit")==0)
	    break;

	else if(strncmp(c,"cyc",3)==0)
	    cyclic_module();
	else if(strncmp(c,"bsd",3)==0)
	    torus_bsdratio();

	else
	    status();
    }
}

template <class F>
void ModularCalculator<F>::init_sk() 
{
    V = H->get_Sk();
    mk = false ; sk = true ;
}

template <class F>
void ModularCalculator<F>::init_mk() 
{
    mk=true; sk=false; 
}

template <class F>
void ModularCalculator<F>::init_V(char c) 
{
    if (c=='?') {
	c=get_char("Select one:\n\
c: CUSP forms\n\
m: MODULAR forms\n\
n: NEW cusp forms\n\
z: INTEGRAL Z-basis for homology\n\
Z: INTEGRAL Z-basis for M_k\n\
e: Eisenstein part\n\
s: save V\n\
l: load V\n\
x: cancel\n> ? ");
    }
    /* u = user defined space (e.g., coming from common eigenspace computation)\n\
       o = S^{old}(N,chi)  cuspidal oldforms\n\ 
    */
    
    if (c =='x' || c=='q') { return; }
    else if (c=='c') {
	sk=true; mk=false; sknew=false; V=H->get_Sk(); 
	
    } 
    else if (c=='e') {
	mk=sk=sknew=false;
	strcpy(Vname, "Eis");
	const Matrix<F>& Eis = H->get_Eis();
	V = Eis;
    }
    else if (c=='m') {
	mk=true; sk=false; sknew=false;
	V.idmat(H->dim_Mk());
	sprintf(Vname, "M_%d(%d)", k, N);
    }
    else if (c=='n') {
	mk=sk=false;
	sknew=true;
	sprintf(Vname, "S_%d(%d)^{new}", k, N);
	H->get_Sknew(V); 
    } 
    else if (c=='z') {
	if(!integral) {
	    cout << "spaceintegral: there is no integral structure."<<endl;
	    return;
	}
	mk=sk=sknew=false;
	strcpy(Vname,"H_1(X_0(N),Z)/H_1^-");
	if(!plus) Vname[13]='\0';
	V = H->get_ZBasis_Sk();
    } else if (c=='Z') {
	if(!integral) {
	    cout << "spaceintegral: there is no integral structure."<<endl;
	    return;
	}
	mk=sk=sknew=false;
	strcpy(Vname,"M_k(Gamma,Z)^+");
	if(!plus) Vname[strlen(Vname)-2]='\0';
	V = H->get_ZBasis_Mk();
    } else if (c=='l') {
	get_str(Vname,1024,"Enter the name of a previously saved V (in this space!): ");
	ifstream in(userdata_read(Vname));
	if (!in) {
	    cout << "Couldn't open userdata '" << Vname << "'." << endl ;
	    return ;
	}
	V.Load(in);
	mk=sk=sknew=false;
    } else if (c=='s') {
	if(mk||sk) {
	    cout<<"It's only interesting to save proper subspaces of Sk or Mk."<<endl; return; }
	get_str(Vname,1024,"Name V : ");
	ofstream out(userdata_write(Vname));
	if (!out) {
	    cout << "Couldn't open userdata '" << Vname << "'." << endl ;
	    return ;
	}
	V.Save(out);
    } else {
	cout << "Command not recognized.  Nothing done."<<endl;
    }
    return;
}

template <class F>
void ModularCalculator<F>::reduce_modp() 
{
#ifdef INST_CHARMOD
    cout << "Reduce mod p."<<endl;
    F a;
    if(field_type(a)!=FT_ALGNUM) {
	cout << "Reduce modp only implimented for algnum fields."<<endl;
	return;
    }
    int p=get_int("p = ");
    if (p<=2 || p>253 || !is_prime(p)) {
	cout << "p must be a prime between 3 and 253."<<endl;
	return;
    }
    InitFieldCharmod(p);
    if (!H->CanReduce_charmod()) {
	cout << "Due to denominators, it is impossible to reduce mod p."<<endl;
	return;
    }
    cout << "What element of F_"<<p<<" should x map to?  Note, that it had"
	 << endl
	 << "better map to a root of the min poly of x mod "
	 << p << "!" << endl;
    charmod z=get_int("x |--> ");
    H1NChi<charmod> *Hmod=new H1NChi<charmod>(Character(1,0,save),2,save);
    H->Reduce_charmod(*Hmod, z);
    
    char buf[512];
    sprintf(buf,"(Now working in the reduction modulo %d.)", p);
    ModularCalculator<charmod>* pHa = new ModularCalculator<charmod>(buf,Hmod);
    pHa->interface1();
    delete pHa;
    cout << "Finished working modulo " << p << "."<<endl
	 << "Now returning to characteristic zero."<<endl;
    status();
#else
    cout << "charmod is not compiled into hecke."<<endl;
#endif
}

template <class F>
void ModularCalculator<F>::compute_discriminant() 
{
    F disc;
    if(mk) disc=H->heckedisc();
    else  disc=H->heckedisc(V);
    cout << "disc(T|V) = " << disc << endl;
}

template <class F>
void ModularCalculator<F>::compute_discriminant_new() 
{
    F discTnew, discTfnewprod, Vol, Idx;
    H->heckedisc_Sknew(discTnew, discTfnewprod, Vol, Idx);
    cout << "disc(Tnew) = " << discTnew << endl
	 << "prod(disc(T_f)) = " << discTfnewprod << endl;
//	 << "Vol = " << Vol << endl
//	 << "Idx = " << Idx << endl;
}

template <class F>
void ModularCalculator<F>::compute_integral_basis() 
{
    cout << "The columns of the following matrix form a basis " << endl
	 << "for the integral cuspidal symbols, expressed in terms " << endl
	 << "of the free basis."<<endl;
    const Matrix<F>& B=H->get_ZBasis();
    cout << "Zbase = " << B << endl <<endl;

/*
    cout << "In terms of Manin symbols the basis is: " << endl;
    for(int i=1;i<=B.ncols();i++) {
	MSymVec<F> m;
	H->VecToMSymVec(m,B.col(i));
	cout << "z[" << i << "] = " << m << endl;
    }
    cout <<endl << "In terms of modular symbols the basis is: " << endl;
    for(int i=1;i<=B.ncols();i++) {
	MSymVec<F> m;
	H->VecToMSymVec(m,B.col(i));
	cout << "z[" << i << "] = " ;
	m.ModSymPrint(cout,k);
	cout << endl;
    }
*/
    
}


template <class F>
void ModularCalculator<F>::cyclic_module()
{
    Vector<F> e;
    H->WindingElement(e);
    Matrix<F> Te;
    H->CyclicHeckeModule(Te,e);
    cout << "{T*e : T in Hecke algebra} = "<<endl;
    cout << "{Te = " << Te << ";}"<<endl;
}

template <class F>
void ModularCalculator<F>::display_basisfree() 
{
    Matrix<F> B;
    B.idmat(H->dim_Mk());
    cout << "Display the free basis." << endl;
    cout << "In terms of Manin symbols the basis is: " << endl;
    for(int i=1;i<=B.ncols();i++) {
	MSymVec<F> m;
	H->VecToMSymVec(m,B.col(i));
	cout << "b[" << i << "] = " << m << endl;
    }
    cout <<endl << "In terms of modular symbols the basis is: " << endl;
    for(int i=1;i<=B.ncols();i++) {
	MSymVec<F> m;
	H->VecToMSymVec(m,B.col(i));
	cout << "b[" << i << "] = " ;
	m.ModSymPrint(cout,k);
	cout << endl;
    }
    
    /*
      cout << "The relations satisfied by Modular symbols are:"<<endl;
      cout << "Warning: this function is untested!" << endl;
      Matrix<F> mrels;
      H->MSymRelations(mrels);
      // The columns of mrels give the relations between the 
      // Manin symbols in the standard list.
      for(int i=1;i<=mrels.ncols();i++) {
      for(int j=1;j<=mrels.nrows();j++) {
      if(mrels(j,i)!=0) {
      H->get_MSymbols().x(j).ModSymPrint(cout,N,k);
      cout << " + " ;
      }
      }
      cout << endl;
      }
    */
}

template <class F>
void ModularCalculator<F>::display_basiscurrent() 
{
    if(!mk) {
	cout << "Columns give basis of V wrt free basis b[i]."<<endl;
	cout << V <<endl<<endl;
    } else {
	cout << "The current space is the full space M_k."<<endl;
	display_basisfree();
    }
}

template <class F>
void ModularCalculator<F>::basisheckedual() 
{
    Matrix<F> VDual;
    if(mk) {
	VDual.idmat(H->dim_Mk());
    } else {
	H->ComputeHeckeDual(VDual, V, 37);
    }
    VDual.transpose();
    cout << "VDual = " << VDual << endl;
}

template <class F>
void ModularCalculator<F>::modular_to_free() 
{
    if(k>2)
	cout << "Enter X^qY^("<<k-2<<"-q)*{a/b,c/d}."<<endl;
    else
	cout << "Enter {a/b,c/d}."<<endl;
    int q=0;
    if(k>2)
	q=get_int("q = ");
    bigint a,b,c,d;
    a=get_int("a = "); b=get_int("b = "); c=get_int("c = "); d=get_int("d = ");
    if((a==0 && b==0) || (c==0 && d==0)) {
	cout << "Invalid input." << endl;
	return;
    }
    Vector<F> v;
    H->modular_to_free(v,a,b,c,d,q);
    cout << v << endl ;
}

template <class F>
void ModularCalculator<F>::manin_to_free() 
{
    if(k>2)
	cout << "Enter [X^qY^"<<k-2<<"-q,(u,v)]."<<endl;
    else
	cout << "Enter (u,v)." <<endl;
    int q,u,v;
    
    if(k>2) {
	q=get_int("q = "); if(q==-1) return;
	if(q>k-2) {
	    cout << "q is too large."<<endl;
	    return;
	}
    } else 
	q=0;
    u=get_int("u = "); v =get_int("v = ");
    if(abs(gcd(u,gcd(v,N)))!=1) {
	cout << "Invalid Manin symbol."<<endl;
	return;
    }
    MSym x(u,v,q);
    Vector<F> vec;
    H->MSymToVec(vec,x);
    cout <<vec << endl;
}

template <class F>
void ModularCalculator<F>::matrix_to_free() 
{
    while(1) {
	if(k!=4) {
	    cout << "matrix_to_free only implimented for weight 4."<<endl;
	    return;
	}
	cout << endl << "Enter a 2x2 matrix in Gamma_0("<<N<<") "<<endl;
	cout << "such that c>0.  (set c=0 to quit.)"<<endl; 
	int a,b,c,d;
	a=get_int("a = "); b=get_int("b = "); 
	c=get_int("c = ");
	if(c==0) return;
	d=get_int("d = ");
	bigint p2=1-bigint(a)*bigint(d),
	    p1=- bigint(b)*(bigint(d)-bigint(a)),
	    p0=bigint(b)*bigint(b);
	
	cout << p2<<"X^2+"<<p1<<"XY+"<<p0<<"Y^2{"
	     <<0<<","<<b<<"/"<<d<<"}"<<endl;
	Vector<F> v;
	H->modular_to_free(v,0,1,b,d,0);
	v*=F(p0);
	
	Vector<F> w;
	H->modular_to_free(w,0,1,b,d,1);
	v+=F(p1)*w;
	
	H->modular_to_free(w,0,1,b,d,2);
	v+=F(p2)*w;
	cout << endl << "="<<v<<endl;
	cout << "Using LoopSymbol gives v = " <<endl;
	H->LoopSymbol(v, a,b,c,d);
	cout << endl << "="<<v<<endl;
    }
}

template <class F>
void ModularCalculator<F>::torusbsd() 
{
    assure_newspace();
    compute_conjeigen(true);
    if(V.ncols()<1) return;


    cout << "Warning: BSD values could be off by a power of 2," << endl
	 << "and by a Manin constant c which is a unit in Z[1/(2m)]," << endl
	 << "where m is the largest square dividing the level N."<<endl; 
    
    MatVecList<F> eigenbasis, anlist;
    ModularFormulas m;
    if(!H->HeckeEigenvalues_an(anlist,eigenbasis,V,
			       int(float(k)*float(m.idxG0(N))/float(12))+1)) {
	cout<<"Could not compute eigenbasis."<<endl;
	return;
    }
    //    for(int i=0;i<anlist.len();i++)
    //      eigenbasis.M(i) = F(get_int("rescale by?"))*eigenbasis.M(i);
    
    Matrix<F> bsd;
    Vector<F> disc;
    cout << endl << "The number of eigenform is " << eigenbasis.len() << endl;
    H->AllBSD (bsd, disc, eigenbasis, anlist);
    for(int i=0;i<anlist.len();i++) {
	cout << endl;
	for(int s=1;s<=k-1;s++) {
	    cout << "\\\\ L(A_f" << i+1 << ","
		 << s << ")/Omega_f = ";
	    Display(bsd(i+1,s)); cout << endl ;
	}
	cout << "\\\\ disc(Z[f]) = " << flush ; Display(disc[i+1]); cout  << endl;
	char ai[512];
	sprintf(ai,"a%d",i+1);
	if(anlist.v(i).poldeg()>1) {
	    cout << ai<<"=Mod(a,";
	    anlist.v(i).polprint(cout,"a");
	    cout << ");"<<endl;
	}
	cout << "f"<<i+1<<" = ";
	for(int j=1;j<=min(7,anlist.M(i).nrows());j++){
	    Vector<F> row;
	    anlist.M(i).row(j,row);
	    if (!row.iszero()) {
		if(j>1) {
		    cout << " + ";
		    if (!(row.poldeg()==0 && row[1]==F(1))) {
			if(row.poldeg()>0) cout << "(";
			if(anlist.v(i).poldeg()<=1)
			    row.polprint(cout,"a");
			else
			    row.polprint(cout,ai);
			if(row.poldeg()>0) cout << ")";
			cout << "*";
		    }
		}
		cout << "q";
		if(j>1)
		    cout << "^"<<j;
	    }
	}
	cout << " + O(q^"<<min(7,anlist.M(i).nrows())+1<<");";
	if(i+1<anlist.len())
	    cout<<endl;
    }
    cout << "}"<<endl;
}

template <class F>
void ModularCalculator<F>::toruscinfinity() 
{
    cout << "NOT IMPLIMENTED."<<endl;
}

template <class F>
void ModularCalculator<F>::torusellc4c6() 
{
    cout << "NOT IMPLIMENTED."<<endl;
}

template <class F>
void ModularCalculator<F>::torusellj() 
{
    cout << "NOT IMPLIMENTED."<<endl;
}

template <class F>
void ModularCalculator<F>::toruslseries() 
{
    cout << "NOT IMPLIMENTED." << endl;
}

template <class F>
void ModularCalculator<F>::torus_bsdratio()
{
    if(!integral) {
	cout << "torusbsdratio requires integral structure."<<endl;
	return;
    }
    int n;
    Vector<F> f;
    cout << "Compute BSD Ratio."<<endl;
    cout << "Selected factor cuts out a subspace of the full S_k."<<endl;
    getheckefactor(n,f);
    if(n==0) return;

    Vector<F> bsd(k-2+1);
    for(int s=1; s<=bsd.size(); s++) 
	H->BSDRatio(bsd[s], n,f,s);
    cout << endl;
    char tn[256];
    ostrstream ost(tn,256);
    ost << "T"<<n<<'\0';

    cout << "P = (";f.polprint(cout,tn);cout<<") subset Hecke algebra"<<endl;
    cout << "A = toric factor corresponding to Sk[P]"<<endl;
    cout << "R(A,s) = L(A,s)/(Omega*(2Pi)^(s-1)*(s-1)!)"<<endl;
    for(int s=1; s<=bsd.size(); s++) {
	cout << "R(A," << s << ") = " ;
	Display(bsd[s]); cout << endl;
    }

    cout << endl << endl;
    for(int s=1; s<=bsd.size(); s++) {
	TexDisplay(bsd[s]);
	cout << ", ";
    }
    cout << endl;

    return;
    if(get_bool("Compute canonical degree? ")) {
	F moddeg= H->ModularDegree(n,f);
	cout << "Canonical degree"<<endl;
	cout << "  = "; Display(moddeg); cout << endl;
    }

    if(get_bool("Compute torsion bound? ")) {
	Messages.off();
	set_V(n,f);
	bigint b=H->TorusTorsionBound(V, 37);
	Messages.on();
	cout << "Torsion bound" << endl;
	cout << "  = "; DisplayFactored(b); cout << endl;
    }


}

template <class F>
void ModularCalculator<F>::torusmoddeg() 
{
    if(!integral) {
	cout << "torusmoddeg requires integral structure."<<endl;
	return;
    }
    int n;
    Vector<F> f;
    cout << "Compute Modular Degree."<<endl;
    cout << "Selected factor cuts out a subspace of the full S_k."<<endl;
    getheckefactor(n,f);
    if(n==0) return;
    F mcusp = H->ModularDegree(n,f);
    cout << endl;
    char tn[256];
    ostrstream ost(tn,256);
    ost << "T"<<n<<'\0';

    cout << "P = (";f.polprint(cout,tn);cout<<") subset Hecke algebra"<<endl;
    cout << "A = toric factor corresponding to Sk[P]"<<endl;
    cout << "m = degree of natural polarization A^{dual}-->A"<<endl;
    cout << "  = "; Display(mcusp); cout << endl;
}

template <class F>
void ModularCalculator<F>::torusintpairing() 
{
    if(!integral) {
	cout << "torusintpairing requires integral structure."<<endl;
	return;
    }
    int n;
    Vector<F> f;
    cout << "Compute Intersection Pairing Matrix."<<endl;
    cout << "Selected factor cuts out a subspace of the full S_k."<<endl;
    getheckefactor(n,f);
    if(n==0) return;
    Matrix<F> Pairing;
    H->IntPairing(Pairing,n,f);
    cout << endl;
    char tn[256];
    ostrstream ost(tn,256);
    ost << "T"<<n<<'\0';

    cout << "P = (";f.polprint(cout,tn);cout<<") subset Hecke algebra"<<endl;
    cout << "A = toric factor corresponding to Sk[P]"<<endl;
    cout << "Pairing Matrix:"<<endl;
    cout << "p= " << Pairing<<endl;
}


template <class F>
void ModularCalculator<F>::torusperiodlattice() 
{
    cout << "NOT IMPLIMENTED." << endl;
}

template <class F>
void ModularCalculator<F>::torustorsion() 
{
    if(!integral) {
	cout << "torustorsion doesn't make sense."<<endl;
	return;
    }
    cout << "compute gcd{2<p<maxp, (p,N)=1: charpoly(T_p|V)(p+1)}" << endl;
    int maxp=get_int("maxp = ");
    if(mk) {
	cout << "Switching to S_k..."<<endl;
	init_V('c');
    }
    bigint b=H->TorusTorsionBound(V, maxp);
    cout << endl << endl << "torustorsionbound = " ;
    DisplayFactored(b);
    cout << endl;
}

template <class F>
void ModularCalculator<F>::torusrealperiod() 
{
    cout << "NOT IMPLIMENTED." << endl;
}

template <class F>
void ModularCalculator<F>::opcharpoly() 
{
    if(charpoly) { 
	charpoly=false;
	cout<<"charpoly display off"<<endl; 
    } else {
	charpoly=true; 
	cout<<"charpoly display on"<<endl;
    }
}

template <class F>
void ModularCalculator<F>::opmatrix() 
{
    if(matrix) {
	matrix=false; 
	cout<<"matrix display off"<<endl; 
    } else {
	matrix=true; 
	cout<<"matrix display on"<<endl; 
    }
}

template <class F>
void ModularCalculator<F>::optrace() 
{
    if(trace) {
	trace=false;
	cout<<"trace display off"<<endl; 
    } else {
	trace=true; 
	cout<<"trace display on"<<endl;
    }
}

template <class F>
void ModularCalculator<F>::compute_Tn() 
{
    int n ;
    cout << "Tn: Enter values of n, then q when done."<<endl;
    while (1) {
	n=get_int(NULL, 0);
	if (n<=0) return ;
	Matrix<F> T;
	if (mk) 
	    H->Tn_Mk(n,T);
	else if (sk)
	    H->Tn_Sk(n,T);
	else
	    H->Tn_On(n,V,T);
	if (matrix)
	    cout << "T"<<n<<"=" << T << ";" << endl ;
	if (charpoly) {
	    cout << "f"<<n<<"=";
	    DisplayCharPoly(T); cout <<";"<< endl ;
	}
	if (trace) {
	    cout << "Tr(T"<<n<<")="<<T.trace()<<endl;
	}
    }
}

template <class F>
void ModularCalculator<F>::act_ZmodNZ()
{
    int u=get_int("Compute action of u in Z/NZ where u = ", 0);
    if(u==0) return;
    Matrix<F> U;
    if (mk) 
	H->ZmodNZ_Mk(U, u);
    else
	H->ZmodNZ_On(U, V, u);

    if (matrix)
	cout << "U " << u << " = " << U << endl ;
    if (charpoly) {
	cout << "charpoly(U"<<u<<") = ";
	DisplayCharPoly(U);cout << endl;
    }
    if (trace) {
	cout << "Tr(U"<<u<<") = "<<U.trace()<<endl;
    }
}

template <class F>
void ModularCalculator<F>::compute_atkin() 
{
    int q=get_int("Compute Wq where p = ", 0);
    if(q==0) return;
    Matrix<F> W;
    if (mk) 
	H->Wq_Mk(W, q);
    else
	H->Wq_On(W, V,q);

    if (matrix)
	cout << "W" << q << " = " << W << endl ;
    if (charpoly) {
	cout << "charpoly(W"<<q<<") = ";
	DisplayCharPoly(W);cout << endl;
    }
    if (trace) {
	cout << "Tr(W"<<q<<") = "<<W.trace()<<endl;
    }
}

template <class F>
void ModularCalculator<F>::compute_L() 
{
    if (k>2) {
	cout << "I only know how to define the winding element when k=2."<<endl;
	return;
    }
    if(mk||sk) { cout << "Please set V first."<<endl; return; }
    cout << "This function will try to compute the Hecke dual of V, and "<<endl
	 << "then compute whether or not <e_w,V>=0, where e_w={0,\\infty}"<<endl
	 << "is the winding element.  If V corresponds to a Galois conjugacy"<<endl
	 << "class of eigenforms, this computes whether or not the corresponding"<<endl
	 << "L-function vanishes.  This function may fail if V is not contained"<<endl
	 << "in the new part of S_2(Gamma_0(N))."<<endl;
    Matrix<F> VDual;
    if(!H->ComputeHeckeDual(VDual, V, 100)) {
	cout << "Unable to find Hecke dual (it may still exist, it's just that the"<<endl
	     << "program only tried used T_n for n<=100 to try to cut out the dual."<<endl;
	return;
    }
    Vector<F> e;
    H->WindingElement(e);
    bool vanishes=true;
    for(int i=1; i<=VDual.ncols() && vanishes==true; i++) 
	if (e*VDual.col(i)!=F(0))
	    vanishes=false;
    if(vanishes)
	cout << "<e,V>==0"<<endl;
    else
	cout << "<e,V>!=0"<<endl;
}

template <class F>
void ModularCalculator<F>::compute_conjugation() 
{
    Matrix<F> C;
    H->Conjugation_Mk(C);
    Matrix<F> B;
    C.RestrictedTo(B, V);
    cout << "C = " << B << endl;
}

template <class F>
void ModularCalculator<F>::compute_conjeigen(bool plus) 
{
    if(V.ncols()<1) return;
    Matrix<F> C, B, I;
    H->Conjugation_Mk(C);
    C.RestrictedTo(B, V);
    I.idmat(B.nrows());
    if(plus)
	B=B-I;
    else
	B=B+I;
    Matrix<F> K;
    Kernel(K,B);
    V=V*K;
    int n=strlen(Vname);
    Vname[n]='^';
    if(plus) 
	Vname[n+1]='+';
    else
	Vname[n+1]='-';
    Vname[n+2]='\0';
    sk=mk=sknew=false;
}

template <class F>
void ModularCalculator<F>::compute_eigen() 
{
    cout << endl << "Compute ker(Tp1-ap1)/\\ker(Tp2-ap2)/\\..."<<endl;
    cout << "Enter a list of prime,eigenvalue pairs (eg, '2 4[ret]  3 10[ret] q'):"<< endl ;
    Array<int> p(500);
    Array<F> e(500);
    int i=0;
    while (1) {
	int n=get_int(NULL,0);
	if(n==0) break; 
	p[i]=n;
	cin >> e[i++];
    }
    if (i==0) return ;
    p.resize(i); e.resize(i);
    Matrix<F> K1, K2;
    if (mk) 
	H->Tn_Mk_Eig(p[0],e[0],K1);
    else if (sk)
	H->Tn_Sk_Eig(p[0],e[0],K1);
    else
	H->Tn_Eig_On(p[0],V,e[0],K1);
    
    for (int j=1; j<p.length(); j++) {
	if (K1.ncols()==0) {
	    V=K1;
	    strcpy(Vname,"{0}");
	    cout << endl << "  dim V=0" << endl ;
	    return ; 
	}
	H->Tn_Eig_On(p[j],K1,e[j],K2);
	K1 = K2 ;
    }
    mk=sk=false;
    ostrstream ost(Vname,1024);
    for(int j=0; j<p.length(); j++) {
	ost << "ker(T" << p[j]<<"+"<<-e[j]<<")";
	if(j+1<p.length())
	    ost<<"/\\";
	else
	    ost<<'\0';
    }
    V=K1;
    cout << "V = " << Vname << endl;
    cout << "Setting V="<<Vname<<", dim V="
	 << V.ncols() << "." << endl;
}


template <class F>
void ModularCalculator<F>::getheckefactor(int& n,
					  Vector<F>& f)
{
    n=get_int("Factor of Tn,  n = ",0);
    if(n<=0) {n=0; return;}
    
    if(H->dim_Mk()<250) {
	cout << "Choose one of the following factors."<<endl;
	Matrix<F> T;
	if (mk) 
	    H->Tn_Mk(n,T);
	else
	    H->Tn_On(n,V,T);
	Vector<F> pol;
        T.charpoly(pol);
	VecFac<F> fac;
	pol.polfactor(fac);
	cout<<endl;
	for(int i=1;i<=fac.numfactors();i++) {
	    cout << "\t" << i << ": " ; 
	    if(fac.exp(i)>1) cout << "(" ;
	    fac.base(i).polprint();
	    if(fac.exp(i)>1) cout << ")^" << fac.exp(i);
	    cout << endl;
	}
	cout << "\t" << fac.numfactors()+1 << ": ALL factors"<<endl;
	int m;
	do {
	    m=get_int("Select a factor: ");
	    if(m<=0) {n=0; return;}
	} while (m>fac.numfactors()+1);
	if(m==fac.numfactors()+1) 
	    f=pol;
	else {
//	    fac.base(m).polpower(f, fac.exp(m));
	    f = fac.base(m);
	}
	    
	cout << "Selected factor = "; f.polprint();
	cout << endl;
    } else {

	cout << "dim(Mk)>=250, so you must enter f."<<endl;
	int d=get_int("  deg f = ", 0);
	Vector<F> f(d+1);
	cout<<"  Enter the " << d+1
	    << " coefficients of f, starting\n  with the coefficient of x^"<<d<<":  ";
	for(int i=d+1;i>=1;i--) {
	    eof_check();
	    cin>>f[i];
	}
    }
}

template<class F>
void ModularCalculator<F>::set_V(int n, const Vector<F>& f)
{
    H->Tn_Mk_Eig(n,f,V);
    mk=sk=false;
    ostrstream ost(Vname,1024);
    Vector<F> fr=f;
    fr.reverse();		// so it will print right.
    ost << "ker(f(T"<<n<<")), f=";
    fr.polprint(ost);
    ost <<'\0';
    cout << "Setting V=ker(f(T" << n << ")), dim="
	 << V.ncols() << "." << endl;
}


template <class F>
void ModularCalculator<F>::compute_gen_eigen() 
{
    cout << endl << "Compute ker(f(Tn))"<<endl;

    int n;
    Vector<F> f;
    getheckefactor(n,f);
    if(n==0) return;
   if (mk) 
	H->Tn_Mk_Eig(n,f,V);
    else if (sk)
	H->Tn_Sk_Eig(n,f,V);
    else {
	Matrix<F> K;
	H->Tn_Eig_On(n,V,f,K);
	V=K; K.init(0);
    }
    mk=sk=false;
    ostrstream ost(Vname,1024);
    Vector<F> fr=f;
    ost << "ker(f(T"<<n<<")|previous), f=";
    fr.polprint(ost);
    ost <<'\0';
    cout << "Setting V=ker(f(T" << n << ")), dim="
	 << V.ncols() << "." << endl;

}


template <class F>
void ModularCalculator<F>::compute_aplist() 
{
    assure_newspace();
    compute_conjeigen(true);
    if(V.ncols()<1) return;

    MatVecList<F> W;
    MatVecList<F> aplist;
    if(!H->HeckeEigenvalues_ap(aplist,W,V,get_int("max p="))) {
	cout<<"Couldn't compute basis."<<endl;
	return;
    }
    cout << endl << " *** LIST OF HECKE EIGENVALUES for V *** "<<endl;
    for(int i=0;i<aplist.len();i++) {
	cout << "E"<<i+1<<" = [";
	if(aplist.v(i).poldeg()>1)
	    aplist.v(i).polprint(cout);
	cout<<",[";
	for(int j=1;j<=aplist.M(i).nrows();j++){
	    Vector<F> row;
	    aplist.M(i).row(j,row);
	    row.polprint(cout);
	    if(j<aplist.M(i).nrows()) cout<<",";
	}
	cout<<"]];"<<endl;
    }
}

template <class F>
void ModularCalculator<F>::assure_newspace() 
{
    if(mk||sk) { 
	cout<<"Switching to S_k^{new}"<<endl;
	mk=sk=false;
	strcpy(Vname, "S_k^{new}");
	H->get_Sknew(V); 
    } else if (strcmp(Vname,"S_k^{new}")){
	cout<<"Warning: make sure V is contained in S_k^{new}."<<endl;
    }
}

template <class F>
void ModularCalculator<F>::compute_anlist() 
{
    assure_newspace();
    compute_conjeigen(true);
    if(V.ncols()<1) return;

    MatVecList<F> anlist;
    if(!H->HeckeEigenvalues_an(anlist,V,get_int("max n="))) {
	cout<<"Couldn't compute basis."<<endl;
	return;
    }
    if(field_type(foo)!=FT_BIGRAT && field_type(foo)!=FT_ALGNUM) {
	cout << "*WARNING*: I have not yet programed hecke to factor polynomials"<<endl
	     << "over fields other than Q and Q(zeta_n), and hence hecke assumed all polynomials"<<endl
	     << "encountered are irreducible when computing this basis."<<endl<<endl;
    }
    cout << endl << " *** BASIS OF NORMALIZED EIGENFORMS for "<<Vname<<" *** "<<endl;
    cout << endl << "{"<<endl;
    for(int i=0;i<anlist.len();i++) {
	char ai[512];
	sprintf(ai,"s",i+1);
	if(anlist.v(i).poldeg()>1) {
	    cout << "s"<<i+1<<"=";
	    anlist.v(i).polprint(cout,"t");
	    cout << ";  ";
	    cout << ai<<"=Mod(t,";
	    anlist.v(i).polprint(cout,"t");
	    cout << ");"<<endl;   
	}
	cout << "f"<<i+1<<" = ";
	for(int j=1;j<=anlist.M(i).nrows();j++){
	    Vector<F> row;
	    anlist.M(i).row(j,row);
	    if (!row.iszero()) {
		if(j>1) {
		    cout << " + ";
		    if (!(row.poldeg()==0 && row[1]==F(1))) {
			if(row.poldeg()>0) cout << "(";
			if(anlist.v(i).poldeg()<=1)
			    row.polprint(cout,"t");
			else
			    row.polprint(cout,ai);
			if(row.poldeg()>0) cout << ")";
			cout << "*";
		    }
		}
		cout << "q";
		if(j>1)
		    cout << "^"<<j;
	    }
	}
	cout << " + O(q^"<<anlist.M(i).nrows()+1<<");";
	if(i+1<anlist.len())
	    cout<<endl<<endl;
    }
    cout << endl << "}"<<endl;
}


template <class F>
void ModularCalculator<F>::basisrational() 
{
    cout << "NOT IMPLIMENTED."<<endl;
}


template <class F>
void ModularCalculator<F>::basisrationalnew() 
{
    assure_newspace();
    compute_conjeigen(true);
    if(V.ncols()<1) return;

    Matrix<F> Q;
    if(!H->Basis_q_expansions(Q,V,get_int("number of terms? "))) {
	cout<<"Couldn't compute basis."<<endl;
	return;
    }
    
    cout << endl << "F-rational basis of q-expansions (NOT eigenforms) for V"<<endl;
    for(int i=1;i<=Q.ncols();i++) {
	Vector<F> col;
	Q.col(i,col);
	cout << "f["<<i<<"] = ";
	Vector<F> q; q.init(2); q[2]=1; // q = "q".
	col=polmul(q,col); // so starts at q. 
	col.polprint(cout,"q",true);
	cout<<";"<<endl;
    }
}

template <class F>
void ModularCalculator<F>::compute_fast_Tp() 
{
    cout << endl
	 << "** Compute Tp on a small subspace quickly. **" << endl
	 << "Enter a list of prime-eigenvalue pairs.  These cut out a rational " << endl
	 << "subspace of Mk.  This function will then quickly compute the action"<<endl
	 << "of Hecke on this subspace." << endl;
    
    cout << "Enter a list of prime,eigenvalue pairs (eg, '2 4[ret]  3 10[ret] q'):"<< endl ;
    Array<int> pr(500);
    Array<F> ev(500);
    int i=0;
    while (1) {
	int n=get_int(NULL,0);
	if(n==0) break;
	if(i>=pr.length()) {
	    mesg("Resizing eigenvalue array.");
	    pr.resize(2*pr.length());
	    ev.resize(pr.length());  
	}
	pr[i]=n;
	cin >> ev[i++];
    }
    pr.resize(i); ev.resize(i);
    Matrix<F> V;
    
    Vector<F>* f=new Vector<F>[ev.length()];
    for(int j=0; j<ev.length(); j++) {
	f[j].init(2);		// f = X - ev[j]
	f[j][2]=1;
	f[j][1]=-ev[j];
    }
    
    H->Dual_Eig_Intersect (V,pr,f);
    
    delete[] f;
    
    cout << "Computing dual E data."<<endl;
    Vector<int> e;
    Matrix<F> EVinv;
    H->Dual_E(e,EVinv,V);
    
    int p ;
    cout << "Quick Tp on dual: Enter values of p, then q when done."<<endl;
    while (1) {
	p=get_int(NULL, 0);
	if (p<=0) return ;
	Matrix<F> T;
	reset_timer();
	H->Dual_Tp_On(T, p, V, e, EVinv);
	cout << "(time = " << get_time() << " s)" << endl;
	if (matrix)
	    cout << "T"<<p<<"=" << T << ";" << endl ;
	if (charpoly) {
	    cout << "f"<<p<<"=";
	    DisplayCharPoly(T); cout <<";"<< endl ;
	}
	if (trace) {
	    cout << "Tr(T"<<p<<")="<<T.trace()<<endl;
	}
    }
}

template <class F>
void ModularCalculator<F>::startup() 
{
    init_V('m');
    matrix=false; charpoly=true; trace=false;
    help();
    status();
}

template <class F>
ModularCalculator<F>::ModularCalculator (char* _title,
					      const Character& _chi,
					      const int& _k,
					      bool _save,
					      bool _plus,
					      bool _integral) 
    : k(_k), chi(_chi), save(_save), plus(_plus),
      integral(_integral), title(_title) 
{ 
    N = chi.modulus(); 
    if(integral)
	H = (H1NChi<F>*) new H1NChiZ(chi,k,save,plus);	// polymorphism!
    else
	H = new H1NChi<F>(chi, k, save, plus);
    startup();
}

// ModularCalculator deletes the H1NChi<F> object, so client code better not!
template <class F>
ModularCalculator<F>::ModularCalculator (char* _title,
					 H1NChi<F>* _H) 
    : chi(_H->get_chi()), title(_title)
{
    H=_H;
    k=H->get_k();
    integral=false;
    plus=H->get_plus();
    save=H->get_save();
    N=chi.modulus();
    startup();
}

template <class F>
ModularCalculator<F>::~ModularCalculator() 
{ 
    delete H; 
}


#ifdef INST_FFIELD
template class ModularCalculator<ffield_element> ;
#endif
#ifdef INST_INTMOD
template class ModularCalculator<intmod> ;
#endif
#ifdef INST_CHARMOD
template class ModularCalculator<charmod> ;
#endif
#ifdef INST_BIGRATIONAL
template class ModularCalculator<bigrational> ;
#endif
#ifdef INST_ALGNUM
template class ModularCalculator<algnum> ;
#endif

void StartupModularCalculator(bool expert, bool save)
{
    cout << endl << "Define a space M_k(N,chi;K) of modular symbols."<<endl;
    if(expert) cout << "(Extended mode.)"<<endl;
    int p=0;
    bool plus=true;
    if(expert) {
	do {
	    p=get_int("characteristic of K = ",0);
	    if(p!=0 && !is_prime(p)) cerr << "characteristic must be prime."<<endl;
	} while (p!=0 && !is_prime(p));
	plus=get_bool("Work in plus space M_k/<I>? ", true);
    } 
    
    Character chi(save);
    int k=get_int("weight k = ",2);
    
    if (p>0) {
	int n=chi.degree();
	if (phi(p)%n==0) {
	    if (3<=p && p<=253) {	// charmod
#ifdef INST_CHARMOD
		InitFieldCharmod(p);
		ModularCalculator<charmod> ha("Top Level", chi,
					      k, save, plus);
		ha.interface1();
#else
#ifdef INST_INTMOD
		InitFieldIntmod(p);
		ModularCalculator<intmod> ha("Top Level",chi,
					     k, save, plus);
		ha.interface1();
#else
#ifdef INST_FFIELD
		InitFieldFiniteCyclo(1,p);
		ModularCalculator<ffield_element> ha("Top Level",chi,
						     k, save, plus);
		ha.interface1();
#else
		cerr << "No finite field support." << endl;
#endif
#endif
#endif
	    } else if (253<p && p<=45083) {	// intmod
#ifdef INST_INTMOD
		InitFieldIntmod(p);
		ModularCalculator<intmod> ha("Top Level",chi, k,
					     save, plus);
		ha.interface1();
#else
		
#ifdef INST_FFIELD
		InitFieldFiniteCyclo(1,p);
		ModularCalculator<ffield_element> ha("Top Level",chi, k,
						     save, plus);
		ha.interface1();
#else
		cerr << "No finite field support." << endl;
#endif
#endif
	    } else {
#ifdef INST_FFIELD
		InitFieldFiniteCyclo(1,p);
		ModularCalculator<ffield_element> ha("Top Level",chi, k,
						     save, plus);
		ha.interface1();
#else
		cerr << "No finite field support." << endl;
#endif
	    }
	} else {			// character of degree > 2 or
				//	characteristic > 32749.	
#ifdef INST_FFIELD
	    InitFieldFiniteCyclo(n,p);
	    ModularCalculator<ffield_element> ha("Top Level",chi, k,
						 save, plus);
	    ha.interface1();
#else
	    cerr << "No finite field support." << endl;
#endif
	}
    } else if (p==0) {
	int n=chi.degree();
	if (n==1 || n==2) {
#ifdef INST_BIGRATIONAL
	    UserInitField(FT_BIGRAT);
	    ModularCalculator<bigrational> ha("Top Level",chi,
					      k, save, plus, true);
	    ha.interface1();
#else
#ifdef INST_ALGNUM
	    InitFieldAlgnumCyclo(n);
	    ModularCalculator<algnum> ha("Top Level",chi, k,
					 save, plus);
	    ha.interface1();
#else 
	    cerr << "No characteristic 0 support."<<endl;
#endif
#endif
	} else {
#ifdef INST_ALGNUM
	    InitFieldAlgnumCyclo(n);
	    ModularCalculator<algnum> ha("Top Level",chi, k,
					 save, plus);
	    ha.interface1();
#else
	    cerr << "The character doesn't take values in Q" << endl
		 << "and there is no number field support."<<endl;
#endif
	}
    } else {
	return ;
    }
    
}


