/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

// motive.cc 

/* WISH LIST
 * Add conversions between manin,modular,and free to the
   calculator interface.
   * Fix Invariant_Symbols_Matrices code (in h1nchi.cc)
     so that when the weight is two it runs through generators
     for Gamma_0(N).  Otherwise it fails already when N=66.

*/

// STABILIZE: -- any line marked this way shouldn't be necessary
//               when Hecke is done and the database is new.

#include "motive.h"
#include "modformulas.h"
#include "display.h"
#include "methodofgraphs.h"

const int DEFAULT_PREC=11;

////////////////// class Motive ///////////////////////////////

// Load Motive
Motive::Motive(bool&                    success,
	       MotiveDecomp*            _Decomp,
	       int                      _true_IsoClass)
{
    init_ins();
    Decomp = _Decomp; in_Decomp=true; delete_Decomp=false;
    true_IsoClass = _true_IsoClass; in_true_IsoClass=true;
    IsFull = false; in_IsFull=true;
    success = Load();
}


// New Motive
Motive::Motive(MotiveDecomp*                   _Decomp,
	       const Matrix<bigrational>&      _Fast,
	       const Vector<bigrational>&      _Field)
{
    init_ins();
    Decomp = _Decomp; in_Decomp=true; delete_Decomp=false;
    true_N = new_N = Decomp->get_N(); in_true_N=in_new_N=true;
    IsFull = false; in_IsFull=true;
    set_Fast(_Fast, _Field);
}

// Full motive
Motive::Motive(bool& loaded,
	       MotiveDecomp*                   _Decomp)
{
    init_ins();
    Decomp = _Decomp; in_Decomp=true; delete_Decomp=false;
    true_N = new_N = Decomp->get_N(); in_true_N=in_new_N=true;
    true_IsoClass = new_IsoClass = 0; in_true_IsoClass=in_new_IsoClass=true;
    IsFull=true; in_IsFull=true; // is a full motive. 
    is_new=false; in_is_new=true;  // certainly not a nice new thing.
    loaded=Load();
}

// Build from old motive.
Motive::Motive(MotiveDecomp* Decomp, Motive& old)
{
    mesgln("Building from old, something at level = ", Decomp->get_N());
  cout << old.get_true_N() << letter(old.get_true_IsoClass())<<endl;
    if(!old.Is_New()) {
	mesgln("Motive::Motive(Decomp, old) -- the old motive *must* be NEW.");
	mesgln("It is at level ", old.get_true_N());
	mesgln("But new at level ", old.get_new_N());
	mesgln("It's true isogeny class is ",old.get_true_IsoClass());
	die();
    }

    init_ins();
    IsFull = false; in_IsFull=true;
    set_Decomp(Decomp);
    set_new_N(old.get_new_N());
    set_new_IsoClass(old.get_new_IsoClass());
    set_true_N(Decomp->get_N());
    dimension=old.get_dimension();
    in_dimension=true;

    int dim_goal = old.get_dimension() * get_multiplicity();
    int p=get_next_goodp();
    H1NChiZ* HPlus=get_H_Plus();
    Matrix<bigrational> F;
    Vector<bigrational>* f=new Vector<bigrational>[1];
    f[0]=old.get_charpoly_Tp(p);
    Array<int> pr(1);
    pr[0]=p;
    
  //cout << "pr[0]="<<p<<", f[0]="<<f[0]<<endl;
    HPlus->Dual_Eig_Intersect(F, pr, f, dim_goal);

    int start_at=1;
    const int step_size=4;
    while(F.ncols()>dim_goal) {	// cut down with step_size more hecke operators
	Matrix<bigrational> Start=F;
	// extend pr and f
	int n=pr.size();
	pr.resize(n+step_size);
	Vector<bigrational>* g=new Vector<bigrational>[pr.size()];
	for(int j=0; j<n; j++)
	    g[j]=f[j];
	delete[] f;
	for(int j=n; j<n+step_size; j++) {
	    pr[j]=get_next_goodp(pr[j-1]);
	    g[j]=old.get_charpoly_Tp(pr[j]);
//	    cout << "pr["<<j<<"]="<<pr[j]<<", f[]="<<g[j]<<endl;
	    if(Fast_Tp.size()<=pr[j]) 
		Fast_Tp.resize(pr[j]+pr[j]/2+1);
	    Fast_Tp.v(pr[j])=g[j]; // this is ok since pr[j] doesn't divide level.
	}
	f=g;
	HPlus->Dual_Eig_Intersect(F,pr,f, Start, start_at, dim_goal);
	start_at+=step_size;
    } ;

    delete[] f;

    set_Fast(F);
}

void
Motive::RaiseOldBasis(const Matrix<bigrational>& OldBasis,
		      int old_true_N)
{
    // Use the old qIntegral basis to set this one.
    // If it is spanned by f_1,...,f_r then this one
    // is spanned over Q by f_1,...,f_r and f_1(q^d),...,f_r(q^d)
    // where d=N/M.
    int d=get_true_N()/old_true_N;
    Vector<bigint> divs;
    Divisors(divs,d);

    Matrix<bigrational> RatBasis;
    RatBasis.init(OldBasis.nrows()*divs.size(),OldBasis.ncols());
    for(int r=1;r<=OldBasis.nrows();r++) 
	for(int i=1; i<=divs.size(); i++) {
	    int div=to_int(divs[i]), 
		rowstart  = (i-1)*OldBasis.nrows();
	    for(int c=1; c<=OldBasis.ncols()/div; c++)
		RatBasis(rowstart + r,div*c)=OldBasis(r,c);		
	}
  cerr << "Old RatBasis = " << RatBasis << endl;
    set_qIntegralBasis(RatBasis);
}

Motive::Motive(const Motive&) 
{ 
    die("Motives can't be copied or assigned."); 
}

Motive& Motive::operator=(const Motive&) 
{ 
    die("Motives can't be copied or assigned."); 
}


void Motive::init_ins()
{
    in_H=in_Name=in_Decomp=in_plus=in_IsFull=in_Prime=in_DefaultPrec
        =in_is_new=in_multiplicity=in_new_N=in_true_N
	=in_true_IsoClass=in_weight
        =in_new_IsoClass=in_dimension=in_true_dimension
	=in_disc=in_c_inf=in_c_p=in_c_p_arith
	=in_hmoddeg=in_xmoddeg=in_ximage=in_xeigen
        =in_TorsionBound=in_DegMP=in_MP=in_KerMP=in_BSDRatio=in_ShaBound
        =in_BSDRatioTimesTorsionBound
	=in_CuspOrder=in_qExpansion=in_qExp_f=in_qIntegralBasis
        =in_L=in_Lambda=in_LambdaPlus=in_LambdaMinus
 	=in_OmegaPlus=in_OmegaMinus=in_ellcurve_invariants
	=in_PhiGoodPeriods=in_GoodPeriods=in_GoodPeriodIntegrals
        =in_GoodPeriodIntegralsCoefficients
	=in_ApproxPeriodMap=in_ApproxPeriodMap_fromImPhi
        =in_Em=in_Phi=in_PhiPlus=in_PhiMinus=in_V=in_Vplus=in_Vminus=in_VD
        =in_Conj_V=in_Conj_Phi=in_Tp_V=in_Tp_Vplus=in_Tp_Vminus=in_Wq
	=in_Tp_Phi=in_Tp_PhiPlus=in_Tp_PhiMinus
	=in_PhiSk=in_PhiSkInverse=in_Conj_PhiSk
	=in_PhiSkPlus=in_PhiSkMinus=in_Intersection
	=delete_H=delete_Decomp=in_Fast
        =false;

    WqString=NULL;
}

Motive::~Motive()
{
    Save();
    if(in_Decomp && delete_Decomp)
	delete Decomp;
    delete WqString;
    if(in_H && delete_H)
	delete H;
}


bool 
Motive::operator<(Motive& m)
{
    /* Definition of ordering (_extends_ Cremona to higher dimension).
       'Smallest' forms come *first* in output lists, so:
         (1) If they have different levels then the larger N is smallest.
         (2) Next compare by IsoClasses, if both are old. 
         (3) Smallest dimension is lesser (to agree with Cremona).
         (4) order by Wq eigenvalues, starting with *smallest* p|N and 
             with "+" being less than "-"
         (5) order by abs(trace(a_p)), with + one being smaller in the
             the event of a tie.
     */

    // (0) 
    if(get_true_N() != m.get_true_N())
	die("Motive::operator<: only motives at the same level can be compared.");

    // now --> both have same true level
    // (1) Level
    if(get_new_N() > m.get_new_N())
	return true;
    if(get_new_N() < m.get_new_N())
	return false;

    // now --> levels at which new are equal
    
    // (2) Old -- IsoClass
    if(!Is_New()) {
	if(m.Is_New())
	    die("Motive::operator<: Both should be old but are not.");
	cerr << "\n get_new_IsoClass = " << get_new_IsoClass()<<endl;
	cerr << "m.get_new_IsoClass = " << m.get_new_IsoClass()<<endl;
	if(get_new_IsoClass()== m.get_new_IsoClass()) 
	    die("Motive::operator< ERROR! both the new level and new IsoClass are the same!");
	return get_new_IsoClass() < m.get_new_IsoClass();
    }

    // now --> both are new.

    // (3) Dimension
    if(get_dimension() < m.get_dimension())
	return true;
    if(get_dimension() > m.get_dimension())
	return false;

    // now --> both have same dimension.

    // (4) by Wq
    int npd=get_num_PrimeDivisorsOfN();
    for(int i=1; i<=npd; i++) {
	int p=get_nth_PrimeDivisorsOfN(i);
	int a=get_Wq(p), b=m.get_Wq(p);
	if (a < b)
	    return false;
	if (a > b)
	    return true;
    }

	    
    // now --> Wq are the same

    // (5) by trace(ap)
    int p=2;
    while(1) {
	while(get_true_N() % p == 0) p=next_prime(p);
	bigrational a=get_TraceOf_ap(p), b=m.get_TraceOf_ap(p);
	if(a!=b) {
	    if (abs(a) < abs(b))
		return true;
	    if (abs(a) > abs(b)) 
		return false;
	    if(a>0)
		return true;
	    else
		return false;
	}
	p=next_prime(p);
	if(p>97)		// escape...
	    die("Motive::operator<: no diff up to a_97 -- compare a motive with itself?");
    }
    
    die("Motive::operator<: never reached.");
    return false;  // never reached.
}

void
Motive::set_Phi(const Matrix<bigrational>& _Phi)
{
    Phi=_Phi;
    in_Phi=true;
}


void 
Motive::set_Decomp(MotiveDecomp* p_Decomp)
{
    in_Decomp=true;
    delete_Decomp=false;
    Decomp=p_Decomp;
}


const char*
Motive::get_WqString()
{
    if(WqString != NULL)
	return WqString;
    WqString = new char[256];
    ostrstream out(WqString,256);
    for(int j=1; j<=get_num_PrimeDivisorsOfN(); j++) {
	int s=get_Wq(get_nth_PrimeDivisorsOfN(j));
	if(s==1)
	    out << "+";
	else
	    out << "-";
    }
    out<<'\0';
    return WqString;
}

void
Motive::write(OUTTYPE type, ostream& out, 
	      bool brief, int qPrec)
{
    bool isfirst;
    int n;
    switch(type) {
    case LATEX:
	if(!Is_New()) return;
	out<<"\\motiveA{"<<letter(get_true_IsoClass())
	   <<"}{" << get_dimension()
	   <<"}{"; TexDisplay(get_disc(),out);
	out<<"}{" << get_WqString()
	   <<"}{";
	for(int j=1; j<=get_num_PrimeDivisorsOfN(); j++) {
	    bigint cp=get_c_p_arith(get_nth_PrimeDivisorsOfN(j));
	    if(cp==0) out << "?"; else out << cp;
	}
	out << "}{"; TexDisplay(get_c_inf(),out);
	out << "}{" << get_CuspOrder();
	out << "}{" << get_TorsionUpperBound();
	out << "}{" << get_BSDRatioTimesTorsionBound();
	out << "}";
	out << "\\motiveB"
	    << "{"  << get_ShaBound()
	    << "}{" << get_DegMP();
	out << "}{";
	// Intersections
        isfirst=true;
	for(int j=1;j<=get_Decomp()->num_Motives();j++)
	    if(j!=get_true_IsoClass()) {
		Vector<bigint> inter=get_Intersection(j);
		if(inter.size()==0)
		    continue;
		if(!isfirst) out << "+"; else isfirst=false; 
		out << letter(j); out << "_{" ;
		const Vector<bigint>& v=get_Intersection(j);
		for(int i=1; i<=v.size(); i++) {
		    out<<v[i];
		    if(i<v.size()) out<<",";
		}
		out<<"}";
	    }
	if(isfirst) out << "\\mbox{}";
	out << "}\\\\"<<endl;
	break;
    case TEXT:
	if(Is_Full()) {
	    out << "J_0(" << get_true_N() << "), dim = " << get_dimension() << endl;
	    return;
	} 

	out << get_true_N() << letter(get_true_IsoClass());

	if(Is_New()) out << " (new) "; else
	    out << " (old = " << get_new_N() << letter(get_new_IsoClass()) << ")";
	out << ", dim = " << get_dimension() << endl;
	
	out << endl << "CONGRUENCES:" << endl;
	out << "    Modular Degree = "; Display(get_DegMP(),out);  out<<endl;
	out << "    Ker(ModPolar)  = "; DisplayAbGp(get_KerMP(),out); out <<endl;
	if(get_Decomp()->num_Motives()>1) {
	    out << "                   = ";
	    
	    isfirst=true;
	    for(int j=1;j<=get_Decomp()->num_Motives();j++)
		if(j!=get_true_IsoClass()) {
		    Vector<bigint> inter=get_Intersection(j);
		    if(inter.size()==0)
			continue;
		    if(!isfirst) out << " + "; else isfirst=false; 
		    out << letter(j); out << "(" ;
		    DisplayAbGp(get_Intersection(j),out);
		    out << ")";
		}
	    out<<endl;
	}


	if(Is_New()) {
	    out << endl << "ARITHMETIC INVARIANTS:"<<endl;
	    out << "    W_q            = " << get_WqString() << endl;
	    out << "    discriminant   = "; Display(get_disc(),out); out << endl;
	    if(get_weight()==2) {
		out << " #CompGroup(Fpbar) = ";
		for(int j=1; j<=get_num_PrimeDivisorsOfN(); j++) {
		    bigint cp=get_c_p(get_nth_PrimeDivisorsOfN(j));
		    if(cp==0) out << "?"; else Display(cp,out);
		}out << endl;
		out << "    c_p            = ";
		for(int j=1; j<=get_num_PrimeDivisorsOfN(); j++) {
		    bigint cp=get_c_p_arith(get_nth_PrimeDivisorsOfN(j));
		    if(cp==0) out << "?"; else Display(cp,out);
		}out << endl;
		out << "    c_inf          = "; Display(get_c_inf(),out);
		out << endl;
		out << "    ord((0)-(oo))  = "; Display(get_CuspOrder(),out);
		out<<endl;
		out << "    Torsion Bound  = ";
		    Display(get_TorsionUpperBound(),out); out << endl;
	    }
	    for(int s=1; s<=get_weight()-1; s++) {
		out << "    |L("<<s<<")/Omega|";
		if(s>1) out << "*(s-1)!/(2pi)^(s-1)";
		out << "   = "; Display(get_BSDRatio(s),out); out << endl;
	    }
	    if(get_weight()==2) {
		out << "    Sha Bound      = "; Display(get_ShaBound(),out);
		out << endl;
	    }

	    get_qIntegralBasis(qPrec);

	    out << endl << "ANALYTIC INVARIANTS:" << endl << endl;
	    out << "    Omega+         = "; Display(get_OmegaPlus(),out); out<<endl;
	    out << "    Omega-         = "; Display(get_OmegaMinus(),out); out<<endl;
	    for(int s=1; s<=get_weight()-1; s++) {
		out << "    L("<<s<<")           = ";
		Display(get_L(s),out);
		out<<endl;
	    }
	    if(get_dimension()==1) {
		out << "    w1             = "; Display(get_w1(),out); out<<endl;
		out << "    w2             = "; Display(get_w2(),out); out<<endl;
		out << "    c4             = "; Display(get_c4(),out); out<<endl;
		out << "    c6             = "; Display(get_c6(),out); out<<endl;
		out << "    j              = "; Display(get_ellj(),out); out<<endl;
	    }

	    out << endl << "HECKE EIGENFORM:"<<endl;
	    if(get_qExp_f().poldeg()>1) {
		get_qExp_f().polprint(out,"a");
		out << " = 0,"<<endl;
	    } 
	    out << "f(q) = q + ";
	    for(int n=2; n<=qPrec; n++) {
		const Vector<bigrational>& a =
		    get_qExpansion().col(n);
		if(!a.iszero()) {
		    if(a.numnonzero()>1) out << "(";
		    a.polprint(out,"a");
		    if(a.numnonzero()>1) out << ")";
		    out << "*q^" << n;
		    out << " + ";
		}
	    }
	    out<< " ... " << endl;
	}
	out<<endl;
	out << endl << "-------------------------------------------------------"<<endl ;
	break;
    case PARI:
	out<<"M["<<get_true_N()<<","<<get_true_IsoClass()<<"] = ["
	   << get_new_N()
	   << ", " << get_new_IsoClass();
	if(!Is_New()) { out << "];"<<endl; return; }
	out << ", " << get_dimension()
	   << ", " << get_disc()
	   << ", \"" << get_WqString() << "\""
	   << ", [";
	n=get_num_PrimeDivisorsOfN();
	for(int j=1; j<=n; j++) {
	    bigint cp=get_c_p_arith(get_nth_PrimeDivisorsOfN(j));
	    out << cp;
	    if(j<n) out << ",";
	}
	out <<"], " << get_c_inf() 
	    << ", " << get_CuspOrder()
	    << ", " << get_TorsionUpperBound()
	    << ", " << get_BSDRatio(1)*get_c_inf()*get_TorsionUpperBound()
	    << ", " << get_ShaBound()
	    << ", " << get_DegMP() 
	    << ", ";

	// Intersections
	out << "[";
	isfirst=true;
	for(int j=1;j<=get_Decomp()->num_Motives();j++)
	    if(j!=get_true_IsoClass()) {
		if(!isfirst) out << ",";
		isfirst=false;
		out << "["<<j<<",";
		Vector<bigint> inter=get_Intersection(j);
		if(inter.size()==0) 
		    out << "[0]";
		else 
		    out << get_Intersection(j);
		out << "]";
	    }
	out << "]";

	// finish.
	out << "];" <<endl;
	break;
	
    default:
	die("Motive::write -- unreckognized option.");
    };
    
}

bool
Motive::Is_New()
{
    if(in_is_new)
	return is_new;
    is_new = (get_true_N() == get_new_N());
    in_is_new=true;
    return is_new;
}

bool
Motive::Is_Full()
{
    if(!in_IsFull) 
	die("Motive::Is_Full -- no way to determine if full.");
    return IsFull;
}

bool
Motive::Is_Plus()
{
    if (in_plus)
	return plus;
    plus=get_Decomp()->Is_Plus();
    in_plus=true;
    return plus;
}

int
Motive::get_multiplicity()
{
    if(in_multiplicity)
	return multiplicity;
    int d = get_true_N() / get_new_N();
    multiplicity = num_divs(d);
    in_multiplicity=true;
    if(multiplicity<=0)
	die("Motive::get_multiplicity -- negative, bizarre error.");
    return multiplicity;
}

int
Motive::get_new_N()
{
    if(in_new_N)
	return new_N;
    die("Motive::get_new_N, no way to determine new_N.");
}

int
Motive::get_new_IsoClass()
{
    if(in_new_IsoClass)
	return new_IsoClass;
    mesgln("in_new_IsoClass = ");
    cerr<<in_new_IsoClass;
    mesgln("new_IsoClass = ",new_IsoClass);
    die("Motive::get_new_IsoClass, no way to determine new_IsoClass.");
}

int 
Motive::get_true_N()
{
    if(in_true_N)
	return true_N;
    true_N = get_Decomp()->get_N();
    in_true_N = true;
    return true_N;
}

int
Motive::get_true_IsoClass()
{
    if(in_true_IsoClass)
	return true_IsoClass;
    mesgln("true_IsoClass = ",true_IsoClass);
    mesgln("true_N = ", true_N);
    die("Motive::get_true_IsoClass -- don't know how to get IsoClass.");
    return -1;			// to avoid warning.
}

int
Motive::get_weight()
{
    if(in_weight)
	return weight;
    weight = get_Decomp()->get_weight();
    in_weight= true;
    return weight;
}


void
Motive::set_new_N(int N)
{
    in_new_N=true;
    new_N = N;
}

void      
Motive::set_new_IsoClass(int iso)
{
    in_new_IsoClass=true;
    new_IsoClass=iso;
    if(Is_New()) {
	true_IsoClass=iso;
	in_true_IsoClass=true;
    }
}

void
Motive::set_true_N(int N)
{
    in_true_N=true;
    true_N=N;
}

void
Motive::set_true_IsoClass(int iso)
{
    in_true_IsoClass=true;
    true_IsoClass=iso;
    if(Is_New()) {
	new_IsoClass=iso;
	in_new_IsoClass=true;
    }
}

void              
Motive::set_weight(int _weight)
{
    in_weight=true;
    weight=_weight;
}


int
Motive::get_DefaultPrec()
{
    if(in_DefaultPrec)
	return DefaultPrec;
    DefaultPrec=DEFAULT_PREC;
    in_DefaultPrec=true;
    return DefaultPrec;
}

void
Motive::set_DefaultPrec(int prec)
{
    in_DefaultPrec=true;
    DefaultPrec=prec;
}


const int&  
Motive::get_dimension()
{
    if(in_dimension)
	return dimension;
    dimension=get_Fast().ncols();
    in_dimension=true;
    return dimension;
}

const int&  
Motive::get_true_dimension()
{
    if(in_true_dimension)
	return true_dimension;

    true_dimension = get_Fast().ncols();
    if(!Is_Plus()) true_dimension*=2;
    in_true_dimension=true;
    return true_dimension;
}


const bigrational&
Motive::get_disc()
{
    if(in_disc)
	return disc;
    Matrix<bigrational> Gens;
    ModularFormulas formulas;
    int sturm= get_weight()*formulas.idxG0(get_true_N())/12+1;	  // Sturm's bound.
    get_qExpansion(sturm).transpose(Gens);
    disc = ZAlgDisc(get_qExp_f(), Gens);
    in_disc=true;
    return disc;
}

const bigint&
Motive::get_c_inf() 
{
    if(in_c_inf)
	return c_inf;

    // c_inf <-- 2^(dim(ker(Cmod2-1))-d),    (# connected components)
    Matrix<bigrational> C(get_Conj_PhiSk());
    for(int i=1;i<=C.nrows();i++) C(i,i) = C(i,i)-1;
    int dimker=RankModP(C,2);
    int d=get_dimension();
    power(c_inf,2,dimker-d);
    in_c_inf=true;
    return c_inf;
}

const bigint&                  
Motive::get_c_p(int p) 
{
    if(!is_prime(p))
	die("Motive::get_c_p(p), p not prime.");
    // Why not allow it -- maybe it is true?
    // Just carefully define c_p to be the number computed in this way
    // in the documentation.
    if(!(Is_New()||Is_Full()))
        die("get_c_p -- motive must be full or new.");
    if(get_new_N() % p != 0)
	die("Motive::get_c_p-- p must divide the level.");
    if(in_c_p && c_p[get_PrimeIndex(p)]!=0) // know it and have computed it...
	// the second condition that c_p[]!=0 can be removed once HECKE becomes stable.
	return c_p[get_PrimeIndex(p)];

    // Initialize c_p vector.
    in_c_p=true;
    c_p.init(get_num_PrimeDivisorsOfN());
    if(Is_New() && !Is_Full() && is_prime(get_true_N())) { // know how to compute c_p.
	MethodOfGraphs X(get_true_N());// X = character group of torus of J_0(p).
	IntVecList<bigrational> I(4);
	I.Z(0)=2;
	I.v(0)=get_charpoly_Tp(2);
	bigint image, mx;
	X.coker_mx(image, mx, I); 

	if(mx==0) { // T_2 wasn't enough, add more and try again.
	    cerr << "T_2 not enough for "
		 <<get_true_N()<<letter(get_true_IsoClass())
		 <<endl;
	    int i=1;
	    for(int p=3; p<=7; p=next_prime(p)) {
		I.Z(i)=p;
		I.v(i++)=get_charpoly_Tp(p);
	    }
	    X.coker_mx(image, mx, I); 
	}
	
	if(mx!=0)	{	// success
	    // c_p = image * (homology modular degree) / (mx=compgroup modular degree)
	    bigint m=get_DegMP();
	    if(Is_Plus()) {	// kill all two's
		while(image % 2==0 && image>0) image /=2;
		while(m % 2==0 && m>0) m /=2;
		while(mx % 2==0 && mx>0) mx /=2;
	    }
	    if(m % mx != 0) {
		cerr << "The comp group modular degree = "<<mx
		     << " doesn't seem to divide" <<endl
		     << "the homology modular = " << m
		     << ", a contradiction."<<endl
		     << "N = " << get_true_N() << endl
		     << "iso = " << get_true_IsoClass() << endl;
		die("This contradicts everything!");
	    }
	    c_p[1] = image * m / mx;
	    if(c_p[1] != image) {
		cerr << "The comp group modular degree = "<<mx
		     << " doesn't equal" <<endl
		     << "the homology modular = " << m<<" !!!! "<<endl
		     << "!!!!! YOU'VE FOUND A FASCINATING EXAMPLE IN WHICH THE MAP " << endl
		     << "ON COMP GROUPS DOESN'T SURJECT!!!!"<<endl
		     << "N = " << get_true_N() << endl
		     << "iso = " << get_true_IsoClass() << endl
		     << "PLEASE REPORT TO was@math.berkeley.edu immediately!!!"<<endl;
		die("Shutting down, so that you'll definitely see this message!");
	    }		
	    // Now, because of the previous check, the component
	    // group is cyclic, being the image of a cyclic group,
	    // by Mazur's Eisenstein ideal paper and the fact that
	    // the level is prime.  We can thus compute c_p_arith.
	    in_c_p_arith=true;
	    c_p_arith.init(get_num_PrimeDivisorsOfN());
	    if(get_Wq(get_true_N()) == 1) { // nontrivial Galois action
		if(c_p[1]%2 == 0)
		    c_p_arith[1] = 2; // the cyclic group has even order,
                            	      // so it has nontrivial 2-torsion
		else
		    c_p_arith[1] = 1; // the cyclic group has odd order,
				      // so it has no 2-torsion
	    } else {	// trivial galois action
		c_p_arith[1] = c_p[1];
	    }

	    // Save XImage and XModDeg.
	    XModDeg.init(1); XModDeg[1]=mx; in_xmoddeg=true;
	    XImage.init(1); XImage[1]=image; in_ximage=true;
	}
    }

    return c_p[get_PrimeIndex(p)];
}

const bigint&
Motive::get_c_p_arith(int p)
{
    if(!(Is_New()||Is_Full()))
       die("Motive::get_c_p_arith(p) -- motive must be full or new.");
    if(get_new_N()%p!=0)
	die("Motive::get_c_p_arith-- p must divide the level.");
    if(in_c_p_arith) 
	return c_p_arith[get_PrimeIndex(p)];
    get_c_p(p);		       // this might be able to initialize c_p_arith...
    if(in_c_p_arith) 
	return c_p_arith[get_PrimeIndex(p)];

    // no way to initialize, so set to the "we can't compute" value of 0.
    c_p_arith.init(get_num_PrimeDivisorsOfN());
    return c_p_arith[get_PrimeIndex(p)];
}

const bigint&
Motive::get_HModDeg(int p)
{
    if(!is_prime(p))
	die("Motive::get_HModDeg(p), p not prime.");
    if(get_true_N() % p != 0)
	die("Motive::get_HModDeg-- p must divide the level.");
    
    if(get_true_N() != p)	// will be changed, obviously!
	die("get_HModDeg(p) not implimented for nonprime level!");
    else {
	in_hmoddeg=true;
	HModDeg.init(1);
	HModDeg[1]=get_DegMP();
	return HModDeg[1];
    }
    die("get_HModDeg(p): this point in code shouldn't be reached.");
}

const bigint&
Motive::get_XModDeg(int p)
{
    if(!is_prime(p))
	die("Motive::get_XModDeg(p), p not prime.");
    if(get_true_N() % p != 0)
	die("Motive::get_XModDeg(p) -- p must divide the level.");
    if(get_new_N() % p != 0)
	die("Motive::get_XModDeg(p) -- p must divide the level at which form is new.");
    if(in_xmoddeg && XModDeg[get_PrimeIndex(p)]!=0) // know it and have computed it...
	// the second condition that c_p[]!=0 can be removed once HECKE becomes stable.
	return XModDeg[get_PrimeIndex(p)];

    // the only way to initialize xmoddeg is in the midst of computing c_p, 
    // so we do that.
    mesgln("Motive::get_XModDeg(p) -- Attempting to initialize XModDeg by computing c_p...");
    in_c_p=false;		// STABILIZE: this won't be necessary.
    get_c_p(p);
    if(in_xmoddeg) // found it
	return XModDeg[get_PrimeIndex(p)];

    mesgln("Motive::get_XModDeg(p) -- Couldn't compute XModDeg. Setting equal to 0.");
    XModDeg.init(get_num_PrimeDivisorsOfN());
    return XModDeg[get_PrimeIndex(p)];
}
const bigint&
Motive::get_XImage(int p)
{
    if(!is_prime(p))
	die("Motive::get_XImage(p), p not prime.");
    if(get_true_N() % p != 0)
	die("Motive::get_XImage(p) -- p must divide the level.");
    if(get_new_N() % p != 0)
	die("Motive::get_XImage(p) -- p must divide the level at which form is new.");
    if(in_ximage && XImage[get_PrimeIndex(p)]!=0) // know it and have computed it...
	// the second condition that c_p[]!=0 can be removed once HECKE becomes stable.
	return XImage[get_PrimeIndex(p)];

    // the only way to initialize XImage is in the midst of computing c_p, 
    // so we do that.
    mesgln("Motive::get_XImage(p) -- Attempting to initialize XImage by computing c_p...");
    in_c_p=false;		// STABILIZE: this won't be necessary.
    get_c_p(p);
    if(in_ximage) // found it
	return XImage[get_PrimeIndex(p)];

    mesgln("Motive::get_XImage(p) -- Couldn't compute XImage. Setting equal to 0.");
    XImage.init(get_num_PrimeDivisorsOfN());
    return XImage[get_PrimeIndex(p)];   // SLIGHTLY DANGEROUS...
}

const Vector<bigrational>& 
Motive::get_XEigenf(int ell)
{
    if(!in_xeigen)
	get_XEigenV(ell);
    return XEigenf;
}

const Matrix<bigrational>& 
Motive::get_XEigenV(int ell)
{
    if(!is_prime(get_true_N()))
	die("Motive::get_XEigen only implimented for prime level.");

    if(in_xeigen) return XEigenV;
    
    // Find a single ell so that T_ell acts irreducibly on this motive.
    // I'm using this "silly" algorithm because I would like to find the
    // first example where no Q_f =/= Q(a_ell) for ell=2,3,5,7 (and prime level!)
    int p=2;
    XEigenf=get_charpoly_Tp(p);
    Matrix<bigrational> T, E;
    MethodOfGraphs X(get_true_N());// X = character group of torus of J_0(p).
    X.Tp(T, p);
    Kernel(E, T, XEigenf);
    while(E.ncols()>XEigenf.poldeg()) {
	p=next_prime(p);
	if(p>7) 
	    die("Motive::get_XEigenV -- Not possible to break up the space using only T2, T3, T5, T7");
	XEigenf=get_charpoly_Tp(p);
	X.Tp(T, p);
	Matrix<bigrational> TT, EE;
	T.RestrictedTo(TT, E);
	Kernel(EE, TT, XEigenf);
	E=E*EE;
    }
    p=2;
    while(p<=7 && !get_charpoly_Tp(p).pol_is_irreducible()) 
	p=next_prime(p);
    if(p>7) 
	die("Motive::get_XEigenV -- No T_p, p=2,3,5,7 acts irreduncibly.");

    X.Tp(T, p);
    Matrix<bigrational> A;
    T.RestrictedTo(A, E);
    Matrix<bigrational> V, v;
    Eigenvector(v, A, XEigenf);
    XEigenV = E*v;
    in_xeigen=true;
    return XEigenV;
}

Vector<ffield_element>
Motive::get_XSSBasis(int p)
{
    if(!is_prime(get_true_N()))
	die("Motive::get_XSSBasis only implimented for prime level.");
    MethodOfGraphs X(get_true_N());
    return X.ssbasis();
}

Vector<int>
Motive::get_XMonodromy(int p)
{
    if(!is_prime(get_true_N()))
	die("Motive::get_XSSMonodromy only implimented for prime level.");
    MethodOfGraphs X(get_true_N());
    return X.get_weights();
}

/* get_CuspOrder: Compute the order of the image of the cusp (0)-(oo) in Motive.
                  (The meaning of this in higher weight isn't clear to me now.)
   ALGORITHM: 1. Write w={0,oo} in M_k(Q) in terms of free basis.
              2. Compute the image of w in PhiSk via the Phi map.
              3. Find the smallest multiple of e which lies in PhiSk
                 by writing w in terms of a basis for PhiSk.
*/
const bigint& 
Motive::get_CuspOrder()
{
    if(in_CuspOrder)
	return CuspOrder;

    CuspOrder = get_CuspOrder(0,1,1,0,0);
    in_CuspOrder=true;
    return CuspOrder;
}

// Compute the order of the difference (a/b)-(c/d) of cusps.
bigint
Motive::get_CuspOrder(const bigint& a, const bigint& b,
		      const bigint& c, const bigint& d,
		      int q=0)
{
    Vector<bigrational> x;
    get_H()->modular_to_free(x,a,b,c,d,q);
    Vector<bigrational> v=get_PhiSkInverse()*(get_Phi()*x);
    bigint order;
    Vector_denominator(order, v);
    return order;
}

bigint
Motive::get_CuspOrder(const bigrational& cusp, int q=0)
{
    return get_CuspOrder(numerator(cusp), denominator(cusp), 1, 0, q);
}

const bigint&
Motive::get_TorsionUpperBound(int prec)
{
    prec=next_prime(prec-1);
    if(in_TorsionBound && prec<=TorsionBound_prec)
	return TorsionBound;
    
    int p;
    if(in_TorsionBound) 
	p=TorsionBound_prec;
    else {
	TorsionBound=0;
	p=3;
    }
    while(p<=prec) {
	if(get_true_N()%p!=0) {
	    bigint a=get_charpoly_Tp(p).polsubst(p+1).numerator();
	    if(TorsionBound==0)
		TorsionBound = a;
	    else
		TorsionBound = gcd(TorsionBound,a); 
	}
	p=next_prime(p);
    }
    in_TorsionBound=true;
    TorsionBound_prec=prec;
    return TorsionBound;
}

const Matrix<bigrational>& 
Motive::get_MP()
{
    if (in_MP)
	return MP;
    MP = get_VDual()*get_V();
    in_MP=true;
    return MP;
}


const bigint&
Motive::get_DegMP() 
{
    if(in_DegMP)
	return DegMP;

    bigrational det=abs(get_MP().determinant());
    if(det.denominator()!=1)
	die("Motive::get_DegMP -- not an int?");
    bigint d=numerator(det);
    if(Is_Plus()) {
	mesgln("Motive::get_DegMP -- WARNING, modular degree off by a power of 2.");
	DegMP = d;
    } else {
	if(!issquare(DegMP, d))	// sets DegMP = sqrt(d).
	    die("Motive::get_DegMP -- not a square!");
    }
    in_DegMP=true;
    return DegMP;
}

const Vector<bigint>&     
Motive::get_KerMP()
{
    if(in_KerMP)
	return KerMP;
    Vector<bigint> divs;
    ElementaryDivisors(divs, get_MP());
    int num=0;
    for(int i=1; i<=divs.size(); i++) {
	if(divs[i]!=1)
	    num++;
	if(divs[i]==0) {
	    cerr << "true_N = " << true_N<<endl
		 << "true_IsoClass = " << true_IsoClass<<endl;
	    cerr << "get_MP = " << get_MP()<<endl;
	    die("Motive::get_KerMP -- invalid kernel, has 0's");
	}
    }
    KerMP.init(num);
    int j=1;
    for(int i=1; i<=divs.size(); i++)
	if(divs[i]!=1)
	    KerMP[j++]=divs[i];
    in_KerMP=true;
    return KerMP;
}

const bigrational&
Motive::get_BSDRatio(int s) 
{
    int k=get_weight();
    if(s<1 || s>k-1) 
	die("Motive::get_BSDRatio -- s is not critical.");
    if(in_BSDRatio) {
	if(BSDRatio[s]!=-1)
	    return BSDRatio[s];
    }
    
    if(!in_BSDRatio) {
	BSDRatio.init(k-1);
	for(int i=1;i<=k-1;i++)
	    BSDRatio[i]=-1;
	in_BSDRatio=true;
    }

    return (BSDRatio[s] = get_BSDBound(s, 0));
}

bigrational
Motive::get_BSDBound(int s, int bound)
{
    mesgln("get_BSDBound b=",bound);
    Vector<bigrational> e;
    get_H()->WindingElement(e,s-1);
    Matrix<bigrational> PTe;
    PhiTx(PTe, e, bound);
    mesgln("get_BSDBound done.");
    bigrational val;
    if(s%2)
	val =
	    ModuleIndex(get_PhiSkPlus(), PTe) / bigrational(get_c_inf());
    else
	val =
	    ModuleIndex(get_PhiSkMinus(), PTe) / bigrational(get_c_inf());
    return val;
}


const bigint&                  
Motive::get_BSDRatioTimesTorsionBound()
{
    if(get_weight()>2)
	die("Motive::get_BSDRatioTimesTorsionBound -- don't call when k>2.");
    if(in_BSDRatioTimesTorsionBound)
	return BSDRatioTimesTorsionBound;
    bigrational a = get_TorsionUpperBound() * get_TorsionUpperBound() * get_BSDRatio(1);
    if(a.denominator()!=1) {
	mesgln("Motive::get_BSDRatioTimesTorsionBound -- the BSD conjecture is false.");
	mesgln("true_N = ", get_true_N());
	mesgln("new_N = ", get_new_N());
	mesgln("new_IsoClass = ", get_new_IsoClass());
	die();
    }
    BSDRatioTimesTorsionBound = a.numerator();
    in_BSDRatioTimesTorsionBound = true;
    return BSDRatioTimesTorsionBound;
}

// Analytic Sha is bounded by TorsionBound^2 * BSDRatio / Product c_p_arith.
const bigint&
Motive::get_ShaBound()
{
    if(in_ShaBound)
	return ShaBound;
    bigrational prodc_p=1, a;
    for(int i=1; i<=get_num_PrimeDivisorsOfN(); i++)
	if((a=get_c_p_arith( get_nth_PrimeDivisorsOfN(i) )) > 0)
	    prodc_p *= a;	// 0 means "couldn't compute it".
    a = bigrational(get_BSDRatioTimesTorsionBound()) / prodc_p;
    if(denominator(a)!=1) {
	mesgln("Motive::get_ShaBound -- the BSD conjecture is false?!");
	mesgln("true_N = ", get_true_N());
	mesgln("new_N = ", get_new_N());
	mesgln("new_IsoClass = ", get_new_IsoClass());
	die();
    }
    ShaBound = numerator(a);
    in_ShaBound = true;
    return ShaBound;
}

bool
Motive::Is_LchiNone_nz(const Character& chi)
{
    if (get_weight()!=2) 
	die("Motive::Is_Lchione_nz -- I can't compute this when weight > 2.");
    if (!Is_New())
	die("Motive::Is_Lchione_nz -- motive must be new.");
    if (chi.modulus() != get_true_N())
	die("Motive::Is_Lchione_nz -- character must be modulo level N.");
    /* For now, *ASSUMES* weight 2.
       Let phi be a left (dual) eigenvectors corresponding to this motive.
       We compute whether or not the quantity
                  phi(sum chi(a) {oo, a/N}) 
       vanishes, where the sum is over a in (Z/NZ)^*.
       For now, we assume chi is a Dirichlet character modulo N.
       Equivalently, upon applying Fricke W_N=[0,-1;N,0], 
       we must compute whether or not the quantity 
                 =phi ( sum chi(a) {0,-1/a})
                 =phi ( sum chi(a) [-a:1] )
       is nonzero.  The latter is advantageous because it is a sum 
       of Manin symbols.
       Algorithm: 1) compute phi
                  2) compute the vector s = sum chi(a) [-a:1]  
                     in the space M_2(N) tensor Q(zeta_N).
                  3) compute the dot product and check if it is
                     zero or not.  For the last step we do not
                     have to work over the field of definition of phi,
                     by writing phi = sum v_i * alp^i and checking
                     that v_i(s)=0 for all i (or not). 
    */

    int N = get_true_N();
    // Step 1) Compute phi -- if odd character get 
    //         minus space, if even get the plus space. 
    Matrix<bigcomplex> phi;
    if(chi.of(-1)==rootof1(-1))   // odd character
      to_bigcomplex(phi, get_PhiMinus());
    else                          // even character
      to_bigcomplex(phi, get_PhiPlus());

//    to_bigcomplex(phi,get_Fast());
//    phi.transpose();		// so we can view it as row vectors.

    // OPTIMIZE: one can easily change the definition of S below 
    // so it isn't necessary to transpose phi.  

    // Step 2) Compute s.
    //  Form the sum.
    H1NChiZ* H = get_H();
    Vector<bigcomplex> s;
    s.init(H->dim_Mk());
    for(int a=1; a<N; a++) 
	if(gcd(a,N)==1) {
	    Vector<bigrational> vec;
	    H->modular_to_free(vec,1,0,a,N,0);
	    Vector<bigcomplex> comvec;
	    to_bigcomplex(comvec,vec);
	    s += bigcomplex(chi.of(a))*comvec;
	}

    // Step 3) Determine if L vanishes.
    cout << "Phi = " << phi << endl;
    cout << "s = " << s << endl;
    Matrix<bigcomplex> phis = phi*s;
    cout << "Phi * s = " << phis << endl;
    bigfloat tolerance = 1/bigfloat(10e10);
    for(int r=1; r<=phis.nrows(); r++)
	for(int c=1; c<=phis.ncols(); c++)
	    if(abs(phis(r,c))>tolerance)
		return true;	// it's nonzero!
    return false;		// all entries appear to be zero.
}     


void
Motive::get_Fast_TpE(Vector<bigrational>& v, int p)
{
    if(!in_Fast) get_Fast();
    get_H_Plus()->Tp_StandardSymbol(v, Fast_e_pos, p);
}

const Matrix<bigrational>& 
Motive::get_qExpansion(int prec)
{
    if(!Is_New()) {
	if(in_qExpansion && prec<=qExpansion.ncols())
	    return qExpansion;
	else
	    die("Motive::get_qExpansion -- not written for old!");
    }

    if(prec==0) {
	prec=get_DefaultPrec();
	if(in_qExpansion && qExpansion.ncols()>get_DefaultPrec())
	    prec=qExpansion.ncols();
    }
    if(in_qExpansion) 
	if(prec<=qExpansion.ncols())
	    return qExpansion;

    const Vector<bigrational> K=get_Fast_Field();
//    cout << "Fast_Field = " << K << endl;
//    cout << "Fast_times_e = " << get_Fast_times_e() << endl;
    Vector<bigrational> OneOver = get_Fast_times_e().polinv_polymod(K); // 1/<f,e>
    
    int start;
    if(in_qExpansion) {	       // only *increasing* precision
	start=qExpansion.ncols()+1;
	qExpansion.resize(K.poldeg(), prec);
    } else {
	start=2;
	qExpansion.init(K.poldeg(), prec);
	qExpansion(1,1)=1;
    }

    
    mesgln("Expanding f(q):");
    if(start>2) {
	mesg("f(q) = q + ... + q^",start-1);
	mesg(" + ");
    } else
	mesg("f(q) = q +");
    Matrix<bigrational> tFast; get_Fast().transpose(tFast);
    for(int n=start; n<=prec; n++) {
	Vector<bigrational> a, TpE;
	if(is_prime(n)) {
	    get_Fast_TpE (TpE, n);
	    Vector<bigrational> Prod = tFast * TpE;
	    a = polmul_polymod(Prod,OneOver,K);
	} else {
	    Vector<bigint> base, exp;
	    Factor(base, exp, n);
	    if(base.size()==1) { // prime power
		a=polmul_polymod(qExpansion.col(to_int(n/base[1])), 
				 qExpansion.col(to_int(base[1])), K);
		if(get_true_N() % base[1] != 0) {
		    bigint q; power(q,base[1],get_weight()-1);
		    a -= bigrational(q) * qExpansion.col(to_int(n/(base[1]*base[1])));
		}
	    } else {		// product of primes.
		bigint q1; power(q1,base[1],exp[1]);
		bigint q2=n/q1;
		a = polmul_polymod(qExpansion.col(to_int(q1)),
				   qExpansion.col(to_int(q2)), K);
	    }
	}
	if(!a.iszero()) {
	    a.polprint(cerr,"a");
	    mesg("*q^",n);
	    if(n<prec) mesg(" + ");
	}
	qExpansion.setcol(n, a);
    }
    mesgln("");

    qExp_f = K;
    in_qExpansion=true;
    in_qExp_f=true;
    return qExpansion;
}

const Vector<bigrational>& 
Motive::get_qExp_f()
{
    if(in_qExp_f)
	return qExp_f;
    get_qExpansion();
    return get_qExp_f();
}

const Matrix<bigrational>&     
Motive::get_qIntegralBasis(int prec)
{
    if(prec==0) {
	if(in_qIntegralBasis && qIntegralBasis.ncols()>get_DefaultPrec())
	    prec=qIntegralBasis.ncols();
	else 
	    prec=get_DefaultPrec();
    }

    if(in_qIntegralBasis && qIntegralBasis.ncols()>=prec)
	return qIntegralBasis;

    // Failed, can we recover?

    if(Is_Full()) {		// hard one -- it's a full motive
	mesgln("******** Initialize integral basis.");
	int dim=get_true_dimension();
	cerr << "DIMENSION = " << dim << endl;
	if(!Is_Plus()) dim/=2;
	Matrix<bigrational> Basis(dim,prec);
	mesgln("nco = ",prec);
	int row=0;
	for(int i=1; i<=get_Decomp()->num_Motives(); i++) {
	    const Matrix<bigrational>& B
		= get_Decomp()->get_Motive(i).get_qIntegralBasis(prec);
	    for(int r=1; r<=B.nrows(); r++)
		for(int c=1; c<=prec; c++)
		    Basis(r+row,c)=B(r,c);
	    row+=B.nrows();
	    get_Decomp()->free_Motive(i);
	}
	mesgln("Basis = ");
	cerr<<Basis<<endl;
	set_qIntegralBasis(Basis);
	return qIntegralBasis;
    }

    if(!Is_New()) {		// not new -- need to build the new motive
				// from whence this one arose. 
	MotiveDecomp md(get_new_N(), get_weight(), plus);
	Motive& old=md.get_Motive(get_new_IsoClass());
	RaiseOldBasis(old.get_qIntegralBasis(prec),old.get_true_N());
	return qIntegralBasis;
    }

    // It's a new simple motive, so we generate the qIntegralBasis
    // from the qExpansion, which we can compute.

    // The rows form a rational basis of q-expansion in S_k(Gamma_0(N),Q).
    Matrix<bigrational> RationalBasis = get_qExpansion(prec);
    // Clear denominators and saturate.
    bigint d;
    Matrix_denominator(d, RationalBasis);
    if(d!=1) RationalBasis *= d;
    RationalBasis.transpose();	// Saturate takes column space as input.
    Saturate(qIntegralBasis, RationalBasis);
    qIntegralBasis.transpose();
    in_qIntegralBasis=true;
    return qIntegralBasis;
}

void
Motive::set_qIntegralBasis(Matrix<bigrational> RationalBasis)
{
    int dim=get_true_dimension();
    if(!Is_Plus()) dim/=2;
    if(RationalBasis.nrows() != dim) {
	mesgln("Motive::set_qIntegralBasis -- wrong dimension.");
        mesgln("true_N = ", get_true_N());
	if(in_true_IsoClass)
	    mesgln("true_IsoClass = ", get_true_IsoClass());
	mesgln("True dimension = ",get_true_dimension());
	mesgln("But the rational basis has nrows = ", RationalBasis.nrows());
	if(!Is_Plus())
	    mesgln("Which is not half the true dimension.");
        die();
    }
    mesgln("set_qIntegralBasis: Now saturating.");
    bigint d;
    Matrix_denominator(d, RationalBasis);
    if(d!=1) RationalBasis *= d;
    RationalBasis.transpose();	// Saturate takes column space as input.
    Saturate(qIntegralBasis, RationalBasis);
    qIntegralBasis.transpose();
    in_qIntegralBasis=true;
}

void
Motive::reset_Analytic_Invariants()
{
  in_L=in_Lambda=in_LambdaPlus=in_LambdaMinus
      =in_OmegaPlus=in_OmegaMinus=in_ellcurve_invariants
      =in_GoodPeriodIntegrals=in_GoodPeriodIntegralsCoefficients
      =in_ApproxPeriodMap=in_ApproxPeriodMap_fromImPhi
      =false;
}

const bigcomplex&      
Motive::get_L(int s)
{    
    int k=get_weight();
    if(s<1 || s>k-1)
	die("Motive::get_L -- s not critical?!");
    if(!in_L) {
	L.init(k-1);
	L_defined.init(k-1);
	for(int i=1; i<=k-1; i++) {
	    L[i]=-1;
	    L_defined[i]=0;
	}
	in_L=true;
    } 

    if(in_L && L_defined[s]==1)	
	return L[s];

    if(!Is_New()) {
	// Take the product, with appropritiate multiplicity, of L_{Mold}(s) 
	// over the old motives contained in this one.
	if(!Is_Full()) {// there is exactly one factor at some lower level.
	    MotiveDecomp md(get_new_N(), get_weight(), plus);
	    Motive& old=md.get_Motive(get_new_IsoClass());
	    power(L[s],old.get_L(s),get_multiplicity());
	} else { // this is the full space at level N, take product over all other
	    mesgln("Full L value!");
	    // factors at this level.
	    L[s]=bigcomplex(1);
	    mesgln("Num others = ",get_Decomp()->num_Motives());
	    for(int i=1; i<=get_Decomp()->num_Motives(); i++) {
		bigcomplex val=get_Decomp()->get_Motive(i).get_L(s);
		cerr << "Val of first = " << val << endl;
		get_Decomp()->free_Motive(i); // safe because *this isn't one of these.
		L[s] *= val;
		cerr << "now L=" << L[s]<<endl;
	    }
	}
	L_defined[s]=1;
	return L[s];
    }

    /* Use the following theorem to compute L(s) in terms of Omega.

       WindingElement(w, s-1).
                    (2pi)^(s-1)
       |L(f_i,s)| = ----------- |<w, f_i>|
                      (s-1)!

      and

       x = [Phi(S_k)^{+-} : Phi(Zw)]/c_inf    (module index)
         = |prod_i(<w,f_i>)|/Omega^+-.
    */

    bigrational x = get_BSDRatio(s);

    if(s%2)
	L[s] = x * get_OmegaPlus() * get_c_inf();
    else
	L[s] = x * get_OmegaMinus()* get_c_inf();
    
    bigcomplex fact=bigcomplex(1)/bigcomplex(factorial(bigint(s-1)));
    for(int i=1; i<=s-1; i++)
	fact *= 2*Pi();
    power (fact, fact, get_dimension());

    L[s] *= fact;

    L[s] = abs(L[s]);
    if(s%2==0)
	L[s] *= bigcomplex(0,1);
    L_defined[s] = 1;

    return L[s];
}

const Matrix<bigcomplex>&      
Motive::get_Lambda()
{
    if(in_Lambda)
	return Lambda;
    Matrix<bigcomplex> C; to_bigcomplex(C, get_PhiSk());
    Lambda = get_ApproxPeriodMap_fromImPhi() * C;
    in_Lambda=true;
    return Lambda;
}

const Matrix<bigcomplex>&        
Motive::get_LambdaPlus()
{
    if(in_LambdaPlus)
	return LambdaPlus;
    Matrix<bigcomplex> C; to_bigcomplex(C, get_PhiSkPlus());
    LambdaPlus=get_ApproxPeriodMap_fromImPhi() * C;
    in_LambdaPlus=true;
    return LambdaPlus;
}

const Matrix<bigcomplex>&        
Motive::get_LambdaMinus()
{
    if(in_LambdaMinus)
	return LambdaMinus;
    Matrix<bigcomplex> C; to_bigcomplex(C, get_PhiSkMinus());
    LambdaMinus=get_ApproxPeriodMap_fromImPhi()*C;
    in_LambdaMinus = true;
    return LambdaMinus;
}

const bigcomplex&                
Motive::get_OmegaPlus()
{
    if(in_OmegaPlus)
	return OmegaPlus;
    OmegaPlus = get_LambdaPlus().determinant();
    if(real(OmegaPlus)<0) OmegaPlus*=bigcomplex(-1);
    in_OmegaPlus=true;
    return OmegaPlus;
}

const bigcomplex&                
Motive::get_OmegaMinus()
{
    if(in_OmegaMinus)
	return OmegaMinus;
    OmegaMinus = get_LambdaMinus().determinant();
    if(real(OmegaMinus)<0) OmegaMinus*=bigcomplex(-1);
    in_OmegaMinus=true;
    return OmegaMinus;
}

const bigcomplex&                
Motive::get_tau()
{
    if(in_ellcurve_invariants)
	return tau;
    init_ellcurve_invariants();			
    return tau;
}

const bigcomplex&                
Motive::get_w1()
{
    if(in_ellcurve_invariants)
	return w1;
    init_ellcurve_invariants();			
    return w1;
}

const bigcomplex&                
Motive::get_w2()
{
    if(in_ellcurve_invariants)
	return w2;
    init_ellcurve_invariants();			
    return w2;
}


/******* from Cremona's programs *****/
bigcomplex jwc_normalize(bigcomplex& w1, bigcomplex& w2)
{
    bigcomplex tau = w1/w2, w3;
    if (tau.imag()<0) { w1=-w1 ; tau=-tau ; }
    w1=w1-w2*round(tau.real());
    tau=w1/w2;
    
    for (int i=1;i<50 && (abs(tau)<1);i++) {
    // Just to stop infinite loop due to rounding
	w3=-w1; w1=w2; w2=w3; tau=w1/w2;
	w1=w1-w2*round(tau.real());
	tau=w1/w2;
    }
    return tau;
}
 
void jwc_getc4c6(const bigcomplex& w1, const bigcomplex& w2, 
		 bigcomplex& c4, bigcomplex &c6)
{
    bigcomplex tau= w1/w2;
    bigfloat zero(0), one(1), two(2);
    bigfloat pi(Pi());
    bigfloat x = two*pi*tau.real();
    bigfloat y = two*pi*tau.imag();
    bigcomplex q = exp(-y) *  bigcomplex(cos(x),sin(x));
    bigcomplex f = two*pi/w2;
    bigcomplex f2 = f*f;  
    bigcomplex f4=f2*f2;
    
    bigcomplex term   = bigcomplex(one); 
    bigcomplex qpower = bigcomplex(one);
    bigcomplex sum4   = bigcomplex(zero);
    bigcomplex sum6   = bigcomplex(zero);
    
    bigfloat n, n2;
    
// n<=1000 -- ????????????
    for (n=1; !is_approx_zero(term) && n<= 2000; n+=1) {
	n2      = n*n;
	qpower *= q;
	term    = n*n2*qpower/(one-qpower);
	sum4   += term;
	term   *= n2;
	sum6   += term;
    }
    if(abs(term)>1e-4)
	mesgln("jwc_getc4c6: WARNING: precision very bad!");
    c4= (one +  bigfloat(240)*sum4)*f4;
    c6= (one -  bigfloat(504)*sum6)*f4*f2;
}

const bigcomplex&              
Motive::get_c4()
{
    if(in_ellcurve_invariants)
	return c4;
    init_ellcurve_invariants();			
    return c4;
}

const bigcomplex&              
Motive::get_c6()
{
    if(in_ellcurve_invariants)
	return c6;
    init_ellcurve_invariants();			
    return c6;
}

void
Motive::init_ellcurve_invariants()
{
    mesgln("init_ellcurve_invariants");
    if(in_ellcurve_invariants)
	return;
    if(get_dimension() != 1)
	die("Motive::init_c4c6w1w2-- dimension =/= 1!");
    w1=get_Lambda()(1,1);
    w2=get_Lambda()(1,2);
    cerr<<"w1="<<w1<<endl<<"w2="<<w2<<endl;
    tau=jwc_normalize(w1,w2);	
    cerr<<"normalize to:"<<endl;
    cerr<<"w1="<<w1<<endl<<"w2="<<w2<<endl;
    cerr<<"tau="<<tau<<endl;
    jwc_getc4c6(w1,w2,c4,c6);
    ellj=1728*c4*c4*c4/(c4*c4*c4 - c6*c6);

    in_ellcurve_invariants=true;
}

const bigcomplex& 
Motive::get_ellj()
{
    if(in_ellcurve_invariants)
	return ellj;
    init_ellcurve_invariants();			
    return ellj;
}

const Matrix<bigrational>&
Motive::get_PhiGoodPeriods()
{
    if(in_PhiGoodPeriods)
	return PhiGoodPeriods;
    mesgln("PhiGoodPeriods");
    get_H()->Invariant_Symbol_Matrices(GoodPeriods,
				       PhiGoodPeriods,
				       get_Phi());
    in_PhiGoodPeriods=true;
    in_GoodPeriods=true;
    return PhiGoodPeriods;
}


const MatVecList<bigint>&
Motive::get_GoodPeriods()
{
    if(in_GoodPeriods)
	return GoodPeriods;
    get_PhiGoodPeriods();
    return get_GoodPeriods();
}

/*** Used in numerically computing the periods ***/
bigfloat
LineIntegralProduct(int m, int s)
{
    if(m+1-s>m) return bigfloat(1);
    bigfloat prod=m+1-s;
    for(int i=m+1-s+1; i<=m; i++)
	prod *= i;
    return prod;
}

// Compute Int_{z0}^{inf} exp(2*pi*I*n*z)*z^m dz.
bigcomplex 
LineIntegralValue(const bigcomplex& z0,
		  int m, int n)
{
    bigcomplex c=2*Pi()*bigcomplex(0,1)*n;
    // In pari format, must compute:
    // exp(c*z0) * sum(s=0,m,(-1)^s*  (z0^(m-s)) / (c^(s+1)) * prod(i=m-s+1,m,i));
    bigcomplex ans=exp(c*z0);
    bigcomplex sum=bigcomplex(0);
    for(int s=0; s<=m; s++) {
	bigcomplex summand;
	power(summand, z0, m-s);
	if(s%2) summand*=bigcomplex(-1);
	bigcomplex powerofc;
	power(powerofc, c, s+1);
	summand /= powerofc;
	summand *= LineIntegralProduct(m,s);
	sum += summand;
    }
    return sum*ans;
}

bigcomplex 
LineIntegralValue(const bigfloat& y0,
		  const bigfloat& x1,
		  const bigfloat& x2,
		  int m, int n)
{
    return LineIntegralValue(x2+bigcomplex(0,1)*y0,m,n)
	-  LineIntegralValue(x1+bigcomplex(0,1)*y0,m,n);
}


// Compute Int_{0}^{g(0)} exp(2*Pi*I*n*z)*P(z) dz
//           = Int_{alp}^{g(alp)} exp(2*Pi*I*n*z)*P(z) dz
bigcomplex 
PeriodIntegralValue(const Matrix<bigint>& g,
		    const Vector<bigint>& P,
		    int n, int k)
{
//    cout << "Computing PeriodIntegralValue"<<endl;
//    cout << "g = " << g << endl;
//    cout << "P = " << P << endl;
//    cout << "n = " << n << endl;
    bigfloat y0=1/bigfloat(g(2,1)),
	     x1=-bigfloat(g(2,2))*y0,
	     x2= bigfloat(g(1,1))*y0;

    bigcomplex val(0);
    if(k==2)
	val=LineIntegralValue(y0,x1,x2,0,n);
    else {
	for(int i=0; i<=P.poldeg(); i++)
	    if(P.polgetcoeff(i)!=0)
		val+=P.polgetcoeff(i)*LineIntegralValue(y0,x1,x2,i,n);
    }
//    cout << n << "th int val = " << val << endl;
    return val;
}


const Matrix<bigcomplex>& 
Motive::get_GoodPeriodIntegralsCoefficients()
{
    if(in_GoodPeriodIntegralsCoefficients)
	return GoodPeriodIntegralsCoefficients;
    Matrix<bigcomplex>& cn=GoodPeriodIntegralsCoefficients;
    cn.init(get_GoodPeriods().size(), get_qIntegralBasis().ncols());
    for(int i=0; i<cn.nrows(); i++) {
	Matrix<bigint> g=get_GoodPeriods().cM(i);
	Vector<bigint> P=get_GoodPeriods().cv(i);
	for(int n=1; n<=cn.ncols(); n++) {
//	    cout << "PeriodIntegralValue"<<endl;
//	    cout << "g   = " <<g << endl;
//	    cout << "P   = " <<P << endl;
//	    cout << "n   = " <<n<<endl;

	    cn(i+1,n) = PeriodIntegralValue(g, P, n, get_weight());

//	    cout << "ans = " << 2*Pi()*bigcomplex(0,1)*cn(i+1,n)<<endl;
	}
    }

    cn *= -2*Pi()*bigcomplex(0,1);
    return GoodPeriodIntegralsCoefficients;
}



const Matrix<bigcomplex>& 
Motive::get_GoodPeriodIntegrals()
{
    if(in_GoodPeriodIntegrals)
	return GoodPeriodIntegrals;
    /* Find q-expansion of minimal integral basis. */
    Matrix<bigcomplex> an;
    to_bigcomplex(an,get_qIntegralBasis());
    const Matrix<bigcomplex>& cn
	= get_GoodPeriodIntegralsCoefficients();
    // GoodPeriodIntegrals is the matrix whose i-j entry is
     //     (i-th row of cn) * (j-th row of an).
    an.transpose();
    GoodPeriodIntegrals = cn*an;
    in_GoodPeriodIntegrals=true;
    return GoodPeriodIntegrals;
}

const Matrix<bigcomplex>&
Motive::get_ApproxPeriodMap_fromImPhi()
{
    if(in_ApproxPeriodMap_fromImPhi)
	return ApproxPeriodMap_fromImPhi;

    Matrix<bigrational> A;
    get_PhiGoodPeriods().Inverse(A);
    A.transpose();
    Matrix<bigcomplex> B;
    to_bigcomplex(B, A);
    ApproxPeriodMap_fromImPhi = B * get_GoodPeriodIntegrals();
    ApproxPeriodMap_fromImPhi.transpose();

    in_ApproxPeriodMap_fromImPhi=true;
    return ApproxPeriodMap_fromImPhi;
}


const Matrix<bigcomplex>&      
Motive::get_ApproxPeriodMap()
{
    if(in_ApproxPeriodMap)
	return ApproxPeriodMap;

    Matrix<bigcomplex> B;
    to_bigcomplex(B, get_Phi());
    ApproxPeriodMap = get_ApproxPeriodMap_fromImPhi() * B;

    in_ApproxPeriodMap=true;
    return ApproxPeriodMap;
}

int 
Motive::get_next_goodp(int p=1)
{
    p=next_prime(p);
    while(get_true_N()%p==0) p=next_prime(p);
    return p;
}

const Matrix<bigrational>&     
Motive::get_Em()
{
    if(in_Em)
	return Em;

    int p=get_next_goodp();
    get_H()->Tn_Mk_Eig(p, get_charpoly_Tp(p), Em);

    int i=0, d=get_true_dimension();
    while (Em.ncols() > d) {
	p=get_next_goodp();
	Matrix<bigrational> ES;
	get_H()->Tn_Eig_On(p, Em, get_charpoly_Tp(p), ES);
	Em=ES;
    }
    in_Em = true;
    return Em;
}

const Matrix<bigrational>&     
Motive::get_Phi()
{
    if(in_Phi)
	return Phi;

    int dim_goal = get_true_dimension();
    int p=get_next_goodp();
    Matrix<bigrational> F;

    Vector<bigrational>* f=new Vector<bigrational>[1];
    f[0]=get_charpoly_Tp(p);
    Array<int> pr(1);
    pr[0]=p;

    get_H()->Dual_Eig_Intersect(F, pr, f, dim_goal);

    int start_at=1;
    const int step_size=4;
    while(F.ncols()>dim_goal) {	// cut down with step_size more hecke operators
	Matrix<bigrational> Start=F;
	// extend pr and f
	int n=pr.size();
	pr.resize(n+step_size);
	Vector<bigrational>* g=new Vector<bigrational>[pr.size()];
	for(int j=0; j<n; j++)
	    g[j]=f[j];
	delete[] f;
	for(int j=n; j<n+step_size; j++) {
	    pr[j]=get_next_goodp(pr[j-1]);
	    g[j]=get_charpoly_Tp(pr[j]);
	}
	f=g;
	get_H()->Dual_Eig_Intersect(F,pr,f, Start, start_at, dim_goal);
	start_at+=step_size;
    } ;

    delete[] f;
    F.transpose(Phi);
    in_Phi=true;
    return Phi;
}

const Matrix<bigrational>&
Motive::get_PhiPlus()
{
    if(in_PhiPlus)
	return PhiPlus;

    Matrix<bigrational> C=get_Conj_Phi();
    for(int i=1;i<=C.ncols();i++) 
	C(i,i) = C(i,i) - 1;
    C.transpose();
    IntegerKernel(PhiPlus, C);
    PhiPlus.transpose();
    PhiPlus = PhiPlus*get_Phi();
    in_PhiPlus=true;
    return PhiPlus;

}

const Matrix<bigrational>&
Motive::get_PhiMinus()
{
    if(in_PhiMinus)
	return PhiMinus;

    Matrix<bigrational> C=get_Conj_Phi();
    for(int i=1;i<=C.ncols();i++) 
	C(i,i) = C(i,i) + 1;
    C.transpose();
    IntegerKernel(PhiMinus, C);
    PhiMinus.transpose();
    PhiMinus = PhiMinus*get_Phi();
    in_PhiMinus=true;
    return PhiMinus;
}

const Matrix<bigrational>&     
Motive::get_Conj_Phi()
{
    if(in_Conj_Phi)
	return Conj_Phi;
    Matrix<bigrational> C;
    get_H()->Conjugation_Mk(C);
    Matrix<bigrational> Phitrans;
    get_Phi().transpose(Phitrans);
    C.transpose();
    C.RestrictedTo(Conj_Phi, Phitrans);
    Conj_Phi.transpose();
    in_Conj_Phi=true;
    return Conj_Phi;
}

const Matrix<bigrational>& 
Motive::get_V()
{
    if (in_V)
	return V;

    int i=0;
    int d=get_true_dimension();
    V_Embedded=get_H()->get_ZBasis_Sk();
    int p=get_next_goodp();
    while (V_Embedded.ncols() > d) {
	mesgln("get_V: trying T",p);
	Matrix<bigrational> T;
	get_H()->Tn_On(p, V_Embedded, T) ;
	Matrix<bigrational> K;
	IntegerKernel(K, T, get_charpoly_Tp(p));
	V_Embedded = V_Embedded*K;
	p=get_next_goodp(p);
	mesgln("Sought for dimension = ", d);
	mesgln("dim current V_Embedded = ", V_Embedded.ncols());
    }
    mesgln("Found V_Embedded.");
    Matrix<bigrational> ZBasisInv;
    mesgln("Computing left inverse.");
    get_H()->get_ZBasis_Sk().LeftInverse(ZBasisInv);
    mesgln("Done, now computing V from V_Embedded.");
    V = ZBasisInv * V_Embedded;
    mesgln("Done, now V is computed.");

    in_V = true;
    return V;
}

const Matrix<bigrational>& 
Motive::get_Vplus()
{
    if(in_Vplus)
	return Vplus;

    Matrix<bigrational> C=get_Conj_V();
    for(int i=1;i<=C.ncols();i++) 
	C(i,i) = C(i,i) - 1;
    IntegerKernel(Vplus, C);
    Vplus = V*Vplus;
    
    Vplus_Embedded = get_H()->get_ZBasis_Sk() * Vplus ;

    in_Vplus=true;
    return Vplus;
}

const Matrix<bigrational>& 
Motive::get_Vminus()
{
    if(in_Vminus)
	return Vminus;

    Matrix<bigrational> C=get_Conj_V();
    for(int i=1;i<=C.ncols();i++) 
	C(i,i) = C(i,i) + 1;
    IntegerKernel(Vminus, C);
    Vminus = V * Vminus;

    Vminus_Embedded = get_H()->get_ZBasis_Sk() * Vminus ;

    in_Vminus=true;
    return Vminus;
}

const Matrix<bigrational>&     
Motive::get_V_Embedded()
{
    if(in_V)
	return V_Embedded;
    get_V();
    if(!in_V) die("Motive::get_V_Embedded -- get_V() didn't do its job.");
    return V_Embedded;
}

const Matrix<bigrational>&
Motive::get_Vplus_Embedded()
{
    if(in_Vplus)
	return Vplus_Embedded;
    get_Vplus();
    if(!in_Vplus) die("Motive::get_Vplus_Embedded -- get_Vplus() didn't do its job.");
    return Vplus_Embedded;
}

const Matrix<bigrational>&     
Motive::get_Vminus_Embedded()
{
    if(in_Vminus)
	return Vminus_Embedded;
    get_Vminus();
    if(!in_Vminus) die("Motive::get_Vminus_Embedded -- get_Vminus() didn't do its job.");
    return Vminus_Embedded;
}



const Matrix<bigrational>&
Motive::get_VDual()
{
    if (in_VD)
	return VD;

    // Cut out left VD using DefList and Tp on integral basis
    int i=0;
    int d=get_true_dimension();
    int p=get_next_goodp();
    Matrix<bigrational> T;
    get_H()->Tn_On(p, get_H()->get_ZBasis_Sk(), T);
    T.transpose();
    IntegerKernel(VD, T,  get_charpoly_Tp(p));

    while (VD.ncols() > d) {
	p=get_next_goodp(p);

	Matrix<bigrational> T;
	get_H()->Tn_On(p, get_H()->get_ZBasis_Sk(), T);
	T.transpose();
	Matrix<bigrational> Ton;
	T.RestrictedTo(Ton, VD);
	Matrix<bigrational> K;
	IntegerKernel(K, Ton, get_charpoly_Tp(p));
	VD = VD*K;

/*	cout << "Sought for dimension = " << d << endl;
	cout << "VD = " << VD <<endl;
	cout << "Ton = " << Ton << endl;
	cout << "T = " << T << endl;
*/
    }

    VD.transpose();

    in_VD = true;
    return VD;
}

const Vector<bigrational>&
Motive::get_charpoly_Tp(int p)
{
    if(!in_Fast) get_Fast();
    if(p<=Fast_Tp.len() && Fast_Tp.cv(p).size()>0) 
	return Fast_Tp.cv(p);
    Vector<bigrational> f;
    get_Tp_Fast(p).charpoly(f);
    Fast_Tp.v(p)=f;
    // kill multiplicity..
    if(gcd(get_true_N(),p)==1) {
	if(!Is_New()) {
	    VecFac<bigrational> fac;
	    Fast_Tp.cv(p).polfactor(fac);
	    Fast_Tp.v(p).resize(1);  Fast_Tp.v(p)[1]=1;
	    for(int j=1; j<=fac.numfactors(); j++) {
		if(fac.exp(j) % get_multiplicity() != 0) 
		    die("Motive::get_charpoly_Tp(int p): ERROR!");
		Vector<bigrational> pp;
		fac.base(j).polpower(pp, fac.exp(j)/get_multiplicity());
		Fast_Tp.v(p) = polmul(Fast_Tp.v(p), pp);
	    }
	}
    }
    
    return Fast_Tp.cv(p);
}

const Matrix<bigrational>&     
Motive::get_Tp_Fast(int p)
{
    if(!in_Fast) {
	mesgln("Motive::get_Tp_Fast(int p) -- currently everything");
	mesgln("arises from Fast, and it is not initialized.");
    }

    if(p>=Fast_Tp.size())
	Fast_Tp.resize(p+p/2+1);
    if(p<Fast_Tp.len() || Fast_Tp.cM(p).nrows()==0) {
	get_H_Plus()->Dual_Tp_On(Fast_Tp.M(p), p,
				 Fast,Fast_e, Fast_EVinv);
    }
    return Fast_Tp.cM(p);
}

void
Motive::set_Fast(const Matrix<bigrational>& _Fast,
		 const Vector<bigrational>& _Field)
{
    if(in_Fast) 
	die("Motive::set_Fast -- attempt to call after Fast already set.");
    set_Fast(_Fast);
    Fast_Field=_Field;
}

void
Motive::set_Fast(const Matrix<bigrational>& _Fast)
{
    if(in_Fast) 
	die("Motive::set_Fast -- attempt to call after Fast already set.");
    Fast=_Fast;
    Fast_Tp.init(1);
    get_H_Plus()->Dual_E(Fast_e, Fast_EVinv, Fast);
    if(Fast_e.size()<1)
	die("set_Fast: ERROR");
    Fast_e_pos = Fast_e[1];
    in_Fast=true;
}


const Matrix<bigrational>&     
Motive::get_Fast()
{
    if(!in_Fast) {
	if(Is_Full()) {		// we can do it.
	    // Find dual of Sk.
	    mesgln("Initializing the full dual subspace at level ",get_true_N());
	    mesgln("This might take a while.");
	    Matrix<bigrational> SkDual;
	    get_H_Plus()->HeckeDual_UsingImage(SkDual, get_H_Plus()->get_Sk(), 97);
	    bigint denom; Matrix_denominator(denom, SkDual);
	    SkDual*=denom;
	    set_Fast(SkDual);
	} else 
	    die("Motive::get_Fast -- Fast must be initialized.");
    }
    return Fast;
}

const Matrix<bigrational>&
Motive::get_Fast_EVinv()
{
    if(!in_Fast) get_Fast();
    return Fast_EVinv;
}
 
const Vector<int>&
Motive::get_Fast_e()
{
    if(!in_Fast) get_Fast();
    return Fast_e;
}

Vector<bigrational>
Motive::get_Fast_times_e()
{
    if(!in_Fast) get_Fast();
    return Fast.row(Fast_e_pos);
}

const Vector<bigrational>&     
Motive::get_Fast_Field()
{
    if(!in_Fast) get_Fast();
    return Fast_Field;
}



const Matrix<bigrational>&    
Motive::get_Tp_V(int p)
{
    if(!in_Tp_V) {
	Tp_V.init(p+1);
	in_Tp_V=true;
    }
    if(p>=Tp_V.size())
	Tp_V.resize(p+p/2+1);
    if(Tp_V.cM(p).nrows()==0) {
	get_H()->Tn_On(p, get_V_Embedded(), Tp_V.M(p));
    }
    return Tp_V.cM(p);
}

const Matrix<bigrational>&    
Motive::get_Tp_Vplus(int p)
{
    if(!in_Tp_Vplus) {
	Tp_Vplus.init(p+1);
	in_Tp_Vplus=true;
    }
    if(p>=Tp_Vplus.size())
	Tp_Vplus.resize(p+p/2+1);
    if(Tp_Vplus.cM(p).nrows()==0) 
	get_H()->Tn_On(p, get_Vplus_Embedded(), Tp_Vplus.M(p));
    return Tp_Vplus.cM(p);
}

const Matrix<bigrational>&    
Motive::get_Tp_Vminus(int p)
{
    if(!in_Tp_Vminus) {
	Tp_Vminus.init(p+1);
	in_Tp_Vminus=true;
    }
    if(p>=Tp_Vminus.size())
	Tp_Vminus.resize(p+p/2+1);
    if(Tp_Vminus.cM(p).nrows()==0) 
	get_H()->Tn_On(p, get_Vminus_Embedded(), Tp_Vminus.M(p));
    return Tp_Vminus.cM(p);
}


const Matrix<bigrational>&    
Motive::get_Tp_Phi(int p)
{
    if(!in_Tp_Phi) {
	Tp_Phi.init(p+1);
	Matrix<bigrational> PhiTrans; get_Phi().transpose(PhiTrans);
	get_H()->Dual_E(Tp_Phi_Dual_e, Tp_Phi_Dual_EVinv, PhiTrans);
	in_Tp_Phi=true;
    }
    if(p>=Tp_Phi.size())
	Tp_Phi.resize(p+p/2+1);
    if(Tp_Phi.cM(p).nrows()==0) {
	Matrix<bigrational> PhiTrans; get_Phi().transpose(PhiTrans);
	get_H()->Dual_Tp_On(Tp_Phi.M(p), p, PhiTrans,
			    Tp_Phi_Dual_e,
			    Tp_Phi_Dual_EVinv);
    }
    return Tp_Phi.cM(p);
}

const Matrix<bigrational>&    
Motive::get_Tp_PhiPlus(int p)
{
    if(!in_Tp_PhiPlus) {
	Tp_PhiPlus.init(p+1);
	Matrix<bigrational> PhiPlusTrans;
	get_PhiPlus().transpose(PhiPlusTrans);
	get_H()->Dual_E(Tp_PhiPlus_Dual_e,
			Tp_PhiPlus_Dual_EVinv,
			PhiPlusTrans);
	in_Tp_PhiPlus=true;
    }
    if(p>=Tp_PhiPlus.size())
	Tp_PhiPlus.resize(p+p/2+1);
    if(Tp_PhiPlus.cM(p).nrows()==0) {
	Matrix<bigrational> PhiPlusTrans; get_PhiPlus().transpose(PhiPlusTrans);
	get_H()->Dual_Tp_On(Tp_PhiPlus.M(p), p, PhiPlusTrans,
			    Tp_PhiPlus_Dual_e,
			    Tp_PhiPlus_Dual_EVinv);
    }
    return Tp_PhiPlus.cM(p);
}


const Matrix<bigrational>&    
Motive::get_Tp_PhiMinus(int p)
{
    if(!in_Tp_PhiMinus) {
	Tp_PhiMinus.init(p+1);
	Matrix<bigrational> PhiMinusTrans; get_PhiMinus().transpose(PhiMinusTrans);
	get_H()->Dual_E(Tp_PhiMinus_Dual_e,
			Tp_PhiMinus_Dual_EVinv,
			PhiMinusTrans);
	in_Tp_PhiMinus=true;
    }
    if(p>=Tp_PhiMinus.size())
	Tp_PhiMinus.resize(p+p/2+1);
    if(Tp_PhiMinus.cM(p).nrows()==0) {
	Matrix<bigrational> PhiMinusTrans; get_PhiMinus().transpose(PhiMinusTrans);
	get_H()->Dual_Tp_On(Tp_PhiMinus.M(p), p, PhiMinusTrans,
			    Tp_PhiMinus_Dual_e,
			    Tp_PhiMinus_Dual_EVinv);
    }
    return Tp_PhiMinus.cM(p);
}

bigrational
Motive::get_TraceOf_ap(int p)
{
    return get_Tp_PhiPlus(p).trace();
}

const Matrix<bigrational>&     
Motive::get_Conj_V()
{
    if(in_Conj_V)
	return Conj_V;
    Matrix<bigrational> C;
    get_H()->Conjugation_Mk(C);
    C.RestrictedTo(Conj_V, get_V_Embedded());
    in_Conj_V=true;
    return Conj_V;
}

const Matrix<bigrational>&     
Motive::get_Conj_PhiSk()
{
    if(in_Conj_PhiSk)
	return Conj_PhiSk;

    //   Compute complex conjugation C on Z=PhiSk.
    //       (a) choose n independent columns i_1,...,i_n of P.
    const Matrix<bigrational>& P = get_Phi();
    int n=P.nrows();
    Matrix<bigrational> PP=P;
    Vector<int> elimcol, elimrow;
    PP.SimpleEchelon(elimcol, elimrow);
    Vector<int> ind(n);
    int i=1;
    for(int c=1;c<=PP.ncols();c++)
	if(elimcol[c]!=0)
	    ind[i++]=c;
    
    //       (b) B <-- matrix consisting of n linearly independent 
    //           columns i_1,...,i_n
    Matrix<bigrational> B(n,n);
    for(int i=1;i<=n;i++)
	B.setcol(i,P.col(ind[i]));

    //       (c) A <-- * involution on M_k
    Matrix<bigrational> A;
    get_H()->Conjugation_Mk(A);

    //       (d) A' <-- matrix whose columns are i_1,...,i_n of A.
    Matrix<bigrational> AA(A.nrows(),n);
    for(int i=1;i<=n;i++)
	AA.setcol(i,A.col(ind[i]));

    //       (e) C <-- Z^(-1)*P*A'*B^(-1)*Z
    const Matrix<bigrational>& Z=get_PhiSk();
    Matrix<bigrational> Zinv,Binv;
    Z.Inverse(Zinv);
    B.Inverse(Binv);
    Conj_PhiSk = Zinv*P*AA*Binv*Z;
    in_Conj_PhiSk = true;
    return Conj_PhiSk;
}

int
Motive::get_Wq(int p)
{
    if(!Is_New())
	die("Motive::get_Wq should only be called on new motives.");
    if(!is_prime(p))
	die("Motive::get_Wq: p isn't prime..");
    if(get_true_N()%p !=0)
	return 1;
    if(!in_Wq) {
	Wq.init(get_num_PrimeDivisorsOfN());
	in_Wq=true;
    }
    int i=get_PrimeIndex(p);
    if(Wq[i]!=0)
	return Wq[i];
    // Compute Wq on H_Plus.
    Matrix<bigrational> W = get_H_Plus()->get_Wq(i);
    W.transpose();		// transpose action on fast.
    Matrix<bigrational> WF = W*Fast;
    // WF must equal Wq*Fast, if not, there is a bug.
    if((WF+Fast).iszero()) {
	Wq[i]=-1;
	return -1;
    }
    if((WF-Fast).iszero()) {
	Wq[i]=1;
	return 1;
    }
    die("get_Wq -- there is a bug."); 
}

int 
Motive::get_WqProd()
{
    int WqProd=1;
    for(int j=1; j<=get_num_PrimeDivisorsOfN(); j++) {
	WqProd *= get_Wq(get_nth_PrimeDivisorsOfN(j));
    }
    return WqProd;
}

const Matrix<bigrational>&     
Motive::get_PhiSk()
{
    if(in_PhiSk)
	return PhiSk;
    ZBasis(PhiSk, get_Phi() * get_H()->get_ZBasis_Sk());
    in_PhiSk=true;
    return PhiSk;
}

const Matrix<bigrational>&     
Motive::get_PhiSkInverse()
{
    if(in_PhiSkInverse)
	return PhiSkInverse;
    get_PhiSk().Inverse(PhiSkInverse);
    in_PhiSkInverse=true;
    return PhiSkInverse;
}

const Matrix<bigrational>&     
Motive::get_PhiSkPlus()
{
    if(in_PhiSkPlus)
	return PhiSkPlus;
    Matrix<bigrational> C=get_Conj_PhiSk();
    for(int i=1;i<=C.nrows();i++)
	C(i,i)-=1;
    Matrix<bigrational> KerC;
    IntegerKernel(KerC, C);
    PhiSkPlus=get_PhiSk()*KerC;
    in_PhiSkPlus=true;
    return PhiSkPlus;
}

const Matrix<bigrational>&     
Motive::get_PhiSkMinus()
{
    if(in_PhiSkMinus)
	return PhiSkMinus;
    Matrix<bigrational> C=get_Conj_PhiSk();
    for(int i=1;i<=C.nrows();i++)
	C(i,i)+=1;
    Matrix<bigrational> KerC;
    IntegerKernel(KerC, C);
    PhiSkMinus=get_PhiSk()*KerC;
    in_PhiSkMinus=true;
    return PhiSkMinus;
}

void
Motive::set_Intersection(int i, const Vector<bigint>& congs)
{
    if(i<1 || i>get_Decomp()->num_Motives())
	die("Motive::set_Intersection -- invalid motive number.");

    if(!in_Intersection) {	// force init.
	get_Intersection(get_true_IsoClass());
    }
    Intersection.v(i)=congs;
}

const Vector<bigint>& 
Motive::get_Intersection(int i)
{
    if(!in_Intersection) {
	Intersection.init(get_Decomp()->num_Motives()+1);
	Intersection.v(get_true_IsoClass()).init(1);
	in_Intersection=true;
    }
    if(i!=get_true_IsoClass() && Intersection.cv(i).size()==0) {
              // cong. not known.
	mesgln("Computing intersection of two lattices.");
	Matrix<bigrational> VV;
	get_V().augment(VV, get_Decomp()->get_Motive(i).get_V());
	Vector<bigint> divs;
	mesgln("Intersection: Computing structure.");
	ElementaryDivisors(divs, VV);
	int cnt=0;
	for(int j=1; j<=divs.size();j++)
	    if(divs[j]>1)
		cnt++;
	Vector<bigint> congs(cnt);
	if(cnt>0) {
	    int ell=1;
	    for(int j=1; j<=divs.size(); j++)
		if(divs[j]>1)
		    congs[ell++]=divs[j];
	}
	set_Intersection(i, congs);
	get_Decomp()->get_Motive(i).set_Intersection(get_true_IsoClass(), congs);
	
// WARNING: VERY SCARY LINE!
	get_Decomp()->free_Motive(i);
    }
    return Intersection.cv(i);
}

void
Motive::PhiTx(Matrix<bigrational>& PTx, const Vector<bigrational>& x,
	      int bound=0)
{
    get_H()->CyclicHeckeImage(PTx, get_Phi(), x, bound);
}

int
Motive::get_PrimeIndex(int p)
{
    if(!is_prime(p))
	die("Motive::PrimeIndex -- p is not prime.");
    if(get_true_N()%p != 0)
	die("Motive::PrimeIndex -- p does not divide N.");
    if(in_Prime) {
	for(int i=1; i<=Primes.size(); i++)
	    if(Primes[i]==p) 
		return i;
	die("Motive::get_PrimeIndex -- an error occured.");
    } else {
	get_num_PrimeDivisorsOfN();	// inits PrimeIndex.
	return get_PrimeIndex(p);
    }
}

int
Motive::get_num_PrimeDivisorsOfN()
{
    if(get_true_N()==1)
	return 0;
    if(in_Prime)
	return Primes.size();
    Vector<bigint> bi_p, bi_exp;
    Factor(bi_p, bi_exp, get_true_N());
    Primes.init(bi_p.size());
    Exponents.init(bi_exp.size());
    for(int i=1; i<=bi_p.size(); i++) {
	Primes[i]=to_int(bi_p[i]);
	Exponents[i]=to_int(bi_exp[i]);
    }
    in_Prime=true;
    return Primes.size();
}

int 
Motive::get_nth_PrimeDivisorsOfN(int n)
{
    if(in_Prime) {
	if(n<1 || n>Primes.size())
	    die("Motive::get_nth_DivisorOfN -- n out of range.");
	return Primes[n];
    }

    get_num_PrimeDivisorsOfN();
    return get_nth_PrimeDivisorsOfN(n);
}

int 
Motive::get_nth_PrimeExponentOfN(int n)
{
    if(in_Prime) {
	if(n<1 || n>Exponents.size())
	    die("Motive::get_nth_ExponentOfN -- n out of range.");
	return Exponents[n];
    }

    get_num_PrimeDivisorsOfN();
    return get_nth_PrimeExponentOfN(n);
}

H1NChiZ*
Motive::get_H()
{
    if (in_H)
	return H;
    if(in_Decomp) {
	H = get_Decomp()->get_H();
	in_H = true;
	delete_H = false ;
	return H;
    }
    if(!in_plus) {
	plus=false;
	in_plus=true;
    }
    H = new H1NChiZ(Character(get_true_N()), get_weight(), true, Is_Plus());

    in_H = true;
    delete_H = true;
    return H;
}

H1NChiZ*
Motive::get_H_Plus()
{
    if(in_Decomp) {
	return get_Decomp()->get_H_Plus();
    }

    die("Motive::get_H_Plus -- not attached to a Decomp object.");
    return NULL;
}

   
MotiveDecomp*
Motive::get_Decomp()
{
    if(in_Decomp)
	return Decomp;
    die("Motive::get_Decomp -- not defined. I've not yet worked out how to do this.");
}

char* 
Motive::Name()
{
    if(in_Name)
	return strName;
    strName=new char[1024];
    ostrstream ost(strName, 1024);
    ost<< get_H()->Name() << "Motive" << get_true_IsoClass() << '\0';
    in_Name = true;
    return strName;
}

void
Motive::assure_new(const char* mesg)
{
    if(!Is_New()) {
	cerr << "Motive::";
	die(mesg);
    }
}


const char* MemberToDescription(MEMBER_TYPE mem)
{
static char* MEMBER_DESCRIPTIONS[] = {
    "PRIMES -- list of prime divisors of the level",
    "EXPONENTS -- exponents of the primes dividing the level",
    "DEFAULTPREC -- at worst, q-expansions are computed to this precision",
    "ISFULL -- is it the full J_0(N) object",
    "ISNEW -- yes iff the motive new at the true level",
    "ISPLUS -- working always in the plus one quotient",
    "MULTIPLICITY -- if new =1, otherwise gives the multiplicity to which this occurs in S_k",
    "NEW_N -- the level at which the motive is new",
    "NEW_ISOCLASS -- the isogeny class of the motive where it was new",
    "TRUE_N -- the level at which the motive is embedded",
    "TRUE_ISOCLASS -- the isogeny class of the embedding",
    "WEIGHT -- the weight k of the motive.  an even integer >= 2",
    "DIMENSION -- the dimension of the new version of this motive",
    "TRUE_DIMENSION -- the dimension as it is embedded in homology (includes multiplicity)",
    "DISC -- the discrimant of the ring Z[a_1,a_2,a_3,...]",
    "C_INF -- number of real components",
    "C_P(p) -- number of components of Neron model over Fpbar (lists all if p omitted)",
    "C_P_ARITH(p) -- number of components of Neron model over Fp (lists all if p omitted)",
    "TORSIONBOUND(max) -- gcd { #A(F_p) : 3 <= p <= max }",
    "CUSPORDER(a,b) -- order of image of [alp]-[oo] where alp is a cusp in P^1(Q)",
    "SHABOUND -- upper bound on order of Sha, (assumes Omega=OmegaNeron and BSD conjecture)",
    "MODPOLAR -- matrix of the modular polarization",
    "DEGMP -- square root of the degree of the modular polarization",
    "KERMP -- structure of the kernel of the modular polarization",
    "BSDRATIO(s=1) -- the ratio |L(A,s)/Omega * (s-1)!/(2pi)^(s-1)|",
    "BSDRATIOTIMESTOR(s=1) -- BSDRATIO * (TORSIONBOUND)^2",
    "QEXPANSION(numterms) -- the q-expansion of a newform corresponding to the motive", 
    "QEXPF -- the field over which the newform is defined",
    "QINTBASIS(numterms) -- an integral basis for the subspace of cusp forms corresponding to the motive",
    "PHIGOODPERIODS -- the image of the 'good periods' under PHI",
    "GOODPERIODS -- the 'good periods', used in numerical computation of periods",
    "GOODPERIODINTEGRALS -- the integrals of good periods against integral basis",
    "GOODPERIODINTEGRALCOEFFICIENTS -- used in computing GOODPERIODINTEGRALS", 
    "APPROXPERIODMAP_FROMIMPHI -- numerical period map from Im(Phi) to C^d",
    "APPROXPERIODMAP -- the period map",
    "LVALUES(s) -- numerical special value of L(A,s), s=1,...,weight-1.",
    "LVALUES_DEFINED -- used internally (which LVALUES are actually defined)",
    "LAMBDA -- the period lattice",
    "LAMBDAPLUS -- the part of the period lattice fixed under conjugation",
    "LAMBDAMINUS -- the part of the period lattice acted on as -1 by conjugation",
    "OMEGAPLUS -- the volume of LAMBDAPLUS (multiply by C_INF to get BSD Omega)",
    "OMEGAMINUS -- the volume of LAMBDAMINUS",
    "ELLC4 -- c4 invariant of one dimensional lattice LAMBDA",
    "ELLC6 -- c6 invariant of one dimensional lattice LAMBDA",
    "ELLW1 -- omega_1 invariant of one dimensional lattice LAMBDA",
    "ELLW2 -- omega_1 invariant of one dimensional lattice LAMBDA",
    "ELLTAU -- tau invaraint of one dimensional lattice LAMBDA",
    "ELLJ -- j invariant of one dimensional lattice LAMBDA",
    "EM -- embedding of motive in Modular symbols / Q, w.r.t free basis",
    "PHI -- left eigenspace corresponding to motive, ker(PHI)=ker(Period Map)", 
    "PHIPLUS -- plus one subspace of PHI for complex conjugation",
    "PHIMINUS-- minus one subspace of PHI for complex conjugation",
    "VSPACE --  embedding of motive in cuspidal Modular symbols / Z",
    "VSPACE_EMBEDDED -- embedding of VSPACE in Modular symbols / Q, w.r.t free basis",
    "VPLUS -- plus one subspace of VSPACE", 
    "VPLUS_EMBEDDED -- embedding of plus one subspace of VSPACE in Modular symbols / Q, w.r.t free basis",
    "VMINUS -- minus one subspace of VSPACE", 
    "VMINUS_EMBEDDED -- embedding of VSPACE in Modular symbols / Q, w.r.t free basis",
    "VDUAL_SPACE -- embedding of dual of motive in cuspidal Modular symbols / Z",
    "CONJ_V -- conjugation involution on VSPACE",
    "CONJ_PHI -- conjugation involution on PHI",
    "TP_V(p) -- action of Hecke operator T_p on VSPACE",
    "TP_VPLUS(p) -- action of Hecke operator T_p on VPLUS",
    "TP_VMINUS(p) -- action of Hecke operator T_p on VMINUS",
    "TP_PHI(p) -- action of Hecke operator T_p on PHI",
    "TP_PHIPLUS(p) -- action of Hecke operator T_p on PHIPLUS",
    "TP_PHIMINUS(p) -- action of Hecke operator T_p on PHIMINUS",
    "TP_PHIDUALEVINV -- used internally for quickly computing T_p",
    "TP_PHIPLUS_DUAL_EVINV -- used internally for quickly computing T_p",
    "TP_PHIMINUS_DUAL_EVINV -- used internally for quickly computing T_p",
    "TP_PHI_DUAL_E -- used internally for quickly computing T_p",
    "TP_PHIPLUS_DUAL_E -- used internally for quickly computing T_p",
    "TP_PHIMINUS_DUAL_E -- used internally for quickly computing T_p",
    "WQ -- Signs of Atkin-Lehner involutions W_q on VSPACE, q=power of p dividing N (lists all if p omitted)",
    "CONJ_PHISK -- action of conjugation involution on PHISK", 
    "PHISK -- PHI(Integral cuspidal modular symbols)", 
    "PHISKINVERSE -- used internally", 
    "PHISKPLUS -- Plus one subspace of PHISK for conjugation involution", 
    "PHISKMINUS -- Minus one subspace of PHISK for conjugation involution", 
    "INTERSECTION(isogclass) -- Compute intersection of this motive with the isogclass one. isogclass is a letter.",
    "FAST(p) -- matrix and charpoly of left action of Hecke on 'fast dual space'",
    "HMODDEG(p) -- new at p homology modular degree",
    "XMODDEG(p) -- new at p character group modular degree",
    "XIMAGE(p) -- order of image of component group of J_0(N) at p in component group of A_f",
    "XEIGEN(p) -- eigenform on supersingular basis (N must be prime.)"
} ;

    if(0<=mem && mem<MEMBER_EOF)
	return MEMBER_DESCRIPTIONS[mem];
    die("MemberToDescription: invalid member number.");

}

const char* MemberToKeyword(MEMBER_TYPE mem)
{
static char* MEMBER_NAMES[] = {
    "PRIMES", "EXPONENTS", "DEFAULTPREC",
    "ISFULL", "ISNEW", "ISPLUS", "MULTIPLICITY", "NEW_N", "NEW_ISOCLASS",
    "TRUE_N", "TRUE_ISOCLASS", "WEIGHT", "DIMENSION", "TRUE_DIMENSION",
    "DISC", "C_INF", "C_P", "C_P_ARITH", "TORSIONBOUND", "CUSPORDER", "SHABOUND",
    "MODPOLAR", "DEGMP", "KERMP", "BSDRATIO", "BSDRATIOTIMESTOR", "QEXPANSION", 
    "QEXPF", "QINTBASIS", "PHIGOODPERIODS", "GOODPERIODS", "GOODPERIODINTEGRALS",
    "GOODPERIODINTEGRALCOEFFICIENTS", "APPROXPERIODMAP_FROMIMPHI",
    "APPROXPERIODMAP", "LVALUES", "LVALUES_DEFINED", "LAMBDA",
    "LAMBDAPLUS", "LAMBDAMINUS", "OMEGAPLUS", "OMEGAMINUS", 
    "ELLC4", "ELLC6", "ELLW1", "ELLW2", "ELLTAU", "ELLJ",
    "EM", "PHI", "PHIPLUS", "PHIMINUS", "VSPACE", "VSPACE_EMBEDDED",
    "VPLUS", "VPLUS_EMBEDDED", "VMINUS", "VMINUS_EMBEDDED", "VDUAL_SPACE", 
    "CONJ_V", "CONJ_PHI", "TP_V", "TP_VPLUS", "TP_VMINUS", "TP_PHI", "TP_PHIPLUS",
    "TP_PHIMINUS", "TP_PHIDUALEVINV", "TP_PHIPLUS_DUAL_EVINV",
    "TP_PHIMINUS_DUAL_EVINV", "TP_PHI_DUAL_E", "TP_PHIPLUS_DUAL_E",
    "TP_PHIMINUS_DUAL_E",
    "WQ", "CONJ_PHISK", "PHISK", "PHISKINVERSE", "PHISKPLUS", "PHISKMINUS",
    "INTERSECTION", "FAST",
    "HMODDEG", "XMODDEG", "XIMAGE", "XEIGEN"
} ;
 if(sizeof(MEMBER_NAMES)!=MEMBER_EOF*sizeof(char*)) {
	mesgln("sizeof(MEMBER_NAMES)=",sizeof(MEMBER_NAMES));
        mesgln("MEMBER_EOF*sizeof(char*)=", MEMBER_EOF*sizeof(char*));
	die();
 }
    if(0<=mem && mem<MEMBER_EOF)
	return MEMBER_NAMES[mem];
    die("MemberToKeyword: invalid member name.");

}

MEMBER_TYPE KeywordToMember(const char* word) 
{
    if(word==NULL) return MEMBER_EOF;
    for(int i=0; i<MEMBER_EOF; i++)
	if(strcmp(word,MemberToKeyword(MEMBER_TYPE(i)))==0)
	    return MEMBER_TYPE(i);
    return MEMBER_EOF;		// not found
}

void title(MEMBER_TYPE mem, ostream& out)
{
    out<<"["<< MemberToKeyword(mem) << "]"<<endl;
}

void Motive::Save(MEMBER_TYPE mem, ofstream& out)
{
//    cout << "Motive::Save(" << MemberToKeyword(mem) << ")" << endl;
    bool newline=true;
    switch(mem) {
    case PRIMES: if(in_Prime) { title(mem,out); Primes.Save(out); } else newline=false; break;
    case EXPONENTS: if(in_Prime) { title(mem,out); Exponents.Save(out); } else newline=false; break; 
    case DEFAULTPREC: if(in_DefaultPrec) { title(mem,out); out<<DefaultPrec; } else newline=false; break;
    case ISFULL: if(in_IsFull) { title(mem,out); out<<IsFull; } else newline=false; break; 
    case ISNEW: if(in_is_new) { title(mem,out); out<<is_new;}  else newline=false; break;
    case ISPLUS: if(in_plus) { title(mem,out); out<<plus;} else newline=false; break;
    case MULTIPLICITY: if(in_multiplicity) { title(mem,out); out<<multiplicity;} else newline=false; break;
    case NEW_N: if(in_new_N) { title(mem,out); out<<new_N;}  else newline=false; break;
    case NEW_ISOCLASS: if(in_new_IsoClass) { title(mem,out); out<<new_IsoClass; } else newline=false; break;
    case TRUE_N: if(in_true_N) { title(mem,out); out<<true_N;} else newline=false; break; 
    case TRUE_ISOCLASS: if(in_true_IsoClass) { title(mem,out); out<<true_IsoClass; } else newline=false; break;
    case WEIGHT: if(in_weight) {title(mem,out); out<< weight; } else newline=false; break; 
    case DIMENSION: if(in_dimension) { title(mem,out); out<<dimension; } else newline=false; break;
    case TRUE_DIMENSION: if(in_true_dimension) { title(mem,out); out<<true_dimension; } else newline=false; break;
    case DISC: if(in_disc) { title(mem,out); out<<disc;} else newline=false; break; 
    case C_INF: if(in_c_inf) {title(mem,out); out<<c_inf;} else newline=false; break;
    case C_P: if(in_c_p) {title(mem,out); c_p.Save(out);} else newline=false; break;  
    case C_P_ARITH:  if(in_c_p_arith) {title(mem,out); c_p_arith.Save(out);} else newline=false; break; 
    case TORSIONBOUND: if(in_TorsionBound) {title(mem,out); out<<TorsionBound<<endl<<TorsionBound_prec;} else newline=false; break; 
    case CUSPORDER:  if(in_CuspOrder) {title(mem,out); out<<CuspOrder;} else newline=false; break; 
    case SHABOUND: if(in_ShaBound) {title(mem,out); out<<ShaBound;} else newline=false; break;
    case MODPOLAR: if(in_MP) {title(mem,out); MP.Save(out);} else newline=false; break; 
    case DEGMP: if(in_DegMP) {title(mem,out); out << DegMP;} else newline=false; break; 
    case KERMP: if(in_KerMP) {title(mem,out); KerMP.Save(out);} else newline=false; break;
    case BSDRATIO: if(in_BSDRatio) {title(mem,out); BSDRatio.Save(out); } else newline=false; break;
    case BSDRATIOTIMESTOR: if (in_BSDRatioTimesTorsionBound) {title(mem,out);
    out<<BSDRatioTimesTorsionBound; } else newline=false; break;
    case QEXPANSION: if(in_qExpansion) {title(mem,out); qExpansion.Save(out);} else newline=false; break;
    case QEXPF: if(in_qExp_f) {title(mem,out); qExp_f.Save(out);} else newline=false; break;
    case QINTBASIS: if(in_qIntegralBasis) {title(mem,out); qIntegralBasis.Save(out);} else newline=false; break;
    case PHIGOODPERIODS: if(in_PhiGoodPeriods) {title(mem,out); PhiGoodPeriods.Save(out);} else newline=false; break;
    case GOODPERIODS: if(in_GoodPeriods) {title(mem,out); GoodPeriods.Save(out);} else newline=false; break;
    case GOODPERIODINTEGRALS:  if(in_GoodPeriodIntegrals) 
	{title(mem,out); GoodPeriodIntegrals.Save(out);} else newline=false; break;
    case GOODPERIODINTEGRALCOEFFICIENTS:  if(in_GoodPeriodIntegralsCoefficients) 
	{title(mem,out); GoodPeriodIntegrals.Save(out); } else newline=false; break;
    case APPROXPERIODMAP_FROMIMPHI: if(in_ApproxPeriodMap_fromImPhi) 
	{title(mem,out); ApproxPeriodMap_fromImPhi.Save(out); } else newline=false; break;
    case APPROXPERIODMAP: if(in_ApproxPeriodMap) {title(mem,out); ApproxPeriodMap.Save(out); } else newline=false; break;
    case LVALUES: if(in_L) {title(mem,out); L.Save(out); } else newline=false; break;
    case LVALUES_DEFINED: if(in_L) { title(mem,out); L_defined.Save(out);} else newline=false; break;
    case LAMBDA: if(in_Lambda) {title(mem,out); Lambda.Save(out); } else newline=false; break;
    case LAMBDAPLUS: if(in_LambdaPlus) {title(mem,out); LambdaPlus.Save(out); } else newline=false; break;
    case LAMBDAMINUS:if(in_LambdaMinus) {title(mem,out); LambdaMinus.Save(out); } else newline=false; break;
    case OMEGAPLUS: if(in_OmegaPlus) {title(mem,out); out<<OmegaPlus;} else newline=false; break;
    case OMEGAMINUS: if(in_OmegaMinus) {title(mem,out); out<<OmegaMinus;} else newline=false; break;
    case ELLC4: if(in_ellcurve_invariants) {title(mem,out); out<<c4;} else newline=false; break;
    case ELLC6: if(in_ellcurve_invariants) {title(mem,out); out<<c6;} else newline=false; break;
    case ELLW1: if(in_ellcurve_invariants) {title(mem,out); out<<w1;} else newline=false; break;
    case ELLW2: if(in_ellcurve_invariants) {title(mem,out); out<<w2;} else newline=false; break;
    case ELLTAU: if(in_ellcurve_invariants) {title(mem,out); out<<tau;} else newline=false; break;
    case ELLJ: if(in_ellcurve_invariants) {title(mem,out); out<<ellj;} else newline=false; break;
    case EM: if(in_Em) {title(mem,out); Em.Save(out); } else newline=false; break; 
    case PHI: if(in_Phi) {title(mem,out); Phi.Save(out); } else newline=false; break; 
    case PHIPLUS: if(in_PhiPlus) {title(mem,out); PhiPlus.Save(out); } else newline=false; break; 
    case PHIMINUS: if(in_PhiMinus) {title(mem,out); PhiMinus.Save(out); } else newline=false; break; 
    case VSPACE: if(in_V) {title(mem,out); V.Save(out); } else newline=false; break; 
    case VSPACE_EMBEDDED: if(in_V) {title(mem,out); V_Embedded.Save(out); } else newline=false; break; 
    case VPLUS: if(in_Vplus) {title(mem,out); Vplus.Save(out); } else newline=false; break; 
    case VPLUS_EMBEDDED: if(in_Vplus) {title(mem,out); Vplus_Embedded.Save(out); } else newline=false; break; 
    case VMINUS: if(in_Vminus) {title(mem,out); Vminus.Save(out); } else newline=false; break; 
    case VMINUS_EMBEDDED: if(in_Vplus) {title(mem,out); Vminus_Embedded.Save(out); } else newline=false; break; 
    case VDUAL_SPACE: if(in_VD) {title(mem,out); VD.Save(out); } else newline=false; break; 
    case CONJ_V: if(in_Conj_V) {title(mem,out); Conj_V.Save(out); } else newline=false; break; 
    case CONJ_PHI: if(in_Conj_Phi) {title(mem,out); Conj_Phi.Save(out); } else newline=false; break; 
    case TP_V: if(in_Tp_V) {title(mem,out); Tp_V.Save(out); } else newline=false; break;
    case TP_VPLUS: if(in_Tp_Vplus) {title(mem,out); Tp_Vplus.Save(out); } else newline=false; break;
    case TP_VMINUS: if(in_Tp_Vminus) {title(mem,out); Tp_Vminus.Save(out); } else newline=false; break;
    case TP_PHI:  if(in_Tp_Phi) {title(mem,out); Tp_Phi.Save(out); } else newline=false; break;
    case TP_PHIPLUS: if(in_Tp_PhiPlus) {title(mem,out); Tp_PhiPlus.Save(out); } else newline=false; break;
    case TP_PHIMINUS: if(in_Tp_PhiMinus) {title(mem,out); Tp_PhiMinus.Save(out); } else newline=false; break;
    case TP_PHIDUALEVINV: if(in_Tp_Phi) {title(mem,out); Tp_Phi_Dual_EVinv.Save(out); } else newline=false; break;
    case TP_PHIPLUS_DUAL_EVINV: if(in_Tp_PhiPlus) {title(mem,out); 
    Tp_PhiPlus_Dual_EVinv.Save(out); } else newline=false; break;
    case TP_PHIMINUS_DUAL_EVINV: if(in_Tp_PhiMinus) {title(mem,out); 
    Tp_PhiMinus_Dual_EVinv.Save(out); } else newline=false; break;
    case TP_PHI_DUAL_E:  if(in_Tp_Phi) {title(mem,out); Tp_Phi_Dual_e.Save(out); } else newline=false; break;
    case TP_PHIPLUS_DUAL_E: if(in_Tp_PhiPlus) {title(mem,out); 
    Tp_PhiPlus_Dual_e.Save(out); } else newline=false; break;
    case TP_PHIMINUS_DUAL_E: if(in_Tp_PhiPlus) {title(mem,out);
    Tp_PhiMinus_Dual_e.Save(out); } else newline=false; break;
    case WQ: if(in_Wq) {title(mem,out); Wq.Save(out); } else newline=false; break; 
    case CONJ_PHISK: if(in_Conj_PhiSk) {title(mem,out); Conj_PhiSk.Save(out); }break;
    case PHISK: if(in_PhiSk) {title(mem,out); PhiSk.Save(out); } else newline=false; break;
    case PHISKINVERSE: if(in_PhiSkInverse) {title(mem,out); PhiSkInverse.Save(out); } else newline=false; break;
    case PHISKPLUS: if(in_PhiSkPlus) {title(mem,out); PhiSkPlus.Save(out); } else newline=false; break;
    case PHISKMINUS: if(in_PhiSkMinus) {title(mem,out); PhiSkMinus.Save(out); } else newline=false; break;
    case INTERSECTION: if(in_Intersection) {title(mem,out); Intersection.Save(out); } else newline=false; break;
    case FAST: if(in_Fast) {
	title(mem,out);
	Fast.Save(out); out<<endl;
	Fast_EVinv.Save(out); out<<endl;
	Fast_e.Save(out); out<<endl;
	out << Fast_e_pos <<endl;
	Fast_Field.Save(out); out<<endl;
	Fast_Tp.Save(out); out<<endl;
    } else newline=false; break;
    case HMODDEG: if(in_hmoddeg) {title(mem,out);HModDeg.Save(out);}else newline=false; break;
    case XMODDEG: if(in_xmoddeg) {title(mem,out);XModDeg.Save(out);}else newline=false; break;
    case XIMAGE: if(in_ximage) {title(mem,out);XImage.Save(out);}else newline=false; break;
    case XEIGEN: if(in_xeigen) {title(mem,out);XEigenV.Save(out);XEigenf.Save(out);}else newline=false; break;
    default:
	newline=false;
    };
    if(newline)
	out<<endl;
}

void  
Motive::Save()
{
    ofstream out(data_write(Name(), get_true_N()));
    if(!out) {
	mesgln("WARNING: Could not save motive.");
	return;
    }
//    mesg("Saving motive = ");
//    mesgln(Name());
    for(int i=0; i<int(MEMBER_EOF); i++) 
	Save(MEMBER_TYPE(i),out);
}

void 
Motive::Load(MEMBER_TYPE mem, ifstream& in)
{
//    mesg("Motive::Load  ");
//    mesgln(MemberToKeyword(mem));
    switch(mem) {
    case PRIMES: in_Prime=true;  Primes.Load(in); break;
    case EXPONENTS: in_Prime=true; Exponents.Load(in); break; 
    case DEFAULTPREC: in_DefaultPrec=true; in>>DefaultPrec;  break;
    case ISFULL: in_IsFull=true; in>>IsFull; break; 
    case ISNEW: in_is_new=true; in>>is_new;  break;
    case ISPLUS: in_plus=true; in>>plus; break;
    case MULTIPLICITY: in_multiplicity=true; in>>multiplicity; break;
    case NEW_N: in_new_N=true; in>>new_N;  break;
    case NEW_ISOCLASS: in_new_IsoClass=true; in>>new_IsoClass;  break;
    case TRUE_N: in_true_N=true; in>>true_N; break; 
    case TRUE_ISOCLASS: in_true_IsoClass=true; in>>true_IsoClass;  break;
    case WEIGHT: in_weight=true; in>> weight;  break; 
    case DIMENSION: in_dimension=true; in>>dimension;  break;
    case TRUE_DIMENSION: in_true_dimension=true;  in>>true_dimension;  break;
    case DISC: in_disc=true;  in>>disc; break; 
    case C_INF: in_c_inf=true;  in>>c_inf; break;
    case C_P: in_c_p=true;  c_p.Load(in); break; break; 
    case C_P_ARITH:  in_c_p_arith=true;  c_p_arith.Load(in); break; 
    case TORSIONBOUND: in_TorsionBound=true;  in>>TorsionBound; in>>TorsionBound_prec; break; 
    case CUSPORDER:  in_CuspOrder=true;  in>>CuspOrder; break; 
    case SHABOUND: in_ShaBound=true;  in>>ShaBound; break;
    case MODPOLAR: in_MP=true;  MP.Load(in); break; 
    case DEGMP: in_DegMP=true;  in >> DegMP; break; 
    case KERMP: in_KerMP=true;  KerMP.Load(in); break;
    case BSDRATIO: in_BSDRatio=true;  BSDRatio.Load(in);  break;
    case BSDRATIOTIMESTOR: in_BSDRatioTimesTorsionBound=true; 
	in>>BSDRatioTimesTorsionBound;  break;
    case QEXPANSION: in_qExpansion=true;  qExpansion.Load(in); break;
    case QEXPF: in_qExp_f=true;  qExp_f.Load(in); break;
    case QINTBASIS: in_qIntegralBasis=true;  qIntegralBasis.Load(in); break;
    case PHIGOODPERIODS: in_PhiGoodPeriods=true;  PhiGoodPeriods.Load(in); break;
    case GOODPERIODS: in_GoodPeriods=true;  GoodPeriods.Load(in); break;
    case GOODPERIODINTEGRALS:  in_GoodPeriodIntegrals=true;  GoodPeriodIntegrals.Load(in); break;
    case GOODPERIODINTEGRALCOEFFICIENTS:  in_GoodPeriodIntegralsCoefficients=true;  GoodPeriodIntegrals.Load(in);  break;
    case APPROXPERIODMAP_FROMIMPHI: in_ApproxPeriodMap_fromImPhi=true;  ApproxPeriodMap_fromImPhi.Load(in);  break;
    case APPROXPERIODMAP: in_ApproxPeriodMap=true;  ApproxPeriodMap.Load(in);  break;
    case LVALUES: in_L=true;  L.Load(in);  break;
    case LVALUES_DEFINED: in_L=true;  L_defined.Load(in); break;
    case LAMBDA: in_Lambda=true;  Lambda.Load(in);  break;
    case LAMBDAPLUS: in_LambdaPlus=true;  LambdaPlus.Load(in);  break;
    case LAMBDAMINUS:in_LambdaMinus=true;  LambdaMinus.Load(in);  break;
    case OMEGAPLUS: in_OmegaPlus=true;  in>>OmegaPlus; break;
    case OMEGAMINUS: in_OmegaMinus=true;  in>>OmegaMinus; break;
// POSSIBLE BUG: If somehow only the j one of these invariants is loaded
// the Motive object thinks they all are. 
    case ELLC4:  in>>c4; break;
    case ELLC6:   in>>c6; break;
    case ELLW1:   in>>w1; break;
    case ELLW2:  in>>w2; break;
    case ELLTAU:in>>tau; break;
    case ELLJ: in_ellcurve_invariants=true;  in>>ellj; break;
    case EM: in_Em=true;  Em.Load(in);  break; 
    case PHI: in_Phi=true;  Phi.Load(in);  break; 
    case PHIPLUS: in_PhiPlus=true;  PhiPlus.Load(in);  break; 
    case PHIMINUS: in_PhiMinus=true;  PhiMinus.Load(in);  break; 
    case VSPACE: in_V=true;  V.Load(in);  break; 
    case VSPACE_EMBEDDED: in_V=true;  V_Embedded.Load(in);  break; 
    case VPLUS: in_Vplus=true;  Vplus.Load(in);  break; 
    case VPLUS_EMBEDDED: in_Vplus=true;  Vplus_Embedded.Load(in);  break; 
    case VMINUS: in_Vminus=true;  Vminus.Load(in);  break; 
    case VMINUS_EMBEDDED: in_Vplus=true;  Vminus_Embedded.Load(in);  break; 
    case VDUAL_SPACE: in_VD=true;  VD.Load(in);  break; 
    case CONJ_V: in_Conj_V=true;  Conj_V.Load(in);  break; 
    case CONJ_PHI: in_Conj_Phi=true;  Conj_Phi.Load(in);  break; 
    case TP_V: in_Tp_V=true;  Tp_V.Load(in);  break;
    case TP_VPLUS: in_Tp_Vplus=true;  Tp_Vplus.Load(in);  break;
    case TP_VMINUS: in_Tp_Vminus=true;  Tp_Vminus.Load(in);  break;
    case TP_PHI:  in_Tp_Phi=true;  Tp_Phi.Load(in);  break;
    case TP_PHIPLUS: in_Tp_PhiPlus=true;  Tp_PhiPlus.Load(in);  break;
    case TP_PHIMINUS: in_Tp_PhiMinus=true;  Tp_PhiMinus.Load(in);  break;
    case TP_PHIDUALEVINV: Tp_Phi_Dual_EVinv.Load(in);  break;
    case TP_PHIPLUS_DUAL_EVINV:  Tp_PhiPlus_Dual_EVinv.Load(in);  break;
    case TP_PHIMINUS_DUAL_EVINV: Tp_PhiMinus_Dual_EVinv.Load(in);  break;
    case TP_PHI_DUAL_E:  Tp_Phi_Dual_e.Load(in);  break;
    case TP_PHIPLUS_DUAL_E: Tp_PhiPlus_Dual_e.Load(in);  break;
    case TP_PHIMINUS_DUAL_E: Tp_PhiMinus_Dual_e.Load(in);  break;
    case WQ: in_Wq=true;  Wq.Load(in);  break; 
    case CONJ_PHISK: in_Conj_PhiSk=true;  Conj_PhiSk.Load(in); break;
    case PHISK: in_PhiSk=true;  PhiSk.Load(in);  break;
    case PHISKINVERSE: in_PhiSkInverse=true;  PhiSkInverse.Load(in);  break;
    case PHISKPLUS: in_PhiSkPlus=true;  PhiSkPlus.Load(in);  break;
    case PHISKMINUS: in_PhiSkMinus=true;  PhiSkMinus.Load(in);  break;
    case INTERSECTION: in_Intersection=true;  Intersection.Load(in);  break;
    case FAST: in_Fast=true;
	Fast.Load(in);
	Fast_EVinv.Load(in);
	Fast_e.Load(in);
	in>>Fast_e_pos;
	Fast_Field.Load(in);
	Fast_Tp.Load(in);
	break;
    case HMODDEG: in_hmoddeg=true; HModDeg.Load(in); break;
    case XMODDEG: in_xmoddeg=true; XModDeg.Load(in); break;
    case XIMAGE:  in_ximage=true;  XImage.Load(in);  break;
    case XEIGEN:  in_xeigen=true;  XEigenV.Load(in); XEigenf.Load(in); break;
    default: 
	mesgln("A keyword in the data file was not recognized.");
	mesgln("This suggests the datafile may be corrupted.");
	mesgln("Proceed with caution.");
	die();
    };
}

char* get_next_keyword(istream& in)
{
    char c;
    do {
	if(in.eof()) return NULL;
	in.get(c);
    } while (c!='[');
    char* buf=new char[1024]; // no keyword is this long.
    int i=0;
    do {
	if(in.eof()) { delete buf; return NULL; }
	in.get(c);
	if(c!=']') buf[i++]=c;
	if(i>=1024) { mesgln("Keyword = "); mesgln(buf);
	die("get_next_keyword: Keyword buffer length exceeded. \
Input file probably corrupted."); }
    } while (c!=']');
    buf[i]='\0';
    do { in.get(c); } while(!in.eof() && c!='\n');
    return buf;
}

bool
Motive::Load()
{
    ifstream in(data_read(Name(),get_true_N()));
    if(!in) return false;
    char* kw;
    while((kw=get_next_keyword(in)) != NULL) {
	Load(KeywordToMember(kw), in);
	delete kw;
    }
    return true;
}

///////////////////////////////////////////////////////////////
////////////////// class MotiveDecomp //////////////////
///////////////////////////////////////////////////////////////

MotiveDecomp::MotiveDecomp(int _N, int _weight, bool _plus=false)
{
    in_ins();
    N=_N; 
    weight=_weight;
    plus=_plus;
    Load();
}

void 
MotiveDecomp::in_ins()
{
    in_H=in_H_Plus=in_MotiveList=in_FullMotive=in_Name=false;
}
    
MotiveDecomp::~MotiveDecomp()
{
    Save();
    if(in_MotiveList) {
	for(int i=0; i<numMotives; i++)
	    delete MotiveList[i];
	delete[] MotiveList;
    }
    if (in_FullMotive) {
	delete FullMotive;
    }
    if(in_H)
	delete H;
}
    
void                           
MotiveDecomp::write(OUTTYPE type, ostream& out, 
		    bool brief, int qPrec)
{
    if(num_Motives()==0)
	return;
    bool first;
    switch(type) {
    case TEXT:
	out << "Gamma_0("<<get_N()<<")"<<endl;
	out << "Weight " << get_weight()<< ""<<endl;
	out << endl << "-------------------------------------------------------"<<endl;
      break;

    case LATEX:
	out << "\\vspace{1ex}" << endl
	    << "\\parbox{\\textwidth}{"
	    << "\\motivehead"
	    << "{" << get_N() << "}{"<<get_N();
	if (!is_prime(get_N())) {
	    out << "=";
	    TexDisplay(bigint(get_N()),out);
	} 
	out << "}" << endl
	    << "\\fbox{\\parbox{\\textwidth}{" << endl
	    << "\\motivearr" << endl;

	first=true;
	for(int i=1; i<=num_Motives(); i++) {
	    Motive& m=get_Motive(i);
	    if(m.Is_New()) {
		m.write(type,out,brief,qPrec);
	    } else {
		if(first) {
		    out << "\\end{array}\\)" << endl
			<< "\\motiveoldkey{";
		    first=false;
		}
		else { out << ", \\quad "; }
		out << "\\mathbf{" << letter(i) << "}="
		    << m.get_new_N()
		    << letter(m.get_new_IsoClass());
	    }
	    free_Motive(i);
	}
	if(!first)
	    out << "}" << endl;
	else
	    out << "\\end{array}\\)" << endl;
	out << "}}}"<<endl;
	break;
    case PARI:
	for(int i=1; i<=num_Motives(); i++) {
	    Motive& m=get_Motive(i);
	    m.write(PARI,out,brief,qPrec);
	    free_Motive(i);
	}
	break;
    default:
	die("MotiveDecomp::write -- unreckognized type");
    };
}

/*
\newcommand{\motivehead}[2]{
 \parbox{\textwidth}{{\Large $\mathbf{#1}$} \hfill $N=#2$ \hfill {\Large $\mathbf{#1}$}}\vspace{1ex}\\
}
\newcommand{\motivearr}{\(
\begin{array}{lccccccccccl}
 &d&\Delta&w&c_p&c_\infty&|C|&|T|&\frac{|T|^2L}{\Omega}&|S|&m&\text{Cong.}\\
}
\newcommand{\motiveA}[9]{\mathbf{#1}&#2&#3&#4&#5&#6&#7&#8&#9}
\newcommand{\motiveB}[3]{&#1&#2&\parbox[t]{11em}{$#3$}}
\newcommand{\motiveoldkey}[1]{\\\mbox{}\dotfill\mbox{}\\\vspace{1ex}$#1$}
*/

/* Algorithm:
   I)  NEW: 
       1) Decompose the new space using Hecke operators.
       2) Construct each motive.
  II)  OLD: 
       1) For each M|N such that p=N/M is prime (and in order so M_0 < M_1, ...)
        a) Construct the Motive at level M. 
        b) Using Hecke operators, cut out and define the old 
           subMotives arising from it at level N.
 III)  ORDER:
       Using the order <, extending Cremona's, order the forms. 
*/

void 
MotiveDecomp::init_MotiveList()
{
    mesgln("\n\n*************************************");
    mesgln("init_MotiveList, N=",get_N());
    mesgln("*************************************\n\n");
    if(get_N()==0 || get_H_Plus()->dim_Sk()==0) {
	numMotives = 0;
	in_MotiveList=true;
	MotiveList=NULL;
    }

    MotiveList = new (Motive*)[get_H_Plus()->dim_Sk()];
    in_MotiveList=true;
    numMotives=0;
    if(get_H_Plus()->dim_Sk()==0) {
	mesgln("There are no motives at level ",get_N());
	return;
    }

    //       1) Decompose the new space using Hecke operators.
    if(get_H_Plus()->dim_Sknew()>0) {
	MatVecList<bigrational> E;
	Matrix<bigrational> V; get_H_Plus()->get_SknewPlus(V);
	cerr << "Dimension SknewPlus = " << V.ncols() << endl;
	if(!get_H_Plus()->DualBasisOfEigenforms(E,V))
	    die("MotiveDecomp::init_MotiveList -- couldn't decompose new space.");
	mesg("At level ",get_N());
	mesgln(" the number of new factors is ",E.len());
	//       2) Construct each motive.
	for(int i=0; i<E.len(); i++) {
	    MotiveList[numMotives] = new Motive(this, E.cM(i), E.cv(i));
	    MotiveList[numMotives]->set_true_IsoClass(numMotives+1);
	    numMotives++;
//	    free_Motive(numMotives);       // will need this later for sorting.
	}
    }
    mesgln("Constructed each new motive.");
    mesgln("There were ",numMotives);

    //       3) Sort. (# items small, so use "bubble sort").
    mesgln("Now sorting them.");
    for(int i=0; i<numMotives-1; i++) {
	if(!(*(MotiveList[i]) < *(MotiveList[i+1]))) {
	    cerr << "swapping " << i+1 << " and " << i+2 << endl;
	       Motive* t=MotiveList[i];
   	       MotiveList[i] = MotiveList[i+1];
	       MotiveList[i+1] = t;
	       i=-1;
	}
    }
    for(int i=0; i<numMotives; i++) {
	MotiveList[i]->set_true_IsoClass(i+1);
	MotiveList[i]->set_new_IsoClass(i+1);
    }

    // Save memory by freeing new motives from memory.
    // Their destructors will save them to disk for easy reload.
    in_MotiveList=true;		// to "fool" free_Motive().
    free_AllMotives();
    in_MotiveList=false;

    //       4) For each M|N get the newforms from level M.
    //          The divisors M of N proceed from largest to smallest, so that
    //          no additional sorting is necessary.
    if(get_H_Plus()->dim_Sknew() == get_H_Plus()->dim_Sk()) {
	mesgln("There are no old motives.");
    } else {
	Vector<bigint> divs;
	Divisors(divs, get_N());
	int savenumMotives=numMotives;

	for(int i=divs.size(); i>=1; i--) { 
	    int M=to_int(divs[i]);
	    if(M==get_N()) continue;
	    MotiveDecomp oldforms(M, get_weight(), Is_Plus());
	    cerr << "The number of forms coming from level " << M
		 << " is " << oldforms.num_Motives() << endl;
	    for(int j=1; j<=oldforms.num_Motives(); j++) 
		if(oldforms.get_Motive(j).Is_New()) {
		    MotiveList[numMotives] = new
			Motive(this, oldforms.get_Motive(j));
		    mesg("Initialized ", M); mesgln(letter(oldforms.get_Motive(j).get_true_IsoClass()));
		    oldforms.free_Motive(j);
		    MotiveList[numMotives]->set_true_IsoClass(numMotives+1);

                    // free, in order to save space. 
		    free_Motive(numMotives+1);
				// the destructor will save it to disk.
		    numMotives++;
		}
	}
	mesg("There were ",numMotives-savenumMotives);
	mesgln(" old motives.");
    }

    in_MotiveList=true;
}

int
MotiveDecomp::num_Motives()
{
    if(in_MotiveList)
	return numMotives;
    init_MotiveList();
    if(!in_MotiveList) die("MotiveDecomp::get_Motive(i) error with init_MotiveList.");
    return numMotives;
}

Motive&
MotiveDecomp::get_Motive(int i)
{
    if (i==0) 
	return get_FullMotive();
    if(in_MotiveList) {
	if (1<=i && i<=num_Motives()) {
	    if(MotiveList[i-1]==NULL) {
		bool success;
		MotiveList[i-1] = new Motive(success, this, i);
		if(!success) 
		    die("MotiveDecomp::get_Motive(i) -- unable to load previously saved motive.");
	    }
	    return (*(MotiveList[i-1]));
	}
	die("MotiveDecomp::get_Motive(i) : i out of range.");
    }
    init_MotiveList();
    if(!in_MotiveList) die("MotiveDecomp::get_Motive(i) error with init_MotiveList.");
    return get_Motive(i);
}

Motive&
MotiveDecomp::get_FullMotive()
{
    if(in_FullMotive)
	return *FullMotive;
    if(num_Motives()<1)
	die("MotiveDecomp::get_FullMotive -- there is none.");
    bool loaded;
    FullMotive = new Motive(loaded,this);
    in_FullMotive=true;
    if(loaded) return *FullMotive;
    
    mesgln("Creating full motive at level ", get_N());

    return *FullMotive;
}

void   
MotiveDecomp::free_Motive(int i)
{
    if(!in_MotiveList) 
	return;
    delete MotiveList[i-1];
    MotiveList[i-1]=NULL;
}

void
MotiveDecomp::free_AllMotives()
{
    for(int i=1; i<=num_Motives(); i++)
	free_Motive(i);
}

void
MotiveDecomp::free_FullMotive()
{
    if(!in_FullMotive)
	return;
    delete FullMotive;
    FullMotive=NULL;
    in_FullMotive=false;
}


H1NChiZ* 
MotiveDecomp::get_H()
{
    if(in_H)
	return H;
    H=new H1NChiZ(Character(get_N()), get_weight(), true, Is_Plus());
    in_H=true;
    return H;
}

H1NChiZ* 
MotiveDecomp::get_H_Plus()
{
    if(in_H_Plus)
	return H_Plus;
    H_Plus=new H1NChiZ(Character(get_N()), get_weight(), true, true);
    in_H_Plus=true;
    return H_Plus;
}

int
MotiveDecomp::get_N()
{
    return N;
}

int                            
MotiveDecomp::get_weight()
{
    return weight;
}

bool
MotiveDecomp::Is_Plus()
{
    return plus;
}

void
MotiveDecomp::Save()
{
    if(!in_MotiveList) return;
    ofstream out(data_write(Name(), get_N()));
    if(!out) {
	mesgln("WARNING: Could not save MotiveDecomp!");
	return;
    }
    out << num_Motives() << endl;
}

bool
MotiveDecomp::Load()
{
    if(in_MotiveList) return true; // no need.

    in_ins();
    ifstream in(data_read(Name(),get_N()));
    if(!in) return false;
    in >> numMotives;
    MotiveList = new (Motive*)[numMotives];
    for(int i=0; i<numMotives; i++)
	MotiveList[i]=NULL;
    in_MotiveList = true;

    // for safety -- should NOT be necessary
    mesgln("Verify that all new motives have already been correctly computed.");
    for(int i=1; i<=num_Motives(); i++) {
	bool success;
	Motive* m = new Motive(success, this, i);
	delete m;
	if(!success) {		// regenerate all new motives.
	    in_MotiveList=false;
	    return false;
	}
    }
    return true;
}
    
char*
MotiveDecomp::Name()
{
    if(in_Name)
	return strName;
    strName=new char[1024];
    ostrstream ost(strName, 1024);
    ost<< get_H()->Name() << "MotiveDecomp" << '\0';
    in_Name = true;
    return strName;
}




/////////// MOTIVECALCULATOR ////////////////
MotiveCalculator::MotiveCalculator()
{
    in_Decomp=false;
}

MotiveCalculator::~MotiveCalculator()
{
    if(in_Decomp)
	delete Decomp;
}

void
MotiveCalculator::Interface()
{
    cout << "Welcome to the motive calculator." << endl;
    plus = get_bool("WORK IN FAST MODE (+1 quotient, certain powers of 2 *wrong*)?  ");
    cout << "Type ? for help."<< endl << endl;
    while(1) {
	int N, weight, Iso;
	char* cmd; char* arg; char* user;
	cout << endl;
	if(ParseCommand(N,Iso,weight,cmd,arg,user=GetUserCommand())) {
	    if(N==0) { // system command
		// strip space?
		if(user[0]=='\\'&&user[1]=='Q')
		    return;
		if(user[0]=='?') {
		    if(user[1]=='\0')
			Help();
		    else
			Help(user+1);
		    continue;
		}
		if(strcmp(user,"TABLE_SPECIALVALUES")==0) {
		    int Nstart=get_int("N start = ");
		    int Nstop=get_int("N stop = ");
		    int weight=get_int("weight = ");
		    bool primeonly=get_bool("Prime level only [y/n]? ");
		    bool old=get_bool("include old factors = ");
		    
		    char buf[512];
		    ostrstream ost(buf,512);
		    ost << "SpecialValues_";
		    if(primeonly) ost<<"p"; else ost<<"N";
		    ost << Nstart << "-"
			<< Nstop << "_k"<< weight << ".gp" << '\0';
		    ofstream out(userdata_write(buf));
		    if(!out) {
			cout << "Error opening output file." << endl;
			continue;
		    }
		    if(old)
			PARITable_SpecialValuesOld(Nstart, Nstop, weight,
						   primeonly, out);
		    else
			PARITable_SpecialValues(Nstart, Nstop, weight,
						primeonly, out);
		    continue;
		}
		if(strcmp(user,"TABLE_CP")==0) {
		    int Nstart=get_int("N start = ");
		    int Nstop=get_int("N stop = ");
		    bool primeonly=get_bool("Prime level only [y/n]? ");
		    char buf[512];
		    ostrstream ost(buf,512);
		    ost << "cp_";
		    if(plus) ost<<"plus_";
		    if(primeonly) ost<<"p"; else ost<<"N";
		    ost << Nstart << "-"
			<< Nstop  << ".gp" << '\0';
		    ofstream out(userdata_write(buf));
		    if(!out) {
			cout << "Error opening output file." << endl;
			continue;
		    }
		    PARITable_cp(Nstart, Nstop, primeonly, out);
		    continue;
		}
		if(strcmp(user,"TABLE_SPLITTING")==0) {
		    int Nstart=get_int("N start = ");
		    int Nstop=get_int ("N stop  = ");
		    int weight=get_int("weight  = ");
		    bool primeonly=get_bool("Prime level only [y/n]? ");
		    char buf[512];
		    ostrstream ost(buf,512);
		    ost << "splitting_k" << weight << "_";
		    if(plus) ost<<"plus_";
		    if(primeonly) ost<<"p"; else ost<<"N";
		    ost << Nstart << "-"
			<< Nstop  << ".gp" << '\0';
		    ofstream out(userdata_write(buf));
		    if(!out) {
			cout << "Error opening output file." << endl;
			continue;
		    }
		    PARITable_splitting(Nstart, Nstop, weight,
					primeonly, out);
		    continue;
		}
		if(strcmp(user,"TABLE_NONMAXIMAL")==0) {
		    int Nstart=get_int("N start = ");
		    int Nstop=get_int ("N stop  = ");
		    int weight=get_int("weight  = ");
		    bool primeonly=get_bool("Prime level only [y/n]? ");
		    char buf[512];
		    ostrstream ost(buf,512);
		    ost << "nonmaximal_k" << weight << "_";
		    if(primeonly) ost<<"p"; else ost<<"N";
		    ost << Nstart << "-"
			<< Nstop  << ".gp" << '\0';
		    ofstream out(userdata_write(buf));
		    if(!out) {
			cout << "Error opening output file." << endl;
			continue;
		    }
		    PARITable_nonmaximal(Nstart, Nstop, weight,
					 primeonly, out);
		    continue;
		}
		if(strcmp(user,"TABLE_SS")==0) {
		    int Pstart=get_int("P start = ");
		    int Pstop=get_int("P stop = ");
		    char buf[512];
		    ostrstream ost(buf,512);
		    ost << "ss_p";
		    ost << Pstart << "-"
			<< Pstop  << ".gp" << '\0';
		    ofstream out(userdata_write(buf));
		    if(!out) {
			cout << "Error opening output file." << endl;
			continue;
		    }
		    PARITable_ssbasis(Pstart, Pstop, out);
		    continue;
		}
		if(strcmp(user,"TABLE_LCHINZ")==0) {
		    int Pstart=get_int("P start = ");
		    int Pstop=get_int("P stop = ");
		    char buf[512];
		    ostrstream ost(buf,512);
		    ost << "lchinz_p";
		    ost << Pstart << "-"
			<< Pstop  << ".gp" << '\0';
		    ofstream out(userdata_write(buf));
		    if(!out) {
			cout << "Error opening output file." << endl;
			continue;
		    }
		    PARITable_lchinz(Pstart, Pstop, out);
		    continue;
		}
		if(strcmp(user,"TABLE_EIGEN")==0) {
		    int Nstart=get_int("N start = ");
		    int Nstop=get_int ("N stop = ");
		    int weight=get_int("weight = ");
		    int prec=next_prime(get_int(
			"Compute a2, ..., ap, where p = ")-1);
		    char buf[512];
		    ostrstream ost(buf,512);
		    ost << "eigen_k" << weight 
			<< "_N" << Nstart << "-" << Nstop
			<< "_prec" << prec << ".gp" << '\0';
		    ofstream out(userdata_write(buf));
		    if(!out) {
			cout << "Error opening output file." << endl;
			continue;
		    }
		    PARITable_eigenvalues(Nstart, Nstop, weight, prec, out);
		    continue;
		}
		if(strcmp(user,"TABLE")==0) {
		    cout << "Create a LaTeX-able table (sans macros)."<<endl;
		    cout << "Output will be sent to the userdata directory."<<endl;
		    int Nstart=get_int("N start = ");
		    int Nstop=get_int("N stop = ");
		    int weight=get_int("weight = ");
		    bool primeonly=get_bool("Prime level only [y/n]? ");
		    
		    char buf[512];
		    ostrstream ost(buf,512);
		    ost << "MotiveArith_";
		    if(primeonly) ost<<"p"; else ost<<"N";
		    ost << Nstart << "-"
			<< Nstop << "_k"<< weight << ".gp" << '\0';
		    ofstream out(userdata_write(buf));
		    if(!out) {
			cout << "Error opening output file." << endl;
			continue;
		    }
		    Table_Arithmetic(Nstart, Nstop, weight, primeonly, out);
		    continue;
		} 
		cout << "Command " << user << " not recognized. " << endl;
		cout << "Please type ? for a list of commands.";
	    } else { // motive command
		Motive* m;
		if((m=get_Motive(N, Iso, weight)) != NULL) 
		    MotiveCommand(*m, cmd, arg);
		else 
		    cout << N << "k" << weight << letter(Iso) << " doesn't exist." << endl;
	    }
	} else {
	    cout << "Invalid input." << endl;
	    cout << "Please type ? for a list of commands.";
	}
	delete cmd; delete arg; delete user;
    }
}

void
MotiveCalculator::Welcome()
{
    cout << "MC: Motives From Modular Forms Calculator" << endl;
    cout << "Type ? for help." << endl;
}

void
MotiveCalculator::Help()
{
    cout << "                    THE MOTIVE CALCULATOR"<<endl;
    cout << endl;
    cout << "A motive object is specified by its weight, level, and isogeny class."<<endl;
    cout << "For example 37k4B denotes the second isogeny class for Gamma_0(37) of weight 4."<<endl;
    cout << "If omitted, the weight is assumed to be 2.  If the isogeny class"<<endl;
    cout << "is omitted then the motive consisting of all classes is specified."<<endl;
    cout << endl;
    cout << "Commands are issued as follows.  Feel free to try these out."<<endl
	 << "At any time, you can type '?' to see this screen again." << endl
	 <<endl;
    cout << "1. Giving a level and weight by itself lists all isogeny classes:"<<endl;
    cout << "      100            = list all isogeny classes of level 100 and weight 2"<<endl;
    cout << "      37k4           = list all classes of level 37 and weight 4"<<endl;
    cout << endl ;
    cout << "2. There are a LARGE number of methods which can be called"<<endl;
    cout << "   for a given motive.  These are called as in C++ by giving"<<endl;
    cout << "   the name of the motive, a dot '.', the method's name, and"<<endl;
    cout << "   any arguments enclosed in parenthesis."<<endl;
    cout << "      37B.tp_v(3)    = Hecke operator on integral modular symbols"<<endl;
    cout << "      37A.wq         = list signs of Atkin-Lehner involutions on 37A"<<endl;
    cout << "      27A.ellj       = j-invariant of the first CM curve of conductor 27"<<endl;
    cout << "      9k4A.ellj      = j-invariant of a certain weight 4 intermediate Jacobian"<<endl;
    cout << "      100A.fast(97)  = computes T_97 on dual of modular symbols"<<endl;
    cout << endl;
    cout << "3. Other" <<endl;
    cout << " * If the program can't determine a quantity it displays an '@' sign."<<endl;
    cout << " * (Almost) ALL computations are remembered and saved to disk for future use." << endl;
    cout << " * The parenthesis for a method taking no arguments are optional."<<endl;
    cout << " * A '.' refers to the last motive used, for example,"<<endl;
    cout << "      .dimension     = dimension of last motive (100k2A)"<<endl;

    cout << " * To compute an analytic invariant using more terms of the q-expansion:"<<endl
	 << "       1) use the method RESETANALYTIC to erase any saved values,"<<endl
	 << "       2) use QINTBASIS to increase the precision,"<<endl
	 << "       3) ask for the analytic invariant again." << endl;
    cout << endl
	 << "Further Commands:"<<endl
	 << "??       = list all methods"<<endl;
    cout << "?method  = description of a method" << endl;
    cout << "\\q       = quit."<<endl;
    cout << endl;

}

void
MotiveCalculator::ListCommands()
{
    for(int i=0; i<MEMBER_EOF; i++) 
	cout << MemberToKeyword(MEMBER_TYPE(i)) << endl;
    cout <<"RESETANALYTIC"<<endl
	 <<"WRITE"<<endl
	 <<"WRITELATEX"<<endl
	 <<"TABLE"<<endl
	 <<"TABLE_CP"<<endl
	 <<"TABLE_EIGEN"<<endl
	 <<"TABLE_LCHINZ"<<endl
	 <<"TABLE_NONMAXIMAL"<<endl
         <<"TABLE_SPECIALVALUES"<<endl
	 <<"TABLE_SPLITTING"<<endl
	 <<"TABLE_SS"<<endl
	 <<"BSDBOUND"<<endl
	 <<"LCHINZ"<<endl;
    cout << endl << endl;
}

void
MotiveCalculator::Help(const char* cmd)
{
    MEMBER_TYPE mem=KeywordToMember(cmd);
    if(mem!=MEMBER_EOF)
	cout << MemberToDescription(mem);
    else if(strcmp(cmd,"?")==0)
	ListCommands();
    else if(strcmp(cmd,"WRITE")==0)
	cout << "WRITE -- display most data about the motive.";
    else if(strcmp(cmd,"WRITELATEX")==0)
	cout << "WRITELATEX -- display data about the motive in Latex format.";
    else if(strcmp(cmd,"TABLE")==0)
	cout << "TABLE -- create latex table of arithmetic invariants.";
    else if(strcmp(cmd,"TABLE_SPECIALVALUES")==0)
	cout << "TABLE_SPECIALVALUES -- create PARI table of rational parts of special values.";
    else if(strcmp(cmd,"TABLE_CP")==0)
	cout << "TABLE_CP -- create PARI table of component group data.";
    else if(strcmp(cmd,"TABLE_SS")==0)
	cout << "TABLE_SS -- create PARI table of eigenforms on supersingular basis.";
    else if(strcmp(cmd,"TABLE_SPLITTING")==0)
	cout << "TABLE_SPLITTING -- create table of splittings of the new part";
    else if(strcmp(cmd,"TABLE_LCHINZ")==0)
	cout << "TABLE_LCHINZ -- table of degrees of nonvanishing twists (Merel project)";
    else if(strcmp(cmd,"TABLE_EIGEN")==0)
	cout << "TABLE_EIGEN -- create table of Hecke eigenvalues.";
    else if(strcmp(cmd,"\\Q")==0)
	cout << "\\q -- quit.";
    else if(strcmp(cmd,"RESETANALYTIC")==0)
	cout << "RESETANALYTIC -- reset all analytic invariants.  Thus next time they are"<<endl
	     << "computed the precision from the latex call to QEXPANSION will be used."<<endl;
    else if(strcmp(cmd,"LCHINZ")==0)
	cout << "LCHINZ -- determines whether or not L(M_f,chi,1)=/=0 (computation is approximate, so beware...)";
    else {
	cout << "There is no help on '"<<cmd<<"'." << endl;
	cout << "Please recheck and try again, or type '?'"<<endl;
    }
    cout<<endl;
}

void
MotiveCalculator::MotiveCommand(Motive& m,
				const char* cmd,
				const char* arg)
{
    if(cmd==NULL) {// summary
	mesgln("MOTIVECOMMAND = Summarize level.");
//	cerr << "Motive's class = "<<endl;
//	cerr << m.get_true_IsoClass() << endl;
	if(m.get_true_IsoClass() == 0) 
	    SummarizeLevel(m.get_true_N(), m.get_weight());
	else
	    SummarizeMotive(m);
	return;
    }
    if(strcmp(cmd,"WRITELATEX")==0) {
	if(m.get_true_IsoClass() == 0) 
	    WriteLevel(m.get_true_N(), m.get_weight(),LATEX);
	else
	    m.write(LATEX);
	return;
    }
    if(strcmp(cmd,"WRITE")==0) {
	if(m.get_true_IsoClass() == 0) 
	    WriteLevel(m.get_true_N(), m.get_weight(),TEXT);
	else
	    m.write(TEXT);
	return;
    }
    if(strcmp(cmd,"RESETANALYTIC")==0) {
	m.reset_Analytic_Invariants();
	cout << "Analytic invariants reset."<<endl;
	return;
    }

    if(strcmp(cmd,"BSDBOUND")==0) {
	int bound=0;
	if(arg!=NULL) bound=parse_int(arg);
	if(m.Is_New()) {
	    if(bound>0)
		mesgln("Computing BSD at 1 to precision = ", bound);
	    Display(m.get_BSDBound(1,bound));
	}
	else
	    cout << "Not new, so not computing BSD."<<endl;
	return;
    }
    
    if(strcmp(cmd,"LCHINZ")==0) {
	if (m.get_weight()!=2) {
	    cout << "LCHINZ: For now, weight must be 2."<<endl;
	    return; 
	}
	if (!m.Is_New()) {
	    cout << "LCHINZ: For now, motive must be new."<<endl;
	    return;
	}
	int d;
	bool is;
        if((d=parse_int(arg))>0) {
	    Array<int> degs(1); degs[0]=d;
	    if(d<=0 || (m.get_true_N()-1)%d!=0) {
		cout << "LCHINZ: invalid degree = " << d << endl;
		return;
	    }
	    Character chi(m.get_true_N(), 1, true, &degs);
	    is=m.Is_LchiNone_nz(chi);
	} else {
	    cout << "Define character:"<<endl;
	    Character chi;
	    if (chi.modulus() != m.get_true_N()) {
		cout << "LCHINZ: For now, character "
		     << "must be modulo level N."<<endl;
		return;
	    }
	    is=m.Is_LchiNone_nz(chi);
	}
	cout << "L(M, chi, 1)";
	if(is) 
	    cout << " =/= 0"<<endl;
	else
	    cout << " = 0"<<endl;
	return;
    }


    MEMBER_TYPE mem = KeywordToMember(cmd);
    int p,b,s,prec,isog;	// used in certain cases. 
    switch(mem) {
    case PRIMES: cout << "Internal use only."; break;
    case EXPONENTS: cout << "Internal use only."; break;
    case DEFAULTPREC: cout << m.get_DefaultPrec(); break;
    case ISFULL: cout << m.Is_Full(); break; 
    case ISNEW: cout << m.Is_New(); break;
    case ISPLUS: cout << m.Is_Plus(); break;
    case MULTIPLICITY: cout << m.get_multiplicity(); break;
    case NEW_N: cout << m.get_new_N() << " = "; Display(m.get_new_N()); break;
    case NEW_ISOCLASS: cout << letter(m.get_new_IsoClass()); break;
    case TRUE_N: cout << m.get_true_N() << " = "; Display(m.get_true_N()); break; 
    case TRUE_ISOCLASS: cout << letter(m.get_true_IsoClass()); break;
    case WEIGHT: cout << m.get_weight(); break; 
    case DIMENSION: cout << m.get_dimension(); break;
    case TRUE_DIMENSION: cout << m.get_true_dimension(); break;
    case DISC: if(!m.Is_New() || m.Is_Full()) {cout << "Must be new."; break; }
	Display(m.get_disc()); break; 
    case C_INF:
	cout << m.get_c_inf(); break;
    case C_P: 
        if((p=parse_int(arg))==0) {
	    for(int i=1; i<=m.get_num_PrimeDivisorsOfN(); i++) {
		bigint cp=m.get_c_p(m.get_nth_PrimeDivisorsOfN(i));
		if(cp==0) cout << "@"; else cout << cp;
		if(i<m.get_num_PrimeDivisorsOfN()) cout <<",";
	    }
	} else if(m.get_true_N()%p==0) {
		bigint cp=m.get_c_p(p);
		if(cp==0) cout << "@"; else Display(m.get_c_p(p));
	}
	break; 
    case C_P_ARITH:  
        if((p=parse_int(arg))==0) {
	    for(int i=1; i<=m.get_num_PrimeDivisorsOfN(); i++) {
		bigint cp=m.get_c_p_arith(m.get_nth_PrimeDivisorsOfN(i));
		if(cp==0) cout << "@"; else cout << cp;
		if(i<m.get_num_PrimeDivisorsOfN()) cout <<",";
	    }
	} else if(m.get_true_N()%p==0) {
		bigint cp=m.get_c_p_arith(p);
		if(cp==0) cout << "@"; else Display(m.get_c_p_arith(p));
	}
	break; 
    case TORSIONBOUND:
	b=parse_int(arg);
	if(b<=3) b=3;
	cout << "gcd {#M(F_p) : (p,2N)=1, p<="<<b<<"} = " << endl;
	Display(m.get_TorsionUpperBound(b));
    break;
    case CUSPORDER:
	if(m.get_weight()==2) 
	    Display(m.get_CuspOrder(parse_bigrational(arg))); 
	else {
	    b=get_int("q=",0,m.get_weight()-2);
	    Display(m.get_CuspOrder(parse_bigrational(arg),b)); 
	}
	break;
    case SHABOUND: if(!m.Is_New()){cout << "@"; break; }
	Display(m.get_ShaBound()); break;
    case MODPOLAR: cout << m.get_MP(); break; 
    case DEGMP: Display(m.get_DegMP()); break; 
    case KERMP:  DisplayAbGp(m.get_KerMP(),cout); break;
    case BSDRATIO:  
	if(arg==NULL) s=1; else s=parse_int(arg);
	if(1<=s && s<=m.get_weight()-1) {
	    if(m.Is_New())
		Display(m.get_BSDRatio(s));
	    else 
		if(s%2)
		    cout << m.get_L(s) / (m.get_c_inf()*m.get_OmegaPlus());
		else
		    cout << m.get_L(s) / (m.get_c_inf()*m.get_OmegaMinus());
	} else { cout << "@"; } break;
    case BSDRATIOTIMESTOR:
	if(arg==NULL) s=1; else s=parse_int(arg);
	if(1<=s && s<=m.get_weight()-1) {
	    bigint b=m.get_TorsionUpperBound()*m.get_TorsionUpperBound();
	    if(m.Is_New())
		Display(m.get_BSDRatio(s)*b);
	    else
		if(s%2)
		    cout << b*m.get_L(s) / (m.get_c_inf()*m.get_OmegaPlus());
		else
		    cout << b*m.get_L(s) / (m.get_c_inf()*m.get_OmegaMinus());
	}
	else { cout << "@"; } break;
    case QEXPANSION:
	if(!m.Is_New()) {
	    cout <<"Must be new.  Use QINTBASIS instead.";
	    break;
	}
	prec=parse_int(arg);
	if(prec>=0) {
	    cout<<endl;
	    m.get_qExp_f().polprint(cout,"a");
	    cout << " = 0"<<endl<<endl;
	    const Matrix<bigrational>& qexp=m.get_qExpansion(prec);
	    if(prec==0) prec=qexp.ncols();
	    DisplayQExpansion(qexp,m.get_qExp_f(), prec, "a", false);
	} else
	    cout<<"QEXPANSION takes a positive argument.";
    break;
    case QEXPF: if(!m.Is_New()){cout << "Must be new."; break; } m.get_qExp_f().polprint(); break; 
    case QINTBASIS: prec=parse_int(arg);
	if(prec>=0) 
  	    DisplayQExpansion(m.get_qIntegralBasis(prec), prec);
        else
	    cout<<"QINTBASIS takes a positive argument.";

    break; 
    case PHIGOODPERIODS: cout << m.get_PhiGoodPeriods(); break; 
    case GOODPERIODS: cout<<"Internal use only."; break;
    case GOODPERIODINTEGRALS: cout << m.get_GoodPeriodIntegrals(); break;
    case GOODPERIODINTEGRALCOEFFICIENTS: cout << m.get_GoodPeriodIntegralsCoefficients(); break;
    case APPROXPERIODMAP_FROMIMPHI: cout << m.get_ApproxPeriodMap_fromImPhi(); break;
    case APPROXPERIODMAP: cout << m.get_ApproxPeriodMap(); break; 
    case LVALUES:
	s=parse_int(arg); if(1<=s && s<=m.get_weight()-1)
	Display(m.get_L(s)); break;
    case LVALUES_DEFINED: cout << "Internal use only."; break;
    case LAMBDA:  Display(m.get_Lambda()); break;
    case LAMBDAPLUS: Display(m.get_LambdaPlus()); break; 
    case LAMBDAMINUS: Display(m.get_LambdaMinus()); break; 
    case OMEGAPLUS: Display(m.get_OmegaPlus()); break;
    case OMEGAMINUS: Display(m.get_OmegaMinus()); break;
    case ELLC4: if(!m.Is_New()){cout << "Must be a simple new factor."; break; }
	if(m.get_dimension()!=1){cout << "Must be one dimensional."; break; }
	if(m.Is_Plus()){cout << "Not available in fast +1 mode."; break; }
	Display(m.get_c4()); break; 
    case ELLC6: if(!m.Is_New()){cout << "Must be a simple new factor."; break; }
	if(m.get_dimension()!=1){cout << "Must be one dimensional."; break; }
	if(m.Is_Plus()){cout << "Not available in fast +1 mode."; break; }
	Display(m.get_c6()); break; 
    case ELLW1: if(!m.Is_New()){cout << "Must be a simple new factor."; break; }
	if(m.get_dimension()!=1){cout << "Must be one dimensional."; break; }
	if(m.Is_Plus()){cout << "Not available in fast +1 mode."; break; }
        Display(m.get_w1()); break; 
    case ELLW2: if(!m.Is_New()){cout << "Must be a simple new factor."; break; }
	if(m.get_dimension()!=1){cout << "Must be one dimensional."; break; }
	if(m.Is_Plus()){cout << "Not available in fast +1 mode."; break; }
	Display(m.get_w2()); break;  
    case ELLTAU: if(!m.Is_New()){cout << "Must be a simple new factor."; break; }
	if(m.get_dimension()!=1){cout << "Must be one dimensional."; break; }
	Display(m.get_tau()); break; 
    case ELLJ: if(!m.Is_New()){cout << "Must be simple new factor."; break; }
	if(m.get_dimension()!=1){cout << "Must be one dimensional."; break; }
	if(m.Is_Plus()){cout << "Not available in fast +1 mode."; break; }
	Display(m.get_ellj()); break;
    case EM: cout << m.get_Em(); break; 
    case PHI: cout << m.get_Phi(); break; 
    case PHIPLUS: cout << m.get_PhiPlus(); break; 
    case PHIMINUS: cout << m.get_PhiMinus(); break;
    case VSPACE: cout << m.get_V(); break; 
    case VSPACE_EMBEDDED: cout << m.get_V_Embedded(); break;
    case VPLUS: cout << m.get_Vplus(); break; 
    case VPLUS_EMBEDDED: cout << m.get_Vplus_Embedded(); break; 
    case VMINUS: cout << m.get_Vminus(); break; 
    case VMINUS_EMBEDDED: cout << m.get_Vminus_Embedded(); break; 
    case VDUAL_SPACE: cout << m.get_VDual(); break;
    case CONJ_V: cout << m.get_Conj_V(); break; 
    case CONJ_PHI: cout << m.get_Conj_Phi(); break;
    case TP_V:  p=parse_int(arg); if(p>=1) cout << m.get_Tp_V(p); break;
    case TP_VPLUS: p=parse_int(arg); if(p>=1) cout << m.get_Tp_Vplus(p); break; 
    case TP_VMINUS: p=parse_int(arg); if(p>=1) cout << m.get_Tp_Vminus(p); break; 
    case TP_PHI:  p=parse_int(arg); if(is_prime(p)) cout << m.get_Tp_Phi(p); break; 
    case TP_PHIPLUS:p=parse_int(arg); if(is_prime(p)) cout << m.get_Tp_PhiPlus(p); break; 
    case TP_PHIMINUS: p=parse_int(arg); if(is_prime(p)) cout << m.get_Tp_PhiMinus(p); break; 
    case TP_PHIDUALEVINV: cout << "internal use only."; break;
    case TP_PHIPLUS_DUAL_EVINV: cout << "internal use only."; break;
    case TP_PHIMINUS_DUAL_EVINV: cout << "internal use only."; break;
    case TP_PHI_DUAL_E: cout << "internal use only."; break;
    case TP_PHIPLUS_DUAL_E: cout << "internal use only."; break;
    case TP_PHIMINUS_DUAL_E: cout << "internal use only."; break;
    case WQ:  if(!m.Is_New()){cout << "Must be a simple new factor."; break; }
	p=parse_int(arg);
	if(p==0)
	    cout << m.get_WqString(); 
	else
	    cout << m.get_Wq(p);
	break;
    case CONJ_PHISK: cout << m.get_Conj_PhiSk(); break;
    case PHISK: cout << m.get_PhiSk(); break;
    case PHISKINVERSE: cout << m.get_PhiSkInverse(); break;
    case PHISKPLUS: cout << m.get_PhiSkPlus(); break;
    case PHISKMINUS: cout << m.get_PhiSkMinus(); break;
    case INTERSECTION: isog=parse_letter_to_num(arg);
	if(1<=isog && isog<=m.get_Decomp()->num_Motives() && isog!=m.get_true_IsoClass())
	    cout << m.get_Intersection(isog);
	else
	    cout << "Invalid isogeny class.";
	break;
    case FAST: p=parse_int(arg);
	Display(m.get_Tp_Fast(p)); cout << endl;
        DisplayPol(m.get_charpoly_Tp(p),1,1);
	break;
    case HMODDEG: 
        if((p=parse_int(arg))==0) {
	    for(int i=1; i<=m.get_num_PrimeDivisorsOfN(); i++) {
		bigint deg=m.get_HModDeg(m.get_nth_PrimeDivisorsOfN(i));
		if(deg==0) cout << "@"; else cout << deg;
		if(i<m.get_num_PrimeDivisorsOfN()) cout <<",";
	    }
	} else if(m.get_true_N()%p==0) {
		bigint deg=m.get_HModDeg(p);
		if(deg==0) cout << "@"; else Display(deg);
	}
	break; 
    case XMODDEG: 
        if((p=parse_int(arg))==0) {
	    for(int i=1; i<=m.get_num_PrimeDivisorsOfN(); i++) {
		bigint deg=m.get_XModDeg(m.get_nth_PrimeDivisorsOfN(i));
		if(deg==0) cout << "@"; else cout << deg;
		if(i<m.get_num_PrimeDivisorsOfN()) cout <<",";
	    }
	} else if(m.get_true_N()%p==0) {
		bigint deg=m.get_XModDeg(p);
		if(deg==0) cout << "@"; else Display(deg);
	}
	break; 
    case XIMAGE: 
        if((p=parse_int(arg))==0) {
	    for(int i=1; i<=m.get_num_PrimeDivisorsOfN(); i++) {
		bigint deg=m.get_XImage(m.get_nth_PrimeDivisorsOfN(i));
		if(deg==0) cout << "@"; else cout << deg;
		if(i<m.get_num_PrimeDivisorsOfN()) cout <<",";
	    }
	} else if(m.get_true_N()%p==0) {
		bigint deg=m.get_XImage(p);
		if(deg==0) cout << "@"; else Display(deg);
	}
	break; 
    case XEIGEN:
	if(m.get_num_PrimeDivisorsOfN()!=1) {
	    cout << "Sorry, XEIGEN currently only programmed for prime level."<<endl;
	} else {
	    int p = m.get_true_N();
	    const Matrix<bigrational>& v=m.get_XEigenV();
	    Vector<ffield_element> b=m.get_XSSBasis();
	    cout << endl;
	    cout << "Supersingular basis: " << endl
		 << "b = " << b << "; " << endl << endl;
	    if(m.get_XEigenf().poldeg()>1) {
		cout << "0 = ";
		m.get_XEigenf().polprint(cout,"a");
		cout << endl;
	    }
	    cout << "0 = " << ffield_element::get_def_base() ;
	    cout << endl;
	    bool first=true;
	    cout << "Sum representation: "<<endl;
	    for(int i=1; i<=b.size(); i++) 
		if(!v.row(i).iszero()) {
		    if(!first)
			cout << " + " ;
		    else first=false;
		    int d=v.row(i).poldeg();
		    if(d>0) cout << "(";
		    v.row(i).polprint(cout, "a");
		    if(d>0) cout << ")";
		    cout << "*[" << b[i] << "]";
		}
	    cout << endl << endl << "Vector representation: "<<endl;
	    cout << "v = [";
	    m.get_XEigenf().polprint(cout,"a");
	    cout << ", [";
	    first=true;
	    for(int i=1; i<=b.size(); i++) {
		if(!first) cout << ", " ; else first=false;
		v.row(i).polprint(cout, "a");
	    }
	    cout << "]];"<<endl;
	}
	break;
	
    default:
	cout << "Command " << cmd << " not recognized. " << endl;
	cout << "Please type ? for a list of commands.";
	break;

    };
    cout << endl;
}

void 
MotiveCalculator::Table_Arithmetic(int Nstart, int Nstop, int weight,
				      bool primeonly, ostream& out)
{
    if(Nstart<=0 || Nstop<=0 || Nstart>Nstop
       || weight<2 || weight%2!=0)
	die("MotiveCalculator::Table_Arithmetic -- invalid parameters.");

    out << "\\\\M[trueN,trueIso]=[newN,newIso,dim,disc,Wqstring,[cpvec],c_inf,\n"
	<< "\\\\     order(0-oo),Torsbound,(L(1)/Omega)*Torsbound*c_inf,\n"
        << "\\\\     ShaBound,modulardegree,[[1,[A intersect this],...]]];"<<endl;
    for(int N=Nstart; N<=Nstop; N++) 
	if(!primeonly || is_prime(N)) {
	    MotiveDecomp md(N, weight);
	    md.write(PARI,out);
	}
}

void
MotiveCalculator::PARITable_SpecialValues(int Nstart, int Nstop, int weight,
					  bool primeonly, ostream& out)
{
    if(Nstart<=0 || Nstop<=0 || Nstart>Nstop
       || weight<2 || weight%2!=0)
	die("MotiveCalculator::PARITable_SpecialValues -- invalid parameters.");
    
    out<<"\\\\SV[num] = [N,weight,isoclass,dimension,prod w_q, [T*L(1)/O,T],...]"<<endl;
    out<<"\\\\SV = vector(M,x,0); where M is big enough."<<endl<<endl;
    int j=1;
    for(int N=Nstart; N<=Nstop; N++) 
	if(!primeonly || is_prime(N)) {
	    MotiveDecomp md(N, weight);
	    for(int i=1; i<=md.num_Motives(); i++) {
		Motive& m=md.get_Motive(i);
		if(m.Is_New()) {
		    out<<"SV["<<j++<<"] \t= \t["
		       <<N<<","<<weight<<","<<i<<","
		       <<m.get_dimension()<<","
		       <<m.get_WqProd()<<",";
		    for(int s=1;s<=m.get_weight()-1;s++) {
			bigrational bsd=m.get_BSDRatio(s)*m.get_c_inf();
			bigint T=m.get_CuspOrder(0,s-1);
			out<<"["<<T*bsd<<","<<T<<"]";
			if(s<m.get_weight()-1) out<<",";
		    }
		    out<<"];"<<endl;
		}
	    }
	}
    
}
    
void
MotiveCalculator::PARITable_SpecialValuesOld(int Nstart, int Nstop, int weight,
					  bool primeonly, ostream& out)
{
    if(Nstart<=0 || Nstop<=0 || Nstart>Nstop
       || weight<2 || weight%2!=0)
	die("MotiveCalculator::PARITable_SpecialValues -- invalid parameters.");
    
    out<<"\\\\SV[num] = [N,weight,isoclass,dimension,prod w_q, [T*L(1)/O,T],...]"<<endl;
    out<<"\\\\SV = vector(M,x,0); where M is big enough."<<endl<<endl;
    int j=1;
    for(int N=Nstart; N<=Nstop; N++) 
	if(!primeonly || is_prime(N)) {
	    MotiveDecomp md(N, weight);
	    for(int i=1; i<=md.num_Motives(); i++) {
		Motive& m=md.get_Motive(i);
		out<<"SV["<<j++<<"] \t= \t["
		   <<N<<","<<weight<<","<<i<<","
		   <<m.get_dimension()<<",";
		for(int s=1;s<=m.get_weight()-1;s++) {
		    bigint T=m.get_CuspOrder(0,s-1);
		    if(m.Is_New()) {
			bigrational bsd=m.get_BSDRatio(s)*m.get_c_inf();
			out<<"["<<T*bsd<<","<<T<<"]";
		    } else {
			bigcomplex bsd;
			bigcomplex twopi;
			power(twopi,2*Pi(),s-1);
			bigcomplex f=factorial(bigint(s-1))/twopi;
			if(s%2) 
			    bsd=m.get_L(s) / m.get_OmegaPlus() * f;
			else
			    bsd=m.get_L(s) / m.get_OmegaMinus() * f;
			out<<"["<<T*bsd<<","<<T<<"]";
		    }
		    if(s<m.get_weight()-1) out<<",";
		}
		out << "];"<<endl;
	    }
	}
}

void
MotiveCalculator::PARITable_cp(int Pstart, int Pstop,
			       bool primeonly, ostream& out)
{
    if(!primeonly) {
	cout << "MotiveCalculator::PARITable_cp only implimented if primeonly=true."<<endl;
	out << "MotiveCalculator::PARITable_cp only implimented if primeonly=true."<<endl;
	return;
    }
    out << "\\\\cp[N,iso] = " << endl
	<< "\\\\ [dimension, " << endl
	<< "\\\\  m = modular degree, " << endl
	<< "\\\\      [wq = atkin, " << endl
	<< "\\\\       mh = homology p-new mod deg (=m if p is prime), " <<endl
	<< "\\\\       mx = char. group p-new mod deg,"<<endl
	<< "\\\\       im = order of image of comp group of J in comp group of A ],"<<endl
	<< "\\\\      [...], ..., [...] ];"<<endl
        << "\\\\ There is one entry [...] for each p dividing N," <<endl
	<< "\\\\ from smallest to largest."<<endl<<endl
        << "\\\\ Theorem: #(comp group over Fpbar) = im * (m / mx)."<<endl
	<< "\\\\ Conjecture: mx = mh."<<endl
	<< endl << "\\\\ " << signature_stamp() << endl 
	<< "\\\\ " << date_stamp() << endl << endl;
    
    if(!is_prime(Pstart)) Pstart=next_prime(Pstart);
    for (int p=Pstart; p<=Pstop; p=next_prime(p)) {
	MotiveDecomp* md = new MotiveDecomp(p, 2);
	for(int i=1; i<=md->num_Motives(); i++) {
	    Motive& m=md->get_Motive(i);
	    out << "cp[" << p << "," << i << "] = ["
		<< m.get_dimension() << ","
		<< m.get_DegMP() << ",["
		<< m.get_Wq(p) << ","
		<< m.get_HModDeg(p) << ","
		<< m.get_XModDeg(p) << ","	
		<< m.get_XImage(p) << "]];" << endl;
	    md->free_Motive(i);
	}
	delete md;
    }
}

void
MotiveCalculator::PARITable_nonmaximal(int Nstart, int Nstop, int weight,
				       bool primeonly, ostream& out)
{
    out << "\\\\ D[N,k] = [[disc(O_f), disc(K_f)],...];"<<endl
	<< "\\\\  gives discriminants of ring Z[...,an,...] and "<<endl
	<< "\\\\  of the maximal order for each newform of "<<endl
	<< "\\\\  level N and weight k."<<endl
	<< "\\\\ " << signature_stamp() << endl;
    
    if(primeonly && !is_prime(Nstart)) Nstart=next_prime(Nstart);
    for (int N=Nstart; N<=Nstop;) {
	MotiveDecomp* md = new MotiveDecomp(N, weight);
	out << "D[" << N << ","<<weight<<"] = [";
	bool first=true;
	for(int i=1; i<=md->num_Motives(); i++) {
	    Motive& m=md->get_Motive(i);
	    if(!m.Is_New()) break;
	    if(first) first=false; else out<<", ";
	    out << "[" << m.get_disc() << flush;
            out << "," << nfdisc(m.get_qExp_f()) << flush;
	    out << "]";
	}
	out << "];" << endl;
	delete md;
	if(primeonly) N=next_prime(N); else N++;
    }
}
void
MotiveCalculator::PARITable_splitting(int Nstart, int Nstop, int weight,
				      bool primeonly, ostream& out)
{
    out << "\\\\ S[N,k] = [[dim,[wp]], ..., [dim,[wp]]]" << endl
	<< "\\\\        = ordered list of new factors of level N, weight k."<<endl
        << "\\\\    dim = dimension of factor." << endl
	<< "\\\\     wp = signs of Atkin-Lehner involutions at level N."<<endl
	<< "\\\\ " << endl
	<< "\\\\ " << signature_stamp() << endl 
	<< "\\\\ " << date_stamp << endl << endl;
    
    if(primeonly && !is_prime(Nstart)) Nstart=next_prime(Nstart);
    for (int N=Nstart; N<=Nstop;) {
	MotiveDecomp* md = new MotiveDecomp(N, weight);
	out << "S[" << N << ","<<weight<<"] = [";
	bool first=true;
	for(int i=1; i<=md->num_Motives(); i++) {
	    Motive& m=md->get_Motive(i);
	    if(!m.Is_New()) break;
	    if(first) first=false; else out<<", ";
	    out << "["
		<< m.get_dimension()<<",[";
	    for(int n=1; n<=m.get_num_PrimeDivisorsOfN(); n++) {
		int p=m.get_nth_PrimeDivisorsOfN(n);
		if(m.get_Wq(p)==1)
		    out << "1";
		else
		    out << "-1";
		if(n<m.get_num_PrimeDivisorsOfN())
		    out << ",";
	    }
	    out << "]]";
	}
	out << "];" << endl;
	delete md;
	if(primeonly) N=next_prime(N); else N++;
    }
}

void
MotiveCalculator::PARITable_ssbasis(int Pstart, int Pstop,
				    ostream& out)
{
    out << "\\\\ ssbasis[p]   = [f(x), [j1(x), ..., jn(x)]];" << endl
	<< "\\\\ ssmon  [p]   = [w1, ..., wn]; "<<endl
	<< "\\\\ sseigen[p,1] = [g(x), [a1(x), ..., an(x)]];" << endl
	<< "\\\\  ...................." << endl
	<< "\\\\ sseigen[p,m] = [g(x), [a1(x), ..., an(x)]];" << endl
	<< "\\\\ " << endl
	<< "\\\\ ssbasis = supersingular basis"<<endl
	<< "\\\\ ssmon   = monodromy pairing weights"<<endl
	<< "\\\\ sseigen = ordered list of eigenvectors on supersingular basis"<<endl
	<< endl << "\\\\ " << signature_stamp() << endl << endl;
    
    if(!is_prime(Pstart)) Pstart=next_prime(Pstart);
    for (int p=Pstart; p<=Pstop; p=next_prime(p)) {
	MotiveDecomp* md = new MotiveDecomp(p, 2);
	if(md->num_Motives()<1) continue;
	out<<"ssbasis["<<p<<"] \t= [";
	Vector<ffield_element> basis = md->get_Motive(1).get_XSSBasis();
	Vector<bigint> f; ffield_element::irred_polynomial(f);
	f.polprint(out);
	out<< ", "<<basis<<"];"<<endl;
	out<<"ssmon["<<p<<"] \t= "
	   << md->get_Motive(1).get_XMonodromy() << ";"<<endl;

	for(int i=1; i<=md->num_Motives(); i++) {
	    Motive& m=md->get_Motive(i);
	    out << "sseigen[" << p << "," << i << "] \t= [";
	    m.get_XEigenf().polprint(out);
	    out << ", [";
	    const Matrix<bigrational>& v=m.get_XEigenV();
	    Vector<ffield_element> b=m.get_XSSBasis();
	    bool first=true;
	    for(int j=1; j<=b.size(); j++) {
		if(!first) out << ", " ; else first=false;
		v.row(j).polprint(out);
	    }
	    out << "]];"<<endl;
	    md->free_Motive(i);
	}
	delete md;
	out << endl;
    }
}

void
MotiveCalculator::PARITable_lchinz(int Pstart, int Pstop,
				   ostream& out)
{
    out << "\\\\ lchinz[p,i]  = [divisors d of p-1 such that" << endl
	<< "\\\\                 L(A_f,chi_d,1)=/=0, where "<<endl
	<< "\\\\                 chi_d:(Z/pZ)^*-->C^* is the "<<endl
	<< "\\\\                 character of degree d."<<endl
	<< "\\\\ Computation was done using modular symbols"<<endl
	<< "\\\\ and we consider L(A_f,chi_d,1)=/=0 if the value"<<endl
	<< "\\\\ of a vector dual to f on the modular symbol" << endl
	<< "\\\\   s = sum_{a in (Z/pZ)^*} chi_d(a) {0,-1/a}" << endl
	<< "\\\\ is < 10^(-10)." << endl
	<< endl << "\\\\ " << signature_stamp() << endl << endl;
    
    if(!is_prime(Pstart)) Pstart=next_prime(Pstart);
    for (int p=Pstart; p<=Pstop; p=next_prime(p)) {
	MotiveDecomp* md = new MotiveDecomp(p, 2);
	if(md->num_Motives()<1) continue;
	for(int i=1; i<=md->num_Motives(); i++) {
	    Motive& m=md->get_Motive(i);
	    out << "lchinz[" << p << "," << i << "] \t= [";
	    Vector<bigint> divs;
	    Divisors(divs, p-1);
	    bool first=true;
	    if(m.get_BSDRatio(1)!=0) {
		if(!first) out << ", "; else first=false;
		out << "1";
	    }
	    for(int j=2; j<=divs.size(); j++) {
		Array<int> degs(1); degs[0]=to_int(divs[j]);
		Character chi(p, 1, true, &degs);
		if(m.Is_LchiNone_nz(chi)) {
		    if(!first) out << ", "; else first=false;
		    out << divs[j];
		}
	    }
	    out << "];"<<endl;
	    md->free_Motive(i);
	}
	delete md;
    }
}

void
MotiveCalculator::PARITable_eigenvalues(int Nstart, int Nstop,
					int weight, int prec,
					ostream& out)
{
    out << "\\\\ This is a table of the first " << prec
	<<     " Hecke eigenvalues"<<endl
	<< "\\\\ of newforms for Gamma_0(N)."<<endl
	<< "\\\\ Notation:" << endl
	<< "\\\\ f[N,k]  = [ " << endl
	<< "\\\\            [g(x), [wq's], [a2(x),...,ap(x)]],"<<endl
        << "\\\\                  ...  "<<endl
        << "\\\\            [g(x), [wq's], [a2(x),...,ap(x)]]" <<endl
        << "\\\\          ];"<<endl
        << "\\\\ Where Q_f = Q(...a_n...) = Q[x]/g(x)."<<endl
	<< "\\\\ The newforms are ordered by dimension, "<< endl
	<< "\\\\ then W_q's, then traces."<<endl
	<< endl << "\\\\ " << signature_stamp() << endl << endl;

    out << "\\\\ uncomment the following two lines to make PARI-readable."<<endl;
    out << "\\\\ {" << endl;
    out << "\\\\ f = matrix(" << Nstop << "," << weight<<");" << endl<<endl;
    for (int N=Nstart; N<=Nstop; N++){
	MotiveDecomp* md = new MotiveDecomp(N, weight);
	if(md->num_Motives()<1 || !md->get_Motive(1).Is_New()) continue;
	out<<"f["<<N<<"," << weight <<"]=["<<endl;
	for(int i=1; i<=md->num_Motives(); i++) {
	    Motive& m=md->get_Motive(i);
	    if(!m.Is_New()) break;
	    if(i>1) out << ","<<endl;
	    out << "[";
	    m.get_qExp_f().polprint(out);
	    out << ", [";
	    for(int n=1; n<=m.get_num_PrimeDivisorsOfN(); n++) {
		if(n>1) out<<",";
		out << m.get_Wq(m.get_nth_PrimeDivisorsOfN(n));
	    }
	    out << "], [";
	    const Matrix<bigrational>& Q = m.get_qExpansion(prec);
	    for(int p=2; p<=prec; p=next_prime(p)) {
		if(p>2) out << ",";
		Q.col(p).polprint(out);
	    }
	    out << "]]";
	}
	out << "];"<<endl<<endl;
	delete md;
    }
}

void
MotiveCalculator::SummarizeLevel(int N, int weight)
{
    mesgln("****** SUMMARIZE LEVEL.");
    if(N <= 0) 
	cout << "The level should be a positive integer." << endl;
    if(weight<2 || weight%2 !=0) {
	cout << "The weight should be even and >= 2." << endl;
	return;
    }
    MotiveDecomp& md=get_MotiveDecomp(N,weight);
    cout << N << "k"<<weight<<":\tdim\tW"<<endl;
    for(int i=1; i<=md.num_Motives(); i++) {
	Motive& m = md.get_Motive(i);
	cout << letter(m.get_true_IsoClass());
	if(!m.Is_New())
	    cout << "(" << m.get_new_N() << letter(m.get_new_IsoClass()) << ")";
	cout << "\t" << m.get_dimension();
	if(!m.Is_New()) cout << "*" << m.get_multiplicity();
	if(m.Is_New()) cout << "\t"<<m.get_WqString();
	cout << endl;
    }
}

void
MotiveCalculator::WriteLevel(int N, int weight, OUTTYPE type)
{
    if(N <= 0) 
	cout << "The level should be a positive integer." << endl;
    if(weight<2 || weight%2 !=0) {
	cout << "The weight should be even and >= 2." << endl;
	return;
    }
    get_MotiveDecomp(N,weight).write(type, cout, false);
}


void 
MotiveCalculator::SummarizeMotive(Motive& m)
{
    cout << m.get_true_N() << "k" << m.get_weight() <<letter(m.get_true_IsoClass())<<": \t";
    if(!m.Is_New())
	cout << "(" << m.get_new_N() << letter(m.get_new_IsoClass()) << ")";
    cout << "\t" << m.get_dimension();
    if(!m.Is_New()) cout << "*" << m.get_multiplicity();
    if(m.Is_New()) cout << "\t"<<m.get_WqString();
    cout << endl;
}

/* Example Commands:
   
   100   - list all motives at level 100, weight 2: N=100, iso=0, cmd=NULL, arg=NULL.
   100k2 - same as "100"
   37B.tp_v(3) =   N=37, k=2, iso=2, cmd="tp_v", arg="3".
   37A.wq(37) = N=37, k=2, iso=1, cmd="wq", arg="37"
   37k4A        = N=37, k=4, iso=1, cmd="", arg=""    (empty command lists basis data)
*/

// NOTE: cmd is always all upper case.
bool 
MotiveCalculator::ParseCommand(int& N, int& iso, int& k,   // N = 0 for "system" commands.
			       char*& cmd, char*& arg,     // must be deleted by caller.
			       char* in)
{
    static int last_N=-1, last_iso=1, last_k=2;
    // Make upper case.
    int n=strlen(in);
    for(int i=0; i<n; i++)		
	if(in[i]>='a' && in[i]<='z')
	    in[i]=in[i]-'a'+'A';

    // Defaults. 
    N=0; k=2; iso=0; cmd=arg=NULL; 

    int s=0,t=0;			// current spot in "in" buffer.
    char* buf=new char[n];
    if(in[0]=='.') {
	if(last_N==-1) {delete buf; return false;} // no previous!
	N=last_N;
	iso=last_iso;
	k=last_k;
	if(n==1) { delete buf; return true;}
    } else {
	// first get N
	while(s<n && in[s]>='0' && in[s]<='9')
	    buf[s]=in[s++];
	if(s==0) { delete buf; last_N=N;last_iso=iso;last_k=k; return true; }
	buf[s]='\0';
	N = atoi(buf);
	
	t=0;
	if(in[s]=='K') {		// get weight
	    t=0; 
	    s++;
	    while(s<n && in[s]>='0' && in[s]<='9')
		buf[t++]=in[s++];
	    if(t==0)
		iso=14;
	    else {
		buf[t]='\0';
		k=atoi(buf);
	    }
	}
	

	// get isoclass
	if(iso==0) {
	    t=0;
	    while(s<n && in[s]>='A' && in[s]<='Z')
		buf[t++]=in[s++];
	    if(t>0) {
		buf[t]='\0';
		iso=parse_letter_to_num(buf);
	    }
	}
	
	if(s==n) { delete buf; last_N=N;last_iso=iso;last_k=k; return true; }
	if(in[s]!='.') { delete buf; return false; }
    }

    // get command
    s++;
    t=0;
    while(s<n && in[s]!='(')
	buf[t++]=in[s++];
    if(t>0) {
	buf[t]='\0';
	cmd = new char[strlen(buf)+1];
	strcpy(cmd,buf);
    } else { 
	delete buf;
	return false;
    }

    if(s>=n) { delete buf; last_N=N;last_iso=iso;last_k=k;return true; }

    // get arguments to command
    if(in[s]!='(') {delete buf; return false; }
    s++;
    t=0;
    while(s<n && in[s]!=')')
	buf[t++]=in[s++];
    if(t>0) {
	buf[t]='\0';
	arg = new char[strlen(buf)+1];
	strcpy(arg,buf);
    }

    if(in[s]!=')' || n<s-1) {delete buf; return false;}

    delete buf;
    last_N=N;last_iso=iso;last_k=k;
    return true;
}

char*
MotiveCalculator::GetUserCommand() // returned string must be deleted by caller.
{
    char* buf=new char[1024];
    get_str(buf, 1024, "? ");
    // delete spaces.
    char* buf2=new char[1024];
    int s=0, n=strlen(buf);
    for(int i=0; i<n; i++)
	if(buf[i]!=' ')
	    buf2[s++]=buf[i];
    buf2[s]='\0';
    delete buf;
    return buf2;
}

MotiveDecomp&
MotiveCalculator::get_MotiveDecomp(int N, int weight)
{
    if(N<=0) 
	die("MotiveCalculator::get_MotiveDecomp(int N):  N<=0, invalid.");
    if(!in_Decomp) {
	Decomp = new MotiveDecomp(N, weight, plus);
	in_Decomp=true;
    } else {
	if(Decomp->get_N() != N || Decomp->get_weight() != weight) {
	    delete Decomp;
	    Decomp = new MotiveDecomp(N, weight, plus);
	}
    }
    return *Decomp;
}

Motive*
MotiveCalculator::get_Motive(int true_N, int true_IsoClass, int weight)
{
    if(true_N > 5000) {
	bool sure=get_bool("WARNING: Level is very large. Proceed?");
	if(!sure) return FALSE;
    }
    if(true_N <= 0 || true_IsoClass<0 || weight<2 || weight%2 !=0) return NULL;
    MotiveDecomp& md=get_MotiveDecomp(true_N, weight);
    if(true_IsoClass > md.num_Motives() || md.num_Motives()==0)
	return NULL;
    return &(md.get_Motive(true_IsoClass));
}
