/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

// motive.h 

#ifndef MOTIVE_H
#define MOTIVE_H

#include <iostream.h>
#include "h1nchiz.h"
#include "matrix.h"
#include "vector.h"
/* WARNINGS:
   * "Deg_MP" means the modular degree -- which is the sqrt(deg modular polarization).
     I shouldn't have called it deg mp, but I'm afraid to change it now for fear
     of mucking something else up.
    
 */

class MotiveDecomp;
enum MEMBER_TYPE {
    PRIMES=0, EXPONENTS, DEFAULTPREC,
    ISFULL, ISNEW, ISPLUS, MULTIPLICITY, NEW_N, NEW_ISOCLASS,
    TRUE_N, TRUE_ISOCLASS, WEIGHT,  DIMENSION, TRUE_DIMENSION,
    DISC, C_INF, C_P, C_P_ARITH, TORSIONBOUND, CUSPORDER, SHABOUND,
    MODPOLAR, DEGMP, KERMP, BSDRATIO, BSDRATIOTIMESTOR, QEXPANSION, 
    QEXPF, QINTBASIS, PHIGOODPERIODS, GOODPERIODS, GOODPERIODINTEGRALS,
    GOODPERIODINTEGRALCOEFFICIENTS, APPROXPERIODMAP_FROMIMPHI,
    APPROXPERIODMAP, LVALUES, LVALUES_DEFINED, LAMBDA,
    LAMBDAPLUS, LAMBDAMINUS, OMEGAPLUS, OMEGAMINUS, 
    ELLC4, ELLC6, ELLW1, ELLW2, ELLTAU, ELLJ,
    EM, PHI, PHIPLUS, PHIMINUS, VSPACE, VSPACE_EMBEDDED,
    VPLUS, VPLUS_EMBEDDED, VMINUS, VMINUS_EMBEDDED, VDUAL_SPACE, 
    CONJ_V, CONJ_PHI, TP_V, TP_VPLUS, TP_VMINUS, TP_PHI, TP_PHIPLUS,
    TP_PHIMINUS, TP_PHIDUALEVINV, TP_PHIPLUS_DUAL_EVINV,
    TP_PHIMINUS_DUAL_EVINV, TP_PHI_DUAL_E, TP_PHIPLUS_DUAL_E,
    TP_PHIMINUS_DUAL_E,
    WQ, CONJ_PHISK, PHISK, PHISKINVERSE, PHISKPLUS, PHISKMINUS,
    INTERSECTION, FAST, 
    HMODDEG,
    XMODDEG, XIMAGE, XEIGEN,
    MEMBER_EOF };

// Display types
enum OUTTYPE { TEXT, LATEX, PARI, PARSABLE } ;

/* DATA STRUCTURE:
   Fast = left subspace of Dual of Modsym^+(Q). 
   Phi = the subspace of Dual of ModSym.
   Em  = the Right subspace of ModSym.
   V   = the Right subspace of CuspSym, viewed as a free abelian group on its own,
         cut out.  Thus V sits in a free Z-module of rank equal
         to dim CuspSym.
   V_Embedded = V but embedded in ModSym
   VDual = the Right subspace of CuspSym cut out by
   VD_Embedded = VDual embedded in ModSym
   Plus, Minus = the +1 and -1 subspaces for complex conj. of each of the above.

   true_N = the level at which the Motive sits. 
   true_IsoClass = isogeny class at the level at which the Motive sits. 
   new_N  = where it was new

 */

class Motive
{
public:				// methods
    // Load motive   IsoClass = -0 (full J_0), 1, 2, 3, ...
    // Usually this called when *loading*.
    // success = true <===> object was successfully loaded. 
    Motive(bool&                           success,
	   MotiveDecomp*                   Decomp,
	   int                             true_IsoClass);
    // Create motive which is new, i.e., true_N=new_N.
    Motive(MotiveDecomp*                   Decomp,
	   const Matrix<bigrational>&      Fast,
	   const Vector<bigrational>&      Fast_Field);
    // Construct motive from one which is old and of lower level. 
    Motive(MotiveDecomp* Decomp, Motive& old);
    // Fullmotive. 
    Motive(bool& loaded,	// true <==> able to load the motive from disk.
	   MotiveDecomp* Decomp);
    void RaiseOldBasis(const Matrix<bigrational>& OldBasis,
		       int old_true_N);

    // Destructor. 
    ~Motive();

    // copying motives is NOT allowed.
    Motive(const Motive&);
    Motive& operator=(const Motive&);
    // Warning! BUG: Please don't compare the motive to itself.
    bool operator<(Motive&);

    void                           set_Phi(const Matrix<bigrational>& Phi);
    void                           set_Decomp(MotiveDecomp* Decomp);

    // I. OUTPUT
    void                           write(OUTTYPE type, ostream& out=cout, 
					 bool brief=false, int qPrec=17);

    // II. INVARIANTS
    // 1. naming
    bool                           Is_New();
    bool                           Is_Full();    // Full J0
    bool                           Is_Plus();
    int                            get_multiplicity();
    int                            get_new_N();
    void                           set_new_N(int N);
    int                            get_new_IsoClass();
    void                           set_new_IsoClass(int iso);
    int                            get_true_N();
    void                           set_true_N(int N);
    int                            get_true_IsoClass();
    void                           set_true_IsoClass(int iso);
    int                            get_weight();
    void                           set_weight(int weight);
    int                            get_DefaultPrec(); // for q-expansions.
    void                           set_DefaultPrec(int prec);


    // 2. arithmetic
    const int&                     get_dimension();
    const int&                     get_true_dimension();
    const bigrational&             get_disc();
    const bigint&                  get_c_inf() ;
    const bigint&                  get_c_p(int p) ;
    const bigint&                  get_c_p_arith(int p) ;
    const bigint&                  get_HModDeg(int p); // p-new homology mod. deg
    const bigint&                  get_XModDeg(int p); // p-new character group mod. deg.
    const bigint&                  get_XImage(int p); // |image of comp group in Phi_A|
    // get_XEigen: eigenform on ss basis of D. (only imp for prime level, now).
    // The columns of returned matrix represent elements of the
    // field Q(x)/f(x).  E.g., a column [a,b,c]~ corr to a+b*x+c*x^2 mod f(x).
    const Matrix<bigrational>&     get_XEigenV(int p=0); 
    const Vector<bigrational>&     get_XEigenf(int p=0);
    // get_Xssbasis: (only imp for prime level, now).
    // Return supersingular j's of basis of D. 
    // The character group X of torus is the degree 0 divisors inside D. 
    Vector<ffield_element>         get_XSSBasis(int p=0);
    Vector<int>                    get_XMonodromy(int p=0);

    const bigint&                  get_CuspOrder();
    bigint                         get_CuspOrder(const bigint& a, const bigint& b,
						 const bigint& c, const bigint& d,
						 int q=0);
    bigint                         get_CuspOrder(const bigrational& cusp,
						 int q=0);
    const bigint&                  get_TorsionUpperBound(int prec=37);
    const Matrix<bigrational>&     get_MP();
    const bigint&                  get_DegMP() ;
    const Vector<bigint>&          get_KerMP();
    // Use only the first bound Hecke operators, but if bound=0,
    // uses Sturm to get true BSD Value.
    bigrational                    get_BSDBound(int s, int bound=0);
    const bigrational&             get_BSDRatio(int s) ;
    const bigint&                  get_BSDRatioTimesTorsionBound();
    const bigint&                  get_ShaBound();

    // Returns true iff L(f,chi,1)=/=0.
    // For now, assumes k=2, cond(chi) = N, and Is_New=true.
    bool                           Is_LchiNone_nz(const Character& chi);

    // 3. modular forms, q-expansion at oo. 
    const Matrix<bigrational>&     get_qExpansion(int prec=0); // if 0, default_prec.
    const Vector<bigrational>&     get_qExp_f();
    const Matrix<bigrational>&     get_qIntegralBasis(int prec=0);
				// saturates its input :)
    void                           set_qIntegralBasis(Matrix<bigrational>
	                                              RationalBasis);
       // used by q-expansion functions:
    Vector<bigrational>            get_LeftEigenE();
    Vector<bigrational>            get_LeftEigenTpE(int p);
    const Matrix<bigrational>&     get_LeftEigenvectorVector();
    const Vector<bigrational>&     get_LeftEigenvectorField();
    void                           set_LeftEigenvector(const Matrix<bigrational>& vector,
						       const Vector<bigrational>& field);

    // 4. analytic
    void                           reset_Analytic_Invariants();
    const bigcomplex&              get_L(int s);   // sign info is lost. 
				// lattice is spanned by the columns of Lambda.
				// the number of columns = dimension.
    const Matrix<bigcomplex>&      get_Lambda();
    const Matrix<bigcomplex>&      get_LambdaPlus();
    const Matrix<bigcomplex>&      get_LambdaMinus();
    const bigcomplex&              get_OmegaPlus();    // identity component
    const bigcomplex&              get_OmegaMinus();
    const bigcomplex&              get_tau();
    const bigcomplex&              get_c4();
    const bigcomplex&              get_c6();
    const bigcomplex&              get_w1();
    const bigcomplex&              get_w2();
    const bigcomplex&              get_ellj();
    const Matrix<bigrational>&     get_PhiGoodPeriods();
    const MatVecList<bigint>&      get_GoodPeriods();
    const Matrix<bigcomplex>&      get_GoodPeriodIntegrals();
    const Matrix<bigcomplex>&      get_GoodPeriodIntegralsCoefficients();
    const Matrix<bigcomplex>&      get_ApproxPeriodMap_fromImPhi();
    const Matrix<bigcomplex>&      get_ApproxPeriodMap();

    // III. EMBEDDINGS
    const Matrix<bigrational>&     get_Em();
    const Matrix<bigrational>&     get_Phi();
    const Matrix<bigrational>&     get_PhiPlus();
    const Matrix<bigrational>&     get_PhiMinus();
    const Matrix<bigrational>&     get_V();  
    const Matrix<bigrational>&     get_Vplus();  
    const Matrix<bigrational>&     get_Vminus();  
    const Matrix<bigrational>&     get_V_Embedded();  
    const Matrix<bigrational>&     get_Vplus_Embedded();  
    const Matrix<bigrational>&     get_Vminus_Embedded();  
    const Matrix<bigrational>&     get_VDual();	

    // IV. OPERATORS
    const Vector<bigrational>&     get_charpoly_Tp(int p);// charpoly on PhiPlus.
    const Matrix<bigrational>&     get_Tp_Fast(int p);
    void                           get_Fast_TpE(Vector<bigrational>& v, int p);
    const Matrix<bigrational>&     get_Tp_Vplus(int p);
    const Matrix<bigrational>&     get_Tp_Vminus(int p);
    const Matrix<bigrational>&     get_Tp_V(int p);
    const Matrix<bigrational>&     get_Tp_Phi(int p);
    const Matrix<bigrational>&     get_Tp_PhiPlus(int p);
    const Matrix<bigrational>&     get_Tp_PhiMinus(int p);
    bigrational                    get_TraceOf_ap(int p);
    const Matrix<bigrational>&     get_Conj_V();
    const Matrix<bigrational>&     get_Conj_Phi();
    const Matrix<bigrational>&     get_Conj_PhiSk();
    const char*                    get_WqString(); // +++---+
    int                            get_Wq(int p);
    int                            get_WqProd();   // prod_p|N w_q.
    const Matrix<bigrational>&     get_PhiSk();
    const Matrix<bigrational>&     get_PhiSkInverse();
    const Matrix<bigrational>&     get_PhiSkPlus();
    const Matrix<bigrational>&     get_PhiSkMinus();

    // V. INTERSECTION: Group structure of intersection of this with
    // the IsoClass factor inside J_0(N).
    // Vector = [0] means this.IsoClass == IsoClass.
    // Vector = [2,7] means the intersection is Z/2 x Z/7.
    // Vector = [] means the intersection is just the zero group.
    const Vector<bigint>&          get_Intersection(int IsoClass);
    void                           set_Intersection(int IsoClass,
						    const Vector<bigint>& congs);

    // VI. OTHER
    // Compute Phi(Tx) where T is the sub Z-module of the Hecke algebra
    // generated by T_1,...,T_bound.  If bound=0 the Sturm bound is used.
    void                           PhiTx(Matrix<bigrational>& PTx,
					 const Vector<bigrational>& x,
					 int bound=0);
    // Assumes p|N.  If so, returns i where p is the ith prime divisor of true_N.
    int                            get_PrimeIndex(int p);
    int                            get_num_PrimeDivisorsOfN();
    int                            get_nth_PrimeDivisorsOfN(int n);
    int                            get_nth_PrimeExponentOfN(int n);

    // VII. DEFINITION
    H1NChiZ*                       get_H();
    H1NChiZ*                       get_H_Plus();
    MotiveDecomp*                  get_Decomp();

    // FAST (F is dual and column vectors!)
    // Note: Fast is a subspace of dual of the PLUS space of modular symbols.
    void                           set_Fast(const Matrix<bigrational>& Fast,
					    const Vector<bigrational>& Field);
    // For a non-new motive Fast is not an eigenvector so there is no field...
    void                           set_Fast(const Matrix<bigrational>& Fast);
    const Matrix<bigrational>&     get_Fast();
    const Matrix<bigrational>&     get_Fast_EVinv();
    Vector<bigrational>            get_Fast_times_e();
    const Vector<int>&             get_Fast_e();
    const Vector<bigrational>&     get_Fast_Field();

private:			// functions
    int                            get_next_goodp(int p=1);
    void                           assure_new(const char* mesg);
    char*                          Name();
    bool                           Load();
    void                           Load(MEMBER_TYPE, ifstream& in);
    void                           Save();
    void                           Save(MEMBER_TYPE, ofstream& out);
    void                           init_ins();
    void                           init_ellcurve_invariants();


private:			// data
    bool     in_H; 
    bool     delete_H;
    H1NChiZ* H;

    bool     in_Name;
    char*    strName;

    // 0. definition
    bool in_Decomp, delete_Decomp, in_plus, in_IsFull, in_Prime, in_DefaultPrec;
    MotiveDecomp*                  Decomp;
    bool                           plus;
    bool                           IsFull;
    Vector<int>                    Primes, Exponents;
    int                            DefaultPrec;

    // 1. naming
    bool in_is_new, in_multiplicity, in_new_N,
	 in_new_IsoClass, in_true_N, in_true_IsoClass, in_weight;
    bool                           is_new;
    int                            multiplicity;
    int                            new_N;
    int                            new_IsoClass;
    int                            true_N;
    int                            true_IsoClass;
    int                            weight;

    // 2. arithmetic
    bool in_dimension, in_true_dimension, in_disc, in_c_inf, in_c_p, in_c_p_arith,
	 in_hmoddeg, in_xmoddeg, in_ximage, in_xeigen,
	 in_TorsionBound,  in_ShaBound,
	 in_DegMP, in_MP, in_KerMP, in_BSDRatio,
	 in_BSDRatioTimesTorsionBound, in_CuspOrder;
    int                            dimension, true_dimension;
    bigrational                    disc;
    bigint                         c_inf;
    Vector<bigint>                 c_p;
    Vector<bigint>                 c_p_arith;
    Vector<bigint>                 HModDeg, XModDeg, XImage;
    Matrix<bigrational>            XEigenV;
    Vector<bigrational>            XEigenf;
    bigint                         TorsionBound;
    int                            TorsionBound_prec;
    bigint                         CuspOrder;
    bigint                         ShaBound;
    Matrix<bigrational>            MP;
    bigint                         DegMP;
    Vector<bigint>                 KerMP;
    Vector<bigrational>            BSDRatio;
    bigint                         BSDRatioTimesTorsionBound;

    // 3. modular forms
    bool in_qExpansion, in_qExp_f, in_qIntegralBasis,
	 in_LeftEigenE, in_LeftEigenvector;
    Matrix<bigrational>            qExpansion;
    Vector<bigrational>            qExp_f;
    Matrix<bigrational>            qIntegralBasis;

    // 4. analytic
    bool in_L, in_Lambda, in_LambdaPlus, in_LambdaMinus, 
 	 in_OmegaPlus, in_OmegaMinus, in_ellcurve_invariants,
	 in_PhiGoodPeriods, in_GoodPeriods, in_GoodPeriodIntegrals,
	 in_GoodPeriodIntegralsCoefficients,
	 in_ApproxPeriodMap_fromImPhi,
	 in_ApproxPeriodMap;
    Matrix<bigrational>            PhiGoodPeriods;
    MatVecList<bigint>             GoodPeriods;
    Matrix<bigcomplex>             GoodPeriodIntegrals;
    Matrix<bigcomplex>             GoodPeriodIntegralsCoefficients;
    Matrix<bigcomplex>             ApproxPeriodMap_fromImPhi;
    Matrix<bigcomplex>             ApproxPeriodMap;
    Vector<bigcomplex>             L;
    Vector<int>                    L_defined;
    Matrix<bigcomplex>             Lambda;
    Matrix<bigcomplex>               LambdaPlus;
    Matrix<bigcomplex>               LambdaMinus;
    bigcomplex                       OmegaPlus;
    bigcomplex                       OmegaMinus;
    // these are the ellcurve_invariants...
    bigcomplex                     c4, c6, w1, w2, tau, ellj;

    // III. EMBEDDINGS
    bool in_Em, in_Phi, in_PhiPlus, in_PhiMinus, in_V, in_Vplus, in_Vminus, in_VD;
    Matrix<bigrational>            Em;
    Matrix<bigrational>            Phi, PhiPlus, PhiMinus;
    Matrix<bigrational>            V, V_Embedded;
    Matrix<bigrational>            Vplus, Vplus_Embedded;
    Matrix<bigrational>            Vminus, Vminus_Embedded;
    Matrix<bigrational>            VD;
    

    // IV. OPERATORS
    bool in_Conj_V, in_Conj_Phi,
	in_Conj_PhiSk, in_Tp_V, in_Tp_Vplus, in_Tp_Vminus, in_Wq, 
	in_PhiSk, in_PhiSkInverse, in_PhiSkPlus, in_PhiSkMinus,
	in_Tp_Phi, in_Tp_PhiPlus, in_Tp_PhiMinus;
    Matrix<bigrational>            Conj_V;
    Matrix<bigrational>            Conj_Phi;
    MatVecList<bigrational>        Tp_V;    
    MatVecList<bigrational>        Tp_Vplus;
    MatVecList<bigrational>        Tp_Vminus;        
    MatVecList<bigrational>        Tp_Phi, Tp_PhiPlus, Tp_PhiMinus;    
    Matrix<bigrational>            Tp_Phi_Dual_EVinv, Tp_PhiPlus_Dual_EVinv,
	                           Tp_PhiMinus_Dual_EVinv;
    Vector<int>                    Tp_Phi_Dual_e, Tp_PhiPlus_Dual_e,
	                           Tp_PhiMinus_Dual_e;
    Vector<int>                    Wq;
    char*                          WqString;
    
    Matrix<bigrational>            Conj_PhiSk;
    Matrix<bigrational>            PhiSk;
    Matrix<bigrational>            PhiSkInverse;
    Matrix<bigrational>            PhiSkPlus;
    Matrix<bigrational>            PhiSkMinus;

    // V. INTERSECTION 
    bool in_Intersection;
    IntVecList<bigint>             Intersection;

    // VI. Fast subspace of dual of + quotient   Hom(M_k^+,Q).
    //     This can be used for computing the Tp and Wq action,
    //     and q-expansions as quickly as can be expected. 
    // NOTE: (1) Fast is viewed as column vectors since it is only 
    // used in calls to Dual_Tp_On, and set_Fast requires 
    // this format.
    //       (2) Fast is a left EIGENVECTOR.
    bool                           in_Fast;
    Matrix<bigrational>            Fast, Fast_EVinv;
    Vector<int>                    Fast_e;
    int                            Fast_e_pos;
    Vector<bigrational>            Fast_Field;
    MatVecList<bigrational>        Fast_Tp;
} ;


// IT IS NEVER GUARANTEED SAFE TO HAVE MORE THAN ONE MOTIVE
// "CHECKED OUT" with get_Motive AT A TIME.
// (Except when using the intersection method above.)

class MotiveDecomp
{
public:
    MotiveDecomp(int N, int weight, bool plus=false);
    ~MotiveDecomp();
    
    // I. OUTPUT
    void                           write(OUTTYPE type, ostream& out=cout,
					 bool brief=false,int qPrec=17);

    // II. ACCESS
    int                            num_Motives();
    // It is oK to use a reference right after you get it, but be careful
    // if the reference has been around a while.  See the warning below
    // about "free_Motive".
    Motive&                        get_Motive(int i);
    Motive&                        get_FullMotive();
    H1NChiZ*                       get_H();
    H1NChiZ*                       get_H_Plus();
    int                            get_N();
    int                            get_weight();
    bool                           Is_Plus();

    // The following 3 functions, for memory saving purposes only, allow
    // the client to purge the indicated object(s) from memory.  
    // BUG WARNING: If outstanding references to such objects exists, then
    // a core dump is waiting to happen!
    // In future, the Motive class should somehow be wrapped by a reference
    // counting object.  Something like:
    // class MotiveWrap { Motive& m(); } ~MotiveWrap (-- calls free_Motive(this))
    // and then free_Motive(this) only decrements a reference count. 
    // The current danger place is Motive::Intersection as it deletes the 
    // Motive being intersected with.
    void                           free_Motive(int i);
    void                           free_AllMotives();
    void                           free_FullMotive();

private:			// functions
    void                           Save();
    bool                           Load();
    char*                          Name();
    void                           in_ins();
    void                           init_MotiveList();
    
private:			// data
    bool                           in_H, in_H_Plus,
	                           in_MotiveList, in_FullMotive, in_Name;
    int                            N;
    int                            weight;
    H1NChiZ*                       H;
    H1NChiZ*                       H_Plus;
    bool                           plus;
    Motive**                       MotiveList;
    Motive*                        FullMotive;
    int                            numMotives;
    char*                          strName;

} ;

class MotiveCalculator 
{
public:
    MotiveCalculator();
    ~MotiveCalculator();
    
    void Interface();
private:
    bool plus;

    void Welcome();
    void Help();
    void Help(const char* cmd);
    void ListCommands();
    void MotiveCommand(Motive& m,
		       const char* cmd, const char* arg);


    void Table_Arithmetic(int Nstart, int Nstop, int weight,
			     bool primeonly, ostream& out);
    void PARITable_SpecialValues(int Nstart, int Nstop, int weight,
				 bool primeonly, ostream& out);
    void PARITable_SpecialValuesOld(int Nstart, int Nstop, int weight,
				 bool primeonly, ostream& out);
    void PARITable_cp(int Pstart, int Pstop,
		      bool primeonly, ostream& out);
    void PARITable_eigenvalues(int Nstart, int Nstop, int weight, int prec, 
			      ostream& out);
    void PARITable_ssbasis(int Pstart, int Pstop, ostream& out);
    void PARITable_lchinz(int Pstart, int Pstop, ostream& out);
    void PARITable_splitting(int Nstart, int Nstop, int weight,
			     bool primeonly, ostream& out);
    void PARITable_nonmaximal(int Nstart, int Nstop, int weight,
			      bool primeonly, ostream& out);

    void SummarizeLevel(int N, int weight);
    void SummarizeMotive(Motive& m);
    void WriteLevel(int N, int weight, OUTTYPE type);
    // returns true on success.
    bool ParseCommand(int& true_N, int& true_IsoClass, int& weight,// true_N = 0 for "system" commands.
		      char*& cmd, char*& arg,         // must be deleted by caller.
		      char* input); 
    char* GetUserCommand(); // returned string must be deleted by caller.
		      
    // do not delete Motive* and do not use after calling get_Motive again.
    Motive* get_Motive(int true_N, int true_IsoClass, int weight); 
    MotiveDecomp& get_MotiveDecomp(int N, int weight);

    MotiveDecomp* Decomp;
    bool in_Decomp;
} ;


#endif

