/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/*
   FILE:   msymbols.cc
   DESC:   
*/

#include <LiDIA/nmbrthry_functions.h>
#include <stdio.h>
#include <iostream.h>
#include "msymbols.h"
#include "system.h"
#include "arith.h"

///////////////////// MSym class /////////////////
void MSym::save (ofstream& out) const 
{
    out << u() << endl << v() << endl << q() << endl ;
}

void MSym::load (ifstream& in) 
{
    bigint _u, _v; 
    int _q ;
    in >> _u >> _v >> _q ;
    u(_u); v(_v); q(_q);
}

ostream& operator<<(ostream& os, const MSym& x) 
{
    os << '[' ;
    if(x.q()==0) os<<"1";
    if(x.q()>0) 
      os << "X";
    if(x.q()>1)
      os << "^"<<x.q();
    os << ",(" << x.u() << ',' << x.v() << ")]";
    return os ;
}

// Input:  * a monomial P(X,Y) = X^q*Y^(k-2-q)
//         * a 2x2 matrix m=[ m[0], m[1]; m[2], m[3] ] = [a,b;c,d]
// Output: * the coefficients of the monomial
//           P(m(X,Y)) = P(a*X+b*Y, c*X+d*Y)
//                     = coeff[0]*X^0*Y^(k-2) + ... + coeff[k-2]*X^(k-2).
void act_on_monomial(Array<bigint>& coeff, 
		     const bigint& a, const bigint& b,
		     const bigint& c, const bigint& d,
		     const int& q, const int& k) 
{
    //  change to do _all_ arithmetic over F, which will be *much* faster!
    Array<bigint> aXbY(k-2+1), cXdY(k-2+1);
    if (a==0) {
	power (aXbY[0], b, q);
    } else if (b==0) {
	power (aXbY[q], a, q);
    } else {
	for (int i=0; i<=q; i++) {
	    bigint ai, bqi;
	    power(ai,a,i); power(bqi,b,q-i);
	    aXbY[i]=ai*bqi*binom(bigint(q),bigint(i));
	}
    }
    
    if (c==0) {
	power(cXdY[0],d,k-2-q);
    } else if (d==0) {
	power(cXdY[k-2-q],c,k-2-q);
    } else {
	for (int i=0; i<=k-2-q; i++) {
	    bigint ci, dk2qi;
	    power(ci,c,i); power(dk2qi,d,k-2-q-i);
	    cXdY[i]=ci*dk2qi*binom(bigint(k-2-q),bigint(i));
	}
    }
    
    for(int qq=0; qq<=k-2; qq++) 
    				// all ways to sum two naturals and obtain i.
	for (int j=intmax(0,qq + q-k-2); j<=intmin(qq, q); j++) {
	    coeff[qq] += aXbY[j]*cXdY[qq-j];
	}
}

void MSym::ModSymPrint(ostream& os, int N, int k) const
{
    bigint a,b,c,d;
    abcd(a,b,c,d,N);
    if (k>2) {
      Array<bigint> coeff; coeff.reset(k-2+1);
      act_on_monomial(coeff, d, -b, -c, a, qq, k);
      bool prev=false;

      os<<"(";
      for(int i=k-2; i>=0; i--) 
	if (coeff[i]!=0) {
	  if(coeff[i]>0)
	    { if(prev) os << " + "; }
	  else
	    os<<" - ";
	  coeff[i]=abs(coeff[i]);
	  if(coeff[i]!=1)
	    os<<coeff[i];
	  if(i>0) os << "X";
	  if(i>1) os << "^"<<i;
	  if(k-2-i>0) os << "Y";
	  if(k-2-i>1) os << "^"<<k-2-i;
	  prev=true;
	}
      os<<")";

    }
    os<<"{";
    if(d==0)
	os << "oo";
    else
	os << bigrational(b)/d;
    os<<",";
    if(c==0)
	os << "oo";
    else
	os << bigrational(a)/c;
    os<<"}";
}

void MSym::assert(bool assertion, const char* func_name,
		  const char* mesg) const 
{
    if (!assertion) {
	cerr << "MSym assertion failed in function " << func_name << endl
	     << "'" << mesg << "'" << endl ;
	exit(-1);
    }
}

void MSym::abcd (bigint& a, bigint& b, bigint& c, bigint& d,
		 const bigint& N) const
{
    c=uu; d=vv;
    if(c==0) c+=N; if(d==0) d+=N;
    bigint m=c, g ;
    while ((g=gcd(m,d)) != 1) { m/=g; }
    while ((g=gcd(N,m)) != 1) { m/=g; }
    d += N*m ;
    xgcd(b,a,-c,d);
//#ifdef DEBUG
//    assert(a*d-b*c == 1, "abcd", "ERROR in xgcd");
//#endif
}

void MSym::abcd (Matrix<bigint>& M,
		 const bigint& N) const
{
    M.init(2,2);
    M(2,1)=uu; M(2,2)=vv;
    if(M(2,1)==0) M(2,1)+=N; if(M(2,2)==0) M(2,2)+=N;
    bigint m=M(2,1), g ;
    while ((g=gcd(m,M(2,2))) != 1) { m/=g; }
    while ((g=gcd(N,m)) != 1) { m/=g; }
    M(2,2) += N*m ;
    xgcd(M(1,2),M(1,1),-M(2,1),M(2,2));
}


///////////////////// MSymbols class /////////////////

void MSymbols::filename (char* fname) const
{
    sprintf(fname, "%s_MSymbols_%d", chi.name(), k);
}


void MSymbols::assert(bool assertion, const char* func_name,
		      const char* mesg) const 
{
    if (!assertion) {
	cerr << "MSymbols assertion failed in function " << func_name << endl
	     << "'" << mesg << "'" << endl ;
	exit(-1);
    }
}

void MSymbols::save() const
{
    if (!do_save) return ;
    char fname[512] ;
    filename (fname) ;
    ofstream out(data_write(fname,N));
    assert(out, "save", "Error saving MSymbols object.");
    out << size << endl ;
    for (int i = 0 ; i < size ; i++) 
	list[i].save (out) ;
    cerr << endl;
    for (int i = 0 ; i < N ; i++)
	out << GCD[i] << ' ' << INV[i] << ' ' ;
}

bool MSymbols::load()
{
    if (!do_save) return false ;
    char fname[512] ;
    filename(fname) ;
    ifstream in(data_read(fname,N)) ;
    if (!in) return false ; 
    in >> size ;
    mesg("Loading Manin symbols list (size=",size);
    mesg(", name = "); mesg(fname); mesg("):\n\t");
    list = new MSym[size];
    for (int i = 0 ; i < size ; i++) {
	if(i%((size/15)+1)==0) {mesg(i*100/size);mesg("% ");}
	list[i].load (in) ;
    }
    mesg("\n");
    GCD = new bigint[N];
    INV = new bigint[N]; 
    for (int i=0; i < N; i++)
	in >> GCD[i] >> INV[i] ; 
    return true ;
}

void MSymbols::MSym_to_list(int& n, int& lambda,
			    const MSym& x) const
{
#ifdef DEBUG
    assert(gcdN(gcd(x.u(), x.v())) == 1, "MSym_to_list", 
	   "(u,v) part of MSymbol is invalid");
    assert(0 <= x.q() && x.q() <= k-2, "MSym_to_list",
	   "q part of MSymbol is invalid") ;
#endif
    
    
    bigint u=x.u()%N, v=x.v()%N;
    int q=x.q();
    if (u<0) u+=N;  if (v<0) v+=N;
    if (chi.type() == -1) {
	lambda=1;
	if (u>N/2) { u=N-u; v=N-v; if (k%2) lambda=-1;}
	else if (u==0 && v>N/2) {
	    v=N-v;
	    if (k%2)
		lambda=-1;
	}
	else if (N%2==0 && u==N/2 && v>N/2) {
	    v = N-v;
	    if (k%2)
		lambda=-1;
	}
    } else {
	if (u==0) {
	    v.intify(lambda);
	    v=1;
	} else {
	    bigint c=inv(u);
	    u*=c; u%=N;
	    v*=c; v%=N;
	    inv(c).intify(lambda);
	    
#ifdef DEBUG
	    assert (N%u==0, "MSym_to_list", "u doesn't divide N");
#endif
	    
	    bigint uu,vv; int lam;
	    if (!first_on_line(uu,vv,lam,u,v)){
		u=uu; v=vv;
		bigint L=lambda; 
		L*=lam; L%=N; if(L<0)L+=N; 
#ifdef DEBUG
		assert(is_int(L),"MSym_to_list", "VVERFLOW!");
#endif
		L.intify(lambda);
	    }
#ifdef DEBUG
	    // REMOVE after a while -- very wasteful effiency wise. 
	    assert(first_on_line(uu,vv,lam,u,v),
		   "MSym_to_list", "first_on_line failed");
#endif
	}
    }
    u%=N; v%=N ;
    if (u<0)
	u+=N;
    if (v<0)
	v+=N;
    
#ifdef DEBUG
    if (chi.type() != -1) {
	if ((v*lambda-x.v())%N!=0)
	    cerr << "("<<u<<","<<v<<") "<<lambda<<",    "<<x<<endl;
	assert((u*lambda-x.u()) % N == 0,"MSym_to_list","u result is invalid");
	assert((v*lambda-x.v()) % N == 0,"MSym_to_list","v result is invalid");
	assert(gcdN(lambda)==1,"MSym_to_list", "lambda is not comprime to N");
    }
#endif
    
    n=find(MSym(u,v,q))+1;
}


void MSymbols::abcd (bigint& a, bigint& b, bigint& c, bigint& d,
		     const int& n) const
{
#ifdef DEBUG
    assert(1<=n && n<=size, "abcd", "invalid n");
#endif
    MSym m(list[n-1].u(), d=list[n-1].v());
    m.abcd(a,b,c,d,N);		// let the MSym object do the work.
}

int MSymbols::find (const MSym& x) const
{
    // binary search
    int left = 0, right, m ;
    
    right = size - 1 ;
    while (right >= left) {
	m = (right + left) / 2 ;
	if (x < list[m])
	    right = m - 1 ;
	else
	    left = m + 1 ;
	
	if (x == list[m])
	    return m ;
    }
    
#ifdef DEBUG
    cerr << x << endl;
    assert (false, "find", "MSymbol not found.");
#endif
}

// I think this algorithm could be greatly optimized
// with a little more careful thought. 
bool MSymbols::first_on_line (bigint& uu, bigint& vv, int& lambda,
			      const bigint& u, const bigint& v) const
{
    
#ifdef DEBUG
    assert(u!=0&&N%u==0,"first_on_line","assertion 1");
#endif
    bool is_first=true;
    uu=u; vv=v; lambda=1;
    
    for(int alpha=0; alpha<uu; alpha++) {
	bigint L=(1+alpha*N/uu) % N ; 
	if (gcd(L,N)==1) {
	    bigint Lv = L; Lv*=v; Lv%=N; // prevent overflow.
	    if (Lv < vv) {
		vv=Lv;
		lambda=to_int(inv(L)%N);
		is_first=false; 
	    }
	}
    }
#ifdef DEBUG
    assert((lambda*uu-u)%N==0 && (lambda*vv-v)%N==0, "first_on_line", "invalid");
#endif
    return is_first;
}

void MSymbols::create_list () 
{
    mesg("MSymbols::create_list... ");
    int n = 0 ;
    if (chi.type() == -1) {
	bigint a=k-1;		// for the weight part
	// #{(u,v) : 0<=u,v<N, gcd(u,v,N)=1} = prod_{p|N} (p^{2n}-p^{2n-2})
	// proof: use the orbit stabilizer theorem applied to P^1(N) and the
	// fact that #P^1(N)=prod_{p|N} (p^n+p^{n-1}).    
	
	// Cremona, Lemma 3.1:  Coset representatives for Gamma_1(N)
	// in PSL_2 are {(u,v) : 0<=u,v<N, gcd(u,v,N)=1} / ~
	// where (u,v) ~ (-u,-v).  
	// Thus we are overcounting by a factor of 2, except for
	// points fixed by (u,v)|-->(-u,-v), i.e., u=v=0.
	if (N==2) {
	    a*=3;			// There are 3 M-symbols at level 2,
				// namely (0,1), (1,0), and (1,1).
				// Furthermore, 1=-1 so we *don't* mod
				// out by the action of -1. 
	} else
	    if (N>2) {
		rational_factorization f(N);
		f.factor();
		for (int i=0; i<f.no_of_comp(); i++) {
		    bigint p2n;
		    power(p2n,f.base(i),2*f.exponent(i));
		    a *= p2n - p2n/(f.base(i)*f.base(i));
		}
		a/=2;   // working in PSL_2 so divide by 2. 
	    }
	
	assert (is_int(a),"create_list", "N is too big!");
	a.intify(size);
	list = new MSym[size];
	for (bigint v=1; v<=N/2;v++)
	    if (gcdN(v)==1)
		for (int q=0; q<=k-2; q++)
		    list[n++] = MSym(0,v,q);
	
	for (bigint u=1; u<N/2 || (N%2 && u<=N/2); u++)
	    for (bigint v=0; v<N; v++)
		if (gcdN(gcd(u,v))==1)
		    for (int q=0; q<=k-2; q++)
			list[n++] = MSym(u,v,q);
	
	if (N%2 == 0)
	    for (bigint v=0; v<=N/2; v++)
		if (gcdN(gcd(N/2,v))==1)
		    for (int q=0; q<=k-2; q++)
			list[n++] = MSym(N/2,v,q);
	
	if (N==1) 
	    for (int q=0; q<=k-2; q++)
		list[n++] = MSym(0,0,q);
	
	assert (n==size,"create_list", "error in programming!");
	
    } else {
	//   make a sorted list of:
	//     * <q,(0,1)>, <q,(1,1)>, <q,(1,2)>, ..., <q,(1,N)>.
	//     * All <q,(u,v)>, u|N, gcd(u,v)=1
	//     where 0<=q<=k-2.
	// compute #(P^1(N))
	bigint a = k-1;		// for the weight part
	rational_factorization f(N);
	f.factor();
	for (int i=0; i<f.no_of_comp(); i++) {
	    bigint pn;
	    power(pn,f.base(i),f.exponent(i));
	    a *= pn + pn/f.base(i);
	}
	assert (is_int(a),"create_list", "N is too big!");
	a.intify(size);
	list = new MSym[size];
	
	for (int q=0; q<=k-2; q++) {
	    list[n++] = MSym(0,1,q);
	}
	for (bigint v=0; v < N; v++)
	    for (int q=0; q<=k-2; q++) 
		list[n++] = MSym(1,v,q);
	
	sort_vector<bigint> divs = divisors(N);
	if (N==1) divs.set_size(1);	// because of a bug in LiDIA.
	for (int i=0; i<divs.size(); i++) {
	    bigint u = divs[i]; 
	    if (u>=2 && u<N)
		for (bigint v=1; v < N; v++)
		    if (gcd(u,v)==1) {
			bigint dum1, dum2; int dum3;
			if (first_on_line (dum1, dum2, dum3, u, v))
			    for (int q=0; q<=k-2; q++) {
				list[n] = MSym(u,v,q);
				n++;
			    }
		    }
	    mesg(100*i/divs.size());
	    mesg("% ");
	}
        mesg("100%");
    }
    
#ifdef DEBUG
    cerr << endl ;
    assert (n == size, "create_list", "n != size");
    for (int i=0; i<size-1; i++)
	assert(list[i]<list[i+1],"create_list", "list is not sorted");
#endif
    
    mesg("\n");
}

void MSymbols::create_modN_tables () 
{
    // initialize the table of inverses mod N and gcd's with N:
    INV = new bigint[N] ;
    GCD = new bigint[N] ;
    assert(INV!=0,"create_modN_tables","error allocating memory");
    assert(GCD!=0,"create_modN_tables","error allocating memory");
    GCD[0] = N ;
    mesg("Creating modN tables... ");
    for(int u = 1 ; u < N ; u++) {
	GCD [u] = gcd(u,N) ;
	if(gcdN(u) == 1) {
	    bigint Uinv, dummy;	// avoid overflow
	    xgcd (Uinv, dummy, u, bigint(N));
	    Uinv %= N ; if (Uinv <0) Uinv+=N;
	    INV[u] = Uinv ;
	} else {
	    // find some x such that gcdN(x) == 1 and (u*x % N) | N.
	    // Do this by finding an x such that
	    //     ux + Ny = gcd(u,N) = g, say
	    // and x is coprime to N.  
	    //     ux = g (mod N).
	    //    u/g * x' = 1 (mod N/g).
	    //    x = x' + k*N/g.   
	    // say p|N and p|x.  if p|(N/g) then p|x', a contradiction 
	    // since x' is coprime to N/g.  thus N/g is nonzero mod p
	    // so there is a solution to x' =/= k_p N/g (mod p). 
	    // using CRT lift to a solution k s.t. x=x'+k*N/g can't be
	    // divisible by any p s.t p|N and p|x, i.e., gcd(N,x)=1.
	    bigint g = gcd (u,N) ;
	    bigint xprime, y;
	    xgcd(INV[u],y,u/g,N/g);
	    INV[u] %= N;
	    while (gcd (INV[u], N) != 1) {
		INV[u] += N/g ; INV[u] %= N ; 
	    }
	    if (INV[u] < 0) INV[u] += N ;
#ifdef DEBUG
	    bigint temp=(INV[u]*u)%N;
	    assert (gcd (INV[u], N) == 1 &&  N%temp == 0,
		    "create_modN_tables", "Error in computing table.");
#endif
	}
    }
    mesg("\n");
}

MSymbols::MSymbols (const Character& _chi, int _k,
		    bool _save) 
    : chi(_chi), k(_k), do_save(_save)
{
    N = chi.modulus () ;
    if (!load ()) {
	create_modN_tables () ;
	create_list();
	save();
    }
}

MSymbols& 
MSymbols::operator=(const MSymbols& msymbols) 
{
    N=msymbols.N;
    k=msymbols.k;
    chi=msymbols.chi;
    do_save=msymbols.do_save;
    delete[] INV;
    delete[] GCD;
    delete[] list;
    GCD = new bigint[N];
    INV = new bigint[N]; 
    for(int i=0;i<N;i++) {
	GCD[i]=msymbols.GCD[i];
	INV[i]=msymbols.INV[i];
    }
    size=msymbols.size;
    list = new MSym[size];
    for(int i=0;i<size;i++) 
       list[i]=msymbols.list[i];
    return *this;
}


MSymbols::~MSymbols() 
{
    delete[] INV ;
    delete[] GCD ;
    delete[] list ;
}



