/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/*
   FILE:   msymbols.h
   DESC:   
*/

#ifndef MSYMBOLS_H
#define MSYMBOLS_H

#include <LiDIA/bigint.h>
#include "array.h"
#include "matrix.h"
#include "character.h"



class MSym {
  public:
    MSym (bigint _u=0, bigint _v=0, int _q=0) : uu(_u), vv(_v), qq(_q) {} 
    bigint u() const { return uu; } 
    void u(const bigint& _u) { uu=_u; }
    bigint v() const { return vv; } 
    void v(const bigint& _v) { vv=_v; }
    int q() const { return qq; } 
    void q(const int& _q) { qq=_q; }
    bool operator<(const MSym& x) const ;
    bool operator == (const MSym& x) const ;
    void save (ofstream& out) const ;
    void load (ifstream& in) ;
    void abcd (bigint& a, bigint& b,
	       bigint& c, bigint& d,
	       const bigint& N) const;
    void abcd (Matrix<bigint>& M,
	       const bigint& N) const;
    friend ostream& operator<<(ostream& os, const MSym& x) ;
    void ModSymPrint(ostream& os, int N, int k) const;
  private:
    bigint uu, vv;
    int qq ;
    void assert(bool assertion, const char* func_name="?",
		const char* mesg="assertion failed") const;
} ;




class MSymbols 
{ 
  public:
    MSymbols (const Character& chi, int k=2,
	      bool save=false) ;
    MSymbols (const MSymbols& msymbols) { 
	die("MSymbols (const MSymbols& msymbols) not yet implimented.");
    }
    MSymbols& operator=(const MSymbols& msymbols) ;
    ~MSymbols() ;
    
    // Given a Manin symbol x find the Manin symbol y in the standard 
    // list such that
    //        x = chi(lambda)y
    // n is the position of y in standard list. 
    // NOTE, in the standard list, it is guaranteed (because of the ordering)
    // that it goes <0,(u,v)>, <1,(u,v)>, ..., <k-2,(u,v)>.
    void MSym_to_list(int& n, int& lambda,
		      const MSym& x) const ;
    
    // Find a 2x2 matrix (a,b;c,d) in SL2(Z) s.t. c=u, d=v(mod N)
    // where <q,(u,v)> is the n-th Manin symbol in the standard list.
    void abcd (bigint& a, bigint& b, bigint& c, bigint& d,
	       const int& n) const ;
    void abcd (Matrix<bigint>& M,
	       const int& n) const ;
    
    // return the n-th Manin symbol in the standard list.
    MSym x(const int& n) const {
#ifdef DEBUG
	assert (1<=n && n<=size,"x","n out of range");
#endif
	return list[n-1] ; } 
    
    int length () const { return size ; }    
    
  private:
    void assert(bool assertion, const char* func_name="?",
		const char* mesg="assertion failed") const;
    int k, N ;			// weight, level
    Character chi ;		// character
    bool do_save ;
    
    // a table of inverses mod N. 

    bigint* INV;
    bigint inv(const bigint& c) const { 
	int ic=to_int(c%N);
	return INV [ic + ((ic<0)?N:0)] ; 
    }
    
    // a table of gcd's with N.
    bigint* GCD;
    bigint gcdN(const bigint& c) const { 
	int ic=to_int(c%N);
	return GCD [ic + ((ic<0)?N:0)] ; 
    }
    
    void create_modN_tables();
    
    //////////////// "STANDARD LIST" ///////////////////
    
    MSym* list ;			// standard list of M-symbols
    int   size ;			// length of standard list
    void create_list();		
    bool first_on_line (bigint& uu, bigint& vv, int& lambda,
			const bigint& u, const bigint& v) const ;
    
				// places filename under which to store this
				// object in memory pointed to by "fname".
    void filename (char* fname) const ; 
    void save() const ;		
    bool load() ;			
    
    
    // Find x in the standard list.  In DEBUG mode, if x is not
    // in the standard list, an error occurs, and the program then stops. 
    int find (const MSym& x) const ;
} ;


//////////// Inlined Functions ////////////////

inline bool MSym::operator< (const MSym& x) const
{
    if (u()<x.u()) 
	return true ;
    if (u()==x.u() && v()<x.v()) 
	return true ;
    if (u()==x.u() && v()==x.v() && q()<x.q()) 
	return true ;
    return false ;
}

inline bool MSym::operator == (const MSym& x) const
{
    return (u()==x.u() && v()==x.v() && q()==x.q()) ;
}

#endif 


