/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/*
   FILE:   my_ntl.cc
   DESC:   interface with NTL
*/

#include "my_ntl.h"
#include "system.h"
#include "NTL/pair_lzz_pX_long.h"
#include "NTL/lzz_pXFactoring.h"

const int BUFLEN=4096;		        // too big? too small?
static char buf[BUFLEN];		// used for conversions.


// initialize ntl's F_q to be the current charmod, the current
// intmod, or the current ffield. 
void ntl_init_to_charmod()
{
    mesgln("ntl_init_to_charmod");
    zz_p::init(charmod::get_pr());
}

void ntl_init_to_intmod()
{
    mesgln("ntl_init_to_intmod");
    zz_p::init(intmod::get_pr());
}

void ntl_init_to_ffield()
{
    mesgln("ntl_init_to_ffield");
    const Fp_polynomial f=ffield_element::irred_polynomial();
    ZZ p;
    my_to_ntl(p, f.modulus());
    ZZ_p::init(p);
    ZZ_pX gen;			// gen = 0.
    for(int i=0; i<=f.degree(); i++) {
	ZZ_p c;
	my_to_ntl(c,f[i]);
	gen += ZZ_pX(i,c);
    }
    ZZ_pE::init(gen);
}


///////////// CONVERSIONS ///////////////////////////
// Conversions between my types and NTL types:
//     (scalars, polynomials, vectors, matrices, etc.)
// scalars
void ntl_to_my(charmod& x,                  const zz_p& y)
{ x=rep(y); }

void ntl_to_my(intmod&  x,                  const zz_p& y)
{ x=rep(y); }

void ntl_to_my(ffield_element& x,           const ZZ_pE& y)
{die("my_ntl function not implimented");}

void ntl_to_my(bigrational& x,              const ZZ& y)
{
    if(y==0) { x=0; return; }
    ostrstream ost(buf,BUFLEN);
    ost<<y<<'\0';
    bigint n;
    string_to_bigint(buf, n);
    x=n;
}

void my_to_ntl(zz_p& x,                     const charmod& y)
{
    x=lift(y);
}

void my_to_ntl(zz_p& x,                     const intmod&  y)
{
    x=lift(y);
}

void my_to_ntl(ZZ_pE& x,                    const ffield_element& y)
{die("my_ntl function not implimented");}


void my_to_ntl(ZZ& x,                        const bigint& y)
{
    if(y.bit_length()/3 + 10 >= BUFLEN) {  // not enough space
	char bigbuf[y.bit_length()/3+11];
	bigint_to_string(y,bigbuf);
	x=ZZ(INIT_VAL, bigbuf);
    } else {
	bigint_to_string (y, buf);
	x=ZZ(INIT_VAL, buf);
    }
}

void my_to_ntl(ZZ_p& x,                      const bigint& y)
{die("my_ntl function not implimented");}


// polynomials
void ntl_to_my(Vector<charmod>& f,          const zz_pX& g)
{
    f.init(deg(g)+1); 
    for(int i=0;i<=deg(g);i++) 
	ntl_to_my(f[i+1], coeff(g,i));
}

void ntl_to_my(Vector<intmod>&  f,          const zz_pX& g)
{
    f.init(deg(g)+1); 
    for(int i=0;i<=deg(g);i++) 
	ntl_to_my(f[i+1], coeff(g,i));
}

void ntl_to_my(Vector<ffield_element>&  f,  const ZZ_pEX& g)
{die("my_ntl function not implimented");}

void ntl_to_my(Vector<bigrational>&  f,     const ZZ_pEX& g)
{die("my_ntl function not implimented");}

void my_to_ntl(zz_pX& f,                    const Vector<charmod>& g)
{ 
    f=0;
    for(int i=0; i<=g.poldeg(); i++) {
	zz_p a;
	my_to_ntl(a,g.polgetcoeff(i));
	f += zz_pX(i, a);
    }
}

void my_to_ntl(zz_pX& f,                    const Vector<intmod>& g)
{
    f=0;
    for(int i=0; i<=g.poldeg(); i++) {
	zz_p a;
	my_to_ntl(a,g.polgetcoeff(i));
	f += zz_pX(i, a);
    }
}

void my_to_ntl(ZZ_pEX& f,                   const Vector<ffield_element>& g)
{die("my_ntl function not implimented");}

void my_to_ntl(ZZX& f, bigint& denom,       const Vector<bigrational>& g)
{die("my_ntl function not implimented");}



// vectors
void ntl_to_my(Vector<charmod>& v,          const vec_zz_p&  w)
{die("my_ntl function not implimented");}

void ntl_to_my(Vector<intmod>&  v,          const vec_zz_p&  w)
{die("my_ntl function not implimented");}

void ntl_to_my(Vector<ffield_element>&  v,  const vec_ZZ_pE& w)
{die("my_ntl function not implimented");}

void ntl_to_my(Vector<bigrational>&  v,     const vec_ZZ&    w)
{die("my_ntl function not implimented");}

void my_to_ntl(vec_zz_p& v,                 const Vector<charmod>& w)
{die("my_ntl function not implimented");}

void my_to_ntl(vec_zz_p& v,                 const Vector<intmod>&  w)
{die("my_ntl function not implimented");}

void my_to_ntl(vec_ZZ_pE& v,                const Vector<ffield_element>& w)
{die("my_ntl function not implimented");}

void my_to_ntl(vec_ZZ& v,   bigint& denom,  const Vector<bigrational>&    w)
{die("my_ntl function not implimented");}



// matrices
void ntl_to_my(Matrix<charmod>& v,          const mat_zz_p&  w)
{die("my_ntl function not implimented");}

void ntl_to_my(Matrix<intmod>&  v,          const mat_zz_p&  w)
{die("my_ntl function not implimented");}

void ntl_to_my(Matrix<ffield_element>&  v,  const mat_ZZ_pE& w)
{die("my_ntl function not implimented");}

void ntl_to_my(Matrix<bigrational>&  v,     const mat_ZZ&    w)
{
    v.init(w.NumRows(),w.NumCols());
    for(int r=1;r<=v.nrows();r++)
	for(int c=1;c<=v.ncols();c++)
	    ntl_to_my(v(r,c),w(r,c));
}

void my_to_ntl(mat_zz_p& A,                 const Matrix<charmod>& B)
{
    A.SetDims(B.nrows(),B.ncols());
    for(int r=1;r<=B.nrows();r++)
	for(int c=1;c<=B.ncols();c++)
	    my_to_ntl(A(r,c),B(r,c));
}

void my_to_ntl(mat_zz_p& A,                 const Matrix<intmod>&  B)
{
    A.SetDims(B.nrows(),B.ncols());
    for(int r=1;r<=B.nrows();r++)
	for(int c=1;c<=B.ncols();c++)
	    my_to_ntl(A(r,c),B(r,c));
}

void my_to_ntl(mat_ZZ_pE& v,                const Matrix<ffield_element>& w)
{die("my_ntl function not implimented");}

void my_to_ntl(mat_ZZ& v,    bigint& d, const Matrix<bigrational>& M)
{
    // find lcm of denominators in M.
    d = 1;
    for (int r=1 ; r<=M.nrows() ; r++)
	for (int c=1 ; c<=M.ncols() ; c++) 
	    if (M(r,c) != 0)
		d = lcm (d, denominator(M(r,c)) ) ;
    d = abs(d);			// necessary???
    v.SetDims(M.nrows(), M.ncols());
    for (int r=1 ; r<=M.nrows() ; r++)
	for (int c=1 ; c<=M.ncols() ; c++) 
	    my_to_ntl(v(r,c),(M(r,c)*d).numerator());
}



///////////////// FUNCTIONS  /////////////////////

// characteristic polynomials
void ntl_charpoly(Vector<charmod>& cpoly,      const Matrix<charmod>& M)
{
//    ntl_init_to_charmod();
    mat_zz_p A;
    my_to_ntl(A, M);
    zz_pX f;
    CharPoly(f, A);		// NTL charpoly.
    ntl_to_my(cpoly, f);
}

void ntl_charpoly(Vector<intmod>&  cpoly,          const Matrix<intmod>& M)
{
//    ntl_init_to_intmod();
    mat_zz_p A;
    my_to_ntl(A, M);
    zz_pX f;
    CharPoly(f, A);		// NTL charpoly.
    ntl_to_my(cpoly, f);
}

void ntl_charpoly(Vector<ffield_element>&  f,  const Matrix<ffield_element>& A)
{die("my_ntl function not implimented");}

void ntl_charpoly(Vector<bigrational>& f,      const Matrix<bigrational>& A)
{die("my_ntl function not implimented");}


// factorization
void ntl_factor(VecFac<charmod>& fac,          const Vector<charmod>& f)
{
    mesgln("ntl_factor");
//    ntl_init_to_charmod();
    zz_pX g;
    my_to_ntl(g, f);
    vec_pair_zz_pX_long ntlfac;
    CanZass(ntlfac, g, 1);
    fac.resize(ntlfac.length());
    for(int i=1; i<=ntlfac.length(); i++) {
	ntl_to_my(fac.base(i), ntlfac(i).a);
	fac.exp(i) = int(ntlfac(i).b);
    }
}

void ntl_factor(VecFac<intmod>&  fac,          const Vector<intmod>& f)
{
//    ntl_init_to_intmod();
    mesgln("ntl_factor");
    zz_pX g;
    my_to_ntl(g, f);
    vec_pair_zz_pX_long ntlfac;
    CanZass(ntlfac, g, 1);
    fac.resize(ntlfac.length());
    for(int i=1; i<=ntlfac.length(); i++) {
	ntl_to_my(fac.base(i), ntlfac(i).a);
	fac.exp(i) = int(ntlfac(i).b);
    }
}

void ntl_factor(VecFac<ffield_element>&  fac,  const Vector<ffield_element>& f)
{die("my_ntl function not implimented");}

void ntl_factor(VecFac<bigrational>& fac,      const Vector<bigrational>& f)
{die("my_ntl function not implimented");}


// polynomial substitution
void ntl_subst(Matrix<charmod>& fM,            const Matrix<charmod>& M,
	                                       const Vector<charmod>& f)
{die("my_ntl function not implimented");}

void ntl_subst(Matrix<intmod>& fM,             const Matrix<intmod>& M,
	                                       const Vector<intmod>& f)
{die("my_ntl function not implimented");}

void ntl_subst(Matrix<ffield_element>& fM,     const Matrix<ffield_element>& M,
	                                       const Vector<ffield_element>& f)
{die("my_ntl function not implimented");}

void ntl_subst(Matrix<bigrational>& fM,        const Matrix<bigrational>& M,
	                                       const Vector<bigrational>& f)
{die("my_ntl function not implimented");}

// kernels
void ntl_kernel(Matrix<charmod>& KerM,         const Matrix<charmod>& M)
{die("my_ntl function not implimented");}

void ntl_kernel(Matrix<intmod>& KerM,          const Matrix<intmod>& M)
{die("my_ntl function not implimented");}

void ntl_kernel(Matrix<ffield_element>& KerM,  const Matrix<ffield_element>& M)
{die("my_ntl function not implimented");}

void ntl_kernel(Matrix<bigrational>& KerM,     const Matrix<bigrational>& M)
{die("my_ntl function not implimented");}


// images
void ntl_image(Matrix<charmod>& KerM,          const Matrix<charmod>& M)
{die("my_ntl function not implimented");}

void ntl_image(Matrix<intmod>& KerM,           const Matrix<intmod>& M)
{die("my_ntl function not implimented");}

void ntl_image(Matrix<ffield_element>& KerM,   const Matrix<ffield_element>& M)
{die("my_ntl function not implimented");}

void ntl_image(Matrix<bigrational>& KerM,      const Matrix<bigrational>& M)
{die("my_ntl function not implimented");}

// LLL: find good basis for the rows of M -- a good first step
// before computing a kernel.
/*
void my_LLL_FP(mat_ZZ& A, mat_ZZ& U) ;

void ntl_LLL(Matrix<bigrational>& L,           const Matrix<bigrational>& M)
{die("my_ntl function not implimented");}

void ntl_cheapLLL(Matrix<bigrational>& L,      const Matrix<bigrational>& M)
{
    mesgln("ntl_cheapLLL");
    mat_ZZ A, U;
    bigint d;
    my_to_ntl(A, d, M);		// M = d*A
    mesgln("ntl_cheapLLL: calling NTL's approximate LLL reduction");
    my_LLL_FP(A, U);
    mesgln("The matrix U has determinant ");
    ZZ det=determinant(U);
    cerr << det<<endl;
    if(det!=1 && det!=-1) {
	mesgln("The determinant isn't 1, so using a slower version of LLL...");
	my_to_ntl(A, d, M);		// M = d*A
	die("not written yet.");
    }
    ntl_to_my(L, A);
    L /= d;

///    mesgln("Now running a check which should be completely redundant.");
//    Matrix<bigrational> cU;
//  ntl_to_my(cU,U);
//    if(cU*M != L) {
//	die("OH NO!  contradiction!  programming error!");
//    } else {
//	mesgln("Test passed.");
//    }
}
*/
