/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/*
   FILE:   my_ntl.h
   DESC:   interface with NTL
*/

#ifndef MY_NTL
#define MY_NTL

#include "NTL/lzz_p.h"
#include "NTL/mat_poly_lzz_p.h"
#include "NTL/lzz_pEX.h"
#include "NTL/ZZX.h"
#include "NTL/ZZ_pX.h"
#include "NTL/ZZ_pEX.h"
#include "NTL/mat_lzz_p.h"
#include "NTL/mat_ZZ_pE.h"
#include "NTL/mat_ZZ.h"
#include "matrix.h"
#include "vector.h"
#include "fields.h"
#include <LiDIA/bigrational.h>

class ffield_element;

// initialize ntl's F_q to be the current charmod, the current
// intmod, or the current ffield. 
void ntl_init_to_charmod();
void ntl_init_to_intmod();
void ntl_init_to_ffield();

///////////// CONVERSIONS ///////////////////////////
// Conversions between my types and NTL types:
//     (scalars, polynomials, vectors, matrices, etc.)
// scalars
void ntl_to_my(charmod& x,                  const zz_p& y);
void ntl_to_my(charmod& x,                  const ZZ& y);
void ntl_to_my(intmod&  x,                  const zz_p& y);
void ntl_to_my(intmod& x,                   const ZZ& y);
void ntl_to_my(ffield_element& x,           const ZZ_pE& y);
void ntl_to_my(bigrational& x,              const ZZ& y);
void my_to_ntl(zz_p& x,                     const charmod& y);
void my_to_ntl(zz_p& x,                     const intmod&  y);
void my_to_ntl(ZZ_pE& x,                    const ffield_element& y);
void my_to_ntl(ZZ& x,                       const bigint& y);
void my_to_ntl(ZZ_p& x,                     const bigint& y);


// polynomials
void ntl_to_my(Vector<charmod>& f,          const zz_pX& g);
void ntl_to_my(Vector<intmod>&  f,          const zz_pX& g);
void ntl_to_my(Vector<ffield_element>&  f,  const ZZ_pEX& g);
void ntl_to_my(Vector<bigrational>&  f,     const ZZ_pEX& g);
void my_to_ntl(zz_pX& f,                    const Vector<charmod>& g);
void my_to_ntl(zz_pX& f,                    const Vector<intmod>& g);
void my_to_ntl(ZZ_pEX& f,                   const Vector<ffield_element>& g);
void my_to_ntl(ZZX& f,                      const Vector<bigrational>& g);


// vectors
void ntl_to_my(Vector<charmod>& v,          const vec_zz_p&  w);
void ntl_to_my(Vector<intmod>&  v,          const vec_zz_p&  w);
void ntl_to_my(Vector<ffield_element>&  v,  const vec_ZZ_pE& w);
void ntl_to_my(Vector<bigrational>&  v,     const vec_ZZ&    w);
void my_to_ntl(vec_zz_p& v,                 const Vector<charmod>& w);
void my_to_ntl(vec_zz_p& v,                 const Vector<intmod>&  w);
void my_to_ntl(vec_ZZ_pE& v,                const Vector<ffield_element>& w);
void my_to_ntl(vec_ZZ& v,   bigint& denom,  const Vector<bigrational>&    w);


// matrices
void ntl_to_my(Matrix<charmod>& v,          const mat_zz_p&  w);
void ntl_to_my(Matrix<intmod>&  v,          const mat_zz_p&  w);
void ntl_to_my(Matrix<ffield_element>&  v,  const mat_ZZ_pE& w);
void ntl_to_my(Matrix<bigrational>&  v,     const mat_ZZ&    w);
void my_to_ntl(mat_zz_p& v,                 const Matrix<charmod>& w);
void my_to_ntl(mat_zz_p& v,                 const Matrix<intmod>&  w);
void my_to_ntl(mat_ZZ_pE& v,                const Matrix<ffield_element>& w);
void my_to_ntl(mat_ZZ& v,  bigint& denom,   const Matrix<bigrational>&    w);


///////////////// FUNCTIONS  /////////////////////

// characteristic polynomials
void ntl_charpoly(Vector<charmod>& f,          const Matrix<charmod>& A);
void ntl_charpoly(Vector<intmod>&  f,          const Matrix<intmod>& A);
void ntl_charpoly(Vector<ffield_element>&  f,  const Matrix<ffield_element>& A);
void ntl_charpoly(Vector<bigrational>& f,      const Matrix<bigrational>& A);

// factorization
void ntl_factor(VecFac<charmod>& fac,          const Vector<charmod>& f);
void ntl_factor(VecFac<intmod>&  fac,          const Vector<intmod>& f);
void ntl_factor(VecFac<ffield_element>&  fac,  const Vector<ffield_element>& f);
void ntl_factor(VecFac<bigrational>& fac,      const Vector<bigrational>& f);

// polynomial substitution
void ntl_subst(Matrix<charmod>& fM,            const Matrix<charmod>& M,
	                                       const Vector<charmod>& f);
void ntl_subst(Matrix<intmod>& fM,             const Matrix<intmod>& M,
	                                       const Vector<intmod>& f);
void ntl_subst(Matrix<ffield_element>& fM,     const Matrix<ffield_element>& M,
	                                       const Vector<ffield_element>& f);
void ntl_subst(Matrix<bigrational>& fM,        const Matrix<bigrational>& M,
	                                       const Vector<bigrational>& f);
// kernels
void ntl_kernel(Matrix<charmod>& KerM,         const Matrix<charmod>& M);
void ntl_kernel(Matrix<intmod>& KerM,          const Matrix<intmod>& M);
void ntl_kernel(Matrix<ffield_element>& KerM,  const Matrix<ffield_element>& M);
void ntl_kernel(Matrix<bigrational>& KerM,     const Matrix<bigrational>& M);

// images
void ntl_image(Matrix<charmod>& KerM,          const Matrix<charmod>& M);
void ntl_image(Matrix<intmod>& KerM,           const Matrix<intmod>& M);
void ntl_image(Matrix<ffield_element>& KerM,   const Matrix<ffield_element>& M);
void ntl_image(Matrix<bigrational>& KerM,      const Matrix<bigrational>& M);

// LLL: find an LLL reduced basis for the *ROWS* of M -- a good first step
// before computing a kernel.
void ntl_LLL(Matrix<bigrational>& L,           const Matrix<bigrational>& M);

// cheapLLL: use fast approximate LLL to find a transition matrix U
// approximately taking M to an LLL reduced bases.
// L is set to U*M.
// For now I'm going to laboriously check that det(U)=+-1
// so for now this is guaranteed.
void ntl_cheapLLL(Matrix<bigrational>& L,      const Matrix<bigrational>& M);

#endif // MY_NTL
