/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

// sparse.cc

#ifndef SPARSE_CC
#define SPARSE_CC

#include <values.h>

#define ROWS_SLACK    1
#define COLUMNS_SLACK 1

#include "system.h"
#include "sparse.h"

template <class T> void 
Sparse<T>::AuxLists::change(int row, int oldn, int newn)
{
  int next = forwards[row];
  int back = backwards[row];
  
  if( next ) backwards[next] = back;
  if( back == 0 )  heads[oldn] = next;
  else forwards[back] = next;
  
  if( newn == 0 ) return;
  int oldh = heads[newn];
  backwards[row] = 0;
  forwards[row] = oldh;
  heads[newn] = row;
  backwards[oldh] = row;
}
template <class T> void 
Sparse<T>::AuxLists::clean_up()
{
  delete [] forwards;
  delete [] backwards;
  delete [] heads;
}

template <class T> void 
Sparse<T>::AuxLists::take_out(int r, int d)
{
  int next = forwards[r];
  int back = backwards[r];
  if( next ) backwards[next] = back;
  if( back ) forwards[back] = next;
  else heads[d] = next;
}

template <class T>
void Sparse<T>::error(const char* msg) const
{
  cerr << "Fatal error in class Sparse. " << msg << endl;
  exit(-1);
}

template <class T>
Sparse<T>::Sparse(const SparseMatrix<T>& sm) : SparseMatrix<T>(sm)
{
  init();
}

template <class T>
void Sparse<T>::BuildFromMatrix (const Matrix<T>& M, int s) 
{
  SparseMatrix<T>::BuildFromMatrix (M, s) ;
  init();
}

template <class T>
void Sparse<T>::BuildFromSparseRep (const Vector<int>& r,
				    const Vector<int>& c,
				    const Vector<T>& val,
				    int _nro, int _nco) 
{
  SparseMatrix<T>::BuildFromSparseRep (r, c, val, _nro, _nco) ;
  init();
}


template <class T>
Sparse<T>::Sparse( const char* filename) : SparseMatrix<T>(filename)
{
  init();
}

template <class T>
Sparse<T>::Sparse (long r, long c, double _MAX_DEN) : SparseMatrix<T>(r,c)
{
  MAX_DEN = _MAX_DEN ;
}

template <class T> Sparse<T>::~Sparse()
{
  row_lists.clean_up();
  col_lists.clean_up();
}

// Function `init()' initializes Sparse class, assuming its
// SparseMatrix part has already been constructed.
//
template <class T> void Sparse<T>::init()
{
  mesg("Initializing sparse matrix. ");
  gauss_done = false;
  ncola = nco;     // number of active columns
  nrowa = nro;     // number of active rows
  active = new bool [nro+1];
  for( int i = 0; i <= nro; i++ ) active[i] = true;
  
  // The object rows contains the actual data for
  // all of the rows of the sparse matrix.
  // GUESS that the number of nonzero entries (+ garbage accumulation)
  // will never exceed about 3/2 * (number of nonzero entries at 
  // start of elimination).  
  // The sparse matrix is stored as a one dimensional array.
  // When more space for a row is needed, garbage collection
  // is used to move things about and obtain more space in
  // this long array.  If this array is exhausted, the sparse
  // function will terminate. 
  
  // This is a good guess, *if* the matrix is sparse.  
  bigint b_rows_size=bigint(int(double(nro)*double(nco)*(MAX_DEN+.1)))+5000;
  // If it is not sparse, this may be way too small to even fit the matrix!
  // Remember, a "Sparse" type matrix can be very dense!
  if(2*count() > b_rows_size) {
    b_rows_size = 2*count(); 
    // count() is defined in sparsematrix.h, from which Sparse is derived. 
    mesg("Note, object of type SparseMatrix is very dense.\n");
  }
  if(b_rows_size>20000000) {
      mesg("matrix size parameter way too large. reducing.");
      b_rows_size=20000000;
  }
  
  int rows_size; b_rows_size.intify(rows_size);
  
  mesg("Sparse: size parameter = "); mesg(rows_size);
  mesg(", max density = "); mesg(MAX_DEN); mesg("\n");
  
  rows->resize(rows_size);
  rows->set_slack(ROWS_SLACK);
  
  row_lists.heads = new long [nco+1];
  for( int i = 0; i <= nco; i++ ) row_lists.heads[i] = 0; 
  row_lists.forwards = new long [nro+1];
  row_lists.backwards = new long [nro+1];
  
  for( int i = 1; i <= nro; i++ ) {
    long d = rows -> nline(i);
    if( d == 0 ) { nrowa--; active[i] = false; }
    
    // set AuxLists row_list;
    int old_head = row_lists.heads[d];
    row_lists.forwards[i] = old_head;
    row_lists.backwards [i] = 0;
    row_lists.heads[d] = i;
    row_lists.backwards [old_head] = i;
  }
  
  // init sparse column representation
  columns = new Lines (nco);
  columns -> resize ( rows->get_size() );
  columns -> set_slack( rows->get_slack() );
  
  // here we do two passes on entries of spr. First to check number of
  // elements in each column and then to write the row numbers
  for( int r = 1; r <= nro; r++ ) {                // first pass
    long nr = rows -> nline(r);
    for( int l = 0; l < nr; l++ ) {
      long col = rows -> Lines::entry( r, l);
      columns->nline(col) += 1;
    }
  }
  Vector<int> colstart (nco);
  for( int c = 1; c <= nco; c++ )     // now does all the column starts
    colstart[c] = columns->start(c) 
      = columns->start(c-1) + columns->nline(c-1);
  for( int r = 1; r <= nro; r++ ) {              // second pass
    long nr = rows -> nline(r) ;
    for( int l = 0; l < nr; l++ ) {
      long col = rows -> entry( r, l );
      columns->set( colstart[col]++, r); // promoted startcopy[c] !
    }
  }
  col_lists.heads = new long [nro+1];
  for( int i = 0; i <= nro; i++ ) col_lists.heads[i] = 0; 
  col_lists.forwards = new long [nco+1];
  col_lists.backwards = new long [nco+1];
  
  // set AuxList col_list
  for( int i = 1; i <= nco; i++ ) {
    int nc = columns -> nline(i);
    if( nc == 0 ) ncola--;
    else {
      int old_head = col_lists.heads[nc];
      col_lists.forwards[i] = old_head;
      col_lists.backwards [i] = 0;
      col_lists.heads[nc] = i;
      col_lists.backwards [old_head] = i;
    }
  }
}

// Function Gauss performs Gauss elimination using
// the Marcovitz ordering of pivots
//
template <class T> void Sparse<T>::Gauss()
{
  // Initializes lots of stuff needed during Gauss elimination
  low_mem_warning = false ;
  rank = 0;
  elimcol.init(nco);
  elimcol.zero();
  col_index = new int [nco+1];
  pivots.init(nco);
  
  int MaxCount = count();
  mesg("[%nonzero, rank]");
  
  while( ncola ) {   // while there are active columns
    
    // checks if matrix is too dense. If it is, switch to Dense elim.
    
    float Count = count();
    float den = Count/(nrowa*ncola);
    if( den > MAX_DEN || low_mem_warning) {
      mesg("\nSwitching to dense elimination on a ");
      mesg(nrowa);mesg("x");
      mesg(ncola);mesg(" matrix with "); 
      mesg(Count);mesg(" nonzero entries.\n");
      columns -> clean();  // frees all space of array columns.
      Dense();
      mesg("\n Dense complete. Now there are ",count()); 
      mesg(" nonzero entries.\n");
      continue;
    }
    
    // get the pivot for this step
    
    int col, row = 0;
    pivot_selection(row,col);
    
    // elimination of row, col
    normalize(row,col);
    clear_col(row,col);
    eliminate(row,col);
    
    // statistics
    
    Count = count();
//    MaxCount = int( MaxCount >? Count );
    if(int(Count)>MaxCount) MaxCount = int(Count) ;
    if (rank%50==0) {
      mesg("["); mesg(100*Count/(nrowa*ncola)); 
      mesg("%,"); mesg(rank); mesg("]");
    }
    
  }    // no active columns in matrix
  mesg("\nrank = ",rank);
  mesg(", maximum number nonzero = ",MaxCount);
  mesg(", final number nonzero = ",count());
  mesg("\n");
  
  gauss_done = true;
  if( nrowa != 0 ) error("In function Gauss()");
  delete [] active;
  delete [] col_index;
}

// Function clear_col(r, c) loops through all rows in column c
// deleting column col from each of them, except row r.
// Row r must be normalized.
//
template <class T> void Sparse<T>::clear_col(long row, long col)
{
  int nrow = columns -> nline(col);
  if( nrow == 1 ) return;                // nothing to do!
  
  // now make a local copy of all rows in this col.
  
  long* buffer = new long [nro]; 
  long* bf = buffer;
  long pos = columns -> start( col );
#ifdef DEBUG
  if (nrow>=nro) error("Error in clear_col.");
#endif
  for( int c = 0; c < nrow; c++ ) *bf++ = columns -> entry( pos++ );
  
  // clear column col with pivot (row,col)
  
  for( int i = 0; i < nrow; i++ ) {
    int row2 = buffer[i];
    if( row2 == row ) continue;
    T alpha = this->operator()(row2,col);
    if( alpha == T(0) ) error("In function clear_col.\n");
    sum_rows(row,row2,-alpha);       // row2 -= row*alpha
  }
  delete [] buffer;
}

// Function eliminate(int r, int c) does a lot of bookeeping
// necessary after an elimination step with pivot (r,c)
//
template <class T> void Sparse<T>::eliminate(int r, int c) 
{
  int d = rows -> nline(r);
  long st = rows -> start(r);
  for( int i = 0; i < d; i++ ) {
    long col = rows -> entry( st++ );
    del_row_from_col(r, col);
  }
  row_lists.take_out(r,d);
  active[r] = false;     // row r is now inactive
  nrowa--;
  elimcol[c] = r;        // column c is pivotal with pivot at row r
  pivots[++rank] = c;      // column c was rank-th pivotal column;
} 

// Function sum_rows(row1,row2,a) performs the operation
// row2 += row1*alpha
// and updates all the row and column information arrays
//
template <class T> void Sparse<T>::sum_rows( int r1, int r2, T alpha )
{
  int* ic = col_index+1;
  for( int i = 0; i < nco; i++ ) *ic++ = -1;
  
  long d1 = rows -> nline(r1);
  long d2 = rows -> nline(r2);
  long pos1 = rows -> start(r1);
  long pos2 = rows -> start(r2);
  T*   val1 = rows->values + pos1;
  T*   val2 = rows->values + pos2;
  int old2 = d2, index = 0;
  for( int i = 0; i < d2; i++ ) {
    int col = rows -> entry(pos2++);
    col_index[ col ] = index++;  // colunas presentes na linha 2.
  }
  pos2 = rows -> start(r2);
  for( int i = 0; i < d1; i++ ) {
    int col = rows -> entry( pos1+i );
    int index = col_index[col];
    if( index > -1 ) {  // column col was already in row2
      val2[index] += val1[i] * alpha;
      if( val2[index] == T(0) ) {
	int col2 = rows -> entry( pos2+index );
	if( col2 != col ) error( "In function sum_rows()");
	int p2 = rows -> entry( pos2 + --d2 );
	rows -> nline( r2 ) -= 1;
	rows -> set( pos2 + index, p2 ); 
	val2[index] = val2[d2];
	col_index[p2] = index;
	rows -> set( pos2 + d2, 0 );
	del_row_from_col(r2,col);
      }
    }
    else {                 // column col was not in row2
      if( rows -> entry( pos2+d2 ) != 0 ) {
	get_more_space(r2, linha);
	pos2 = rows -> start(r2);
	pos1 = rows -> start(r1);
	val2 = rows->values+pos2;
	val1 = rows->values+pos1;
      }
      rows -> set( pos2+d2, col );
      val2[d2++] = val1[i] * alpha;
      rows -> nline(r2)++;
      put_row_in_col(r2,col);
    }
  }
  row_lists.change(r2,old2,d2);
  if( d2 == 0 ) { nrowa--; active[r2] = false; }
}

// Function get_more_space(int r, line l, int n = 0) copies row r 
// to a position in array rows where it can take n entries. n 
// defaults to number of entries+1+slack in row r. If can't do that 
// then garbage collects the array and tries again. If it still can't do
// it the program exits.
//
template <class T> void 
Sparse<T>::get_more_space(int r, line l, int needed = 0)
{
  Lines* L = ( l == linha ? rows : columns );
  int d = L -> nline(r);
  if( needed == 0 ) needed = d + 1 + L->get_slack();
  if( L->available() <= needed ) {      // have to garbage collect
    mesg("Garbage Collect: available=",L->available());
    mesg(", needed=",needed); mesg("\n");
    // prepare array with the info about active elements
    int nrows = 0;
    for( int i = 1; i <= L->nrows(); i++ )
      if( L -> nline(i) > 0 ) nrows++;
    Array<line_info> RI (nrows);
    for( int i = 1, j = 0; i <= L->nrows(); i++ ) 
      if( L->nline(i) > 0 ) RI.set(j++,line_info(i, L->start(i)));
    RI.qsort();           // quick sort in order of allocation
    L -> garbage_collect(RI);
    
    if( l == linha ) {  // have to garbage collect values array
      for( int i = 0; i < nrows; i++ ) {
	int row = RI[i].row;
	int d = L->nline(row);
	T* val1 = rows->values + RI[i].start;
	T* val2 = rows->values + L->start(row);
	for( int k = 0; k < d; k++ ) *val2++ = *val1++;
      }
    }
    // print statistics
    static int ind = 0;
    ind++;
    mesg("After # ",ind); 
    mesg(" gc: ",L->available()); 
    mesg(" positions in ");
    const char* name = ( l == linha ? "rows" : "columns" );
    mesg(name); mesg(" with nrow = ",nrowa);
    mesg(" and ncol = ",ncola);mesg("\n");

    //	if( L->available() <= needed+needed/2) {
    //	  low_mem_warning = true ;
    //#ifdef VERBOSE
    //	  cerr << "Activating low memory warning." << endl ;
    //#endif
    //	}
  }
  if( L->available() > needed ) {
    int oldstart = L->start(r);
    L->copy_to_end(r);
    if( l == linha ) {
      T* val2 = rows->values + L->start(r);
      T* val1 = rows->values + oldstart;
      for( int j = 0; j < L->nline(r); j++ )  *val2++ = *val1++;
    }
  }
  else {
    cerr << "Pop = " << count() <<  "with nrow = " << nrowa 
	 << " and ncol = " <<  ncola << endl;
    error("Not enough space after get_more_space");
  }
}

// Function `del_row_from_col(int r, int c)' deletes row r from 
// column c in the column array.
//
template <class T> Sparse<T>::del_row_from_col(int row, int col)
{
  columns -> take_out (col, row);
  int d = columns->nline(col);
  if( d == 0 ) ncola--;
  col_lists.change(col,d+1,d);   // update count of column
}

// Function `put_row_in_col(int r, int c)' puts row r in 
// column c in the column array. Row is assumed not to be already there.
// If necessary garbage collection is performed.
//
template <class T> Sparse<T>::put_row_in_col(int r, int c)
{
  int d = columns -> nline(c);
  int pos = columns->start(c) + d;  // position where I want to write to
  if( columns->entry(pos) != 0 ) {   // no space at the end, so
    get_more_space(c, coluna);
    pos = columns->start(c) + d;
  }
  if( columns->entry(pos) != 0 ) error("In function put_row_in_col");
  columns->set(pos, r);
  columns->nline(c)++;
  col_lists.change(c,d,d+1);   // update count of column
} 

// Function `pivot_selection(int& row, int& col)' chooses a pivot
// entry using the Marcovitz count function, ie. an entry which minimizes
// (r-1)(c-1), and stores the result in the 2 int arguments.
//
template <class T> void Sparse<T>::pivot_selection(int& row, int& col)
{
  int max = MAXLONG;
  for( int i = 1; i <= nro; i++ ) {
    if( (i-1)*(i-1) >= max ) break;
    
    // investigate columns with i entries
    
    int c = col_lists.heads[i];   // first column with i entries
    while( c != 0 & max > 0 ) {
      int dc = columns -> nline(c);
      long pos = columns->start(c);
      for( int j = 0; j < dc; j++ ) {
	int r = columns->entry(pos++);
	int tmp = (rows->nline(r)-1)*(i-1);
	if( tmp < max ) {
	  max = tmp;
	  row = r;
	  col = c;
	}
      }
      c = col_lists.forwards[c];
    }
    
    // now investigate rows with i entries
    
    if( (i-1)*i >= max ) break;
    if ( i > nco ) break;
    int r = row_lists.heads[i];   // first column with i entries
    while( r != 0 & max > 0) {
      int dr = rows->nline(r);
      long pos = rows->start(r);
      for( int j = 0; j < dr; j++ ) {
	int c = rows->entry(pos++);
	int tmp = (columns->nline(c)-1)*(i-1);
	if( tmp < max ) {
	  max = tmp; 
	  row = r;
	  col = c;
	}
      }
      r = row_lists.forwards[r];
    }
  }
  /*
    if( max < MAXLONG) mesg("max = ", max);
  */
}

// Function `kernel(vector& pc, vector& npc, Matrix<T>& basis)'
// finds the kernel of the matrix. It stores the result in matrix
// basis, which is resized. Pivot column numbers go to vector pc
// and non-pivot ones to vector npc.
// 
// If the DELETE_FREE_ROWS option is set to true then kernel computes
// a basis for the kernel and then deletes all of the rows corresponding
// to numbers in npc.  This is useful because these rows contain exactly
// one 1 and have all other entries = 0.  Thus knowing npc and the
// matrix with the deleted rows gives the original matrix but takes 
// nullity^2*sizeof(T) less bytes to store.   If this option is set,
// pc is set equal to the array newpos which describes how to correctly
// use this matrix. 

template <class T> void 
Sparse<T>::kernel(Vector<int>& pc, Vector<int>& npc, Matrix<T>& basis,
		  bool DELETE_FREE_ROWS)
{
  mesg("Computing sparse kernel\n");
  if( !gauss_done ) Gauss();  // must first reduce to upper triangular form 
  mesg("\nGauss elimination done.  Now back substituting.\n<");
  // set-up vectors pc & npc 
  
  long nullity = nco - rank;
  long ny = 0, k = 0;
  pc.init( rank );
  npc.init( nullity );
  Vector<int> invpiv(nco);
  for( int i = 1; i <= rank; i++ ) invpiv[pivots[i]] = i;
  for( long i = 1; i <= nco; i++ ) {
    if( elimcol[i] ) pc[++k] = i;
    else { npc[++ny] = i; invpiv[i] = ny; }
  }
  if( k != rank ) error("Problems in function kernel"); 
  
  Vector<int> newpos;
  if (DELETE_FREE_ROWS) {
    newpos.init(nco);
    for(int i=1; i<=npc.size(); i++)
      newpos[npc[i]]=-i;
    int j=1;
    for(int i=1; i<=nco; i++)
      if (newpos[i]==0)
	newpos[i]=j++;
  }
  
  // write basis for kernel
  if (DELETE_FREE_ROWS) 
    basis.init(nco-nullity,nullity);
  else
    basis.init(nco,nullity);
  if (nullity == 0) return ;
  const T denom(1);
  if (!DELETE_FREE_ROWS)
    for( long n = 1; n <= nullity; n++ ) 
      basis.set(npc[n],n,denom);
  
  Vector<T> ans(nullity); ans.zero();
  Vector<T> aux(nullity);
  
  for( int n = rank; n; n-- ) {
    if (n%50==0) {mesg(" ",(rank-n)*100/rank);mesg("%");}
    
    long col = pivots[n];
    long row = elimcol[col];
    long d = rows->nline(row);
    long pos = rows->start(row);
    T* val = rows->values + pos;
    
    for( int i = 0; i < d; i++ ) {
      long c = rows->entry(pos++);
      if( c == col ) continue;
      if (DELETE_FREE_ROWS) {
	if (newpos[c]>0)
	  basis.row(newpos[c],aux);
	else {
	  aux.init(nullity) ;
	  aux[-newpos[c]]=T(1);
	}
      } else 
	basis.row(c,aux);
      
      ans -= val[i] * aux;
    }
    if (DELETE_FREE_ROWS) {
      if (newpos[col]>0)
	basis.setrow(newpos[col],ans);
    } else
      basis.setrow( col, ans );
    
    ans.zero();
  }
  if (DELETE_FREE_ROWS) 
    pc = newpos ;
  mesg(">\n");
}

// Function `Dense' peforms dense elimination on the active part 
// of the matrix. Auxlists row_list and col_list are used here to
// store relation between row and col numbers in dense matrix and 
// their values in sparse matrix. At the end of the function their
// values are reset.
//
template <class T> void Sparse<T>::Dense()
{
  Matrix<T> densemat (nrowa,ncola);
  
  // first constructs densemat and set the translation arrays
  
  int nc = 0;
  for( int c = 1; c <= nco; c++ )
    if( columns->nline(c) > 0 ) {  // if col is active
      col_lists.forwards[c] = ++nc;
      col_lists.backwards[nc] = c;
    }
  if( nc != ncola ) error("In function Dense()");
  int nr = 0;
  Vector<T> vec (nc);
  
  for( int r = 1; r <= nro; r++ )
    if( active[r] ) {             // if row is active
      row_lists.forwards[r] = ++nr;
      row_lists.backwards[nr] = r;
      
      int d = rows->nline(r);
      long pos = rows->start(r);
      T* val = rows->values + pos;
      vec.zero();
      for( long i = 0; i < d; i++ ) {
	int col = col_lists.forwards[rows->entry(pos++)];
	vec[col] = *val++;
      }
      densemat.setrow(nr, vec);
    }
  if( nr != nrowa ) error("In function Dense()");
  
  // now performs the dense elimination
  
  Vector<int> dense_elimcol, dense_elimrow;
  densemat.SimpleEchelon(dense_elimcol, dense_elimrow);
  // then redo all the necessary control parameters and arrays
  
  ncola -= nc;
  nrowa -= nr;
  if( ncola == 0 ) {   // much less work since Gauss() is finished
    for( int c = 1; c <= nc; c++ ) {
      int col = col_lists.backwards[c];
      int r = dense_elimcol[c];
      if( r > 0 ) {       // is pivotal
	elimcol[col] = row_lists.backwards[r];
	pivots[++rank] = col;
      }
    }
    for( int r = 1; r <= nr; r++ ) {
      int row = row_lists.backwards[r];
      if( dense_elimrow[r] == 0 ) rows->nline(row) = 0;
      else {
	densemat.row(r,vec);
	const T* vecval = vec.get_vec();
	int d = 0;
	for( int i = 0; i < nc; i++ ) if( vecval[i] != T(0) ) d++;
	if( d > rows->nline(row) )   // have to get more space
	  get_more_space(row,linha,d);
	rows->nline(row) = d;
	long pos = rows->start(row);
	T* val = rows->values + pos;
	for( int i = 0; i < nc; i++ ) 
	  if( vecval[i] != T(0) ) {
	    rows->set(pos++, col_lists.backwards[i+1]);
	    *val++ = vecval[i];
	  }
      }
    }
  }
}



//////////////////////////////////LINEAR ALGEBRA ///////////////////


template <class T>
void Kernel(Matrix<T>& K, const Matrix<T>& A, double MAX_DEN=_MAX)
{
  mesg("Ker(A), A an ",A.nrows());mesg("x",A.ncols());mesgln(" matrix.");
  Sparse<T> S(0,0,MAX_DEN);
  Vector<int> v1,v2;
  int numnonzero=A.count();
/*
  if(double(A.count())/(A.nrows()*A.ncols())>=MAX_DEN) {
      mesg("\nThe matrix is not sparse, therefore we'll perform a simple\n");
      mesg("row reduction strategy before computing the kernel.\n");
      Matrix<T> B=A;
      B.SimpleEchelon(v1,v2);
      S.BuildFromMatrix(B);
  } else {
      S.BuildFromMatrix(A, numnonzero) ;
  }
  S.kernel(v1, v2, K) ; 
*/

  S.BuildFromMatrix(A, numnonzero) ;
  S.kernel(v1, v2, K) ; 

  mesgln("Ker(A) done.");
}

template <class T>
void Kernel(Matrix<T>& K, const Matrix<T>& A, const Vector<T>& f, double MAX_DEN=_MAX)
{
    mesg("Ker(A), A=f(B) with B an ",A.nrows());mesg("x",A.ncols());
    mesgln(" matrix and");
    mesgln("f(X) a polynomial of degree ",f.poldeg());
    Matrix<T> fA;
    A.polsubst(fA, f);
    mesgln("f(B) computed.");
    Kernel(K, fA, MAX_DEN);
}

template <class T>
void Image(Matrix<T>& I, const Matrix<T>& A, double MAX_DEN=_MAX)
{
    mesg("Image(A), A an ",A.nrows());mesg("x",A.ncols());mesgln(" matrix.");
    Matrix<T> B;
    A.transpose(B);
    Vector<int> elimcol,elimrow;
    B.SimpleEchelon(elimcol,elimrow);
    // make I the matrix whose columns are the pivot rows of B.
    int i=0;
    for(int r=1;r<=B.nrows();r++)
	if(elimrow[r]!=0) i++;
    I.init(A.nrows(),i);
    int j=1;
    for(int r=1;r<=B.nrows();r++)
	if(elimrow[r]!=0) 
	    I.setcol(j++,B.row(r));
    mesgln("Image(A) done.");
}

template <class T>
void Image(Matrix<T>& I, const Matrix<T>& A, const Vector<T>& f, double MAX_DEN=_MAX)
{
    mesg("Image(A), A=f(B) with B an ",A.nrows());mesg("x",A.ncols());
    mesgln(" matrix and");
    mesgln("f(X) a polynomial of degree ",f.poldeg());
    Matrix<T> fA;
    A.polsubst(fA, f);
    mesgln("A=f(B) computed.");
    Image(I, fA, MAX_DEN);
}


template <class F>
void Eigenvector(Matrix<F>& v, const Matrix<F>& M, const Vector<F>& f)
{				
    mesg("Eigenvector of M (",M.nrows());mesg("x",M.ncols());mesgln(")");
				// c = start vector
    Vector<F> c; c.init(M.ncols()); c[1]=F(1);
    // f = (x-alpha)h(x), h in Qbar[x].
    while(1) {
	Matrix<F> iterates;
	M.iterate(iterates, c, M.nrows()-1); // powers of M times start vector
	int d=f.poldeg();
	v.init(M.nrows(),d);
	Vector<F> coeff; coeff.init(d); coeff[1]=1;
	Vector<F> x(2); x[1]=0; x[2]=1;
	for(int i=M.nrows();i>0;i--) {
	    Vector<F> vi;
	    iterates.col(i,vi);
	    for(int j=1; j<=d; j++)
		if(coeff[j]!=F(0))
		    v.addtocol(j,coeff[j]*vi);
	    coeff = polmul_polymod(coeff, x, f);
	    coeff[1] += f[i];
	}
	if(!v.iszero()) return;
	mesg("Eigenvector: generating new random start vector.\n");
	for(int j=1; j<=M.ncols(); j++) {
	    bigint rand; rand.randomize(100);
	    c[j]=F(rand); 
	}
    }
    mesgln("Eigenvector done.");
}


#endif
