/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

// sparse.h

#ifndef _SPARSE_H
#define _SPARSE_H
#include "sparsematrix.h"

#define _MAX 0.2

// Class Sparse implements Gaussian elimination for sparse matrices.
// It derives from SparseMatrix and includes a representation by
// sparse columns.
//
enum line { linha, coluna };

template <class T> class Sparse : public SparseMatrix<T> {
private:
  Lines* columns;
    
  bool* active;           // true if row is active
  Vector<int>  elimcol;
  Vector<int>  pivots;
  int* col_index;
  int rank;
  int ncola, nrowa;
  bool gauss_done;
  double MAX_DEN;
  
  struct  AuxLists {
    long* heads;
    long* forwards;
    long* backwards;
    void change(int,int,int);
    void take_out(int,int);
    void clean_up();
  } col_lists, row_lists;
  
  bool low_mem_warning ; 

protected:
  void eliminate(int r, int c);
  void clear_col(long row, long col);
  void get_more_space(int r, line, int needed = 0);
  void pivot_selection(int& row, int& col);
  void sum_rows(int, int, T);
  put_row_in_col(int r, int c);
  del_row_from_col(int r, int c);
  void Dense();
public:
  void init ();
  Sparse(long r = 0,long c = 0, double MAX_DEN=_MAX);
  Sparse( const SparseMatrix<T>& ) ;

				// s = # nonzero entries
				// default: Sparse() computes s. 
  void BuildFromMatrix (const Matrix<T>&, int s = -1) ;
  void BuildFromSparseRep (const Vector<int>& r,
			   const Vector<int>& c,
			   const Vector<T>& val,
			   int nro=-1, int nco=-1) ;
  Sparse( const char* filename);
  ~Sparse();                       
  
  void Gauss();
  void kernel(Vector<int>& pc, Vector<int>& npc, Matrix<T>& basis,
	      bool DELETE_FREE_ROWS=false);
  void setrow(int r, Vector<T> vec);       // 1 <= r <= nro
  
  void error(const char* msg) const;
};

template <class T>
void Kernel(Matrix<T>& K, const Matrix<T>& A, double MAX_DEN=_MAX);

template <class T>
void Kernel(Matrix<T>& K, const Matrix<T>& A, const Vector<T>& f, double MAX_DEN=_MAX);

template <class T>
void Image(Matrix<T>& I, const Matrix<T>& A, double MAX_DEN=_MAX);

template <class T>
void Image(Matrix<T>& I, const Matrix<T>& A, const Vector<T>& f, double MAX_DEN=_MAX);


// Suppose M is an nxn matrix and that M satisfies the irreducible
// polynomial defined by f, i.e., 
//    f[1]+f[2]M+f[3]M^2+...+f[d+1]M^d = 0.
// Note that f is an irreducible factor over the base field F
// of the characteristic polynomial of M.  Also, f is assumed MONIC.
// This function finds *an* eigenvector for the x eigenspace, 
// v is encoded as a matrix.  The rows of v represent the components
// of v with respect to the basis 1,x,x^2,...,x^(d-1) of the F-vector
// space F[x]/f(x). 
//        [ 1+x ]        [1  1  0]
//    v = [ x ]     <--->[0  1  0]
//        [ 1-x^2 ]      [1  0 -1]
// see the manual for a description of how this algorithm works. 
// Throughout I'm using unclever ways of copying bits of matrices
// in order to make the code easier to read.  By far the only
// time consuming part of this algorithm is the call to the
// Kernel functions. 

// WARNING: Even if the eigenspace has dimension > 1, only
// one eigenvector is found. 

template <class F>
void Eigenvector(Matrix<F>& v, const Matrix<F>& M, const Vector<F>& f);

#endif





