/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

// sparsematrix.cc

#ifndef SPARSEMATRIX_CC
#define SPARSEMATRIX_CC

#include "arith.h"
#include "sparsematrix.h"
#include "system.h"      // for mesg.

template <class T>
void SparseMatrix<T>::error(const char* msg) const
{
    cout << "Fatal error in class SparseMatrix.\n " << msg << endl;
    exit(-1);
}

template <class T> long SparseMatrix<T>::count() const
{
    long count = 0;
    for( int i = 1; i <= nro; i++ ) count += rows->nline(i);
    return count;
}

// constructor SparseMatrix(nr,nc)
// creates a zero matrix of nr rows and nc columns
//
template <class T>
SparseMatrix<T>::SparseMatrix(long nr = 0, long nc = 0)
{
    nro = nr;
    nco = nc;
     rows = new LinesVal<T> (nro);
}

template <class T>
SparseMatrix<T>::SparseMatrix(const SparseMatrix<T>& sm)
{
    nro = sm.nro;
    nco = sm.nco;
   /* delete rows; */        // BUG, since the object doesn't exist yet?
    rows = new LinesVal<T>( *sm.rows );
}

				// construct from a Matrix.
template <class T>
void SparseMatrix<T>::BuildFromMatrix (const Matrix<T>& M, int s)
{
  mesg("Building sparse matrix.");
  delete rows ;
  nro = M.nrows () ;
  nco = M.ncols () ;
  rows = new LinesVal<T> (nro);

  if (s == -1) {		// must find # nonzero entries in M.
    s = 0 ;			
    for (int r = 1 ; r <= nro ; r++)
      for (int c = 1 ; c <= nco ; c++)
	if (M(r,c) != T(0))
	  s++ ;
  }

  mesg((double(s)/double(nro))/double(nco)*100);
  mesg("% nonzero entries ");
  
  resize_nelem (s) ;
  long pos[nco] ;
  T val[nco] ;
  for (int r = 1 ; r <= nro ; r++) {
    int t = 0 ;
    for (int c = 1 ; c <= nco ; c++) 
      if (M(r, c) != T(0)) {
	pos[t] = c ;
	val[t++] = M(r,c) ;
      }
    if (t > 0)
      my_setrow (r, t, pos, val) ;
  }
  mesg(" Done.\n");
}

// Build from the sparse representation which gives the nonzero
// entries as a *SORTED* list of coordinate, value pairs. 
// The sort order is the dictionary order with the row value
// having precedence. 
template <class T>
void 
SparseMatrix<T>::BuildFromSparseRep (const Vector<int>& r,
				     const Vector<int>& c,
				     const Vector<T>& val,
				     int _nro, int _nco) 
{
  delete rows ;
  //  nro = intmax(_nro, max(r));
  //  nco = intmax(_nco, max(c));
  nro=_nro;
  for(int i=1; i<=r.size(); i++)
      if(r[i]>nro)
	  nro=r[i];
  nco=_nco;
  for(int i=1; i<=c.size(); i++)
      if(c[i]>nco)
	  nco=c[i];
  rows = new LinesVal<T> (nro);

  int s = r.size();
  resize_nelem (s) ;

  long pos[nco] ;
  T rowval[nco] ;
  int i = 1 ;
  for (int rr = 1 ; rr <= nro ; rr++) {
    if (i <= s && r[i]==rr) {
      int t = 0 ;
      while (i <= s && r[i] == rr) {
	pos[t]=c[i];
	rowval[t]=val[i];
	t++; i++;
      }
      
      if (t > 0)
	my_setrow (rr, t, pos, rowval) ;
    }
  }
}


// constructor from a file. Format of file is:
// First line: nr nc (dimension of matrix).
// Then one line for each row containing: the number of elements
// of the row, all the column numbers then all the values of the 
// entries, given in the same order as the column numbers.
//
template <class T>
SparseMatrix<T>::SparseMatrix( const char* filename )
{
    ifstream infile(filename, ios::in);
    if( !infile ) {
	cout << "\nFile " << filename << "doesn't exist!"; exit(-1);
    }
    infile >> nro >> nco;
    rows = new LinesVal<T> (nro);
    streampos pos = infile.tellg();
    
    // first counts total number of elements and checks that 
    // the file has the right number of lines
    
    long nelem = 0, nr;
    char ch;
    for( long r = 1; r <= nro; r++ ) {
	infile >> nr;
	nelem += nr;
	do { infile.get(ch); } while( ch != '\n' );
	if( infile.eof() )
	    error("Premature end of file in SparseMatrix constructor");
    }
    cout << "Matrix has " << nelem << " elements\n";
    
    // now does a second pass reading the actual values and
    // writing to LinesVal using function rows::setrow
    
    rows->resize(nelem +1);
    infile.seekg(pos);
    long* posi = new long [nco];
    T* val = new T [nco];
    int d;
    for( long r = 1; r <= nro; r++ ) {
      infile >> d;
      for(int i = 0; i < d; i++ ) infile >> posi[i];
      for(int i = 0; i < d; i++ ) infile >> val[i];
      rows->setrow(r,d,posi,val);
    }
    delete [] posi;
    delete [] val;
}


// added by was
template <class T>
void SparseMatrix<T>::resize_nelem ( long nelem )  
{
  rows->resize (nelem + 1) ;
}


template <class T>
void SparseMatrix<T>::my_setrow (long r, int d, long* posi, T* val) 
{
  rows->setrow(r,d,posi,val);
}



// Function as_matrix( Matrix<T>& Mat )
// Writes the SparseMatrix as a dense matrix to Mat
//
template <class T>
void SparseMatrix<T>::as_Matrix( Matrix<T>& Mat ) const
{
    Mat.init( nrows(), ncols() ); 
    for( long i = 1; i <= nro; i++ ) {
	long d = rows -> nline(i);
	Elem<T> elem;
	T* ans = (T*) Mat.getMat() + (i-1)*nco;
	for( int index = 0; index < d; index++ ) {
	    rows->entry( i, index, elem );
	    ans[ elem.pos-1 ] = elem.val;
	}
    }
}

// Operator () (long i, long j)
// Returns value of (i,j) entry, 1 <= i <= nro and 1 <= j <= nco
//
template <class T>
T SparseMatrix<T>::operator() (long i, long j)  const  
{  
    if( (0<i) && (i<=nrows()) && (0<j) && (j<=ncols()) ) {         
	long d = rows -> nline(i);
	Elem<T> elem;
	for( int index = 0; index < d; index++ ) {
	    rows->entry( i, index, elem );
	    if( j == elem.pos ) return elem.val;
	}
	return T(0);
    }
}

template <class T>
SparseMatrix<T>& SparseMatrix<T>::operator = (const SparseMatrix<T>& sm)
{
    if (this==&sm) return *this;
    nro = sm.nro;
    nco = sm.nco;
    delete rows;
    rows = new LinesVal<T> ( *sm.rows );
    return *this;
}

// Multiplication of SparseMatrix by a Matrix
//
template <class T>
Matrix<T> SparseMatrix<T>::operator*( const Matrix<T>& M )
{
    if( nco != M.nrows() )
	error("incompatible SparseMatrix & Matrix in operator*");

    Matrix<T> product( nro, M.ncols() );
    for( long i = 1; i <= nro; i++ ) {
	long d = rows -> nline(i);
	Elem<T> elem;
	for( long j = 1; j <= M.ncols(); j++ ) {
	    T ans(0);  
	    for( long index = 0; index < d; index++ ) {
		rows->entry(i,index,elem);
		ans += elem.val * M(elem.pos, j);
	    }
	    product(i,j) = ans;
	}
    }
}

template <class T>
bool SparseMatrix<T>::equal( const SparseMatrix<T>& sm2 ) const
{
    bool eq = ( nro == sm2.nro ) & (nco == sm2.nco);
    for( long r = 1; r <= nro && eq; r++ ) {
	long d1 = rows -> nline(r); 
	if( d1 != sm2.rows -> nline(r) ) { eq = false; continue; }
	Elem<T> elem;
	for( int i = 0; i < d1 && eq; i++ ) {
	    rows->entry(r,i,elem);
	    eq = ( elem.val == sm2(r, elem.pos) );
	}
    }
    return eq;
}

template <class T> void SparseMatrix<T>::display() const
{
    for( long r = 1; r <= nro; r++ ) {
	cout << "row[" << r << "] =";
	long d = rows -> nline(r);
	Elem<T> elem;
	long n = d-1 > 0 ? d-1 : 0;
	cout << "{ ";
	cout << "values " << "[";
	for( int index = 0; index < d; index++ ) {
	    rows->entry(r,index,elem);
	    cout << elem.val;
	    if( index < d-1 ) cout << ",";
	}
	cout << "]";
	cout << "   positions: " << "[";
	for( int index = 0; index < d; index++ ) {
	    rows->entry(r,index,elem);
	    cout << elem.pos;
	    if( index < d-1 ) cout << ",";
	}
	cout << "] }" << endl;
    }
}

// Function normalize(row,col) make entry at (row,col) equal to 1
template <class T> void SparseMatrix<T>::normalize(int row, int col)
{
  long d = rows -> nline(row);
  T alpha = this->operator()(row,col);
  if( alpha == T(0) ) error("Error in normalize.");
  if( alpha == T(1) ) return;
  else rows->multiply(row, 1/alpha);
}

// Function `row(int r, Vector<T> vec)' stores r-th row of matrix
// (1 <= r <= nro) to Vector<T> vec. 
//
template <class T> void SparseMatrix<T>::row(int r, Vector<T> vec)
{
    if( 1 <= r && r <= nro ) {
	if( dim(vec) != nco ) vec.init(nco);
	long d = rows -> nline(r);
	Elem<T> elem;
	for( int index = 0; index < d; index++ ) {
	    rows->entry(r,index,elem);
	    vec[elem.pos] = elem.val;
	}
    }
    else error( "Bad index in SparseMatrix::row\n" );
}

#include "charmod.h"
//template class SparseMatrix<charmod>;

#endif // SPARSEMATRIX_CC

