/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/*
   FILE:   system.cc
   DESC:   
*/

#include <stdlib.h>
#include <strstream.h>
#include "system.h"

// name of home directory environment variable
//         (e.g., in bash, export HECKE_HOME=/home/was)
const char* HECKE_HOME = "HECKE_HOME";

/////////////////////////////
// I/O INTERFACE
void eof_mesg(bool terminate=true) 
{
  cerr << "Input error: End of file reached." << endl
       << "If you are using a non-interactive input file, then there might " << endl
       << "be no problem at all -- it's just that the end of the file was"<<endl
       << "reached and your file doesn't contain an explicit exit quit command."<<endl;
  if(terminate) exit(-1);
}

void eof_check()
{
  if(cin.eof()) { cerr << "EOF detected. Exiting." << endl; exit(-1); }
}


int get_int(const char* prompt,
	    int min, int max,
	    istream& in,
	    bool terminate_on_eof)
{
  while(1) {
    if(prompt) cout << prompt ;
    if(in.eof()) {
      eof_mesg(terminate_on_eof);
      return 0;
    }

    char c;
    do {
      if(in.eof()) { eof_mesg(terminate_on_eof); return 0; }
      in.get(c);
    } while (c=='\n');
    in.putback(c);
    char buf[64];
    int i=0;
    do {
      if(in.eof()) { eof_mesg(terminate_on_eof); return 0; }
      in.get(buf[i++]);
    } while (i<63&&buf[i-1]!=' ' && buf[i-1]!='\n');
    buf[i-1]='\0';

    int n=atoi(buf);
    if(n<min) {
      cerr << "Sorry, but I can't accept this input." << endl
	   << "Please enter an integer >=" << min<< ":  ";
      continue;
    }
    if(n>max) {
      cerr << "Sorry, but I can't accept this input." << endl
	   << "Please enter an integer <=" << max<< ":  ";
      continue;
    }
    return n;
  }
}

bool get_bool(const char* prompt,
	      bool def,
	      istream& in,
	      bool terminate_on_eof)
{
    while(1) {
	if(prompt) cout << prompt ;
	if(def)
	    cout << " [y] ";
	else
	    cout << " [n] ";
	if(in.eof()) {
	    eof_mesg(terminate_on_eof);
	    return 0;
	}
	
	char c;
	if(in.eof()) { eof_mesg(terminate_on_eof); return 0; }
	do {
	    in.get(c);
	} while (c=='\n');
	if(c=='y' || c=='Y') { in.get(c); return true; }
	if(c=='n' || c=='N') { in.get(c); return false; }
	cout << "Please enter 'y' or 'n'." << endl;
	do {			// clear buffer.
	    if(in.eof()) { eof_mesg(terminate_on_eof); return 0; }
	    in.get(c);
	} while (c=='\n');
	in.putback(c);
    }
}

char get_char(const char* prompt,
	      istream& in,
	      bool terminate_on_eof)
{
  while(1) {
    if(prompt) cout << prompt ;
    if(in.eof()) {
      eof_mesg(terminate_on_eof);
      return 'x';
    }

    char c;
    do {
      if(in.eof()) { eof_mesg(terminate_on_eof); return 'x'; }
      in.get(c);
    } while (c=='\n');
    return c;
  }
}

void get_str(char* str,
	     int maxlen,
	     const char* prompt,
	     istream& in,
	     bool terminate_on_eof)
{
  while(1) {
    if(prompt) cout << prompt ;
    if(in.eof()) {
      eof_mesg(terminate_on_eof);
      str[0]='\0';
      return;
    }
    char c;
    do {
      if(in.eof()) { eof_mesg(terminate_on_eof); return; }
      in.get(c);
    } while (c=='\n');
    in.putback(c);
    if(!in.get(str,maxlen)) {
      cerr << "Sorry, but I can't accept this input.." << endl
	   << "I'm not sure what the problem is."<< endl;
      die("Error in cin.");
    }
    return;
  }
}

#include <sys/stat.h>
#include <sys/types.h>       
#include <fcntl.h>       
#include <unistd.h>
#include <stdio.h>         

char* hecke_home_dir()
{
  static char* s=NULL;
  if(s) return s;
  if(!(s=getenv(HECKE_HOME))) 
    s=".";
  return s;
}

char* username()
{
  static char* s=NULL;
  if(s) return s;
  if(!(s=getenv("USER"))) 
    s=".";
  return s;
}


bool make_data_directory () 
{
  char* s=hecke_home_dir();
  int buflen=strlen(s)+128;
  char buf[buflen];
  ostrstream ost(buf,buflen);
  ost << s << "/data" << '\0';
  if (mkdir(buf,2557) != 0) {
    cerr << "ERROR: Unable to create a directory named " << buf << endl
	 << "This directory is necessary in order to avoid recomputing" << endl 
	 << "data computed by this program.  Please try manually creating" << endl 
	 << "this directory.  Also check that the HECKE_HOME environment" << endl
	 << "variable is set to a valid directory for which you" << endl
	 << "have write permission."<<endl;
    return false;
  }
  return true;
}

bool make_level_directory (int N) 
{
  char* s=hecke_home_dir();
  char buf[strlen(s)+128];
  ostrstream ost(buf,strlen(s)+128);
  ost << s << "/data/" << N << '\0';
  return (mkdir(buf,2557) == 0);
}

bool make_userdata_directory () 
{
  char* s=hecke_home_dir();
  char buf[strlen(s)+128];
  ostrstream ost(buf,strlen(s)+128);
  ost << s << "/userdata" << '\0';
  if (mkdir(buf,2557) != 0) {
    cerr << "ERROR: Unable to create the directory " << buf << "." << endl 
	 << "This directory is necessary for storing data created" << endl 
	 << "from the eigenspace function." << endl 
	 << "Try manually creating a directory called 'userdata'." << endl
         << "Also check that the HECKE_HOME environment variable" << endl
	 << "is set to a valid directory for which you have write" << endl
	 << "permission."<<endl ;
    return false;
  }
  return true;
}


char* data_write (const char* fname, int N)
{
  static char buf[1024];
  ostrstream ost(buf,1024);
  ost<<hecke_home_dir()<<"/data/"<<N<<"/"<<N<<"_"<<fname<<'\0';
  ofstream os(buf);
  if (os.bad()) {
    if(!make_level_directory(N)) 
      if(!make_data_directory()) 
	return NULL; 
      else
	if(!make_level_directory(N))
	  return NULL; 
    ofstream os2(buf);
    if(os2.bad()) {
      cerr << "Error opening data file '" << buf << "'."<<endl ;
      return NULL;
    }
  }
  return buf;
}

char* data_read  (const char* fname, int N)
{
  static char buf[1024];
  ostrstream ost(buf,1024);
  ost<<hecke_home_dir()<<"/data/"<<N<<"/"<<N<<"_"<<fname<<'\0';
  return buf;
}

char* userdata_write (const char* fname)
{
  static char buf[1024];
  ostrstream ost(buf,1024);
  ost<<hecke_home_dir()<<"/userdata/"<<fname<<'\0';
  ofstream os(buf);
  if (os.bad()) {
    if(!make_userdata_directory()) 
	return NULL; 
    ofstream os2(buf);
    if(os2.bad()) {
      cerr << "Error opening userdata file '" << buf << "'." << endl ;
      return NULL;
    }
  }
  return buf;
}

char* userdata_read  (const char* fname)
{
  static char buf[1024];
  ostrstream ost(buf,1024);
  ost<<hecke_home_dir()<<"/userdata/"<<fname<<'\0';
  return buf;
}

/////////////////////////////
// program death

void die(const char* msg)
{
    cerr << "POSSIBLE BUG IN HECKE!?!"<<endl;
    if(msg)
	cerr << msg << endl;
    cerr << "Thus I shall now retire.  Good day."<<endl;
    exit(-1);
}


/////////////////////////////
// TIMING INTERFACE
struct tms buffer ;   
long time0, time1 ;
void reset_timer()
{ 
  times (&buffer) ;
  time0 = buffer.tms_utime ;
}

double get_time ()
{
  times (&buffer) ; 
  time1 = buffer.tms_utime ;
  return (time1 - time0)*1.0 / HZ ;
}

#include<time.h>
char* date_stamp() 
{
    static char buf[1024];
    time_t t;
    time(&t);
    strcpy(buf, ctime(&t));
    buf[strlen(buf)-1]='\0';
    return buf;
}

char* signature_stamp() 
{
    return "William Stein (was@math.berkeley.edu)";
}

// IMPLIMENTATION OF Mesg class.

void Mesg::mesg(const float f)
{
    if(show) 
	if(file)
	    of << f <<flush;
	else
	    cerr << f << flush;
}

void Mesg::mesg(const char* str) 
{
    if(show)
	if(file)
	    of<<str<<flush;
	else
	    cerr << str << flush;
}

void Mesg::mesg(const char* str, const int n) 
{
    if(show)
	if(file)
	    of<<str<<n<<flush;
	else
	    cerr << str << n << flush;
}

void Mesg::mesg(const int n, const char* str) 
{
    if(show)
	if(file)
	    of<<n<<str<<flush;
	else
	    cerr <<n<<str<<flush;
}


void stub(const char* funcname)
{
    cerr << funcname << ": has not yet been written." << endl;
}





