/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu
------------------------------------------------------------------------- */

/*
   FILE:   tables.cc 
   DESC:   
*/

#include "tables.h"
#include "methodofgraphs.h"
#include "display.h"
#include "charpoly.h"
#include "NTL/ZZX.h"
#include "NTL/ZZXFactoring.h"
#include "sparse.h"
#include "modformulas.h"

void 
Tables::processing(int N, char* func) const
{
    if(func) {
	mesg(func);
	mesg(": ");
    }
    mesg("Processing level ",N);
    mesg("\n");
}

/////////////////////////////////////////
// BASIS OF EIGENFORMS
/////////////////////////////////////////

void 
Tables::anlist_g0new() const 
{
    cout << "Compute a pari readable basis of newforms for " << endl
	 << "S_k(Gamma_0(N)) for a specified range of N."<<endl;
    
    int k = get_int("k = ");
    int Nstart = get_int("N start = ");
    int Nstop = get_int("N stop = ");
    int maxn =get_int("Precision: Compute [a1,a2,...,an], where n = ");
    char fname[1024];
    sprintf(fname,"an_s%dg0new_%d-%d.gp",k,Nstart,Nstop);
    ofstream os(userdata_write(fname));
    if(!os) { cout << fname << ": ERROR opening output file."<<endl; return; }
    os << "\\\\ " << fname << endl
       << "\\\\ This file contains a basis for the new subspace of "<< endl
       << "\\\\  S_"<<k<<"(Gamma_0(N)),  "
       << Nstart << " <= N <= " << Nstop << "." << endl
       << "\\\\ Each basis element sum a_n q^n is given in the form "<<endl
       << "\\\\     E[N,i] = [f(x), [a1(x), a2(x), a3(x), a4(x), ..., an(x)]]."<<endl
       << "\\\\ where each ai(x) is a polynomial in a root of f(x)."<<endl
       << "\\\\ WARNING: To avoid displaying fractions the "<<endl
       << "\\\\ vector [a1(x),...] has been normalized."<<endl
       << "\\\\ The number of a_n computed is always at least "<< maxn <<"." << endl
       << "\\\\ The newforms are sorted in ascending dictionary order by trace." << endl
       << "\\\\ " << signature_stamp() << endl
       << "\\\\ " << date_stamp() << endl
       << endl << endl;
    
    for(int N=Nstart; N<=Nstop; N++) {
	processing(N);
	Character chi (N,0,save) ;
	H1NChiZ* H=new H1NChiZ(chi,k,save);
	
	Matrix<bigrational> V; H->get_Sknew(V);
	if(V.ncols()==0) { delete H; continue; }
	MatVecList<bigrational> anlist;
	int the_maxn=maxn;
	while(!H->HeckeEigenvalues_an(anlist,V,maxn)) {
	    cout<<"At level N = " << N << ", couldn't compute basis using " << endl
		<< maxn << " precision.  Thus increasing to " ;
	    maxn = maxn + maxn/3 + 5 ;
	    cout << maxn << " precision."<<endl;
	}
	maxn=the_maxn;

	H->OrderByTrace(anlist);

	for(int i=0;i<anlist.len();i++) {
	    os << "E["<<N<<","<<i+1<<"] = [";
	    if (anlist.v(i).poldeg()>1)
		anlist.v(i).polprint(os);
	    else
		os<<"x";
	    os<<", [";
	    for(int j=1;j<=anlist.M(i).nrows();j++){
		// clear denominators from the matrix of rational numbers anlist[i]:
		bigint d=1;
		for (int r = 1 ; r <= anlist.M(i).nrows() ; r++)
		    for (int c = 1 ; c <= anlist.M(i).ncols() ; c++) 
			if (anlist.M(i)(r,c) != 0)
			    d = lcm (d, denominator(anlist.M(i)(r,c)) ) ;
		anlist.M(i) *= d;
		
		Vector<bigrational> row;
		anlist.M(i).row(j,row);
		row.polprint(os);
		if(j<anlist.M(i).nrows()) os<<",";
	    }
	    os<<"]];"<<endl;
	}
	os<<endl;
    }
}

void 
Tables::aplist_g1new() const 
{
    cout << "Compute a PARI readable list of a_p for a basis of newforms in" << endl
	 << "S_k(Gamma_1(N)) for a specified range of N."<<endl;
    
    int Nstart = get_int("N start = ");
    int Nstop = get_int("N stop = ");
    int k = get_int("k = ");
    int maxp = get_int("compute a2,a3,a5,...,ap, where p = ");
    char fappend[512],fname[1024];
    sprintf(fname,"ap_s%dg1new_%d-%d.gp",k,Nstart,Nstop);
    ofstream os(userdata_write(fname));
    if(!os) { cout << "aplist_g1new: ERROR opening output file."<<endl; return; }
    os << "\\\\ " << fname << endl 
       << "\\\\ This is a PARI readable list of Hecke eigenvalues a_p of a "<<endl
       << "\\\\ basis of newforms in S_"<<k<<"(Gamma_1(N)) for "<<endl
       << "\\\\        " << Nstart << " <= N <= " << Nstop << "." << endl
       << "\\\\ For each galois conjugacy class of newform we give a pair"<<endl
       << "\\\\   [f(x), [a_1(x),a_2(x),a_3(x),a_5(x),a_7(x)...,a_"<<maxp<<"(x)]]"<<endl
       << "\\\\ where f(x) is an irred poly and a_p(x) in Q(x)/(f(x)) is"<<endl
       << "\\\\ the p-th coefficient of the newform sum a_n q^n. "<<endl
       << "\\\\ NOTE: If the a_p(x) do not all lie in Z[x] we multiply the " << endl
       << "\\\\ newform through by an integer to clear denominators." << endl
       << "\\\\ " << signature_stamp() << endl 
       << "\\\\ " << date_stamp() << endl
       << endl<<endl;
    
    for(int N=Nstart; N<=Nstop; N++) {
	processing(N);
	Character chi (N,-1,save) ;
	H1NChiZ* H=new H1NChiZ(chi,k,save);
	
	Matrix<bigrational> V; H->get_Sknew(V);
	if(V.ncols()==0) { delete H; continue; }
	MatVecList<bigrational> aplist;
	
	int the_maxp=maxp;
	while(!H->HeckeEigenvalues_ap(aplist,V,maxp)) {
	    cout<<"At level N = " << N << ", couldn't compute basis using " << endl
		<< maxp << " precision.  Thus increasing to " ;
	    maxp = maxp + maxp/3 + 5 ;
	    cout << maxp << " precision."<<endl;
	}
	maxp=the_maxp;
	
	H->OrderByTrace(aplist);
	for(int i=0;i<aplist.len();i++) {
	    os << "E["<<N<<","<<i+1<<"] = [";
	    if (aplist.v(i).poldeg()>1)
		aplist.v(i).polprint(os,"x");
	    else
		os<<"x";
	    os<<", [";
	    // clear denominators from the matrix of rational numbers aplist[i]:
	    bigint d=1;
	    for (int r = 1 ; r <= aplist.M(i).nrows() ; r++)
		for (int c = 1 ; c <= aplist.M(i).ncols() ; c++) 
		    if (aplist.M(i)(r,c) != 0)
			d = lcm (d, denominator(aplist.M(i)(r,c)) ) ;
	    aplist.M(i) *= d;
	    os<<d<<",";
	    for(int j=1;j<=aplist.M(i).nrows();j++){
		Vector<bigrational> row;
		aplist.M(i).row(j,row);
		row.polprint(os,"x");
		if(j<aplist.M(i).nrows()) os<<",";
	    }
	    os<<"]];"<<endl;
	}
	os<<endl;
    }
}
    

void 
Tables::aplist_g0new_modp() const 
{
#ifdef INST_CHARMOD
    cout << "Compute Hecke eigenvalues for a basis of newforms of S_k(Gamma_0(N); Fp)."<<endl
	 << "The computation is really done using modular symbols"<<endl
	 << "so in some special cases this will fail, e.g., probably if p divides"<<endl
	 << "the discriminant of the Hecke algebra."<<endl
	 << "It is also possible, but very unlikely if p>3, that the dimension"<<endl
	 << "of the modp reduction will go up because of spurious torsion."<<endl;
    cout << endl << endl;
    char uname[512];
    get_str(uname, 512, "Userdata filename = ");
    int maxp = get_int("compute a_2, a_3, ..., a_p where p = ");
    cout << "Enter a list of triples 'N k p' with 3<=p<=251. (N=0 when done.)"<<endl;
    Vector<int> Nlist(8), klist(8), plist(8);
    int i=0;			// number of triples
    while(1) {
	int N=get_int("N = ");
	if(N<1) break;
	int k=get_int("k = ");
	if(k<2 || k%2 !=0) {
	    cout << "Triple discarded: k must be even and >=2."<<endl;
	    continue;
	}
	int p=get_int("p = ");
	if(p<3 || p>251 || !is_prime(p)) {
	    cout << "Triple discarded: p MUST be a prime between 3 and 251."<<endl;
	    continue;
	}
	i++;			
	if(Nlist.size() < i) {
	    Nlist.resize(Nlist.size()+Nlist.size()/2);
	    klist.resize(klist.size()+klist.size()/2);
	    plist.resize(plist.size()+plist.size()/2);
	}
	Nlist[i]=N; klist[i]=k; plist[i]=p;
    };
    if (i==0) {
	cout << "No valid triples entered, so doing nothing." << endl;
	return;
    }
    Nlist.resize(i); klist.resize(i); plist.resize(i);

    char fappend[512],fname[1024];
    sprintf(fname,"apmod_%s.gp",uname);
    cout << "Storing output in userdata/" << fname << endl;
    ofstream os(userdata_write(fname));
    if(!os) { cout << "aplist_g0new_modp: ERROR opening output file."<<endl; return; }
    os << "\\\\" << fname << endl
       << "\\\\Compute Hecke eigenvalues for a basis of newforms of" << endl
       << "\\\\S_k(Gamma_0(N); Fp).  This is really computed using modular symbols"<<endl
       << "\\\\so in some special cases the computation may fail, e.g., maybe, if p divides"<<endl
       << "\\\\the discriminant of the Hecke algebra."<<endl
       << "\\\\It is also possible, but very unlikely if p>3, that the dimension"<<endl
       << "\\\\of the modp reduction will go up because of \"spurious torsion.\""<<endl
       << endl
       << "\\\\ Notation: This table is destined to be input into PARI."<<endl
       << "\\\\ Unfortunately, PARI doesn't support n-dimensional arrays."<<endl
       << "\\\\ Thus for now the output format is" << endl
       << "\\\\    apmod_k7p13[N,i] = [g(x), [a2(x), a3(x), a5(x), ...]]. "<<endl
       << "\\\\ where k=7,p=13 are examples,"<<endl
       << "\\\\ i is the conjugacy class (in no particular order), " << endl
       << "\\\\ g(x) is an irreducible poly over Fp, and the Hecke eigenvalues a2, a3, ..." << endl
       << "\\\\ are expressed as polynomials in a fixed root of g(x)."<<endl<<endl
       << "\\\\ " << signature_stamp() << endl 
       << "\\\\ " << date_stamp() << endl
       << endl<<endl;
    
    for(i=1; i<=Nlist.size(); i++) {
	int N=Nlist[i], k=klist[i], p=plist[i];
	processing(N, "apmod");
	cout << "N = " << N << ", k = " << k << ", p = " << p << endl;
	Character chi (N,0,save) ;
	InitFieldCharmod(p);
	H1NChi<charmod>* H=new H1NChi<charmod>(chi,k,save);
	
	Matrix<charmod> V; H->get_Sknew(V);
	if(V.ncols()==0) { delete H; continue; }
	MatVecList<charmod> aplist;
	
	if(!H->HeckeEigenvalues_ap(aplist,V,maxp)) {
	    cout<<"At level N = " << N << ", couldn't compute basis using " << endl
		<< maxp << " Hecke operators.  Maybe there isn't multiplicity one " << endl
		<< "for the triple N="<<N<<", k="<<k<<", p="<<p<<endl
		<< "Thus skipping this one."<<endl;
	    continue;
	}
	
	for(int i=0;i<aplist.len();i++) {
	    os << "apmod_k"<<k<<"p"<<p;
	    os << "["<<N<<","<<i+1<<"] = [";
	    if (aplist.v(i).poldeg()>1)
		aplist.v(i).polprint(os,"x");
	    else
		os<<"x";
	    os<<", [";
	    for(int j=1;j<=aplist.M(i).nrows();j++){
		Vector<charmod> row;
		aplist.M(i).row(j,row);
		row.polprint(os,"x");
		if(j<aplist.M(i).nrows()) os<<",";
	    }
	    os<<"]];"<<endl;
	}
	os<<endl;
	delete H;
    }
#else
    cout << "mod p support not included in this compilation of HECKE."<<endl;
#endif
}

    
////////////////////////////////////
// DISCRIMINANTS OF HECKE ALGEBRAS
////////////////////////////////////

void
Tables::heckedisc_Sk() const
{
    int N1 = get_int("N start = ");
    int N2 = get_int("N stop = ");
    int k = get_int("k = ");
    bool prime=false;
    char c=get_char("Prime levels only?  [n]  ");
    if(c=='y') prime=true;

    bool factor=false;

    char fappend[512],fname[1024];
    if(prime)
	sprintf(fname,"disc_s%d_prime_%d-%d.gp",k,N1,N2);
    else
	sprintf(fname,"disc_s%d_%d-%d.gp",k,N1,N2);
    ofstream os(userdata_write(fname));
    if(!os) { cout << "heckedisc_Sk: ERROR opening output file."<<endl; return; }
    os << "\\\\ " << fname << endl
       << "\\\\ This is a PARI readable file of discriminants of Hecke " << endl
       << "\\\\ algebras acting on S_"<<k<<"(Gamma_0(N)) for N ";
    if(prime) os << "prime";
    os << endl 
       << "\\\\ in the range:  " << N1 << " <= N <= " << N2 << "." << endl
    << "\\\\ " << signature_stamp() << endl << endl;
    
    for(int N=N1; N<=N2; N++) 
	if (!prime || is_prime(N)) {
	    processing(N);
	    Character chi(N,0,save) ;
	    H1NChiZ* H=new H1NChiZ(chi,k,save);
	    const Matrix<bigrational>& V = H->get_Sk();
	    if(V.ncols()==0) { delete H; continue; }
	    os<<"discsk["<<N<<"] = ";
	    bigrational disc=H->heckedisc(V);
	    if(factor)
		DisplayFactored(disc,os);
	    else
		os<<disc;
	    os<<";"<<endl;
	    delete H;
	}
}

void
Tables::heckedisc_Sknew() const
{
    int N1 = get_int("N start = ");
    int N2 = get_int("N stop = ");
    int k = get_int("k = ");
    bool factor=false;

    char fappend[512],fname[1024];
    sprintf(fname,"disc_s%dnew_%d-%d.gp",k,N1,N2);
    ofstream os(userdata_write(fname));
    if(!os) { cout << "heckedisc_Sknew: ERROR opening output file."<<endl; return; }
    os << "\\\\ " << fname << endl
       << "\\\\ This is a PARI readable file of discriminants of Hecke " << endl
       << "\\\\ algebras acting on S_"<<k<<"^new(Gamma_0(N)) for N "<<endl
       << "\\\\ in the range:  " << N1 << " <= N <= " << N2 << "." << endl
       << "\\\\ along with the products prod disc(O_f) = product of discriminants"<<endl
       << "\\\\ of the rings Z[a_1,a_2,a_3,...] associated to a basis of newforms f."<<endl
       << "\\\\ Computed by William Stein (was@math.berkeley.edu), Feb, 1999." << endl << endl;
    
    for(int N=N1; N<=N2; N++) {
	processing(N);
	Character chi(N,0,save) ;
	H1NChiZ* H=new H1NChiZ(chi,k,save);
	if(H->dim_Sknew()==0) { delete H; continue; }
	bigrational discTnew, discTfnewprod, Vol, Idx;
        H->heckedisc_Sknew(discTnew,discTfnewprod, Vol, Idx);
	os<<"discnew["<<N<<"] = " <<discTnew<<";"<<endl;
	os<<"discnewprod["<<N<<"] = "<<discTfnewprod<<";"<<endl;
//	os<<"Vol["<<N<<"] = "<<Vol<<endl;
//	os<<"Idx["<<N<<"] = "<<Idx<<endl;
	delete H;
    }
}


void
Tables::heckedisc_normalization_Sknew() const
{
    cout << "heckedisc_normalization_Sknew "<<endl;
    int N1 = get_int("N start = ");
    int N2 = get_int("N stop = ");
    int k = get_int("k = ");

    char fappend[512],fname[1024];
    sprintf(fname,"disc_s%d_normal_%d-%d.gp",k,N1,N2);
    ofstream os(userdata_write(fname));
    if(!os) { cout << "heckedisc_T1Tg: ERROR opening output file."<<endl; return; }
    os << "\\\\ " << fname << endl
       << "\\\\ This is a PARI readable file of discriminants of the normalization of the Z-module "<<endl
       << "\\\\ generated by T_1,...,T_g which act on "<<endl
       << "\\\\ S_"<<k<<"^{new}(Gamma_0(N)) for N "<<endl
       << "\\\\ in the range:  " << N1 << " <= N <= " << N2 << "." << endl
       << "\\\\ This discriminant is the product of the discriminants of the number"<<endl
       << "\\\\ fields cut out by a basis of newforms."<<endl
       << "\\\\ Computed by William Stein (was@math.berkeley.edu)" << endl << endl;
    
    for(int N=N1; N<=N2; N++) {
	processing(N);
	Character chi(N,0,save) ;
	H1NChiZ* H=new H1NChiZ(chi,k,save);
	if(H->dim_Sknew()==0) { delete H; continue; }
	os<<"discnorm["<<N<<"] = ";
	os<<H->heckedisc_normalization_Sknew()<<";"<<endl;
	delete H;
    }
}

void 
Tables::heckedisc_T1_Tg() const
{
    int N1 = get_int("N start = ");
    int N2 = get_int("N stop = ");
    int k = get_int("k = ");
    bool factor=false;
    bool newsubspace=false;

    char fappend[512],fname[1024];
    sprintf(fname,"disc_s%dT1Tg_%d-%d.gp",k,N1,N2);
    ofstream os(userdata_write(fname));
    if(!os) { cout << "heckedisc_T1Tg: ERROR opening output file."<<endl; return; }
    os << "\\\\ " << fname << endl
       << "\\\\ This is a PARI readable file of discriminants of the Z-module "<<endl
       << "\\\\ generated by T_1,...,T_g which act on "<<endl
       << "\\\\ S_"<<k<<"(Gamma_0(N)) for N "<<endl
       << "\\\\ in the range:  " << N1 << " <= N <= " << N2 << "." << endl
       << "\\\\ Computed by William Stein (was@math.berkeley.edu)" << endl << endl;
    
    for(int N=N1; N<=N2; N++) {
	processing(N);
	Character chi(N,0,save) ;
	H1NChiZ* H=new H1NChiZ(chi,k,save);
	if(H->dim_Sknew()==0) { delete H; continue; }
	os<<"disct1tg["<<N<<"] = ";
	bool dontgenerate;
	bigrational disc=H->heckedisc_T1_Tg(dontgenerate);
	if(dontgenerate) {
	    os << "[]";
	} else {
	    if(factor)
		DisplayFactored(disc,os);
	    else
		os<<disc;
	} 
	os<<";"<<endl;
	delete H;
    }
}


//////////////////////////////////////////
// BIRCH AND SWINNERTON-DYER DATA
//////////////////////////////////////////

void
Tables::bsd() const
{
    cout << "------ bsd ------"<<endl;
    int prec = get_int("prec = ");
    int nvals = get_int("Number of values to compute (0 for range) = ");
    int N1, N2;
    Vector<int> vals;
    if(nvals<=0) {
	N1 = get_int("N start = ");
	N2 = get_int("N stop  = ");
	vals.init(N2-N1+1);
	for(int i=1; i<=vals.size(); i++)
	    vals[i]=N1+(i-1);
    } else {
	cout << "Enter " << nvals << " values: "<<endl;
	vals.init(nvals);
	for(int i=1; i<=vals.size(); i++)
	    vals[i]=get_int("");
    }
    int k = get_int("k = ");

    char fname[2048];
    ostrstream ost(fname,2048);
    ost << "bsd_prec" << prec << "_" ; 
    if(k>2) ost << "k"<<k<<"_";
    if(nvals<=0)
	ost << N1 << "-" << N2;
    else {
	for(int i=1;i<vals.size();i++)
	    ost << vals[i] << "_";
	ost << vals[vals.size()];
    }
    ost << ".gp" << '\0';

    ofstream os(userdata_write(fname));
    if(!os) { cout << "bsd: ERROR opening output file."<<endl; return; }
    os << "\\\\ " << fname << endl 
       << "\\\\ ---------------------------------------------------------------"<<endl
       << "\\\\ N,i      = i-th newform at level N (forms ordered by trace)." << endl 
       << "\\\\ bsd[N,i] = odd_part(L(A_f,1)/Omega*(manin constant))" << endl
       << "\\\\ deg[N,i] = odd_part(modular degree)" << endl
       << "\\\\ dsc[N,i] = disc(O_f)" << endl
       << "\\\\ E  [N,i] = [g(x), denom*[a_1(x), a_2(x), ..., a_"<< prec << "(x)]]" << endl;
    if(nvals<=0)
	os << "\\\\ " << N1 << " <= N <= " << N2 << endl;
    os << "\\\\ " << endl 
       << "\\\\ " << date_stamp() << endl
       << "\\\\ William Stein (was@math.berkeley.edu)" << endl 
       << "\\\\ ---------------------------------------------------------------"<<endl;

    
    for(int m=1; m<=vals.size(); m++) {
	int N=vals[m];
	processing(N);
	Character chi(N,0,save) ;
	H1NChiZ* H=new H1NChiZ(chi,k,save);
	if(H->dim_Sknew()==0) { delete H; continue; }

	MatVecList<bigrational> eigenbasis, anlist;
	ModularFormulas m;
	if(!H->HeckeEigenvalues_an(anlist,eigenbasis,H->get_Sknew(),
				   int(float(k)*float(m.idxG0(N))/float(12))+1)) 
	    die("Could not compute eigenbasis!");

	mesgln("About to order by trace!");
	H->OrderByTrace(anlist, eigenbasis);
	mesgln("Just succeeded!");

	Vector<bigrational> disc, moddeg;
	Matrix<bigrational> bsd;
	H->AllBSD (bsd, disc, eigenbasis, anlist);
	H->AllModularDegree(moddeg, anlist);
	os << endl ;
	for(int i=0;i<anlist.len();i++) {
	    for(int s=1;s<=k-1;s++)
		if(bsd(i+1,s)!=0) {
		    while(numerator(bsd(i+1,s))%2==0){bsd(i+1,s)/=2;}
		    while(denominator(bsd(i+1,s))%2==0){bsd(i+1,s)*=2;}
		}     

	    if(moddeg[i+1]!=0) {
		while(numerator(moddeg[i+1])%2==0){moddeg[i+1]/=2;}
		while(denominator(moddeg[i+1])%2==0){moddeg[i+1]*=2;}
	    }     


	    os << endl;
	    os << "bsd[" << N << "," << i+1 << "] = [";
	    for(int s=1;s<=k-1;s++) {
		DisplayFactored(bsd(i+1,s),os);
		if(s<k-1)
		    os << ", ";
	    }
	    os << "];" << endl ;
	    os << "deg[" << N << "," << i+1 << "] = " ;
	    DisplayFactored(moddeg[i+1],os); os << " ;" << endl; 

	    os << "dsc[" << N << "," << i+1 << "] = " << disc[i+1] <<endl;
	    os << "E  [" << N << "," << i+1 <<"] = [";
	    if (anlist.v(i).poldeg()>1)
		anlist.v(i).polprint(os,"x");
	    else
		os<<"x";
	    os<<", [";
	    for(int k=1;k<=min(prec,anlist.M(i).nrows());k++){
		// clear denominators from the matrix of rational numbers anlist[i]:
		bigint d=1;
		for (int r = 1 ; r <= anlist.M(i).nrows() ; r++)
		    for (int c = 1 ; c <= anlist.M(i).ncols() ; c++) 
			if (anlist.M(i)(r,c) != 0)
			    d = lcm (d, denominator(anlist.M(i)(r,c)) ) ;
		anlist.M(i) *= d;
		
		Vector<bigrational> row;
		anlist.M(i).row(k,row);
		row.polprint(os,"x");
		if(k<min(prec,anlist.M(i).nrows())) os<<",";
	    }
	    os<<"]];"<<endl;
	}

	delete H;
    }
}



//////////////////////////////////////////
// SUPERSINGULAR BASIS
//////////////////////////////////////////

void
Tables::supersingular() const
{
}

/*
typedef Vector<bigrational> vb;

// N
// a b     [where a+bx+x^2=0 defines a quadratic extension of F_N]
// n       [=number of s.s. j's in char. N = g+1.]
// a1 a2   b1 b2  ...   [n pairs, where a1+x*a2 is a s.s. j-invariant in char. N]
// for each eigenform...
// d  a0 a1 ... a_{d-1} [coeff of poly a0+a1x+...+a_{d-1}x^{d-1}+x^d over 
//                       which the form is defined.]
// a0 a1 ... a_{d-1}   b0 b1 ... b_{d-1}   [a_2, a_3, a_5, a_7 for the eigenform]
// a0 a1 ... a_{d-1} ...    [n coefficients = expression of the
//                                            eigenform on the s.s. basis.]

void
Tables::methodofgraphs(int Nstart, int Nstop,
		       const char* fname) const
{
#ifdef DEBUG
    if(Nstart<1) error("methodofgraphs: Nstart<1");
    if(Nstop<Nstart) error("methodofgraphs: Nstop<Nstart");
#endif

    ofstream out(userdata_write(fname));
    if (!out) error("methodofgraphs: couldn't open output file.");
    
    for(int N=Nstart; N<=Nstop; N++) 
	if (is_prime(N)) {
	    cout << "Computing N = " << N << endl;
	    out<<N<<endl;
	    MethodOfGraphs MOG(N);
	    Matrix<bigrational>* T=new Matrix<bigrational>[4];
	    MOG.Tp(T[0],2);      
	    MOG.Tp(T[1],3);
	    MOG.Tp(T[2],5);
	    MOG.Tp(T[3],7);
	    
	    const Fp_polynomial pol
		= ffield_element::get_def_base().irred_polynomial() ;
	    out<<pol[0]<< ' ' << pol[1]<<endl;
	    const Vector<ffield_element>& ss = MOG.ssbasis();
	    out<<ss.size()<<endl;
	    for (int i=1; i<=ss.size(); i++) {
		const Fp_polynomial rep=ss[i].polynomial_rep();
		if(rep.degree()==1)
		    out << rep[0] << " " << rep[1] << "   " ;
		else if (rep.degree()==0)
		    out << rep[0] << " 0   ";
		else 
		    out << "0 0   ";
	    }
	    out << endl;
	    
	    Eigen                E;
	    int                  numforms;
	    polynomial<bigint>*  f;
	    Matrix<bigrational>* basis;
	    Matrix<bigrational>*  eigenbasis;
	    
	    E.eigen_expansion(f, basis, eigenbasis, numforms, T, 4);
	    
	    if (numforms==-1) 
		out << "[* * N = " << N << " must be done by a different method! * * ]"<<endl;
	    else {
		for (int i=0; i<numforms; i++) {
		    // poly over which eigenform is defined
		    out << f[i].degree() << "  " ;
		    for (int j=0; j<f[i].degree(); j++)
			out << (f[i])[j]<<" ";
		    out << endl ;
		    // first few a_p
		    for (int j=1; j<=4; j++) {
			for (int z=1; z<=f[i].degree(); z++)
			    out << basis[i](j,z) << " ";
			out<<"   ";
		    }
		    out << endl;
		    // eigenform on ss basis
		    for (int j=1; j<=eigenbasis[i].nrows(); j++) {
			for (int z=1; z<=f[i].degree(); z++)
			    out << eigenbasis[i](j,z) << " ";
			out<<"  ";
		    }
		    out << endl ;
		}
		delete[] f;
		delete[] basis;
		delete[] eigenbasis;
	    }
	    out << endl << endl ;
	    
	    delete[] T;
	}
    out << "-1"<<endl;		// eof
}

void
Tables::methodofgraphs_tolatex(const char* infname, const char* outfname) const
{
    ifstream in(infname);
    ofstream out(outfname);
    if(!in) error("methodofgraphs_tolatex: couldn't open input file");
    if(!out) error("methodofgraphs_tolatex: couldn't open output file");
    
    // header
    out << "%this document automatically generated by the "<<endl
	<< "%Tables::methodofgraphs_tolatex() method of the "<<endl
	<< "%hecke program.  William Stein."<<endl
	<< "\\documentclass{article}"<<endl
	<< "\\textwidth=480pt"<<endl
	<< "\\hoffset=-75pt"<<endl
	<< "\\textheight=700pt"<<endl
	<< "\\voffset=-89pt"<<endl
	<< "\\usepackage{amsfonts}"<<endl
	<< "\\usepackage{boxedminipage}"<<endl
	<< "\\usepackage{amsmath}"<<endl
	<< "\\newcommand{\\C}{\\mathbb C}"<<endl
	<< "\\newcommand{\\g}{\\gamma}"<<endl
	<< "\\title{Eigenforms on the supersingular basis}"<<endl
	<< "\\author{William A. Stein}"<<endl
	<< "\\begin{document}"<<endl
	<< "\\maketitle"<<endl
	<< endl
	<< "\\section{Introduction}"<<endl
	<< "For any prime number $p$, let $M_2(p)$ denote the complex"<<endl
	<< "vector space spanned by the supersingular $j$-invariants"<<endl
	<< "in characteristic $p$.  For primes $\\ell\\neq p$, the "<<endl
	<< "Hecke operator $T_{\\ell}$" << endl
	<< "acts on $M_2(p)$ by sending $j$ to the sum, counted with"<<endl
	<< "multiplicity, of all $j$-invariants of elliptic curves" << endl
	<< "$\\ell$-isogeneous to a fixed elliptic curve of $j$-invariant"<<endl
	<< "$j$.  Equipped with this Hecke action, $M_2(p)$ is "<<endl
	<< "non-canonically isomorphic to the space of weight 2 modular "<<endl
	<< "forms of level $p$."<<endl<<endl
	<< "The following table lists, for each prime number $p\\leq ??$,"<<endl
	<< "a basis of eigenforms for the action of the "<<endl
	<< "Hecke algebra on $M_2(p)$."<<endl
	<< "The representation of the eigenform on the supersingular"<<endl
	<< "basis is given in round brackets.  Then the first four"<<endl
	<< "Hecke eigenvalues $[a_2, a_3, a_5, a_7]$ are listed."<<endl
	<< endl
	<< "\\section{Table}"<<endl;
    
    int null;
    while (1) {
	int N; 
	in>>N; 
	if (N==-1) break;
	cout<<"Parsing level N="<<N<<endl;
	vb alp(2);
	in>>alp[1]>>alp[2];
	int nss; in>> nss;
	vb* ss=new vb[nss];
	for(int i=0; i<nss;i++) {
	    ss[i].init(2);
	    in >> ss[i][1] >> ss[i][2];
	}
	
	// write header: (e.g.,)
	
	out<<"\\bigskip\\par\\noindent{\\LARGE $\\mathbf{"<<N<<"}$}"<<endl
	   <<"\\begin{boxedminipage}{\\textwidth}"<<endl
	   <<"$M_2("<<N<<")=";
	
	for(int i=0; i<nss;i++) {
	    if(i>0)out<<"\\oplus";
	    out << "\\C[";
	    texprintalgnum(out,ss[i],"\\alpha");
	    out << "]";
	}
	bool allrat=true; 
	for(int i=0;i<nss;i++) 
	    if(ss[i][2]!=0) {allrat=false; break;}
	if(allrat==false) {
	    out<<", \\,\\,";
	    texprintpoly(out,alp,"\\alpha");
	    out<<"$";
	} else out<<"$"<<endl;
	
	out<<"\\vspace{-.18cm}\\\\"<<endl
	   <<"\\mbox{}\\dotfill\\mbox{}\\\\"<<endl;
	
	int n=nss;
	int num=0;
	while(n>0) {
	    int d; in>>d; n-=d;
	    vb poly(d);
	    for(int i=1;i<=d;i++)
		in>>poly[i];
	    
	    vb* ap=new vb[4];
	    for(int i=0;i<4;i++) {
		ap[i].init(d);
		for(int j=1;j<=d;j++)
		    in>>ap[i][j];
	    }
	    vb* ssrep=new vb[nss];
	    for(int i=0;i<nss;i++) {
		ssrep[i].init(d);
		for(int j=1;j<=d;j++)
		    in>>ssrep[i][j];
	    }
	    // write form data:
	    out<<char('A'+num)<<"=";
	    out<<"(";
	    for(int i=0;i<nss;i++) {
		out<<"$";
		texprintalgnum(out,ssrep[i],"\\g");
		out<<"$";
		if(i<nss-1)out<<",";
	    }
	    out<<"), $\\,$"<<endl;
	    
	    out<<"$[$";
	    for(int i=0;i<4;i++) {
		out<<"$";
		texprintalgnum(out,ap[i],"\\g");
		out<<"$";
		if(i<3) out<<",";
	    }
	    out<<"$]";
	    
	    if(d>1) {
		out<<",$ \\,"<<endl;
		out<<"$";
		texprintpoly(out,poly,"\\g");
	    }
	    out<<"$";
	    
	    out<<"\\\\"<<endl;
	    
	    num++;
	    delete[] ap;
	    delete[] ssrep;
	}
	delete[] ss;
	out<<endl<<"\\end{boxedminipage}"<<endl;
    }
    out<<"\\end{document}"<<endl;
    cout<<"Bye."<<endl;
}    
*/


////////////////////////////////////////////
//  DIMENSION TABLES
////////////////////////////////////////////

void 
Tables::dimg0() const
{
    cout << endl << endl
	 << "======== dim S_k(Gamma_0(N)) ========="<<endl<<endl;
    
    int Nstart, Nstop, kstop;
    Nstart=get_int("Nstart = ");
    Nstop=get_int("Nstop = ");
    kstop=get_int("kstop = ");
    char fname[1024];
    sprintf(fname, "dim_skg0_N%d-%d_k%d-%d.gp", Nstart, Nstop, 2, kstop);
    
    if(kstop%2!=0) kstop--;
    ofstream out(userdata_write(fname));

    if(!out) {
	mesg("dimg0: couldn't open output file",fname);
	die("dimg0: file open error.");
    }
    out<<"\\\\ " << fname << endl
	<< "\\\\ This is a table of dimensions of S_k(Gamma_0(N)) "<<endl
	<<"\\\\ " << signature_stamp() << endl
	<<"\\\\ " << date_stamp() << endl
	<<endl;
    out<<"{\\\\ dim[N,k]=dim(S_{2k}(Gamma_0(N))), 2k>=2"<<endl
       <<"dim=matrix("<<Nstop<<","<<kstop/2<<",m,n,0);"<<endl<<endl;
    ModularFormulas formula;
    for (int N=Nstart; N<=Nstop; N++) {
	processing(N);
	out<<"dim["<<N<<","<<"]=[";
	for (int k=2; k<=kstop; k+=2) {
	    if (k%(kstop/10+1)==0) cout <<k<<" "<<flush;
	    out<<formula.dimSk(N,k);
	    if (k<kstop-1)out<<",";
	}
	out<<"];"<<endl;
    }
    out<<"}"<<endl;
}

void 
Tables::dimg0new() const
{
    cout << endl << endl << "======== dim S_k^{new}(Gamma_0(N)) ========="<<endl<<endl;
    int Nstart, Nstop, kstop;
    Nstart=get_int("Nstart = ");
    Nstop=get_int("Nstop = ");
    kstop=get_int("kstop = ");

    char fname[1024];
    sprintf(fname, "dim_skg0new_N%d-%d_k%d-%d.gp", Nstart, Nstop, 2, kstop);

    if(kstop%2!=0) kstop--;
    ofstream out(userdata_write(fname));
    if(!out) {
	mesg("dimg0new: couldn't open output file",fname);
	die("dimg0new: file open error.");
    }
    out<<"\\\\ " << fname << endl
       <<"\\\\ This is a table of dimensions of the spaces" << endl
       <<"\\\\             S_k^{new}(Gamma_0(N)) "<<endl
       <<"\\\\ of weight k cuspidal newforms for Gamma_0(N)."<<endl
       <<"\\\\ " << signature_stamp() << endl
       <<"\\\\ " << date_stamp() << endl
       <<endl<<endl;
    out<<"{\\\\ dim[N,k]=dim(S_{2k}^{new}(Gamma_0(N))), 2k>=2"<<endl
       <<"dim=matrix("<<Nstop<<","<<kstop/2<<",m,n,0);"<<endl<<endl;
    ModularFormulas formula;
    for (int N=Nstart; N<=Nstop; N++) {
	processing(N);
	out<<"dim["<<N<<","<<"]=[";
	for (int k=2; k<=kstop; k+=2) {
	    if (k%(kstop/10+1)==0) cout <<k<<" "<<flush;
	    out<<formula.dimSknew(N,k);
	    if (k<kstop-1)out<<",";
	}
	out<<"];"<<endl;
    }
    out<<"}"<<endl;
}

void 
Tables::dimg1() const
{
    cout << endl << endl << "======== charpoly's on S_k(Gamma_1) ========="<<endl<<endl;
    int Nstart = get_int("N start = ");
    int Nstop = get_int("N stop = ");
    int kstop = get_int("kstop = ");
    
    char fname[1024];
    sprintf(fname, "dim_skg1_N%d-%d_k%d-%d.gp", Nstart, Nstop, 2, kstop);

    if(kstop%2!=0) kstop--;
    ofstream out(userdata_write(fname));
    if(!out) { mesg("dimg1: couldn't open output file",fname); 
       die("dimg1: file open error.");}
    
    out<<"\\\\ " << fname <<endl
       <<"\\\\ This is a table of dimensions of S_k(Gamma_1(N))."<<endl
       <<"\\\\ " << signature_stamp() << endl
       <<"\\\\ " << date_stamp() << endl
       <<endl<<endl;
    out<<"{\\\\ dim[N,k]=dim(S_k(Gamma_1(N))), k>=2"<<endl
       <<"dim=matrix("<<Nstop<<","<<kstop<<",m,n,0);"<<endl<<endl;
    ModularFormulas formula;
    for (int N=Nstart; N<=Nstop; N++) {
	processing(N);
	out<<"dim["<<N<<","<<"]=[-1,";
	for (int k=2; k<=kstop; k++) {
	    if (k%(kstop/10+1)==0) cout <<k<<" "<<flush;
	    out<<formula.dimSk1(N,k);
	    if (k<kstop)out<<",";
	}
	out<<"];"<<endl;
    }
    out<<"}"<<endl;
}

void 
Tables::dimg1new() const
{
    cout << endl << endl << "======== dim S_k(Gamma_1(N)) ========="<<endl<<endl;
    int Nstart = get_int("N start = ");
    int Nstop = get_int("N stop = ");
    int kstop = get_int("kstop = ");
    
    char fname[1024];
    sprintf(fname, "dim_skg1new_N%d-%d_k%d-%d.gp", Nstart, Nstop, 2, kstop);

    if(kstop%2!=0) kstop--;
    ofstream out(userdata_write(fname));
    if(!out) {
	mesg("dimg1new: couldn't open output file",fname); 
	die("dimg1new: file open error.");
    }

    out<<"\\\\ " << fname << endl
       <<"\\\\ This is a table of dimensions of S_k(Gamma_1(N))^new."<<endl
       <<"\\\\ " << signature_stamp() << endl
       <<"\\\\ " << date_stamp() << endl
       <<endl<<endl;
    out<<"{\\\\ dim[N,k]=dim(S_k^{new}(Gamma_1(N))), k>=2"<<endl
       <<"dim=matrix("<<Nstop<<","<<kstop<<",m,n,0);"<<endl<<endl;
    ModularFormulas formula;
    for (int N=Nstart; N<=Nstop; N++) {
	processing(N);
	out<<"dim["<<N<<","<<"]=[-1,";
	for (int k=2; k<=kstop; k++) {
	    if (k%(kstop/10+1)==0) cout <<k<<" "<<flush;
	    out<<formula.dimSk1new(N,k);
	    if (k<kstop)out<<",";
	}
	out<<"];"<<endl;
    }
    out<<"}"<<endl;
}

////////////////////////////////////
// CHARACTERISTIC POLYNOMIALS
////////////////////////////////////

void 
Tables::cpg0() const
{
    cout << endl << endl << "======== charpoly's on S_k(Gamma_0) ========="<<endl<<endl;
    int Nstart = get_int("N start = ");
    int Nstop = get_int("N stop = ");
    int pstop = get_int("pstop = ");
    int k = get_int("k = ");

    char fname[1024];
    sprintf(fname, "charpoly_s%dg0_T2-%d_N%d-%d.gp", k, pstop, Nstart, Nstop);
    ofstream out(userdata_write(fname));
    if(!out) {
	mesg("cpg0: can't open ",fname);
	die("cpg0 file open error.");
    }
    out<<"\\\\ " << fname << endl
       <<"\\\\ This file contains the characteristic polynomials of the" << endl
       <<"\\\\ Hecke operators T_p acting on S_"<<k<<"(Gamma_0(N)) "<<endl
       <<"\\\\ " << signature_stamp() << endl
       <<"\\\\ " << date_stamp() << endl
       <<endl<<endl;
    out<<"{"<<endl
       <<"T=matrix("<<Nstop<<","<<pstop<<",m,n,0);"<<endl;
    for (int N=Nstart; N<=Nstop; N++) {
	processing(N);
	Character chi (N,0,save) ;
	H1NChi<bigrational>* H=new H1NChi<bigrational>(chi,k,save);
	if(H->dim_Sk()==0) {delete H; continue;}
	for (int p=2; p<=pstop; p=next_prime(p)) 
	    if (is_prime(p)) {
		out<<"T["<<N<<","<<p<<"]=";
		Matrix<bigrational> T;
		H->Tn_Sk(p,T);
		ZZX f;
		CharPoly(T,f);
		Display(f,1,1,out);
		out<<";"<<endl;
	    }
	out<<endl;
	delete H;
    }
    out<<"}"<<endl;
}

void 
Tables::cpg0new() const
{
    cout << endl << endl
	 << "======== charpoly's on S_k(Gamma_0)^new ========="<<endl<<endl;
    int Nstart = get_int("N start = ");
    int Nstop = get_int("N stop = ");
    int pstop = get_int("pstop = ");
    int k = get_int("k = ");

    char fname[1024];
    sprintf(fname, "charpoly_s%dg0new_T2-%d_N%d-%d.gp", k, pstop, Nstart, Nstop);
    ofstream out(userdata_write(fname));
    if(!out) {
	mesg("cpg0new: can't open ",fname);
	die("cpg0new file open error.");
    }
    out<<"\\\\ " << fname << endl
       <<"\\\\ This file contains the characteristic polynomials of the" << endl
       <<"\\\\ Hecke operators T_p acting on S_"<<k<<"(Gamma_0(N))^new "<<endl
       <<"\\\\ " << signature_stamp() << endl
       <<"\\\\ " << date_stamp() << endl
       <<endl<<endl;

    out<<"{"<<endl
       <<"T=matrix("<<Nstop<<","<<pstop<<",m,n,0);"<<endl;
    for (int N=Nstart; N<=Nstop; N++) {
	processing(N);
	if(is_prime(N)) continue;	// prime levels are uninteresting,
				// since all forms are new, and
				// we've already done that case before.
	Character chi (N,0,save) ;
	H1NChi<bigrational>* H=new H1NChi<bigrational>(chi,k,save);
	Matrix<bigrational> V;
	H->get_Sknew(V);
	if(H->dim_Sk()-V.ncols()==0 && V.ncols()==0) {delete H; continue;}
	for (int p=2; p<=pstop; p=next_prime(p)) 
	    if (is_prime(p)) {
		out<<"T["<<N<<","<<p<<"]=";
		Matrix<bigrational> T;
		H->Tn_On(p,V,T);
		ZZX f;
		CharPoly(T,f);
		Display(f,1,1,out);
		out<<";"<<endl;
	    }
	out<<endl;
	delete H;
    }
    out<<"}"<<endl;
}

void 
Tables::cpg1() const
{
    cout << endl << endl
	 << "======== charpoly's on S_k(Gamma_1(N)) ========="<<endl<<endl;
    int Nstart = get_int("N start = ");
    int Nstop = get_int("N stop = ");
    int pstop = get_int("pstop = ");
    int k = get_int("k = ");

    char fname[1024];
    sprintf(fname, "charpoly_s%dg1_T2-%d_N%d-%d.gp", k, pstop, Nstart, Nstop);

    ofstream out(userdata_write(fname));
    if(!out) {
	mesg("cpg1: can't open ",fname);
	die("cpg1: file open error.");
    }
    out<<"\\\\ " << fname << endl
       <<"\\\\ This is a table of characteristic polynomials of the" << endl
       <<"\\\\ Hecke operators T_p acting on the space S_"<<k<<"(Gamma_1(N)) "<<endl
       <<"\\\\ of weight "<<k<<" cusp forms for Gamma_1(N)."<<endl
       <<"\\\\ " << signature_stamp() << endl
       <<"\\\\ " << date_stamp() << endl
       <<endl<<endl;
    out<<"{"<<endl
       <<"T=matrix("<<Nstop<<","<<pstop<<",m,n,0);"<<endl;
    for (int N=Nstart; N<=Nstop; N++) {
	processing(N);
	Character chi(N,-1,save) ;
	H1NChi<bigrational>* H=new H1NChi<bigrational>(chi,k,save);
	if(H->dim_Sk()==0) { delete H; continue; }
	for (int p=2; p<=pstop; p=next_prime(p)) 
	    if (is_prime(p)) {
		out<<"T["<<N<<","<<p<<"]=";
		Matrix<bigrational> T;
		H->Tn_Sk(p,T);
		ZZX f;
		CharPoly(T,f);
		Display(f,1,1,out);
		out<<";"<<endl;
	    }
	out<<endl;
	delete H;
    }
    out<<"}"<<endl;
}

void 
Tables::cpg1new() const
{
    cout << endl << endl
	 << "======== charpoly's on S_k(Gamma_1(N))^new ========="<<endl<<endl;
    int Nstart = get_int("N start = ");
    int Nstop = get_int("N stop = ");
    int pstop = get_int("pstop = ");
    int k = get_int("k = ");

    char fname[1024];
    sprintf(fname, "charpoly_s%dg1new_T2-%d_N%d-%d.gp", k, pstop, Nstart, Nstop);

    ofstream out(userdata_write(fname));
    if(!out) {
	mesg("cpg1new: can't open ",fname);
	die("cpg1new: file open error.");
    }
    out<<"\\\\ " << fname << endl
       <<"\\\\ This is a table of characteristic polynomials of the Hecke" << endl
       <<"\\\\ operators T_p acting on the spaces S_"<<k<<"(Gamma_1(N))^new "<<endl
       <<"\\\\ of weight "<<k<<" cusp forms for Gamma_1(N)."<<endl
       <<"\\\\ " << signature_stamp() << endl
       <<"\\\\ " << date_stamp() << endl
       <<endl<<endl;
    out<<"{"<<endl
       <<"T=matrix("<<Nstop<<","<<pstop<<",m,n,0);"<<endl;
    for (int N=Nstart; N<=Nstop; N++) {
	processing(N);
	Character chi(N,-1,save) ;
	H1NChi<bigrational>* H=new H1NChi<bigrational>(chi,k,save);
	if(H->dim_Sk()==0) { delete H; continue; }
	Matrix<bigrational> V;
	H->get_Sknew(V);
	if(H->dim_Sk()-V.ncols()==0 && V.ncols()==0) {delete H; continue;}
	for (int p=2; p<=pstop; p=next_prime(p)) 
	    if (is_prime(p)) {
		out<<"T["<<N<<","<<p<<"]=";
		Matrix<bigrational> T;
		H->Tn_On(p,V,T);
		ZZX f;
		CharPoly(T,f);
		Display(f,1,1,out);
		out<<";"<<endl;
	    }
	out<<endl;
	delete H;
    }
    out<<"}"<<endl;
}


/////////////////////////////////////////
// GENERATING THE HECKE ALGEBRA
/////////////////////////////////////////

void 
Tables::heckegens_skg0() const
{
    cout << "heckegens_skg0: Compute a list of Hecke operators " << endl
	 << "which are guaranteed to generate the Hecke algebra " << endl
	 << "associated to S_k(Gamma_0(N)) as a Z-module."<<endl;
    
    int Nstart = get_int("N start = ");
    int Nstop = get_int("N stop = ");
    int k = get_int("k = ");
    char fname[1024];
    sprintf(fname,"Tn_s%dg0_%d-%d.gp",k,Nstart,Nstop);
    ofstream os(userdata_write(fname));
    if(!os) { cout << "heckegens_skg0: ERROR opening output file."<<endl; return; }
    os << "\\\\ " << fname << endl
       << "\\\\ This is a PARI readable set of Z-module generators for the Hecke"<<endl
       << "\\\\ algebra associated to for S_"<<k<<"(Gamma_0(N)) for N "<<endl
       << "\\\\ in the range:  " << Nstart << " <= N <= " << Nstop << "." << endl
       << "\\\\ The number of T_n computed is sufficient to generate the Hecke algebra."<<endl
       << "\\\\ William Stein (was@math.berkeley.edu)" << endl << endl;
    
    for(int N=Nstart; N<=Nstop; N++) {
	processing(N);
	Character chi (N,0,save) ;
	H1NChi<bigrational>* H=new H1NChi<bigrational>(chi,k,save);
	
	ModularFormulas formulas;
	int maxn = k*formulas.idxG0(N)/12;	// Sturm's bound on the number needed.
	if(H->dim_Sk()==0) { delete H; continue; }
	os<<"T["<<N<<"]=[";
	for(int n=1;n<=maxn;n++) {
	    Matrix<bigrational> T;
	    H->Tn_Sk(n,T);
	    Display(T,2,os);
	    if(n<maxn)os<<",";
	}
	os<<"];"<<endl<<endl;
    }
}


////////////////////////////////////////////
//  EXCEPTIONAL CASE OF SERRE'S CONJECTURE
////////////////////////////////////////////

void
Tables::exceptional(int N) const
{
#ifdef INST_ALGNUM
    // 0. [Preconditions] Verify that N is a prime = 1 (mod 12).
    if(N%12 != 1) {
	cout << "The input to the exceptional function must be a prime = 1 (mod 12)."<<endl;
	return;
    }

    // 1. [Check class number condition] Let h <-- class number of Q(sqrt(-3*N))
    bigint h = classno(-3*N);
    cout << "classno(-3N) = " << h << endl;
    int dummy;			// ignored
    int o3 = ordp(3,N-1,dummy);
    cout << "ord_3(N-1) = ord_3("<<N-1<<") = "<<o3<<endl;
    if(h % 3 == 0) {
	cout << "Jones implies nothing about rho_"<<N<<"."<<endl;
	if(N<1000) {
	    cout << "The explicit computations in his thesis" << endl
		 << "imply that rho_"<<N<<" is not exceptional."<<endl;
	    cout << "            d("<<N<<") = 2"<<endl;
	} else {
	    cout << "Further work must be done to determine d("<<N<<")"<<endl;
	    return;
	}
    }
    
    // 2. [Check simple divisibility condition]
    cout << "Jones implies that rho_"<<N<<" is exceptional."<<endl;
    // if N-1 is exactly divisible by either 3
    if (o3==1) {
	cout << "But since 3 exactly divides N-1="<<N-1<<" we conclude that"<<endl
	     << "               d("<<N<<") = 6"<<endl;
	return;
    }
    
    // 3. [Use modular symbols to determine the correct character]
    // Let 3, 3^2, ..., 3^a be the powers of 3 dividing N-1.
    // In this step we try to rule out small powers of 3, until hopefully
    // we are left with only 3^a.  A power 3^i can be ruled out by
    // showing that for the character of degree 2*3^i some Hecke
    // operator T_p with p=2 (mod 3) has nonzero determinent modulo
    // any prime lieing over 3. 

    Array<int> degs(1); 
    degs[0]=2;
    int i=0;
    bool all_elim=true;
    InitFieldCharmod(3);
    while(1) {
	degs[0] *= 3;
	i++;
	if(i==o3) {
	    cout << "All nonmaximal characters have been analyzed."<<endl;
	    if(all_elim) 
		cout << "Thus               d("<<N<<") ="<< degs[0]<<endl;
	    else
		cout << "Not all lower degrees could be eliminated."<<endl;
	    return;
	}
	cout << endl << "=====> Analyzing degree " << degs[0] << " <=====" << endl;
	// chi <-- character of degree degs.
	InitFieldAlgnumCyclo(degs[0]);
	Character chi(N,1,save,&degs) ;
	// M <-- M_2(Gamma_0(N),chi).
	H1NChi<algnum>* M=new H1NChi<algnum>(chi,2,save);
	Matrix<algnum> T;
	int p, stop_val=23;
	for(p=2; p<stop_val; ) {
	    M->Tn_Sk(p,T);		// getting action on Mk is faster and probably won't hurt.
	    // The algnum field is the cyclotomic field Q(zeta_n) where
            // n = 2*3^i.  The n-th cyclotomic polynomial factors, modulo 3,
            // as  (x+1)^phi(n).  Thus we must
            //     (a) replace the indeterminate in T by Mod(2,3) to obtain Tmod := T mod 3.
            //     (b) compute the determinant of Tmod (or maybe just the rank!)
            // Note: In the unlikely event that (a) is not possible because there are
            // "3's in the denominator" we terminate, saying so, and think of a 
            // better algorithm.
            //     (c) If det(Tmod) is not 0 then the representation can not
            //         arise from this space.
//	    cout << "T_"<<p<<" = "<<T<<endl;
	    bool canreduce=true;
	    for(int r=1;r<=T.nrows();r++)
		for(int c=1;c<=T.ncols();c++)
		    if(!can_reduce_charmod(T(r,c))) {
			canreduce=false;
			r=T.nrows()+1; 
			c=T.ncols()+1;
		    }

	    if(!canreduce) {
		cout << "Matrix rep of T_"<<p<<" can not be reduced mod 3, so SKIPPING."<<endl;
	    } else {
		charmod a(-1);
		Matrix<charmod> Tmod(T.nrows(),T.ncols());
		for(int r=1;r<=T.nrows();r++)
		    for(int c=1;c<=T.ncols();c++)
			Tmod(r,c)=to_charmod(T(r,c),a);
		//	cout << "Tmod = " << Tmod<<endl;
		// Compute rank of Tmod. 
		if(Tmod.is_invertible()) {
		    cout << "det(T_"<<p<<") =/= 0 (mod 3) "
			 << "thus d("<<N<<")=/="<<degs[0]<<endl;
		    break;
		} else {
		cout << "det(T_"<<p<<") = 0 (mod 3)"<<endl;
		}
	    }
	    
	    do {
		p=next_prime(p);
	    } while(p%3!=2);
	}
	if(p>=stop_val) {
	    cout << "Couldn't eliminate degree " << degs[0] << " "
		  << "hence d("<<N<<") <=?= " << degs[0]<<"."<<endl<<endl;
	    all_elim=false;
	} 
    }
#else
    cout << "Algnum is compiled into hecke."<<endl;
#endif
}

void 
Tables::go_non_exceptional() const
{
    int Nstart=get_int("Start N = ");
    int Nstop=get_int("Stop N = ");
    int N=Nstart-1;
    while(1) {
	do {
	    N=next_prime(N);
	} while((N-1)%12 != 0 ||  (classno(-3*N)%3 !=0));
	if(N>Nstop) break;
	non_exceptional(N);
    }
}

void
Tables::non_exceptional(int N) const
{
    cout << endl << "N = " << N <<", ";
    // 0. [Preconditions] Verify that N is a prime = 1 (mod 12).
    if(N%12 != 1) {
	cout << "The input to the exceptional function must be a prime = 1 (mod 12)."<<endl;
	return;
    }

    // 1. [Check class number condition] Let h <-- class number of Q(sqrt(-3*N))
    bigint h = classno(-3*N);
    cout << "classno(-3N) = " << h <<endl;
    if(h % 3 != 0) {
	cout << " Exceptional."<<endl;
	return;
    }
    // 3. [Use modular symbols to perhaps rule out the occurence of rho_N]

    Array<int> degs(1); 
    degs[0]=2;
    InitFieldCharmod(3);
    
    // chi <-- character of degree 2. 
    Character chi(N,1,save,&degs) ;
    // M <-- M_2(Gamma_0(N),chi).
    H1NChi<bigrational>* M=new H1NChi<bigrational>(chi,2,save);
    Matrix<bigrational> T;
    int p, stop_val=11;
    for(p=2; p<=stop_val; ) {
	M->Tn_Mk(p,T);		// getting action on Mk is faster and probably won't hurt.
	bool canreduce=true;
	for(int r=1;r<=T.nrows();r++)
	    for(int c=1;c<=T.ncols();c++)
		if(denominator(T(r,c))%3==0) {
		    canreduce=false;
		    r=T.nrows()+1; 
		    c=T.ncols()+1;
		}
	
	if(!canreduce) {
	    cout << "Matrix rep of T_"<<p<<" can not be reduced mod 3, so SKIPPING."<<endl;
	} else {
	    charmod a(-1);
	    Matrix<charmod> Tmod(T.nrows(),T.ncols());
	    for(int r=1;r<=T.nrows();r++)
		for(int c=1;c<=T.ncols();c++)
		    Tmod(r,c)=charmod(numerator(T(r,c)))/charmod(denominator(T(r,c)));
	    // Compute rank of Tmod. 
	    if(Tmod.is_invertible()) {
		cout << "det(T_"<<p<<") =/= 0 (mod 3) "
		     << "thus d("<<N<<")=/="<<degs[0]<<endl;
		break;
	    } else {
		cout << "det(T_"<<p<<") = 0 (mod 3)"<<endl;
	    }
	}
	
	do {
	    p=next_prime(p);
	} while(p%3!=2);
    }
    if(p>=stop_val) 
	cout << "The form may occur in this space."<<endl;
    else
	cout << "*** The form might NOT occur here."<<endl;
}



/////////////////////////////////////////
// MENU
/////////////////////////////////////////

void Tables::help() const
{
    cout << "Table making functions:"<<endl
	 << "  ang0:       basis of newforms for S_k(Gamma_0(N))"<<endl
	 << "  apg1:       Hecke eigenvalues for S_k(Gamma_1(N))"<<endl
	 << "  apg0mod:    Hecke eigenvalues for S_k(Gamma_0(N);Fp)"<<endl
	 << "  ss:         list of eigenforms wrt supersingular basis"<<endl
	 << endl

	 << "  bsd:        Birch and Swinnerton-Dyer data"<<endl
	 << endl

	 << "  cpg0:       charpoly on S_k(Gamma_0(N))"<<endl
	 << "  cpg0new:    charpoly on S_k(Gamma_0(N))^new"<<endl
	 << "  cpg1:       charpoly on S_k(Gamma_1(N))"<<endl
	 << "  cpg1new:    charpoly on S_k(Gamma_1(N))^new"<<endl
	 << endl

	 << "  dimg0:      dim S_k(Gamma_0(N))."<<endl
	 << "  dimg0new:   dim S_k^{new}(Gamma_0(N))."<<endl
	 << "  dimg1:      dim S_k(Gamma_1(N))."<<endl
	 << "  dimg1new:   dim S_k^{new}(Gamma_1(N))."<<endl
	 << endl

	 << "  disc        disc(T)"<<endl
	 << "  discnew     disc(T|new)"<<endl
         << "  discnorm    disc(Ttilde|new)"<<endl
	 << "  disct1tg    disc(Z{T1,...,Tg})"<<endl
	 << endl
	 << "  exc:        exceptional case of Serre."<<endl
	 << "  nonexc:     non-exceptional case of Serre."<<endl
	 << endl

	 << "  tn:         generators for the Hecke algebra"<<endl
	 << "  x:          Exit"<<endl<<endl;
}

void
Tables::interactive() const
{ 
    cout << "Table making functions"<<endl
	 << endl
	 << "Type ? for help."<<endl<<endl;
    while(1) {
	char buf[256];
	get_str(buf, 256, "\n\nTable? ");
	if (strcmp(buf,"x")==0 || strcmp(buf,"q")==0) break;
	if (strcmp(buf,"?")==0)  
	    help(); 
	else if (strcmp(buf, "ang0")==0) 
	    anlist_g0new();
	else if (strcmp(buf, "apg1")==0)
	    aplist_g1new();
	else if (strcmp(buf, "apg0mod")==0)
	    aplist_g0new_modp();
	else if (strcmp(buf, "ss")==0)
	    supersingular();
	else if (strcmp(buf, "bsd")==0)
	    bsd();

	else if (strcmp(buf, "cpg0")==0)
	    cpg0();
	else if (strcmp(buf, "cpg0new")==0)
	    cpg0new();
	else if (strcmp(buf, "cpg1")==0)
	    cpg1();
	else if (strcmp(buf, "cpg1new")==0)
	    cpg1new();
	
	else if (strcmp(buf, "dimg0")==0)
	    dimg0();
	else if (strcmp(buf, "dimg0new")==0)
	    dimg0new();
	else if (strcmp(buf, "dimg1")==0)
	    dimg1(); 
	else if (strcmp(buf, "dimg1new")==0)
	    dimg1new();

	else if (strcmp(buf,"disc")==0) 
	    heckedisc_Sk();
	else if (strcmp(buf,"discnew")==0) 
	    heckedisc_Sknew();
	else if (strcmp(buf,"discnorm")==0) 
	    heckedisc_normalization_Sknew();
	else if (strcmp(buf,"disct1tg")==0) 
	    heckedisc_T1_Tg();

	else if(strcmp(buf,"exc")==0) 
	    exceptional(get_int("N = "));
	else if(strcmp(buf,"nonexc")==0) 
	    go_non_exceptional();
	else if(strcmp(buf,"tn")==0) 
	    heckegens_skg0(); 
    }
}

