/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

// vector.cc 
#ifndef VECTOR_H
#define VECTOR_H

#include "vector.h"
#include "NTL/ZZXFactoring.h"
#include "NTL/ZZX.h"
#include "system.h"
#include "my_ntl.h"

typedef polynomial<bigrational> polbigrat;

#include<stdlib.h>
template <class T>
void 
Vector<T>::random(int max)
{
    for(int i=1;i<=dim;i++)
	(*this)[i]=T(::random()%max);
}


template <class T>
void 
Vector<T>::reorder(const Vector<int>& s) 
{
    if(s.size()!=size())
	die("Vector::reorder -- invalid permutation length.");
    Vector<T> t(size());
    for(int i=1; i<=size(); i++)
	t[i]=(*this)[s[i]];
    *this = t;
}


/////////// polmul ////////////////////////////////
bool _polmul(Vector<bigcomplex>&,
    const Vector<bigcomplex>&, const Vector<bigcomplex>&) { return false; }
bool _polmul(Vector<polbigrat>&,
    const Vector<polbigrat>&, const Vector<polbigrat>&) { return false; }
bool _polmul(Vector<bigint>&,
    const Vector<bigint>&, const Vector<bigint>&) { return false; }
bool _polmul(Vector<int>&,
    const Vector<int>&, const Vector<int>&) { return false; }
bool _polmul(Vector<charmod>& fg,
    const Vector<charmod>& f, const Vector<charmod>& g) {
    zz_pX x, y;
    my_to_ntl(x,f); 
    my_to_ntl(y,g);
    ntl_to_my(fg, x*y);
    return true;
}
bool _polmul(Vector<intmod>&,
    const Vector<intmod>&, const Vector<intmod>&) { return false; }
bool _polmul(Vector<algnum>&,
    const Vector<algnum>&, const Vector<algnum>&) { return false; }
bool _polmul(Vector<ffield_element>&,
    const Vector<ffield_element>&, const Vector<ffield_element>&) { return false; }
bool _polmul(Vector<bigrational>& ab,
    const Vector<bigrational>& a, const Vector<bigrational>& b) 
{
    ZZX  A, B;
    bigint da, db;
    to_ZZX(A,da,a);   // A = da*a
    to_ZZX(B,db,b);   // B = db*b
    ZZX C=A*B;
    to_bigrationalvec(ab, C);	// C = A*B = da*a*db*b = (da*db)*(a*b).
    ab/=(da*db);
    return true;
}



template<class T>
Vector<T> polmul(const Vector<T>& f, const Vector<T>& g)
{
    Vector<T> fg;
    if(_polmul(fg,f,g)) 
	return fg ;
    else {
	mesg("[slow polmul");
	Vector<T> fg;
	fg.init(f.size()+g.size()-1);
	for(int i=1; i<=f.size(); i++)
	    for(int j=1; j<=g.size(); j++)
		fg[i+j-1]+=f[i]*g[j];
	mesg("]");
	return fg;
    }
}



/////////// polmod ///////////////////////////////
bool _polmod(Vector<polbigrat>&,
	     const Vector<polbigrat>&, const Vector<polbigrat>&) { return false; }
bool _polmod(Vector<bigint>&,
	     const Vector<bigint>&, const Vector<bigint>&) { return false; }
bool _polmod(Vector<int>&,
	     const Vector<int>&, const Vector<int>&) { return false; }
bool _polmod(Vector<bigcomplex>&,
	     const Vector<bigcomplex>&, const Vector<bigcomplex>&) { return false; }
bool _polmod(Vector<charmod>& fmodg,
	     const Vector<charmod>& f, const Vector<charmod>& g) 
{
    zz_pX x, y;
    my_to_ntl(x,f); 
    my_to_ntl(y,g);
    ntl_to_my(fmodg, x%y);
    return true;
}

bool _polmod(Vector<intmod>&,
	     const Vector<intmod>&, const Vector<intmod>&) { return false; }
bool _polmod(Vector<algnum>&,
	     const Vector<algnum>&, const Vector<algnum>&) { return false; }
bool _polmod(Vector<ffield_element>&,
	     const Vector<ffield_element>&, const Vector<ffield_element>&) { return false; }
bool _polmod(Vector<bigrational>& r,
	     const Vector<bigrational>& a, const Vector<bigrational>& b) 
{
    ZZX  A, B;
    bigint da, db;
    to_ZZX(A,da,a);   // A = da*a
    to_ZZX(B,db,b);
    
    if(db!=1) {
	mesg("[polmod(bigrational)-resorting to slow...]");
	return false; // can't use NTL, so use slow algorithm.
    }
    ZZX C=A%B;
    to_bigrationalvec(r, C);
    r /= da;
    return true;
}

template<class T>
Vector<T> polmod(const Vector<T>& f, const Vector<T>& g)
{
    Vector<T> r;
    if(_polmod(r, f,g)) 
	return r;
    else {
	mesg("[slow polmod");
	if(f.poldeg()<0 || g.poldeg()==0) {
	    Vector<T> z(0); return z;
	}
	if(g.poldeg()<0) { cerr << "polmod: divide by zero."<<endl; die(); }
	int m=f.poldeg(), n=g.poldeg(); 
	Vector<T> ff=f;
	for(int k=m-n+1; k>=1; k--) {
	    T q = ff[n+k]/g[n+1];
	    for(int j=n+k; j>=k; j--)
		ff[j] -= q*g[j-k+1];
	}
	ff.resize(g.size()-1);
	mesg("]");
	return ff;
    }
}





/////////// pol_xgcd ///////////////////////////////
bool _polxgcd(Vector<polbigrat>&, Vector<polbigrat>&, Vector<polbigrat>&, 
	      const Vector<polbigrat>&,
	      const Vector<polbigrat>&) { return false; }
bool _polxgcd(Vector<bigint>&, Vector<bigint>&,  Vector<bigint>&, 
             const Vector<bigint>&, const Vector<bigint>&) { return false; }
bool _polxgcd(Vector<int>&, Vector<int>&, Vector<int>&, 
             const Vector<int>&, const Vector<int>&) { return false; }
bool _polxgcd(Vector<bigcomplex>&, Vector<bigcomplex>&,Vector<bigcomplex>&,
             const Vector<bigcomplex>&, const Vector<bigcomplex>&) { return false; }
bool _polxgcd(Vector<charmod>& x, Vector<charmod>& y,Vector<charmod>& gcd,
             const Vector<charmod>& a, const Vector<charmod>& b) 
{
    mesgln("ntl xgcd.");
    zz_pX f, g, s, t, d;
    my_to_ntl(f, a); my_to_ntl(g, b);
    XGCD(d, s, t, f, g);
    ntl_to_my(gcd,d);
    ntl_to_my(x,s);
    ntl_to_my(y,t);
    return true;
}

bool _polxgcd(Vector<intmod>&, Vector<intmod>&,  Vector<intmod>&, 
             const Vector<intmod>&, const Vector<intmod>&) { return false; }
bool _polxgcd(Vector<algnum>&, Vector<algnum>&,  Vector<algnum>&, 
             const Vector<algnum>&, const Vector<algnum>&) { return false; }
bool _polxgcd(Vector<ffield_element>&, Vector<ffield_element>&, Vector<ffield_element>&,
             const Vector<ffield_element>&, const Vector<ffield_element>&) { return false; }
bool _polxgcd(Vector<bigrational>& x, Vector<bigrational>& y, Vector<bigrational>& gcd,
             const Vector<bigrational>& a, const Vector<bigrational>& b) 
{
    // use NTL
    mesg("[NTL XGCD");
    ZZX  A, B, X, Y;
    bigint da, db;
    to_ZZX(A,da,a);
    to_ZZX(B,db,b);
    ZZ res;
    XGCD(res, X, Y, A, B, 0);
	 //  using nondeterministic, but OK since checking answer below!!
    if(res==0) 
	die("_polxgcd(bigrational): a and b are not coprime!");
    else {
	gcd.resize(1);
	gcd[1]=1;
    }

    to_bigrationalvec(x, X);
    to_bigrationalvec(y, Y);

    // now x*da*a + y*db*b = res, so 
    //     (x*da/res)*a + (y*db)/res * b = 1
    x *= bigrational(da) / to_bigint(res);
    y *= bigrational(db) / to_bigint(res);

    // CHECK
    Vector<bigrational> check=poladd(polmul(x,a),polmul(y,b));
    if(check.poldeg()!=0 || check[1]!=1) {
	cout << "check = " << check << endl;
	die("pol_xgcd(bigrational) error!!!!!!!!!");
    }
    mesg("]\n");
    
    return true;
}

template<class T>
void pol_xgcd(Vector<T>& xx, Vector<T>& yy, Vector<T>& gcd,
	     const Vector<T>& aa, const Vector<T>& bb)
{
    if(!_polxgcd(xx,yy,gcd,aa,bb)) {
	mesg("[resorting to *slow* polxgcd ");
	Vector<T> a = aa, b = bb, x; 
	Vector<T> oldx; oldx.polconst(T(1));
	Vector<T> y=oldx, oldy, q, r;
	while(!b.iszero()) {
	    Vector<T> q; q=poldiv(a,b);
	    Vector<T> c; c=poladd(a,-polmul(q,b)); a=b; b=c;
	    Vector<T> newx=poladd(oldx,-polmul(q,x)); oldx=x; x=newx;
	    Vector<T> newy=poladd(oldy,-polmul(q,y)); oldy=y; y=newy;
	}
	xx=oldx; yy=oldy; gcd=a;
	if(gcd.poldeg()==0) {	// normalize constant to be = 1. 
	    xx = xx/gcd[1];
	    yy = yy/gcd[1];
	    gcd[1]=T(1);
	}   
	mesg("]\n");
    }
}

///////////////////////////////////////////////////


// most of these do nothing -- producing the empty factorization --
// which symbolizes failure. 

void Factor(VecFac<charmod>& fac, const Vector<charmod>& f) { 
    ntl_factor(fac, f);
}

void Factor(VecFac<intmod>& fac, const Vector<intmod>& f) { 
    ntl_factor(fac, f);
}

void Factor(VecFac<bigcomplex>& fac, const Vector<bigcomplex>& f) { 
  cerr<<"Warning: Factor bigcomplex not implimented!!!!!!!!!!!"<<endl; 
}

void Factor(VecFac<bigrational>& fac, const Vector<bigrational>& f) 
{
    ZZX g;
    bigint denom=1;
    to_ZZX (g, denom, f);
    vec_pair_ZZX_long ntlfac ;
    ZZ c;
    mesg("[ntl factor]");
    factor(c,ntlfac,g);
    fac.resize(ntlfac.length());
    for(int i=1; i<=ntlfac.length(); i++) {
	fac.exp(i)=ntlfac(i).b;
	to_bigrationalvec(fac.base(i),ntlfac(i).a);
    } 
    fac.base(1)/=denom;
}





/*******************************************************************************
 *******************************************************************************/

/* **** Factorization of Polynomials over a Number Field ****
   This algorithm is taken from section 3.6.2 of Cohen's book.

   Let K = Q(theta) be a number field, T in Q[X] the minimal monic polynomial
   of theta.  Let A(X) be a non-zero polynomial in K[X].  This algorithm finds
   a complete factorization of A in K[X].

 */ 

template class Vector<polbigrat>;	// instantiate.
typedef Vector<polbigrat> PolQX;

bigrational polcontent(const Vector<bigrational>& f)
{return (f.poldeg()<0)?0:1;}
algnum polcontent(const Vector<algnum>& f)
{return (f.poldeg()<0)?0:1;}
ffield_element polcontent(const Vector<ffield_element>& f)
{return (f.poldeg()<0)?0:1;}
charmod polcontent(const Vector<charmod>& f)
{return (f.poldeg()<0)?0:1;}
bigcomplex polcontent(const Vector<bigcomplex>& f)
{return (f.poldeg()<0)?bigcomplex(0):bigcomplex(1);}
intmod polcontent(const Vector<intmod>& f)
{return (f.poldeg()<0)?0:1;}


polbigrat 
polcontent(const Vector<polbigrat>& f) 
{
    if(f.poldeg()<0) return polbigrat(0);
    polbigrat g=f[1];
    for(int i=2;i<=f.size();i++)
	g=gcd(g,f[i]);
    return g;
}
int
polcontent(const Vector<int>& f) 
{
    if(f.poldeg()<0) return 0;
    int g=f[1];
    for(int i=2;i<=f.size();i++)
	g=gcd(g,f[i]);
    return g;
}
bigint
polcontent(const Vector<bigint>& f) 
{
    if(f.poldeg()<0) return 0;
    bigint g=f[1];
    for(int i=2;i<=f.size();i++)
	g=gcd(g,f[i]);
    return g;
}

polbigrat operator/(const polbigrat& f, const polbigrat& g) 
{
    polbigrat q,r;
    div_rem(q,r,f,g);
    if(r.degree()>=0) die("Error polbigrat operator/, g doesn't divide f.");
    return q;
}

bool operator<(const polbigrat& f, const polbigrat& g) {
    die("terminate operator< polbigrat");
    return true;
}
bool operator>(const polbigrat& f, const polbigrat& g) {
    die("terminate operator< polbigrat");
    return true;
}


// Pseudo-Division
// Algorithm 3.1.2 of Cohen: 
// Let A and B be two polynomials in T[X] with B\neq 0, and set
// m <-- deg(A), n<--deg(B), d<--l(B)=leading coeff(B).  Assume
// that m>=n.  This algorithm finds Q and R such that
// d^{m-n+1}A = BQ+R and deg(R)<deg(B).

template<class T>
void pol_pseudodiv(Vector<T>& Q, Vector<T>& R,
		   const Vector<T>& A, const Vector<T>& B)
{
    if(B.poldeg()<0) die("Vector<T>::pseudo_div called with B=0.");
    // 1. [Initialize] Set R<--A, Q<--0, e<--m-n+1.
    int m=A.poldeg(), n=B.poldeg();
    int e=m-n+1;
    R=A;
    Q.polconst(T(0));
    T d=B.pol_l();
    
    // 2. [Finished] If deg(R)<deg(B) then set q<--d^e, Q<--qQ, R<--qR
    // and terminate the algorithm.
    while(R.poldeg() >= B.poldeg()) {
	// 3. [Find coefficient] Set
	//           S <-- l(R) X^{deg(R)-deg(B)}
	//    then Q<--dQ+S, R<--dR-SB, e=e-1 and go to step 2.
	Vector<T> S; S.init(R.poldeg()-B.poldeg()+1);
	S.polcoeff(R.poldeg()-B.poldeg()) = R.pol_l();
	Q = poladd(d*Q,S);
	R = poladd(d*R, polmul(T(-1)*S,B));
	e--;
    } ;
    
    T q;
    power(q, d, e);
    Q = q*Q;
    R = q*R;
}

// Sub-Resultant
// Algorithm 3.3.7 of Cohen: Given two polynomial A and B with coefficients
// in the unique factorization domain T, this algorithm computes the resultant 
// of A and B.

template<class T>
T pol_resultant(Vector<T> A, Vector<T> B)
{
    // 1. [Initializations and reductions] If A=0 or B=0, output 0 and terminate the
    // algorithm.  Otherwise, set a<--cont(A), b<--cont(B), A<--A/a, B<--B/b,
    // g<--1, h<--1, s<--1, and t<--a^{deg(B)}b^{deg(A)}.  Finally, if deg(A)<deg(B) 
    // exchange A and B and if in addition deg(A) and deg(B) are odd set s <-- -1.

//    cout<<"pol_resultant of ";A.polprint();cout<<"  and  ";B.polprint();cout<<endl;  
    if(A.poldeg()<0 || B.poldeg()<0) return T(0);
    T a = polcontent(A), b = polcontent(B);
//    cout<<"polcontent(A)="<<a<<", polcontent(B)="<<b<<endl;
    A /= a; 
    B /= b;

    T g(1), h(1), s(1);
    T t1, t2; power(t1,a,B.poldeg()); power(t2,b,A.poldeg());
    T t = t1*t2;
    if(A.poldeg()<B.poldeg()) {
	Vector<T> temp=A; A=B; B=temp;
    }
    if(A.poldeg()%2!=0 && B.poldeg()%2!=0) s=-s;

    do {
	// 2. [Pseudo division] Set d<--deg(A)-deg(B).  If deg(A) and deg(B) are odd, set
	// s <- -s.  Finally, compute R such that l(B)^{d+1}A = BQ+R using Algorithm 3.1.2
	// (pol_pseudodiv). 
	int d = A.poldeg()-B.poldeg();
	if(A.poldeg()%2!=0 && B.poldeg()%2!=0) s=-s;
	Vector<T> Q, R;
	pol_pseudodiv(Q,R,A,B);
	
	// 3. [Reduce remainder] Set A<--B and B<--R/(gh^d).
//	cout<<"[Reduce remainder] R=";R.polprint();cout<<endl;
	A = B;
	T hd; power(hd,h,d);
	B = R;
//	cout<<"B=";B.polprint();cout<<endl<<"hd="<<hd<<", g="<<g<<endl;
	B /= (g*hd);
	
	// 4. [Finished?] Set g<--l(A), h<--h^{1-d}g^d. If deg(B)>0 go to step 2, otherwise
	// set h<--h^{1-deg(A)}l(B)^deg(A) output s*t*h and terminate the algorithm.
	g=A.pol_l();
	T t1, t2;
	if(d>0) {
	    power(t1,h,d-1);
	    power(t2,g,d);
//	    cout<<"t1="<<t1<<", t2="<<t2<<endl;
	    h = t2/t1;
	} else {
	    h = T(1)/h;
	}
    } while (B.poldeg()>0);
    if(A.poldeg()>0) {
	power(t1,h,A.poldeg()-1); 
	power(t2,B.pol_l(),A.poldeg());
//	cout<<"outside t1="<<t1<<", t2="<<t2<<endl;
	h = t2/t1;
    } else {
	h = T(1)/h;
    }
    return s*t*h;
}


algnum gcd(algnum x, algnum y) {
    if(x==algnum(0) && y==algnum(0)) return 0;
    return algnum(1);
}
bigrational gcd(bigrational x, bigrational y) {
    if(x==0 && y==0) return 0;
    return 1;
}
charmod gcd(charmod x, charmod y) {
    if(x==0 && y==0) return 0;
    return charmod(1);
}
bigcomplex gcd(bigcomplex x, bigcomplex y) {
    if(x==bigcomplex(0) && y==bigcomplex(0)) return bigcomplex(0);
    return bigcomplex(1);
}

intmod gcd(intmod x, intmod y) {
    if(x==0 && y==0) return 0;
    return intmod(1);
}
ffield_element gcd(ffield_element x, ffield_element y) {
    if(x==0 && y==0) return 0;
    return ffield_element(1);
}


// Sub-Resultant GCD.  Algoirthm 3.3.1 from Cohen.  Given two polynomials A and B
// with coefficients in T, this algorithm computes a GCD of A and B, using only 
// operations in T. 
template<class T>
Vector<T> polgcd(Vector<T> A, Vector<T> B)
{
//    cout<<"polgcd(";A.polprint();cout<<", ";B.polprint();cout<<")"<<endl;
    // 1. [Initializations and reductions] If deg(B)>deg(A) exchange A and B.  Now
    // if B=0, output A and terminate the algorithm, otherwise, set a<--cont(A),
    // b<--cont(B), d<--gcd(a,b), A<--A/a, B<--B/b, g<--1, h<--1 and c<--0.
    if(B.poldeg()>A.poldeg()) {
	Vector<T> temp=A; A=B; B=temp;
    }
    if(B.poldeg()<0) return A;
    T a = polcontent(A), b = polcontent(B);
    T d = gcd(a,b);
    A /= a; 
    B /= b;
    T g(1), h(1);
    int c=0;

    while(1) {
	// 2. [Pseudo division] Set del<--deg(A)-deg(B). Using Algorithm 3.1.2, compute 
	// R such that l(B)^{del+1}A=BQ+R.  If R=0 go to step 4.  If deg(R)=0, set
	// B<-1 and go to step 4.
	int del=A.poldeg()-B.poldeg();
	Vector<T> Q,R;
	pol_pseudodiv(Q,R,A,B);
	if(R.poldeg()<0) break;
	if(R.poldeg()==0) { B.polconst(T(1)); break; }
	
	// 3. [Reduce remainder] Set A<--B, B<--R/(gh^del) and c<--c+1.  If c<10, set
	// g<--l(A), h<--h^(1-d)g^del and go to step 2.  Otherwise, set A<--A/cont(A),
	// B<--B/cont(B), g<--1, h<--1, c<--0 and go to step 2. 
	A = B;
	T hd; power(hd,h,del);
	B = R;
	B /= (g*hd);
	c++;
	if(c<10) {
	    g=A.pol_l();
	    T t1, t2;
	    if(del>0) {
		power(t1,h,del-1);
		power(t2,g,del);
		h = t2/t1;
	    } else {
		h = T(1)/h;
	    }
	} else {
	    A /= polcontent(A);
	    B /= polcontent(B);
	    g = T(1);
	    h = T(1);
	    c = 0;
	}
    }
    
    // 4. [Terminate] Output d*B/cont(B) and terminate the algorithm.
    B = d*(B/polcontent(B));
//    cout << "gcd ans is "; B.polprint(); cout << endl;
    return B;
}

template<class T>
T Vector<T>::polsubst(const T& x) const
{
    T y(0);
    for (int i=size(); i>=1; i--) {
	y += (*this)[i];
	if (i>1) y *= x;
    }
    return y;
}

template<class T>
Vector<T> Vector<T>::pol_translate(const T& k) const
{
    const Vector<T>& f=*this;
    Vector<T> ans;
    ans.polconst(f.polgetcoeff(0));
    Vector<T> lin; lin.init(2);
    lin.polcoeff(0)=k;
    lin.polcoeff(1)=T(1);
    Vector<T> t=lin;
    
    for(int i=1; i<=f.poldeg(); i++) {
	Vector<T> s=f.polgetcoeff(i)*t;
	ans = poladd(ans,s);
	t = polmul(t,lin);
    }
    return ans;
}

// Input: G(X,Y) = polynomial in Y with coefficients in Q[X]
//               = u_n(X)Y^n + u_{n-1}(X)Y^{n-1} + ... + u_1(X)Y + u_0(X).
// Output: G(X,Y-kX) (translate the Y by -kX).
PolQX get_Gk(const PolQX& G,  const bigrational& k)
{
//    cout << "get_Gk(G,"<<k<<"), G="; G.polprint(); cout<<endl;
    if(k==0) return G;
    bigrational v[2];
    v[0]=0;
    v[1]=-k;
    polbigrat t(v,1);
//    cout << "get_Gk: t = " << t << endl;

    PolQX p = G.pol_translate(t);
//    cout << "G.pol_translate(t) = "; p.polprint(); cout<<endl;

    return G.pol_translate(t);
}

// switchXY:
// Input: G(X,Y) = polynomial in Y with coefficients in Q[X]
//               = u_n(X)Y^n + u_{n-1}(X)Y^{n-1} + ... + u_1(X)Y + u_0(X).
// Output: H(X,Y) = G(Y,X) = u_n(Y)X^n + u_{n-1}(Y)X^{n-1} + ... + u_1(Y)X + u_0(Y),
//               again a poly in Y with coefficients in Q[X].            
// Uses convert, which takes a poly u(X) and returns the corr. poly u(Y), whose
// coefficients are all constant.             
PolQX convert(polbigrat u)
{
//    cout<<"convert("<<u<<")=";
    PolQX U(u.degree()+1);
    for(int i=0;i<=u.degree();i++)
	U.polcoeff(i)=polbigrat(u[i]);
//    U.polprint();cout<<endl;
    return U;
}

PolQX switchXY(const PolQX& G)
{
//    cout<<"switchXY(";G.polprint();cout<<")"<<endl;
    PolQX H=convert(G.polgetcoeff(0));	// H = u_0(Y).
    bigrational v[2]; v[0]=0; v[1]=1;
    polbigrat x(v,1);		// x = X
    polbigrat xn=x;
    for(int i=1; i<=G.poldeg(); i++) {
	H=poladd(H,xn*convert(G.polgetcoeff(i)));
	xn*=x;
    }
//    cout<<"=";H.polprint();cout<<endl;
    return H;
}


// Input: f = (irreducible) polynomial i Q[x]
//        k = rational number
// Output: f(X+k*theta) where theta is a root of the irred_polynomial
//         which defines the current algebraic number field.
Vector<algnum> get_Ni(const Vector<bigrational>& f, const bigrational& k)
{
    Vector<algnum> Ni;
    Ni.polconst(f.polgetcoeff(0));
    Vector<algnum> lin; lin.init(2);
    lin.polcoeff(0).set_rootirredpoly(); lin.polcoeff(0)*=algnum(k);
    lin.polcoeff(1)=1;
    Vector<algnum> t=lin;
    
    for(int i=1; i<=f.poldeg(); i++) {
	Vector<algnum> s = algnum(f.polgetcoeff(i)) * t;
	Ni = poladd(Ni, s);
	t = polmul(t,lin);
    }

    return Ni;
}

bool pol_issqrfree(const Vector<bigrational>& f)
{
    ZZX g;
    g.SetMaxLength(f.size());
    bigint denom=1;
    for(int i=1;i<=f.size();i++) {
	denom=lcm(denom,denominator(f[i]));
    }
    for(int i=1;i<=f.size();i++) {
	SetCoeff(g,i-1,to_ZZ( numerator(f[i])*denom));
    }
//    cout<<"g="<<g<<endl<<"diff(g)="<<diff(g)<<endl<<"GCD()="<<GCD(g,diff(g))<<endl<<"deg="<<deg(GCD(g,diff(g)))<<endl;
    return deg(GCD(g,diff(g)))<1;
}

void Factor(VecFac<algnum>& fac, const Vector<algnum>& A) { 
    // 1. [Reduce to squarefree] Set U <- A/gcd(A,A') 
//cout<<"Factor ";A.polprint();cout<<endl;
    mesg("[factor algnum poly]");
    algnum lead=A.pol_l();
    Vector<algnum> g = polgcd(A,A.polderiv());
    g.pol_makemonic();
    Vector<algnum> U = poldiv(A,g);
//cout<<"U=";U.polprint();cout<<endl;

    // 2. [Initialize search] Let U(X) = sum_{0<=i<=m} u_i X^i in K[X] and
    //    write u_i=g_i(theta) for some polynomial g_i in Q[X].  Set
    //       G(X,Y) <-- sum_{0<=i<=m} g_i(Y)X^i in Q[X,Y] and k<--0.
    PolQX G;
    G.init(U.poldeg()+1);
    for(int i=1;i<=G.size();i++) 
	G[i] = U[i].lift();
//    cout<<"G=";G.polprint();cout<<endl;

    const polbigrat& irred=algnum::irred_polynomial();
    PolQX T(irred.degree()+1);
    for(int i=0;i<=irred.degree();i++)
	T.polcoeff(i)=polbigrat(irred[i]);
//    cout<<"T=";T.polprint();cout<<endl;

    bigrational k=-1;
    Vector<bigrational> N;
//    mesg("[find squareless norm]");
    do {
	k=k+1;

	// 3. [Search for squarefree norm] Using the sub-resultant algorithm
	//    over the UFD Q[X], compute N(X) <-- R_Y(T(Y),G(X-kY,Y)) where
	//    R_Y denotes the resultant with respect to the variable Y.  
	//    If N(X) is not squarefree, set k<--k+1.
	PolQX Gk=get_Gk(G,k);
//	cout<<"k="<<k<<endl;
//	cout<<"Gk=";Gk.polprint();cout<<endl;
	to_bigrationalvec(N,pol_resultant(T, switchXY(Gk)));
//	cout<<"N=";N.polprint();cout<<endl;
    } while (!pol_issqrfree(N));

//    cout << "N = " ; N.polprint();
//    cout << "\n about to factor N..."<<flush;
     
    // 4. [Factor norm] (Here N(X) is squarefree)  Let N=prod_{1<=i<=g} N_i
    //    be a factorization of N in Q[X].
    VecFac<bigrational> facN;
    Factor(facN, N);

//    cout << "done!"<<endl << "now to output answer..."<<endl;

    // 5. [Output factorization] For i=1,...,g set A_i(X)<--gcd(U(X),N_i(X+k(theta)))
    //    computed in K[X], output A_i and the exponent of A_i in A.
    fac.resize(facN.numfactors());
    // mesg("[gcd out algnum factors]");
    for(int i=1; i<=fac.numfactors(); i++) {
//	cout<<"facN("<<i<<")=("; facN.base(i).polprint(); cout<<")^"<<facN.exp(i)<<endl;
	if(facN.exp(i)!=1) die("Error in Factor algnum, facN.exp(i) != 1.");
	mesg("[N",i);mesg("]");
	Vector<algnum> Ni=get_Ni(facN.base(i),k); 
//	mesg("[polgcd(U,Ni)]");
	fac.base(i)=polgcd(U, Ni);
//	mesg("[makenonic]");
	(fac.base(i)).pol_makemonic();
	int deg=fac.base(i).poldeg();
	mesg("[deg=",deg);mesg("]");
	if(1<=deg && deg+1<fac.base(i).size())
	    fac.base(i).resize(deg+1);
	mesg("[pol_ord]");
	fac.exp(i) = A.pol_ord(fac.base(i));
//	mesg("\n");
    }
    mesg("\n");
    fac.base(1)*=lead;
}

/*******************************************************************************
 *******************************************************************************/

/*
void Factor(VecFac<algnum>& fac, const Vector<algnum>& f) { 
  cerr<<"Warning: Factor algnum only works for IRREDUCIBLE polys now."<<endl; 
  fac.resize(1);
  fac.exp(1)=1;
  fac.base(1)=f;
}
*/

void Factor(VecFac<ffield_element>& fac, const Vector<ffield_element>& f) { 
  cerr<<"Warning: Factor ffield_element not implimented"<<endl; }
void Factor(VecFac<int>& fac, const Vector<int>& f) { 
  cerr<<"Warning: Factor int not implimented"<<endl; }
void Factor(VecFac<polbigrat>& fac, const Vector<polbigrat>& f) { 
  cerr<<"Warning: Factor int not implimented"<<endl; }
void Factor(VecFac<bigint>& fac, const Vector<bigint>& f) { 
  cerr<<"Warning: Factor bigint not implimented"<<endl; }

template<class T>
void Vector<T>::polfactor(VecFac<T>& fac) const
{
  // unless T is bigrational, I don't really know what to do at this time. 
  fac.resize(0);
  Factor(fac, *this);
}

template<class T>
bool Vector<T>::pol_is_irreducible() const
{
    VecFac<T> fac;
    Factor(fac,*this);
    return fac.numfactors()==1 && fac.exp(1)==1;
}


template <class T> 
void Vector<T>::reverse() 
{
  for(int i=1; i<=dim/2; i++) {
    T t=(*this)[i];
    (*this)[i] = (*this)[dim-i+1];
    (*this)[dim-i+1] = t;
  }
}


template <class T> 
int Vector<T>::numnonzero() const 
{
  int num=0;
  for(int i=0; i<size(); i++) 
    if (vec[i] != T(0)) num++;
  return num ;
}


/*
template <class T> 
T Vector<T>::max() const
{
  if (dim == 0) 
    error ("Cannot take max of the empty vector.");
  T m = vec[0] ;
  for (int i=1; i<dim; i++)
    if (vec[i]>m) m=vec[i];
  return m ;
}
*/

template <class T>
int Vector<T>::find (const T& x) const 
{
  for (int i=0; i<dim; i++)
    if (vec[i]==x) return i+1;
  return 0;

}

template <class T> void Vector<T>::error(const char* msg) const
{
  cerr << "Error in class Vector:\n" << msg << endl;
  exit(1);
}

template <class T>
Vector<T>::Vector(long n, const T* array) 
  //array must have at least n elements!
{
  if(n<0) n==0;
  dim=n;
  if (n>0) vec = new T[dim]; else vec=NULL;
  for( int i = 0; i < n; i++ ) vec[i] = *array++;
}

template <class T>
Vector<T>::Vector(const Vector<T>& v)
{
  vec = new T[dim=v.dim];
  T *v1 = vec, *v2 = v.vec;
  long n = dim;
  while(n--) *v1++ = *v2++;
}

template <class T>
Vector<T>::Vector(const Array<T>& A) 
{
  vec = new T[dim =A.length()];
  ArrayLooper<T> loop(A);
  T* v = vec;
  while( loop.OK() ) *v++ = loop.next();
}

template <class T>
Vector<T>& Vector<T>::operator=(const Vector<T>& v)   // assignment
{
  if( this == &v ) return *this;
  if( dim != v.dim )  { // no point in deleting if new is same size
    delete [] vec;
    vec = new T [dim = v.dim]; 
  }
  T *v1 = vec, *v2 = v.vec; long n = dim; 
  while(n--) *v1++ = *v2++;
  return *this;
}

template <class T>
Vector<T>& Vector<T>::operator+=(const Vector<T>& q2)
{
  T* vi = vec, *wi = q2.vec; long i=dim;
  if( dim == q2.dim ) { while(i--)(*vi++) += (*wi++); }
  else error( "Incompatible Vectors in Vector::operator+=\n" );
  return *this;
}

template <class T>
Vector<T>& Vector<T>::operator-=(const Vector<T>& q2)
{
  T* vi = vec, *wi = q2.vec; long i=dim;
  if( dim == q2.dim ) { while(i--)(*vi++) -= (*wi++); }
  else error( "Incompatible Vectors in Vector::operator-=\n" );
  return *this;
}

template <class T>
Vector<T>& Vector<T>::operator*=(const T& scalar)
{
  T* vi = vec; long i = dim;
  while (i--) (*vi++) *= scalar;
  return *this;
}

template <class T>
Vector<T>& Vector<T>::operator/=(const T& scalar)
{
  T* vi = vec; long i = dim;
  while (i--) (*vi++) /= scalar;
  return *this;
}

template <class T> void Vector<T>::init(long d) 
{
  if(d<0) d=0;
  if( d != dim ) {
    delete [] vec;
    vec = new T [dim=d];
  }
  T * v = vec;
  for( long j = 0; j < d; j++ ) *v++ = T(0);
}

template <class T> void Vector<T>::resize(long d) 
{
  if(d<0) d=0;
  if( d != dim ) {
    T* old = vec ;
    vec = new T [d];
    for (long j = 0; j < d; j++)
      if (j < dim)
	vec[j] = old[j] ;
      else
	vec[j] = T(0) ;
    delete [] old ;
    dim = d ;
  }
}

template <class T>
Vector<T> Vector<T>::slice(long first, long last) const  // returns subVector
{
  if (last == -1) {last = first; first = 1;}
  long n = last-first+1;
  Vector<T> ans(n);
  T *veci=vec+(first-1), *ansi=ans.vec; long i=n;
  while (i--) *ansi++ = *veci++;
  return ans;
}

template <class T>
Vector<T> Vector<T>::operator[] ( const Array<int>& index ) const  
  // returns v[index[j]]
{
  long i = index.length(); 
  Vector<T> w(i);
  T* wi = w.vec;
  for( long j = 0; j < i; j++ ) (*wi++) = vec[index[j]-1];
  return w;
}

template <class T>
void Vector<T>::set(long i, const T& x)
{
  if ((i>0) && (i<=dim)) vec[i-1] = x;
  else error( "bad subscript in Vector::set\n" );
}

template <class T>
void Vector<T>::add(long i, const T& x)
{
  if ((i>0) && (i<=dim)) vec[i-1] += x;
  else error( "bad subscript in Vector::add\n" );
}

template <class T>
bool Vector<T>::isMember(const T& X) const
{
  bool ans = false; 
  T* vi = vec;
  for( int i = 0; i < dim && !ans; i++ ) ans = ( X == (*vi++));
  return ans;
}

template <class T>
bool Vector<T>::isTrivial() const
{
  bool ans = true;
  T* vi = vec;
  for( int i = 0; i < dim && ans; i++ ) ans = ((*vi++) == T(0));
  return ans;
}

template <class T>
bool Vector<T>::iszero() const
{
  T* vi=vec;
  for(int i=0; i<dim; i++)
    if((*vi++) != T(0))
      return false;
  return true;
}


template <class T>
bool Vector<T>::isEqual( const Vector<T>& W) const
{
  if(dim != W.dim) return false;
  for( int i = 0; i < dim; i++ )
    if(vec[i]!=W.vec[i]) return false;
  return true;
}

// Function `invert' inverts the order of entries in vector
//
template <class T> void Vector<T>::invert()
{
  T* tmp = new T [dim];
  for( int i = 0; i < dim; i++ ) *tmp++ = vec[i];
  tmp--;
  for( int i = 0; i < dim; i++ ) vec[i] = *tmp--;
  delete [] tmp;
}

// Definitions of non-member, friend operators and functions

template <class T>
T operator*(const Vector<T>& v, const Vector<T>& w)
{
  long Dim = v.dim;
  T dot = T(0);
  T* vi = v.vec, *wi = w.vec;
  if( Dim == w.dim ) 
    while( Dim-- ) dot += (*vi++)*(*wi++);
  else v.error( "Unequal dimensions in dot product\n" );
  return dot;
}

template <class T>
ostream& operator<<(ostream& s, const Vector<T>& v)
{
  long i = v.dim; 
  T* vi=v.vec;
  s << "[";
  while (i--) { s << (*vi++); if(i) s<<","; }
  s << "]";
  return s;
}

template <class T>
istream& operator>>(istream& s, Vector<T>& v)
{
  v.resize(get_int());
  for(int i=1; i<=v.size();i++)
    s>>v[i];
  return s;
}



bool do_polprint(const Vector<charmod>& p, ostream& os, const char* variable, bool reverse)
{return false;}
bool do_polprint(const Vector<bigcomplex>& p, ostream& os, const char* variable, bool reverse)
{return false;}
bool do_polprint(const Vector<intmod>& p, ostream& os, const char* variable, bool reverse)
{return false;}
bool do_polprint(const Vector<algnum>& p, ostream& os, const char* variable, bool reverse)
{return false;}
bool do_polprint(const Vector<ffield_element>& p, ostream& os, const char* variable, bool reverse)
{return false;}
bool do_polprint(const Vector<int>& p, ostream& os, const char* variable, bool reverse)
{return false;}
bool do_polprint(const Vector<bigint>& p, ostream& os, const char* variable, bool reverse)
{return false;}
bool do_polprint(const Vector<polbigrat>& p, ostream& os, const char* variable, bool reverse)
{return false;}
bool do_polprint(const Vector<bigrational>& p, ostream& os, const char* variable, bool reverse)
{
  int d = p.poldeg() ;
  if (d == -1) {
    os << "0";
    return true;
  }
  int i;
  bool first=true;
  if (reverse) i=0; else i=d;
  while((reverse && i<=d) || (!reverse && i>=0)) {
    if(p[i+1] != 0) {
      if(i==0) {
	if (!first && p[1]>0) os << "+"; 
	os << p[1];
      } else {
	if (p[i+1] < 0 && first) 
	  os << "-" ;
	if (!first) {
	  if (p[i+1] > 0)
	    os << "+"; 
	  else
	    os << "-"; 
	}
	if (abs(p[i+1]) != 1)  {
	  if(p[i+1]>0)
	    os << p[i+1] ;
	  else 
	    os << -p[i+1];
	  os << "*";
	}
	os<<variable; if (i>1) os << "^" << i;
      }
      first=false;
    }
    if (reverse) i++; else i--;
  }
  return true;
}

template<class T>
void
Vector<T>::polprint(ostream& os, const char* variable, bool reverse, bool tex) const
{
  const Vector<T>& p=*this;
  if (!do_polprint(p,os,variable,reverse)) {
    int d = poldeg() ;
    if (d == -1) {
      os << "0";
      return;
    }
    int i;
    bool first=true;
    if (reverse) i=0; else i=d;
    while((reverse && i<=d) || (!reverse && i>=0)) {
      if(p[i+1] != T(0)) {
	if (!first) os << "+"; 
	if(i==0) 
	  os << p[1];
	else {
	  if (p[i+1] != T(1)) {
	    os << p[i+1] ;
	    if(!tex)  os << "*";
	  }
	  os << variable;
	  if (i>1) {
	      os << "^";
	      if (tex && i>=10) os<<"{";
	      os << i;
	      if (tex && i>=10) os<<"}";
	  }
	}
	first=false;
      }
      if (reverse) i++; else i--;
    }
  }
  if (reverse)
      if(tex)
	  os << "+\\cdots";
      else
	  os << "+O(q^"<<poldeg()+1<<")";
}

template <class T>
T Vector<T>::norm(const Vector<T>& f) const 
{
  die("Vector norm not yet implimented.");
}


template <class T>
T Vector<T>::disc() const 
{
  die("Vector disc not yet implimented.");
}

template<class T>
void Vector<T>::Save(ofstream& of) const
{
  Matrix<T> m(*this);
  m.Save(of);                // wasteful in some ways, but simple.
}

template<class T>
void Vector<T>::Load(ifstream& in) 
{
  Matrix<T> m;
  m.Load(in);			// wasteful in some ways, but simple.
  if(m.nrows()!=1) die("Vector::Load -- error loading from file.");
  *this = m.row(1);
}


template<class F>
void 
IntVecList<F>::Save(ofstream& of) const
{
    of<<thesize<<endl<<thelen<<endl;
    for(int i=0;i<thesize;i++) {
	of << pZ[i] << endl;
	pv[i].Save(of); of<<endl;
    }
}

template<class F>
void 
IntVecList<F>::PrintAsPolyList(ostream& of, const char* variable="x") const
{
    int n=strlen(variable);
    char* var=new char[n+512];
    for(int i=0; i<n; i++) var[i]=variable[i];
    of << "[";
    for(int i=0;i<len();i++) {
	char* s=new char[pZ[i].bit_length()/3 + 11];
	bigint_to_string(pZ[i],s);
	strcpy(var+n, s);
	pv[i].polprint(of,var);
	delete s;
	if(i<len()-1) of<<", ";
    }
    of << "]";
    delete var;
}


template<class F>
void 
IntVecList<F>::Load(ifstream& in) 
{
    in>>thesize;
    init(thesize);
    in>>thelen;
    mesg("Loading IntVecList, size=",thelen);mesg("\n[");
    for(int i=0;i<thesize;i++) {
	if(i<thelen) {
	    if(i>0) mesg(" ");
	    mesg(i+1);
	}
	in >> pZ[i];
	pv[i].Load(in); 
    }
    mesgln("]");
}


#endif


