/* ----------------------------------------------------------------------
    HECKE: Program for computing with modular forms
    Copyright (C) 1999  William A. Stein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    William A. Stein
    2041 Francisco Street, Apt. 14
    Berkeley, CA  94709
    was@math.berkeley.edu

    This source file is based on code of John Cremona and Luiz. 
------------------------------------------------------------------------- */

// vector.h 

#ifndef __VECTOR_H_TEMPLATE
#define __VECTOR_H_TEMPLATE   

#include <iostream.h>
#include "array.h"
#include "system.h"
#include <LiDIA/bigint.h>

template <class T> class Matrix;
template <class T> class VecFac;

template <class T>
class Vector 
{
   friend class Matrix<T>;
public:
    Vector(long n=0);          // Vector of dimension n
    Vector(long n, const T* P);  // Vector of the n first numbers in array P
    Vector(const Array<T>&);
    Vector(const Vector<T>&);  
    ~Vector();        

    void random(int max);	// Set entries to be random integers
				// in the interval [0,max).
				// Uses the stdlib function random, thus
				// srandom can be used to set the seed.
    
    void Save(ofstream& of) const;
    void Load(ifstream& in) ;
    
    Vector& operator = ( const Vector<T>& );     
    T& operator [] ( long i ) const {     // starts at 1, not 0 as in Array 
      if ((i>0) && (i<=dim)) return vec[i-1];
      else error( "bad subscript in Vector::operator[]\n" );
    }
    Vector& operator += ( const Vector<T>& );
    Vector& operator -= ( const Vector<T>& );
    Vector& operator *= ( const T& );
    Vector& operator /= ( const T& );
    
    void invert();
    Vector slice( long, long=-1 ) const;  // returns subVector
    void init( long d = 0 );              // resizes to d, makes entries = 0
    void resize (long d) ;	        // resizes to d, copy old entries.
    void iota();                          // makes v[i] = i
    void zero();                          // makes all v[i] = 0
    bool iszero() const;                 // true iff all entries are zero.
    Vector operator [] ( const Array<int>& ) const;
    long size() const { return dim ; }
    int numnonzero() const ;	        // #{i : v[i]!=0}
    void reverse();		        // swaps v[i] & v[d-i] for i<=d/2.
    
    // reorder: v[i] <-- v[s[i]] for i=1,...,dim. 
    void reorder(const Vector<int>& s);
    
    
    void set( long i, const T& x );                  // sets v[i]=x
    void add( long i, const T& x );                  // v[i]+=x
    int find (const T& x) const ; // returns smallest i s.t.
				// v[i]==x, or 0 if no such i exists.
    
    const T* getVec() const { return vec; }
    const T* get_vec() const { return vec; }   
    bool isTrivial() const;              // is a zero vector ?
    bool isMember( const T& X ) const;   // is v[i] =X for some X?
    bool isEqual( const Vector<T>& ) const;
    
    // I've defined the standard arithmetic operations so
    // that polynomial style arithmetic can be done with
    // vectors.  This is useful for working in rings of
    // the form T[x]/(f(x)), etc.  Note that a[i] is the
    // coefficient of x^(i-1), NOT x^i.  
    // -----------------------------------------
    // Factor the monic polynomial 
    //   a[1]+a[2]x+...+a[n]x^(n-1)
    // defined by *this = [a[1],...a[n]].
    // for most datatype, this does nothing -- producing the empty factorization --
    // which symbolizes failure. 
    void polfactor(VecFac<T>& fac) const;
    bool pol_is_irreducible() const;
    // return 1/Mod(*this, polymod).
    Vector<T> polinv_polymod(const Vector<T>& polymod) const;
    friend Vector<T> polmul_polymod(const Vector<T>& f, const Vector<T>& g,
				    const Vector<T>& polymod);
    void polderiv(Vector<T>& deriv) const;

    // returns (*this)(X+k), where *this is viewed as a polynomial.
    Vector<T> pol_translate(const T& k) const;
    T& polcoeff(int i);
    T polgetcoeff(int i) const;
    T pol_l() const;		// leading coefficient.
    T polsubst(const T& x) const;
    bool pol_issqrfree() const ;
    Vector<T> polderiv() const;
    // pol_xgcd assumes (a,b)=1, and then returns x, y so that
    //    x*a + y*b = 1.
    friend void pol_xgcd(Vector<T>& x, Vector<T>& y, Vector<T>& gcd,
			 const Vector<T>& a, const Vector<T>& b);

    friend Vector<T> polgcd(Vector<T> f, Vector<T> g);
    friend Vector<T> poladd(const Vector<T>& f, const Vector<T>& g);
    friend Vector<T> polsub(const Vector<T>& f, const Vector<T>& g);
    friend Vector<T> polmul(const Vector<T>& f, const Vector<T>& g);
    // poldiv computes f/g and discard the remainder 
    friend Vector<T> poldiv(const Vector<T>& f, const Vector<T>& g);
    // polpower sets g = (*this)^n.
    void polpower(Vector<T>& g,
		  int n);
    // poldiv gives the remainder of f modulo g.
    friend Vector<T> polmod(const Vector<T>& f, const Vector<T>& g);
    friend T polcontent(const Vector<T>& f);	// gcd of coefficients
    
    int pol_ord(const Vector<T>& p) const;  // returns ord_p(*this). 
    void pol_makemonic();	// divides (*this) through by leading coefficient to make monic.
    void polconst(const T& a) { init(1); (*this)[1]=a; }
    int poldeg() const ;
    void polprint(ostream& os=cout, const char* variable="x",
		  bool reverse=false, bool tex=false) const;
    // norm: compute the norm of v=*this, considered as an element of
    // the ring Q[x]/(f(x)), expressed on a power basis.  The norm
    // is the determinent of left multiplication.
    T norm(const Vector<T>& f) const ;
    // disc: compute the discriminant of *this, viewed as a polynomial.
    T disc() const ;  

    // Pseudo-Division
    // Algorithm 3.1.2 of Cohen: Let K be the polynomial ring Q[X], 
    // A and B be two polynomials in Q[X] with B\neq 0, and set
    // m <-- deg(A), n<--deg(B), d<--l(B)=leading coeff(B).  Assume
    // that m>=n.  This algorithm finds Q and R such that
    // d^{m-n+1}A = BQ+R and deg(R)<deg(B).
    friend void pol_pseudodiv(Vector<T>& Q, Vector<T>& R,
			   const Vector<T>& A, const Vector<T>& B);
    friend T pol_resultant(Vector<T> A, Vector<T> B);
    
    
    friend long dim(const Vector<T>&);             // the dimension
    friend T operator*(const Vector<T>&, const Vector<T>&);   // dot product
    
    friend Vector<T> operator*(const Matrix<T>& m, const Vector<T>& v);
    friend Vector<T> operator*(const Vector<T>& v, const Matrix<T>& m);
    
    friend bool operator == ( const Vector<T>&, const Vector<T>& );
    friend bool operator != ( const Vector<T>&, const Vector<T>& );
    
    friend Vector operator + ( const Vector<T>& ); // unary
    friend Vector operator - ( const Vector<T>& ); // unary
    
    friend void swapvec(Vector<T>&, Vector<T>&);
    
    friend ostream& operator<< (ostream&s, const Vector<T>&);
    friend istream& operator>> (istream&s, Vector<T>&);
    
    void    error(const char* msg) const;     // error handling
    
private:
    long dim;            // the dimension 
    T * vec;
};



template<class F>
class IntVecList {
public:
    IntVecList(int _size=1) { allocate(_size); }
    void init(int newsize) { delete[] pZ; delete[] pv; allocate(newsize); }
    IntVecList(const IntVecList<F>& list) {
	allocate(list.thesize);
	for(int i=0; i<size(); i++) { Z(i)=list.pZ[i]; v(i)=list.pv[i]; }
    }
    IntVecList& operator=(const IntVecList<F>& list) {
	init(list.thesize);
	for(int i=0; i<size(); i++) { Z(i)=list.pZ[i]; v(i)=list.pv[i]; }
	return *this;
    }
    ~IntVecList() { delete[] pZ; delete[] pv; }
    void resize(int newsize) {
        bigint* oldpZ=pZ;
	Vector<F>* oldpv=pv;
	pZ=NULL; pv=NULL;
	int oldsize=size();
	init(newsize);
	for(int i=0; i<oldsize; i++)
	    { pZ[i]=oldpZ[i]; pv[i]=oldpv[i]; }
	delete[] oldpZ;
	delete[] oldpv;
    }

    bigint& Z(int i) { validate(i); return pZ[i]; }
    Vector<F>& v(int i) { validate(i); return pv[i]; }
    const bigint& cZ(int i) const { constvalidate(i); return pZ[i]; }
    const Vector<F>& cv(int i) const { constvalidate(i); return pv[i]; }
    int size() const { return thesize; }
    int len() const { return thelen; }	 // largest array element set. 
    void Save(ofstream& of) const;
    void Load(ifstream& in);
    void PrintAsPolyList(ostream& os, const char* variable="x") const;
private:
    int thesize, thelen;
    bigint* pZ;
    Vector<F>* pv;
    void validate(int i) {
	constvalidate(i);
	if(i>=thelen) thelen=i+1;
    }
    void constvalidate(int i) const {
	if(!(0<=i && i<size())) {
	    cerr<<"IntVecList: assert(0<="<<i<<"<"
		<<size()<<") failed"<<endl; die(); 
	}
    }
    void allocate(int _size) {
	thesize=_size>=1?_size:1;
	thelen=0;
	pZ=new bigint[thesize];
	pv=new Vector<F>[thesize];
    }
};









// inline function definitions

template <class T> 
Vector<T>::Vector(long n)
{
    if(n<0) n=0;
    dim=n;
    if (n>0) {
	vec=new T[n];
//	for(int i=0; i<n; i++)
//	    vec[i]=T(0);
    } else
	vec=NULL;
}

template <class T> inline
long dim(const Vector<T>& v) { return v.dim; }


template <class T> inline
void Vector<T>::iota()
{
    T* veci = vec; 
    for(long i=0; i<dim;) *veci++ = T(++i);
}

template <class T> inline
void Vector<T>::zero()
{
    T* veci = vec; 
    for( long i = 0; i < dim; i++) *veci++ = T(0);
}

template <class T> inline 
bool operator!=(const Vector<T>& v, const Vector<T>& w) 
{ 
    return !v.isEqual(w);
}

template <class T> inline 
Vector<T> operator+(const Vector<T>& v) { return v;}

template <class T> inline 
Vector<T>::~Vector() 
{
    delete [] vec;
}

template <class T> inline 
Vector<T> operator-(const Vector<T>& v) 
{  
    //    for( int i = 0; i < d; i++ ) vec[i] *= -1;
    Vector<T> ans(v); 
    ans *= T(-1); 
    return ans; 
}

template <class T> inline 
Vector<T> operator+(const Vector<T>& v1, const Vector<T>& v2)
{ 
    Vector<T> ans(v1); 
    ans+=v2; 
    return ans; 
}

template <class T> inline 
Vector<T> operator-(const Vector<T>& v1, const Vector<T>& v2)
{ 
    Vector<T> ans(v1); 
    ans-=v2; 
    return ans; 
}

template <class T> inline 
Vector<T> operator*(const T& scalar, const Vector<T>& v)
{ 
    Vector<T> ans(v); 
    ans *= scalar; 
    return ans; 
}

/*
template <class T> inline 
Vector<T> operator*(int scalar, const Vector<T>& v)
{ 
    Vector<T> ans(v); 
    ans *= T(scalar); 
    return ans; 
}
*/

template <class T> inline 
Vector<T> operator/(const Vector<T>& v, const T& scalar)
{ 
    Vector<T> ans(v); 
    ans/=scalar; 
    return ans; 
}

template <class T>
Vector<T> operator*(const Matrix<T>& m, const Vector<T>& v)
{
    long Dim = v.dim;
    if (Dim != m.nco)
	v.error("Incompatible dimensions in matrix * vector\n");
    Vector<T> w(m.nro);
    for (int i=0; i <m.nro; i++) {
	w.vec[i] = 0;
	T* vi=v.vec;
	T* mij=m.mat+i*m.nco;
	for (int j=0; j<Dim; j++)
	    w.vec[i]+=(*vi++)*(*mij++);
    }
    return w;
}

template <class T>
Vector<T> operator*(const Vector<T>& v, const Matrix<T>& m)
{
    long d = v.dim;
    if (d != m.nrows())
	v.error("Incompatible dimensions in vector * matrix\n");
    Vector<T> w(m.ncols());
    for (int i=1; i <=m.ncols(); i++) {
	w[i] = 0;
	for(int j=1; j<=d; j++)
	    w[i] += v[j]*m(j,i);
    }
    return w;
}


template <class T> inline 
bool operator==(const Vector<T>& V, const Vector<T>& W)
{
    return V.isEqual(W);
}

template <class T>
void swapvec(Vector<T>& v, Vector<T>& w)
{
    if( v.dim == w.dim ) { T *temp=v.vec; v.vec=w.vec; w.vec=temp;}
    else v.error( "Attempt to swap Vectors of different lengths!\n" );
}

/* Notes:
 * Vector V(n) just allocates space for n entries. It doesn't fill in      *
 * with zeros. A Vector<T> V with dim n and zero entries can be created by *
 *  ' Vector<T> V(n); V.zero();'                                           *
 * 
 * If V is of type Vector<T>, then when multiplication or division by      *
 * is performed, say a*V, a integer, then a is converted to type T before  *
 * the multiplication. Sometimes it is more efficient to provide a direct  *
 * multiplication and division by builtin types.                           *
 */


///////////////////////////// Factor structure class///////////////////////
template <class T>
class VecFac {
public:
    VecFac() { pBase=NULL; pExp=NULL; }
    resize(int n) {
	size=n;
	delete[] pBase; delete[] pExp; 
	if(n>0) {
	    pBase=new Vector<T>[n];
	    pExp=new int[n];      
	}
    }
    ~VecFac() { resize(0); }
    Vector<T>& base(int i) {
	if(1<=i && i<=size) return pBase[i-1]; 
	else {
	    cerr<<"VecFac base access out of bounds"<<endl
		<<"i="<<i<<", but should have 1<=i<="<<size<<endl;
	    die(); 
	}
    }
    int& exp(int i) {
	if(1<=i && i<=size) return pExp[i-1]; 
	else { 
	    cerr<<"VecFac exponent access out of bounds"<<endl
		<<"i="<<i<<", but should have 1<=i<="<<size<<endl;
	    die(); 
	}
    }
    int numfactors() const { return size; }
private:
    Vector<T>* pBase;
    int* pExp;
    int size;
};


////////////// POLYMOD VECTORS ///////////// 

template <class T>
int Vector<T>::poldeg() const 
{ 
    int i=size();
    while(i>0&&(*this)[i]==T(0)) i--;
    return i-1;
}

template <class T>
void Vector<T>::pol_makemonic()
{
    *this /= pol_l();
}

template<class T>
void Vector<T>::polpower(Vector<T>& b, int n)
{
    b.init(1); b[1]=T(1); 
    if (n==0)
	return ;
    if (n<0) 
	die("polpower: negative exponent not defined!");
    
    Vector<T> t=(*this);
    while (n>=1) {
	if (n%2) 
	    b = polmul(b,t);
	t = polmul(t,t);
	n/=2;
    }
}


template<class T>
Vector<T> poladd(const Vector<T>& f, const Vector<T>& g)
{
    Vector<T> a;
    a.init(intmax(f.size(),g.size()));
    int m=intmin(f.size(),g.size());
    for(int i=1;i<=m;i++)
	a[i]=f[i]+g[i];
    int s=a.size();
    if(f.size()>g.size())
	for(int i=m+1;i<=s;i++)
	    a[i]=f[i];
    else
	for(int i=m+1;i<=s;i++)
	    a[i]=g[i];
    return a;
}

template<class T>
Vector<T> polsub(const Vector<T>& f, const Vector<T>& g)
{
    return poladd(f,-g);
}

template<class T>
Vector<T> poldiv(const Vector<T>& f, const Vector<T>& g)
{
    if(f.poldeg()<0) {
	Vector<T> z; z.init(0); return z;
    }
    if(g.poldeg()<0) { cerr << "poldiv: divide by zero."<<endl; die(); }
    if(g.poldeg()==0) return f/g[1];
    int m=f.poldeg(), n=g.poldeg(); 
    Vector<T> ff=f;
    Vector<T> q; q.init(m-n+1);
    for(int k=q.size(); k>=1; k--) {
	q[k] = ff[n+k]/g[n+1];
	for(int j=n+k; j>=k; j--)
	    ff[j] -= q[k]*g[j-k+1];
    }
    return q;
}

 
template<class T>
Vector<T> polmul_polymod(const Vector<T>& f, const Vector<T>& g,
			 const Vector<T>& polymod)
{ 
    Vector<T> fg=polmul(f,g);
    Vector<T> a = polmod(fg,polymod);
    a.resize(polymod.size()-1);
    return a;
}


template<class T>
void Vector<T>::polderiv(Vector<T>& deriv) const
{
    deriv.init(poldeg());
    for(int i=1;i<=deriv.size();i++)
	deriv[i]=(*this)[i+1]*T(i);
}


template<class T>
Vector<T> Vector<T>::polinv_polymod(const Vector<T>& polymod) const
{
    // keep in mind: (*this)*oneover + polymod*b = gcd.
    Vector<T> b, gcd, oneover;
    pol_xgcd(oneover, b, gcd, *this, polymod);
    if(gcd[1] == T(0)) {
	cerr << "Vector::oneover_polymod, not invertible."<<endl; die();
    }
    for(int i=2;i<=gcd.size();i++) 
	if(gcd[i] != T(0)) {
	    cerr << "Vector::oneover_polymod, not invertible."<<endl; die(); 
	}
    oneover /= gcd[1];
    oneover.resize(polymod.size()-1);
    return oneover;
}

template<class T>
T& Vector<T>::polcoeff(int i)
{
    return (*this)[i+1];
}

template<class T>
T Vector<T>::polgetcoeff(int i) const
{
    return (*this)[i+1];
}

template<class T>
T Vector<T>::pol_l() const
{
    if(poldeg()<0) return T(0);
    return polgetcoeff(poldeg());
}

template<class T>
bool Vector<T>::pol_issqrfree() const 
{
    Vector<T> deriv;
    polderiv(deriv);
    return polgcd(*this, deriv).poldeg()==0 ;
}

template<class T>
Vector<T> Vector<T>::polderiv() const
{
    Vector<T> v;
    polderiv(v);
    return v;
}

template<class T>
int Vector<T>::pol_ord(const Vector<T>& p) const
{
    if(poldeg()<0) die("pol_ord called for zero polynomial!");
    Vector<T> a=*this;
    int i=0;
    while(polmod(a,p).poldeg()<0) {
	a=poldiv(a,p);
	i++;
    }
    return i;
}





#endif


